/*-------------------------------------------------------------------------
 *
 * miscam.c--
 *    Access methods for relations that require special naming conventions.
 *    So far:
 *	- Archive relations.
 *	- Mariposa copy fragments.
 *	- Mariposa side files.
 *	- Oid hash indices for Mariposa copies.
 *
 * Copyright (c) 1995, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/access/common/miscam.c,v 1.5 1996/02/18 02:57:00 aoki Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <stdio.h>

#include "postgres.h"

#include "access/heapam.h"
#include "catalog/heap.h"
#include "catalog/pg_class.h"
#include "tcop/tcopprot.h"
#include "utils/elog.h"
#include "utils/rel.h"

#include "access/miscam.h"

/* ----------------
 *  Currently, archive relation names are formed by appending A to
 *  the oid (in hex format) of the fragment being archived.
 * ----------------
 */

/* ----------------------------------------------------------------
 *	IsArchiveRelation
 *
 *	Judges whether or not rname represents an archive relation.
 *	Used in vacuum.c.
 *
 * ----------------------------------------------------------------
 */
bool
IsArchiveRelation(Name rname)
{
    int i, len;

    if ((len = strlen(rname->data)) > 9)
	return (false);

    for (i = 0; i < (len - 1); i++)
	if (!(rname->data[i] >= '0' && rname->data[i] <= '9') &&
	    !(rname->data[i] >= 'a' && rname->data[i] <= 'f'))
	{
	    return (false);
	}

    return (bool)(rname->data[i] == 'A');
}

/* ----------------------------------------------------------------
 *	CreateArchiveRelationFromScratch
 *
 *	Creates an archive relation for a relation that does not exist yet.
 *	Default archive location is magnetic disk.
 *
 * ----------------------------------------------------------------
 */
void
CreateArchiveRelationFromScratch(Oid relid, int natts, TupleDesc tupdesc)
{	
    NameData		archiveName;
    int			archloc = 0; /* Archive location, default magdisk. */
    Oid			id;

    sprintf(archiveName.data, "%xA", relid);
	
    id = heap_create(archiveName.data,
		     'n',		/* archive isn't archived */
		     natts,
		     archloc,
		     tupdesc);
    DefineRelationFirstFragment(id);
}

/* ----------------------------------------------------------------
 *	CreateArchiveRelation
 *
 *	Creates an archive relation for a relation.
 *	Default archive location is magnetic disk.
 *
 * ----------------------------------------------------------------
 */
void
CreateArchiveRelation(Relation rdesc)
{	
    NameData		archiveName;
    int			archloc = 0; /* Archive location, default magdisk. */
    Oid			id;

    Assert(RelationIsValid(rdesc));

    sprintf(archiveName.data, "%xA", rdesc->rd_id);
	
    id = heap_create(archiveName.data,
		     'n',		/* archive isn't archived */
		     rdesc->rd_rel->relnatts,
		     archloc,
		     rdesc->rd_att);
    DefineRelationFirstFragment(id);
}

/* ----------------------------------------------------------------
 *	DestroyArchiveRelation
 *
 *	Destroys a relation's archive relation.
 *
 * ----------------------------------------------------------------
 */
void
DestroyArchiveRelation(Oid relid)
{	
    NameData		archiveName;

    sprintf(archiveName.data, "%xA", relid);

    heap_destroy(archiveName.data);
}

/* ----------------------------------------------------------------
 *	OpenArchiveRelation
 *
 *	Opens a relation's archive relation.
 *
 * ----------------------------------------------------------------
 */
Relation
OpenArchiveRelation(Oid relid)
{	
    NameData		archiveName;

    sprintf(archiveName.data, "%xA", relid);

    return heap_openr((Name) &archiveName);
}

/* ----------------------------------------------------------------
 *	GenerateCopyName
 *
 *	Given pg_class tuple for a fragment, and new relid, generate
 *	name for its copy. For now, just appends oid to first 8
 *	characters of original fragment name.
 *
 * ----------------------------------------------------------------
 */
void
GenerateCopyName(Name name, Form_pg_class classtup, Oid relid)
{
    sprintf(name->data, "%.8s%x", classtup->relname.data, relid);
}

/* ----------------------------------------------------------------
 *	GenerateSideFileName
 *
 *	Given reldesc for fragment, generate name for its side-file.
 *	For now, just appends 'S' to fragment name, for debugging
 *	purposes.
 *
 *	@@@ In future, should be (fragment oid (in hex)) + S.
 *
 * ----------------------------------------------------------------
 */
void
GenerateSideFileName(Name name, Relation rdesc)
{
    sprintf(name->data, "%sS", rdesc->rd_rel->relname.data);
}

/* ----------------------------------------------------------------
 *	DestroySideFile
 *
 *	Destroys a relation's side file
 *
 * ----------------------------------------------------------------
 */
void
DestroySideFile(Relation rdesc)
{	
    NameData		sideFileName;

    GenerateSideFileName((Name) &sideFileName, rdesc);

    heap_destroy(sideFileName.data);
}

/* ----------------
 *  Currently, names for hash indices on oid are formed by
 *  appending the oid of the fragment being indexed to string 'hashoid'.
 * ----------------
 */

/* ----------------------------------------------------------------
 *	DefineRelationOidHashIndex
 *
 *	Creates a hash index on oid for a relation.
 *
 * ----------------------------------------------------------------
 */
void
DefineRelationOidHashIndex(Relation rdesc)
{	
    char		query[1000];

    Assert(RelationIsValid(rdesc));

    sprintf(query,
	    "CREATE INDEX hashoid%x ON %s USING hash (oid oid_ops)",
	    rdesc->rd_id,
	    rdesc->rd_rel->relname.data);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
}

/* ----------------------------------------------------------------
 *	OpenRelationOidHashIndex
 *
 *	Opens relation's oid hash index.
 *
 * ----------------------------------------------------------------
 */
Relation
OpenRelationOidHashIndex(Oid relid)
{	
    NameData		indexName;

    sprintf(indexName.data, "hashoid%x", relid);

    return (Relation) index_openr((Name) &indexName);
}
