/*-------------------------------------------------------------------------
 *
 * tidmap.c--
 *    TID translation support code.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Id: tidmap.c,v 1.1 1995/09/05 09:27:06 aoki Exp $
 *
 *
 * NOTES
 *
 * conventions:
 * - external interface routines take in and return "void *" for their
 *   opaque handles.  this is for modularity reasons (i prefer not to
 *   export these structures without good reason).
 *-------------------------------------------------------------------------
 */

#include <string.h>

#include "c.h"

#include "storage/block.h"
#include "storage/off.h"
#include "storage/itemptr.h"

#include "utils/elog.h"
#include "utils/hsearch.h"
#include "utils/palloc.h"
#include "lib/hasht.h"

#include "miscadmin.h"

/*-------------------------------------------------------------------------
 * the code below implements a variation on the TID-translation
 * technique from: W. Sun, W. Meng, C. Yu and W. Kim, "An Efficient
 * Way to Reestablish B+ Trees in a Distributed Environment,"
 * Information Sciences 77 (3-4), March 1994, pp. 227-251.  the main
 * difference is that we do not bother with the MAP array; the block
 * numbers assigned on the sender are assumed to be acceptable on the
 * receiver.  (the cost of this mapping should be negligible.)
 *
 * it's not clear from Sun et al. why they choose to use a hash table
 * instead of an array for the sender translation mechanism.  it's not
 * because the table is sparse; the index will contain pointers to
 * every sender page.  it's not because we can't predict the size;
 * even though you do need to keep some information per *receiver*
 * page, there are exactly as many table entries as there are *sender*
 * pages.
 *-------------------------------------------------------------------------
 */

/*
 * we fake byte offsets here by using the usual slot number
 * (OffsetNumber).  the logic is the same, the required amount of
 * memory is the same, the only difference is the content.
 */
typedef OffsetNumber	ByteOffset;

/*
 * hash table search tag.  because of the way the dynahash code
 * (mainly HashTableWalk) works, the size of the tag must be such that
 * the PTidHashEntry payload is LONGALIGN(sizeof(PTidHashTag)) bytes
 * from the start of the PTidHashEntry structure.
 */
typedef struct {
    BlockNumber	pht_sblkno;	/* sender blkno */
    int32	pht_dummy;
} PTidHashTag;

/*
 * translation information.  the size of this depends on the number of
 * offsets.  the hash table code stores byte offsets, requiring one
 * 'phd_off' element per offset.
 */
typedef struct {
    BlockNumber	phd_rblkno;	/* first receiver blkno */
    uint16	phd_noff;	/* size of phd_slast */
    ByteOffset	phd_slast[1];	/* array of last offsets on sender */
    /* VARIABLE-LENGTH STRUCTURE */
} PTidHashData;

/*
 * we store the payload in the hash table as a pointer because this is
 * (more or less) the only way we can store variable-length data using
 * dynahash.  unfortunately, this means that we have to free those
 * pointers by hand when we destroy the hash table or otherwise remove
 * hash table entries.
 */
typedef struct {
    PTidHashTag		phe_tag;
    PTidHashData	*phe_data;
} PTidHashEntry;

/*
 * destroy a hash table entry.
 *
 * (this is a helper routine for HashTableWalk.)
 */
void
PTidHashDataDestroy(PTidHashData **datap)
{
    pfree((void *) *datap);
    *datap = (PTidHashData *) NULL;
}

/*
 * allocate a PTid hash table.
 *
 * we need as many hash table entries as we have disk blocks on the
 * sender.  fortunately, it's not very hard to find out how many
 * blocks a given table or index has (e.g., we can get an approximate
 * number by consulting the table's reldesc, or we can get an exact
 * number by calling smgrnblocks).
 */
void *
PTidHashCreate(int nblocks)
{
    HTAB *htab;
    HASHCTL info;
    int nelem = nblocks * 1.5;	/* fudge factor */

    (void) memset((char *) &info, 0, sizeof(HASHCTL));
    info.keysize = sizeof(PTidHashTag);
    info.datasize = sizeof(PTidHashData *);
    if ((htab = hash_create(nelem, &info, HASH_ELEM)) == (HTAB *) NULL) {
	elog(WARN, "PTidHashInit: out of memory");
    }
    return((void *) htab);
}

/*
 * destroy a PTid hash table.
 */
void
PTidHashDestroy(void *handle)
{
    HTAB *htab = (HTAB *) handle;

    HashTableWalk(htab, (HashtFunc) PTidHashDataDestroy, 0);
    hash_destroy(htab);
}


/*
 * return a TID that points to the correct page on the receiver.  we
 * do need the sender's offset to figure that out.  however, we can't
 * figure out the right offset without searching the receiver heap
 * page.
 */
void
PTidHashMap(void *handle, ItemPointer sitem, ItemPointer ritem)
{
    HTAB *htab = (HTAB *) handle;
    BlockNumber sblkno = ItemPointerGetBlockNumber(sitem);
    ByteOffset soff = (ByteOffset) ItemPointerGetOffsetNumber(sitem);
    bool found;
    PTidHashTag tag;
    PTidHashEntry *entry;
    PTidHashData *p;
    int i;

    (void) memset((char *) &tag, 0, sizeof(PTidHashTag));
    tag.pht_sblkno = sblkno;
    entry =
	(PTidHashEntry *) hash_search(htab, (char *) &tag, HASH_FIND, &found);
    if (entry == (PTidHashEntry *) NULL || !found) {
	ItemPointerSetInvalid(ritem);
	return;
    }
    p = entry->phe_data;
    for (i = 0; i < p->phd_noff && soff > p->phd_slast[i]; ++i)
	;
    if (i < p->phd_noff) {
	ItemPointerSet(ritem, p->phd_rblkno + i, InvalidOffsetNumber);
    } else {
	ItemPointerSetInvalid(ritem);
    }
}

/*
 * maintain one PTid hash table entry per sender blkno.  however, each
 * sender page entry can have various byte offsets corresponding to
 * the offset of the last item placed in rblkno,rblkno+1,rblkno+2,...
 * so we need to look and see if we must add an offset to an existing
 * entry before blindly trying to stick in a new one.
 *
 * XXX memory allocation is expensive enough that we should probably
 * not realloc one entry at a time; we should do something like grow
 * the offset array by doubling.  on the other hand, this has minimal
 * memory utilization.
 */
void
PTidHashInsert(void *handle, ItemPointer sitem, ItemPointer ritem)
{
    HTAB *htab = (HTAB *) handle;
    BlockNumber sblkno = ItemPointerGetBlockNumber(sitem);
    BlockNumber rblkno = ItemPointerGetBlockNumber(ritem);
    ByteOffset soffset = ItemPointerGetOffsetNumber(sitem);
    ByteOffset roffset = ItemPointerGetOffsetNumber(ritem);
    bool found;
    PTidHashTag tag;
    PTidHashEntry *entry;
    PTidHashData *data;

    (void) memset((char *) &tag, 0, sizeof(PTidHashTag));
    tag.pht_sblkno = sblkno;
    entry = (PTidHashEntry *)
	hash_search(htab, (char *) &tag, HASH_FIND, &found);
    if (entry == (PTidHashEntry *) NULL || !found) {
	data = (PTidHashData *) palloc(sizeof(PTidHashData));
	data->phd_rblkno = rblkno;
	data->phd_noff = 1;
	data->phd_slast[0] = soffset;

	entry = (PTidHashEntry *)
	    hash_search(htab, (char *) &tag, HASH_ENTER, &found);
	if (entry == (PTidHashEntry *) NULL || found) {
	    elog(WARN, "PTidHashInsert: duplicate blkno %d", sblkno);
	}
    } else {
	Assert(entry != (PTidHashEntry *) NULL);
	Assert(entry->phe_data != (PTidHashData *) NULL);
	data = (PTidHashData *)
	    repalloc(entry->phe_data,
		     sizeof(PTidHashData) +
		     (entry->phe_data->phd_noff * sizeof(uint16)));
	data->phd_slast[data->phd_noff] = soffset;
	++data->phd_noff;
    }
    /*
     * directly modify the entry in the table.  this obviously relies
     * on hash_search returning a pointer to the actual hash table
     * entry (as opposed to returning a copy).
     */
    entry->phe_data = data;
}

/*-------------------------------------------------------------------------
 *
 *-------------------------------------------------------------------------
 */

/*
 * 'pad_slast' is a variable-length array.  normally, we would either
 * just make it a pointer or make the whole PTidArray an array of
 * pointers to variable-length structures.  on an alpha (64-bit
 * pointers), if you're going to include a pointer, you may as well
 * make your structure size a multiple of 8 bytes because that's the
 * boundary to which it will align anyway.  but if we have that much
 * space, we may as well *use* that space for data until we *need* the
 * extra expansion space provided by the pointer indirection.  hence,
 * if we have < 5 offsets, they are stored in pad_slast.off; if we have
 * >= 5 offsets, pad_slast.offp is a pointer to an OffsetNumber array.
 */
typedef struct {
    BlockNumber		pad_rblkno;	/* first receiver blkno */
    OffsetNumber	pad_rfirst;	/* first receiver offset */
    uint16		pad_noff;	/* size of pad_slast */
    union {
	OffsetNumber off[4];
	OffsetNumber *offp;
    }			pad_slast;
} PTidArrayData;

typedef struct {
    BlockNumber		pa_nblocks;
    PTidArrayData	*pa_data;
} PTidArray;

/*
 *
 */
void *
PTidArrayCreate(int nblocks)
{
    PTidArrayData *data;
    PTidArray *array;

    data = (PTidArrayData *) palloc(nblocks * sizeof(PTidArrayData));
    memset((char *) data, 0, nblocks * sizeof(PTidArrayData));

    array = (PTidArray *) palloc(sizeof(PTidArray));
    array->pa_nblocks = nblocks;
    array->pa_data = data;

    return((void *) array);
}

/*
 *
 */
void
PTidArrayDestroy(void *handle)
{
    PTidArray *array = (PTidArray *) handle;
    int i;

    for (i = 0; i < array->pa_nblocks; ++i) {
	if (array->pa_data[i].pad_noff > 4) {
	    pfree((void *) array->pa_data[i].pad_slast.offp);
	}
    }
    pfree((void *) array->pa_data);
    pfree((void *) array);
}

/*
 *
 */
void
PTidArrayMap(void *handle, ItemPointer sitem, ItemPointer ritem)
{
    PTidArray *array = (PTidArray *) handle;
    BlockNumber sblkno = ItemPointerGetBlockNumber(sitem);
    OffsetNumber soff = ItemPointerGetOffsetNumber(sitem);
    PTidArrayData *p;
    OffsetNumber *offp;
    int i;

    p = &(array->pa_data[sblkno]);
    Assert(p->pad_noff > 0);
    offp = (p->pad_noff <= 4) ? p->pad_slast.off : p->pad_slast.offp;
    for (i = 0; i < p->pad_noff && soff > offp[i]; ++i)
	;
    if (i == 0) {
	ItemPointerSet(ritem, p->pad_rblkno, (p->pad_rfirst - 1) + soff);
    } else if (i < p->pad_noff) {
	OffsetNumber last = offp[i - 1];

	ItemPointerSet(ritem, p->pad_rblkno + i, soff - last);
    } else {
	ItemPointerSetInvalid(ritem);
    }
}  

/*
 *
 */
void
PTidArrayInsert(void *handle, ItemPointer slast, ItemPointer rfirst)
{
    PTidArray *array = (PTidArray *) handle;
    OffsetNumber sblkno = ItemPointerGetBlockNumber(slast);
    OffsetNumber soff = ItemPointerGetOffsetNumber(slast);
    PTidArrayData *p;

    Assert(sblkno < array->pa_nblocks);
    p = &(array->pa_data[sblkno]);

    if (p->pad_noff < 4) {
	if (p->pad_noff == 0) {
	    p->pad_rblkno = ItemPointerGetBlockNumber(rfirst);
	    p->pad_rfirst = ItemPointerGetOffsetNumber(rfirst);
	}
	p->pad_slast.off[p->pad_noff] = soff;
    } else {
	if (p->pad_noff == 4) {
	    OffsetNumber *offp = (OffsetNumber *)
		palloc(5 * sizeof(OffsetNumber));
	    
	    /* copy the existing OffsetNumbers into the new array */
	    (void) memcpy((char *) offp, (char *) &(p->pad_slast),
			  4 * sizeof(OffsetNumber));
	    p->pad_slast.offp = offp;
	} else {
	    /* realloc the array of OffsetNumbers to make room */
	    p->pad_slast.offp = (OffsetNumber *)
		repalloc(p->pad_slast.offp,
			 (p->pad_noff + 1) * sizeof(OffsetNumber));
	}
	p->pad_slast.offp[p->pad_noff] = soff;
    }
    ++p->pad_noff;
}

#define TIDTEST

#ifdef TIDTEST

#include "utils/memutils.h"
#include "storage/bufmgr.h"
#include "storage/item.h"
#include "storage/itemid.h"
#include "storage/page.h"
#include "storage/smgr.h"

#include "catalog/pg_attribute.h"

#include "access/heapam.h"
#include "access/genam.h"

bool PTidUseHash = false;

/*
 * for the array TID translation to work, we must preserve the number
 * and spacing of TIDs on the sender and receiver pages.  this means
 * we can't get rid of old ItemId holes, which in turn means that we
 * can't use the regular heap_* routines.
 */
void *
heap_tidmap(Name from_heap, Name to_heap, int blocksz)
{
    Relation srel;
    int nblocks;
    void *handle;
    TupleDesc tupdesc;
    Oid roid;
    Relation rrel;
    int i;
    BlockNumber sblkno;
    BlockNumber rblkno;
    Buffer sbuf;
    Buffer rbuf;
    Page spage;
    Page rpage;
    OffsetNumber maxoff;
    OffsetNumber soff;
    OffsetNumber roff;
    ItemId itemid;
    Item sitem;
    Item ritem;
    Size itemsz;
    ItemPointerData slast;
    ItemPointerData rfirst;

    FlushBufferPool(0);
    srel = heap_openr(from_heap);
    if (!RelationIsValid(srel)) {
	elog(WARN, "heap_tidmap: could not open relation '%s'", from_heap);
    }
    nblocks = smgrnblocks(srel->rd_rel->relsmgr, srel);
    if (PTidUseHash) {
	handle = PTidHashCreate(nblocks);
    } else {
	handle = PTidArrayCreate(nblocks);
    }

    tupdesc = (TupleDesc) palloc(sizeof(AttributeTupleForm) *
				 srel->rd_rel->relnatts);
    for (i = 0; i < srel->rd_rel->relnatts; ++i) {
	tupdesc[i] = (AttributeTupleForm) palloc(ATTRIBUTE_TUPLE_SIZE);
	(void) memcpy(tupdesc[i], srel->rd_att[i], ATTRIBUTE_TUPLE_SIZE);
    }
    roid = heap_create((char *) to_heap, srel->rd_rel->relarch,
		       srel->rd_rel->relnatts, srel->rd_rel->relsmgr, tupdesc);
    for (i = 0; i < srel->rd_rel->relnatts; ++i) {
	pfree((void *) tupdesc[i]);
    }
    pfree((void *) tupdesc);

    CommandCounterIncrement();
    {
	char query[1000];

	(void) sprintf(query, "UPDATE pg_class SET relkey='%d 0 0 0 0 0 0 0' WHERE oid='%d'::oid",
		       blocksz, roid);
	pg_eval(query, (char *) NULL, (Oid *) NULL, 0);
    }
    CommandCounterIncrement();

    rrel = heap_open(roid);

    /*
     * this is gross, but if we do this the right way (invalidating
     * and reconstructing the reldesc) then the 'islocal' flag gets
     * cleared..
     */
    rrel->rd_blocksz = rrel->rd_rel->relkey[0] = blocksz;

    i = 0;
    rbuf = InvalidBuffer;
    for (sblkno = 0; sblkno < nblocks; ++sblkno) {
	sbuf = ReadBuffer(srel, sblkno);
	spage = BufferGetPage(sbuf);
	maxoff = PageGetMaxOffsetNumber(spage);
	for (soff = FirstOffsetNumber;
	     soff <= maxoff;
	     soff = OffsetNumberNext(soff)) {
	    itemid = PageGetItemId(spage, soff);
	    if (ItemIdIsUsed(itemid)) {
		sitem = PageGetItem(spage, itemid);
		itemsz = ItemIdGetLength(itemid);
		itemsz = DOUBLEALIGN(itemsz);
		if (BufferIsInvalid(rbuf) ||
		    PageGetFreeSpace(rpage) < itemsz) {
		    if (BufferIsInvalid(rbuf)) {
			rblkno = 0;
			rbuf = ReadBuffer(rrel, rblkno);
		    } else {
			if (PTidUseHash) {
			    PTidHashInsert(handle, &slast, &rfirst);
			} else {
			    PTidArrayInsert(handle, &slast, &rfirst);
			}
			(void) WriteBuffer(rbuf);
			rbuf = ReadBuffer(rrel, ++rblkno);
		    }
		    rpage = BufferGetPage(rbuf);
		    PageInit(rpage, BufferGetPageSize(rbuf), 0);
		    roff = FirstOffsetNumber;
		    /*
		     * reset 'rfirst'
		     */
		    ItemPointerSet(&rfirst, rblkno, roff);
		}
		/*
		 * insert the item, remembering to fix the TID in the
		 * new heap tuple so that heapgetnext will work.
		 */
		PageManagerModeSet(OverwritePageManagerMode);
		(void) PageAddItem(rpage, sitem, itemsz, roff, LP_USED);
		PageManagerModeSet(ShufflePageManagerMode);
		ritem = PageGetItem(rpage, PageGetItemId(rpage, roff));
		ItemPointerSet(&(((HeapTuple) ritem)->t_ctid), rblkno, roff);

		/*
		 * 'slast' tracks the last item inserted for a given
		 * sender block; 'rfirst' tracks the first item
		 * inserted for a given receiver block.
		 */
		ItemPointerSet(&slast, sblkno, soff);
		if (soff == FirstOffsetNumber) {
		    ItemPointerSet(&rfirst, rblkno, roff);
		}
		++i;
	    }
	    roff = OffsetNumberNext(roff);
	}
	if (PTidUseHash) {
	    PTidHashInsert(handle, &slast, &rfirst);
	} else {
	    PTidArrayInsert(handle, &slast, &rfirst);
	}
	ReleaseBuffer(sbuf);
    }
    (void) WriteBuffer(rbuf);

    heap_close(srel);
    heap_close(rrel);
    FlushBufferPool(0);		/* so heap_beginscan sees the #$! pages.. */
    RelationFlushRelation(&rrel, true);
    CommandCounterIncrement();

    printf("saw %d valid/invalid tuples\n", i);
    return((void *) handle);
}

int
PTidTest(Name from_heap, Name from_index, int blocksz, bool usehash)
{
    Oid oid;
    char to_heap[sizeof(NameData)+1];
    char to_index[sizeof(NameData)+1];
    void *handle;
    Relation srel;
    Relation rrel;
    Relation irel;
    HeapScanDesc sscan;
    HeapScanDesc rscan;
    Buffer sbuf;
    Buffer rbuf;
    HeapTuple stup;
    HeapTuple rtup;
    int i;

    PTidUseHash = usehash;

    oid = newoid();
    (void) sprintf(to_heap, "H%x_%x", oid, getpid());
    (void) sprintf(to_index, "I%x_%x", oid, getpid());

    ResetBufferUsage();

    BufferPoolBlowaway();
    elog(NOTICE, "=== heap START ===");
    handle = heap_tidmap(from_heap, (Name) to_heap, blocksz);
    elog(NOTICE, "=== heap DONE ===");

#if 0
    /* scan the heap */
    srel = heap_openr(from_heap);
    rrel = heap_openr((Name) to_heap);
    sscan = heap_beginscan(srel, 0, NowTimeQual, 0, (ScanKey) NULL);
    rscan = heap_beginscan(rrel, 0, NowTimeQual, 0, (ScanKey) NULL);
    i = 0;
    sbuf = rbuf = InvalidBuffer;
    while (HeapTupleIsValid(stup = heap_getnext(sscan, 0, &sbuf))) {
	ItemPointerData check;

	++i;
	if (!PTidUseHash) {
	    PTidArrayMap(handle, &(stup->t_ctid), &check);
	    if (!HeapTupleIsValid(rtup = heap_fetch(rrel, NowTimeQual,
						    &check, &rbuf))) {
		printf("ERROR: s=%s array=%s\n", tidout(&check),
		       tidout(&(stup->t_ctid)));
	    } else if (ItemPointerEquals(&(rtup->t_ctid), &check)) {
		if (DebugLvl > 0) {
		    printf("VALID: s=%s r=%s\n", 
			   tidout(&(stup->t_ctid)), tidout(&(rtup->t_ctid)));
		}
	    } else {
		printf("ERROR: s=%s r=%s array=%s\n",
		       tidout(&check),
		       tidout(&(stup->t_ctid)), tidout(&(rtup->t_ctid)));
	    }
	}
	ReleaseBuffer(rbuf);
    }
    printf("saw %d valid tuples\n", i);
    heap_endscan(sscan);
    heap_endscan(rscan);
    heap_close(srel);
    heap_close(rrel);
#endif

    BufferPoolBlowaway();
    rrel = heap_openr((Name) to_heap);
    elog(NOTICE, "=== btree LEAF ===");
    _bt_tidmap(from_index, to_index, rrel, handle, blocksz);
    elog(NOTICE, "=== btree UPPER ===");
    irel = index_openr((Name) to_index);
    _bt_upperbuild(irel);
    elog(NOTICE, "=== btree DONE ===");

    {
	extern int ReadBufferCount, BufferHitCount, BufferFlushCount;
	extern long NDirectFileRead, NDirectFileWrite;

	printf("buffer(%d): r=%d w=%d\n", rrel->rd_blocksz,
	       ReadBufferCount - BufferHitCount, BufferFlushCount);
	printf("direct(%d): r=%d w=%d\n", LocalBlockSize,
	       NDirectFileRead, NDirectFileWrite);
    }

    heap_close(rrel);
    index_close(irel);

    if (PTidUseHash) {
#if 0
	PTidHashDestroy(handle);
#endif
    } else {
	PTidArrayDestroy(handle);
    }

    return(0);
}

#endif /* TIDTEST */

