/*-------------------------------------------------------------------------
 *
 * btdbg.c--
 *    code to debug btrees.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 * IDENTIFICATION
 *    $Id: nbtdbg.c,v 1.1 1996/05/09 02:49:10 aoki Exp $
 *
 * NOTES
 *    this code is not for use under multiuser conditions.
 *
 *-------------------------------------------------------------------------
 */

#include "c.h"

#include "access/genam.h"
#include "access/nbtree.h"

#include "utils/elog.h"

#define	NEXTPAGE(irel, buf, page, opaque, blk, idx) \
    _bt_relbuf(irel, buf, BT_READ); \
    if (!BufferIsValid(buf = _bt_getbuf(irel, blk, BT_READ))) { \
	elog(WARN, "btdump: could not get blk %d of index '%.*s'", \
	     blk, NAMEDATALEN, idx); \
    } \
    page = BufferGetPage(buf); \
    opaque = (BTPageOpaque) PageGetSpecialPointer(page)

/*
 *  bt_dumptup() --
 *
 *  XXX use type output routines
 */
void
_bt_dumptup(int lvl,
	    Relation irel, IndexTuple itup, BlockNumber blk, OffsetNumber off)
{
    bool isnull = false;
    Datum d = index_getattr(itup, 1, irel->rd_att, &isnull);
    
    if (isnull == true) {
	printf("%d [ %d %d ] NULL\n", lvl, blk, off);
    } else {
	printf("%d [ %d %d ] %d\n", lvl, blk, off, (int) d);
    }
}

/*
 *  btdump() -- dump level 'lvl' of btree 'idx', where level 0 is the
 *		root.  this routine traverses the tree manually and is
 *		therefore unsuitable for use under multiuser conditions.
 *
 * XXX modify to return the number of levels
 */
int
btdump(Name idx, int lvl)
{
    Relation irel;
    Buffer buf;
    Page page;
    BTPageOpaque opaque;
    BlockNumber blk;
    OffsetNumber off;
    OffsetNumber maxoff;
    BTItem bti;
    IndexTuple itup;
    int i;

    if (!NameIsValid(idx)) {
	elog(WARN, "btdump: no index name provided");
    }
    if (lvl < 0) {
	elog(WARN, "btdump: index level must be greater than 0");
    }

    if (!RelationIsValid(irel = index_openr(idx))) {
	elog(WARN, "btdump: could not open index '%.*s'",
	     NAMEDATALEN, idx);
    }

    /*
     * drop down the left edge of the btree to level 'lvl'.
     */
    if (!BufferIsValid(buf = _bt_getroot(irel, BT_READ))) {
	elog(WARN, "btdump: could not get root of index '%.*s'",
	     NAMEDATALEN, idx);
    }
    blk = BufferGetBlockNumber(buf);
    page = BufferGetPage(buf);
    opaque = (BTPageOpaque) PageGetSpecialPointer(page);
    if (!(opaque->btpo_flags & BTP_ROOT)) {
	elog(WARN, "btdump: retrieved invalid root for index '%.*s'",
	     NAMEDATALEN, idx);
    }
    for (i = 0; i < lvl; ++i) {
	if (opaque->btpo_flags & BTP_LEAF) {
	    elog(WARN,
		 "btdump: could not reach level %d of index '%.*s' (max = %d)",
		 lvl, NAMEDATALEN, idx, i);
	}

	/*
	 * if the leftmost pages of this level are empty, we have to
	 * walk right until we find a page with data.
	 */
	while (PageIsEmpty(page)) {
	    if (P_RIGHTMOST(opaque)) {
		elog(WARN, "btdump: level %d of btree '%.*s' is empty!",
		     i, NAMEDATALEN, idx);
	    }

	    blk = opaque->btpo_next;
	    NEXTPAGE(irel, buf, page, opaque, blk, idx);
	}

	/*
	 * now descend to the page to which the first "data" (i.e.,
	 * non-high-key) index tuple points.
	 */
	off = P_RIGHTMOST(opaque) ? P_HIKEY : P_FIRSTKEY;
	bti = (BTItem) PageGetItem(page, PageGetItemId(page, off));

	blk = ItemPointerGetBlockNumber(&(bti->bti_itup.t_tid));
	NEXTPAGE(irel, buf, page, opaque, blk, idx);
    }

    /*
     * we have reached the leftmost page of the desired btree level.
     * now we can just scan right, printing index tuples as we go.
     *
     * we don't have to check whether the item is valid because the
     * btree code keeps the items packed into sorted, contiguous
     * arrays.
     */
    do {
	while (PageIsEmpty(page)) {
	    if (P_RIGHTMOST(opaque)) {
		elog(WARN, "btdump: done.");	/* XXX */
	    }

	    blk = opaque->btpo_next;
	    NEXTPAGE(irel, buf, page, opaque, blk, idx);
	}

	off = P_RIGHTMOST(opaque) ? P_HIKEY : P_FIRSTKEY;
	maxoff = PageGetMaxOffsetNumber(page);
	for (; off <= maxoff; off = OffsetNumberNext(off)) {
	    bti = (BTItem) PageGetItem(page, PageGetItemId(page, off));
	    _bt_dumptup(lvl, irel, &(bti->bti_itup), blk, off);
	}

	blk = opaque->btpo_next;
	NEXTPAGE(irel, buf, page, opaque, blk, idx);
    } while (!P_RIGHTMOST(opaque));
    _bt_relbuf(irel, buf, BT_READ);

    return(0);
}
