/*-------------------------------------------------------------------------
 *
 * btmap.c--
 *    BTree-specific TID mapping code
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Id: nbtmap.c,v 1.1 1995/09/05 09:27:30 aoki Exp $
 *
 *  NOTES
 *
 *-------------------------------------------------------------------------
 */
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "utils/memutils.h"
#include "storage/bufmgr.h"
#include "storage/item.h"
#include "storage/itemid.h"
#include "storage/page.h"

#include "access/genam.h"
#include "access/heapam.h"
#include "access/itup.h"
#include "access/nbtree.h"
#include "access/valid.h"

#include "catalog/pg_index.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/syscache.h"
#include "miscadmin.h"

extern void *heap_tidmap(Name, Name);

#define	BTITEMSZ(btitem) \
    ((btitem) ? \
     (IndexTupleDSize((btitem)->bti_itup) + \
      (sizeof(BTItemData) - sizeof(IndexTupleData))) : \
     0)

extern bool PTidUseHash;

/*
 * generate a ScanKey for 'irel' using an equality test.
 *
 * hack: the key generated by _bt_mkscankey uses this type's
 * BTORDER_PROC, which will return {-1,0,1} depending on the relative
 * size of the argument.  if we set SK_NEGATE, heap_keytest will make
 * things work as if BTORDER_PROC returned {0,1,0}.  This is just a
 * way to get an equality test without messing around with strategies
 * and the like.
 */
static ScanKey
_bt_mkheapscankey(Relation irel)
{
    ScanKey skey;
    HeapTuple iheap;
    FormData_pg_index *iform;
    RegProcedure proc = index_getprocid(irel, 1, BTORDER_PROC);

    skey = (ScanKey) palloc(sizeof(ScanKeyData));
    iheap = SearchSysCacheTuple(INDEXRELID, (char *) irel->rd_id,
				(char *) NULL, (char *) NULL, (char *) NULL);
    if (!HeapTupleIsValid(iheap)) {
	elog(WARN, "_bt_mkheapscankey: index tuple for '%d' not found",
	     irel->rd_id);
    }
    iform = (FormData_pg_index *) GETSTRUCT(iheap);
    ScanKeyEntryInitialize(&skey[0], SK_NEGATE, iform->indkey[0],
			   proc, (Datum) NULL);
    return(skey);
}

/*
 * returns the offset of the tuple on the heap page to which 'itup'
 * points that matches 'itup's key.  returns InvalidOffsetNumber if no
 * matching tuple is found.
 */
OffsetNumber
_bt_getheapoff(Relation hrel, Relation irel, ScanKey skey, IndexTuple itup)
{
    /* XXX ItemPointerGetBlockNumber checks the #$%! OffsetNumber.. */
    BlockNumber blkno = BlockIdGetBlockNumber(&(itup->t_tid.ip_blkid));
    bool isnull;
    Datum key = index_getattr(itup, 1, irel->rd_att, &isnull);
    Buffer buf;
    Page page;
    OffsetNumber off;
    OffsetNumber maxoff;
    HeapTuple htup;

    skey[0].sk_argument = key;

    buf = ReadBuffer(hrel, blkno);
    page = BufferGetPage(buf);
    if (PageIsEmpty(page)) {
	return(InvalidOffsetNumber);
    }
    maxoff = PageGetMaxOffsetNumber(page);
    for (off = FirstOffsetNumber;
	 off <= maxoff;
	 off = OffsetNumberNext(off)) {
	htup = heap_tuple_satisfies(PageGetItemId(page, off), hrel,
				    (PageHeader) page, NowTimeQual, 1, skey);
	if (HeapTupleIsValid(htup)) {
	    ReleaseBuffer(buf);
	    return(off);
	}
    }
    ReleaseBuffer(buf);

    /* either there's no such heap tuple, or it was not committed. */
    return(InvalidOffsetNumber);
}

/*
 * these implement several options for doing bulk i/o.
 *	USE_PAGEIO	read one page at a time through the buffer manager
 *	USE_BLOCKIO	read several pages at a time using private buffers
 *	USE_BULKIO	read the whole file into memory ...
 *		USE_BULK_FILE	... using private buffers
 *		USE_BULK_MMAP	... using mmap(2)
 * these various options seem to make very little real difference.
 */

/*#define USE_BULKIO*/
/*#define USE_BLOCKIO*/
#define USE_PAGEIO

#ifdef USE_BULKIO

#define USE_BULK_MMAP
/*#define USE_BULK_FILE*/

/*
 * if USE_BULK_FILE is set, reads the entire index into private buffers.
 * if USE_BULK_MMAP is set, mmap(2)'s the entire index file.
 */
static char *
_bt_mapfile(char *filename, int *len)
{
    int fd;
    char *region = (char *) NULL;
    struct stat statbuf;
    int pagesz;

    FlushBufferPool(0);
    if ((fd = open(filename, O_RDONLY, 0666)) < 0) {
	elog(WARN, "_bt_mapfile: open('%s') failed: %m", filename);
    }
    if (fstat(fd, &statbuf) < 0) {
	elog(WARN, "_bt_mapfile: fstat('%s') failed: %m", filename);
    }
#ifdef USE_BULK_MMAP
    if ((pagesz = sysconf(_SC_PAGE_SIZE)) < 0) {
	elog(WARN, "_bt_mapfile: sysconf(_SC_PAGE_SIZE) failed: %m");
    }
    *len = (statbuf.st_size + (pagesz - 1)) & (~pagesz);
    region = mmap((caddr_t) NULL, *len, PROT_READ,
		  MAP_FILE|MAP_VARIABLE|MAP_PRIVATE, fd, 0);
    if (region == (char *) -1 || region == (char *) NULL) {
	elog(WARN, "_bt_mapfile: mmap('%s') failed: %m", filename);
    }
#endif /* USE_BULK_MMAP */
#ifdef USE_BULK_FILE
    *len = statbuf.st_size;
    region = palloc(statbuf.st_size);
    if (read(fd, region, statbuf.st_size) != statbuf.st_size) {
	pfree((void *) region);
	elog(WARN, "_bt_mapfile: read('%s') failed: %m", filename);
    }
#endif /* USE_BULK_FILE */
    if (close(fd) < 0) {
	elog(WARN, "_bt_mapfile: close('%s') failed: %m", filename);
    }
    return(region);
}

/*
 * frees the resources allocated by _bt_mapfile.
 */
static void
_bt_unmapfile(char *region, int len)
{
#ifdef USE_BULK_MMAP
    if (munmap(region, len) < 0) {
	elog(WARN, "_bt_unmapfile: munmap failed: %m");
    }
#endif /* USE_BULK_MMAP */
#ifdef USE_BULK_FILE
    pfree((void *) region);
#endif /* USE_BULK_FILE */
}

#endif /* USE_BULKIO */

/*
 * 'from_heap' has been copied into 'to_heap'; 'map' is the TID map as
 * generated by heap_tidmap().  we now want to copy 'from_index' into
 * 'to_index', fixing the ItemPointers.
 */
void
_bt_tidmap(Name from_index, Name to_index, Relation to_rel, void *map,
	   int blocksz)
{
    Relation from_irel;
    Relation to_irel;
    NameData to_dummy;
    IndexScanDesc scan;
    RetrieveIndexResult res;
    ScanKey skey;
    void *state;
#ifdef USE_BULKIO
    char *from_file;
    int from_len;
#endif /* USE_BULKIO */
#ifdef USE_BLOCKIO
    File from_fd;
    char *from_buf;
    BlockNumber first_blk;
    BlockNumber last_blk;
#endif /* USE_BLOCKIO */
#ifdef USE_PAGEIO
    Buffer rbuf;
#endif /* USE_PAGEIO */
    Page rpage;
    BTPageOpaque ropaque;
    BlockNumber blk;
    BlockNumber lastblk;
    BTItem obti;
    ItemId oiid;
    OffsetNumber firstoff;
    OffsetNumber off;
    OffsetNumber maxoff;
    OffsetNumber noff;
    Size btisz;
    BTItem nbti = (BTItem) NULL;
    Size nbtisz = 0;
    Oid to_relid;
    Oid to_irelid;
    int ntups = 0;

    from_irel = index_openr(from_index);
    if (!RelationIsValid(from_irel)) {
	elog(WARN, "_bt_tidmap: could not open '%.*s'",
	     NAMEDATALEN, from_index);
    }

    /* copy_index creates "temp_xxxxxxxx"; fix it. */
    CommandCounterIncrement();
    {
	extern bool _doindexbuild_;

	_doindexbuild_ = false;
	copy_index(from_irel->rd_id, to_rel->rd_id);
	_doindexbuild_ = true;
    }
    CommandCounterIncrement();
    sprintf((char *) &to_dummy, "temp_%x", from_irel->rd_id);

    /* find the first leaf page in the 'from' index */
    scan = index_beginscan(from_irel, false, 0, (ScanKey) NULL);
    res = index_getnext(scan, ForwardScanDirection);
    blk = ItemPointerGetBlockNumber(&(res->index_iptr));
    index_endscan(scan);
    
    /* initialize scan information for the 'to' index */
    to_irel = index_openr(&to_dummy);
    if (!RelationIsValid(to_irel)) {
	elog(WARN, "_bt_tidmap: could not open '%.*s'",
	     NAMEDATALEN, &to_dummy);
    }
    if (to_irel->rd_blocksz != to_rel->rd_blocksz || 
	to_irel->rd_blocksz != blocksz) {
	elog(WARN, "_bt_tidmap: bad block size for '%.*s': %d",
	     NAMEDATALEN, &to_dummy, to_irel->rd_blocksz);
    }
    _bt_metapinit(to_irel);

    /* lock the whole index and turn off page locking */
#if 0
    FileAsync(to_irel->rd_fd, true);
#endif
    BuildingBtree = true;
    RelationSetLockForRead(from_irel);

#ifdef USE_BULKIO
    from_file = _bt_mapfile((char *) &(from_irel->rd_rel->relname), &from_len);
    lastblk = from_len / from_irel->rd_blocksz;
#endif /* USE_BULKIO */
#ifdef USE_BLOCKIO
    from_fd = FileNameOpenFile((char *) &(from_irel->rd_rel->relname),
			       O_RDONLY, 0666);
    from_buf = (char *) palloc(MAXBLCKSZ << 3);
    first_blk = last_blk = InvalidBlockNumber;
#endif /* USE_BLOCKIO */

    /* for each page... */
    skey = _bt_mkheapscankey(to_irel);
    state = _bt_pagestate(to_irel, BTP_LEAF, 0, false);

    do {
#if 0
	printf("blk=%d\n", blk);
#endif
#ifdef USE_BULKIO
	rpage = (Page) (from_file + (blk * from_irel->rd_blocksz));
#endif /* USE_BULKIO */
#ifdef USE_BLOCKIO
	if (first_blk != InvalidBlockNumber &&
	    (blk < first_blk || blk > last_blk)) {
	    int from_len;

	    FileSeek(from_fd, (blk * from_irel->rd_blocksz), SEEK_SET);
	    from_len = FileRead(from_fd, from_buf, from_irel->rd_blocksz << 3);
	    first_blk = blk;
	    last_blk = blk + (from_len / from_irel->rd_blocksz);
	}
	rpage = from_buf + ((blk - first_blk) * from_irel->rd_blocksz);
#endif /* USE_BLOCKIO */
#ifdef USE_PAGEIO
	rbuf = _bt_getbuf(from_irel, blk, BT_READ);
	rpage = BufferGetPage(rbuf);
#endif /* USE_PAGEIO */
	ropaque = (BTPageOpaque) PageGetSpecialPointer(rpage);
	
	if (!PageIsEmpty(rpage)) {
	    maxoff = PageGetMaxOffsetNumber(rpage);
	    firstoff = P_RIGHTMOST(ropaque) ? P_HIKEY : P_FIRSTKEY;

	    /* for each item... */
	    for (off = firstoff;
		 off <= maxoff;
		 off = OffsetNumberNext(off)) {
		oiid = PageGetItemId(rpage, off);
		if (!ItemIdIsValid(oiid)) {
		    continue;
		}
		obti = (BTItem) PageGetItem(rpage, oiid);
		if (PTidUseHash) {
		    btisz = BTITEMSZ(obti);
		    btisz = DOUBLEALIGN(btisz);
		    if (btisz > nbtisz) {
			if (nbti) {
			    pfree((void *) nbti);
			}
			nbti = (BTItem) palloc(btisz);
			nbtisz = btisz;
		    }
		    memcpy((char *) nbti, (char *) obti, btisz);
		    PTidHashMap(map, &(obti->bti_itup.t_tid),
				&(nbti->bti_itup.t_tid));

		    /*
		     * the itemptr offnum is still unmapped; we need a
		     * key and a valid block number to find it.
		     */
		    noff = _bt_getheapoff(to_rel, to_irel, skey,
					  &(nbti->bti_itup));
		    ItemPointerSetOffsetNumber(&(nbti->bti_itup.t_tid), noff);
		    if (ItemPointerIsValid(&(nbti->bti_itup.t_tid))) {
			(void) _bt_buildadd(to_irel, state, nbti, BTP_LEAF);
			++ntups;
		    }
		} else {
		    nbti = _bt_buildadd(to_irel, state, obti, BTP_LEAF);
		    PTidArrayMap(map, &(obti->bti_itup.t_tid),
				 &(nbti->bti_itup.t_tid));
		    ++ntups;
		}
	    }
	}
	blk = ropaque->btpo_next;
#ifdef USE_PAGEIO
	_bt_relbuf(from_irel, rbuf, BT_READ);
#endif /* USE_PAGEIO */
    } while (blk != P_NONE);
#ifdef USE_BULKIO
    _bt_unmapfile(from_file, from_len);
#endif /* USE_BULKIO */
#ifdef USE_BLOCKIO
    pfree((void *) from_buf);
    FileClose(from_fd);
#endif /* USE_BLOCKIO */
	
    if (PTidUseHash && nbti != (BTItem) NULL) {
	pfree((void *) nbti);
    }
    pfree((void *) skey);

    _bt_uppershutdown(to_irel, state);
    {
	Buffer wbuf = _bt_getbuf(to_irel, 1, BT_WRITE);
	Page wpage = BufferGetPage(wbuf);
	BTPageOpaque wopaque = (BTPageOpaque) PageGetSpecialPointer(wpage);

	wopaque->btpo_flags |= BTP_ROOT;
	_bt_wrtbuf(to_irel, wbuf);
    }
    _bt_metaproot(to_irel, 1);

    RelationUnsetLockForRead(from_irel);
    BuildingBtree = false;
#if 0
    FileAsync(to_irel->rd_fd, false);
#endif

    to_relid = to_rel->rd_id;
    to_irelid = to_irel->rd_id;

    index_close(from_irel);
    index_close(to_irel);

    UpdateStats(to_relid, ntups, true);
    UpdateStats(to_irelid, ntups, false);

    /*
     * do the renamerel after everything else -- it makes the index
     * non-local.  be sure to flush out any dirty local buffers so
     * that we don't lose the updates (global buffer searches won't
     * find them, of course).
     */
    LocalBufferSync();
    CommandCounterIncrement();
    renamerel((char *) &to_dummy, to_index);
    CommandCounterIncrement();
}
