/*-------------------------------------------------------------------------
 *
 * indexing.h--
 *    This include provides some definitions to support indexing 
 *    on system catalogs
 *
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 * $Id: indexing.h,v 0.5 1996/06/07 04:54:47 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#ifndef INDEXING_H
#define INDEXING_H

#include "utils/rel.h"

/*
 * Some definitions for indices on pg_attribute
 */
#define Num_pg_attr_indices	3
#define Num_pg_proc_indices	3
#define Num_pg_type_indices	2
#define Num_pg_class_indices	2
#ifndef P95
#define Num_pg_nsvcattr_indices 1
#define Num_pg_nsvcclass_indices 1
#define Num_pg_nsvcfrag_indices 1
#endif /* P95 */


/*
 * Names of indices on system catalogs
 */
#ifndef P95
extern Name AttributeNameIndex;
extern Name AttributeNumIndex;
extern Name ProcedureNameIndex;
extern Name ProcedureOidIndex;
extern Name ProcedureSrcIndex;
extern Name TypeNameIndex;
extern Name TypeOidIndex;
extern Name ClassNameIndex;
extern Name ClassOidIndex;
extern Name NsvcAttrOidIndex;
extern Name NsvcClassOidIndex;
extern Name NsvcFragOidIndex;
#else /* P95 */
#define AttributeNameIndex "pg_attnameind"
#define AttributeNumIndex  "pg_attnumind"
#define AttributeRelidIndex "pg_attrelidind"
#define ProcedureNameIndex "pg_procnameind"
#define ProcedureOidIndex  "pg_procidind"
#define ProcedureSrcIndex  "pg_procsrcind"
#define TypeNameIndex      "pg_typenameind"
#define TypeOidIndex       "pg_typeidind"
#define ClassNameIndex     "pg_classnameind"
#define ClassOidIndex      "pg_classoidind"
#endif /* P95 */

extern char *Name_pg_attr_indices[];
extern char *Name_pg_proc_indices[];
extern char *Name_pg_type_indices[];
extern char *Name_pg_class_indices[];
#ifndef P95
extern char *Name_pg_nsvcattr_indices[];
extern char *Name_pg_nsvcclass_indices[];
extern char *Name_pg_nsvcfrag_indices[];
#endif /* P95 */

#ifndef P95
extern Name IndexedCatalogNames[];
#else /* P95 */
extern char *IndexedCatalogNames[];
#endif /* P95 */

/*
 * indexing.c prototypes 
 *
 * Functions for each index to perform the necessary scan on a cache miss.
 */
extern void CatalogOpenIndices(int nIndices, char *names[], Relation idescs[]);
extern void CatalogCloseIndices(int nIndices, Relation *idescs);
extern void CatalogIndexInsert(Relation *idescs,
			       int nIndices,
			       Relation heapRelation,
			       HeapTuple heapTuple);
extern bool CatalogHasIndex(char *catName, Oid catId);

extern HeapTuple AttributeNameIndexScan(Relation heapRelation,
					Oid relid,
					char *attname);

extern HeapTuple AttributeNumIndexScan(Relation heapRelation,
				       Oid relid,
				       AttrNumber attnum);
extern HeapTuple ProcedureOidIndexScan(Relation heapRelation, Oid procId);
extern HeapTuple ProcedureNameIndexScan(Relation heapRelation,
	char *procName, int nargs, Oid *argTypes);
extern HeapTuple ProcedureSrcIndexScan(Relation heapRelation, text *procSrc);
extern HeapTuple TypeOidIndexScan(Relation heapRelation, Oid typeId);
extern HeapTuple TypeNameIndexScan(Relation heapRelation, char *typeName);
extern HeapTuple ClassNameIndexScan(Relation heapRelation, char *relName);
extern HeapTuple ClassOidIndexScan(Relation heapRelation, Oid relId);


/*
 * What follows are lines processed by genbki.sh to create the statements
 * the bootstrap parser will turn into DefineIndex commands.
 *
 * The keyword is DECLARE_INDEX every thing after that is just like in a
 * normal specification of the 'define index' POSTQUEL command.
 */
#ifndef P95
DECLARE_INDEX(pg_attnameind on pg_attribute using btree (mkoidchar16(attrelid, attname) oidchar16_ops));
#else /* P95 */
DECLARE_INDEX(pg_attnameind on pg_attribute using btree (mkoidname(attrelid, attname) oidname_ops));
#endif /* P95 */
DECLARE_INDEX(pg_attnumind  on pg_attribute using btree (mkoidint2(attrelid, attnum) oidint2_ops));
DECLARE_INDEX(pg_attrelidind on pg_attribute using btree (attrelid oid_ops));

DECLARE_INDEX(pg_procidind on pg_proc using btree (Oid oid_ops));
#ifndef P95
DECLARE_INDEX(pg_procnameind on pg_proc using btree (proname char16_ops));
#else /* P95 */
DECLARE_INDEX(pg_procnameind on pg_proc using btree (proname name_ops));
#endif /* P95 */
DECLARE_INDEX(pg_procsrcind on pg_proc using btree (prosrc text_ops));

DECLARE_INDEX(pg_typeidind on pg_type using btree (Oid oid_ops));
#ifndef P95
DECLARE_INDEX(pg_typenameind on pg_type using btree (typname char16_ops));
#else /* P95 */
DECLARE_INDEX(pg_typenameind on pg_type using btree (typname name_ops));
#endif /* P95 */

#ifndef P95
DECLARE_INDEX(pg_classnameind on pg_class using btree (relname char16_ops));
#else /* P95 */
DECLARE_INDEX(pg_classnameind on pg_class using btree (relname name_ops));
#endif /* P95 */
DECLARE_INDEX(pg_classoidind on pg_class using btree (Oid oid_ops));
#ifndef P95

DECLARE_INDEX(hashoid72 on pg_nsvcattr using hash (Oid oid_ops));
DECLARE_INDEX(hashoid73 on pg_nsvcclass using hash (Oid oid_ops));
DECLARE_INDEX(hashoid74 on pg_nsvcfrag using hash (Oid oid_ops));
#endif /* P95 */

/* now build indices in the initialization scripts */
BUILD_INDICES

#endif /* INDEXING_H */
