/*-------------------------------------------------------------------------
 *
 * pg_aggregate.c--
 *    routines to support manipulation of the pg_aggregate relation
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/catalog/pg_aggregate.c,v 0.5 1996/06/07 04:54:48 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <string.h>
#include "postgres.h"

#include "access/heapam.h"
#include "access/relscan.h"
#include "access/skey.h"
#include "access/htup.h"
#ifdef P95
#include "access/tupdesc.h"
#endif /* P95 */
#include "utils/rel.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#ifdef P95
#include "utils/builtins.h"
#endif /* P95 */
#include "fmgr.h"

#include "catalog/catname.h"
#include "utils/syscache.h"
#include "catalog/pg_operator.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "catalog/pg_aggregate.h"

/* ----------------
	[Mariposa]
 * AggregateDefine
	[Postgres95]
 * AggregateCreate
 *
 * [Postgres95]
 * aggregates overloading has been added.  Instead of the full
 * overload support we have for functions, aggregate overloading only
 * applies to exact basetype matches.  That is, we don't check the 
 * the inheritance hierarchy
 * [end Postgres95]
 * OLD COMMENTS:
 *	Currently, redefining aggregates using the same name is not
 *	supported.  In such a case, a warning is printed that the 
 *	aggregate already exists.  If such is not the case, a new tuple
 *	is created and inserted in the aggregate relation.  The fields
 *	of this tuple are aggregate name, owner id, 2 transition functions
 *	(called aggtransfn1 and aggtransfn2), final function (aggfinalfn),
 *	type of data on which aggtransfn1 operates (aggbasetype), return
 *	types of the two transition functions (aggtranstype1 and 
 *	aggtranstype2), final return type (aggfinaltype), and initial values
 *	for the two state transition functions (agginitval1 and agginitval2).
 *	All types and functions must have been defined
 *	prior to defining the aggregate.
 * ---------------
 */
void
#ifndef P95
AggregateDefine(Name aggName,
		Name aggtransfn1Name,
		Name aggtransfn2Name,
		Name aggfinalfnName, 
		Name aggbasetypeName,
		Name aggtransfn1typeName,
		Name aggtransfn2typeName,
		String agginitval1,
		String agginitval2)
#else /* P95 */
AggregateCreate(char *aggName,
		char *aggtransfn1Name,
		char *aggtransfn2Name,
		char *aggfinalfnName, 
		char *aggbasetypeName,
		char *aggtransfn1typeName,
		char *aggtransfn2typeName,
		char *agginitval1,
		char *agginitval2)
#endif /* P95 */
{
    register		i;
    Relation		aggdesc;
    HeapTuple		tup;
    char		nulls[Natts_pg_aggregate];
    Datum               values[Natts_pg_aggregate];
    Form_pg_proc	proc;
    Oid		xfn1 = InvalidOid;
    Oid		xfn2 = InvalidOid;
    Oid		ffn = InvalidOid;
    Oid		xbase = InvalidOid;
    Oid		xret1 = InvalidOid;
    Oid		xret2 = InvalidOid;
    Oid		fret = InvalidOid;
    Oid		fnArgs[8];
#ifdef P95
    TupleDesc    tupDesc;
#endif /* P95 */
    
    memset(fnArgs, 0, 8 * sizeof(Oid)); 
    
    /* sanity checks */
#ifndef P95
    if (!NameIsValid(aggName))
	elog(WARN, "AggregateDefine: no aggregate name supplied");
    tup = SearchSysCacheTuple(AGGNAME, aggName->data,
			      (char *) NULL, (char *) NULL, (char *) NULL);
    if (HeapTupleIsValid(tup))
	elog(WARN, "AggregateDefine: aggregate \"%-.*s\" already exists",
	     sizeof(NameData), aggName->data);
#else /* P95 */
    if (!aggName)
	elog(WARN, "AggregateCreate: no aggregate name supplied");
#endif /* P95 */
    
#ifndef P95
    if (!NameIsValid(aggtransfn1Name) && !NameIsValid(aggtransfn2Name))
	elog(WARN, "AggregateDefine: aggregate must have at least one transition function");
#else /* P95 */
    if (!aggtransfn1Name && !aggtransfn2Name)
	elog(WARN, "AggregateCreate: aggregate must have at least one transition function");
#endif /* P95 */
    
#ifndef P95
    if (NameIsValid(aggtransfn1Name)) {
	tup = SearchSysCacheTuple(TYPNAME, (char *) aggbasetypeName,
				  (char *) NULL, (char *) NULL, (char *) NULL);
#else /* P95 */
    tup = SearchSysCacheTuple(TYPNAME, 
			      PointerGetDatum(aggbasetypeName),
			      0,0,0);
    if(!HeapTupleIsValid(tup))
	elog(WARN, "AggregateCreate: Type '%s' undefined",aggbasetypeName);
    xbase = tup->t_oid;

    if (aggtransfn1Name) {
	tup = SearchSysCacheTuple(TYPNAME, 
				  PointerGetDatum(aggtransfn1typeName),
				  0,0,0);
#endif /* P95 */
	if(!HeapTupleIsValid(tup))
#ifndef P95
	    elog(WARN, "AggregateDefine: Type \"%-.*s\" undefined",
		 sizeof(NameData), aggbasetypeName);
	xbase = tup->t_oid;
	
	tup = SearchSysCacheTuple(TYPNAME, (char *) aggtransfn1typeName,
				  (char *) NULL, (char *) NULL, (char *) NULL);
	if(!HeapTupleIsValid(tup))
	    elog(WARN, "AggregateDefine: Type \"%-.*s\" undefined",
		 sizeof(NameData), aggtransfn1typeName);
#else /* P95 */
	    elog(WARN, "AggregateCreate: Type '%s' undefined",
		 aggtransfn1typeName);
#endif /* P95 */
	xret1 = tup->t_oid;
	
	fnArgs[0] = xret1;
	fnArgs[1] = xbase;
	tup = SearchSysCacheTuple(PRONAME,
#ifndef P95
				  (char *) aggtransfn1Name->data,
				  (char *) Int32GetDatum(2),
				  (char *) fnArgs,
				  (char *) NULL);
#else /* P95 */
				  PointerGetDatum(aggtransfn1Name),
				  Int32GetDatum(2),
				  PointerGetDatum(fnArgs),
				  0);
#endif /* P95 */
	if(!HeapTupleIsValid(tup))
#ifndef P95
	    elog(WARN, "AggregateDefine: \"%-.*s\"(\"%-.*s\", \"%-.*s\",) does not exist",
		 sizeof(NameData), aggtransfn1Name->data,
		 sizeof(NameData), aggtransfn1typeName->data,
		 sizeof(NameData), aggbasetypeName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: '%s('%s', '%s') does not exist",
		 aggtransfn1Name,  aggtransfn1typeName, aggbasetypeName);
#endif /* P95 */
	if (((Form_pg_proc) GETSTRUCT(tup))->prorettype != xret1)
#ifndef P95
	    elog(WARN, "AggregateDefine: return type of \"%-.*s\" is not \"%-.*s\"",
		 sizeof(NameData), aggtransfn1Name->data,
		 sizeof(NameData), aggtransfn1typeName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: return type of '%s' is not '%s'",
		 aggtransfn1Name,
		 aggtransfn1typeName);
#endif /* P95 */
	xfn1 = tup->t_oid;
	if (!OidIsValid(xfn1) || !OidIsValid(xret1) ||
	    !OidIsValid(xbase))
#ifndef P95
	    elog(WARN, "AggregateDefine: bogus function \"%-.*s\"",
		 sizeof(NameData), aggfinalfnName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: bogus function '%s'", aggfinalfnName);
#endif /* P95 */
    }
    
#ifndef P95
    if (NameIsValid(aggtransfn2Name)) {
	tup = SearchSysCacheTuple(TYPNAME, (char *) aggtransfn2typeName,
				  (char *) NULL, (char *) NULL, (char *) NULL);
#else /* P95 */
    if (aggtransfn2Name) {
	tup = SearchSysCacheTuple(TYPNAME, 
				  PointerGetDatum(aggtransfn2typeName),
				  0,0,0);
#endif /* P95 */
	if(!HeapTupleIsValid(tup))
#ifndef P95
	    elog(WARN, "AggregateDefine: Type \"%-.*s\" undefined",
		 sizeof(NameData), aggtransfn2typeName);
#else /* P95 */
	    elog(WARN, "AggregateCreate: Type '%s' undefined",
		 aggtransfn2typeName);
#endif /* P95 */
	xret2 = tup->t_oid;
	
	fnArgs[0] = xret2;
	fnArgs[1] = 0;
#ifndef P95
	tup = SearchSysCacheTuple(PRONAME, (char *) aggtransfn2Name->data,
				  (char *) Int32GetDatum(1),
				  (char *) fnArgs,
				  (char *) NULL);
#else /* P95 */
	tup = SearchSysCacheTuple(PRONAME, 
				  PointerGetDatum(aggtransfn2Name),
				  Int32GetDatum(1),
				  PointerGetDatum(fnArgs),
				  0);
#endif /* P95 */
	if(!HeapTupleIsValid(tup))
#ifndef P95
	    elog(WARN, "AggregateDefine: \"%-.*s\"(\"%-.*s\") does not exist",
		 sizeof(NameData), aggtransfn2Name->data,
		 sizeof(NameData), aggtransfn2typeName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: '%s'('%s') does not exist",
		 aggtransfn2Name, aggtransfn2typeName);
#endif /* P95 */
	if (((Form_pg_proc) GETSTRUCT(tup))->prorettype != xret2)
#ifndef P95
	    elog(WARN, "AggregateDefine: return type of \"%-.*s\" is not \"%-.*s\"",
		 sizeof(NameData), aggtransfn2Name->data,
		 sizeof(NameData), aggtransfn2typeName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: return type of '%s' is not '%s'",
		 aggtransfn2Name, aggtransfn2typeName);
#endif /* P95 */
	xfn2 = tup->t_oid;
	if (!OidIsValid(xfn2) || !OidIsValid(xret2))
#ifndef P95
	    elog(WARN, "AggregateDefine: bogus function \"%-.*s\"",
		 sizeof(NameData), aggfinalfnName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: bogus function '%s'",aggfinalfnName);
#endif /* P95 */
    }
    
#ifdef P95
    tup = SearchSysCacheTuple(AGGNAME, PointerGetDatum(aggName),
			      ObjectIdGetDatum(xbase),  
			      0,0);
    if (HeapTupleIsValid(tup))
	elog(WARN, 
	     "AggregateCreate: aggregate '%s' with base type '%s' already exists",
	     aggName, aggbasetypeName);

#endif /* P95 */
    /* more sanity checks */
#ifndef P95
    if (NameIsValid(aggtransfn1Name) && NameIsValid(aggtransfn2Name) &&
	!NameIsValid(aggfinalfnName))
	elog(WARN, "AggregateDefine: Aggregate must have final function with both transition functions");
#else /* P95 */
    if (aggtransfn1Name && aggtransfn2Name && !aggfinalfnName)
	elog(WARN, "AggregateCreate: Aggregate must have final function with both transition functions");
#endif /* P95 */
    
#ifndef P95
    if ((!NameIsValid(aggtransfn1Name) || !NameIsValid(aggtransfn2Name)) &&
	NameIsValid(aggfinalfnName))
	elog(WARN, "AggregateDefine: Aggregate cannot have final function without both transition functions");
#else /* P95 */
    if ((!aggtransfn1Name || !aggtransfn2Name) && aggfinalfnName)
	elog(WARN, "AggregateCreate: Aggregate cannot have final function without both transition functions");
#endif /* P95 */
    
#ifndef P95
    if (NameIsValid(aggfinalfnName)) {
#else /* P95 */
    if (aggfinalfnName) {
#endif /* P95 */
        fnArgs[0] = xret1;
	fnArgs[1] = xret2;
	tup = SearchSysCacheTuple(PRONAME,
#ifndef P95
				  (char *) aggfinalfnName->data,
				  (char *) Int32GetDatum(2),
				  (char *) fnArgs,
				  (char *) NULL);
#else /* P95 */
				  PointerGetDatum(aggfinalfnName),
				  Int32GetDatum(2),
				  PointerGetDatum(fnArgs),
				  0);
#endif /* P95 */
	if(!HeapTupleIsValid(tup))
#ifndef P95
	    elog(WARN, "AggregateDefine: \"%-.*s\"(\"%-.*s\", \"%-.*s\") does not exist",
		 sizeof(NameData), aggfinalfnName->data,
		 sizeof(NameData), aggtransfn1typeName->data,
		 sizeof(NameData), aggtransfn2typeName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: '%s'('%s','%s') does not exist",
		 aggfinalfnName, aggtransfn1typeName, aggtransfn2typeName);
#endif /* P95 */
	ffn = tup->t_oid;
	proc = (Form_pg_proc) GETSTRUCT(tup);
	fret = proc->prorettype;
	if (!OidIsValid(ffn) || !OidIsValid(fret))
#ifndef P95
	    elog(WARN, "AggregateDefine: bogus function \"%-.*s\"",
		 sizeof(NameData), aggfinalfnName->data);
#else /* P95 */
	    elog(WARN, "AggregateCreate: bogus function '%s'", aggfinalfnName);
#endif /* P95 */
    }
    
    /*
     * If transition function 2 is defined, it must have an initial value,
     * whereas transition function 1 does not, which allows man and min
     * aggregates to return NULL if they are evaluated on empty sets.
     */
#ifndef P95
    if (OidIsValid(xfn2) && !PointerIsValid(agginitval2))
	elog(WARN, "AggregateDefine: transition function 2 MUST have an initial value");
#else /* P95 */
    if (OidIsValid(xfn2) && !agginitval2)
	elog(WARN, "AggregateCreate: transition function 2 MUST have an initial value");
#endif /* P95 */
    
    /* initialize nulls and values */
    for(i=0; i < Natts_pg_aggregate; i++) {
	nulls[i] = ' ';
	values[i] = (Datum)NULL;
    }
#ifndef P95
    values[Anum_pg_aggregate_aggname-1] = (Datum)aggName;
#else /* P95 */
    values[Anum_pg_aggregate_aggname-1] = PointerGetDatum(aggName);
#endif /* P95 */
    values[Anum_pg_aggregate_aggowner-1] =
	Int32GetDatum(GetUserId());
    values[Anum_pg_aggregate_aggtransfn1-1] =
	ObjectIdGetDatum(xfn1);
    values[Anum_pg_aggregate_aggtransfn2-1] =
	ObjectIdGetDatum(xfn2);
    values[Anum_pg_aggregate_aggfinalfn-1] =
	ObjectIdGetDatum(ffn);
    
#ifdef P95
    values[Anum_pg_aggregate_aggbasetype-1] =
	ObjectIdGetDatum(xbase);
#endif /* P95 */
    if (!OidIsValid(xfn1)) {
#ifndef P95
	values[Anum_pg_aggregate_aggbasetype-1] =
	    ObjectIdGetDatum(InvalidOid);
#endif /* P95 */
	values[Anum_pg_aggregate_aggtranstype1-1] =
	    ObjectIdGetDatum(InvalidOid);
	values[Anum_pg_aggregate_aggtranstype2-1] =
	    ObjectIdGetDatum(xret2);
	values[Anum_pg_aggregate_aggfinaltype-1] =
	    ObjectIdGetDatum(xret2);
    }
    else if (!OidIsValid(xfn2)) {
#ifndef P95
	values[Anum_pg_aggregate_aggbasetype-1] =
	    ObjectIdGetDatum(xbase);
#endif /* P95 */
	values[Anum_pg_aggregate_aggtranstype1-1] =
	    ObjectIdGetDatum(xret1);
	values[Anum_pg_aggregate_aggtranstype2-1] =
	    ObjectIdGetDatum(InvalidOid);
	values[Anum_pg_aggregate_aggfinaltype-1] =
	    ObjectIdGetDatum(xret1);
    }
    else {
#ifndef P95
	values[Anum_pg_aggregate_aggbasetype-1] =
	    ObjectIdGetDatum(xbase);
#endif /* P95 */
	values[Anum_pg_aggregate_aggtranstype1-1] =
	    ObjectIdGetDatum(xret1);
	values[Anum_pg_aggregate_aggtranstype2-1] =
	    ObjectIdGetDatum(xret2);
	values[Anum_pg_aggregate_aggfinaltype-1] =
	    ObjectIdGetDatum(fret);
    }
    
#ifndef P95
    if (OidIsValid(agginitval1))
	values[Anum_pg_aggregate_agginitval1-1] = (Datum)agginitval1;
#else /* P95 */
    if (agginitval1)
	values[Anum_pg_aggregate_agginitval1-1] = PointerGetDatum(textin(agginitval1));
#endif /* P95 */
    else
	nulls[Anum_pg_aggregate_agginitval1-1] = 'n';
    
#ifndef P95
    if (OidIsValid(agginitval2))
	values[Anum_pg_aggregate_agginitval2-1] = (Datum)agginitval2;
#else /* P95 */
    if (agginitval2)
	values[Anum_pg_aggregate_agginitval2-1] = PointerGetDatum(textin(agginitval2));
#endif /* P95 */
    else
	nulls[Anum_pg_aggregate_agginitval2-1] = 'n';
    
    if (!RelationIsValid(aggdesc = heap_openr(AggregateRelationName)))
#ifndef P95
	elog(WARN, "AggregateDefine: could not open \"%-.*s\"",
	     sizeof(NameData), AggregateRelationName);
    if (!HeapTupleIsValid(tup = heap_formtuple(Natts_pg_aggregate,
					       aggdesc->rd_att,
#else /* P95 */
	elog(WARN, "AggregateCreate: could not open '%s'",
	     AggregateRelationName);

    tupDesc = aggdesc->rd_att;
    if (!HeapTupleIsValid(tup = heap_formtuple(tupDesc,
#endif /* P95 */
					       values,
					       nulls)))
#ifndef P95
	elog(WARN, "AggregateDefine: heap_formtuple failed");
#else /* P95 */
	elog(WARN, "AggregateCreate: heap_formtuple failed");
#endif /* P95 */
    if (!OidIsValid(heap_insert(aggdesc, tup)))
#ifndef P95
	elog(WARN, "AggregateDefine: heap_insert failed");
#else /* P95 */
	elog(WARN, "AggregateCreate: heap_insert failed");
#endif /* P95 */
    heap_close(aggdesc);

}

char *
#ifndef P95
AggNameGetInitVal(char *aggName, int xfuncno, bool *isNull)
#else /* P95 */
AggNameGetInitVal(char *aggName, Oid basetype, int xfuncno, bool *isNull)
#endif /* P95 */
{
    HeapTuple	tup;
    Relation	aggRel;
    int		initValAttno;
    Oid	transtype;
    text	*textInitVal;
    char	*strInitVal, *initVal;
    extern char	*textout();
    
    Assert(PointerIsValid(aggName));
    Assert(PointerIsValid(isNull));
    Assert(xfuncno == 1 || xfuncno == 2);

#ifndef P95
    tup = SearchSysCacheTuple(AGGNAME, aggName, (char *) NULL,
			      (char *) NULL, (char *) NULL);
#else /* P95 */
    tup = SearchSysCacheTuple(AGGNAME, 
			      PointerGetDatum(aggName),
			      PointerGetDatum(basetype),
			      0,0);
#endif /* P95 */
    if (!HeapTupleIsValid(tup))
#ifndef P95
	elog(WARN, "AggNameGetInitVal: cache lookup failed for aggregate \"%-.*s\"",
	     sizeof(NameData), aggName);
#else /* P95 */
	elog(WARN, "AggNameGetInitVal: cache lookup failed for aggregate '%s'",
	     aggName);
#endif /* P95 */
    if (xfuncno == 1) {
	transtype = ((Form_pg_aggregate) GETSTRUCT(tup))->aggtranstype1;
	initValAttno = Anum_pg_aggregate_agginitval1;
    }
    else if (xfuncno == 2) {
	transtype = ((Form_pg_aggregate) GETSTRUCT(tup))->aggtranstype2;
	initValAttno = Anum_pg_aggregate_agginitval2;
    }
    
    aggRel = heap_openr(AggregateRelationName);
    if (!RelationIsValid(aggRel))
	elog(WARN, "AggNameGetInitVal: could not open \"%-.*s\"",
#ifndef P95
	     sizeof(NameData), AggregateRelationName->data);
#else /* P95 */
	     AggregateRelationName);
#endif /* P95 */
    /* 
     * must use fastgetattr in case one or other of the init values is NULL
     */
#ifndef P95
    textInitVal = (text *) fastgetattr(tup, initValAttno, aggRel->rd_att,
#else /* P95 */
    textInitVal = (text *) fastgetattr(tup, initValAttno, 
				       RelationGetTupleDescriptor(aggRel),
#endif /* P95 */
				       isNull);
    if (!PointerIsValid(textInitVal))
	*isNull = true;
    if (*isNull) {
	heap_close(aggRel);
	return((char *) NULL);
    }
    strInitVal = textout(textInitVal);
    heap_close(aggRel);
    
#ifndef P95
    tup = SearchSysCacheTuple(TYPOID, (char *) ObjectIdGetDatum(transtype),
			      (char *) NULL, (char *) NULL, (char *) NULL);
#else /* P95 */
    tup = SearchSysCacheTuple(TYPOID, ObjectIdGetDatum(transtype),
			      0,0,0);
#endif /* P95 */
    if (!HeapTupleIsValid(tup)) {
	pfree(strInitVal);
	elog(WARN, "AggNameGetInitVal: cache lookup failed on aggregate transition function return type");
    }
#ifndef P95
    initVal = fmgr(((TypeTupleForm) GETSTRUCT(tup))->typinput, strInitVal);
#else /* P95 */
    initVal = fmgr(((TypeTupleForm) GETSTRUCT(tup))->typinput, strInitVal, -1);
#endif /* P95 */
    pfree(strInitVal);
    return(initVal);
}
