/*-------------------------------------------------------------------------
 *
 * pg_contract.h--
 *    definition of the global system "contract" relation (pg_contract)
 *    Records contracted update streams for keeping copies at other sites
 *    up to date.
 *
 * Copyright (c) 1995, Regents of the University of California
 *
 * $Id: pg_contract.h,v 1.6 1996/05/31 17:13:04 jsidell Exp $
 *
 * NOTES
 *    the genbki.sh script reads this file and generates .bki
 *    information from the DATA() statements.
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_CONTRACT_H
#define PG_CONTRACT_H

/* ----------------
 *	postgres.h contains the system type definitions and the
 *	CATALOG(), BOOTSTRAP and DATA() sugar words so this file
 *	can be read by both genbki.sh and the C compiler.
 * ----------------
 */
#include "postgres.h"
#include "utils/nabstime.h"

/* ----------------
 *	pg_contract definition.  cpp turns this into
 *	typedef struct FormData_pg_contract
 *
 *	@@@ Note! typemask must always be initialized to 64.
 *
 *	@@@ Note! If you change the schema, also change the struct
 *          UpdateContractInfo in rpc/copies.x so that it is the same as
 *	    the new pg_contract schema.
 * ----------------
 */
CATALOG(pg_contract) BOOTSTRAP {
    Oid		contractid;	/* unique id, same for both send/recv ends */
    Oid		sendfrastoreid;	/* fragment that sends updates */
    Oid		recvfrastoreid; /* fragment that receives updates */
    Oid		sidefrastoreid; /* side-file store oid */
    int4	lastupdcommit;	/* commit time for last update send/recv */
    int4	otherloc;	/* location to send/from recv updates */
    int4	period;		/* update period of this contract */
    char	typemask;	/* bit-field for type of contract */
} FormData_pg_contract;

/* ----------------
 *	Form_pg_contract corresponds to a pointer to a tuple with
 *	the format of pg_contract relation.
 * ----------------
 */
typedef FormData_pg_contract		*Form_pg_contract;

/* ----------------
 *	compiler constants for pg_contract
 * ----------------
 */
#define Name_pg_contract			"pg_contract" 
#define Natts_pg_contract			8
#define Anum_pg_contract_contractid		1
#define Anum_pg_contract_sendfrastoreid		2
#define Anum_pg_contract_recvfrastoreid		3
#define Anum_pg_contract_sidefrastoreid		4
#define Anum_pg_contract_lastupdcommit		5
#define Anum_pg_contract_otherloc		6
#define Anum_pg_contract_period			7
#define Anum_pg_contract_typemask		8

/* ----------------
 *	pg_contract is initially empty
 * ----------------
 */

/* ----------------
 *	Access methods for pg_contract tuples.
 * ----------------
 */

/*
 * Information stored in typemask.
 */
#define CONTRACT_TYPE		0x01	/* User copy (0) or name service (1) */
#define	COPY_TYPE		0x02	/* Read-write (0) or read-only (1) */
#define DIRECTION		0x04	/* Sending (0) or receiving (1) */
#define NAMESERVICE_TYPE	0x08	/* Specific (0) or remote (1) */

/*
 * Access methods for information stored in typemask.
 */
#define ContractSetUserCopy(tuple) \
        ((Form_pg_contract) (tuple))->typemask &= ~CONTRACT_TYPE
#define ContractSetNameService(tuple) \
	((Form_pg_contract) (tuple))->typemask |= CONTRACT_TYPE
#define ContractSetReadWrite(tuple) \
        ((Form_pg_contract) (tuple))->typemask &= ~COPY_TYPE
#define ContractSetReadOnly(tuple) \
	((Form_pg_contract) (tuple))->typemask |= COPY_TYPE
#define ContractSetSend(tuple) \
        ((Form_pg_contract) (tuple))->typemask &= ~DIRECTION
#define ContractSetReceive(tuple) \
	((Form_pg_contract) (tuple))->typemask |= DIRECTION
#define ContractSetNSSpecific(tuple) \
        ((Form_pg_contract) (tuple))->typemask &= ~NAMESERVICE_TYPE
#define ContractSetNSRemote(tuple) \
	((Form_pg_contract) (tuple))->typemask |= NAMESERVICE_TYPE

#define ContractIsUserCopy(tuple) \
        (!(((Form_pg_contract) (tuple))->typemask & CONTRACT_TYPE))
#define ContractIsNameService(tuple) \
	(((Form_pg_contract) (tuple))->typemask & CONTRACT_TYPE)
#define ContractIsReadWrite(tuple) \
        (!(((Form_pg_contract) (tuple))->typemask & COPY_TYPE))
#define ContractIsReadOnly(tuple) \
	(((Form_pg_contract) (tuple))->typemask & COPY_TYPE)
#define ContractIsSend(tuple) \
        (!(((Form_pg_contract) (tuple))->typemask & DIRECTION))
#define ContractIsReceive(tuple) \
	(((Form_pg_contract) (tuple))->typemask & DIRECTION)
#define ContractIsNSSpecific(tuple) \
        (!(((Form_pg_contract) (tuple))->typemask & NAMESERVICE_TYPE))
#define ContractIsNSRemote(tuple) \
	(((Form_pg_contract) (tuple))->typemask & NAMESERVICE_TYPE)

/*
 * Pretty-print a pg_contract tuple.
 */

#define PrintContract(ref, contract) \
        printf("%s:\n\tcontractid %d\n\tsendfrastoreid %d\n\trecvfrastoreid \
%d\n\tsidefrastoreid %d\n\tlastupdcommit %d\n\totherloc %d\n\tperiod \
%d\n\ttypemask %d\n\tContract Type: %s\n\tName Service Type: %s\n\tCopy Type: \
%s\n\tDirection: %s\n", \
	ref, \
	((Form_pg_contract) (contract))->contractid, \
	((Form_pg_contract) (contract))->sendfrastoreid, \
	((Form_pg_contract) (contract))->recvfrastoreid, \
	((Form_pg_contract) (contract))->sidefrastoreid, \
	((Form_pg_contract) (contract))->lastupdcommit, \
	((Form_pg_contract) (contract))->otherloc, \
	((Form_pg_contract) (contract))->period, \
	((Form_pg_contract) (contract))->typemask, \
	ContractIsNameService(contract) ? "Name Service" : "User Copy", \
	ContractIsNSSpecific(contract) ? "Specific" : "Remote", \
	ContractIsReadOnly(contract) ? "Read-only" : "Read-write", \
	ContractIsReceive(contract) ? "Receive" : "Send")

#endif /* PG_CONTRACT_H */
