/*-------------------------------------------------------------------------
 *
 * pg_proc.c--
 *    routines to support manipulation of the pg_proc relation
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/catalog/pg_proc.c,v 0.5 1996/06/07 04:55:07 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <string.h>
#include "postgres.h"

#include "access/heapam.h"
#include "access/relscan.h"
#include "access/skey.h"
#include "utils/rel.h"
#include "fmgr.h"
#include "utils/elog.h"
#include "utils/builtins.h"
#include "utils/sets.h"

#include "nodes/pg_list.h"

#include "catalog/catname.h"
#include "utils/syscache.h"
#include "catalog/pg_proc.h"
#include "catalog/indexing.h"
#include "tcop/dest.h"
#include "parser/parse_query.h"
#include "tcop/tcopprot.h"
#include "catalog/pg_type.h"
#include "parser/catalog_utils.h"
#include "utils/lsyscache.h"
#include "optimizer/internal.h"
#include "optimizer/planner.h"

/* ----------------------------------------------------------------
 *	ProcedureDefine
 * ----------------------------------------------------------------
 */
Oid
#ifndef P95
ProcedureDefine(Name procedureName,
#else /* P95 */
ProcedureCreate(char *procedureName,
#endif /* P95 */
		bool returnsSet,
#ifndef P95
		Name returnTypeName,	
		Name languageName,
#else /* P95 */
		char *returnTypeName,	
		char *languageName,
#endif /* P95 */
		char *prosrc,
		char *probin,
		bool canCache,
		bool trusted,
		int32 byte_pct,
		int32 perbyte_cpu,
		int32 percall_cpu,
		int32 outin_ratio,
		List *argList,
		CommandDest dest)
{
    register	i;
    Relation 	rdesc;
    HeapTuple 	tup;
    bool        defined;
    uint16 	parameterCount;
    char	nulls[ Natts_pg_proc ];
    Datum 	values[ Natts_pg_proc ];
    Oid 	languageObjectId;
    Oid		typeObjectId;
    List 	*x;
    QueryTreeList *querytree_list;
    List	*plan_list;
    Oid		typev[8];
    Oid 	relid;
    Oid 	toid;
    text 	*prosrctext;
#ifdef P95
    TupleDesc   tupDesc;
#endif /* P95 */
    
    /* ----------------
     *	sanity checks
     * ----------------
     */
#ifndef P95
    Assert(NameIsValid(procedureName));
    Assert(NameIsValid(returnTypeName));
    Assert(NameIsValid(languageName));
#endif /* P95 */
    Assert(PointerIsValid(prosrc));
    Assert(PointerIsValid(probin));
    
    parameterCount = 0;
    memset(typev, 0, 8 * sizeof(Oid));
    foreach (x, argList) {
	Value *t = lfirst(x);
	
	if (parameterCount == 8)
	    elog(WARN, "Procedures cannot take more than 8 arguments");
	
	if (strcmp(strVal(t), "opaque") == 0) {
#ifndef P95
	    if (namestrcmp(languageName, "sql") == 0) {
#else /* P95 */
	    if (strcmp(languageName, "sql") == 0) {
#endif /* P95 */
		elog(WARN, "ProcedureDefine: sql functions cannot take type \"opaque\"");
	    }
	    else
		toid = 0;
	} else {
#ifndef P95
	    toid = TypeGet((Name)strVal(t), &defined);
#else /* P95 */
	    toid = TypeGet(strVal(t), &defined);
#endif /* P95 */
	    
	    if (!OidIsValid(toid)) {
#ifndef P95
		elog(WARN, "ProcedureDefine: arg type '%s' is not defined",
#else /* P95 */
		elog(WARN, "ProcedureCreate: arg type '%s' is not defined",
#endif /* P95 */
		     strVal(t));
	    }
	    
	    if (!defined) {
#ifndef P95
		elog(NOTICE, "ProcedureDefine: arg type '%s' is only a shell",
#else /* P95 */
		elog(NOTICE, "ProcedureCreate: arg type '%s' is only a shell",
#endif /* P95 */
		     strVal(t));
	    }
	}
	
	typev[parameterCount++] = toid;
    }
    
    tup = SearchSysCacheTuple(PRONAME,
#ifndef P95
			      (char *) procedureName,
			      (char *) UInt16GetDatum(parameterCount),
			      (char *) typev,
			      (char *) NULL);
#else /* P95 */
			      PointerGetDatum(procedureName),
			      UInt16GetDatum(parameterCount),
			      PointerGetDatum(typev),
			      0);
#endif /* P95 */
    
    if (HeapTupleIsValid(tup))
#ifndef P95
	elog(WARN, "ProcedureDefine: procedure %s already exists with same arguments",
#else /* P95 */
	elog(WARN, "ProcedureCreate: procedure %s already exists with same arguments",
#endif /* P95 */
	     procedureName);
    
#ifndef P95
    if (!namestrcmp(languageName, "sql"))  {
#else /* P95 */
    if (!strcmp(languageName, "sql"))  {
#endif /* P95 */
	/* If this call is defining a set, check if the set is already
	 * defined by looking to see whether this call's function text
	 * matches a function already in pg_proc.  If so just return the 
	 * OID of the existing set.
	 */
#ifndef P95
	if (!namestrcmp(procedureName, GENERICSETNAME)) {
#else /* P95 */
	if (!strcmp(procedureName, GENERICSETNAME)) {
#endif /* P95 */
	    prosrctext = textin(prosrc);
	    tup = SearchSysCacheTuple(PROSRC,
#ifndef P95
				      (char *) prosrctext,
				      (char *) NULL,
				      (char *) NULL,
				      (char *) NULL);
#else /* P95 */
				      PointerGetDatum(prosrctext),
				      0,0,0);
#endif /* P95 */
	    if (HeapTupleIsValid(tup))
		return tup->t_oid;
	}
    }
    
    tup = SearchSysCacheTuple(LANNAME,
#ifndef P95
			      (char *) languageName,
			      (char *) NULL,
			      (char *) NULL,
			      (char *) NULL);
#else /* P95 */
			      PointerGetDatum(languageName),
			      0,0,0);
#endif /* P95 */
    
    if (!HeapTupleIsValid(tup))
#ifndef P95
	elog(WARN, "ProcedureDefine: no such language %s",
#else /* P95 */
	elog(WARN, "ProcedureCreate: no such language %s",
#endif /* P95 */
	     languageName);
    
    languageObjectId = tup->t_oid;
    
#ifndef P95
    if (namestrcmp(returnTypeName, "opaque") == 0) {
	if (namestrcmp(languageName, "sql") == 0) {
	    elog(WARN, "ProcedureDefine: sql functions cannot return type \"opaque\"");
#else /* P95 */
    if (strcmp(returnTypeName, "opaque") == 0) {
	if (strcmp(languageName, "sql") == 0) {
	    elog(WARN, "ProcedureCreate: sql functions cannot return type \"opaque\"");
#endif /* P95 */
	}
	else
	    typeObjectId = 0;
    }
    
    else {
	typeObjectId = TypeGet(returnTypeName, &defined);
	
	if (!OidIsValid(typeObjectId)) {
#ifndef P95
	    elog(NOTICE, "ProcedureDefine: type '%s' is not yet defined",
#else /* P95 */
	    elog(NOTICE, "ProcedureCreate: type '%s' is not yet defined",
#endif /* P95 */
		 returnTypeName);
#if 0
#ifndef P95
	    elog(NOTICE, "ProcedureDefine: creating a shell for type '%s'",
#else /* P95 */
	    elog(NOTICE, "ProcedureCreate: creating a shell for type '%s'",
#endif /* P95 */
		 returnTypeName);
#endif	    
	    typeObjectId = TypeShellMake(returnTypeName);
	    if (!OidIsValid(typeObjectId)) {
#ifndef P95
		elog(WARN, "ProcedureDefine: could not create type '%s'",
#else /* P95 */
		elog(WARN, "ProcedureCreate: could not create type '%s'",
#endif /* P95 */
		     returnTypeName);
	    }
	}
	
	else if (!defined) {
#ifndef P95
	    elog(NOTICE, "ProcedureDefine: return type '%s' is only a shell",
#else /* P95 */
	    elog(NOTICE, "ProcedureCreate: return type '%s' is only a shell",
#endif /* P95 */
		 returnTypeName);
	}
    }
    
    /* don't allow functions of complex types that have the same name as
       existing attributes of the type */
    if (parameterCount == 1 && 
#ifndef P95
	(toid = TypeGet((Name)strVal(lfirst(argList)), &defined)) &&
#else /* P95 */
	(toid = TypeGet(strVal(lfirst(argList)), &defined)) &&
#endif /* P95 */
	defined &&
	(relid = typeid_get_relid(toid)) != 0 &&
	get_attnum(relid, procedureName) != InvalidAttrNumber)
	elog(WARN, "method %s already an attribute of type %s",
	     procedureName, strVal(lfirst(argList)));
    
    
    /*
     *  If this is a postquel procedure, we parse it here in order to
     *  be sure that it contains no syntax errors.  We should store
     *  the plan in an Inversion file for use later, but for now, we
     *  just store the procedure's text in the prosrc attribute.
     */
    
#ifndef P95
    if (namestrcmp(languageName, "sql") == 0) {
#else /* P95 */
    if (strcmp(languageName, "sql") == 0) {
#endif /* P95 */
	plan_list = pg_plan(prosrc, typev, parameterCount,
			    &querytree_list, dest);
	
	/* typecheck return value */
	pg_checkretval(typeObjectId, querytree_list);
    }
    
    for (i = 0; i < Natts_pg_proc; ++i) {
	nulls[i] = ' ';
	values[i] = (Datum)NULL;
    }
    
    i = 0;
#ifndef P95
    values[i++] = (Datum) procedureName;
#else /* P95 */
    values[i++] = PointerGetDatum(procedureName);
#endif /* P95 */
    values[i++] =  Int32GetDatum(GetUserId());
    values[i++] =  ObjectIdGetDatum(languageObjectId);
    
    /* XXX isinherited is always false for now */
    
    values[i++] = Int8GetDatum((bool) 0);
    
    /* XXX istrusted is always false for now */
    
    values[i++] =  Int8GetDatum(trusted);
    values[i++] =  Int8GetDatum(canCache);
    values[i++] =  UInt16GetDatum(parameterCount);
    values[i++] =  Int8GetDatum(returnsSet);
    values[i++] =  ObjectIdGetDatum(typeObjectId);
    
    values[i++] = (Datum) typev;
    /*
     * The following assignments of constants are made.  The real values
     * will have to be extracted from the arglist someday soon.
     */
    values[i++] =  Int32GetDatum(byte_pct); /* probyte_pct */
    values[i++] =  Int32GetDatum(perbyte_cpu); /* properbyte_cpu */
    values[i++] =  Int32GetDatum(percall_cpu); /* propercall_cpu */
    values[i++] =  Int32GetDatum(outin_ratio); /* prooutin_ratio */
    
    values[i++] = (Datum)fmgr(TextInRegProcedure, prosrc);	/* prosrc */
    values[i++] = (Datum)fmgr(TextInRegProcedure, probin);   /* probin */
    
    rdesc = heap_openr(ProcedureRelationName);
#ifndef P95
    tup = heap_formtuple(Natts_pg_proc,
			rdesc->rd_att,
			values,
			nulls);
#else /* P95 */
    
    tupDesc = rdesc->rd_att;
    tup = heap_formtuple(tupDesc,
			 values,
			 nulls);
#endif /* P95 */
    
    heap_insert(rdesc, tup);
    
    if (RelationGetRelationTupleForm(rdesc)->relhasindex)
	{
	    Relation idescs[Num_pg_proc_indices];
	    
	    CatalogOpenIndices(Num_pg_proc_indices, Name_pg_proc_indices, idescs);
	    CatalogIndexInsert(idescs, Num_pg_proc_indices, rdesc, tup);
	    CatalogCloseIndices(Num_pg_proc_indices, idescs);
	}
    heap_close(rdesc);
    return tup->t_oid;
}

