/*-------------------------------------------------------------------------
 *
 * pg_type.c--
 *    routines to support manipulation of the pg_type relation
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/catalog/pg_type.c,v 0.6 1996/06/07 04:55:12 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <string.h>
#include "postgres.h"

#include "access/heapam.h"
#include "access/relscan.h"
#include "access/skey.h"
#ifdef P95
#include "access/tupdesc.h"
#endif /* P95 */
#include "utils/builtins.h"
#include "utils/rel.h"
#include "utils/palloc.h"
#include "fmgr.h"
#include "utils/elog.h"
#include "parser/catalog_utils.h"

#include "catalog/catname.h"
#include "utils/syscache.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "catalog/indexing.h"
#include "storage/lmgr.h"

/* ----------------------------------------------------------------
 * 	TypeGetWithOpenRelation
 *
 *	preforms a scan on pg_type for a type tuple with the
 *	given type name.
 * ----------------------------------------------------------------
 *	pg_type_desc		 -- reldesc for pg_type
 *	typeName		 -- name of type to be fetched 
 *	defined			 -- has the type been defined? 
 */
#ifndef P95
Oid
#else /* P95 */
static Oid
#endif /* P95 */
TypeGetWithOpenRelation(Relation pg_type_desc,
#ifndef P95
			Name typeName, 
#else /* P95 */
			char* typeName, 
#endif /* P95 */
			bool *defined)
{
    HeapScanDesc	scan;
    HeapTuple		tup;
    
    static ScanKeyData	typeKey[1] = {
	{ 0, Anum_pg_type_typname, NameEqualRegProcedure }
    };
    
    /* ----------------
     *	initialize the scan key and begin a scan of pg_type
     * ----------------
     */
    fmgr_info(NameEqualRegProcedure,
	      &typeKey[0].sk_func, &typeKey[0].sk_nargs);
#ifndef P95
    typeKey[0].sk_argument = NameGetDatum(typeName);
#else /* P95 */
    typeKey[0].sk_argument = PointerGetDatum(typeName);
#endif /* P95 */
    
    scan = heap_beginscan(pg_type_desc,
			  0,
			  SelfTimeQual,
			  1,
			  typeKey);
    
    /* ----------------
     *	get the type tuple, if it exists.
     * ----------------
     */
    tup = heap_getnext(scan, 0, (Buffer *) 0);
    
    /* ----------------
     *	if no type tuple exists for the given type name, then
     *  end the scan and return appropriate information.
     * ----------------
     */
    if (! HeapTupleIsValid(tup)) {
	heap_endscan(scan);
	*defined = false;
	return InvalidOid;
    }
    
    /* ----------------
     *	here, the type tuple does exist so we pull information from
     *  the typisdefined field of the tuple and return the tuple's
     *  oid, which is the oid of the type.
     * ----------------
     */
    heap_endscan(scan);
    *defined = (bool) ((TypeTupleForm) GETSTRUCT(tup))->typisdefined;
    
    return
	tup->t_oid;
}

/* ----------------------------------------------------------------
 * 	TypeGet
 *
 *	Finds the ObjectId of a type, even if uncommitted; "defined"
 *	is only set if the type has actually been defined, i.e., if
 *	the type tuple is not a shell.
 *
 *	Note: the meat of this function is now in the function
 *	      TypeGetWithOpenRelation().  -cim 6/15/90
 *
 *	Also called from util/remove.c
 * ----------------------------------------------------------------
 */
Oid
#ifndef P95
TypeGet(Name typeName,		/* name of type to be fetched */
#else /* P95 */
TypeGet(char* typeName,		/* name of type to be fetched */
#endif /* P95 */
	bool *defined)		/* has the type been defined? */
{
    Relation	pg_type_desc;
    Oid		typeoid;
    
#ifndef P95
    /* ----------------
     *	sanity checks
     * ----------------
     */
    Assert(NameIsValid(typeName));
    Assert(PointerIsValid(defined));
    
#endif /* P95 */
    /* ----------------
     *	open the pg_type relation
     * ----------------
     */
    pg_type_desc = heap_openr(TypeRelationName);
    
    /* ----------------
     *	scan the type relation for the information we want
     * ----------------
     */
    typeoid = TypeGetWithOpenRelation(pg_type_desc,
				      typeName,
				      defined);
    
    /* ----------------
     *	close the type relation and return the type oid.
     * ----------------
     */
    heap_close(pg_type_desc);
    
    return
	typeoid;
}

/* ----------------------------------------------------------------
 *	TypeShellMakeWithOpenRelation
 *
 * ----------------------------------------------------------------
 */
Oid
#ifndef P95
TypeShellMakeWithOpenRelation(Relation pg_type_desc, Name typeName)
#else /* P95 */
TypeShellMakeWithOpenRelation(Relation pg_type_desc, char *typeName)
#endif /* P95 */
{
    register int 	i;
    HeapTuple 		tup;
    Datum               values[ Natts_pg_type ];
    char		nulls[ Natts_pg_type ];
    Oid			typoid;
#ifdef P95
    TupleDesc           tupDesc;
#endif /* P95 */
    
    /* ----------------
     *	initialize our nulls[] and values[] arrays
     * ----------------
     */
    for (i = 0; i < Natts_pg_type; ++i) {
	nulls[i] = ' ';
	values[i] = (Datum)NULL; 	/* redundant, but safe */
    }
    
    /* ----------------
     *	initialize values[] with the type name and 
     * ----------------
     */
    i = 0;
    values[i++] = (Datum) typeName;	/* 1 */
    values[i++] = (Datum) InvalidOid;	/* 2 */
    values[i++] = (Datum) (int16) 0;	/* 3 */
    values[i++] = (Datum) (int16) 0;	/* 4 */
    values[i++] = (Datum) (bool) 0;	/* 5 */
    values[i++] = (Datum) (bool) 0;	/* 6 */
    values[i++] = (Datum) (bool) 0;	/* 7 */
    values[i++] = (Datum) (bool) 0;	/* 8 */
    values[i++] = (Datum) InvalidOid;	/* 9 */
    values[i++] = (Datum) InvalidOid;	/* 10 */
    values[i++] = (Datum) InvalidOid;	/* 11 */
    values[i++] = (Datum) InvalidOid;	/* 12 */
    values[i++] = (Datum) InvalidOid;	/* 13 */
    values[i++] = (Datum) InvalidOid;	/* 14 */
    values[i++] = (Datum) 'i';		/* 15 */

    /*
     * ... and fill typdefault with a bogus value
     */
    values[i++] =
#ifndef P95
	(Datum)fmgr(TextInRegProcedure, (char *)typeName);	/* 15 */
#else /* P95 */
	(Datum)fmgr(TextInRegProcedure, typeName);	/* 15 */
#endif /* P95 */
    
    /* ----------------
     *	create a new type tuple with FormHeapTuple
     * ----------------
     */
#ifndef P95
    tup = heap_formtuple(Natts_pg_type,
			 pg_type_desc->rd_att,
			 values,
			 nulls);
#else /* P95 */
    tupDesc = pg_type_desc->rd_att;

    tup = heap_formtuple(tupDesc, values, nulls);
#endif /* P95 */
    
    /* ----------------
     *	insert the tuple in the relation and get the tuple's oid.
     * ----------------
     */
    heap_insert(pg_type_desc, tup);
    typoid = tup->t_oid;
    
    if (RelationGetRelationTupleForm(pg_type_desc)->relhasindex)
	{
	    Relation idescs[Num_pg_type_indices];
	    
	    CatalogOpenIndices(Num_pg_type_indices, Name_pg_type_indices, idescs);
	    CatalogIndexInsert(idescs, Num_pg_type_indices, pg_type_desc, tup);
	    CatalogCloseIndices(Num_pg_type_indices, idescs);
	}
    /* ----------------
     *	free the tuple and return the type-oid
     * ----------------
     */
    pfree(tup);
    
    return
	typoid;   
}

/* ----------------------------------------------------------------
 *	TypeShellMake
 *
 *	This procedure inserts a "shell" tuple into the type
 *	relation.  The type tuple inserted has invalid values
 *	and in particular, the "typisdefined" field is false.
 *
 *	This is used so that a tuple exists in the catalogs.
 *	The invalid fields should be fixed up sometime after
 *	this routine is called, and then the "typeisdefined"
 *	field is set to true. -cim 6/15/90
 * ----------------------------------------------------------------
 */
Oid
#ifndef P95
TypeShellMake(Name typeName)
#else /* P95 */
TypeShellMake(char *typeName)
#endif /* P95 */
{
    Relation 	pg_type_desc;
    Oid	typoid;
    
    Assert(PointerIsValid(typeName));
    
    /* ----------------
     *	open pg_type
     * ----------------
     */
    pg_type_desc = heap_openr(TypeRelationName);
    
    /* ----------------
     *	insert the shell tuple
     * ----------------
     */
#ifndef P95
    typoid = TypeShellMakeWithOpenRelation(pg_type_desc,
					   typeName);
#else /* P95 */
    typoid = TypeShellMakeWithOpenRelation(pg_type_desc, typeName);
#endif /* P95 */
    
    /* ----------------
     *	close pg_type and return the tuple's oid.
     * ----------------
     */
    heap_close(pg_type_desc);
    
    return
	typoid;
}

/* ----------------------------------------------------------------
 * [Mariposa]
 *	TypeDefine
 * [Postgres95]
 *	TypeCreate
 *
 *	This does all the necessary work needed to define a new type.
 * ----------------------------------------------------------------
 */
Oid
#ifndef P95
TypeDefine(Name typeName,
#else /* P95 */
TypeCreate(char *typeName,
#endif /* P95 */
	   Oid relationOid,	/* only for 'c'atalog typeTypes */
	   int16 internalSize,
	   int16 externalSize,
	   char typeType,
	   char typDelim,
#ifndef P95
	   Name inputProcedure,
	   Name outputProcedure,
	   Name sendProcedure,
	   Name receiveProcedure,
	   Name elementTypeName,
#else /* P95 */
	   char *inputProcedure,
	   char *outputProcedure,
	   char *sendProcedure,
	   char *receiveProcedure,
	   char *elementTypeName,
#endif /* P95 */
	   char *defaultTypeValue, /* internal rep */
	   bool	passedByValue,
	   char alignment)
{
    register		i, j;
    Relation 		pg_type_desc;
    HeapScanDesc 	pg_type_scan;
    
    Oid            typeObjectId;
    Oid            elementObjectId = InvalidOid;
    
    HeapTuple 		tup;
    char 		nulls[Natts_pg_type];
    char	 	replaces[Natts_pg_type];
    Datum               values[Natts_pg_type];
    
    Buffer		buffer;
#ifndef P95
    NameData 		procname;
    Name		procs[4];
#else /* P95 */
    char               *procname;
    char               *procs[4];
#endif /* P95 */
    bool 		defined;
    ItemPointerData	itemPointerData;
#ifdef P95
    TupleDesc           tupDesc;
#endif /* P95 */

    Oid argList[8];

    
    static ScanKeyData	typeKey[1] = {
	{ 0, Anum_pg_type_typname, NameEqualRegProcedure }
    };
    
#ifndef P95
    /* ----------------
     *	sanity checks
     * ----------------
     */
    Assert(NameIsValid(typeName));  
    Assert(NameIsValid(inputProcedure) && NameIsValid(outputProcedure));
    
#endif /* P95 */
    fmgr_info(NameEqualRegProcedure,
	      &typeKey[0].sk_func, &typeKey[0].sk_nargs);
    
    /* ----------------
     *	check that the type is not already defined.
     * ----------------
     */
    typeObjectId = TypeGet(typeName, &defined);
    if (OidIsValid(typeObjectId) && defined) {
#ifndef P95
	elog(WARN, "TypeDefine: type %s already defined", typeName);
#else /* P95 */
	elog(WARN, "TypeCreate: type %s already defined", typeName);
#endif /* P95 */
    }
    
    /* ----------------
     *	if this type has an associated elementType, then we check that
     *  it is defined.
     * ----------------
     */
#ifndef P95
    if (elementTypeName != NULL) {
#else /* P95 */
    if (elementTypeName) {
#endif /* P95 */
	elementObjectId = TypeGet(elementTypeName, &defined);
    	if (!defined) {
#ifndef P95
	    elog(WARN, "TypeDefine: type %s is not defined", elementTypeName);
#else /* P95 */
	    elog(WARN, "TypeCreate: type %s is not defined", elementTypeName);
#endif /* P95 */
	}
    }
    
    /* ----------------
     *	XXX comment me
     * ----------------
     */
    if (externalSize == 0) {
	externalSize = -1;		/* variable length */
    }
    
    /* ----------------
     *	initialize arrays needed by FormHeapTuple
     * ----------------
     */
    for (i = 0; i < Natts_pg_type; ++i) {
	nulls[i] = ' ';
	replaces[i] = 'r';
	values[i] = (Datum)NULL; 	/* redundant, but nice */
    }
    
    /*
     * XXX
     *
     * Do this so that user-defined types have size -1 instead of zero if
     * they are variable-length - this is so that everything else in the
     * backend works.
     */
    
    if (internalSize == 0)
	internalSize = -1; 
    
    /* ----------------
     *	initialize the values[] information
     * ----------------
     */
    i = 0;
#ifndef P95
    values[i++] = (Datum) typeName;	/* 1 */
#else /* P95 */
    values[i++] = PointerGetDatum(typeName);	/* 1 */
#endif /* P95 */
    values[i++] = (Datum) GetUserId();	/* 2 */
    values[i++] = (Datum) internalSize;	/* 3 */
    values[i++] = (Datum) externalSize;	/* 4 */
    values[i++] = (Datum) passedByValue; /* 5 */
    values[i++] = (Datum) typeType; 	/* 6 */
    values[i++] = (Datum) (bool) 1; 	/* 7 */
    values[i++] = (Datum) typDelim; 	/* 8 */
    values[i++] = (Datum) (typeType == 'c' ? relationOid : InvalidOid); /* 9 */
    values[i++] = (Datum) elementObjectId; /* 10 */
    
    /*
     * arguments to type input and output functions must be 0
     */
    memset(argList, 0, 8 * sizeof(Oid)); 
    
    procs[0] = inputProcedure;
    procs[1] = outputProcedure;
#ifndef P95
    procs[2] = NameIsValid(receiveProcedure)
	? receiveProcedure : inputProcedure;
    procs[3] = NameIsValid(sendProcedure)
	? sendProcedure : outputProcedure;
#else /* P95 */
    procs[2] = (receiveProcedure) ? receiveProcedure : inputProcedure;
    procs[3] = (sendProcedure) ? sendProcedure : outputProcedure;
#endif /* P95 */
    
    for (j = 0; j < 4; ++j) {
#ifndef P95
        namecpy(&procname,procs[j]);
#else /* P95 */
	procname = procs[j];
#endif /* P95 */
	
	tup = SearchSysCacheTuple(PRONAME,
#ifndef P95
				  procname.data,
				  (char*)1,
				  (char*)argList,
				  (char *) NULL);
#else /* P95 */
				  PointerGetDatum(procname),
				  Int32GetDatum(1),
				  PointerGetDatum(argList),
				  0);
#endif /* P95 */
	
	if (!HeapTupleIsValid(tup)) {
	    /*
	     * it is possible for the input/output procedure
	     * to take two arguments, where the second argument
	     * is the element type (eg array_in/array_out)
	     */
	    if (OidIsValid(elementObjectId)) {
		tup = SearchSysCacheTuple(PRONAME,
#ifndef P95
					  procname.data,
					  (char*)2,
					  (char*)argList, 
					  (char *) NULL);
#else /* P95 */
					  PointerGetDatum(procname),
					  Int32GetDatum(2),
					  PointerGetDatum(argList), 
					  0);
#endif /* P95 */
	    }
	    if (!HeapTupleIsValid(tup)) {
#ifndef P95
		func_error("TypeDefine", procname.data, 1, (int*)argList);
#else /* P95 */
		func_error("TypeCreate", procname, 1, (int*)argList);
#endif /* P95 */
	    }
	}
	
	values[i++] = (Datum)tup->t_oid; 	/* 11 - 14 */
    }

    /* ----------------
     * set default alignment
     * ----------------
     */
    values[i++] = (Datum)alignment;		/* 15 */

    /* ----------------
     *	initialize the default value for this type.
     * ----------------
     */
    values[i] = (Datum)fmgr(TextInRegProcedure,		/* 16 */
		     PointerIsValid(defaultTypeValue)
		     ? defaultTypeValue : "-");	/* XXX default typdefault */
    
    /* ----------------
     *	open pg_type and begin a scan for the type name.
     * ----------------
     */
    pg_type_desc = heap_openr(TypeRelationName);
    
    /* -----------------
     * Set a write lock initially so as not upgrade a read to a write
     * when the heap_insert() or heap_replace() is called.
     * -----------------
     */
    RelationSetLockForWrite(pg_type_desc);
    
#ifndef P95
    typeKey[0].sk_argument = NameGetDatum(typeName);
#else /* P95 */
    typeKey[0].sk_argument = PointerGetDatum(typeName);
#endif /* P95 */
    pg_type_scan = heap_beginscan(pg_type_desc,
				  0,
				  SelfTimeQual,
				  1,
				  typeKey);
    
    /* ----------------
     *  define the type either by adding a tuple to the type
     *  relation, or by updating the fields of the "shell" tuple
     *  already there.
     * ----------------
     */
    tup = heap_getnext(pg_type_scan, 0, &buffer);
    if (HeapTupleIsValid(tup)) {
	tup = heap_modifytuple(tup,
			       buffer,
			       pg_type_desc,
			       values,
			       nulls,
			       replaces);
	
	/* XXX may not be necessary */
	ItemPointerCopy(&tup->t_ctid, &itemPointerData);
	
	setheapoverride(true);
	(void) heap_replace(pg_type_desc, &itemPointerData, tup);
	setheapoverride(false);
	
	typeObjectId = tup->t_oid;
    } else {
#ifndef P95
	tup = heap_formtuple(Natts_pg_type,
			     pg_type_desc->rd_att,
#else /* P95 */
	tupDesc = pg_type_desc->rd_att;

	tup = heap_formtuple(tupDesc,
#endif /* P95 */
			     values,
			     nulls);
	
	heap_insert(pg_type_desc, tup);
	
	typeObjectId = tup->t_oid;
    }
    
    /* ----------------
     *	finish up
     * ----------------
     */
    heap_endscan(pg_type_scan);
    
    if (RelationGetRelationTupleForm(pg_type_desc)->relhasindex)
	{
	    Relation idescs[Num_pg_type_indices];
	    
	    CatalogOpenIndices(Num_pg_type_indices, Name_pg_type_indices, idescs);
	    CatalogIndexInsert(idescs, Num_pg_type_indices, pg_type_desc, tup);
	    CatalogCloseIndices(Num_pg_type_indices, idescs);
	}
    RelationUnsetLockForWrite(pg_type_desc);
    heap_close(pg_type_desc);
    
    
    return
	typeObjectId;
}

/* ----------------------------------------------------------------
 *	TypeRename
 *
 *	This renames a type 
 * ----------------------------------------------------------------
 */
void
#ifndef P95
TypeRename(Name oldTypeName, Name newTypeName)
#else /* P95 */
TypeRename(char *oldTypeName, char *newTypeName)
#endif /* P95 */
{
    Relation 		pg_type_desc;
    Relation		idescs[Num_pg_type_indices];
    Oid            type_oid;
    HeapTuple 		tup;
    bool 		defined;
    ItemPointerData	itemPointerData;
    
#ifndef P95
    Assert(NameIsValid(oldTypeName));
    Assert(NameIsValid(newTypeName));
    
#endif /* P95 */
    /* check that that the new type is not already defined */
    type_oid = TypeGet(newTypeName, &defined);
    if (OidIsValid(type_oid) && defined) {
	elog(WARN, "TypeRename: type %s already defined", newTypeName);	
    }
    
    /* get the type tuple from the catalog index scan manager */
    pg_type_desc = heap_openr(TypeRelationName);
#ifndef P95
    tup = TypeNameIndexScan(pg_type_desc, oldTypeName->data);
#else /* P95 */
    tup = TypeNameIndexScan(pg_type_desc, oldTypeName);
#endif /* P95 */
    
    /* ----------------
     *  change the name of the type
     * ----------------
     */
    if (HeapTupleIsValid(tup)) {
	
#ifndef P95
	namecpy(& (((TypeTupleForm) GETSTRUCT(tup))->typname),newTypeName);
#else /* P95 */
	namestrcpy(& (((TypeTupleForm) GETSTRUCT(tup))->typname),newTypeName);
#endif /* P95 */
	
	ItemPointerCopy(&tup->t_ctid, &itemPointerData);
	
	setheapoverride(true);
#ifndef P95
	(void) heap_replace(pg_type_desc, &itemPointerData, tup);
#else /* P95 */
	heap_replace(pg_type_desc, &itemPointerData, tup);
#endif /* P95 */
	setheapoverride(false);
	
	/* update the system catalog indices */
	CatalogOpenIndices(Num_pg_type_indices, Name_pg_type_indices, idescs);
	CatalogIndexInsert(idescs, Num_pg_type_indices, pg_type_desc, tup);
	CatalogCloseIndices(Num_pg_type_indices, idescs);
	
	/* all done */
	pfree(tup);
	
    } else {
	elog(WARN, "TypeRename: type %s not defined", oldTypeName);	
    }
    
    /* finish up */
    heap_close(pg_type_desc);
#ifdef P95
}

/*
 * makeArrayTypeName(typeName);
 *    - given a base type name, make an array of type name out of it
 *
 * the CALLER is responsible for pfreeing the 
 */

char*
makeArrayTypeName(char* typeName)
{
    char *arr;

    if (!typeName) return NULL;
    arr = palloc (strlen(typeName) + 2);
    arr[0] = '_';
    strcpy(arr+1, typeName);

    return arr;
	
#endif /* P95 */
}
