#ifndef P95

/*-------------------------------------------------------------------------
 *
 * creatinh.c--
 *    POSTGRES create/destroy relation with inheritance utility code.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/commands/creatinh.c,v 1.12 1996/06/07 04:55:30 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>	/* for sprintf() */
#include <string.h>
#include "postgres.h"

#include "tcop/tcopdebug.h"

#include "utils/builtins.h"
#include "utils/elog.h"
#include "utils/palloc.h"

#include "nodes/pg_list.h"
#include "nodes/primnodes.h"
#include "nodes/plannodes.h"
#include "nodes/parsenodes.h"
#include "nodes/execnodes.h"

#include "utils/syscache.h"
#include "catalog/catname.h"
#include "catalog/pg_type.h"
#include "catalog/pg_inherits.h"
#include "catalog/pg_ipl.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_fragment.h"

#include "commands/creatinh.h"

#include "access/tupdesc.h"
#include "access/heapam.h"
#include "access/xact.h"
#include "access/miscam.h"

#include "copies.h"

/* ----------------
 *	local stuff
 * ----------------
 */

static List *MergeAttributes(List *schema, List *supers);
static void StoreCatalogInheritance(Oid relationId, List *supers);
static int checkAttrExists(Name attributeName, Name attributeType, List *schema);

/* ----------------------------------------------------------------
 *	DefineRelation --
 *		Creates a new relation.
 * ----------------------------------------------------------------
 */
void
DefineRelation(CreateStmt *stmt)
{
    char		*relname	= stmt->relname;
    List		*schema		= stmt->tableElts;
    int			numberOfAttributes;
    Oid			relationId;
    char		archChar;
    List		*inheritList 	= NULL;
    Name		relationName 	= NULL;
    TupleDesc		descriptor;
    int			heaploc, archloc;
    static char	*modes[] = { "none", "random", "Roundrobin", "key", "hash" };
    int			i;
    int			partmode	= 'n';
    AttrNumber		attnum		= 0;
    List		*entry;
    ColumnDef		*coldef;
    char		query[1000];

    /* ----------------
     *  Check fragment partition mode.
     * ----------------
     */
    for (i = 0; i < 5; ++i) {
	if (!strcasecmp(modes[i], stmt->partmode)) {
	    partmode = modes[i][0];
	    break;
	}
    }
    if (partmode == 'n') {
	elog(WARN, "DefineRelation: bad partition mode '%s'", stmt->partmode);
    }

    /* ----------------
     *	minimal argument checking follows
     * ----------------
     */
    AssertArg(NameIsValid(relname));

    if ( strlen(relname) > NAMEDATALEN)
	elog(WARN, "the relation name %s is > %d characters long", relname,
	     NAMEDATALEN);
    
    relationName = (Name)palloc(NAMEDATALEN + 1);
    namestrcpy(relationName, relname);
    
    /* ----------------
     * 	Handle parameters
     * 	XXX parameter handling missing below.
     * ----------------
     */
    inheritList = stmt->inhRelnames;
    
    /* ----------------
     *	determine archive mode
     * 	XXX use symbolic constants...
     *
     *  @@@
     *  Ok, we take the user's choice for whether the fragment is archived or
     *  not, and this will decide if the archive is maintained. However for
     *  now we end up creating it anyway (see the end of the function)
     *  because the fragment move/ fragment copy/ splitfragment code depends
     *  on the archive relation being there.
     * ----------------
     */

    /* ----------------
     *  determine archive mode
     *  XXX use symbolic constants...
     * ----------------
     */
    switch (stmt->archiveType) {
    case ARCH_NONE:
	archChar = 'n';
	break;
    case ARCH_LIGHT:
	archChar = 'l';
	break;
    case ARCH_HEAVY:
	archChar = 'h';
	break;
    default:
	elog(WARN, "Botched archive mode %d, ignoring", stmt->archiveType);
	break;
    }
    
    if (stmt->location == -1)
	heaploc = 0;
    else
	heaploc = stmt->location;
    
    /*
     *  For now, any user-defined relation defaults to the magnetic
     *  disk storage manager.  --mao 2 july 91
     */
    if (stmt->archiveLoc == -1) {
	archloc = 0;
    } else {
	if (archChar == 'n') {
	    elog(WARN, "Set archive location, but not mode, for %s",
		 relationName);
	}
	archloc = stmt->archiveLoc;
    }
    
    /* ----------------
     *	generate relation schema, including inherited attributes.
     * ----------------
     */
    schema = MergeAttributes(schema, inheritList);
    
    numberOfAttributes = length(schema);
    if (numberOfAttributes <= 0) {
	elog(WARN, "DefineRelation: %s",
	     "please inherit from a relation or define an attribute");
    }

    /* ----------------
     *  Check partition attribute name.
     *
     *  @@@ This assumes BuildDescForRelation puts the attributes into the
     *	    relation descriptor the same way they were listed in the schema.
     *	    (Since we use the attnum in the query to update pg_class.)
     * ----------------
     */
    if (stmt->partattname) {
	attnum = 1;
	foreach (entry, schema) {
	    coldef = (ColumnDef *)lfirst(entry);
	    if (!strcmp(coldef->colname, stmt->partattname)) {
		break;
	    }
	    ++attnum;
	}
	if (attnum > numberOfAttributes) {
	    elog(WARN, "DefineRelation: no partition attribute '%s' in schema",
		 stmt->partattname);
	}
    }
    
    /* ----------------
     *	create a relation descriptor from the relation schema
     *  and create the relation.  
     * ----------------
     */
    descriptor = BuildDescForRelation(schema, relationName);
    relationId = heap_create(relationName->data,
			     archChar,
			     numberOfAttributes,
			     heaploc,
			     descriptor);
    
    StoreCatalogInheritance(relationId, inheritList);
    
    /* ----------------
     *	create an archive relation if necessary
     *
     *  @@@ For now, we create it automatically.
     * ----------------
     */
#ifdef 0
    if (archChar != 'n') {
#endif
	/*
	 *  Need to create an archive relation for this heap relation.
	 *  We cobble up the command by hand, and increment the command
	 *  counter ourselves.
	 */
	CommandCounterIncrement();
	CreateArchiveRelationFromScratch(relationId, numberOfAttributes,
					 descriptor);
#ifdef 0
    }
#endif

    /* ----------------
     *  Update the system catalogs with fragment information.
     *
     *  pg_class gets updated with the partition specification;
     *  pg_fragment gets an entry for the new class' initial fragment.
     * ----------------
     */

    /*
     * @@@ Hack! What really should happen, since all tables are
     *     fragmented, is that the partition info be passed to
     *     heap_create. I will put this on my to-do list. -Rex 8.5.1995
     */
    CommandCounterIncrement();

    sprintf(query, "UPDATE pg_class \
SET relpartcmp='%s', relparthash='%s', relpartatt=%d, relpartmode='%c' \
WHERE relname='%s'",
	    stmt->partcmpfn ? stmt->partcmpfn : "",
	    stmt->parthashfn ? stmt->parthashfn : "",
	    attnum, partmode, relationName->data);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    DefineRelationFirstFragment(relationId);
}

/*
 * RemoveRelation --
 *	Deletes a fragments, and its pg_fragment entries.
 *
 *	@@@ Should do something more sophisticated for copies, but we must
 *	    define semantics for that so I didn't put anything specific.
 *	    We could just grab a write lock, send out update streams for each
 *	    remaining contract, before deleting-- that was my first thought.
 *
 *	@@@ Should check if there's an archive relation, and delete that too!!
 *
 * Exceptions:
 *	BadArg if name is invalid.
 *
 * Note:
 *	If the relation has indices defined on it, then the index relations
 * 	themselves will be destroyed, too.
 */
void
RemoveRelation(Name name)
{
    Relation rel, arel, frel;
    Oid reloid;
    HeapScanDesc sdesc;
    HeapTuple tup;
    ScanKeyData skey;
    int myHostId = getMyHostId();
    
    AssertArg(NameIsValid(name));

    /*
     * Open relation 'name' to get its supposed reloid.
     */
    rel = heap_openr(name);
    if (!RelationIsValid(rel)) {
	elog(WARN, "RemoveRelation: no such table '%s'", name);
    }
    reloid = rel->rd_id;

    /*
     * Set a write lock so people don't add stuff while we delete this guy.
     * Also so that update stream sendoffs and receives will fail, in case
     * this is a copy.
     */
    RelationSetLockForWrite(rel);

    /*
     * If this is a copied fragment, need to break the contracts on it and
     * remove the side file if it has one. If it has a hash index on oid
     * that will get removed automatically later, when the fragment is
     * heap_destroy'd.
     */
    if (rel->rd_iscopied) {
	CancelContracts(false, false, 0, reloid);

	/*
	 * If relation has a side-file (had a send contract) then we
	 * remove it.
	 */
	if (rel->rd_copysidef != 0) {
	    DestroySideFile(rel);
	}
    }
    heap_close(rel);

    /*
     * if this table has an archive, remove it first.
     *
     * normally, we would trust the test "relarch == 'n'" as an
     * indication of whether or not we have to do this, but we can't
     * in Mariposa because of the hack that every user-defined table
     * has an archive table (whether it is maintained or not).
     */
    arel = OpenArchiveRelation(reloid);
    if (RelationIsValid(arel)) {
	NameData archname;

	(void) namecpy(&archname, &(arel->rd_rel->relname));
	heap_close(arel);
	RemoveRelation(&archname);
    }

    /*
     * The code section below, I'm not sure we want. In Mariposa it is
     * more of an issue to delete whole classes, locating their
     * fragments... that will come later.
     *
     * XXX presumably the issue is that we need (1) some kind of test
     * to make sure that we don't destroy the last fragment, and (2)
     * some kind of notification.  i'm not sure what the former is,
     * but the latter should be the metadata replication (aka "name
     * service") system.  anyway, i think littering the world with
     * fragment metadata is a crock, so i'm putting this back.
     * -- pma 02/18/96
     *
     * YYY and furthermore....  :-)
     *
     * Changed the code below so it only deletes fragments that are
     * stored locally.  It was trying to call heap_open for fragments
     * that were actually stored elsewhere.  Another possibility (for
     * now) would be to check if there are any remote fragments and
     * refuse to drop the table if there are.  This is somewhat semantically
     * cleaner.
     *
     *					--jps 03/06/96
     */
#if 1
    /*
     * Scan pg_fragment on the reloid, looking for all the fragments.
     * Destroy each fragment. This is so that if this were a class,
     * then we'd destroy all of its fragments.
     */
    ScanKeyEntryInitialize(&skey, 0x0, Anum_pg_fragment_fraclassid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(reloid));
    frel = heap_openr((Name)Name_pg_fragment);
    sdesc = heap_beginscan(frel, false, NowTimeQual, 1, &skey);
    while (HeapTupleIsValid(tup = heap_getnext(sdesc, 0, NULL))) {
	Relation fragrel;
	NameData fragname;
	Form_pg_fragment pgFragTuple;

	pgFragTuple = (Form_pg_fragment)GETSTRUCT(tup);

	if (pgFragTuple->fralocation == myHostId) {
	    fragrel = heap_open(pgFragTuple->frastoreid);
	    if (!RelationIsValid(fragrel)) {
		elog(WARN, "RemoveRelation: can't open class '%s' fragment",
		     name->data);
	    }
	    (void) namecpy(&(fragname), &(fragrel->rd_rel->relname));
	    heap_close(fragrel);

	    fprintf(stderr, "found (%s %s) %d %d %d\n",
		    name->data, fragname.data,
		    pgFragTuple->fraclassid,
		    pgFragTuple->frastoreid,
		    pgFragTuple->fralogicalid);

	    if (pgFragTuple->frastoreid != reloid) {
		RemoveRelation(&fragname);
		CommandCounterIncrement();
	    }
	}
    }

    heap_endscan(sdesc);
    heap_close(frel);
#endif

    DeletePgFragmentTuples(reloid, Anum_pg_fragment_frastoreid);
    fprintf(stderr, "actually destroying %s\n", name->data);
    heap_destroy(name->data);
}

/*
 * MergeAttributes --
 *	Returns new schema given initial schema and supers.
 *
 *
 * 'schema' is the column/attribute definition for the table. (It's a list
 *	of ColumnDef's.) It is destructively changed.
 * 'inheritList' is the list of inherited relations (a list of Value(str)'s).
 *
 * Notes:
 *    The order in which the attributes are inherited is very important.
 *    Intuitively, the inherited attributes should come first. If a table
 *    inherits from multiple parents, the order of those attributes are
 *    according to the order of the parents specified in CREATE TABLE.
 *
 *    Here's an example:
 *
 *	create table person (name text, age int4, location point);
 *	create table emp (salary int4, manager char16) inherits(person);
 *	create table student (gpa float8) inherits (person);
 *	create table stud_emp (percent int4) inherits (emp, student);
 *
 *    the order of the attributes of stud_emp is as follow:
 *
 *
 *                          person {1:name, 2:age, 3:location}
 *                          /    \
 *             {6:gpa}  student   emp {4:salary, 5:manager}
 *                          \    /
 *                         stud_emp {7:percent}
 */
static List *
MergeAttributes(List *schema, List *supers)
{
    List *entry;
    List *inhSchema = NIL;
    
    /*
     * Validates that there are no duplications.
     * Validity checking of types occurs later.
     */
    foreach (entry, schema) {
	List	*rest;
	ColumnDef *coldef = lfirst(entry);
	
	foreach (rest, lnext(entry)) {
	    /*
	     * check for duplicated relation names
	     */
	    ColumnDef *restdef = lfirst(rest);
	    
	    if (!strcmp(coldef->colname, restdef->colname)) {
		elog(WARN, "attribute \"%s\" duplicated",
		     coldef->colname);
	    }
	}
    }
    foreach (entry, supers) {
	List	*rest;
	
	foreach (rest, lnext(entry)) {
	    if (!strcmp(strVal(lfirst(entry)), strVal(lfirst(rest)))) {
		elog(WARN, "relation \"%s\" duplicated",
		     strVal(lfirst(entry)));
	    }
	}
    }
    
    /*
     * merge the inherited attributes into the schema
     */
    foreach (entry, supers) {
        NameData        relNameData;
	char		*name = strVal(lfirst(entry));
	Relation	relation;
	List		*partialResult = NIL;
	AttrNumber	attrno;
	TupleDesc	tupleDesc;
	
	/*
	 * grab tuple descriptor for next superclass relation
	 */
	namestrcpy(&relNameData,name);
	relation =  heap_openr(&relNameData);
	if (relation==NULL) {
	    elog(WARN,
		 "MergeAttr: Can't inherit from non-existent superclass '%s'",
		 name);
	}
	tupleDesc = RelationGetTupleDescriptor(relation);
	
	for (attrno = relation->rd_rel->relnatts - 1; attrno >= 0; attrno--) {
	    AttributeTupleForm	attribute = tupleDesc[attrno];
	    Name	attributeName;
	    Name	attributeType;
	    HeapTuple	tuple;
	    ColumnDef	*def;
	    TypeName	*typename;
	    
	    /*
	     * form name and type
	     */
	    attributeName = &(attribute->attname);
	    tuple =
		SearchSysCacheTuple(TYPOID,
				    (char *)ObjectIdGetDatum(attribute->atttypid),
				    (char *) NULL, (char *) NULL,
				    (char *) NULL);
	    AssertState(HeapTupleIsValid(tuple));
	    attributeType =
		&(((TypeTupleForm)GETSTRUCT(tuple))->typname);
	    /*
	     * check validity
	     *
	     */
	    if (checkAttrExists(attributeName, attributeType, inhSchema) ||
		checkAttrExists(attributeName, attributeType, schema)) {
		/*
		 * this entry already exists
		 */
		continue;
	    }

	    /*
	     * add an entry to the schema
	     */
	    def = makeNode(ColumnDef);
	    typename = makeNode(TypeName);
	    def->colname = (char*)makeName((char*)attributeName);
	    typename->name = (char*)makeName((char*)attributeType); 
	    def->typename = typename;
	    partialResult = lcons(def, partialResult);
	}
	
	/*
	 * iteration cleanup and result collection
	 */
	heap_close(relation);

	/*
	 * wants the inherited schema to appear in the order they are
	 * specified in CREATE TABLE
	 */
	inhSchema = nconc(inhSchema, partialResult);
    }

    /*
     * put the inherited schema before our the schema for this table
     */
    schema = nconc(inhSchema, schema);
    
    return (schema);
}

/*
 * StoreCatalogInheritance --
 *	Updates the system catalogs with proper inheritance information.
 */
static void
StoreCatalogInheritance(Oid relationId, List *supers)
{
    Relation	relation;
    TupleDesc	desc;
    int16	seqNumber;
    List	*entry;
    List	*idList;
    HeapTuple	tuple;
    
    /* ----------------
     *	sanity checks
     * ----------------
     */
    AssertArg(OidIsValid(relationId));
    
    if (supers==NIL)
	return;
    
    /* ----------------
     * Catalog INHERITS information.
     * ----------------
     */
    relation = heap_openr( InheritsRelationName );
    desc = RelationGetTupleDescriptor(relation);

    seqNumber = 1;
    idList = NIL;
    foreach (entry, supers) {
	Datum		datum[ Natts_pg_inherits ];
	char		nullarr[ Natts_pg_inherits ];
	
	tuple = SearchSysCacheTuple(RELNAME, (char *) strVal(lfirst(entry)),
				    (char *) NULL, (char *) NULL,
				    (char *) NULL);
	AssertArg(HeapTupleIsValid(tuple));
	
	/*
	 * build idList for use below
	 */
	idList = lappendi(idList, tuple->t_oid);
	
	datum[0] = ObjectIdGetDatum(relationId);	/* inhrel */
	datum[1] = ObjectIdGetDatum(tuple->t_oid);	/* inhparent */
	datum[2] = Int16GetDatum(seqNumber);		/* inhseqno */
	
	nullarr[0] = ' ';
	nullarr[1] = ' ';
	nullarr[2] = ' ';
	
	tuple = heap_formtuple(Natts_pg_inherits, desc,
			       datum, nullarr);
	
	(void) heap_insert(relation, tuple);
	pfree(tuple);
	
	seqNumber += 1;
    }
    
    heap_close(relation);
    
    /* ----------------
     * Catalog IPL information.
     *
     * Algorithm:
     *	0. list superclasses (by Oid) in order given (see idList).
     *	1. append after each relationId, its superclasses, recursively.
     *	3. remove all but last of duplicates.
     *	4. store result.
     * ----------------
     */
    
    /* ----------------
     *	1.
     * ----------------
     */
    foreach (entry, idList) {
	HeapTuple		tuple;
	Oid			id;
	int16			number;
	List			*next;
	List			*current;
	
	id = (Oid)lfirsti(entry);
	current = entry;
	next = lnext(entry);
	
	for (number = 1; ; number += 1) {
	    tuple = SearchSysCacheTuple(INHRELID,
					(char *) ObjectIdGetDatum(id),
					(char *) Int16GetDatum(number),
					(char *) NULL, (char *) NULL);
	    
	    if (! HeapTupleIsValid(tuple))
		break;
	    
	    lnext(current) =
		lconsi(((InheritsTupleForm)
			 GETSTRUCT(tuple))->inhparent,
			NIL);
	    
	    current = lnext(current);
	}
	lnext(current) = next;
    }
    
    /* ----------------
     *	2.
     * ----------------
     */
    foreach (entry, idList) {
	Oid		name;
	List		*rest;
	bool		found = false;
	
    again:
	name = lfirsti(entry);
	foreach (rest, lnext(entry)) {
	    if (name == lfirsti(rest)) {
		found = true;
		break;
	    }
	}
	if (found) {
	    /*
	     * entry list must be of length >= 2 or else no match
	     *
	     * so, remove this entry.
	     */
	    lfirst(entry) = lfirst(lnext(entry));
	    lnext(entry) = lnext(lnext(entry));
	    
	    found = false;
	    goto again;
	}
    }
    
    /* ----------------
     *	3.
     * ----------------
     */
    relation = heap_openr( InheritancePrecidenceListRelationName );
    desc = RelationGetTupleDescriptor(relation);
    
    seqNumber = 1;
    
    foreach (entry, idList) {
	Datum	datum[ Natts_pg_ipl ];
	char	nullarr[ Natts_pg_ipl ];
	
	datum[0] = ObjectIdGetDatum(relationId);	/* iplrel */
	datum[1] = ObjectIdGetDatum(lfirsti(entry));
	/*iplinherits*/
	datum[2] = Int16GetDatum(seqNumber);		/* iplseqno */
	
	nullarr[0] = ' ';
	nullarr[1] = ' ';
	nullarr[2] = ' ';
	
	tuple =
	    heap_formtuple(Natts_pg_ipl, desc, datum, nullarr);
	
	(void) heap_insert(relation, tuple);
	pfree(tuple);
	
	seqNumber += 1;
    }

    heap_close(relation);
}

/*
 * returns 1 if attribute already exists in schema, 0 otherwise.
 */
static int
checkAttrExists(Name attributeName, Name attributeType, List *schema)
{
    List *s;

    foreach (s, schema) {
	ColumnDef *def = lfirst(s);

	if (!namestrcmp(attributeName, def->colname)) {
	    /*
	     * attribute exists. Make sure the types are the same.
	     */
	    if (namestrcmp(attributeType, def->typename->name) != 0) {
		elog(WARN, "%s and %s conflict for %s",
		     attributeType, def->typename->name, attributeName);
	    }
	    return 1;
	}
    }
    return 0;
}

/*
 * inserts a single pg_fragment entry corresponding to a new table.
 *
 * this should be used any place where heap_create is called to create
 * a cataloged table, since all cataloged tables must be fragmented.
 *
 * XXX this is a bad place to use pg_eval, since this should be used
 * in the bootstrap code as well as the server.
 */
void
DefineRelationFirstFragment(Oid relationId)
{
    Relation rd = heap_openr(Name_pg_fragment);
    HeapTuple tup;
    Datum values[Natts_pg_fragment];
    char nulls[Natts_pg_fragment];
    Oid id;
    int i;

    if (!RelationIsValid(rd)) {
	return;
    }

    for (i = 0; i < Natts_pg_fragment; ++i) {
	values[i] = (Datum) NULL;
	nulls[i] = 'n';
    }

    values[Anum_pg_fragment_frapages - 1] =
	values[Anum_pg_fragment_fratuples - 1] = 0;
    values[Anum_pg_fragment_fraclassid - 1] = 
	values[Anum_pg_fragment_frastoreid - 1] = 
	    values[Anum_pg_fragment_fralogicalid - 1] = relationId;
    values[Anum_pg_fragment_fralocation - 1] = getMyHostId();

    nulls[Anum_pg_fragment_frapages - 1] =
	nulls[Anum_pg_fragment_fratuples - 1] =
	    nulls[Anum_pg_fragment_fraclassid - 1] = 
		nulls[Anum_pg_fragment_frastoreid - 1] = 
		    nulls[Anum_pg_fragment_fralogicalid - 1] =
			nulls[Anum_pg_fragment_fralocation - 1] = ' ';

    tup = heap_formtuple(Natts_pg_fragment, rd->rd_att, values, nulls);
    if (!HeapTupleIsValid(tup)) {
	(void) heap_close(rd);
	elog(WARN, "DefineRelationFirstFragment: formtuple failed");
    }
    id = heap_insert(rd, tup);

    pfree(tup);
    (void) heap_close(rd);

    if (!OidIsValid(id)) {
	elog(WARN, "DefineRelationFirstFragment: insertion failed: id='%d'",
	     relationId);
    }
}
#else /* P95 */

/*-------------------------------------------------------------------------
 *
 * creatinh.c--
 *    POSTGRES create/destroy relation with inheritance utility code.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>	/* for sprintf() */
#include <string.h>
#include "postgres.h"

#include "tcop/tcopdebug.h"

#include "utils/builtins.h"
#include "utils/elog.h"
#include "utils/palloc.h"

#include "nodes/pg_list.h"
#include "nodes/primnodes.h"
#include "nodes/plannodes.h"
#include "nodes/parsenodes.h"
#include "nodes/execnodes.h"

#include "utils/syscache.h"
#include "utils/relcache.h"
#include "catalog/catname.h"
#include "catalog/pg_type.h"
#include "catalog/pg_inherits.h"
#include "catalog/pg_ipl.h"
#include "parser/catalog_utils.h"

#include "commands/creatinh.h"

#include "access/tupdesc.h"
#include "access/heapam.h"
#include "access/xact.h"

/* ----------------
 *	local stuff
 * ----------------
 */

static int checkAttrExists(char *attributeName, 
 			   char *attributeType, List *schema);
static List *MergeAttributes(List *schema, List *supers);
static void StoreCatalogInheritance(Oid relationId, List *supers);

/* ----------------------------------------------------------------
 *	DefineRelation --
 *		Creates a new relation.
 * ----------------------------------------------------------------
 */
void
DefineRelation(CreateStmt *stmt)
{
    char *relname = stmt->relname;
    List *schema = stmt->tableElts;
    int			numberOfAttributes;
    Oid			relationId;
    char		archChar;
    List		*inheritList 	= NULL;
    char *archiveName 	= NULL;
    TupleDesc	descriptor;
    int			heaploc, archloc;
    
    char*   typename = NULL;  /* the typename of this relation. not useod for now */

    if ( strlen(relname) > NAMEDATALEN)
	elog(WARN, "the relation name %s is > %d characters long", relname,
	     NAMEDATALEN);
    
    /* ----------------
     * 	Handle parameters
     * 	XXX parameter handling missing below.
     * ----------------
     */
    inheritList = stmt->inhRelnames;
    
    /* ----------------
     *	determine archive mode
     * 	XXX use symbolic constants...
     * ----------------
     */
    archChar = 'n';
    
    switch (stmt->archiveType) {
    case ARCH_NONE:
	archChar = 'n';
	break;
    case ARCH_LIGHT:
	archChar = 'l';
	break;
    case ARCH_HEAVY:
	archChar = 'h';
	break;
    default:
	elog(WARN, "Botched archive mode %d, ignoring",
	     stmt->archiveType);
	break;
    }
    
    if (stmt->location == -1)
	heaploc = 0;
    else
	heaploc = stmt->location;
    
    /*
     *  For now, any user-defined relation defaults to the magnetic
     *  disk storgage manager.  --mao 2 july 91
     */
    if (stmt->archiveLoc == -1) {
	archloc = 0;
    } else {
	if (archChar == 'n') {
	    elog(WARN, "Set archive location, but not mode, for %s",
		 relname);
	}
	archloc = stmt->archiveLoc;
    }
    
    /* ----------------
     *	generate relation schema, including inherited attributes.
     * ----------------
     */
    schema = MergeAttributes(schema, inheritList);
    
    numberOfAttributes = length(schema);
    if (numberOfAttributes <= 0) {
	elog(WARN, "DefineRelation: %s",
	     "please inherit from a relation or define an attribute");
    }
    
    /* ----------------
     *	create a relation descriptor from the relation schema
     *  and create the relation.  
     * ----------------
     */
    descriptor = BuildDescForRelation(schema, relname);
    relationId = heap_create(relname,
			     typename,
			     archChar,
			     heaploc,
			     descriptor);
    
    StoreCatalogInheritance(relationId, inheritList);
    
    /* ----------------
     *	create an archive relation if necessary
     * ----------------
     */
    if (archChar != 'n') {
	/*
	 *  Need to create an archive relation for this heap relation.
	 *  We cobble up the command by hand, and increment the command
	 *  counter ourselves.
	 */
	
	CommandCounterIncrement();
	archiveName = MakeArchiveName(relationId);
	
	relationId = heap_create(archiveName,
				 typename,
				 'n',		/* archive isn't archived */
				 archloc,
				 descriptor);
	
	pfree(archiveName);
    }
}

/*
 * RemoveRelation --
 *	Deletes a new relation.
 *
 * Exceptions:
 *	BadArg if name is invalid.
 *
 * Note:
 *	If the relation has indices defined on it, then the index relations
 * themselves will be destroyed, too.
 */
void
RemoveRelation(char *name)
{
    AssertArg(name);
    heap_destroy(name);
}


/*
 * MergeAttributes --
 *	Returns new schema given initial schema and supers.
 *
 *
 * 'schema' is the column/attribute definition for the table. (It's a list
 *	of ColumnDef's.) It is destructively changed.
 * 'inheritList' is the list of inherited relations (a list of Value(str)'s).
 *
 * Notes:
 *    The order in which the attributes are inherited is very important.
 *    Intuitively, the inherited attributes should come first. If a table
 *    inherits from multiple parents, the order of those attributes are
 *    according to the order of the parents specified in CREATE TABLE.
 *
 *    Here's an example:
 *
 *	create table person (name text, age int4, location point);
 *	create table emp (salary int4, manager char16) inherits(person);
 *	create table student (gpa float8) inherits (person);
 *	create table stud_emp (percent int4) inherits (emp, student);
 *
 *    the order of the attributes of stud_emp is as follow:
 *
 *
 *                          person {1:name, 2:age, 3:location}
 *                          /    \
 *             {6:gpa}  student   emp {4:salary, 5:manager}
 *                          \    /
 *                         stud_emp {7:percent}
 */
static List *
MergeAttributes(List *schema, List *supers)
{
    List *entry;
    List *inhSchema = NIL;
    
    /*
     * Validates that there are no duplications.
     * Validity checking of types occurs later.
     */
    foreach (entry, schema) {
	List	*rest;
	ColumnDef *coldef = lfirst(entry);
	
	foreach (rest, lnext(entry)) {
	    /*
	     * check for duplicated relation names
	     */
	    ColumnDef *restdef = lfirst(rest);
	    
	    if (!strcmp(coldef->colname, restdef->colname)) {
		elog(WARN, "attribute \"%s\" duplicated",
		     coldef->colname);
	    }
	}
    }
    foreach (entry, supers) {
	List	*rest;
	
	foreach (rest, lnext(entry)) {
	    if (!strcmp(strVal(lfirst(entry)), strVal(lfirst(rest)))) {
		elog(WARN, "relation \"%s\" duplicated",
		     strVal(lfirst(entry)));
	    }
	}
    }
    
    /*
     * merge the inherited attributes into the schema
     */
    foreach (entry, supers) {
	char		*name = strVal(lfirst(entry));
	Relation	relation;
	List		*partialResult = NIL;
	AttrNumber	attrno;
	TupleDesc	tupleDesc;
	
	relation =  heap_openr(name);
	if (relation==NULL) {
	    elog(WARN,
		 "MergeAttr: Can't inherit from non-existent superclass '%s'",
		 name);
	}
	tupleDesc = RelationGetTupleDescriptor(relation);
	
	for (attrno = relation->rd_rel->relnatts - 1; attrno >= 0; attrno--) {
	    AttributeTupleForm	attribute = tupleDesc->attrs[attrno];
	    char *attributeName;
	    char *attributeType;
	    HeapTuple	tuple;
	    ColumnDef	*def;
	    TypeName	*typename;
	    
	    /*
	     * form name and type
	     */
	    attributeName = (attribute->attname).data;
	    tuple =
		SearchSysCacheTuple(TYPOID,
				    ObjectIdGetDatum(attribute->atttypid),
				    0,0,0);
	    AssertState(HeapTupleIsValid(tuple));
	    attributeType =
		(((TypeTupleForm)GETSTRUCT(tuple))->typname).data;
	    /*
	     * check validity
	     *
	     */
	    if (checkAttrExists(attributeName, attributeType, inhSchema) ||
		checkAttrExists(attributeName, attributeType, schema)) {
		/*
		 * this entry already exists
		 */
		continue;
	    }

	    /*
	     * add an entry to the schema
	     */
	    def = makeNode(ColumnDef);
	    typename = makeNode(TypeName);
	    def->colname = pstrdup(attributeName);
	    typename->name = pstrdup(attributeType); 
	    def->typename = typename;
	    partialResult = lcons(def, partialResult);
	}
	
	/*
	 * iteration cleanup and result collection
	 */
	heap_close(relation);

	/*
	 * wants the inherited schema to appear in the order they are
	 * specified in CREATE TABLE
	 */
	inhSchema = nconc(inhSchema, partialResult);
    }

    /*
     * put the inherited schema before our the schema for this table
     */
    schema = nconc(inhSchema, schema);
    
    return (schema);
}

/*
 * StoreCatalogInheritance --
 *	Updates the system catalogs with proper inheritance information.
 */
static void
StoreCatalogInheritance(Oid relationId, List *supers)
{
    Relation	relation;
    TupleDesc	desc;
    int16	seqNumber;
    List	*entry;
    List	*idList;
    HeapTuple	tuple;
    
    /* ----------------
     *	sanity checks
     * ----------------
     */
    AssertArg(OidIsValid(relationId));
    
    if (supers==NIL)
	return;
    
    /* ----------------
     * Catalog INHERITS information.
     * ----------------
     */
    relation = heap_openr( InheritsRelationName );
    desc = RelationGetTupleDescriptor(relation);

    seqNumber = 1;
    idList = NIL;
    foreach (entry, supers) {
	Datum		datum[ Natts_pg_inherits ];
	char		nullarr[ Natts_pg_inherits ];
	
	tuple = SearchSysCacheTuple(RELNAME, 
				    PointerGetDatum(strVal(lfirst(entry))),
				    0,0,0);
	AssertArg(HeapTupleIsValid(tuple));
	
	/*
	 * build idList for use below
	 */
	idList = lappendi(idList, tuple->t_oid);
	
	datum[0] = ObjectIdGetDatum(relationId);	/* inhrel */
	datum[1] = ObjectIdGetDatum(tuple->t_oid);	/* inhparent */
	datum[2] = Int16GetDatum(seqNumber);		/* inhseqno */
	
	nullarr[0] = ' ';
	nullarr[1] = ' ';
	nullarr[2] = ' ';
	
	tuple = heap_formtuple(desc,datum, nullarr);
	
	(void) heap_insert(relation, tuple);
	pfree(tuple);
	
	seqNumber += 1;
    }
    
    heap_close(relation);
    
    /* ----------------
     * Catalog IPL information.
     *
     * Algorithm:
     *	0. list superclasses (by Oid) in order given (see idList).
     *	1. append after each relationId, its superclasses, recursively.
     *	3. remove all but last of duplicates.
     *	4. store result.
     * ----------------
     */
    
    /* ----------------
     *	1.
     * ----------------
     */
    foreach (entry, idList) {
	HeapTuple		tuple;
	Oid			id;
	int16			number;
	List			*next;
	List			*current;
	
	id = (Oid)lfirsti(entry);
	current = entry;
	next = lnext(entry);
	
	for (number = 1; ; number += 1) {
	    tuple = SearchSysCacheTuple(INHRELID,
					ObjectIdGetDatum(id),
					Int16GetDatum(number),
					0,0);
	    
	    if (! HeapTupleIsValid(tuple))
		break;
	    
	    lnext(current) =
		lconsi(((InheritsTupleForm)
			 GETSTRUCT(tuple))->inhparent,
			NIL);
	    
	    current = lnext(current);
	}
	lnext(current) = next;
    }
    
    /* ----------------
     *	2.
     * ----------------
     */
    foreach (entry, idList) {
	Oid		name;
	List		*rest;
	bool		found = false;
	
    again:
	name = lfirsti(entry);
	foreach (rest, lnext(entry)) {
	    if (name == lfirsti(rest)) {
		found = true;
		break;
	    }
	}
	if (found) {
	    /*
	     * entry list must be of length >= 2 or else no match
	     *
	     * so, remove this entry.
	     */
	    lfirst(entry) = lfirst(lnext(entry));
	    lnext(entry) = lnext(lnext(entry));
	    
	    found = false;
	    goto again;
	}
    }
    
    /* ----------------
     *	3.
     * ----------------
     */
    relation = heap_openr( InheritancePrecidenceListRelationName );
    desc = RelationGetTupleDescriptor(relation);
    
    seqNumber = 1;
    
    foreach (entry, idList) {
	Datum	datum[ Natts_pg_ipl ];
	char	nullarr[ Natts_pg_ipl ];
	
	datum[0] = ObjectIdGetDatum(relationId);	/* iplrel */
	datum[1] = ObjectIdGetDatum(lfirsti(entry));
	/*iplinherits*/
	datum[2] = Int16GetDatum(seqNumber);		/* iplseqno */
	
	nullarr[0] = ' ';
	nullarr[1] = ' ';
	nullarr[2] = ' ';
	
	tuple = heap_formtuple( desc, datum, nullarr);
	
	(void) heap_insert(relation, tuple);
	pfree(tuple);
	
	seqNumber += 1;
    }

    heap_close(relation);
}

/*
 * returns 1 if attribute already exists in schema, 0 otherwise.
 */
static int
checkAttrExists(char *attributeName, char *attributeType, List *schema)
{
    List *s;

    foreach (s, schema) {
	ColumnDef *def = lfirst(s);

	if (!strcmp(attributeName, def->colname)) {
	    /*
	     * attribute exists. Make sure the types are the same.
	     */
	    if (strcmp(attributeType, def->typename->name) != 0) {
		elog(WARN, "%s and %s conflict for %s",
		     attributeType, def->typename->name, attributeName);
	    }
	    return 1;
	}
    }
    return 0;
}

/*
 * MakeArchiveName
 *    make an archive rel name out of a regular rel name
 *
* the CALLER is responsible for freeing the memory allocated
 */

char*
MakeArchiveName(Oid relationId)
{
    char *arch;

    /*
     *  Archive relations are named a,XXXXX where XXXXX == the OID
     *  of the relation they archive.  Create a string containing
     *  this name and find the reldesc for the archive relation.
     */
    arch = palloc(NAMEDATALEN); 
    sprintf(arch, "a,%d",relationId);

    return arch;
}

#endif /* P95 */
