/*-------------------------------------------------------------------------
 *
 * define.c--
 *    POSTGRES "define" utility code.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/commands/define.c,v 0.4 1996/06/07 04:55:32 am Exp $
 *
 * DESCRIPTION
 *    The "DefineFoo" routines take the parse tree and pick out the
 *    appropriate arguments/flags, passing the results to the
 *    corresponding "FooDefine" routines (in src/catalog) that do
 *    the actual catalog-munging.
 *
 * NOTES
 *    These things must be defined and committed in the following order:
 *	"define function":
 *		input/output, recv/send procedures
 *	"define type":
 *		type
 *	"define operator":
 *		operators
 *
 *	Most of the parse-tree manipulation routines are defined in
 *	commands/manip.c.
 *
 *-------------------------------------------------------------------------
 */
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "postgres.h"

#include "access/heapam.h"
#include "access/htup.h"
#include "utils/tqual.h"
#include "catalog/catname.h"
#include "catalog/pg_aggregate.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "utils/syscache.h"
#include "nodes/pg_list.h"
#include "nodes/parsenodes.h"
#include "fmgr.h"		/* for fmgr */

#include "utils/builtins.h"	/* prototype for textin() */

#include "utils/elog.h"
#include "utils/palloc.h"
#include "commands/defrem.h"
#include "optimizer/xfunc.h"
#include "tcop/dest.h"

#ifndef P95
static Name defGetName(DefElem *def);
#endif /* P95 */
static char *defGetString(DefElem *def);
#ifndef P95
static int  defGetInteger(DefElem *def);
#else /* P95 */
static int  defGetTypeLength(DefElem *def);
#endif /* P95 */

#define	DEFAULT_TYPDELIM	','

/*
 * DefineFunction --
 *	Registers a new function.
 *
 * Exceptions:
 *	XXX
 */
void
DefineFunction(ProcedureStmt *stmt, CommandDest dest)
{
    List   	*parameters = stmt->withClause;
#ifndef P95
    String	proname_str = stmt->funcname;
    String	probin_str;
    String	prosrc_str;
    char16	pronameData;
    char16	prorettypeData;
    Name	languageName;
#else /* P95 */
    char        *proname = stmt->funcname;
    char*	probin_str;
    char*	prosrc_str;
    char *prorettype;
    char *languageName;
#endif /* P95 */
    bool	canCache;
    bool        trusted = TRUE;
    List	*argList;
    int32       byte_pct, perbyte_cpu, percall_cpu, outin_ratio;
    bool	returnsSet;
    int		i;
    
#ifndef P95
    /* ----------------
     * Note:
     * XXX	Checking of "name" validity (16 characters?) is needed.
     * ----------------
     */
    AssertArg(NameIsValid(proname_str));
    
    (void) namestrcpy(&pronameData, proname_str);
    
#endif /* P95 */
    /* ----------------
     * figure out the language and convert it to lowercase.
     * ----------------
     */
#ifndef P95
    languageName = makeName(stmt->language);
    for (i = 0; i < NAMEDATALEN && languageName->data[i]; ++i) {
	languageName->data[i] = tolower(languageName->data[i]);
#else /* P95 */
    languageName = stmt->language;
    for (i = 0; i < NAMEDATALEN && languageName[i]; ++i) {
	languageName[i] = tolower(languageName[i]);
#endif /* P95 */
    }
    
    /* ----------------
     * handle "returntype = X".  The function could return a singleton
     * value or a set of values.  Figure out which.
     * ----------------
     */
    if (nodeTag(stmt->returnType)==T_TypeName) {
	TypeName *setType = (TypeName *)stmt->returnType;
	/* a set of values */
#ifndef P95
	(void) namestrcpy(&prorettypeData,
			  setType->name);
#else /* P95 */
	prorettype = setType->name,
#endif /* P95 */
	returnsSet = true;
    }else {
	/* singleton */
#ifndef P95
	(void) namestrcpy(&prorettypeData, strVal(stmt->returnType));
#else /* P95 */
	prorettype = strVal(stmt->returnType);
#endif /* P95 */
	returnsSet = false;
    }
    
#ifndef P95
    /* Next attributes only definable for C functions */
    if (!namestrcmp(languageName, "c")) {
#else /* P95 */
    /* Next attributes are only defined for C functions */
    if ( strcmp(languageName, "c") == 0 ||
	 strcmp(languageName, "internal") == 0 )  {
#endif /* P95 */
	List *pl;

	/* the defaults */
	canCache = FALSE;
	byte_pct = BYTE_PCT;
	perbyte_cpu = PERBYTE_CPU;
	percall_cpu = PERCALL_CPU;
	outin_ratio = OUTIN_RATIO;

	foreach(pl, parameters) {
	    int count;
	    char *ptr;
	    ParamString *param = (ParamString*)lfirst(pl);

	    if (!strcasecmp(param->name, "isacachable")) {
		/* ----------------
		 * handle "[ iscachable ]": figure out if Postquel functions 
		 * are cacheable automagically?
		 * ----------------
		 */
		canCache = TRUE;
	    }else if (!strcasecmp(param->name, "trusted")) {
		/*
		 * we don't have untrusted functions any more. The 4.2
		 * implementation is lousy anyway so I took it out.
		 *				   	   -ay 10/94
		 */
		elog(WARN, "untrusted function has been decommissioned.");
	    }else if (!strcasecmp(param->name, "byte_pct")) {
		/*
		 ** handle expensive function parameters
		 */
		byte_pct = atoi(param->val);
	    }else if (!strcasecmp(param->name, "perbyte_cpu")) {
		if (!sscanf(param->val, "%d", &perbyte_cpu)) {
		    for (count = 0, ptr = param->val; *ptr != '\0'; ptr++) {
			if (*ptr == '!') {
			    count++;
			}
		    }
		    perbyte_cpu = (int) pow(10.0, (double) count);
		}
	    }else if (!strcasecmp(param->name, "percall_cpu")) {
		if (!sscanf(param->val, "%d", &percall_cpu)) {
		    for (count = 0, ptr = param->val; *ptr != '\0'; ptr++) {
			if (*ptr == '!') {
			    count++;
			}
		    }
		    percall_cpu = (int) pow(10.0, (double) count);
		}
	    }else if (!strcasecmp(param->name, "outin_ratio")) {
		outin_ratio = atoi(param->val);
	    }
	}
#ifndef P95
    } else if (!namestrcmp(languageName, "sql")) {
#else /* P95 */
    } else if (!strcmp(languageName, "sql")) {
#endif /* P95 */
	canCache = false;
	trusted = true;
	
#ifndef P95
	/* query optimizer groks postquel, these are meaningless */
#else /* P95 */
	/* query optimizer groks sql, these are meaningless */
#endif /* P95 */
	perbyte_cpu = percall_cpu = 0;
	byte_pct = outin_ratio = 100;
    } else {
#ifndef P95
	elog(WARN, "DefineFunction: language \"-.*s\" is not supported",
	     NAMEDATALEN, languageName->data);
#else /* P95 */
	elog(WARN, "DefineFunction: language '%s' is not supported",
	     languageName);
#endif /* P95 */
    }
    
    /* ----------------
     * handle "[ arg is (...) ]"
     * XXX fix optional arg handling below
     * ----------------
     */
    argList = stmt->defArgs;
    
#ifndef P95
    if (!namestrcmp(languageName, "c")) {
#else /* P95 */
    if ( strcmp(languageName, "c") == 0 ||
	 strcmp(languageName, "internal") == 0 ) {
#endif /* P95 */
	prosrc_str = "-";
	probin_str = stmt->as;
    } else {
	prosrc_str = stmt->as;
	probin_str = "-";
    }
    
    /* C is stored uppercase in pg_language */
#ifndef P95
    if (!namestrcmp(languageName, "c")) {
	languageName->data[0] = 'C';
#else /* P95 */
    if (!strcmp(languageName, "c")) {
	languageName[0] = 'C';
#endif /* P95 */
    }
    
    /* ----------------
     *	now have ProcedureDefine do all the work..
     * ----------------
     */
#ifndef P95
    ProcedureDefine(&pronameData,
#else /* P95 */
    ProcedureCreate(proname,
#endif /* P95 */
		    returnsSet,
#ifndef P95
		    &prorettypeData,
#else /* P95 */
		    prorettype,
#endif /* P95 */
		    languageName,
		    prosrc_str,		/* converted to text later */
		    probin_str,		/* converted to text later */
		    canCache,
		    trusted,
		    byte_pct,
		    perbyte_cpu,
		    percall_cpu, 
		    outin_ratio,
		    argList,
		    dest);
    
#ifndef P95
    /* XXX free palloc'd memory */
#endif /* P95 */
}

/* --------------------------------
 * DefineOperator--
 *
 *	this function extracts all the information from the
 *	parameter list generated by the parser and then has
 * [Mariposa]
 *	OperatorDefine() do all the actual work.
 * [Postgres95]
 *	OperatorCreate() do all the actual work.
 *
 * 'parameters' is a list of DefElem
 * --------------------------------
 */
void
#ifndef P95
DefineOperator(Name name,	/* XXX actually a String */
#else /* P95 */
DefineOperator(char *oprName,	
#endif /* P95 */
	       List *parameters)
{
    uint16 	precedence=0; 		/* operator precedence */
    bool	canHash=false; 		/* operator hashes */
    bool	isLeftAssociative=true;	/* operator is left associative */
#ifndef P95
    char16	oprNameData;		/* operator name */
    Name 	functionName=NULL; 	/* function for operator */
    Name 	typeName1=NULL;	 	/* first type name */
    Name 	typeName2=NULL;	 	/* second type name */
    Name 	commutatorName=NULL; 	/* optional commutator operator name */
    Name 	negatorName=NULL; 	/* optional negator operator name */
    Name 	restrictionName=NULL;	/* optional restrict. sel. procedure */
    Name 	joinName=NULL;	 	/* optional join sel. procedure name */
    Name 	sortName1=NULL;	 	/* optional first sort operator */
    Name 	sortName2=NULL;	 	/* optional second sort operator */
#else /* P95 */
    char *functionName=NULL; 	/* function for operator */
    char *typeName1=NULL;	 	/* first type name */
    char *typeName2=NULL;	 	/* second type name */
    char *commutatorName=NULL; 	/* optional commutator operator name */
    char *negatorName=NULL; 	/* optional negator operator name */
    char *restrictionName=NULL;	/* optional restrict. sel. procedure */
    char *joinName=NULL;	 	/* optional join sel. procedure name */
    char *sortName1=NULL;	 	/* optional first sort operator */
    char *sortName2=NULL;	 	/* optional second sort operator */
#endif /* P95 */
    List	*pl;

#ifndef P95
    /* ----------------
     *	sanity checks
     *
     * XXX	Checking of operator "name" validity
     *		(16 characters?) is needed.
     * ----------------
     */
    AssertArg(NameIsValid(name));
    
    (void) namestrcpy(&oprNameData, (char *) name);

#endif /* P95 */
    /*
     * loop over the definition list and extract the information we need.
     */
    foreach (pl, parameters) {
	DefElem *defel = (DefElem *)lfirst(pl);

	if (!strcasecmp(defel->defname, "leftarg")) {
	    /* see gram.y, must be setof */
	    if (nodeTag(defel->arg)==T_TypeName) 
		elog(WARN, "setof type not implemented for leftarg");

	    if (nodeTag(defel->arg)==T_String) {
#ifndef P95
		typeName1 = defGetName(defel);
#else /* P95 */
		typeName1 = defGetString(defel);
#endif /* P95 */
	    }else {
		elog(WARN, "type for leftarg is malformed.");
	    }
	} else if (!strcasecmp(defel->defname, "rightarg")) {
	    /* see gram.y, must be setof */
	    if (nodeTag(defel->arg)==T_TypeName) 
		elog(WARN, "setof type not implemented for rightarg");

	    if (nodeTag(defel->arg)==T_String) {
#ifndef P95
		typeName2 = defGetName(defel);
#else /* P95 */
		typeName2 = defGetString(defel);
#endif /* P95 */
	    }else {
		elog(WARN, "type for rightarg is malformed.");
	    }
	} else if (!strcasecmp(defel->defname, "procedure")) {
#ifndef P95
	    functionName = defGetName(defel);
#else /* P95 */
	    functionName = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "precedence")) {
#ifndef P95
	    precedence = defGetInteger(defel);
#else /* P95 */
	    /* NOT IMPLEMENTED (never worked in v4.2) */
	    elog(NOTICE, "CREATE OPERATOR: precedence not implemented");
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "associativity")) {
#ifndef P95
	    /* ----------------
	     * handle "[ associativity = (left|right|none|any) ]"
	     *	
	     * XXX Associativity code below must be fixed when the catalogs and
	     * XXX the planner/executor support proper associativity semantics.
	     * ----------------
	     */
	    String string = defGetString(defel);
	
	    if (!strcmp(string, "right")) {
		isLeftAssociative = false;
	    } else if (strcmp(string, "left") &&
		       strcmp(string, "none") &&
		       strcmp(string, "any")) {
		elog(WARN, "Define: precedence = what?");
	    } else {
		isLeftAssociative = true;
	    }
#else /* P95 */
	    /* NOT IMPLEMENTED (never worked in v4.2) */
	    elog(NOTICE, "CREATE OPERATOR: associativity not implemented");
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "commutator")) {
#ifndef P95
	    commutatorName = defGetName(defel);
#else /* P95 */
	    commutatorName = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "negator")) {
#ifndef P95
	    negatorName = defGetName(defel);
#else /* P95 */
	    negatorName = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "restrict")) {
#ifndef P95
	    restrictionName = defGetName(defel);
#else /* P95 */
	    restrictionName = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "join")) {
#ifndef P95
	    joinName = defGetName(defel);
#else /* P95 */
	    joinName = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "hashes")) {
	    canHash = TRUE;
	} else if (!strcasecmp(defel->defname, "sort1")) {
	    /* ----------------
	     * XXX ( ... [ , sort1 = oprname ] [ , sort2 = oprname ] ... )
	     * XXX is undocumented in the reference manual source as of
	     * 89/8/22.
	     * ----------------
	     */
#ifndef P95
	    sortName1 = defGetName(defel);
#else /* P95 */
	    sortName1 = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "sort2")) {
#ifndef P95
	    sortName2 = defGetName(defel);
#else /* P95 */
	    sortName2 = defGetString(defel);
#endif /* P95 */
	} else {
	    elog(NOTICE, "DefineOperator: attribute \"%s\" not recognized",
		 defel->defname);
	}
    }

    /*
     * make sure we have our required definitions
     */
    if (functionName==NULL) {
	elog(WARN, "Define: \"procedure\" unspecified");
    }
    
    /* ----------------
	 * [Mariposa]
     *	now have OperatorDefine do all the work..
	 * [Postgres95]
     *	now have OperatorCreate do all the work..
     * ----------------
     */
#ifndef P95
    OperatorDefine(&oprNameData,	/* operator name */
#else /* P95 */
    OperatorCreate(oprName,	/* operator name */
#endif /* P95 */
		   typeName1,		/* first type name */
		   typeName2,		/* second type name */
		   functionName,	/* function for operator */
		   precedence,		/* operator precedence */
		   isLeftAssociative,	/* operator is left associative */
		   commutatorName,	/* optional commutator operator name */
		   negatorName,		/* optional negator operator name */
		   restrictionName,	/* optional restrict. sel. procedure */
		   joinName,		/* optional join sel. procedure name */
		   canHash,		/* operator hashes */
		   sortName1,		/* optional first sort operator */
		   sortName2);		/* optional second sort operator */
    
#ifndef P95
    /* XXX free palloc'd memory */
#endif /* P95 */
}

/* -------------------
 *  DefineAggregate
 * ------------------
 */
void
#ifndef P95
DefineAggregate(Name name,		/* XXX actually a String */
		List *parameters)
#else /* P95 */
DefineAggregate(char *aggName, List *parameters)

#endif /* P95 */
{
#ifndef P95
    char16		aggNameData;
    Name		stepfunc1Name = (Name) NULL;
    Name		stepfunc2Name = (Name) NULL;
    Name		finalfuncName = (Name) NULL;
    Name		baseType = (Name) NULL;
    Name		stepfunc1Type = (Name) NULL;
    Name		stepfunc2Type = (Name) NULL;
    struct varlena	*primVal = (struct varlena *) NULL;
    struct varlena	*secVal = (struct varlena *) NULL;
    List		*pl;
#else /* P95 */
    char *stepfunc1Name = NULL;
    char *stepfunc2Name = NULL;
    char *finalfuncName = NULL;
    char *baseType = NULL;
    char *stepfunc1Type = NULL;
    char *stepfunc2Type = NULL;
    char *init1 = NULL;
    char *init2 = NULL;
    List *pl;
#endif /* P95 */
    
#ifndef P95
    /* sanity checks...name validity */
    
    AssertArg(NameIsValid(name));
    
    (void) namestrcpy(&aggNameData, (char *) name);

#endif /* P95 */
    foreach (pl, parameters) {
	DefElem *defel = (DefElem *)lfirst(pl);

	/*
	 * sfunc1
	 */
	if (!strcasecmp(defel->defname, "sfunc1")) {
#ifndef P95
	    stepfunc1Name = defGetName(defel);
#else /* P95 */
	    stepfunc1Name = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "basetype")) {
#ifndef P95
	    baseType = defGetName(defel);
#else /* P95 */
	    baseType = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "stype1")) {
#ifndef P95
	    stepfunc1Type = defGetName(defel);
#else /* P95 */
	    stepfunc1Type = defGetString(defel);
#endif /* P95 */

	/*
	 * sfunc2
	 */
	} else if (!strcasecmp(defel->defname, "sfunc2")) {
#ifndef P95
	    stepfunc2Name = defGetName(defel);
#else /* P95 */
	    stepfunc2Name = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "stype2")) {
#ifndef P95
	    stepfunc2Type = defGetName(defel);
#else /* P95 */
	    stepfunc2Type = defGetString(defel);
#endif /* P95 */
	/*
	 * final
	 */
	} else if (!strcasecmp(defel->defname, "finalfunc")) {
#ifndef P95
	    finalfuncName = defGetName(defel);
#else /* P95 */
	    finalfuncName = defGetString(defel);
#endif /* P95 */
	/*
	 * initial conditions
	 */
	} else if (!strcasecmp(defel->defname, "initcond1")) {
#ifndef P95
	    primVal = textin(defGetString(defel));
#else /* P95 */
	    init1 = defGetString(defel);
#endif /* P95 */
	} else if (!strcasecmp(defel->defname, "initcond2")) {
#ifndef P95
	    secVal = textin(defGetString(defel));
#else /* P95 */
	    init2 = defGetString(defel);
#endif /* P95 */
	} else {
	    elog(NOTICE, "DefineAggregate: attribute \"%s\" not recognized",
		 defel->defname);
	}
    }

    /*
     * make sure we have our required definitions
     */
#ifdef P95
    if (baseType==NULL) 
	elog(WARN, "Define: \"basetype\" unspecified");
#endif /* P95 */
    if (stepfunc1Name!=NULL) {
#ifndef P95
	if (baseType==NULL) 
	    elog(WARN, "Define: \"basetype\" unspecified");
#endif /* P95 */
	if (stepfunc1Type==NULL) 
	    elog(WARN, "Define: \"stype1\" unspecified");
    }
    if (stepfunc2Name!=NULL) {
	if (stepfunc2Type==NULL)
	    elog(WARN, "Define: \"stype2\" unspecified");
    }

    /*
	 * [Mariposa/Postgres95]
     * Most of the argument-checking is done inside of AggregateDefine
     * Most of the argument-checking is done inside of AggregateCreate
     */
#ifndef P95
    AggregateDefine(&aggNameData, 	/* aggregate name */
#else /* P95 */
    AggregateCreate(aggName,         	/* aggregate name */
#endif /* P95 */
		    stepfunc1Name,	/* first step function name */
		    stepfunc2Name,	/* second step function name */
		    finalfuncName,	/* final function name */
		    baseType,		/* type of object being aggregated */
		    stepfunc1Type,	/* return type of first function */
		    stepfunc2Type,	/* return type of second function */
#ifndef P95
		    (char *) primVal,	/* first initial condition */
		    (char *) secVal);	/* second initial condition */
#else /* P95 */
		    init1,	/* first initial condition */
		    init2);	/* second initial condition */
#endif /* P95 */
    
    /* XXX free palloc'd memory */
}

/*
 * DefineType --
 *	Registers a new type.
 *
 * Exceptions:
 *	XXX
 */
void
#ifndef P95
DefineType(Name name,		/* XXX actually a String */
	   List *parameters)
#else /* P95 */
DefineType(char *typeName, List *parameters)
#endif /* P95 */
{
    int16		internalLength= 0;	/* int2 */
    int16		externalLength= 0;	/* int2 */
#ifndef P95
    char16		typeNameData;
    Name		elemName = NULL;
    Name		inputName = NULL;
    Name		outputName = NULL;
    Name		sendName = NULL;
    Name		receiveName = NULL;
#else /* P95 */
    char *elemName = NULL;
    char *inputName = NULL;
    char *outputName = NULL;
    char *sendName = NULL;
    char *receiveName = NULL;
#endif /* P95 */
    char		*defaultValue = NULL;	/* Datum */
    bool		byValue = false;
    char		delimiter = DEFAULT_TYPDELIM;
#ifndef P95
    char16		shadow_type;
    static char16	ain = { "array_in" };
    static char16	aout = { "array_out" };
#else /* P95 */
    char *shadow_type;
#endif /* P95 */
    List		*pl;
    char		alignment = 'i';	/* default alignment */

#ifndef P95
    /* ----------------
     *	sanity checks
     *
     * XXX	Checking of operator "name" validity
     *		(16 characters?) is needed.
     * ----------------
     */
    AssertArg(NameIsValid(name));
    
    (void) namestrcpy(&typeNameData, (char *) name);
    
#endif /* P95 */
    /*
     * Type names can only be 15 characters long, so that the shadow type
     * can be created using the 16th character as necessary.
     */
#ifndef P95
    if (NameComputeLength(&typeNameData) >= (NAMEDATALEN - 1)) {
#else /* P95 */
    if (strlen(typeName) >= (NAMEDATALEN - 1)) {
#endif /* P95 */
	elog(WARN, "DefineType: type names must be %d characters or less",
	     NAMEDATALEN - 1);
    }

    foreach(pl, parameters) {
	DefElem *defel = (DefElem*)lfirst(pl);

	if (!strcasecmp(defel->defname, "internallength")) {
#ifndef P95
	    internalLength = defGetInteger(defel);
#else /* P95 */
	    internalLength = defGetTypeLength(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "externallength")) {
#ifndef P95
	    externalLength = defGetInteger(defel);
#else /* P95 */
	    externalLength = defGetTypeLength(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "input")) {
#ifndef P95
	    inputName = defGetName(defel);
#else /* P95 */
	    inputName = defGetString(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "output")) {
#ifndef P95
	    outputName = defGetName(defel);
#else /* P95 */
	    outputName = defGetString(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "send")) {
#ifndef P95
	    sendName = defGetName(defel);
#else /* P95 */
	    sendName = defGetString(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "delimiter")) {
	    char *p = defGetString(defel);
	    delimiter = p[0];
	}else if (!strcasecmp(defel->defname, "receive")) {
#ifndef P95
	    receiveName = defGetName(defel);
#else /* P95 */
	    receiveName = defGetString(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "element")) {
#ifndef P95
	    elemName = defGetName(defel);
#else /* P95 */
	    elemName = defGetString(defel);
#endif /* P95 */
	}else if (!strcasecmp(defel->defname, "default")) {
	    defaultValue = defGetString(defel);
	}else if (!strcasecmp(defel->defname, "passedbyvalue")) {
	    byValue = true;
	}else if (!strcasecmp(defel->defname, "alignment")) {
	    char *a = defGetString(defel);
	    if (!strcasecmp(a, "double")) {
		alignment = 'd';
	    } else if (!strcasecmp(a, "int")) {
		alignment = 'i';
	    } else {
		elog(WARN, "DefineType: \"%s\" alignment  not recognized",
		     a);
	    }
	}else {
	    elog(NOTICE, "DefineType: attribute \"%s\" not recognized",
		 defel->defname);
	}
    }

    /*
     * make sure we have our required definitions
     */
    if (inputName==NULL)
	elog(WARN, "Define: \"input\" unspecified");
    if (outputName==NULL)
	elog(WARN, "Define: \"output\" unspecified");
    
    /* ----------------
	 * [Mariposa/Postgres95]
     *	now have TypeDefine do all the real work.
     *	now have TypeCreate do all the real work.
     * ----------------
     */
#ifndef P95
    (void) TypeDefine(&typeNameData,	/* type name */
#else /* P95 */
    (void) TypeCreate(typeName,	/* type name */
#endif /* P95 */
		      InvalidOid,  /* relation oid (n/a here) */
		      internalLength,	/* internal size */
		      externalLength,	/* external size */
		      'b',		/* type-type (base type) */
		      delimiter,	/* array element delimiter */
		      inputName,	/* input procedure */
		      outputName,	/* output procedure */
		      sendName,		/* send procedure */
		      receiveName,	/* receive procedure */
		      elemName,		/* element type name */
		      defaultValue,	/* default type value */
		      byValue,		/* passed by value */
		      alignment);
    
    /* ----------------
     *  When we create a true type (as opposed to a complex type)
     *  we need to have an shadow array entry for it in pg_type as well.
     * ----------------
     */
#ifndef P95
    namestrcpy(&shadow_type, "_");
    namestrcat(&shadow_type, (char *) name);
    
    (void) TypeDefine(&shadow_type,	/* type name */
#else /* P95 */
    shadow_type = makeArrayTypeName(typeName);

    (void) TypeCreate(shadow_type,	/* type name */
#endif /* P95 */
		      InvalidOid,  /* relation oid (n/a here) */
		      -1,		/* internal size */
		      -1,		/* external size */
		      'b',		/* type-type (base type) */
		      DEFAULT_TYPDELIM,	/* array element delimiter */
#ifndef P95
		      &ain,		/* input procedure */
		      &aout,		/* output procedure */
		      &aout,		/* send procedure */
		      &ain,		/* receive procedure */
		      &typeNameData,	/* element type name */
#else /* P95 */
		      "array_in",		/* input procedure */
		      "array_out",		/* output procedure */
		      "array_out",		/* send procedure */
		      "array_in",		/* receive procedure */
		      typeName,	/* element type name */
#endif /* P95 */
		      defaultValue,	/* default type value */
		      false,		/* never passed by value */
		      alignment);
    
#ifndef P95
    /* XXX free palloc'd memory */
#else /* P95 */
    pfree(shadow_type);
#endif /* P95 */
}

#ifndef P95
static Name
defGetName(DefElem *def)
{
    Name name;

    if (nodeTag(def->arg)!=T_String)
	elog(WARN, "Define: \"%s\" = what?", def->defname);
    if (strlen((char*)strVal(def->arg)) > NAMEDATALEN) {
	elog(WARN, "Define: \"%s\" is too long to be an identifier",
	     strVal(def->arg));
    }
    if (!(name = (Name) palloc(sizeof(char16) + 1))) {
	elog(WARN, "Define: palloc failed!");
    }
    memset(name->data,0,NAMEDATALEN);
    (void) namestrcpy(name, strVal(def->arg));
    
    return (name);
}

#endif /* P95 */
static char *
defGetString(DefElem *def)
{
    if (nodeTag(def->arg)!=T_String)
	elog(WARN, "Define: \"%s\" = what?", def->defname);
    return (strVal(def->arg));
}

static int 
#ifndef P95
defGetInteger(DefElem *def)
#else /* P95 */
defGetTypeLength(DefElem *def)
#endif /* P95 */
{
#ifndef P95
    if (nodeTag(def->arg)!=T_Integer)
	elog(WARN, "Define: \"%s\" = what?", def->defname);
    return (intVal(def->arg));
#else /* P95 */
    if (nodeTag(def->arg)==T_Integer)
	return (intVal(def->arg));
    else if (nodeTag(def->arg)==T_String &&
	     !strcasecmp(strVal(def->arg),"variable"))
	return -1;	/* variable length */

    elog(WARN, "Define: \"%s\" = what?", def->defname);
    return -1;
#endif /* P95 */
}
