/*-------------------------------------------------------------------------
 *
 * fragments.c--
 *    Commands dealing with fragments.
 *
 *    Interface routines are FragmentMove, FragmentTake, FragmentCopy, and
 *    SplitFragment. These are called via commands to a postgres backend.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/commands/fragments.c,v 1.45 1996/06/14 20:22:39 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <rpc/types.h>
#include <pthread.h>
#include "postgres.h"
#include "commands/fragments.h"
#include "mariposa_rpc.h"
#include "rpc/rpcutils.h"
#include "rpc/rexec.h"
#include "sitemgr/sitemgr.h"
#include "utils/elog.h"

#include "access/attnum.h"
#include "access/heapam.h"
#include "storage/bufmgr.h"	/* for access/transam.h */
#include "storage/fd.h"
#include "access/transam.h"	/* for GetNewObjectId */
#include "access/xact.h"
#include "catalog/catname.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_contract.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "catalog/indexing.h"
#include "rpc/fragutils.h"
#include "utils/syscache.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "utils/tqual.h"
#include "nodes/parsenodes.h"	/* for MoveFragStmt */
#include "nameserver_mk.h"

#define SENDBUFSIZE 7168		/* copied from exchange.c */

extern unsigned short tcpPort;		/* set in ??? */

/* ----------------
 *  These do the basic work for FragmentMove, FragmentTake, and are also
 *  used by AsyncMoveFragment. (tcop/asynchFrag.c)
 * ----------------
 */
void ExecFragmentMove(int rExecAddr, int rExecPort, FragXArg *fragxarg,
		      int newloc, char *relname);
void ExecFragmentTake(int rExecAddr, int rExecPort, FragXArg *fragxarg);

/* ----------------
 *  Support routines for fragment transfer/copying.
 * ----------------
 */
static void StartFragX(OfferBuffer *offbuf, int loc, int *execAddr,
		       int *execPort);
static void ExtractPrep(RpcConn *rpcc, FragXArg *fragxarg);

static void PutFragCatalog(RpcConn *rpcc, FragXArg *fragxarg);
static void PutScan(RpcConn *rpcc, FragDataBuffer *databuf, Relation rdesc,
		    TimeQual tqual);
static void PutFragData(RpcConn *rpcc, FragXArg *fragxarg);
static void GetFragCatalog(RpcConn *rpcc, FragXArg *fragxarg,
			   Relation *newFragDesc, Relation *newFragArchDesc);
static void GetFragData(RpcConn *rpcc, FragXArg *fragxarg,
			Relation newFragDesc, Relation newArchFragDesc);

/* ----------------
 *  Support routines for splitting fragments.
 * ----------------
 */
static Oid GetNewFragmentRel(char *fragName, Relation rdesc);
static void EstablishNewFragment(char *fragName, Relation rdesc, char *lokey,
				 char *hikey, Oid frastoreid, Oid fralogicalid);
Oid FindMedianOID(HeapScanDesc sd);
static void SplitFragDoInserts(HeapScanDesc sdrel, HeapScanDesc sdarch,
			       Relation obrel, Relation nrel1, Relation nrel2,
			       Relation arel1, Relation arel2);

/* ----------------
 *  Support routines for parser.
 * ----------------
 */
void FillFragStmt(MoveFragStmt *stmt, char *fragname, int location,
		  int direction);
static void FetchPgFragTuple(Form_pg_fragment fillMeIn, Oid frastoreid);


/*****************************************************************************
 * Routines for transferring/copying fragments.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	StartFragX
 *
 *	Connects to remote sitemgr that we wish to participate in the
 *	fragment transfer/copy. If it's ok with them, a transaction is
 *	started at their site, we get the id for two phase commit reference,
 *	and we get the address of the backend they've allocated.
 *
 * ----------------------------------------------------------------
 */
static void
StartFragX(OfferBuffer *offbuf, int loc, int *execAddr, int *execPort)
{
    RpcConn		*sm_rpcc;
    TransactionId	remotexactid;

    RExecInit();

    /*
     * Notify the remote site of our intentions to perform transfer according
     * to offbuf, so it can prepare. If it is prepared, it sends us a valid
     * xactid.
     */
    sm_rpcc = Rpc_connectSitemgr(loc, 0, "StartFragX", RPC_ONERR_WARN);
    if (!Rpc_call(sm_rpcc, RPC_RECEIVESHORTOFFER, offbuf, &remotexactid)) {
	Rpc_disconnect(sm_rpcc);
	elog(WARN, "StartFragX: Offer RPC call failed.");
    }
    Rpc_disconnect(sm_rpcc);

    if (remotexactid <= 0) {	/* 0 and -1 are invalid transaction ids. */
	elog(WARN, "StartFragX: Offer rejected.");
    }
    offbuf->details.OfferDetails_u.shortBuy.remotexactid = remotexactid;

    /*
     * Once prepared, the remote site will deliver us an address to one of
     * of its execs, which we connect to.
     */
    RExecWait(execAddr, execPort, NULL);
}

/* ----------------------------------------------------------------
 *    	ExtractPrep
 *
 *	For fragment takes and copies, does the initialization stuff
 *	that needs to run in the main transaction at the remote site.
 *
 * ----------------------------------------------------------------
 */
static void
ExtractPrep(RpcConn *rpcc, FragXArg *fragxarg)
{
    RPCErrorCode	result;

    if (!Rpc_call(rpcc, RPC_EXTRACTPREPARATIONS, fragxarg, &result)) {
	Rpc_disconnect(rpcc);
	elog(WARN, "ExtractPrep: EXTRACTPREPARATIONS rpc call failed.");
    }

    if (result != SUCCESS) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "ExtractPrep: EXTRACTPREPARATIONS error %d.", result);
    }
}

/* ----------------------------------------------------------------
 *    	FragmentMove
 *
 *	Moves fragment to another site (specified by newloc.)
 *
 * ----------------------------------------------------------------
 */
void
FragmentMove(Oid fraclassid, Oid frastoreid, int newloc)
{
    OfferBuffer		offbuf;
    FragXArg		*fragxarg;
    int			execAddr;
    int			execPort;
    Relation		rdesc;
    NameData		relNameData;
    char		relname[MAXPATHLEN];

    rdesc = heap_open(frastoreid);
    if (!RelationIsValid(rdesc)) {
	elog(WARN, "FragmentMove: Can't open fragment %d.", frastoreid);
    }

    if (rdesc->rd_iscopied) {
	elog(WARN, "We're not allowed to move copied fragments yet!");
    }

    namecpy(&relNameData, rdesc->rd_rel->relname.data);	/* unterminated */
    namestrcpy(relname, rdesc->rd_rel->relname.data);	/* terminated */

    heap_close(rdesc);

    /*
     * Prepare the short protocol offer for receiving site to buy the
     * fragment.
     */
    (void) memset((void *) &offbuf, 0, sizeof(offbuf));
    offbuf.budget.cost = 100.0;         /* bogus */
    offbuf.budget.delay = 100.0;        /* bogus */
    offbuf.details.protocol = SHORT_BUY_FRAGMENT;
    offbuf.requester.addr = getMyAddr();
    offbuf.requester.port = tcpPort;
    offbuf.requester.hostid = getMyHostId();
    fragxarg = &(offbuf.details.OfferDetails_u.shortBuy);

    fragxarg->fragxtype = BUYFRAGMENT;
    fragxarg->qepFragId = newQepFragId();
    fragxarg->sender.addr = offbuf.requester.addr;
    fragxarg->sender.port = tcpPort;
    fragxarg->sender.hostid = offbuf.requester.hostid;
    fragxarg->receiver.hostid = newloc;
    fragxarg->fraclassid = fraclassid;
    fragxarg->frastoreid = frastoreid;

    /*
     * Send offer to remote sitemgr and wait for address of backend.
     */
    StartFragX(&offbuf, newloc, &execAddr, &execPort);

    /*
     * Perform the move.
     */
    ExecFragmentMove(execAddr, execPort, fragxarg, newloc, relname);
}

/* ----------------------------------------------------------------
 *	ExecFragmentMove
 *
 *	Does the actual catalog & data transfer and updates the catalog.
 *	Also called by AsyncMoveFragment.
 *
 * ----------------------------------------------------------------
 */
void
ExecFragmentMove(int rExecAddr, int rExecPort, FragXArg *fragxarg, int newloc,
		 char *relationName)
{
    char		*relname;
    char		rnamebuf[MAXPATHLEN];
    Relation		fragdesc;
    RpcConn		*exec_rpcc;
    char		query[1000];

    fragdesc = heap_open(fragxarg->frastoreid);
    if (!RelationIsValid(fragdesc)) {
	elog(WARN, "ExecFragmentMove: Can't open fragment %d.",
	     fragxarg->frastoreid);
    }

    if (fragdesc->rd_iscopied) {
	elog(WARN, "We're not allowed to move copied fragments yet!");
    }

    /*
     * If we don't know the relationName, look it up and
     * set relname to point to the static buffer. Otherwise,
     * set relname to point to relationName.
     */
    if (relationName == NULL) {
	namestrcpy(rnamebuf, fragdesc->rd_rel->relname.data); /* terminated */
	relname = rnamebuf;
    }
    else {
        relname = relationName;
    }

    /*
     * Before transferring, vacuum the fragment.
     */
    vacuum(relname);

    exec_rpcc = Rpc_connectExec(rExecAddr, rExecPort, "ExecFragmentMove",
				RPC_ONERR_WARN);

    /*
     * Set write lock on the fragment, so no one else messes with it while.
     * we move it.
     */
    RelationSetLockForWrite(fragdesc);

    /*
     * Transfer catalog information (pg_class, pg_attribute,) then transfer
     * data.
     */
    PutFragCatalog(exec_rpcc, fragxarg);
    PutFragData(exec_rpcc, fragxarg);

    /*
     * Destroy archive relation at our site.
     */
    DestroyArchiveRelation(fragxarg->frastoreid);

    /*
     * Destroy as much as we can of what's left of the moved fragment.
     */
    if (fragxarg->fraclassid != fragxarg->frastoreid) {
	heap_destroy(relname);
    } else {
	File fd;

	RelationRemoveIndexes(fragdesc);
	fd = FileNameOpenFile(relname, O_WRONLY, 0600);
	if (fd >= 0) {
	    (void) FileTruncate(fd, 0);
	    (void) FileClose(fd);
	}
    }

    heap_close(fragdesc);

    /*
     * Now update pg_fragment on our site, to reflect the move.
     */
    sprintf(query,
	    "DELETE FROM pg_fragment WHERE fraclassid='%d'::oid and frastoreid='%d'::oid",
	    fragxarg->fraclassid,
	    fragxarg->frastoreid);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    /*
     * Update pg_class and pg_type as well.  Eventually, we should
     * make the fragment movement independent of classes and types.
     */
    sprintf(query,
	    "DELETE FROM pg_class WHERE oid = '%d'::oid",
	    fragxarg->fraclassid);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    sprintf(query,
	    "DELETE FROM pg_type WHERE typrelid = '%d'::oid",
	    fragxarg->fraclassid);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    /*
     * Tell remote site to commit.
     */
    CommitRemoteSite(exec_rpcc, fragxarg->remotexactid);
}

/* ----------------------------------------------------------------
 *	FragmentTake
 *
 *	Takes fragment from another site (specified by sourceloc.)
 *
 *	A SELLFRAGMENT RPC is sent over to the site that has the
 *	fragment now; FragXArg is reused here to send the
 *	parameters that describe the desired fragment.
 *
 * ----------------------------------------------------------------
 */
void
FragmentTake(Oid fraclassid, Oid frastoreid, int sourceloc)
{
    OfferBuffer		offbuf;
    FragXArg		*fragxarg;
    int			execAddr;
    int			execPort;
    
    /*
     * Prepare the short protocol offer for source site to sell the
     * fragment.
     */
    (void) memset((void *) &offbuf, 0, sizeof(offbuf));
    offbuf.budget.cost = 100.0;         /* bogus */
    offbuf.budget.delay = 100.0;        /* bogus */
    offbuf.details.protocol = SHORT_SELL_FRAGMENT;
    fragxarg = &(offbuf.details.OfferDetails_u.shortSell);

    fragxarg->fragxtype = SELLFRAGMENT;
    fragxarg->qepFragId = newQepFragId();
    fragxarg->sender.addr = getMyAddr();
    fragxarg->sender.port = tcpPort;
    fragxarg->sender.hostid = getMyHostId();
    fragxarg->receiver.hostid = getMyHostId();
    fragxarg->fraclassid = fraclassid;
    fragxarg->frastoreid = frastoreid;

    /*
     * Send offer to remote sitemgr and wait for address of backend.
     */
    StartFragX(&offbuf, sourceloc, &execAddr, &execPort);

    /*
     * Perform the take.
     */
    ExecFragmentTake(execAddr, execPort, fragxarg);
}

/* ----------------------------------------------------------------
 *	ExecFragmentTake
 *
 *	Does the actual catalog & data transfer and updates the catalog.
 *	Also called by AsyncMoveFragment.
 *
 * ----------------------------------------------------------------
 */
void
ExecFragmentTake(int rExecAddr, int rExecPort, FragXArg *fragxarg)
{
    RpcConn		*exec_rpcc;
    Relation		newFragDesc;
    Relation		newFragArchDesc;

    exec_rpcc = Rpc_connectExec(rExecAddr, rExecPort,
                                "ExecFragmentTake", RPC_ONERR_WARN);
    
    /*
     * Tell remote site how to set up for us.
     */
    ExtractPrep(exec_rpcc, fragxarg);

    /*
     * Transfer catalog information (pg_class, pg_fragment, pg_attribute,)
     * and then transfer the data.
     */
    GetFragCatalog(exec_rpcc, fragxarg, &newFragDesc, &newFragArchDesc);
    GetFragData(exec_rpcc, fragxarg, newFragDesc, newFragArchDesc);

    /*
     * Tell the remote site to commit.
     */
    CommitRemoteSite(exec_rpcc, fragxarg->remotexactid);
}

/* ----------------------------------------------------------------
 *	FragmentCopy
 *
 *	Sets up replica relationship with fragment from another site
 *	(specified by sourceloc.) 
 *
 *	For user copies, notify the source site, get the catalog info and
 *	data, and then setup update stream contracts and schedules at both
 *	the source and copy sites.
 *
 *	This routine is also used to setup metadata exchange for name
 *	service (if isnamesvc is set to true.) In that case, only the
 *	contracts are set up.
 *
 * ----------------------------------------------------------------
 */
void
FragmentCopy(Oid fraclassid, Oid frastoreid, int sourceloc, int period,
	     bool isreadonly, bool isnamesvc, bool isnsspecific)
{
    OfferBuffer		offbuf;
    FragXArg		*fragxarg;
    RPCErrorCode	result;
    RPCErrorCode	endresult;
    RpcConn		*sm_rpcc;
    RpcConn		*exec_rpcc;
    int			execAddr;
    int			execPort;
    Relation		newFragDesc;
    Relation		newFragArchDesc;
    Oid			contractid;
    CopySetupInfo	copysetup;
    char		query[1000];
    
    /*
     * Prepare the short protocol offer for source site to sell us
     * a copy of the fragment.
     */
    (void) memset((void *) &offbuf, 0, sizeof(offbuf));
    offbuf.budget.cost = 100.0;         /* bogus */
    offbuf.budget.delay = 100.0;        /* bogus */
    offbuf.details.protocol = SHORT_COPY_FRAGMENT;
    fragxarg = &(offbuf.details.OfferDetails_u.shortCopy);

    fragxarg->fragxtype = COPYFRAGMENT;
    fragxarg->qepFragId = newQepFragId();
    fragxarg->sender.addr = getMyAddr();
    fragxarg->sender.port = tcpPort;
    fragxarg->sender.hostid = getMyHostId();
    fragxarg->receiver.hostid = getMyHostId();
    fragxarg->fraclassid = fraclassid;
    fragxarg->frastoreid = frastoreid;

    /*
     * Send offer to remote sitemgr and wait for address of backend.
     */
    StartFragX(&offbuf, sourceloc, &execAddr, &execPort);
    exec_rpcc =  Rpc_connectExec(execAddr, execPort, "FragmentCopy",
				 RPC_ONERR_WARN);
    
    /*
     * For user copies, we gotta get the catalog info and data from them.
     * For name servcice none of that is necessary.
     */
    if (!isnamesvc) {
	/*
	 * Tell remote site how to set up for the copy transfer.
	 */
	ExtractPrep(exec_rpcc, fragxarg);

	/*
	 * Transfer copy of catalog information
	 * (pg_class, pg_fragment, pg_attribute) except for new filename for
	 * relation and new frastoreid for the fragment copy.
	 * Then transfer the data.
	 */
	GetFragCatalog(exec_rpcc, fragxarg, &newFragDesc, &newFragArchDesc);
	GetFragData(exec_rpcc, fragxarg, newFragDesc, newFragArchDesc);
    }

    Rpc_setCallerName(exec_rpcc, "FragmentCopy");

    /*
     * Generate the contractid.
     */
    GetNewObjectId(&contractid);

    /*
     * Setup copy at source site.
     */
    copysetup.contractid = contractid;
    copysetup.ambuyer = false;
    copysetup.isreadonly = isreadonly;
    copysetup.isnamesvc = isnamesvc;
    copysetup.isnsspecific = isnsspecific;
    copysetup.hereloc = sourceloc;
    copysetup.thereloc = getMyHostId();
    if (!isnamesvc) {
	copysetup.ourcopystoreid = frastoreid;
	copysetup.theircopystoreid = newFragDesc->rd_id;
    }
    else {
	copysetup.ourcopystoreid = copysetup.theircopystoreid = 0;
    }
    copysetup.period = period;

    /*
     * In both these error-handling cases, we don't have to deschedule the
     * send contract because that is the last thing that is done-- and if
     * there's an error, that couldn't have succeeded.
     */
    if (!Rpc_call(exec_rpcc, RPC_ESTABLISHCOPY, &copysetup, &result))
    {
	AbortRemoteSite(exec_rpcc, fragxarg->remotexactid);
	elog(WARN, "FragmentCopy: Remote EstablishCopy RPC call failed.");
    }
    if (result != SUCCESS) {
	AbortRemoteSite(exec_rpcc, fragxarg->remotexactid);
	elog(WARN, "FragmentCopy: Remote EstablishCopy error %d", result);
    }

    /*
     * Setup copy at this site.
     */
    copysetup.ambuyer = true;
    copysetup.hereloc = copysetup.thereloc;
    copysetup.thereloc = sourceloc;
    if (!isnamesvc) {
	copysetup.ourcopystoreid = newFragDesc->rd_id;
	copysetup.theircopystoreid = frastoreid;
    }
    else {
	copysetup.ourcopystoreid = copysetup.theircopystoreid = 0;
    }

    result = EstablishCopy(&copysetup);

    if (result != SUCCESS) {
	/*
	 * We messed up, cancel contract from remote site's Update Stream
	 * Queue, and tell remote site to abort.
	 */
	sm_rpcc = Rpc_connectSitemgr(sourceloc, 0, "FragmentCopy",
				     RPC_ONERR_WARN);
	Rpc_call(sm_rpcc, RPC_CP_UNSCHEDULESTREAM, &contractid, &endresult);
	Rpc_disconnect(sm_rpcc);

	AbortRemoteSite(exec_rpcc, fragxarg->remotexactid);
	elog(WARN, "FragmentCopy: Local EstablishCopy error %d.", result);
    }

    if (!isnamesvc) {
	/*
	 * To construct a hash index on the new fragment, we call
	 * CommandCounterIncrement so that the index creating procedure can
	 * access the changes to the system catalog that reflect the new
	 * fragment.
	 *
	 * When we copy an fragment, we force both the master and the copy to
	 * have hash indices on classoid. That way, when we want to pass delete
	 * updates we don't have to do a sequential scan to find the tuples at
	 * the receiving site.
	 */
	CommandCounterIncrement();

	DefineRelationOidHashIndex(newFragDesc);
    }

    /*
     * Tell remote site to commit.
     */
    CommitRemoteSite(exec_rpcc, fragxarg->remotexactid);
}


/*****************************************************************************
 * Support routines for fragment transferring/copying.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	PutFragCatalog
 *
 *	Gathers system catalog information about fragment to be transferred
 *	and ships it to the remote site to be stored there.
 *
 * ----------------------------------------------------------------
 */
void
PutFragCatalog(RpcConn *rpcc, FragXArg *fragxarg)
{
    FragCatalogInfo	catInfo;
    Relation		rd;
    Relation		rdesc;
    HeapScanDesc	sdesc;
    HeapTuple		tup;
    ScanKeyData		skey;
    bool		nullAttr;
    FragPartInfo	ofpi;
    int			natts;
    int			i;
    Tuple		*tupdesc;
    RPCErrorCode	result;

    elog(NOTICE, "Entering PutFragCatalog.");
    Rpc_setCallerName(rpcc, "PutFragCatalog");
    
    CopyFragXArg(fragxarg, &catInfo.transferInfo);

    /*
     * Handle pg_class (2 tuples: class and fragment.)
     */
    tup = SearchSysCacheTuple(RELOID,
			      (char *) ObjectIdGetDatum(fragxarg->fraclassid),
			      (char *) NULL,
			      (char *) NULL,
			      (char *) NULL);
    if (!HeapTupleIsValid(tup)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragCatalog: Cannot find class tuple in pg_class.");
    }
    catInfo.classTuple.Tuple_len = tup->t_len;
    catInfo.classTuple.Tuple_val = (char *)palloc(tup->t_len);
    memmove(catInfo.classTuple.Tuple_val, tup, tup->t_len);

    tup = SearchSysCacheTuple(RELOID,
			      (char *) ObjectIdGetDatum(fragxarg->frastoreid),
			      (char *) NULL,
			      (char *) NULL,
			      (char *) NULL);
    if (!HeapTupleIsValid(tup)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragCatalog: Cannot find fragment tuple in pg_class.");
    }
    catInfo.fragTuple.Tuple_len = tup->t_len;
    catInfo.fragTuple.Tuple_val = (char *)palloc(tup->t_len);
    memmove(catInfo.fragTuple.Tuple_val, tup, tup->t_len);

    /*
     * Handle pg_fragment.
     */
    ScanKeyEntryInitialize(&skey, 0x0, Anum_pg_fragment_frastoreid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(fragxarg->frastoreid));
    rd = heap_openr((Name)Name_pg_fragment);
    sdesc = heap_beginscan(rd, false, NowTimeQual, 1, &skey);

    if (!HeapTupleIsValid(tup = heap_getnext(sdesc, 0, NULL))) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragCatalog: Cannot find frag tuple in pg_fragment.");
    }

    catInfo.pgfragTuple.Tuple_len = tup->t_len;
    catInfo.pgfragTuple.Tuple_val = (char *)palloc(tup->t_len);
    memmove(catInfo.pgfragTuple.Tuple_val, tup, tup->t_len);

    /*
     * Get fragment info from fragment relation. This relation descriptor also
     * used to handle pg_attributes.
     */
    rdesc = heap_open(fragxarg->frastoreid);
    if (!RelationIsValid(rdesc)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragCatalog: Cannot open fragment %d",
	     fragxarg->frastoreid);
    }

    heap_endscan(sdesc);
    heap_close(rd);

    /*
     * Handle pg_attributes.
     */
    natts = rdesc->rd_rel->relnatts;
    tupdesc = (Tuple *)palloc(sizeof(Tuple) * natts);
    catInfo.attrTuples.TupleArray_len = natts;
    catInfo.attrTuples.TupleArray_val = tupdesc;
    for (i = 0; i < natts; i++) {
	tupdesc[i].Tuple_len = ATTRIBUTE_TUPLE_SIZE;
	tupdesc[i].Tuple_val = (char*)rdesc->rd_att[i];
    }
    heap_close(rdesc);
    
    /*
     * The RPC call here updates pg_fragment on the remote site.
     */
    if (!Rpc_call(rpcc, RPC_FRAGPUTCAT, &catInfo, &result)) {
	Rpc_disconnect(rpcc);
	elog(WARN, "PutFragCatalog: FRAGPUTCAT rpc call failed.");
    }

    if (result != SUCCESS) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragCatalog: FRAGPUTCAT error %d.", result);
    }

    elog(NOTICE, "Leaving PutFragCatalog.");

    return;
}

/* ----------------------------------------------------------------
 *	PutScan
 *
 *    	Scans relation, and sends all its tuples over the rpc connection.
 *
 * ----------------------------------------------------------------
 */
void PutScan(RpcConn *rpcc, FragDataBuffer *databuf, Relation rdesc,
	     TimeQual tqual)
{
    HeapScanDesc	scan;
    HeapTuple		htup;
    char		buffer[SENDBUFSIZE];
    int			nbytes = 0;
    int			numTuples = 0;
    int			totalTuples = 0;	/* for debugging */
    RPCErrorCode	result;
    
    elog(NOTICE, "Entering PutScan.");
    Rpc_setCallerName(rpcc, "PutScan");
    
    databuf->data.Data_val = buffer;

    /*
     * Fill in tuples until done.
     */
    scan = heap_beginscan(rdesc,
			  0,
			  tqual,
			  0, NULL);
    do {
	htup = heap_getnext(scan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    int len = htup->t_len;

	    if ((len + nbytes) > SENDBUFSIZE) {
		/*
		 * Send the buffer off and start again.
		 */
		databuf->numTuples = numTuples;
		databuf->data.Data_len = nbytes;

		if (!Rpc_call(rpcc, RPC_FRAGPUTDATA, databuf, &result)) {
		    Rpc_disconnect(rpcc);
		    elog(WARN, "PutScan: FRAGPUTDATA rpc call failed.");
		}

		if (result != SUCCESS) {
		    AbortRemoteSite(rpcc, databuf->transferInfo.remotexactid);
		    elog(WARN, "PutScan: FRAGPUTDATA error %d.", result);
		}

		totalTuples += numTuples;

		numTuples = 0;
		nbytes = 0;
	    }

	    memmove(buffer + nbytes, htup, len);
	    numTuples++;
	    nbytes += len;
	}
    } while (htup);

    heap_endscan(scan);

    /*
     * Send last buffer.
     */
    if (nbytes > 0) {
	databuf->numTuples = numTuples;
	databuf->data.Data_len = nbytes;

	if (!Rpc_call(rpcc, RPC_FRAGPUTDATA, databuf, &result)) {
	    Rpc_disconnect(rpcc);
	    elog(WARN, "PutScan: FRAGPUTDATA rpc call failed.");
	}

	if (result != SUCCESS) {
	    AbortRemoteSite(rpcc, databuf->transferInfo.remotexactid);
	    elog(WARN, "PutScan: FRAGPUTDATA error %d.", result);
	}
    }

    elog(NOTICE, "Leaving PutScan.");
}

/* ----------------------------------------------------------------
 *	PutFragData
 *
 *    	Transfer the actual tuples of the fragment to the remote site.
 *
 * ----------------------------------------------------------------
 */
void
PutFragData(RpcConn *rpcc, FragXArg *fragxarg)
{
    FragDataBuffer	databuf;
    Relation		fragdesc;
    RPCErrorCode	result;
    TimeQual		fetchArchiveTimeQual;
    
    elog(NOTICE, "Entering PutFragData.");
    
    CopyFragXArg(fragxarg, &databuf.transferInfo);
    
    /*
     * Open archive part of fragment and send it over.
     */
    fragdesc = (Relation) OpenArchiveRelation(fragxarg->frastoreid);
    if (!RelationIsValid(fragdesc)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragData: Cannot open fragment %d archive.",
	     fragxarg->frastoreid);
    }

    databuf.which = ARCHIVE_PART;
    fetchArchiveTimeQual =
	TimeFormRangedTimeQual(EPOCH_ABSTIME, CURRENT_ABSTIME);
    PutScan(rpcc, &databuf, fragdesc, fetchArchiveTimeQual);
    pfree(fetchArchiveTimeQual);

    heap_close(fragdesc);

    /*
     * Open current part of fragment and send it over.
     */
    fragdesc = heap_open(fragxarg->frastoreid);
    if (!RelationIsValid(fragdesc)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragData: Cannot open fragment %d.",
	     fragxarg->frastoreid);
    }

    databuf.which = CURRENT_PART;
    PutScan(rpcc, &databuf, fragdesc, SelfTimeQual);

    heap_close(fragdesc);

    /*
     * Signals end of transfer.
     */
    databuf.numTuples = 0;
    databuf.data.Data_len = 0;

    Rpc_setCallerName(rpcc, "PutFragData");
    if (!Rpc_call(rpcc, RPC_FRAGPUTDATA, &databuf, &result)) {
	Rpc_disconnect(rpcc);
	elog(WARN, "PutFragData: FRAGPUTDATA rpc call failed.");
    }

    if (result != SUCCESS) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "PutFragData: FRAGPUTDATA error %d.", result);
    }

    elog(NOTICE, "Leaving PutFragData.");

    return;
}

/* ----------------------------------------------------------------
 *	GetFragCatalog
 *
 *	We call remote site and ask for the system catalog information
 *	concerning the fragment being transferred over here.
 *	Sets newFragDesc and newFragArchDesc which are used by getFragData.
 *
 *	If this is being done to create a copy, generates a new oid
 *	and name for the copy.
 *
 * ----------------------------------------------------------------
 */
void
GetFragCatalog(RpcConn *rpcc, FragXArg *fragxarg, Relation *newFragDesc,
	       Relation *newFragArchDesc)
{
    FragCatalogInfo	catInfo;
    Oid			copyfrastoreid;
    Form_pg_fragment	pg_frag_tup;
    Form_pg_class	pg_class_tup;
    HeapTuple		classTup;
    HeapTuple		fragTup;
    HeapTuple		pgfragTup;
    int			natts;
    AttributeTupleForm	*tupdesc;
    int			i;
    MemoryContext	oldcxt;
    Relation		pgfragdesc;
    char		*fralokey;
    char		*frahikey;
    bool		loisnull;
    bool		hiisnull;
    Oid			loout;
    Oid			hiout;
    char		query[1000];

    elog(NOTICE, "Entering GetFragCatalog.");
    Rpc_setCallerName(rpcc, "GetFragCatalog");
    
    memset(&catInfo, 0, sizeof(FragCatalogInfo));

    if (!Rpc_call(rpcc, RPC_FRAGGETCAT, fragxarg, &catInfo)) {
	Rpc_disconnect(rpcc);
	elog(WARN, "GetFragCatalog: FRAGGETCAT rpc call failed.");
    }

    if (catInfo.error != SUCCESS) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "GetFragCatalog: FRAGGETCAT error %d.", catInfo.error);
    }

    if (catInfo.classTuple.Tuple_len == 0 || catInfo.fragTuple.Tuple_len == 0)
    {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "GetFragCatalog: Cannot get pg_class tuples.");
    }
    
    classTup = (HeapTuple)catInfo.classTuple.Tuple_val;
    fragTup = (HeapTuple)catInfo.fragTuple.Tuple_val;
    pgfragTup = (HeapTuple)catInfo.pgfragTuple.Tuple_val;

    if (fragxarg->fragxtype == COPYFRAGMENT) {
	/*
	 * Since we are making a copy of the fragment we must assign a new oid
	 * and a unique relation name for the copy. We'll base it on the oid.
	 * Must stamp the new oid into the pg_fragment tuple also.
	 */
	GetNewObjectId(&copyfrastoreid);

	fragTup->t_oid = copyfrastoreid;
	pg_class_tup = (Form_pg_class)GETSTRUCT(fragTup);
	GenerateCopyName((Name) &pg_class_tup->relname, pg_class_tup,
		     copyfrastoreid);

	pg_frag_tup = (Form_pg_fragment)GETSTRUCT(pgfragTup);
	pg_frag_tup->frastoreid = copyfrastoreid;
    }

    /*
     * Prepare a tuple descriptor; the palloc for tuple descriptor need to
     * persist at the end of transaction (Relation descriptor is cached.)
     */
    if (!CacheCxt) {
	CacheCxt = CreateGlobalMemory("Cache");
    }
    oldcxt = MemoryContextSwitchTo((MemoryContext)CacheCxt);
    
    natts = catInfo.attrTuples.TupleArray_len;
    tupdesc = (TupleDesc)palloc(sizeof(AttributeTupleForm)*natts);
    for(i = 0; i < natts; i++) {
	AttributeTupleForm tup;
	int len;

	len = catInfo.attrTuples.TupleArray_val[i].Tuple_len;
	tup = (AttributeTupleForm) palloc(len);
	memmove(tup,
		catInfo.attrTuples.TupleArray_val[i].Tuple_val,
		len);
	tupdesc[i] = tup;
    }

    MemoryContextSwitchTo(oldcxt);

    *newFragDesc = 
	MakeFragmentRelationDesc(classTup, fragTup, natts, tupdesc);

    /*
     * Create archive relation for fragment, since all fragments are archived
     * in Mariposa.
     */
    CreateArchiveRelation(*newFragDesc);
    CommandCounterIncrement();
    *newFragArchDesc = (Relation) OpenArchiveRelation((*newFragDesc)->rd_id);

    if (!RelationIsValid(*newFragArchDesc)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "GetFragCatalog: Can't open newly created archive relation for fragment %d.",
	     (*newFragDesc)->rd_id);
    }

    if (fragxarg->fragxtype == SELLFRAGMENT) {
	/*
	 * Only need to try to delete old pg_fragment tuple if this isn't a
	 * copy operation (in that case, it's a new tuple.)
	 */
	pg_frag_tup = (Form_pg_fragment)GETSTRUCT(pgfragTup);

	DeletePgFragmentTuples(pg_frag_tup->frastoreid,
			       Anum_pg_fragment_frastoreid);
    }

    /*
     * Open pg_fragment for its tupdesc, so we can heap_getattr from
     * pgfragTup fralokey and frahikey.
     */
    pgfragdesc = heap_openr((Name) Name_pg_fragment);
    if (!RelationIsValid(pgfragdesc)) {
	AbortRemoteSite(rpcc, fragxarg->remotexactid);
	elog(WARN, "GetFragCatalog: Can't open pg_fragment.");
    }
    heap_close(pgfragdesc);

    fralokey = heap_getattr(pgfragTup, InvalidBuffer,
			    Anum_pg_fragment_fralokey, pgfragdesc->rd_att,
			    &loisnull);
    loout =
	typtoout((Oid)
		 pgfragdesc->rd_att[Anum_pg_fragment_fralokey - 1]->atttypid);

    frahikey = heap_getattr(pgfragTup, InvalidBuffer,
			    Anum_pg_fragment_frahikey, pgfragdesc->rd_att,
			    &hiisnull);
    hiout =
	typtoout((Oid)
		 pgfragdesc->rd_att[Anum_pg_fragment_frahikey - 1]->atttypid);
	
    sprintf(query,
	    "INSERT INTO pg_fragment\
 (frapages,fratuples,fraclassid,frastoreid,fralogicalid,fralocation%s%s)\
 values (%d,%d,%d,%d,%d,%d%s%s%s%s%s%s);",
	    /* On the fragment keys, check first char of string for NULL */
	    !loisnull ? ", fralokey" : "",
	    !hiisnull ? ", frahikey" : "",
	    0, 0, /* To be filled in after all tuples put */
	    pg_frag_tup->fraclassid, pg_frag_tup->frastoreid,
	    pg_frag_tup->fralogicalid, getMyHostId(),
	    !loisnull ? ",'" : "",
	    !loisnull ? fmgr(loout, fralokey, gettypelem(pgfragdesc->rd_att[Anum_pg_fragment_fralokey - 1]->atttypid)) : "",
	    !loisnull ? "'::text" : "",
	    !hiisnull ? ",'" : "",
	    !hiisnull ? fmgr(hiout, frahikey, gettypelem(pgfragdesc->rd_att[Anum_pg_fragment_frahikey - 1]->atttypid)) : "",
	    !hiisnull ? "'::text" : "");
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    elog(NOTICE, "Leaving GetFragCatalog.");
}

/* ----------------------------------------------------------------
 *	GetFragData
 *
 *	Snatches tuples of the fragment being transferred here from
 *	the remote site.
 *
 * ----------------------------------------------------------------
 */
void
GetFragData(RpcConn *rpcc, FragXArg *fragxarg, Relation newFragDesc,
	    Relation newFragArchDesc)
{
    FragDataBuffer	databuf;
    int			i;
    int			numTuples = 0;
    HeapTuple		tup;
    char		*data;
    char		query[1000];

    elog(NOTICE, "Entering GetFragData.");
    Rpc_setCallerName(rpcc, "GetFragData");
    
    for (;;) {
	memset(&databuf, 0, sizeof(FragDataBuffer));
	if (!Rpc_call(rpcc, RPC_FRAGGETDATA, fragxarg, &databuf)) {
	    Rpc_disconnect(rpcc);
	    elog(WARN, "GetFragData: FRAGGETDATA rpc call failed.");
	}

	if (databuf.error != SUCCESS) {
	    AbortRemoteSite(rpcc, fragxarg->remotexactid);
	    elog(WARN, "GetFragData: FRAGGETDATA error %d.", databuf.error);
	}

	if (databuf.numTuples == 0) {
	    /*
	     * Transfer finished.
	     */
	    
	    /* XXX number of pages is clearly wrong but we better have an
	       entry because the vacuum cleaner does expects the attribute
	       to be there. */
	    sprintf(query,
	    	    "UPDATE pg_fragment SET frapages = %d, fratuples = %d WHERE frastoreid='%d'::oid",
	    	    0/*XXX*/, numTuples, newFragDesc->rd_id);
	    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

	    break;
	}

	data = databuf.data.Data_val;

	/*
	 * If archive part, we insert into archive. If current part, we insert
	 * into new fragment, and keep track of the number of tuples.
	 */
	if (databuf.which == ARCHIVE_PART) {
	    for(i = 0; i < databuf.numTuples; i++) {
		tup = (HeapTuple)data;

		/*
		 * We use the copy update stream access method here, because we
		 * want to preserve the tuples' time info and origin.
		 */
		updtup_heap_insert(newFragArchDesc, tup);
		data += tup->t_len;
	    }
	}
	else {
	    for(i = 0; i < databuf.numTuples; i++) {
		tup = (HeapTuple)data;

		/*
		 * We use the copy update stream access method here, because we
		 * want to preserve the tuples' time info and origin.
		 */
		updtup_heap_insert(newFragDesc, tup);
		data += tup->t_len;
	    }

	    numTuples += databuf.numTuples;
	}

	Rpc_freeres(rpcc, RPC_FRAGGETDATA, &databuf);
    }

    heap_close(newFragArchDesc);
    heap_close(newFragDesc);

    elog(NOTICE, "Leaving GetFragData.");

    return;
}


/*****************************************************************************
 * Support routines for splitting fragments. 
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	GetNewFragmentRel
 *
 *	Finds or creates a new relation for the new fragment 'fragName.'
 *
 * ----------------------------------------------------------------
 */
Oid
GetNewFragmentRel(char *fragName, Relation rdesc)
{
    Relation		fragdesc;
    Oid			frastoreid;
    
    /*
     * If the relation named "fragName" already exists, we just get
     * its oid; otherwise we create a new table for the fragment.
     */
    fragdesc = heap_openr((Name) fragName);
    if (!RelationIsValid(fragdesc)) {
	frastoreid = heap_create(fragName, 'n', rdesc->rd_rel->relnatts,
				 DEFAULT_SMGR, rdesc->rd_att);
	/*
	 * normally, we'd need a call to DefineRelationFirstFragment
	 * here, but direct scans on storage fragments are explicitly
	 * handled in the fragmenter.
	 */
	CommandCounterIncrement();
    } else {
	frastoreid = fragdesc->rd_id;
	heap_close(fragdesc);
    }

    return(frastoreid);
}

/* ----------------------------------------------------------------
 *	EstablishNewFragment
 *
 *	Update system catalogs to reflect new fragment
 *      (identified by frastoreid).
 *
 * ----------------------------------------------------------------
 */
void
EstablishNewFragment(char *fragName, Relation rdesc, char *lokey, char *hikey,
		     Oid frastoreid, Oid fralogicalid)
{
    char		query[1000];

    sprintf(query, "UPDATE pg_class \
SET relpartcmp='%s', relparthash='%s', relpartatt=%d, relpartmode='%c' \
WHERE relname='%s'",
	    regprocout(rdesc->rd_rel->relpartcmp),
	    regprocout(rdesc->rd_rel->relparthash),
	    rdesc->rd_rel->relpartatt,
	    rdesc->rd_rel->relpartmode,
	    fragName);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    sprintf(query,
	    "INSERT INTO pg_fragment (frapages, fratuples, fraclassid, frastoreid, fralogicalid, fralocation%s%s)\
 values (0,0,%d,%d,%d,%d%s%s%s%s%s%s);",
	    lokey ? ", fralokey" : "",
	    hikey ? ", frahikey" : "",
	    rdesc->rd_id, frastoreid, fralogicalid, getMyHostId(),
	    lokey ? ",'" : "", lokey ? lokey : "", lokey ? "'::text" : "",
	    hikey ? ",'" : "", hikey ? hikey : "", hikey ? "'::text" : "");
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
}

/*
 * ----------------------------------------------------------------
 * 	FindMedianOID--
 *
 *	Input: scan descriptor
 *
 *	Output: median OID value
 *
 *	Implements the algorithm in "Introduction to Algorithms"
 *	(aka "CLR", "The Mobile Book") page 190 to select the ith
 * 	element from an unsorted array of elements.  Had to be
 * 	modified slightly to work with files instead of in-memory
 *	arrays.
 *
 *	The primary differences between this implementation and
 *	the one in CLR are:
 *
 *	- We use files instead of arrays
 *
 *	- This one is iterative, not recursive, and so uses three
 *	  files in the place of one array: one file to hold the
 *	  original "array" of oid's, two others to find the medians,
 *	  the medians-of-medians, etc. (if you don't understand this,
 *	  look in CLR).
 *
 * ----------------------------------------------------------------
 */
Oid
FindMedianOID(HeapScanDesc sd) {

#define GROUPSIZE 5
    
    HeapTuple		htup;
    int			fd1, fd2, basefd;
    char		*fromFile, *toFile;
    extern char* 	DataDir;
    char 		*fName1, *fName2, *fName3;
    int 		bufSize = sizeof(Oid);
    int 		nRead, nWritten;
    int			nTups;
    Oid			oidArr[GROUPSIZE];
    Oid			medianOID;
    char 		*temp;
    int			medianPos = 0;
    int			nLower, nHigher;

    /*
     * Create three temporary files
     */
    fName1 = (char *)palloc(strlen(DataDir) + 20);
    fName2 = (char *)palloc(strlen(DataDir) + 20);
    fName3 = (char *)palloc(strlen(DataDir) + 20);

    sprintf(fName1, "%s/pg_median1%d", DataDir, getpid());
    sprintf(fName2, "%s/pg_median2%d", DataDir, getpid());
    sprintf(fName3, "%s/pg_median3%d", DataDir, getpid());

    unlink(fName1);
    unlink(fName2);
    unlink(fName3);
    
    fd1 = creat(fName1, 0600);
    fd2 = creat(fName2, 0600);
    basefd = creat(fName3, 0600);

    nTups = 0;
    
    /*
     * Make initial pass: Scan over the scan descriptor passed,
     * putting sub-medians into file1.
     */
    do {
	int n;
	int j;

	n = 0;
	while (n < GROUPSIZE &&
	       HeapTupleIsValid(htup = heap_getnext(sd, 0, (Buffer *) NULL))) {
	    nTups++;
	    nWritten = write(basefd, &htup->t_oid, bufSize);
	    oidArr[n] = htup->t_oid;
	    n++;
	}

	/*
	 * Insertion sort the n elements
	 */
	for (j = 1; j < n; j++) {
	    Oid key = oidArr[j];
	    int i = j - 1;
	    while (i > 0 && oidArr[i] > key) {
		oidArr[i+1] = oidArr[i];
		i--;
	    }
	    oidArr[i+1] = key;
	}

	j = (int)(n/2);
	nWritten = write(fd1, &oidArr[j], bufSize);
	
    } while (HeapTupleIsValid(htup));

    close(fd1);
    close(basefd);
    
    heap_rescan(sd, FALSE, sd->rs_key);

    medianPos = (int) (nTups/2);
    
    /*
     * The pseudocode for the rest of the algorithm is:
     *
     *	fromFile, toFile: scratch files
     *	baseFile: corresponds to array A in CLR, page 190:
     *		  the sub-array from which we're extracting
     *		  the medianPos'th element
     *
     *	DO {
     *	    while (NTUPS > THRESHHOLD) {
     *		open fromFile, toFile
     *		do {
     *		    read GROUPSIZE elements from fromFile
     *		    NTUPS = NTUPS + GROUPSIZE
     *		    sort the elements
     *	 	    write the median to toFile
     *		} while (!EOF(fromFile))
     *		close fromFile, toFile
     *
     *		if (NTUPS > THRESHHOLD) {
     *		    swap fromFile, toFile
     *		} else {
     *		    read all elements in toFile
     *		    sort them
     *		    find the median
     *		}
     *	    }
     *
     *	    partition baseFile around median into lowerFile, upperFile
     *	    nLower = # of elements in lowerFile, similarly for upperFile
     *
     *	    if (medianPos < nLower) {
     *		baseFile = upperFile
     *		NTUPS = nLower
     *	    }
     *
     *	    if (medianPos > nLower) {
     *		baseFile = lowerFile
     *		medianPos = medianPos - nLower
     *		NTUPS = nUpper
     *	    }
     *
     *	    if (NTUPS <= THRESHOLD) {
     *		sort the rest of the elements into array A[]
     *		median = A[medianPos]
     *	    }
     *
     *  } while (NTUPS > THRESHHOLD)
     */
    fromFile = fName1;
    toFile = fName2;

    /*
     * do while NTUPS > THRESHOLD - keep finding the medians
     * using files until there are few enough tuples to sort
     * using insertion sort.
     */
    if (nTups > 0) {
	do {

	    /*
	     *  Find the partitioning element by taking the medians
	     *  of the sub-groups, writing them to a file, taking
	     *  the medians of the subgroups of that file and writing
	     *  them to the original file, etc. until there are
	     *  fewer than THRESHOLD tuples.
	     */
	    while ((int)(nTups/GROUPSIZE) > GROUPSIZE) {
		int n;
		int j;
	
		nTups = 0;
		fd1 = open(fromFile, O_RDONLY);
		fd2 = open(toFile, O_WRONLY | O_TRUNC);

		/*
		 * Find the sub-medians in fromFile, write them out
		 * to to-file.
		 */

		do {
		    n = 0;
		    do {

			nRead = read(fd1, &oidArr[n], bufSize);

			if (nRead == bufSize) {
			    nTups++;
			    n++;
			}
		    } while (nRead == bufSize && n < GROUPSIZE);

		    if (n > 0) {
			/*
			 * Insertion sort the n elements
			 */
			for (j = 1; j < n; j++) {
			    Oid key = oidArr[j];
			    int i = j - 1;
			    while (i > 0 && oidArr[i] > key) {
				oidArr[i+1] = oidArr[i];
				i--;
			    }
			    oidArr[i+1] = key;
			}

			/*
			 * Write the median to the other file
			 */
			j = (int)(n/2);
			nWritten = write(fd2, &oidArr[j], bufSize);
		    }
		} while (nRead == bufSize);
	    
		close(fd1);
		close(fd2);
	

		/*
		 * If still have too many tuples, keep finding medians-of-medians
		 */
		if ((int)(nTups/GROUPSIZE) > GROUPSIZE) {
		    /*
		     * Switch from/to files
		     */
		    temp = fromFile;
		    fromFile = toFile;
		    toFile = temp;

		    /*
		     * Otherwise, just sort the remaining ones in an array and take
		     * the middle value.
		     */
		} else {
		    Oid *tempArr = (Oid *)palloc(nTups * sizeof(Oid));
		    fd1 = open(fromFile, O_RDONLY);
		    for (j = 0; j < nTups; j++) {
			nRead = read(fd1, &tempArr[j], bufSize);
		    }
		    /*
		     * Insertion sort the n elements
		     */
		    for (j = 1; j < nTups; j++) {
			Oid key = tempArr[j];
			int i = j - 1;
			while (i > 0 && tempArr[i] > key) {
			    tempArr[i+1] = tempArr[i];
			    i--;
			}
			tempArr[i+1] = key;
		    }

		    nTups = (int)(nTups/2);
		    medianOID = tempArr[nTups];
		    pfree(tempArr);
		}
	
	
	    }
	
	    /*
	     * Partition the oid's around the medianOID
	     */
	    fd1 = open(fName1, O_WRONLY | O_TRUNC);
	    fd2 = open(fName2, O_WRONLY | O_TRUNC);
	    basefd = open(fName3, O_RDONLY);
	    nLower = 0;
	    nHigher = 0;
	
	    do {
		Oid oid;

		nRead = read(basefd, &oid, bufSize);

		if (nRead == bufSize) {
		    if (oid <= medianOID) {
			write(fd1, &oid, bufSize);
			nLower++;
		    } else {
			write(fd2, &oid, bufSize);
			nHigher++;
		    }
		}
	    
	    } while (nRead == bufSize);

	    close(fd1);
	    close(fd2);
	    close(basefd);


	    /*
	     * Figure out in which partition the median resides
	     */
	    if (medianPos < nLower) {
		fromFile = fName1;
		toFile = fName2;
		nTups = nLower;
	    }
	
	    if (medianPos > nLower) {
		fromFile = fName2;
		toFile = fName1;
		medianPos = medianPos - nLower;
		nTups = nHigher;
	    }


	    /*
	     * If there are still too many tuples in the partition that
	     * contains the median, keep using files.
	     */
	    if ((int)(nTups/GROUPSIZE) > GROUPSIZE) {
		char buf[BLCKSZ];
	    
		basefd = open(fName3, O_WRONLY | O_TRUNC);
		fd1 = open(fromFile, O_RDONLY);

		do {
		    nRead = read(fd1, buf, BLCKSZ);
		    nWritten = write(basefd, buf, nRead);
		} while (nRead == BLCKSZ);

		close(basefd);
		close(fd1);

		/*
		 * Otherwise, read the remaining tuples into an array
		 * and take the medianPos'th one - which is the real
		 * median of the original file.
		 */
	    } else {
		Oid *tempArr = (Oid *)palloc(nTups * sizeof(Oid));
		int j;
		fd1 = open(fromFile, O_RDONLY);
		for (j = 0; j < nTups; j++) {
		    nRead = read(fd1, &tempArr[j], bufSize);
		}
		/*
		 * Insertion sort the n elements
		 */
		for (j = 1; j < nTups; j++) {
		    Oid key = tempArr[j];
		    int i = j - 1;
		    while (i > 0 && tempArr[i] > key) {
			tempArr[i+1] = tempArr[i];
			i--;
		    }
		    tempArr[i+1] = key;
		}

		medianOID = tempArr[medianPos];
		pfree(tempArr);
	    }
	
	} while ((int)(nTups/GROUPSIZE) > GROUPSIZE);
    } /* if (nTups > 0) */
    /*
     * Clean up.
     */
    unlink(fName1);
    unlink(fName2);
    unlink(fName3);
    
    pfree(fName1);
    pfree(fName2);
    pfree(fName3);
    

   
    return medianOID;
    
}

/* ----------------------------------------------------------------
 *	SplitFragDoInserts
 *
 *	Scans relation, and based on its partitioning mode, puts its
 *	tuples into nrel1 and nrel2.  The partitioning information
 *	is extracted from the base table descriptors (obrel, nbrel1,
 *	nbrel2) -- nrel1 and nrel2 may be archive tables.
 *
 *	Uses updtup_heap_insert which preserves timestamp information.
 *
 * ----------------------------------------------------------------
 */
void SplitFragDoInserts(HeapScanDesc sdrel, HeapScanDesc sdarch,
			Relation obrel, Relation nrel1, Relation nrel2,
			Relation arel1, Relation arel2)
{
    HeapTuple		htup;
    HeapTuple		copytup;
    int			nrel1ct = 0;
    int			nrel2ct = 0;
    int			orelct = 0;
    int			nOldTups = 0;
    unsigned int 	seed;		/* For RANDOM */
    Oid			medianOID;	/* For Round-Robin */
    
    Assert(HeapScanIsValid(sdrel));
    Assert(HeapScanIsValid(sdarch));
    Assert(RelationIsValid(obrel));
    Assert(RelationIsValid(nrel1));
    Assert(RelationIsValid(nrel2));
    Assert(RelationIsValid(arel1));
    Assert(RelationIsValid(arel2));

    if (obrel->rd_rel->relpartmode == PARTITIONMODE_ROUND) {
	medianOID = FindMedianOID(sdrel);
    }

    if (obrel->rd_rel->relpartmode == PARTITIONMODE_RANDOM) {
	seed = rand();
	srand(seed);
    }
    
    /*
     * Insert into the new fragments
     */
    while (HeapTupleIsValid(htup = heap_getnext(sdrel, 0, (Buffer *) NULL))) {
	++orelct;
	copytup = heap_copytuple(htup);		/* Insert modifies tuple. */
	
	switch (obrel->rd_rel->relpartmode) {
	case PARTITIONMODE_ROUND:
	    if (htup->t_oid < medianOID) {
		updtup_heap_insert(nrel2, copytup);
		++nrel2ct;
	    } else {
		updtup_heap_insert(nrel1, copytup);
		++nrel1ct;
	    }
	    break;
	case PARTITIONMODE_RANDOM:
	    /*
	     * we don't really care where a particular tuple goes, so
	     * just do something that splits them up more or less
	     * evenly.
	     *
	     * (we used to do something less deterministic and
	     * skew-prone here, but we have the problem that we need
	     * the historical tuples to go to the same place as the
	     * current tuple -- so we have to do something based on
	     * oid.)
	     */
	    
	    if (rand() % 2 == 0) {
		updtup_heap_insert(nrel2, copytup);
		++nrel2ct;
	    } else {
		updtup_heap_insert(nrel1, copytup);
		++nrel1ct;
	    }
	    break;
	case PARTITIONMODE_KEY:
	case PARTITIONMODE_HASH:
	    if (ExecPartitionCheck(nrel1, copytup)) {
		updtup_heap_insert(nrel1, copytup);
		++nrel1ct;
	    }
	    if (ExecPartitionCheck(nrel2, copytup)) {
		updtup_heap_insert(nrel2, copytup);
		++nrel2ct;
	    }
	    break;
	default:
	    elog(WARN, "SplitFragDoInserts: bad partition mode '%d.'",
		 obrel->rd_rel->relpartmode);
	    break;
	}
	pfree((void *) copytup);
    }

    Assert(orelct == nrel1ct + nrel2ct);

    orelct = 0;
    nrel1ct = 0;
    nrel2ct = 0;
    
    /*
     * Insert into the new archive fragments
     */
    srand(seed);
    while (HeapTupleIsValid(htup = heap_getnext(sdarch, 0, (Buffer *) NULL))) {
	++orelct;
	copytup = heap_copytuple(htup);		/* Insert modifies tuple. */
	
	switch (obrel->rd_rel->relpartmode) {
	case PARTITIONMODE_ROUND:
	    if (htup->t_oid < medianOID) {
		updtup_heap_insert(nrel2, copytup);
		++nrel2ct;
	    } else {
		updtup_heap_insert(nrel1, copytup);
		++nrel1ct;
	    }
	    break;
	case PARTITIONMODE_RANDOM:
	    /*
	     * we don't really care where a particular tuple goes, so
	     * just do something that splits them up more or less
	     * evenly.
	     *
	     * (we used to do something less deterministic and
	     * skew-prone here, but we have the problem that we need
	     * the historical tuples to go to the same place as the
	     * current tuple -- so we have to do something based on
	     * oid.)
	     */

	    if (rand() % 2 == 0) {
		updtup_heap_insert(arel2, copytup);
		++nrel2ct;
	    } else {
		updtup_heap_insert(arel1, copytup);
		++nrel1ct;
	    }
	    break;
	case PARTITIONMODE_KEY:
	case PARTITIONMODE_HASH:
	    if (ExecPartitionCheck(nrel1, copytup)) {
		updtup_heap_insert(arel1, copytup);
		++nrel1ct;
	    }
	    if (ExecPartitionCheck(nrel2, copytup)) {
		updtup_heap_insert(arel2, copytup);
		++nrel2ct;
	    }
	    break;
	default:
	    elog(WARN, "SplitFragDoInserts: bad partition mode '%d.'",
		 obrel->rd_rel->relpartmode);
	    break;
	}
	pfree((void *) copytup);
    }

    Assert(orelct == nrel1ct + nrel2ct);

}

/* ----------------------------------------------------------------
 *	SplitFragment
 *
 *	Creates two new fragments, and puts all the tuples from the fragment
 *	designated by 'oldFragName' into both new fragments. Updates system
 *	catalog information to reflect the split.
 *
 * ----------------------------------------------------------------
 */
void
SplitFragment(char *oldFragName, char *newFragName1, char *newFragName2,
	      char *splitVal)
{
    Relation		orel;
    Relation		brel;
    Relation		nrel1;
    Relation		nrel2;
    Relation		orelarch;
    Relation		nrelarch1;
    Relation		nrelarch2;
    Oid			ostoreid;
    Oid			classid;
    Oid			nstoreid1;
    Oid			nstoreid2;
    FragPartInfo	ofpi;
    Datum		splitDatum = 0;
    HeapScanDesc	sdrel, sdarch;
    TimeQual		fetchArchiveTimeQual;
    NameData		oarchname;

    /*
     * Look up information on the original fragment.
     */
    orel = heap_openr((Name) oldFragName);
    if (!RelationIsValid(orel)) {
	elog(WARN, "SplitFragment: no such fragment '%s'", oldFragName);
    }

    if (orel->rd_iscopied) {
	elog(WARN, "We're not allowed to split copied fragments yet!");
    }

    ostoreid = orel->rd_id;
    ofpi = orel->rd_fraginfo;
    classid = ofpi->fpi_class;

    /*
     * Ensure that 'splitVal' lies between the bounds of the existing
     * fragment.
     */
    if (PointerIsValid(splitVal)) {
	Oid atttype, typin;

	if (orel->rd_rel->relpartmode != PARTITIONMODE_KEY &&
	    orel->rd_rel->relpartmode != PARTITIONMODE_HASH) {
	    elog(WARN, "SplitFragment: Split value specified for non-partitioned table");
	}

	/*
	 * Figure out the input function for 'relpartatt.'
	 */
	if (orel->rd_rel->relpartmode == PARTITIONMODE_KEY) {
	    atttype = orel->rd_att[orel->rd_rel->relpartatt - 1]->atttypid;
	    typin = (Oid) SearchSysCacheGetAttribute(TYPOID,
						     Anum_pg_type_typinput,
						     (char *) atttype,
						     (char *) NULL,
						     (char *) NULL,
						     (char *) NULL);
	    splitDatum = (Datum) fmgr(typin, splitVal);
	} else {
	    splitDatum = (Datum) int4in(splitVal);
	}
	if (!fragmentPartitionCheck(orel, splitDatum)) {
	    heap_close(orel);
	    elog(WARN, "SplitFragment: '%s' out of bounds", splitVal);
	}
    } else {
	if (orel->rd_rel->relpartmode == PARTITIONMODE_KEY ||
	    orel->rd_rel->relpartmode == PARTITIONMODE_HASH) {
	    elog(WARN, "SplitFragment: Split value not specified for partitioned table");
	}
    }

    fetchArchiveTimeQual =
	TimeFormRangedTimeQual(EPOCH_ABSTIME, CURRENT_ABSTIME);

    /*
     * Create the new fragments (with the new range bounds, if any.)
     */
    brel = heap_open(classid);
    if (!RelationIsValid(brel)) {
	heap_close(orel);
	elog(WARN, "SplitFragment: no base table '%d' for fragment '%s.'",
	     classid, oldFragName);
    }
    nstoreid1 = GetNewFragmentRel(newFragName1, brel);
    nstoreid2 = GetNewFragmentRel(newFragName2, brel);

    nrel1 = heap_openr((Name) newFragName1);
    if (!RelationIsValid(nrel1)) {
	heap_close(orel);
	elog(WARN, "SplitFragment: could not create first new table");
    }
    nrel2 = heap_openr((Name) newFragName2);
    if (!RelationIsValid(nrel2)) {
	heap_close(orel);
	heap_close(nrel1);
	elog(WARN, "SplitFragment: could not create second new table");
    }
    EstablishNewFragment(newFragName1, brel, ofpi->fpi_lokeytext, splitVal,
			 nstoreid1, nstoreid1);
    EstablishNewFragment(newFragName2, brel, splitVal, ofpi->fpi_hikeytext,
			 nstoreid2, nstoreid2);

    /*
     * Create archive relation for new fragments, since all fragments are
     * archived in Mariposa.
     */
    CreateArchiveRelation(nrel1);
    CommandCounterIncrement();
    CreateArchiveRelation(nrel2);
    CommandCounterIncrement();

    heap_close(nrel1);
    heap_close(nrel2);

    /*
     * Copy the data into the new fragments. We do the hokey-pokey of
     * using RelationFlushRelation to blow away the rdescs because the
     * other interfaces keep us from deleting rdescs for local
     * relations including any newly-created ones.
     *
     * Hmmmm... This assumes that heap_close just decrements the refcount and
     * does nothing else, so the relation descriptor hangs around ok. So then
     * we use RelationFlushRelation to kick it out of the cache, so it reflects
     * any updates the next time we open the relation.
     */
    RelationFlushRelation(&nrel1, false);
    RelationFlushRelation(&nrel2, false);
    nrel1 = heap_openr((Name) newFragName1);
    if (!RelationIsValid(nrel1)) {
	heap_close(orel);
	elog(WARN, "SplitFragment: could not reopen first new table");
    }
    nrel2 = heap_openr((Name) newFragName2);
    if (!RelationIsValid(nrel2)) {
	heap_close(orel);
	heap_close(nrel1);
	elog(WARN, "SplitFragment: could not reopen second new table");
    }

/*     sd = heap_beginscan(orel, 0, fetchArchiveTimeQual, 0, (ScanKey) NULL); */
/*     SplitFragDoInserts(sd, orel, nrel1, nrel2, nrel1, nrel2); */
/*     heap_endscan(sd); */

    /*
     * Copy archived data into new fragment archives.
     */
    orelarch = (Relation) OpenArchiveRelation(orel->rd_id);
    if (!RelationIsValid(orelarch)) {
	heap_close(orel);
	heap_close(nrel1);
	heap_close(nrel2);
	elog(WARN, "SplitFragment: Can't open original archive table");
    }	
    (void) namecpy(&oarchname, &(orelarch->rd_rel->relname));
    nrelarch1 = (Relation) OpenArchiveRelation(nrel1->rd_id);
    if (!RelationIsValid(nrelarch1)) {
	heap_close(orel);
	heap_close(nrel1);
	heap_close(nrel2);
	heap_close(orelarch);
	elog(WARN, "SplitFragment: Can't open first new archive table");
    }	
    nrelarch2 = (Relation) OpenArchiveRelation(nrel2->rd_id);
    if (!RelationIsValid(nrelarch2)) {
	heap_close(orel);
	heap_close(nrel1);
	heap_close(nrel2);
	heap_close(orelarch);
	heap_close(nrelarch1);
	elog(WARN, "SplitFragment: Can't open second new archive table");
    }	

    sdrel = heap_beginscan(orel, 0, fetchArchiveTimeQual, 0, (ScanKey) NULL);
    sdarch = heap_beginscan(orelarch, 0, fetchArchiveTimeQual, 0, (ScanKey) NULL);
    SplitFragDoInserts(sdrel, sdarch, orel, nrel1, nrel2, nrelarch1, nrelarch2);
    heap_endscan(sdarch);
    heap_endscan(sdrel);

    heap_close(nrelarch2);
    heap_close(nrelarch1);
    heap_close(orelarch);

    heap_close(nrel2);
    heap_close(nrel1);
    heap_close(orel);

    heap_close(brel);

    pfree(fetchArchiveTimeQual);

    /*
     * Only destroy the fragment if it's not the base table fragment,
     * otherwise just chuck the tuples from the base table fragment.
     */
    if (ostoreid != classid) {
	RemoveRelation(oldFragName);
    } else {
	File fd;
	
	RemoveRelation(&oarchname);

	RelationRemoveIndexes(orel);

	fd = FileNameOpenFile(oldFragName, O_WRONLY, 0600);
	if (fd >= 0) {
	    (void) FileTruncate(fd, 0);
	    (void) FileClose(fd);
	}
    }

    /*
     * Delete the old fragment from pg_fragment.
     */
    DeletePgFragmentTuples(ostoreid, Anum_pg_fragment_frastoreid);
}


/*****************************************************************************
 * Routines used by parser.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	FillFragStmt
 *
 *	Given fragment name, figures out the store and class oids,
 *	and fills in the statement passed in.
 *
 * ----------------------------------------------------------------
 */
void
FillFragStmt(MoveFragStmt *stmt, char *fragname, int location, int direction)
{
    int			fraclassid;
    int			frastoreid;
    Relation		rdesc;
    NameData		relName;
    FormData_pg_fragment pgFragTuple;
    NsvcLookupClassResult *nsvcClassInfo;
    NsvcLookupFragsResult *nsvcFragInfo;
#define TO 348
#define FROM 297

    /*
     * If moving to somewhere else, look up information in our
     * local system catalogs.
     */
    if (direction == TO) {
	namestrcpy(&relName, fragname);
	if (!RelationIsValid(rdesc = heap_openr(&relName))) {
	    elog(WARN, "Can't open fragment %s.", fragname);
	}
	frastoreid = rdesc->rd_id;
	heap_close(rdesc);
	FetchPgFragTuple(&pgFragTuple, frastoreid);
	fraclassid = pgFragTuple.fraclassid;

	
    } else {
	nsvcClassInfo = nsvc_lookupClass(fragname);
	nsvcFragInfo = nsvc_lookupFrags(fragname, ((HeapTuple)nsvcClassInfo->classTuple.Tuple_val)->t_oid);
    

	frastoreid = ((HeapTuple)nsvcClassInfo->classTuple.Tuple_val)->t_oid;
	fraclassid = ((Form_pg_fragment)GETSTRUCT(nsvcFragInfo->fragTuples.TupleArray_val[0].Tuple_val))->fraclassid;
    }
    
    
    stmt->fraclassid = fraclassid;
    stmt->frastoreid = frastoreid;
    stmt->newloc = location;
    stmt->direction = direction;
}

/* ----------------------------------------------------------------
 *	FetchPgFragTuple
 *
 *	Grabs tuple from pg_fragment so you can extract fields from it.
 *	Note: Relies on GETSTRUCT hack. You can't access the text fields
 *	fralokey and frahikey.
 *
 * ----------------------------------------------------------------
 */
void
FetchPgFragTuple(Form_pg_fragment fillMeIn, Oid frastoreid)
{
    Relation		fragdesc;
    HeapScanDesc	fragscan;
    HeapTuple		htup;
    ScanKeyData		skey[1];
    bool		foundValid = FALSE;

    fragdesc = heap_openr((Name)Name_pg_fragment);
    if (!RelationIsValid(fragdesc)) {
	elog(WARN, "Can't open pg_fragment.");
    }

    ScanKeyEntryInitialize(&skey[0],
			   0x0,
			   Anum_pg_fragment_frastoreid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(frastoreid));

    fragscan = heap_beginscan(fragdesc, 0, SelfTimeQual, 1, (ScanKey)skey);
    do {
	htup = heap_getnext(fragscan, 0, NULL);

	if (HeapTupleIsValid(htup)) {
	    bcopy((char *)GETSTRUCT(htup), (char *)fillMeIn,
		  sizeof(FormData_pg_fragment));
	    foundValid = TRUE;
	    break;
	}
    } while (htup);
    heap_endscan(fragscan);
    heap_close(fragdesc);

    /*
     * Should ask name server here.
     */
    if (!foundValid) {
	elog(WARN, "Can't find pg_fragment tuple for %d\n", frastoreid);
    }
}
