/*-------------------------------------------------------------------------
 *
 * remove.c--
 *    POSTGRES remove (function | type | operator ) utilty code.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/commands/remove.c,v 0.5 1996/06/07 04:55:38 am Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <string.h>
#include "c.h"

#include "access/attnum.h"
#include "access/heapam.h"
#include "access/htup.h"
#include "access/skey.h"
#include "utils/builtins.h"
#include "utils/tqual.h"	/* for NowTimeQual */
#include "catalog/catname.h"
#include "commands/defrem.h"
#include "utils/elog.h"

#include "miscadmin.h"

#include "catalog/pg_aggregate.h"
#include "catalog/pg_language.h"
#include "catalog/pg_operator.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "utils/syscache.h"
#include "parser/catalog_utils.h"
#include "storage/bufmgr.h"
#include "fmgr.h"

/*
 * RemoveOperator --
 *	Deletes an operator.
 *
 * Exceptions:
 *	BadArg if name is invalid.
 *	BadArg if type1 is invalid.
 *	"WARN" if operator nonexistant.
 *	...
 */
void
#ifndef P95
RemoveOperator(Name operatorName, /* operator name */
	       Name typeName1,	/* first type name */
	       Name typeName2)	/* optional second type name */
#else /* P95 */
RemoveOperator(char *operatorName, /* operator name */
	       char *typeName1,	/* first type name */
	       char *typeName2)	/* optional second type name */
#endif /* P95 */
{
    Relation 	relation;
    HeapScanDesc scan;
    HeapTuple 	tup;
    Oid 	typeId1 = InvalidOid;
    Oid 	typeId2 = InvalidOid;
    bool 	defined;
    ItemPointerData	itemPointerData;
    Buffer      buffer;
    ScanKeyData operatorKey[3];
#ifndef P95
    Name	userName;
#else /* P95 */
    char *userName;
#endif /* P95 */
    
#ifndef P95
    Assert(NameIsValid(operatorName));
    Assert(NameIsValid(typeName1) || NameIsValid(typeName2));
    
    if (NameIsValid(typeName1)) {
#else /* P95 */
     if (typeName1) {
#endif /* P95 */
	typeId1 = TypeGet(typeName1, &defined);
	if (!OidIsValid(typeId1)) {
#ifndef P95
	    elog(WARN, "RemoveOperator: type \"%-.*s\" does not exist",
		 NAMEDATALEN, typeName1);
#else /* P95 */
	    elog(WARN, "RemoveOperator: type '%s' does not exist", typeName1);
#endif /* P95 */
	    return;
	}
    }
    
#ifndef P95
    if (NameIsValid(typeName2)) {
#else /* P95 */
    if (typeName2) {
#endif /* P95 */
	typeId2 = TypeGet(typeName2, &defined);
	if (!OidIsValid(typeId2)) {
#ifndef P95
	    elog(WARN, "RemoveOperator: type \"%-.*s\" does not exist",
		 NAMEDATALEN, typeName2);
#else /* P95 */
	    elog(WARN, "RemoveOperator: type '%s' does not exist", typeName2);
#endif /* P95 */
	    return;
	}
    }
    
    ScanKeyEntryInitialize(&operatorKey[0], 0x0,
			   Anum_pg_operator_oprname,
			   NameEqualRegProcedure,
#ifndef P95
			   NameGetDatum(operatorName));
#else /* P95 */
			   PointerGetDatum(operatorName));
#endif /* P95 */
    
    ScanKeyEntryInitialize(&operatorKey[1], 0x0,
			   Anum_pg_operator_oprleft,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(typeId1));
    
    ScanKeyEntryInitialize(&operatorKey[2], 0x0,
			   Anum_pg_operator_oprright,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(typeId2));
    
    relation = heap_openr(OperatorRelationName);
    scan = heap_beginscan(relation, 0, NowTimeQual, 3, operatorKey);
    tup = heap_getnext(scan, 0, &buffer);
    if (HeapTupleIsValid(tup)) {
#ifndef NO_SECURITY
	userName = GetPgUserName();
#ifndef P95
	if (!pg_ownercheck(userName->data,
#else /* P95 */
	if (!pg_ownercheck(userName,
#endif /* P95 */
			   (char *) ObjectIdGetDatum(tup->t_oid),
			   OPROID))
#ifndef P95
	    elog(WARN, "RemoveOperator: operator \"%-.*s\": permission denied",
		 NAMEDATALEN, operatorName);
#else /* P95 */
	    elog(WARN, "RemoveOperator: operator '%s': permission denied",
		 operatorName);
#endif /* P95 */
#endif
	ItemPointerCopy(&tup->t_ctid, &itemPointerData);
	heap_delete(relation, &itemPointerData);
    } else {
	if (OidIsValid(typeId1) && OidIsValid(typeId2)) {
#ifndef P95
	    elog(WARN, "RemoveOperator: binary operator \"%-.*s\" taking \"%-.*s\" and \"%-.*s\" does not exist",
		 NAMEDATALEN, operatorName,
		 NAMEDATALEN, typeName1,
		 NAMEDATALEN, typeName2);
#else /* P95 */
	    elog(WARN, "RemoveOperator: binary operator '%s' taking '%s' and '%s' does not exist",
		 operatorName,
		 typeName1,
		 typeName2);
#endif /* P95 */
	} else if (OidIsValid(typeId1)) {
#ifndef P95
	    elog(WARN, "RemoveOperator: right unary operator \"%-.*s\" taking \"%-.*s\" does not exist",
		 NAMEDATALEN, operatorName,
		 NAMEDATALEN, typeName1);
#else /* P95 */
	    elog(WARN, "RemoveOperator: right unary operator '%s' taking '%s' does not exist",
		 operatorName,
		 typeName1);
#endif /* P95 */
	} else {
#ifndef P95
	    elog(WARN, "RemoveOperator: left unary operator \"%-.*s\" taking \"%-.*s\" does not exist",
		 NAMEDATALEN, operatorName,
		 NAMEDATALEN, typeName2);
#else /* P95 */
	    elog(WARN, "RemoveOperator: left unary operator '%s' taking '%s' does not exist",
		 operatorName,
		 typeName2);
#endif /* P95 */
	}
    }
    heap_endscan(scan);
    heap_close(relation);
}

#ifdef NOTYET
/*
 * this stuff is to support removing all reference to a type
 * don't use it  - pma 2/1/94
 */
/*
 *  SingleOpOperatorRemove
 *	Removes all operators that have operands or a result of type 'typeOid'.
 */
static void
SingleOpOperatorRemove(Oid typeOid)
{
    Relation 	rdesc;
    ScanKeyData 	key[3];
    HeapScanDesc 	sdesc;
    HeapTuple 	tup;
    ItemPointerData itemPointerData;
    Buffer 		buffer;
    static 		attnums[3] = { 7, 8, 9 }; /* left, right, return */
    register	i;
    
    ScanKeyEntryInitialize(&key[0],
			   0, 0, ObjectIdEqualRegProcedure, (Datum)typeOid);
#ifndef P95
    rdesc = heap_openr(OperatorRelationName->data);
#else /* P95 */
    rdesc = heap_openr(OperatorRelationName);
#endif /* P95 */
    for (i = 0; i < 3; ++i) {
	key[0].sk_attno = attnums[i];
	sdesc = heap_beginscan(rdesc, 0, NowTimeQual, 1, key);
#ifndef P95
	while (PointerIsValid(tup = heap_getnext(sdesc, 0, &buffer))) {
#else /* P95 */
	while (PointerIsValid(tup = heap_getnext(sdesc,  0, &buffer))) {
#endif /* P95 */
	    ItemPointerCopy(&tup->t_ctid, &itemPointerData);   
	    /* XXX LOCK not being passed */
	    heap_delete(rdesc, &itemPointerData);
	}
	heap_endscan(sdesc);
    }
    heap_close(rdesc);
}

/*
 *  AttributeAndRelationRemove
 *	Removes all entries in the attribute and relation relations
 *	that contain entries of type 'typeOid'.
 *      Currently nothing calls this code, it is untested.
 */
static void
AttributeAndRelationRemove(Oid typeOid)
{
    struct oidlist {
	Oid		reloid;
	struct oidlist	*next;
    };
    struct oidlist	*oidptr, *optr;
    Relation 		rdesc;
    ScanKeyData		key[1];
    HeapScanDesc 	sdesc;
    HeapTuple 		tup;
    ItemPointerData	itemPointerData;
    Buffer 		buffer;
    
    /*
     * Get the oid's of the relations to be removed by scanning the
     * entire attribute relation.
     * We don't need to remove the attributes here,
     * because amdestroy will remove all attributes of the relation.
     * XXX should check for duplicate relations
     */
    
    ScanKeyEntryInitialize(&key[0],
			   0, 3, ObjectIdEqualRegProcedure, (Datum)typeOid);
    
    oidptr = (struct oidlist *) palloc(sizeof(*oidptr));
    oidptr->next = NULL;
    optr = oidptr; 
#ifndef P95
    rdesc = heap_openr(AttributeRelationName->data);
#else /* P95 */
    rdesc = heap_openr(AttributeRelationName);
#endif /* P95 */
    sdesc = heap_beginscan(rdesc, 0, NowTimeQual, 1, key);
#ifndef P95
    while (PointerIsValid(tup = heap_getnext(sdesc, 0, &buffer))) {
#else /* P95 */
    while (PointerIsValid(tup = heap_getnext(sdesc, 0,  &buffer))) {
#endif /* P95 */
	ItemPointerCopy(&tup->t_ctid, &itemPointerData);   
	optr->reloid = ((AttributeTupleForm)GETSTRUCT(tup))->attrelid;
	optr->next = (struct oidlist *) palloc(sizeof(*oidptr));     
	optr = optr->next;
    }
    optr->next = NULL;
    heap_endscan(sdesc);
    heap_close(rdesc);
    
    
    ScanKeyEntryInitialize(&key[0], 0,
			   ObjectIdAttributeNumber,
			   ObjectIdEqualRegProcedure, (Datum)0);
    optr = oidptr;
#ifndef P95
    rdesc = heap_openr(RelationRelationName->data);
#else /* P95 */
    rdesc = heap_openr(RelationRelationName);
#endif /* P95 */
    while (PointerIsValid((char *) optr->next)) {
#ifndef P95
	key[0].sk_argument = (DATUM) (optr++)->reloid;
#else /* P95 */
	key[0].sk_argument = (Datum) (optr++)->reloid;
#endif /* P95 */
	sdesc = heap_beginscan(rdesc, 0, NowTimeQual, 1, key);
	tup = heap_getnext(sdesc, 0, &buffer);
	if (PointerIsValid(tup)) {
#ifndef P95
	    Name	name;
#else /* P95 */
	    char *name;
#endif /* P95 */
	    
#ifndef P95
	    name = &((Form_pg_class)GETSTRUCT(tup))->relname;
	    heap_destroy(name->data);
#else /* P95 */
	    name = (((Form_pg_class)GETSTRUCT(tup))->relname).data;
	    heap_destroy(name);
#endif /* P95 */
	}
    }
    heap_endscan(sdesc);
    heap_close(rdesc);
}
#endif /* NOTYET */

/*
 *  TypeRemove
 *	Removes the type 'typeName' and all attributes and relations that
 *	use it.
 */
void
#ifndef P95
RemoveType(Name typeName)    /* type name to be removed */
#else /* P95 */
RemoveType(char *typeName)    /* type name to be removed */
#endif /* P95 */
{
    Relation 	relation;  
    HeapScanDesc 	scan;      
    HeapTuple 	tup;
    Oid	typeOid;
    ItemPointerData	itemPointerData;
    static ScanKeyData typeKey[1] = {
	{ 0, Anum_pg_type_typname, NameEqualRegProcedure }
    };
#ifndef P95
    char16 arrayNameData;
    Name arrayName = & arrayNameData;
    Name userName;
#else /* P95 */
    char *shadow_type;
    char *userName;
#endif /* P95 */
    
#ifndef P95
    Assert(NameIsValid(typeName));
    
#endif /* P95 */
#ifndef NO_SECURITY
    userName = GetPgUserName();
#ifndef P95
    if (!pg_ownercheck(userName->data, typeName->data, TYPNAME))
	elog(WARN, "RemoveType: type \"%-.*s\": permission denied",
	     NAMEDATALEN, typeName);
#else /* P95 */
    if (!pg_ownercheck(userName, typeName, TYPNAME))
	elog(WARN, "RemoveType: type '%s': permission denied",
	     typeName);
#endif /* P95 */
#endif
    
    relation = heap_openr(TypeRelationName);
    fmgr_info(typeKey[0].sk_procedure, &typeKey[0].sk_func,
	      &typeKey[0].sk_nargs);
    
    /* Delete the primary type */
    
#ifndef P95
    typeKey[0].sk_argument = NameGetDatum(typeName);
#else /* P95 */
    typeKey[0].sk_argument = PointerGetDatum(typeName);
#endif /* P95 */
    
    scan = heap_beginscan(relation, 0, NowTimeQual, 1, typeKey);
    tup = heap_getnext(scan, 0, (Buffer *) 0);
    if (!HeapTupleIsValid(tup)) {
	heap_endscan(scan);
	heap_close(relation);
#ifndef P95
	elog(WARN, "RemoveType: type \"%-.*s\" does not exist",
	     NAMEDATALEN, typeName);
#else /* P95 */
	elog(WARN, "RemoveType: type '%s' does not exist",
	     typeName);
#endif /* P95 */
    }
    typeOid = tup->t_oid;
    ItemPointerCopy(&tup->t_ctid, &itemPointerData);
    heap_delete(relation, &itemPointerData);
    heap_endscan(scan);
    
    /* Now, Delete the "array of" that type */
#ifdef P95
    shadow_type = makeArrayTypeName(typeName);
    typeKey[0].sk_argument = NameGetDatum(shadow_type);
#endif /* P95 */
    
#ifndef P95
    arrayName->data[0] = '_';
    arrayName->data[1] = '\0';
    
/*    strncat(arrayName->data, typeName, 15); */
    namecat(arrayName,typeName);
    
    typeKey[0].sk_argument = NameGetDatum(arrayName);
    
#endif /* P95 */
    scan = heap_beginscan(relation, 0, NowTimeQual,
			  1, (ScanKey) typeKey);
#ifndef P95
    tup = heap_getnext(scan, 0, (Buffer *) 0);
#else /* P95 */
    tup = heap_getnext(scan, 0,  (Buffer *) 0);
#endif /* P95 */
    
    if (!HeapTupleIsValid(tup))
	{
#ifndef P95
	    elog(WARN, "RemoveType: type \"%-.*s\": array stub not found",
		 NAMEDATALEN, typeName->data);
#else /* P95 */
	    elog(WARN, "RemoveType: type '%s': array stub not found",
		 typeName);
#endif /* P95 */
	}
    typeOid = tup->t_oid;
    ItemPointerCopy(&tup->t_ctid, &itemPointerData);
    heap_delete(relation, &itemPointerData);
    heap_endscan(scan);
    
    heap_close(relation);
}

/*
 * RemoveFunction --
 *	Deletes a function.
 *
 * Exceptions:
 *	BadArg if name is invalid.
 *	"WARN" if function nonexistant.
 *	...
 */
void
#ifndef P95
RemoveFunction(Name functionName, /* type name to be removed */
#else /* P95 */
RemoveFunction(char *functionName, /* function name to be removed */
#endif /* P95 */
	       int nargs,
#ifndef P95
	       List *argNameList)
#else /* P95 */
	       List *argNameList /* list of TypeNames */)
#endif /* P95 */
{
    Relation         relation;
    HeapScanDesc         scan;
    HeapTuple        tup;
    Buffer           buffer = InvalidBuffer;
    bool		 bufferUsed = FALSE;
    Oid	 argList[8];
    Form_pg_proc	 the_proc;
#ifndef P95
#ifdef USEPARGS
    Oid         oid;
#endif
#endif /* P95 */
    ItemPointerData  itemPointerData;
    static ScanKeyData key[3] = {
	{ 0, Anum_pg_proc_proname, NameEqualRegProcedure }
    };
#ifndef P95
    Name	userName;
    char16	typename;
#else /* P95 */
    char *userName;
    char *typename;
#endif /* P95 */
    int 		i;
    
#ifndef P95
    Assert(NameIsValid(functionName));
    
#endif /* P95 */
    memset(argList, 0, 8 * sizeof(Oid));
    for (i=0; i<nargs; i++) {
#ifndef P95
	(void) namestrcpy(&typename, strVal(lfirst(argNameList)));
#else /* P95 */
/*	typename = ((TypeName*)(lfirst(argNameList)))->name; */
	typename = strVal(lfirst(argNameList));
#endif /* P95 */
	argNameList = lnext(argNameList);
	
#ifndef P95
	if (strncmp(typename.data, "opaque", NAMEDATALEN) == 0)
#else /* P95 */
	if (strcmp(typename, "opaque") == 0)
#endif /* P95 */
	    argList[i] = 0;
	else {
#ifndef P95
	    tup = SearchSysCacheTuple(TYPNAME, (char *) &typename,
				      (char *) NULL, (char *) NULL,
				      (char *) NULL);
#else /* P95 */
	    tup = SearchSysCacheTuple(TYPNAME, PointerGetDatum(typename),
				      0,0,0);
#endif /* P95 */
	    
	    if (!HeapTupleIsValid(tup)) {
#ifndef P95
		elog(WARN, "RemoveFunction: type \"%-.*s\" not found",
		     NAMEDATALEN, &typename);
#else /* P95 */
		elog(WARN, "RemoveFunction: type '%s' not found",typename);
#endif /* P95 */
	    }
#ifndef P95
	    
#endif /* P95 */
	    argList[i] = tup->t_oid;
	}
    }
    
#ifndef P95
    tup = SearchSysCacheTuple(PRONAME, functionName->data,
			      (char *) Int32GetDatum(nargs),
			      (char *) argList, (char *) NULL);
#else /* P95 */
    tup = SearchSysCacheTuple(PRONAME, PointerGetDatum(functionName),
			      Int32GetDatum(nargs),
			      PointerGetDatum(argList),0);
#endif /* P95 */
    if (!HeapTupleIsValid(tup))
#ifndef P95
	func_error("RemoveFunction", functionName->data, nargs, (int*)argList);
#else /* P95 */
	func_error("RemoveFunction", functionName, nargs, (int*)argList);
#endif /* P95 */
    
#ifndef NO_SECURITY
    userName = GetPgUserName();
#ifndef P95
    if (!pg_func_ownercheck(userName->data, (char *) functionName, 
			    nargs, argList)) {
	elog(WARN, "RemoveFunction: function \"%-.*s\": permission denied",
	     NAMEDATALEN, functionName);
#else /* P95 */
    if (!pg_func_ownercheck(userName, functionName, nargs, argList)) {
	elog(WARN, "RemoveFunction: function '%s': permission denied",
	     functionName);
#endif /* P95 */
    }
#endif
    
#ifndef P95
    key[0].sk_argument = NameGetDatum(functionName);
#else /* P95 */
    key[0].sk_argument = PointerGetDatum(functionName);
#endif /* P95 */
    
    fmgr_info(key[0].sk_procedure, &key[0].sk_func, &key[0].sk_nargs);
    
    relation = heap_openr(ProcedureRelationName);
    scan = heap_beginscan(relation, 0, NowTimeQual, 1, key);
    
    do { /* hope this is ok because it's indexed */
	if (bufferUsed) {
	    ReleaseBuffer(buffer);
	    bufferUsed = FALSE;
	}
#ifndef P95
	tup = heap_getnext(scan, 0, (Buffer *) &buffer);
#else /* P95 */
	tup = heap_getnext(scan,  0, (Buffer *) &buffer);
#endif /* P95 */
	if (!HeapTupleIsValid(tup))
	    break;
	bufferUsed = TRUE;
	the_proc = (Form_pg_proc) GETSTRUCT(tup);
#ifndef P95
    } while (NameIsEqual(&(the_proc->proname), functionName) &&
#else /* P95 */
    } while ( (namestrcmp(&(the_proc->proname), functionName) == 0) &&
#endif /* P95 */
	     (the_proc->pronargs != nargs ||
	      !oid8eq(&(the_proc->proargtypes[0]), &argList[0])));
    
    
#ifndef P95
    if (!HeapTupleIsValid(tup) || !NameIsEqual(&(the_proc->proname), 
					       functionName))
#else /* P95 */
    if (!HeapTupleIsValid(tup) || namestrcmp(&(the_proc->proname), 
					     functionName) != 0)
#endif /* P95 */
	{	
	    heap_endscan(scan);
	    heap_close(relation);
#ifndef P95
	    func_error("RemoveFunction", functionName->data,
		       nargs, (int*)argList);
#else /* P95 */
	    func_error("RemoveFunction", functionName,nargs, (int*)argList);
#endif /* P95 */
	}
    
    /* ok, function has been found */
#ifndef P95
#ifdef USEPARGS
    oid = tup->t_oid;
#endif
#endif /* P95 */
    
    if (the_proc->prolang == INTERNALlanguageId)
	elog(WARN, "RemoveFunction: function \"%-.*s\" is built-in",
	     NAMEDATALEN, functionName);
    
    ItemPointerCopy(&tup->t_ctid, &itemPointerData);
    heap_delete(relation, &itemPointerData);
    heap_endscan(scan);
    heap_close(relation);
#ifndef P95
    
    /*  
     * Remove associated parameters from PARG catalog
     *
     * XXX NOT CURRENTLY DONE
     */
#ifdef USEPARGS
    ScanKeyEntryInitialize(&typeKey, 0, 1, ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(oid));
    key[0].argument = NameGetDatum(functionName);
    relation = heap_openr(ProcedureArgumentRelationName);
    scan = heap_beginscan(relation, 0, NowTimeQual,
			  1, (ScanKey) typeKey);
    tup = heap_getnext(scan, 0, (Buffer *) 0);
    while (PointerIsValid(tup)) {
	ItemPointerCopy(&tup->t_ctid, &itemPointerData);
	heap_delete(relation, &itemPointerData);
	tup = heap_getnext(scan, 0, (Buffer *) 0);
    }
    heap_endscan(scan);
    heap_close(relation);
#endif
#endif /* P95 */
}

void
#ifndef P95
RemoveAggregate(Name aggName)
#else /* P95 */
RemoveAggregate(char *aggName)
#endif /* P95 */
{
    Relation 	relation;
    HeapScanDesc	scan;
    HeapTuple	tup;
    ItemPointerData	  itemPointerData;
    static ScanKeyData key[3] = {
	{ 0, Anum_pg_aggregate_aggname, NameEqualRegProcedure }
    };
    
#ifndef P95
    Assert(NameIsValid(aggName));
#else /* P95 */
    key[0].sk_argument = PointerGetDatum(aggName);
#endif /* P95 */
    
#ifndef P95
#ifndef NO_SECURITY
    /* XXX there's no pg_aggregate.aggowner?? */
#endif
    
    key[0].sk_argument = NameGetDatum(aggName);
    
#endif /* P95 */
    fmgr_info(key[0].sk_procedure, &key[0].sk_func, &key[0].sk_nargs);
    relation = heap_openr(AggregateRelationName);
    scan = heap_beginscan(relation, 0, NowTimeQual, 1, key);
    tup = heap_getnext(scan, 0, (Buffer *) 0);
    if (!HeapTupleIsValid(tup)) {
	heap_endscan(scan);
	heap_close(relation);
#ifndef P95
	elog(WARN, "RemoveAggregate: aggregate \"%-.*s\" does not exist",
	     NAMEDATALEN, aggName);
#else /* P95 */
	elog(WARN, "RemoveAggregate: aggregate '%s' does not exist",
	     aggName);
#endif /* P95 */
    }
    ItemPointerCopy(&tup->t_ctid, &itemPointerData);
    heap_delete(relation, &itemPointerData);
    heap_endscan(scan);
    heap_close(relation);
}
