/*-------------------------------------------------------------------------
 *
 * nodeBuffer.c--
 *    Support routines for buffering node (supports Mark and Restr)
 *    only buffers as little as necessary to allow a Restr to the 
 *    latest marked position
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/executor/nodeBuffer.c,v 1.12 1995/09/05 09:30:51 aoki Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include "c.h"
#include "executor/executor.h"
#include "executor/nodeBuffer.h"
#include "utils/lselect.h"
#include "machine.h"

/* 
 * ExecBuffer -
 *    fetches the next tuple either from the buffer file or the outer plan
 */
TupleTableSlot *
ExecBuffer(TupleBuffer *node)
{
    TupleTableSlot *slot, *resultSlot;
    TupleBufferState	*state;
    Plan *outerPlan;
    HeapTuple tuple;
    FILE *fileDesc;

    state = node->buffstate;

    Assert(state->buff_initialized);

    fileDesc = (FILE *) state->buff_file;
    resultSlot = state->cstate.cs_ResultTupleSlot;
    if (state->buff_nextPos == state->buff_fileLen) {
	/* get new tuple from outer plan (and possibly buffer it) */
	outerPlan = outerPlan(node);
	slot = ExecProcNode(outerPlan, (Plan*)node);
	tuple = slot->val;
	if (TupIsNull(slot)) {
	    return (slot); /* we're done */
	}
	if (tuple->t_len < 1) {
	    elog(WARN, "ExecBuffer: invalid tuple (t_len=%d)",
		 tuple->t_len);
	}
	state->buff_fileLen += tuple->t_len;
	state->buff_nextPos = state->buff_fileLen;
	if (fwrite((void *) tuple, tuple->t_len, 1, fileDesc) < 1) {
	    elog(WARN, "ExecBuffer: can't write to temporary file");
	}
	return (slot);
    } else {
	/* we are reading tuples from the buffer file */
	if (fread((void *) state->buff_tuple, sizeof(Size), 1, fileDesc) < 1) {
	    elog(WARN, "ExecBuffer: can't read from temporary file");
	}
	if (fread((void *) &state->buff_tuple->t_ctid, state->buff_tuple->t_len
	        - sizeof(Size), 1, fileDesc) < 1) {
	    elog(WARN, "ExecBuffer: can't read from temporary file");
	}
	state->buff_nextPos += state->buff_tuple->t_len;
	return
	    ((TupleTableSlot*) ExecStoreTuple(state->buff_tuple,
					      resultSlot,
					      InvalidBuffer,
					      false));
    }
}

 
/* 
 * ExecInitBuffer -
 *    creates and initializes a buffer node
 */
bool
ExecInitBuffer(TupleBuffer *node, EState *estate, Plan *parent)
{
    TupleBufferState *state;
    Plan *outerPlan;
    int fd;

    node->plan.state = (EState *)estate;

    /* create plan state */
    state = makeNode(TupleBufferState);
    node->buffstate = state;

    ExecAssignNodeBaseInfo(estate, &state->cstate, parent);

    outerPlan = outerPlan(node);
    ExecInitNode(outerPlan, estate, (Plan *) node);

#define BUFFER_NSLOTS	1
    ExecInitResultTupleSlot(estate, &state->cstate);
    ExecAssignResultTypeFromOuterPlan((Plan*) node, &state->cstate);

    state->buff_fileLen = 0;
    state->buff_nextPos = 0;

    /* we have yet to open the buffer file */
    (void) strcpy((char *) &(state->buff_name), "pg_buf.XXXXXX");
    (void) mktemp((char *) &(state->buff_name));
    if ((fd = open((char *) &(state->buff_name),
		   O_RDWR|O_EXCL|O_CREAT, 0600)) < 0) {
	elog(WARN, "ExecInitBuffer: can't create temporary file %s: %m",
	     &(state->buff_name));
    }
    if ((state->buff_file = (Pointer) fdopen(fd, "wb+")) == NULL) {
	elog(WARN, "ExecInitBuffer: can't open temporary file %s: %m",
	     &(state->buff_name));
    }
    if (unlink((char *) &(state->buff_name)) < 0) {
	elog(WARN, "ExecInitBuffer: can't unlink temporary file %s: %m",
	     &(state->buff_name));
    }
    
    /* safe assumption that no tuple is bigger than LocalBlockSize */
    state->buff_tuple = (HeapTuple) palloc(LocalBlockSize);
    state->buff_initialized = true;


    state->cstate.cs_ProjInfo = NULL;
    state->cstate.cs_OuterTupleSlot = NULL;
    state->cstate.cs_ExprContext = NULL; 

    return TRUE;
}
 
/* 
 * ExecEndBuffer -
 *     releases any storage allocated
 */
void
ExecEndBuffer(TupleBuffer *node)
{
    TupleBufferState	*state;
    Plan *outerPlan;

    state = node->buffstate;
    ExecClearTuple(state->cstate.cs_ResultTupleSlot);
    pfree(state->buff_tuple);
    if (fclose((FILE * ) state->buff_file) == EOF) {
	elog(WARN, "ExecEndBuffer: can't close temporary file");
    }
    outerPlan = outerPlan(node);
    ExecEndNode((Plan*) outerPlan, (Plan *)node);
}

/*
 * ExecCountSlotsBuffer
 */
int
ExecCountSlotsBuffer(TupleBuffer *node)
{
    return
	(ExecCountSlotsNode(outerPlan(node)) + BUFFER_NSLOTS);
}

/*
 * ExecBufferMarkPos --
 *	marks the current read position in the buffer file 
 */
void
ExecBufferMarkPos(TupleBuffer *node)
{
    TupleBufferState *state;
    HeapTuple tuple;
    FILE *fileDesc;

    state = node->buffstate;
    if (!state->buff_initialized) {
	elog(WARN, "ExecBufferMarkPos: marking an uninitialized buffer");
    }
    if (state->buff_nextPos == state->buff_fileLen) {
	/* 
	 * marking the last tuple retrieved from outer plan
	 * -> we needn't keep the previous ones including the 
	 *    one we've just read (erase them) 
	 */
	fileDesc = (FILE *) state->buff_file;
	if (fseek(fileDesc, 0, SEEK_SET) < 0) {
	    elog(WARN, "ExecBufferMarkPos: can't fseek in temporary file");
	}
	state->buff_nextPos = 0;
	state->buff_fileLen = 0;
	state->buff_markedPos = state->buff_nextPos;
    } else {
	/* we're reading old tuples -> don't erase them */
	state->buff_markedPos = state->buff_nextPos;
    }
    return;
}

/* 
 * ExecBufferRestrPos --
 *	sets the current read position in the buffer file
 *	to the last marked position
 */
void
ExecBufferRestrPos(TupleBuffer *node)
{
    TupleBufferState *state;

    state = node->buffstate;
    state->buff_nextPos = state->buff_markedPos;
    if (fseek((FILE *) state->buff_file, state->buff_nextPos, SEEK_SET) < 0) {
	elog(WARN, "ExecBufferRestrPos: can't fseek in temporary file");
    }
    return;
}

/* ----------------------------------------------------------------
 *   	ExecBufferReScan
 *   
 *   	Rescans the relation.
 * ----------------------------------------------------------------
 */
void
ExecBufferReScan(TupleBuffer *node, ExprContext *exprCtxt, Plan* parent)
{
    ExecBufferRestrPos((TupleBuffer*)node);
    return;
}

/* ------------------------------------------------------------- 
 * AccountBuffer -
 *    creates and initializes a buffer node
 */
List *
AccountBuffer(TupleBuffer *node)
{
    Plan *outerPlan;

    outerPlan = outerPlan(node);
    return AccountNode(outerPlan);
}
