/*-------------------------------------------------------------------------
 *
 * nodeExchangein.c--
 *    Support routines for upper half of exchange pair. reads tuples out of
 *    ExchRecvChannel and passes them up
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/executor/nodeExchangein.c,v 1.14 1996/03/04 09:07:04 aoki Exp $
 *
 *-------------------------------------------------------------------------
 */
#include "executor/executor.h"
#include "executor/nodeExchangein.h"
#include "exchange.h"

#include "miscadmin.h"	/* for MyHostId */

/*
 * ExecExchangeInMarkTuple -
 * ExecExchangeInCheckTuple -
 *    utilities to {mark,check} the destination plan fragment of a 
 *    given tuple (for Channel multiplexing and demultiplexing).
 *
 *    a QepFragmentId of '0' indicates that a tuple is intended for
 *    any and all fragments that receive it.
 *
 *    we do the memmove's instead of assignments so this won't fall
 *    over if the HeapTuple members end up being non-'int'-aligned.
 */
void
ExecExchangeInMarkTuple(HeapTuple tuple, QepFragmentId id, u_int host)
{
    (void) memmove(&(tuple->t_chain), &id, sizeof(id));
    (void) memmove(&(tuple->t_ctid), &host, sizeof(host));
}

bool
ExecExchangeInCheckTuple(HeapTuple tuple, QepFragmentId id, u_int host)
{
    QepFragmentId tupid;
    u_int tuphost;
    bool status;

    if (!HeapTupleIsValid(tuple)) {
	return(false);
    }
    (void) memmove(&tupid, &(tuple->t_chain), sizeof(tupid));
    (void) memmove(&tuphost, &(tuple->t_ctid), sizeof(tuphost));
    (void) memset(&(tuple->t_chain), 0, sizeof(tuple->t_chain));
    (void) memset(&(tuple->t_ctid), 0, sizeof(tuple->t_ctid));
    status = (bool) (((tupid == 0) || (tupid == id)) &&
		     ((tuphost == 0) || (tuphost == host)));
#if 0
    fprintf(stderr,
	    "ExecExchangeInCheckTuple: <qep=%d host=%d> %s <qep=%d host=%d>\n",
	    tupid, tuphost,
	    (status == true) ? "==" : "!=",
	    id, host);
#endif
    return(status);
}

/* 
 * ExecExchangeIn -
 *    returns next tuple read from channel
 */
TupleTableSlot *
ExecExchangeIn(ExchangeIn *node)
{
    TupleTableSlot	*slot;
    ExchangeInState	*state;
    HeapTuple 		tuple;
    QepFragmentId	fid = ((Plan *) node)->qepFragId;
    u_int		hid = getMyHostId();

    state = node->exchstate;
    slot = state->cstate.cs_ResultTupleSlot;
    do {
	recvTuple(state->exch_channel, (Pointer *) &tuple);
	if (ExecExchangeInCheckTuple(tuple, fid, hid) == true) {
	    ExecStoreTuple(tuple, slot, InvalidBuffer, false);
	    return(slot);
	}
    } while (HeapTupleIsValid(tuple));
    (void) ExecClearTuple(slot);
    return(slot);
}

 
/* 
 * ExecInitExchangeIn -
 *    creates and initializes an exchange  node.
 */
bool
ExecInitExchangeIn(ExchangeIn *node, EState *estate, Plan *parent)
{
    Plan	*outerPlan;
    Oid		reloid;
    SiteAddr 	sendSites[16];
    QepFragmentId sendQepFragIds[16];
    int 	nSendSites;
    List 	*sendSiteList, *qepFragList;
    ExchangeInState *state;
    int 	i;
    ExchRecvChannel channel;

    /* assign the node's execution state */
    node->plan.state = estate;

    /* open receive channel */
    sendSiteList = node->exchsendingSites;
    qepFragList = node->exchsendingQepFragIds;
    nSendSites = length(sendSiteList);
    for (i = 0; i < nSendSites; i++) {
	SiteAddrNode *saddr = (SiteAddrNode *)lfirst(sendSiteList);
	sendSites[i].addr = saddr->addr;
	sendSites[i].port = 0;
	sendSites[i].hostid = saddr->hostid;
	sendSiteList = lnext(sendSiteList);
	sendQepFragIds[i] = lfirsti(qepFragList);
	qepFragList = lnext(qepFragList);
    }
    channel = openRecvChannel(sendSites, nSendSites,
        estate->es_qepFragId, sendQepFragIds);

    /* create plan state */
    state = makeNode(ExchangeInState);
    state->exch_channel = channel;
    node->exchstate = state;

    ExecAssignNodeBaseInfo(estate, &state->cstate, parent);

#define EXCHANGEIN_NSLOTS	1
    ExecInitResultTupleSlot(estate, &state->cstate);
    ExecAssignResultTypeFromTL((Plan*) node, &state->cstate);

    state->cstate.cs_ProjInfo = NULL;
    state->cstate.cs_OuterTupleSlot = NULL;
    state->cstate.cs_ExprContext = NULL;

    return TRUE;
}
 
/* 
 * ExecEndExchangeIn -
 *    releases any storage allocated.
 */
void
ExecEndExchangeIn(ExchangeIn *node)
{
    Plan		*outerPlan;
    AttrNumber		*scanAtts;
    ExchangeInState	*state;

    state = node->exchstate;
    ExecClearTuple(state->cstate.cs_ResultTupleSlot);
    closeRecvChannel(state->exch_channel);
}

/*
 * ExecCountSlotsExchangeIn
 */
int
ExecCountSlotsExchangeIn(ExchangeIn *node)
{
    return (EXCHANGEIN_NSLOTS);
}
