/*-------------------------------------------------------------------------
 *
 * nodeMerge.c--
 *    Support routines for merge node takes +2 sorted input plans and
 *    merges them into single stream of tuples
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/executor/nodeMerge.c,v 1.9 1995/04/20 01:23:09 sunita Exp $
 *
 *-------------------------------------------------------------------------
 */
#include "executor/executor.h"
#include "executor/nodeMerge.h"
#include "utils/lselect.h"

/* 
 * ExecMerge -
 *    returns next lowest tuple of all input plans
 */
TupleTableSlot *
ExecMerge(Merge *node)
{
    TupleTableSlot *slot, *resultSlot;
    MergeState	*state;
    HeapTuple tuple = NULL;
    List *mergePlans;
    short nextPlan;
    bool *done;
    int i, iter;
    int nPlans;

    mergePlans = node->mergeplans;
    state = node->mergestate;
    resultSlot = state->cstate.cs_ResultTupleSlot;

    nextPlan = state->merge_nextPlan;
    done = (bool *) state->merge_done;

    nPlans = length(mergePlans);

    for (iter=nextPlan; iter < nPlans + nextPlan; iter++) {
	i = iter % nPlans;

	if (!done[i]) {
	    slot = ExecProcNode((Plan*) nth(i, mergePlans), (Plan*)node);
	
	    if (!TupIsNull(slot)) {
		tuple = slot->val;
		break;	/* get a tuple, stop */
	    } else {
		if (!scheduleQueries) 
		    done[i] = true;
	    }
	}
    }
    if (scheduleQueries) {
	nextPlan = i;
    } else {
	nextPlan++;
    }
    if (nextPlan == nPlans)
	nextPlan = 0;
    state->merge_nextPlan = nextPlan;
    
    return ((TupleTableSlot*)
	    ExecStoreTuple(tuple, resultSlot, InvalidBuffer, false));
}

 
/* 
 * ExecInitMerge -
 *    creates and initializes a merge node
 */
bool
ExecInitMerge(Merge *node, EState *estate, Plan *parent)
{
    MergeState *state;
    List *mergePlans;
    int nMergePlans; /* number of plans to merge */
    int i;
    Plan *inputPlan;
    static int anyInt;
    Plan *firstPlan;

    node->plan.state = (EState*)estate;

    /* create plan state */
    state = makeNode(MergeState);
    node->mergestate = state;

    ExecAssignNodeBaseInfo(estate, &state->cstate, parent);

#define MERGE_NSLOTS	1
    ExecInitResultTupleSlot(estate, &state->cstate);
    /* assign result slot descriptor from first plan in input list */
    mergePlans = node->mergeplans;
    nMergePlans = length(mergePlans);

    /* initialize input nodes */
    for (i = 0; i < nMergePlans; i++) {
	(void) ExecInitNode((Plan*) nth(i, mergePlans), estate, (Plan*) node);
    }

    firstPlan = (Plan*) nth(0, mergePlans);
    ExecAssignResultType(&state->cstate, ExecGetTupType(firstPlan));

    state->cstate.cs_ProjInfo = NULL;
    state->cstate.cs_OuterTupleSlot = NULL;
    state->cstate.cs_ExprContext = NULL; 

    /* initialize merge state information */
    state->merge_done = (bool *) palloc(nMergePlans * sizeof(bool));
    for (i = 0; i < nMergePlans; i++) {
	state->merge_done[i] = false;
    }
    state->merge_nextPlan = 0;
    state->merge_markPos = 0;

    return TRUE;
}
 
/* 
 * ExecEndMerge -
 *    releases any storage allocated
 */
void
ExecEndMerge(Merge *node)
{
    MergeState	*state;
    int nMergePlans;
    List *mergePlans;
    int i;

    state = node->mergestate;
    ExecClearTuple(state->cstate.cs_ResultTupleSlot);

    mergePlans = node->mergeplans;
    nMergePlans = length(mergePlans);
    for (i = 0; i < nMergePlans; i++) {
	ExecEndNode((Plan*) nth(i, mergePlans), (Plan*)node);
    }
}

/*
 * ExecCountSlotsMerge
 */
int
ExecCountSlotsMerge(Merge *node)
{
    int nSlots = 0;
    List *plan;
    List *mergePlans = node->mergeplans;

    foreach (plan, mergePlans) {
	nSlots += ExecCountSlotsNode((Plan*) lfirst(plan));
    }
    return (nSlots + MERGE_NSLOTS);
}

/*
 * ExecMergeMarkPos 
 *      -- mark the current plan being scanned
 */
void
ExecMergeMarkPos(Merge * node)
{
    MergeState *state;
    short markPos;
    int i;
    int nMergePlans;

    state = node->mergestate;
    markPos = state->merge_nextPlan;
    nMergePlans = length(node->mergeplans);

    state->merge_markPos = markPos;
    for(i = 0; i < nMergePlans; i++) {
	ExecMarkPos((Plan*) nth(i, node->mergeplans));
    }
}

/*
 * ExecMergeRestrPos --
 *   -- requires restoring position of the marked plan and rescanning
 *      all the plans between the markedPlan and the nextPlan
 */
void
ExecMergeRestrPos(Merge * node)
{
    MergeState *state;
    int i;
    int nMergePlans;

    state = node->mergestate;
    state->merge_nextPlan = state->merge_markPos;
    nMergePlans = length(node->mergeplans);

    for(i = 0; i < nMergePlans; i++) {
	state->merge_done[i] = false;
	ExecRestrPos((Plan*) nth(i, node->mergeplans));
    }
    
}

/*
 * ExecMergeReScan --
 *     rescan all the plans underneath the merge node
 */
void
ExecMergeReScan(Merge *node, ExprContext *exprCtxt, Plan *parent)
{
    MergeState *state;
    List *mergePlans;
    int nMergePlans; /* number of plans to merge */
    int i;
    Plan *inputPlan;
    static int anyInt;
    Plan *firstPlan;

    state =  node->mergestate;
    mergePlans = node->mergeplans;
    nMergePlans = length(mergePlans);
    state->merge_nextPlan = 0;
    state->merge_markPos = 0;

    /* rescan input nodes */
    for (i = 0; i < nMergePlans; i++) {
	ExecReScan((Plan*) nth(i, mergePlans), exprCtxt, (Plan*) node);
	state->merge_done[i] = false;
    }
}

    
 
/* 
 * AccountMerge -
 *    creates a list of subqueries on tertiary memory relations
 */
List *
AccountMerge(Merge *node)
{
    List *mergePlans;
    int nMergePlans; /* number of plans to merge */
    int i;
    List *list = NIL;

    mergePlans = node->mergeplans;
    nMergePlans = length(mergePlans);

    /* initialize input nodes */
    for (i = 0; i < nMergePlans; i++) {
	list = nconc (list, AccountNode((Plan*) nth(i, mergePlans)));
    }
    return list;
}





