/*-------------------------------------------------------------------------
 *
 * queue.c--
 *      Queue definitions. See queue.h for clarifications.
 *     
 * Copyright (c) 1995, Regents of the University of California
 *
 * $Id: queue.c,v 1.1 1995/07/04 07:38:23 searcher Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include "postgres.h"
#include "lib/queue.h"
#include "utils/elog.h"
#include "utils/palloc.h"

Queue* QMakeQueue()
{
    Queue *newQueue = (Queue *)palloc(sizeof(Queue));
    if (newQueue == NULL) {
	fprintf(stderr, "QMakeQueue: Out of memory!");
    }

    newQueue->front = NULL;
    newQueue->back = NULL;
    newQueue->length = 0;

    return newQueue;
}

bool QIsEmpty(Queue *q)
{
    if (q == NULL) {
	fprintf(stderr, "QIsEmpty: Null Queue");
    }

    return (bool)(q->length == 0);
}

int QLength(Queue *q)
{
    if (q == NULL) {
	fprintf(stderr, "QLength: Null Queue");
    }

    return q->length;
}

void QEnqueue(Queue *q, void *item)
{
    if (q == NULL) {
	fprintf(stderr, "QEnqueue: Null Queue");
    }

    if (q->back == NULL) {
	q->back = (QElem *)palloc(sizeof(QElem));
	q->front = q->back;
    }
    else {
	q->back->next = (QElem *)palloc(sizeof(QElem));
	q->back = q->back->next;
    }
    if (q->back == NULL) {
	fprintf(stderr, "QEnqueue: Out of memory!");
    }

    q->back->next = NULL;
    q->back->key = 0;
    q->back->item = item;

    q->length++;
}

void *QDequeue(Queue *q)
{
    void *oldFront;
    QElem *temp;

    if (q == NULL) {
	fprintf(stderr, "QDequeue: Null Queue");
    }
    if (q->front == NULL) {
	fprintf(stderr, "QDequeue: Queue empty");
    }

    oldFront = q->front->item;

    if (q->front == q->back) {
	pfree((void *)q->front);
	q->front = q->back = NULL;
    }
    else {
	temp = q->front;
	q->front = q->front->next;
	pfree((void *)temp);
    }

    q->length--;

    return oldFront;
}

void *QFrontPeek(Queue *q)
{
    if (q == NULL) {
	fprintf(stderr, "QFrontPeek: Null Queue");
    }
    if (q->front == NULL) {
	fprintf(stderr, "QFrontPeek: Queue empty");
    }

    return q->front->item;
}

void QKeyEnqueue(Queue *q, void *item, Oid key)
{
    if (q == NULL) {
	fprintf(stderr, "QKeyEnqueue: Null Queue");
    }

    if (q->back == NULL) {
	q->back = (QElem *)palloc(sizeof(QElem));
	q->front = q->back;
    }
    else {
	q->back->next = (QElem *)palloc(sizeof(QElem));
	q->back = q->back->next;
    }
    if (q->back == NULL) {
	fprintf(stderr, "QKeyEnqueue: Out of memory");
    }

    q->back->next = NULL;
    q->back->key = key;
    q->back->item = item;

    q->length++;
}

void *QRemoveKey(Queue *q, Oid key)
{
    QElem *i, *iprev;
    void *keyItem;

    if (q == NULL) {
	fprintf(stderr, "QRemoveKey: Out of memory!");
    }

    i = q->front;

    while (i != NULL) {
	if (key == i->key) {
	    keyItem = i->item;

	    if (q->front == i && q->back == i)
		q->front = q->back = NULL;
	    else if (q->back == i) {
		iprev->next = NULL;
		q->back = iprev;
	    }
	    else if (q->front == i)
		q->front = i->next;
	    else
		iprev->next = i->next;

	    pfree((void *)i);

	    q->length--;

	    return keyItem;
	}
	iprev = i;
	i = i->next;
    }

    elog(NOTICE, "QRemoveKey: Key %d not found", key);
    return NULL;
}

void QPrint(Queue *q)
{
    QElem *i;

    if (q == NULL) {
	fprintf(stderr, "QPrint: Null Queue");
    }

    i = q->front;
    while (i != NULL) {
	fprintf(stdout, "%d ", i->key);
	i = i->next;
    }

    fprintf(stdout, "\n");
}
