/*-------------------------------------------------------------------------
 *
 * call_exec.c--
 *    call-facility into executor running in main thread of backend
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/call_exec.c,v 1.10 1996/06/05 19:19:52 sunita Exp $
 *
 * NOTES
 *   implemented as a single-slot input and single-slot output queue for
 *   call argument and call result -> requests are serialized
 * 	
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <rpc/types.h>
#include <pthread.h>
#include "utils/elog.h"
#include "rpc/call_exec.h"
#include "rpc/rpcutils.h"

/* file-local variables */
static thdproc_t callProc;		/* procedure to execute */
static bool_t callInXact;		/* execute in transaction? */
static void *callArg, *callResult;
static FreeProc argFreeProc;
static SVCXPRT *callTransportHandle;
static bool_t callRequestPending = FALSE;
static bool_t callResultPending = FALSE;
static pthread_mutex_t argLock; /* protects callArg */
static pthread_mutex_t resultLock; /* protects callResult */
static pthread_cond_t execGotCalled; /* when exec. got called */
static pthread_cond_t requestCleared; /* when exec. accepted request */
static pthread_cond_t execReturned; /* when exec. returned */
static pthread_cond_t resultCleared; /* when caller read result */

static bool_t initialized = FALSE;
    /* TRUE if above variables have been initialized */
static void
init(void)
{
    pthread_mutex_init(&argLock, pthread_mutexattr_default);
    pthread_mutex_init(&resultLock, pthread_mutexattr_default);
    pthread_cond_init(&execGotCalled, pthread_condattr_default);
    pthread_cond_init(&requestCleared, pthread_condattr_default);
    pthread_cond_init(&execReturned, pthread_condattr_default);
    pthread_cond_init(&resultCleared, pthread_condattr_default);
    initialized = TRUE;
}

void
InitExecVars()
{
  init();
}

/*
 * callExec -
 *    Called by request handler. The argument to the call represented by reqNo
 *    is expected to be converted already and is passed in arg. The result, if 
 *    any, has to be pre-allocated in the request handler. The argument need
 *    not be deallocated in the caller, this can be done later on by the exec.
 *    via a call to freeProc.
 *
 *    All requests to the exec. are serialized, because there's only a single
 *    exec. running. A request blocks if the exec. is busy at the time of
 *    arrival.
 */
void
callExec(thdproc_t reqProc,
	 bool_t reqInXact,
	 void *arg,
	 void *result,
	 FreeProc freeProc,
	 SVCXPRT *transportHandle)
{
    int status;

    if (!initialized) {
	init();
    }
    status = pthread_mutex_lock(&argLock);
    checkAndExit(status, "callExec: can't lock argLock");
    while (callRequestPending) {
	status = pthread_cond_wait(&requestCleared, &argLock);
	checkAndExit(status, "callExec: can't wait on requestCleared");
    }
    callRequestPending = TRUE;
    callProc = reqProc;
    callInXact = reqInXact;
    callArg = arg;
    callResult = result;
    argFreeProc = freeProc;
    callTransportHandle = transportHandle;
    status = pthread_mutex_unlock(&argLock);
    checkAndExit(status, "callExec: can't unlock argLock");

    status = pthread_cond_broadcast(&execGotCalled);
    checkAndExit(status, "callExec: can't signal execGotCalled");

    status = pthread_mutex_lock(&resultLock);
    checkAndExit(status, "callExec: can't lock resultLock");
    while (!callResultPending) {
	status = pthread_cond_wait(&execReturned, &resultLock);
	checkAndExit(status, "callExec: can't wait on execReturned");
    }
    callResultPending = FALSE;
    status = pthread_mutex_unlock(&resultLock);
    checkAndExit(status, "callExec: can't unlock resultLock");

    status = pthread_cond_broadcast(&resultCleared);
    checkAndExit(status, "callExec: can't signal resultCleared");
}

/*
 * getRequest -
 *    Called by exec. in main thread to obtain the next request. The exec. 
 *    is assumed to know the underlying types of 'arg' and 'result'. The
 *    transport handle is necessary to free the argument (as an additional
 *    argument to freeProc beside 'arg'). 
 */
void
getRequest(thdproc_t *reqProc,
	   bool_t *reqInXact,
	   void **argPtr,
	   void **resultPtr,
	   FreeProc *freeProcPtr,
	   SVCXPRT **transportHandlePtr)
{
    int status;

    if (!initialized) {
	init();
    }
    status = pthread_mutex_lock(&argLock);
    checkAndElog(status, "getRequest: can't lock argLock");
    while (!callRequestPending) {
	status = pthread_cond_wait(&execGotCalled, &argLock);
	checkAndElog(status, "getRequest: can't wait on execGotCalled");
    }
    callRequestPending = FALSE;
    *reqProc = callProc;
    *reqInXact = callInXact;
    *argPtr = callArg;
    *resultPtr = callResult;
    *freeProcPtr = argFreeProc;
    *transportHandlePtr = callTransportHandle;
    status = pthread_mutex_unlock(&argLock);
    checkAndElog(status, "getRequest: can't unlock argLock");
    status = pthread_cond_broadcast(&requestCleared);
    checkAndElog(status, "getRequest: can't signal requestCleared");
}

/*
 * returnResult -
 *    Called by exec. to signal that the most recent request is done and it has
 *    to return results. The results are passed implicitly through the 'result'
 *    parameter in getRequest.
 */
void
returnResult(void)
{
    int status;

    if (!initialized) {
	init();
    }
    status = pthread_mutex_lock(&resultLock);
    checkAndElog(status, "returnResult: can't lock resultLock");
    while (callResultPending) {
	status = pthread_cond_wait(&resultCleared, &resultLock);
	checkAndElog(status, "returnResult: can't wait on resultCleared");
    }
    callResultPending = TRUE;
    status = pthread_mutex_unlock(&resultLock);
    checkAndElog(status, "returnResult: can't unlock resultLock");
    status = pthread_cond_broadcast(&execReturned);
    checkAndElog(status, "returnResult: can't signal execReturned");
}









