/*-------------------------------------------------------------------------
 *
 * copies_be.c--
 *    Copies-related calls to the exec./backend
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/copies_be.c,v 1.10 1996/05/31 17:14:11 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <rpc/types.h>
#include <pthread.h>
#include "postgres.h"
#include "utils/elog.h"
#include "mariposa_rpc.h"		/* For procedure constants */
#include "rpc/rpcutils.h"
#include "rpc/call_exec.h"		/* Backend calling facility */
#include "rpc/rexec.h"
#include "sitemgr/rpcerror.h"
#include "copies.h"

#include "access/heapam.h"
#include "access/htup.h"
#include "access/genam.h"
#include "access/itup.h"
#include "access/xact.h"
#include "access/miscam.h"
#include "storage/bufmgr.h"
#include "access/transam.h"
#include "catalog/catname.h"
#include "catalog/pg_class.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_contract.h"
#include "catalog/indexing.h"
#include "utils/syscache.h"
#include "utils/relcache.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "utils/nabstime.h"
#include "commands/vacuum.h"

extern unsigned short tcpPort;

/* ----------------
 *  Routines that handle establishing a copy.
 * ----------------
 */
RPCErrorCode EstablishCopy(CopySetupInfo *copysetup);

static RPCErrorCode InstallSideFile(Oid relid, Oid *sidefrastoreid);
static RPCErrorCode InstallSendContract(CopySetupInfo *copysetup,
					Oid sidefrastoreid);
static RPCErrorCode InstallReceiveContract(CopySetupInfo *copysetup);

/* ----------------
 *  Routines that handle sending an update stream.
 * ----------------
 */
typedef bool (*UpdateStreamFilterProc)(HeapTuple h);

TransactionId Cp_SendUpdateStream(UpdateContractInfo *updatecontract);

static RPCErrorCode Cp_VacuumSideFile(UpdateContractInfo *updatecontract);
static RpcConn *StartUpdateTransfer(SiteAddr *myaddr,
				    TransactionId *remotexactid, int loc);
static RPCErrorCode PutUpdates(RpcConn *rpcc,
			       UpdateContractInfo *updatecontract,
			       Oid actualsend, Oid actualrecv,
			       UpdateStreamFilterProc filter,
			       bool nsvcTransferFinished);
static bool TupleQualifiesForUpdate(HeapTuple htup,
				    UpdateContractInfo *updatecontract);
/* ----------------
 *  Tuple filters for sending update streams
 * ----------------
 */
static bool noFilter(HeapTuple htup);
static bool pgClassFilter(HeapTuple htup);
static bool pgAttrFilter(HeapTuple htup);
static bool pgFragFilter(HeapTuple htup);

/* ----------------
 *  Routines that handle receiving an update stream.
 * ----------------
 */
static RPCErrorCode Cp_PutUpdate(TransferDataBuffer *databuf);
static bool OidLookupNecessary(UpdateContractInfo *contract, HeapTuple htup);

/* ----------------
 *  Routines that handle initializing update streams.
 * ----------------
 */
RPCErrorCode Cp_InitializeUpdateStreams();

/* ----------------
 *  Routines that deal with administering contracts.
 * ----------------
 */
static void CancelRemoteContract(UpdateContractInfo *contract);
static RPCErrorCode CancelContract(Oid contractid);
static void UnscheduleContract(RpcConn *sm_rpcc, UpdateContractInfo *contract);

/* ----------------
 *  Utility routines
 * ----------------
 */
static void CopyUpdateContractInfo(UpdateContractInfo *source,
				   UpdateContractInfo *dest);

/* ----------------
 *  File-scope declarations for update stream reception routines.
 *  This is possible because of twopc mutexes.
 *
 *  Proper RPC way is to return these to sender as opaque type. The sender
 *  would then pass this back to us when calling FragmentPutData. That means
 *  writing xdr routines for all those types. Ugh.
 * ----------------
 */
static Relation xFragDesc = NULL;
static Relation xFragOidHashIndexDesc = NULL;
static IndexScanDesc xFragOidHashIndexscan = NULL;

/*
 * transferNumTuplesPut records how many net tuples actually make it over an
 * the update stream reception, so we can add the result in the pg_fragment
 * entry after Cp_PutUpdate is called the last time.
 */
static int transferNumTuplesPut = 0;

#define SENDBUFSIZE 7168		/* copied from exchange.c */


/*****************************************************************************
 * Routines that handle establishing a copy.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_ESTABLISHCOPY
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_ESTABLISHCOPY(void *args)
{
    char		**argv = (char **)args;
    CopySetupInfo	*rpcarg = (CopySetupInfo *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result;

    elog(NOTICE, "handleCall_ESTABLISHCOPY.");

    mainThreadExec((thdproc_t)EstablishCopy, rpcarg, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_ESTABLISHCOPY: Can't send reply.\n");
    }

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	EstablishCopy
 *
 *	Sets up everything necessary to enable this site to get along
 *	properly with new copy in the system.
 *
 *	1. If necessary, creates side-file to store fragment changes.
 *	2. Decides whether to setup send and/or receive update stream
 *	   contracts.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
 EstablishCopy(CopySetupInfo *setup)
{
    Oid			sidefrastoreid;
    RPCErrorCode	result = SUCCESS;

    elog(NOTICE, "ESTABLISHCOPY");

    /*
     * Case 1: Read-write user copy.
     *	       We install both send and receive contracts, sidefile necessary.
     */
    if (!setup->isnamesvc && !setup->isreadonly) {
	result = InstallSideFile(setup->ourcopystoreid, &sidefrastoreid);
	if (result == SUCCESS)
	    result = InstallReceiveContract(setup);
	if (result == SUCCESS)
	    result = InstallSendContract(setup, sidefrastoreid);
    }
    /*
     * Case 2: Read-only user copy or name-service order, and we are the buyer.
     *	       We install a receive contract.
     */
    else if (setup->isreadonly && setup->ambuyer) {
	result = InstallReceiveContract(setup);
    }
    /*
     * Case 3: Read-only user copy or name-service order, we aren't the buyer.
     *	       We install a send contract, sidefile necessary.
     */
    else if (setup->isreadonly && !setup->ambuyer) {
	if (!setup->isnamesvc)
	    result = InstallSideFile(setup->ourcopystoreid, &sidefrastoreid);
	if (result == SUCCESS)
	    result = InstallSendContract(setup, sidefrastoreid);
    }

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *	InstallSideFile
 *
 *	If no sidefile exists for relation relid, creates one.
 *	In either case writes the sidefile oid to the reference
 *	sidefrastoreid.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
InstallSideFile(Oid relid, Oid *sidefrastoreid)
{
    Relation		fragdesc;
    NameData		sideName;

    fragdesc = heap_open(relid);
    if (!RelationIsValid(fragdesc)) {
	elog(NOTICE, "InstallSideFile: Can't open send fragment.");
	return EXECUTION_ERROR;
    }

    /* ----------------
     * 	Create side file relation if necessary. Stuff sidefile oid in
     *  sidefrastoreid.
     * ----------------
     */

    if (!fragdesc->rd_iscopied) {
	/*
	 * This is to check if an old oid hash index is hangin' around...
	 * If so, keep the one we got. What should happen is after a drop
	 * copy it gets automatically deleted.
	 */
	Relation sidedesc;

	GenerateSideFileName((Name) &sideName, fragdesc);

	sidedesc = heap_openr((Name) &sideName);
	if (!RelationIsValid(sidedesc)) {
	    /*
	     * Create the side-file and get its oid to store in update
	     * contract. The schema for the side-file is same as the fragment.
	     * Deletes are represented by null tuples in the side-file.
	     */
	    *sidefrastoreid = heap_create(sideName.data, 'n',
					  fragdesc->rd_rel->relnatts,
					  DEFAULT_SMGR, fragdesc->rd_att);
	    /*
	     * the side file does not need fragment information since it
	     * is never scanned using the standard query code path (i.e.,
	     * we always scan it ourselves, using the AM interface).
	     */
	    CommandCounterIncrement();
	}
	else {
	    *sidefrastoreid = sidedesc->rd_id;

	    heap_close(sidedesc);
	}
    } else {
	*sidefrastoreid = fragdesc->rd_copysidef;
    }

    heap_close(fragdesc);

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *	InstallSendContract
 *
 *	Make a send contract from information in setup (and sidefrastoreid),
 *	enter it into pg_contract, and schedule the contract on the update
 *	stream queue.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
InstallSendContract(CopySetupInfo *setup, Oid sidefrastoreid)
{
    UpdateContractInfo	updatecontract;
    char		query[1000];
    RpcConn		*sm_rpcc;
    RPCErrorCode	result;

    /* ----------------
     *  Fill in contract, and record in pg_contract.
     * ----------------
     */
    updatecontract.sidefrastoreid = sidefrastoreid;
    updatecontract.lastupdcommit =
	(setup->isnamesvc) ? EPOCH_ABSTIME : GetCurrentTransactionStartTime();

    updatecontract.contractid = setup->contractid;
    updatecontract.sendfrastoreid = setup->ourcopystoreid;
    updatecontract.recvfrastoreid = setup->theircopystoreid;
    updatecontract.otherloc = setup->thereloc;
    updatecontract.period = setup->period;
    /*
     * Typemasks are initialized to 64, so that they print out as recognizable
     * characters, which helps when inserting into pg_contract.
     */
    updatecontract.typemask = (char)64;

    if (setup->isnamesvc) {
	ContractSetNameService(&updatecontract);
    } else {
	ContractSetUserCopy(&updatecontract);
    }
    if (setup->isreadonly) {
	ContractSetReadOnly(&updatecontract);
    } else {
	ContractSetReadWrite(&updatecontract);
    }
    if (setup->isnsspecific) {
	ContractSetNSSpecific(&updatecontract);
    } else {
	ContractSetNSRemote(&updatecontract);
    }
    ContractSetSend(&updatecontract);

    sprintf(query,
	    "INSERT INTO pg_contract\
 (contractid, sendfrastoreid, recvfrastoreid, sidefrastoreid, lastupdcommit,\
 otherloc, period, typemask) values\
 (%d,%d,%d,%d,%d,%d,%d,'%c'::char);",
	    updatecontract.contractid,
	    updatecontract.sendfrastoreid,
	    updatecontract.recvfrastoreid,
	    updatecontract.sidefrastoreid,
	    updatecontract.lastupdcommit,
	    updatecontract.otherloc,
	    updatecontract.period,
	    updatecontract.typemask);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    PrintContract("InstallSendContract", &updatecontract);

    /* ----------------
     *  Call local sitemgr to schedule update stream.
     *
     *  Since this is the last thing that happens at the remote site when a
     *  copy is made from it, if the calling site detects we failed, then it
     *  knows this call must not have succeeded. This is important for error
     *	handling... the caller doesn't have to call CP_UNSCHEDULESTREAM here
     *  if we fail.
     * ----------------
     */
    sm_rpcc = Rpc_connectSitemgr(getMyHostId(), 0, "EstablishCopy",
				 RPC_ONERR_WARN);

    if (!Rpc_call(sm_rpcc, RPC_CP_SCHEDULESTREAM, &updatecontract, &result)) {
	Rpc_disconnect(sm_rpcc);
	elog(NOTICE, "EstablishCopy: CP_SCHEDULESTREAM rpc call failed.");
	return EXECUTION_ERROR;
    }

    Rpc_disconnect(sm_rpcc);

    if (result != SUCCESS) {
	elog(NOTICE, "EstablishCopy: CP_SCHEDULESTREAM error %d.", result);
	return EXECUTION_ERROR;
    }

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *	InstallReceiveContract
 *
 *	Make a receive contract from information in setup, and enter it
 *	into pg_contract.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
InstallReceiveContract(CopySetupInfo *setup)
{
    UpdateContractInfo	updatecontract;
    char		query[1000];
    RpcConn		*sm_rpcc;

    /* ----------------
     *  Fill in contract, and record in pg_contract.
     * ----------------
     */
    updatecontract.sidefrastoreid = 0;
    updatecontract.lastupdcommit =
	(setup->isnamesvc) ? EPOCH_ABSTIME : GetCurrentTransactionStartTime();

    updatecontract.contractid = setup->contractid;
    updatecontract.sendfrastoreid = setup->theircopystoreid;
    updatecontract.recvfrastoreid = setup->ourcopystoreid;
    updatecontract.otherloc = setup->thereloc;
    updatecontract.period = setup->period;
    /*
     * Typemasks are initialized to 64, so that they print out as recognizable
     * characters, which helps when inserting into pg_contract.
     */
    updatecontract.typemask = (char)64;

    if (setup->isnamesvc) {
	ContractSetNameService(&updatecontract);
    } else {
	ContractSetUserCopy(&updatecontract);
    }
    if (setup->isreadonly) {
	ContractSetReadOnly(&updatecontract);
    } else {
	ContractSetReadWrite(&updatecontract);
    }
    if (setup->isnsspecific) {
	ContractSetNSSpecific(&updatecontract);
    } else {
	ContractSetNSRemote(&updatecontract);
    }
    ContractSetReceive(&updatecontract);

    sprintf(query,
	    "INSERT INTO pg_contract\
 (contractid, sendfrastoreid, recvfrastoreid, sidefrastoreid, lastupdcommit,\
 otherloc, period, typemask) values\
 (%d,%d,%d,%d,%d,%d,%d,'%c'::char);",
	    updatecontract.contractid,
	    updatecontract.sendfrastoreid,
	    updatecontract.recvfrastoreid,
	    updatecontract.sidefrastoreid,
	    updatecontract.lastupdcommit,
	    updatecontract.otherloc,
	    updatecontract.period,
	    updatecontract.typemask);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    PrintContract("InstallReceiveContract", &updatecontract);

    return SUCCESS;
}


/*****************************************************************************
 * Routines that handle sending an update stream.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_CP_SENDUPDATESTREAM
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 *	Vacuums sidefile, then sends the update stream.
 *	If the process succeeds returns commit time to caller.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_CP_SENDUPDATESTREAM(void *args)
{
    char		**argv = (char **)args;
    UpdateContractInfo	*rpcarg = (UpdateContractInfo *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    int			lastupdcommit;
    RPCErrorCode	result = SUCCESS;
    SendUpdateStreamResult *returnresult;
    
    returnresult =
	(SendUpdateStreamResult *)palloc(sizeof(SendUpdateStreamResult));
    memset(returnresult, 0, sizeof(SendUpdateStreamResult));

    if (ContractIsUserCopy(rpcarg)) {
	/*
	 * Before sending results of side-file, vacuum it and its repr.
	 * fragment in a separate transaction.
	 */
	mainThreadExecInXact((thdproc_t)Cp_VacuumSideFile, rpcarg, &result);
    }

    if (result != SUCCESS) {
	returnresult->error = EXECUTION_ERROR;
    }
    else {
	/*
	 * Now side-file is ok to be sent. Call the function that sends it.
	 */
	mainThreadExecInXact((thdproc_t)Cp_SendUpdateStream, rpcarg,
			     &lastupdcommit);
    
	if (lastupdcommit == 0) {
	    /*
	     * Signals invalid commit time, meaning the sending of update
	     * stream for this fragment failed. Caller will know what to do
	     * about it.
	     */
	    returnresult->error = EXECUTION_ERROR;
	}
	else {
	    /*
	     * We store the commit time back in the contract. It is
	     * necessary, because we don't want to resend updates for tuples
	     * that we've already covered.
	     */
	    returnresult->committime = lastupdcommit;
	    returnresult->error = SUCCESS;
	}
    }

    if (!svc_sendreply(transp, xdr_SendUpdateStreamResult, returnresult)) {
  	errorExit("handleCall: can't send reply\n");
    }

    pfree(returnresult);
    free(rpcarg);
    free(args);
}

/*
 *      noFilter
 *
 *      Update stream filter that lets every tuple pass.
 */
static bool
noFilter(HeapTuple htup)
{
    return (true);
}

/*
 *      pgClassFilter
 *
 *      Update stream filter which tries to keep catalog information
 *      from being propagated via nameservice. For this, it rejects
 *      any pg_class tuple with oid < transam.h:BootstrapObjectIdData.
 */
static bool
pgClassFilter(HeapTuple htup)
{
    Form_pg_class classTuple;

    if (htup->t_oid < BootstrapObjectIdData) {
        return false;
    }
    return true;
}

/*      pgAttrFilter
 *
 *      Filters out system catalog-related data, indicated by
 *      attrelid < transam.h:BootstrapObjectIdData.
 */
static bool
pgAttrFilter(HeapTuple htup)
{
    AttributeTupleForm attr;

    attr = (AttributeTupleForm) GETSTRUCT(htup);
    if (attr->attrelid < BootstrapObjectIdData) {
        return false;
    }
    return true;
}

/*      pgFragFilter
 *
 *      Filters out system catalog-related data, indicated by
 *      frarelid < transam.h:BootstrapObjectIdData.
 */
static bool
pgFragFilter(HeapTuple htup)
{
    Form_pg_fragment frag;

    frag = (Form_pg_fragment) GETSTRUCT(htup);
    if (frag->fraclassid < BootstrapObjectIdData) {
        return false;
    }
    return true;
}

/* ----------------------------------------------------------------
 *    	Cp_SendUpdateStream
 *
 *	Handles sending of the side-file to a site to satisfy 'updatecontract'.
 *
 *	If User Copy:
 *	  Expects side-file to be vacuumed. Sends updates from sidefile.
 *
 *	If Name Service:
 *	  Depending on the type of service ordered in the contract, sends
 *	  from master system catalogs or nsvc system catalogs.
 *
 *	What happens:
 *	  Calls sitemgr at other site and tells it to get ready, and we wait
 *	  for it to send us an RPC connection to its executor. Then we send
 *	  the updates over the RPC connection (calling PutUpdates.)
 *
 *	  Establishes a "last update commmit time" for this contract, which
 *	  it records in pg_contract, and returns to the handleCall proc which
 *	  will return it to the sitemgr's update stream queue.
 *
 *	  If this fails at any point, we stop and return right away. We'll try
 *	  again next time.
 *
 *	Notes about Name Service system catalogs:
 *	  Currently those catalogs are the attribute, class, and fragment
 *	  ones. If you wish to add more system catalogs to name service, you
 *	  must edit this routine and also _vc_CalculateRelationWatermark.
 *
 * ----------------------------------------------------------------
 */
TransactionId
Cp_SendUpdateStream(UpdateContractInfo *updatecontract)
{
    SiteAddr		myaddr;
    TransactionId	remotexactid;
    RPCErrorCode	result;
    RpcConn		*exec_rpcc;
    int			lastupdatetime;
    char		query[1000];

    PrintContract("Cp_SendUpdateStream", updatecontract);

    myaddr.addr = getMyAddr();
    myaddr.port = tcpPort;
    myaddr.hostid = getMyHostId();

    /*
     * Notify remote sitemgr that we are going to send them updates,
     * and get an RPC connection to the backend. If they refuse we quit
     * out and return invalid commit time.
     */
    exec_rpcc = StartUpdateTransfer(&myaddr, &remotexactid,
				    updatecontract->otherloc);
    if (remotexactid == 0) {
	return 0; /* invalid commit time */
    }

    /*
     * Send the updates over the RPC connection, depends on type of contract.
     */
    if (ContractIsUserCopy(updatecontract)) {
	/*
	 * We send from the side-file to the receiving fragment.
	 */
	result = PutUpdates(exec_rpcc, updatecontract,
			    updatecontract->sidefrastoreid,
			    updatecontract->recvfrastoreid,
			    noFilter, false);
    }
    else if (ContractIsNSSpecific(updatecontract)) {
	/*
	 * We send from pg_attribute, pg_class, pg_fragment to
	 * pg_nsvcattr, pg_nsvcclass, pg_nsvcfrag.
	 */
	result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_attribute,
			    RelOid_pg_nsvcattr, pgAttrFilter, false);
	if (result == SUCCESS)
	    result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_class,
				RelOid_pg_nsvcclass, pgClassFilter, false);
	if (result == SUCCESS)
	    result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_fragment,
				RelOid_pg_nsvcfrag, pgFragFilter, true);
    }
    else if (ContractIsNSRemote(updatecontract)) {
	/*
	 * We send from pg_nsvcattr, pg_nsvcclass, pg_nsvcfrag to
	 * pg_nsvcattr, pg_nsvcclass, pg_nsvcfrag.
	 */
	result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_nsvcattr,
			    RelOid_pg_nsvcattr, pgAttrFilter, false);
	if (result == SUCCESS)
	    result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_nsvcclass,
				RelOid_pg_nsvcclass, pgClassFilter, false);
	if (result == SUCCESS)
	    result = PutUpdates(exec_rpcc, updatecontract, RelOid_pg_nsvcfrag,
				RelOid_pg_nsvcfrag, pgFragFilter, true);
    }
    else {
	elog(WARN, "Yo! What kinda contract is this?!?");
    }

    if (result != SUCCESS) {
	AbortRemoteSite(exec_rpcc, remotexactid);
	return 0;	/* Invalid commit time */
    }

    /*
     * We commit remote site here to avoid deadlock-- unfortunately, it's a
     * likely situation that these updates go on concurrently, and pg_contract
     * needs to be updated by both transactions at each involved site.
     */
    CommitRemoteSite(exec_rpcc, remotexactid);

    /*
     * We will set the transaction start time to be the last update commit
     * time for the contract.
     */
    lastupdatetime = GetCurrentTransactionStartTime();

    sprintf(query,
	    "UPDATE pg_contract SET lastupdcommit = %d WHERE contractid =\
 '%d'::oid AND sendfrastoreid = '%d'::oid",
	    lastupdatetime, updatecontract->contractid,
	    updatecontract->sendfrastoreid);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    printf("\n\n\n\n\n\n\n");

    return lastupdatetime;
}

/* ----------------------------------------------------------------
 *	Cp_VacuumSideFile
 *
 *	Vacuums the side-file associated with send fragment in
 *	updatecontract.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
Cp_VacuumSideFile(UpdateContractInfo *updatecontract)
{
    Relation		sidedesc;

    /* @@@ Also vacuum master fragment? */

    /*
     * Open reldescs for the side-file so we can get its relname,
     * and then call vacuum on the relname.
     */
    sidedesc = heap_open(updatecontract->sidefrastoreid);
    if (!RelationIsValid(sidedesc)) {
	elog(NOTICE, "Cp_VacuumSideFile: Can't open side-file %d.",
	     updatecontract->sidefrastoreid);

	return EXECUTION_ERROR;
    }
    vacuum(sidedesc->rd_rel->relname.data);
    heap_close(sidedesc);

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	StartUpdateTransfer
 *
 *	Connects to remote sitemgr that we wish to participate in the
 *	update transfer. If it's ok with them, a transaction is
 *	started at their site, we get the id for two phase commit reference,
 *	and we get a connection to the backend they've allocated.
 *
 * ----------------------------------------------------------------
 */
static RpcConn *
StartUpdateTransfer(SiteAddr *myaddr, TransactionId *remotexactid, int loc)
{
    RpcConn		*sm_rpcc;
    int			execAddr;
    int			execPort;

    RExecInit();

    /*
     * Call sitemgr with CP_UPDATETRANSFERSETUP to tell it that we're ready to
     * send updates. If it's ready, it will return us a valid transaction id.
     */
    sm_rpcc = Rpc_connectSitemgr(loc, 0, "Cp_SendUpdateStream",
				 RPC_ONERR_WARN);
    if (!Rpc_call(sm_rpcc, RPC_CP_UPDATETRANSFERSETUP, myaddr, remotexactid)) {
	Rpc_disconnect(sm_rpcc);
	elog(NOTICE, "StartUpdateTransfer: CP_UPDATETRANSFERSETUP rpc call failed.");
	*remotexactid = 0; /* Invalid commit time */
	return NULL;
    }
    Rpc_disconnect(sm_rpcc);

    if (*remotexactid <= 0) {
	elog(NOTICE, "StartUpdateTransfer: Receiving site can't prepare (CP_UPDATETRANSFERSETUP)");
	return NULL;
    }

    /*
     * Wait for the exec to forward the side-file tuples to.
     */
    RExecWait(&execAddr, &execPort, NULL);

    /*
     * When we reach here, the remote exec has contacted us and its
     * address is in execAddr/execPort. We will make a few RPC calls
     * to it directly.
     */
    return Rpc_connectExec(execAddr, execPort, "Cp_SendUpdateStream",
			   RPC_ONERR_WARN);
}

/* ----------------------------------------------------------------
 *    	PutUpdates
 *
 *	Scans the relation to send updates from, ships the update tuples
 *	into TransferDataBuffers, and packs them off to the receiving site.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
PutUpdates(RpcConn *rpcc, UpdateContractInfo *updatecontract,
    Oid actualsend, Oid actualrecv, UpdateStreamFilterProc filter,
    bool nsvcTransferFinished)
{
    TransferDataBuffer		databuf;
    Relation			reldesc;
    HeapScanDesc		relscan;
    HeapTuple			htup;
    char			buffer[SENDBUFSIZE];
    int				nbytes = 0;
    int				numTuples = 0;
    int				totalTuples = 0;	/* for debugging */
    RPCErrorCode		result;
    TimeQual			archiveTimeQual;
    
    elog(NOTICE, "Entering PutUpdates (%d.)", actualsend);
    Rpc_setCallerName(rpcc, "PutUpdates");

    CopyUpdateContractInfo(updatecontract, &databuf.transferInfo);
    databuf.transferInfo.recvfrastoreid = actualrecv;
    databuf.data.Data_val = buffer;
    databuf.transferStatus = INITIAL_PACKET;
    databuf.nsvcTransferFinished = false;

    /*
     * Open the relation that we send the updates from, and grab a write-lock
     * on it, so that people don't add tuples to the relation while we send
     * the updates off.
     */
    reldesc = heap_open(actualsend);
    if (!RelationIsValid(reldesc)) {
	elog(NOTICE, "PutUpdates: Can't open relation %d.", actualsend);
	return EXECUTION_ERROR;
    }

    RelationSetLockForWrite(reldesc);

    elog(NOTICE, "PutUpdates: lastupdcommit %d.",
	 updatecontract->lastupdcommit);

    archiveTimeQual = TimeFormRangedTimeQual(EPOCH_ABSTIME, CURRENT_ABSTIME);
    relscan = heap_beginscan(reldesc, 0, archiveTimeQual, 0, NULL);
    do {
	htup = heap_getnext(relscan, 0, NULL);
	if (HeapTupleIsValid(htup) &&
	            TupleQualifiesForUpdate(htup, updatecontract) &&
	            filter(htup)) {
	    int len = htup->t_len;

	    if ((len + nbytes) > SENDBUFSIZE) {
		/*
		 * Send the buffer off and start again.
		 */
		databuf.numTuples = numTuples;
		databuf.data.Data_len = nbytes;

		if (!Rpc_call(rpcc, RPC_CP_PUTUPDATE, &databuf, &result)) {
		    Rpc_disconnect(rpcc);
		    elog(NOTICE,
			 "PutUpdates: CP_PUTUPDATE rpc call failed.");

		    heap_endscan(relscan);
		    heap_close(reldesc);

		    return EXECUTION_ERROR;
		}

		if (result != SUCCESS) {
		    elog(NOTICE, "PutUpdates: CP_PUTUPDATE error %d.",
			 result);

		    heap_endscan(relscan);
		    heap_close(reldesc);

		    return EXECUTION_ERROR;
		}

		totalTuples += numTuples;

		numTuples = 0;
		nbytes = 0;

		databuf.transferStatus = MIDDLE_PACKET;
	    }
	    memmove(buffer + nbytes, htup, len);
	    numTuples++;
	    nbytes += len;
	}
    } while (htup);

    pfree(archiveTimeQual);
    heap_endscan(relscan);
    heap_close(reldesc);

    /*
     * Send last buffer.
     */
    databuf.numTuples = numTuples;
    databuf.data.Data_len = nbytes;
    databuf.transferStatus =
	(databuf.transferStatus == MIDDLE_PACKET) ? FINAL_PACKET : ONLY_PACKET;
    databuf.nsvcTransferFinished = nsvcTransferFinished;

    if (!Rpc_call(rpcc, RPC_CP_PUTUPDATE, &databuf, &result)) {
	Rpc_disconnect(rpcc);
	elog(NOTICE, "PutUpdates: CP_PUTUPDATE rpc call failed.");
	return EXECUTION_ERROR;
    }

    if (result != SUCCESS) {
	elog(NOTICE, "PutUpdates: CP_PUTUPDATE error %d.", result);
	return EXECUTION_ERROR;
    }

    elog(NOTICE, "Leaving PutUpdates.");

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	TupleQualifiesForUpdate
 *
 *	If the tmin of the tuple > last update sendoff, true for either
 *	a sidefile or a table which name service looks at.
 *
 *	If the tmax > last update sendoff, the tuple qualifies for sendoff
 *	only if it's a table which name service looks at and  if the
 *	tmax =! CURRENT/INVALID/NOEND_ABSTIME (it's been deleted since 
 * 	the last update stream was sent).
 *
 * ----------------------------------------------------------------
 */
static bool TupleQualifiesForUpdate(HeapTuple htup,
				    UpdateContractInfo *updatecontract)
{
    AbsoluteTime tmin, tmax;
    int itDoes;

    tmin = (AbsoluteTime) heap_getsysattr(htup, NULL,
        MinAbsoluteTimeAttributeNumber);
    tmax = (AbsoluteTime) heap_getsysattr(htup, NULL,
        MaxAbsoluteTimeAttributeNumber);
    /*elog(NOTICE, "tuple %d, min: %d, max: %d", htup->t_oid, tmin, tmax);*/
    if (tmin > updatecontract->lastupdcommit) {
	elog(NOTICE, "PutUpdates: New tuple %d (%d, %d)",
	    htup->t_oid, tmin, tmax);
	return true;
    }
    if (ContractIsNameService(updatecontract)) {
	itDoes = (tmax > updatecontract->lastupdcommit &&
	    tmax != CURRENT_ABSTIME && tmax != NOEND_ABSTIME &&
	    tmax != INVALID_ABSTIME);
	if (itDoes) {
	    elog(NOTICE, "PutUpdates: Deleted tuple %d (%d, %d)",
	        htup->t_oid, tmin, tmax);
	}
	return itDoes;
    }
}


/*****************************************************************************
 * Routines that handle receiving an update stream.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_CP_PUTUPDATE
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
#define BOGUS 12345678

void
handleCall_CP_PUTUPDATE(void *args)
{
    char		**argv = (char **)args;
    TransferDataBuffer	*rpcarg = (TransferDataBuffer *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result = BOGUS;

    elog(NOTICE, "Entering handleCall_CP_PUTUPDATE.");

    mainThreadExec((thdproc_t)Cp_PutUpdate, rpcarg, &result);
    /*
     * @@@ Been having problems... Cp_PutUpdate sometimes does not return
     *     and set result before it gets sent across the RPC transport,
     *     causing problems. So what we do here is make it wait with this
     *	   while loop. I'm not sure that's such a good thing though.
     *
     *	   But, that bug was a pain in the ass. It must have been some
     *	   memory leak because when I stepped through the code, it should
     *	   have made the function call in time according to my stepping,
     *	   but it simply didn't fucking do it synchronously. Jeff and Paul
     *	   you should remember I showed it to you in the summer.      -Rex
     */
    while (result == BOGUS);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_CP_PUTUPDATE: Can't send reply.\n");
    }

    elog(NOTICE, "Leaving handleCall_CP_PUTUPDATE.");

    if (rpcarg->data.Data_len > 0)
	free(rpcarg->data.Data_val);

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	Cp_PutUpdate
 *
 *	Receives the updates from a contracted update stream.
 *	Stores them in the fragment to be updated.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
Cp_PutUpdate(TransferDataBuffer *databuf)
{
    int			i;
    HeapTuple		tup;
    HeapTuple		targetTup;
    char		*data;
    ScanKeyData         skey[1];
    RetrieveIndexResult res;
    ItemPointer		iptr;
    char		query[1000];
    Oid			contractid = databuf->transferInfo.contractid;
    Oid			recvfrastoreid = databuf->transferInfo.recvfrastoreid;
    Relation		iDescs[Num_pg_nsvcattr_indices];
	/* assumes that Num_pg_nsvc{attr, class, frag}_indices are equal */
    char 		**indexNames;
    Buffer		buffer = InvalidBuffer;

    elog(NOTICE, "Entering Cp_PutUpdate (%d tuples.)", databuf->numTuples);

    /*
     * If we're just starting the move, need to open the the receiving
     * fragment and get its relation descriptor.
     */
    if (databuf->transferStatus == INITIAL_PACKET ||
	databuf->transferStatus == ONLY_PACKET)
    {
	xFragDesc = heap_open(recvfrastoreid);

	if (!RelationIsValid(xFragDesc)) {
	    elog(NOTICE, "Cp_PutUpdate: Can't open receiving fragment %d",
		 recvfrastoreid);
	    return EXECUTION_ERROR;
	}

	xFragOidHashIndexDesc = OpenRelationOidHashIndex(xFragDesc->rd_id);

	if (!RelationIsValid(xFragOidHashIndexDesc)) {
	    elog(NOTICE, "Cp_PutUpdate: Can't open receiving fragment %d's oid\
 hash index",
		 xFragDesc->rd_id);
	    return EXECUTION_ERROR;
	}

	ScanKeyEntryInitialize(&skey[0],
			       0x0,
			       ObjectIdAttributeNumber,
			       ObjectIdEqualRegProcedure,
			       ObjectIdGetDatum(0));

	xFragOidHashIndexscan =
	    index_beginscan(xFragOidHashIndexDesc, false, 1, (ScanKey)skey);

	transferNumTuplesPut = 0;
    }
    
    data = databuf->data.Data_val;

    for(i = 0; i < databuf->numTuples; i++) {
	tup = (HeapTuple)data;

	/*
	 * Do something with the tuple.
	 */
	if (ContractIsNameService(&databuf->transferInfo)) {
	    elog(NOTICE, "Cp_PutUpdate: processing %d (%d, %d)",
		tup->t_oid, tup->t_tmin, tup->t_tmax);
	    Assert(tup->t_tmax > databuf->transferInfo.lastupdcommit ||
	        tup->t_tmax == CURRENT_ABSTIME);
	    /*
	     * For a name service tuple, either have a delete or an insert.
	     * A delete is indicated by tmax != CURRENT_ABSTIME and requires
	     * a tuple lookup.
	     */
	    if (tup->t_tmax != CURRENT_ABSTIME) {
		/*
		 * Restart a scan, keyed on tup's oid, to get the tid of
		 * the tuple.
		 */
		ScanKeyEntryInitialize(&skey[0],
	            0x0, ObjectIdAttributeNumber, ObjectIdEqualRegProcedure,
                    ObjectIdGetDatum(tup->t_oid));
		index_rescan(xFragOidHashIndexscan, false, (ScanKey)skey);
		do {
		    /* 
		     * Loop until we find a tuple that hasn't
		     * been deleted.
		     */
		    res = index_getnext(xFragOidHashIndexscan,
		        ForwardScanDirection);
		    if (res != NULL) {
			iptr = &res->heap_iptr;
			targetTup = heap_fetch(xFragDesc, NowTimeQual,
			    iptr, &buffer);
		    } else {
			break;
		    }
		} while (targetTup == NULL);
	    }

	    if (tup->t_tmax != CURRENT_ABSTIME) {
		/* delete */
		elog(NOTICE, "Cp_PutUpdate: deleting %d", tup->t_oid);
		if (res == NULL) {
		    /*
		     * This tuple was inserted *and* deleted after the previous
		     * update stream got sent and therefore cannot be found
		     * at our site. We need the complete history, so we insert
		     * that tuple. This doesn't affect our fragment statistics.
		     * Also, we don't need to update the indices for this tuple.
		     */
		    updtup_heap_insert(xFragDesc, tup);
		} else {
		    updtup_heap_delete(xFragDesc, iptr, tup->t_tmax);
		    transferNumTuplesPut--;

		    /* 
		     * If a class was dropped, we should also invalidate
		     * the relcache entry for that class.
		     */
		    if (recvfrastoreid == RelOid_pg_nsvcclass) {
		        RelationIdInvalidateRelationCacheByRelationId(
			    tup->t_oid);
		    }
		}
	    } else {
		/* insert */
		elog(NOTICE, "Cp_PutUpdate: inserting %d", tup->t_oid);
		updtup_heap_insert(xFragDesc, tup);

		/* update indices */
		switch (recvfrastoreid) {
		case RelOid_pg_nsvcattr:
		    indexNames = Name_pg_nsvcattr_indices;
		    break;
		case RelOid_pg_nsvcclass:
		    indexNames = Name_pg_nsvcclass_indices;
		    break;
		case RelOid_pg_nsvcfrag:
		    indexNames = Name_pg_nsvcfrag_indices;
		    break;
		default:
		    Assert(0); /* can't be */
		}

		/* we were assuming this when declaring iDescs: */
		Assert(Num_pg_nsvcattr_indices == Num_pg_nsvcclass_indices);
		Assert(Num_pg_nsvcfrag_indices == Num_pg_nsvcclass_indices);

		/* 
		 * We're not using xFragOidHashIndexDesc directly (instead
		 * of iDescs) in case someday someone does the right thing
		 * and adds some more indices for the nameserver catalogs.
		 */
                CatalogOpenIndices(Num_pg_nsvcattr_indices, indexNames, iDescs);
		CatalogIndexInsert(iDescs, Num_pg_nsvcattr_indices, xFragDesc,
		    tup);
		CatalogCloseIndices(Num_pg_nsvcattr_indices, iDescs);
		    
		transferNumTuplesPut++;
	    }
	}
	else {
	    /*
	     * Processing an update stream for a regular relation.
	     * For tuples to be deleted or replaced, gotta lookup the oid.
	     */
	    if (OidLookupNecessary(&databuf->transferInfo, tup)) {
		/*
		 * Restart a scan, keyed on tup's oid, to get the tid of the
		 * tuple.
		 */
		ScanKeyEntryInitialize(&skey[0],
				       0x0,
				       ObjectIdAttributeNumber,
				       ObjectIdEqualRegProcedure,
				       ObjectIdGetDatum(tup->t_oid));

		index_rescan(xFragOidHashIndexscan, false, (ScanKey)skey);
		do {
		    /* Loop until we find a tuple that hasn't been deleted. */
		    res = index_getnext(xFragOidHashIndexscan,
			ForwardScanDirection);
		    if (res != NULL) {
			iptr = &res->heap_iptr;
			targetTup = heap_fetch(xFragDesc, NowTimeQual,
			    iptr, &buffer);
		    } else {
			elog(NOTICE,
			    "Cp_PutUpdate: Bad delete/replace tupleoid %d.",
			     tup->t_oid);
			heap_close(xFragDesc);
			index_endscan(xFragOidHashIndexscan);
			index_close(xFragOidHashIndexDesc);
		
			return EXECUTION_ERROR;
		    }
		} while (targetTup == NULL);
	    }

	    if (UpdateTupleIsInsert(tup)) {
		updtup_heap_insert(xFragDesc, tup);
		transferNumTuplesPut++;
		CatalogIndexInsert(&xFragOidHashIndexDesc, 1, xFragDesc, tup);
	    }
	    else if (UpdateTupleIsReplace(tup)) {
		updtup_heap_replace(xFragDesc, iptr, tup->t_tmin, tup);
		CatalogIndexInsert(&xFragOidHashIndexDesc, 1, xFragDesc, tup);
	    }
	    else if (UpdateTupleIsDelete(tup)) {
		updtup_heap_delete(xFragDesc, iptr, tup->t_tmin);
		transferNumTuplesPut--;
	    }
	}
	data += tup->t_len;
    }

    if (databuf->transferStatus == FINAL_PACKET ||
	databuf->transferStatus == ONLY_PACKET)
    {
	heap_close(xFragDesc);
	index_endscan(xFragOidHashIndexscan);
	index_close(xFragOidHashIndexDesc);

	/*
	 * Update fragment statistics. (@@@ We don't bother to update number
	 * of pages.)
	 */
	if (transferNumTuplesPut != 0) {
	    sprintf(query,
		    "UPDATE pg_fragment SET fratuples = fratuples + %d WHERE\
     frastoreid='%d'::oid",
		    transferNumTuplesPut, recvfrastoreid);
	    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	}

	/*
	 * Update last commit time.
	 */
	if (!ContractIsNameService(&databuf->transferInfo)) {
	    sprintf(query,
	        "UPDATE pg_contract SET lastupdcommit = %d WHERE contractid =\
                '%d'::oid AND recvfrastoreid = '%d'::oid;",
		GetCurrentTransactionStartTime(), contractid, recvfrastoreid);
	    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	} else if (databuf->nsvcTransferFinished) {
	    /* 
	     * Update the nameservice contract. Nameservice catalogs 
	     * have read-only copies and only a single copy contract,
	     * which is why we don't have to specify the recvfrastoreid
	     * in the update.
	     * We only update the commit time of the contract after the last
	     * of the nameservice catalogs has been updated.
	     */
	    sprintf(query,
	        "UPDATE pg_contract SET lastupdcommit = %d WHERE contractid =\
                '%d'::oid;",
		GetCurrentTransactionStartTime(), contractid);
	    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	}

    }

    elog(NOTICE, "Leaving Cp_PutUpdate.");

    return SUCCESS;
}


/* ----------------------------------------------------------------
 *	OidLookupNecessary
 *
 *	Is so if we have a tuple under a name service contract
 *	or if we have a user copy update tuple that's a replace or delete.
 *
 * ----------------------------------------------------------------
 */
static bool
OidLookupNecessary(UpdateContractInfo *contract, HeapTuple htup)
{
    Assert(!ContractIsNameService(contract));
    return (UpdateTupleIsReplace(htup) || UpdateTupleIsDelete(htup));
}


/*****************************************************************************
 * Routines that handle initializing update streams.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_CP_INITIALIZEUPDATESTREAMS
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_CP_INITIALIZEUPDATESTREAMS(void *args)
{
    char		**argv = (char **)args;
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result;

    elog(NOTICE, "handleCall_CP_INITIALIZEUPDATESTREAMS.");

    mainThreadExecInXact((thdproc_t)Cp_InitializeUpdateStreams, NULL, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall: can't send reply\n");
    }

    free(args);
}

/* ----------------------------------------------------------------
 *    	Cp_InitializeUpdateStreams
 *
 *	This is called by CopyUpdateSetup which is called by sitemgrMain
 *	when starting up the sitemgr at a site. It scans pg_contract, and
 *	schedules an update stream into the update stream queue for each
 *	contract.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
Cp_InitializeUpdateStreams()
{
    Relation		contractdesc;
    HeapScanDesc	contractscan;
    HeapTuple		htup;
    Form_pg_contract	contract;
    NSVCProbeArg	nsvc_arg;
    NSVCProbeResult	*nsvc_result;
    UpdateContractInfo	updatecontract;
    RpcConn		*sm_rpcc;
    RPCErrorCode	result;
    RPCErrorCode	returnresult = SUCCESS;
    unsigned int	ourhostid = getMyHostId();

    /*
     * Call local sitemgr to schedule update stream.
     */
    sm_rpcc = Rpc_connectSitemgr(getMyHostId(), 0,
				 "Cp_InitializeUpdateStreams",
				 RPC_ONERR_WARN);

    contractdesc = heap_openr((Name)Name_pg_contract);
    if (!RelationIsValid(contractdesc)) {
	elog(NOTICE, "Cp_InitializeUpdateStreams: cannot open pg_contract.");
	return EXECUTION_ERROR;
    }

    contractscan = heap_beginscan(contractdesc,
				  0,
				  SelfTimeQual,
				  0, NULL);
    do {
	htup = heap_getnext(contractscan, 0, NULL);

	if (HeapTupleIsValid(htup)) {
	    contract = (Form_pg_contract)GETSTRUCT(htup);

	    PrintContract("Cp_InitializeUpdateStreams", contract);

	    /*
	     * We only consider it if it's a send contract.
	     */
	    if (ContractIsReceive(contract))
		continue;

	    /* @@@ This little bit here should be resolved. */
	    /*
	     * Only initialize update stream if send fragment's in our
	     * database! We can check pg_fragment for it.
	     */
	    nsvc_arg.action = NSVC_FINDLOCALFRAG;
	    nsvc_arg.fspec = MASTER;
	    nsvc_arg.oid = contract->sendfrastoreid;
	    nsvc_result = NsvcLocalProbe(&nsvc_arg);

	    if (nsvc_result->found == true ||
		ContractIsNameService(contract))
	    {
		CopyUpdateContractInfo((UpdateContractInfo *)contract,
				       &updatecontract);

		/*
		 * Call the sitemgr to schedule this contract.
		 */
		if (!Rpc_call(sm_rpcc, RPC_CP_SCHEDULESTREAM, &updatecontract,
			      &result))
		{
		    elog(NOTICE, "Cp_InitializeUpdateStreams: CP_SCHEDULESTREAM rpc call failed.");
		    returnresult = EXECUTION_ERROR;
		}

		if (result == EXECUTION_ERROR) {
		    elog(NOTICE, "Cp_InitializeUpdateStreams: CP_SCHEDULESTREAM couldn't schedule contract %d.",
			 updatecontract.contractid);
		    returnresult = EXECUTION_ERROR;
		}
	    }

	    pfree(nsvc_result);
	}
    } while (htup);

    heap_endscan(contractscan);
    heap_close(contractdesc);

    Rpc_disconnect(sm_rpcc);

    return returnresult;
}


/*****************************************************************************
 * Routines that handle cancelling update stream contracts
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	CancelContracts
 *
 *	Cancels all contracts for a relation, if relid specified.
 *	If it's a name service contract, cancels the contract at
 *	each site.
 *
 *	@@@ This routine should really check for errors. If it fails,
 *	    there's a problem. This should be an atomic call-- that
 *	    means call sitemgr's back and reschedule the update streams--
 *	    which is some extra work.
 *
 * ----------------------------------------------------------------
 */
void
CancelContracts(bool isnsspecific, bool isnsremote, unsigned int hostid,
		Oid relid)
{
    Relation		contractdesc;
    HeapScanDesc	contractscan;
    HeapTuple		htup;
    Form_pg_contract	contract;
    RpcConn		*sm_rpcc;

    sm_rpcc = Rpc_connectSitemgr(getMyHostId(), 0, "CancelContracts",
				 RPC_ONERR_WARN);

    contractdesc = heap_openr((Name) Name_pg_contract);
    if (!RelationIsValid(contractdesc)) {
	elog(WARN, "CancelContracts: cannot open pg_contract.");
    }

    contractscan = heap_beginscan(contractdesc,
				  0,
				  SelfTimeQual,
				  0, NULL);
    do {
	htup = heap_getnext(contractscan, 0, NULL);

	if (HeapTupleIsValid(htup)) {
	    contract = (Form_pg_contract)GETSTRUCT(htup);

	    if (isnsspecific && ContractIsNameService(contract) &&
		ContractIsNSSpecific(contract) &&
		contract->otherloc == hostid)
	    {
		/*
		 * We have found the specific name service contract to cancel.
		 * Cancel contract for contractid at remote site.
		 */
		if (ContractIsSend(contract)) {
		    UnscheduleContract(sm_rpcc,
				       (UpdateContractInfo *)contract);
		}
		CancelRemoteContract((UpdateContractInfo *) contract);
		heap_delete(contractdesc, &htup->t_ctid);
	    }
	    else if (isnsremote && ContractIsNameService(contract) &&
		     ContractIsNSRemote(contract) &&
		     contract->otherloc == hostid)
	    {
		/*
		 * We have found the remote name service contract to cancel.
		 * Cancel contract for contractid at remote site.
		 */
		if (ContractIsSend(contract)) {
		    UnscheduleContract(sm_rpcc,
				       (UpdateContractInfo *)contract);
		}
		CancelRemoteContract((UpdateContractInfo *) contract);
		heap_delete(contractdesc, &htup->t_ctid);
	    }
	    else if (ContractIsReadWrite(contract) && ContractIsSend(contract))
	    {
		/*
		 * Read write user copy send contract -- we unschedule the
		 * send contract, and cancel the contracts at the remote site
		 * for this contractid.
		 */
		UnscheduleContract(sm_rpcc, (UpdateContractInfo *)contract);
		CancelRemoteContract((UpdateContractInfo *) contract);
		heap_delete(contractdesc, &htup->t_ctid);
	    }
	    else if (ContractIsReadWrite(contract) &&
		     ContractIsReceive(contract))
	    {
		/*
		 * Read write user copy receive contract -- we do nothing
		 * because when we find the send contract we take the
		 * necessary actions. All we do is delete the tuple.
		 */
		heap_delete(contractdesc, &htup->t_ctid);
	    }
	    else if (ContractIsReadOnly(contract) &&
		     (contract->sendfrastoreid == relid ||
		      contract->recvfrastoreid == relid))
	    {
		/*
		 * Read only user copy send or receive contract -- 
		 * read only contracts are only one per site, so we unschedule
		 * if its a send contract, and cancel the contract at the
		 * remote site for this contractid.
		 */
		if (ContractIsSend(contract)) {
		    UnscheduleContract(sm_rpcc,
				       (UpdateContractInfo *)contract);
		}
		CancelRemoteContract((UpdateContractInfo *) contract);
		heap_delete(contractdesc, &htup->t_ctid);
	    }
	}
    } while (htup);

    heap_endscan(contractscan);
    heap_close(contractdesc);

    Rpc_disconnect(sm_rpcc);
}


/* ----------------------------------------------------------------
 *	CancelRemoteContract
 *
 *	Calls up remote site in update contract, tells them to cancel
 *	the contract indicated by contractid.
 *
 * ----------------------------------------------------------------
 */
static void
CancelRemoteContract(UpdateContractInfo *contract)
{
    RpcConn		*sm_rpcc;
    RPCErrorCode	result;

    sm_rpcc = Rpc_connectSitemgr(contract->otherloc, 0,
				 "CancelRemoteContract", RPC_ONERR_WARN);

    if (!Rpc_call(sm_rpcc, RPC_CANCELCONTRACT, &contract->contractid, &result))
    {
	Rpc_disconnect(sm_rpcc);
	elog(NOTICE, "CancelRemoteContract: Couldn't make rpc call.");
    }

    if (result != SUCCESS) {
	Rpc_disconnect(sm_rpcc);
	elog(NOTICE, "CancelRemoteContract: CANCELCONTRACT rpc call failed.");
    }

    Rpc_disconnect(sm_rpcc);
}

/* ----------------------------------------------------------------
 *    	handleCall_CANCELCONTRACT
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_CANCELCONTRACT(void *args)
{
    char		**argv = (char **)args;
    int			*rpcarg = (int *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result;

    elog(NOTICE, "handleCall_CANCELCONTRACT.");

    mainThreadExecInXact((thdproc_t)CancelContract, (void *) *rpcarg, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_CANCELCONTRACT: Can't send reply.\n");
    }

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *	CancelContract
 *
 *	Cancels a specific contract, and unschedules its update streams.
 *
 *	@@@ This routine should really return errors so the remote caller
 *	    can do some exception handling.
 *
 * ----------------------------------------------------------------
 */
static RPCErrorCode
CancelContract(Oid contractid)
{
    Relation		contractdesc;
    HeapScanDesc	contractscan;
    ScanKeyData		skey[1];
    HeapTuple		htup;
    Form_pg_contract	contract;
    RpcConn		*sm_rpcc;

    sm_rpcc = Rpc_connectSitemgr(getMyHostId(), 0, "CancelContract",
				 RPC_ONERR_WARN);

    contractdesc = heap_openr((Name) Name_pg_contract);
    if (!RelationIsValid(contractdesc)) {
	elog(WARN, "CancelContract: cannot open pg_contract.");
    }

    ScanKeyEntryInitialize(&skey[0],
			   0x0,
			   Anum_pg_contract_contractid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(contractid));

    contractscan = heap_beginscan(contractdesc,
				  0,
				  SelfTimeQual,
				  1, (ScanKey)skey);
    do {
	htup = heap_getnext(contractscan, 0, NULL);

	if (HeapTupleIsValid(htup)) {
	    contract = (Form_pg_contract)GETSTRUCT(htup);

	    /*
	     * First, unschedule update stream if necessary.
	     */
	    if (ContractIsSend(contract)) {
		UnscheduleContract(sm_rpcc, (UpdateContractInfo *)contract);
	    }

	    /*
	     * Then delete the sucker.
	     */
	    heap_delete(contractdesc, &htup->t_ctid);
	}
    } while (htup);

    heap_endscan(contractscan);
    heap_close(contractdesc);

    Rpc_disconnect(sm_rpcc);

    return SUCCESS;
}


/* ----------------------------------------------------------------
 *	UnscheduleContract
 *
 *	Makes the call to sitemgr to unschedule an update contract.
 *
 * ----------------------------------------------------------------
 */
static void
UnscheduleContract(RpcConn *sm_rpcc, UpdateContractInfo *contract)
{
    RPCErrorCode	result;

    if (!Rpc_call(sm_rpcc, RPC_CP_UNSCHEDULESTREAM, &contract->contractid,
		  &result))
    {
	elog(NOTICE, "UnscheduleContract: Couldn't make rpc call.");
    }

    if (result != SUCCESS) {
	elog(NOTICE, "UnscheduleContract: UNSCHEDULESTREAM rpc call failed.");
    }
}


/*****************************************************************************
 * Utility Routines
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	CopyUpdateContractInfo	
 *
 *	Just a routine to copy a UpdateContractInfo struct.
 *
 * ----------------------------------------------------------------
 */
static void
CopyUpdateContractInfo(UpdateContractInfo *source, UpdateContractInfo *dest)
{
    dest->contractid = source->contractid;
    dest->sendfrastoreid = source->sendfrastoreid;
    dest->recvfrastoreid = source->recvfrastoreid;
    dest->sidefrastoreid = source->sidefrastoreid;
    dest->lastupdcommit = source->lastupdcommit;
    dest->otherloc = source->otherloc;
    dest->period = source->period;
    dest->typemask = source->typemask;
}
