/*-------------------------------------------------------------------------
 *
 * datafrag.c--
 *    Data fragment-related calls to the exec./backend.
 *
 *    Includes also some copies-related calls to the exec./backend,
 *    that need to use the fragment-move locks.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/datafrag.c,v 1.28 1996/03/05 23:47:36 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdarg.h>

#include <rpc/types.h>
#include <pthread.h>
#include "utils/elog.h"
#include "mariposa_rpc.h"			/* for procedure constants */
#include "rpc/rpcutils.h"
#include "rpc/call_exec.h"			/* backend calling facility */
#include "rpc/rexec.h"
#include "sitemgr/rpcerror.h"
#include "datafrag.h"

#include "access/heapam.h"
#include "access/xact.h"
#include "access/miscam.h"
#include "catalog/catname.h"
#include "catalog/pg_class.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"
#include "catalog/indexing.h"
#include "rpc/fragutils.h"
#include "storage/itemptr.h"
#include "utils/syscache.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "utils/tqual.h"

#define SENDBUFSIZE 7168		/* copied from exchange.c */
	
/* ----------------
 *  Procedures that implement the RPC calls handled by the fragment transfer
 *  handleCall routines in this file.
 * ----------------
 */

static RPCErrorCode PrepareFragForXToCaller(FragXArg *fragxarg);
static RPCErrorCode AddFragOidHashIndex(Relation rdesc);
static RPCErrorCode ExtractPreparations(FragXArg *fragxarg);

static RPCErrorCode FragmentPutCatalog(FragCatalogInfo *catInfo);
static RPCErrorCode FragmentPutData(FragDataBuffer *databuf);
static void FragmentGetCatalog(FragCatalogInfo *catInfo);
static void FragmentGetData(FragDataBuffer *databuf);

/* ----------------
 *  File-scope declarations shared across catalog and data transfer routines.
 *  This is possible because of twopc mutexes.
 *
 *  Proper RPC way is to return these to sender as opaque type. The sender
 *  would then pass this back to us when calling FragmentPutData. That means
 *  writing xdr routines for all those types. Ugh.
 * ----------------
 */

static Relation xFragDesc;
static Relation xFragArchDesc;
static HeapScanDesc xFragscan;
static HeapScanDesc xFragArchscan;
static TimeQual xFetchArchTimeQual;
/*
 * For FragmentGetData, indicates which part of the fragment we are getting.
 */
static FragRelType whichscan;

/*
 * moveNumTuplesPut records how many tuples actually make it over on the
 * fragment move, so we can store the result in the pg_fragment entry after
 * FragmentPutData is called the last time.
 */
static int moveNumTuplesPut = 0;


/*****************************************************************************
 * Rpc handleCall routines for various fragment transfers.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_FRAGXSETUP
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 *	This handles the for buying, selling, and copying fragments.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_FRAGXSETUP(void *args)
{
    char		**argv = (char **)args;
    FragXArg		*rpcarg = (FragXArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result = SUCCESS;
    
    elog(NOTICE, "Entering handleCall_FRAGXSETUP.");

    /*
     * Reset file-scope transfer variables.
     */
    xFragDesc = NULL;
    xFragArchDesc = NULL;
    xFragscan = NULL;
    xFragArchscan = NULL;

    /*
     * Depending on type of transfer operation, make necessary preparations.
     */
    switch (rpcarg->fragxtype) {
    case SELLFRAGMENT:
	mainThreadExecInXact((thdproc_t) PrepareFragForXToCaller, rpcarg,
			     &result);
	break;
    case COPYFRAGMENT:
	mainThreadExecInXact((thdproc_t) PrepareFragForXToCaller, rpcarg,
			     &result);
	break;
    case BUYFRAGMENT:
	moveNumTuplesPut = 0;
	break;
    }

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_FRAGXSETUP: Can't send reply.\n");
    }

    free(rpcarg);
    free(args);
    
    elog(NOTICE, "Leaving handleCall_FRAGXSETUP.");
}

/* ----------------------------------------------------------------
 *    	handleCall_EXTRACTPREPARATIONS
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_EXTRACTPREPARATIONS(void *args)
{
    char		**argv = (char **)args;
    FragXArg		*rpcarg = (FragXArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result = SUCCESS;
    
    elog(NOTICE, "Entering handleCall_EXTRACTPREPARATIONS.");

    mainThreadExec((thdproc_t)ExtractPreparations, rpcarg, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_EXTRACTPREPARATIONS: Can't send reply.\n");
    }

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_FRAGPUTCAT
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_FRAGPUTCAT(void *args)
{
    char		**argv = (char **)args;
    FragXArg		*rpcarg = (FragXArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result;

    elog(NOTICE, "Entering handleCall_FRAGPUTCAT.");

    mainThreadExec((thdproc_t)FragmentPutCatalog, rpcarg, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_FRAGPUTCAT: Can't send reply.\n");
    }

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_FRAGPUTDATA
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_FRAGPUTDATA(void *args)
{
    char		**argv = (char **)args;
    FragDataBuffer	*rpcarg = (FragDataBuffer *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result;

    elog(NOTICE, "Entering handleCall_FRAGPUTDATA.");

    mainThreadExec((thdproc_t)FragmentPutData, rpcarg, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_FRAGPUTDATA: Can't send reply.\n");
    }

    if (rpcarg->data.Data_len > 0)
	free(rpcarg->data.Data_val);

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_FRAGGETCAT
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_FRAGGETCAT(void *args)
{
    char		**argv = (char **)args;
    FragXArg		*rpcarg = (FragXArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    FragCatalogInfo	result;

    elog(NOTICE, "Entering handleCall_FRAGGETCAT.");

    memset(&result, 0, sizeof(FragCatalogInfo));
    CopyFragXArg(rpcarg, &result.transferInfo);

    mainThreadExec((thdproc_t)FragmentGetCatalog, &result, NULL);

    if (!svc_sendreply(transp, xdr_FragCatalogInfo, &result)) {
  	errorExit("handleCall_FRAGGETCAT: Can't send reply.\n");
    }

    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_FRAGGETDATA
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_FRAGGETDATA(void *args)
{
    char		**argv = (char **)args;
    FragXArg		*rpcarg = (FragXArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    FragDataBuffer	result;
    char		buffer[SENDBUFSIZE];

    elog(NOTICE, "Entering handleCall_FRAGGETDATA.");

    CopyFragXArg(rpcarg, &result.transferInfo);
    result.data.Data_val = buffer;

    mainThreadExec((thdproc_t)FragmentGetData, &result, NULL);

    if (!svc_sendreply(transp, xdr_FragDataBuffer, &result)) {
  	errorExit("handleCall_FRAGGETDATA: Can't send reply.\n");
    }

    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_SCHEDFRAGMENT
 *
 *	Thread start routine;
 *	does synchronous call to backend to process query.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_SCHEDFRAGMENT(void *args)
{
    char		**argv = (char **)args;
    SchedFragmentArg	*rpcarg = (SchedFragmentArg *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    int			pid;
    int			result = 1;

    /*
     * Don't keep site mgr waiting.
     */
    if (!svc_sendreply(transp, xdr_void, NULL)) {
  	errorExit("handleCall_SCHEDFRAGMENT: Can't send reply.\n");
    }

#ifdef HP_JUKEBOX
    if (arg->type == 'C') {
	mainThreadExecInXact((thdproc_t)CacheRelation,
			     (void *)arg->frastoreid, &result);
    } else {
	mainThreadExecInXact((thdproc_t)UnCacheRelation,
			     (void *)arg->frastoreid, &result);
    }
#endif 

    /*
     * Now tell the sitemgr we're done.
     */
    pid = getpid();
    sitemgrBackendDone(pid);

    free(rpcarg);
    free(args);
}


/*****************************************************************************
 * Procedures that implement the RPC calls handled by the fragment transfer
 * handleCall routines in this file.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	PrepareFragForXToCaller
 *
 *	For SELLFRAGMENT and COPYFRAGMENT calls, another site wants our
 *	fragment, so we prepare it to be transferred.
 *
 *	1. Vacuum fragment.
 *	2. Set a write lock on the fragment.
 *
 *	Since we don't allow copied fragments to be "tooken" yet, returns
 *	error if this is SELLFRAGMENT call and relation is copied.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
PrepareFragForXToCaller(FragXArg *fragxarg)
{
    Relation		rdesc;
    NameData		relNameData;

    elog(NOTICE, "Entering PrepareFragForXToCaller.");

    /*
     * If fragxarg->frastoreid is 0, then we are not meant to do anything.
     * (The case of a fragment copy that does a name service contract.)
     */
    if (fragxarg->frastoreid == 0) {
	return SUCCESS;
    }

    /*
     * Vacuum the fragment.
     */
    rdesc = heap_open(fragxarg->frastoreid);
    if (!RelationIsValid(rdesc)) {
	elog(NOTICE, "PrepareFragForXToCaller: Can't open fragment %d",
	     fragxarg->frastoreid);
	return EXECUTION_ERROR;
    }

    if (rdesc->rd_iscopied && fragxarg->fragxtype == SELLFRAGMENT) {
	heap_close(rdesc);
	elog(NOTICE, "PrepareFragForXToCaller: We're not allowed to take copied fragments yet!");
	return EXECUTION_ERROR;
    }

    vacuum(rdesc->rd_rel->relname.data);

    heap_close(rdesc);

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	AddFragOidHashIndex
 *
 *	For COPYFRAGMENT calls. If we don't have a hash index on oid on the
 *	desired fragment already, we add one.
 *
 *	When we copy an fragment, we force both the master and the copy to
 *	have hash indices on classoid. That way, when we want to pass delete
 *	updates we don't have to do a sequential scan to find the tuples at
 *	the receiving site.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
AddFragOidHashIndex(Relation rdesc)
{
    elog(NOTICE, "Entering AddFragOidHashIndex.");

    /*
     * If the rd_iscopied flag has not been set, then this relation has not
     * been copied yet, so it can't yet have the index, so we define it.
     */
    if (!rdesc->rd_iscopied) {
	/*
	 * This is to check if an old oid hash index is hangin' around...
	 * If so, keep the one we got. What should happen is after a drop
	 * copy it gets automatically deleted.
	 */
	Relation ohidesc;

	ohidesc = OpenRelationOidHashIndex(rdesc->rd_id);
	if (!RelationIsValid(ohidesc)) {
	    /*
	     * There isn't one already, so we define one.
	     */
	    DefineRelationOidHashIndex(rdesc);
	}
	else {
	    index_close(ohidesc);
	}
    }

    elog(NOTICE, "Leaving AddFragOidHashIndex.");

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	ExtractPreparations
 *
 *	Does a few things necessary for taking or copying fragments
 *	that can't be done in FRAGXSETUP (need to be done within the
 *	transaction.) -- It opens the fragment, and sets a write lock
 *	on it. If it's going to be copied we also construct a hash
 *	index on oid for the fragment.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
ExtractPreparations(FragXArg *fragxarg)
{
    elog(NOTICE, "Entering ExtractPreparations.");

    xFragDesc = heap_open(fragxarg->frastoreid);
    if (!RelationIsValid(xFragDesc)) {
	elog(NOTICE, "ExtractPreparations: Can't open fragment %d.",
	     fragxarg->frastoreid);
	return EXECUTION_ERROR;
    }

    /*
     * Set write lock on the fragment, so no one else messes with it while
     * we transfer it.
     */
    RelationSetLockForWrite(xFragDesc);

    elog(NOTICE, "Leaving ExtractPreparations.");

    return AddFragOidHashIndex(xFragDesc);
}

/* ----------------------------------------------------------------
 *    	FragmentPutCatalog
 *
 *	When a remote site moves a fragment here, this is called to
 *	transfer the catalog info necessary to establish the existence
 *	of the fragment here.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
FragmentPutCatalog(FragCatalogInfo *catInfo)
{
    HeapTuple		classTup, fragTup, pgfragTup;
    Form_pg_fragment	pgfrag;
    int			natts, i;
    AttributeTupleForm	*tupdesc;
    MemoryContext	oldcxt;
    Relation		pgfragdesc;
    char		*fralokey;
    char		*frahikey;
    bool		loisnull;
    bool		hiisnull;
    Oid			loout;
    Oid			hiout;
    char		query[1000];

    elog(NOTICE, "Entering FragmentPutCatalog.");

    classTup = (HeapTuple)catInfo->classTuple.Tuple_val;
    fragTup = (HeapTuple)catInfo->fragTuple.Tuple_val;
    pgfragTup = (HeapTuple)catInfo->pgfragTuple.Tuple_val;

    /*
     * Prepare a tuple descriptor; the palloc for tuple descriptor need to
     * persist at the end of transaction (Relation descriptor is cached.)
     */
    if (!CacheCxt)
	CacheCxt = CreateGlobalMemory("Cache");
    oldcxt = MemoryContextSwitchTo((MemoryContext)CacheCxt);

    natts = catInfo->attrTuples.TupleArray_len;
    tupdesc = (TupleDesc)palloc(sizeof(AttributeTupleForm)*natts);
    for(i=0; i < natts; i++) {
	AttributeTupleForm tup;
	int len;

	len = catInfo->attrTuples.TupleArray_val[i].Tuple_len;
	tup = (AttributeTupleForm) palloc(len);
	memmove(tup,
		catInfo->attrTuples.TupleArray_val[i].Tuple_val,
		len);
	tupdesc[i] = tup;
    }

    MemoryContextSwitchTo(oldcxt);

    xFragDesc = MakeFragmentRelationDesc(classTup, fragTup, natts, tupdesc);

    /*
     * Create archive relation for fragment, since all fragments are archived
     * in Mariposa.
     */
    CreateArchiveRelation(xFragDesc);
    CommandCounterIncrement();
    xFragArchDesc = OpenArchiveRelation(xFragDesc->rd_id);

    if (!RelationIsValid(xFragArchDesc)) {
	elog(NOTICE, "FragmentPutCatalog: Can't open newly created archive relation for fragment %d.",
	     xFragDesc->rd_id);
	return EXECUTION_ERROR;
    }

    pgfrag = (Form_pg_fragment)GETSTRUCT(pgfragTup);

    sprintf(query,
	    "DELETE FROM pg_fragment WHERE frastoreid='%d'::oid",
	    pgfrag->frastoreid);
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    /*
     * Open pg_fragment for its tupdesc, so we can heap_getattr from
     * pgfragTup fralokey and frahikey.
     */
    pgfragdesc = heap_openr((Name) Name_pg_fragment);
    if (!RelationIsValid(pgfragdesc)) {
	elog(WARN, "FragmentPutCatalog: Can't open pg_fragment.");
	return EXECUTION_ERROR;
    }
    heap_close(pgfragdesc);

    fralokey = heap_getattr(pgfragTup, InvalidBuffer,
			    Anum_pg_fragment_fralokey, pgfragdesc->rd_att,
			    &loisnull);
    loout =
	typtoout((Oid)
		 pgfragdesc->rd_att[Anum_pg_fragment_fralokey - 1]->atttypid);

    frahikey = heap_getattr(pgfragTup, InvalidBuffer,
			    Anum_pg_fragment_frahikey, pgfragdesc->rd_att,
			    &hiisnull);
    hiout =
	typtoout((Oid)
		 pgfragdesc->rd_att[Anum_pg_fragment_frahikey - 1]->atttypid);
	
    sprintf(query,
	    "INSERT INTO pg_fragment\
 (frapages,fratuples,fraclassid,frastoreid,fralogicalid,fralocation%s%s)\
 values (%d,%d,%d,%d,%d,%d%s%s%s%s%s%s);",
	    /* On the fragment keys, check first char of string for NULL */
	    !loisnull ? ", fralokey" : "",
	    !hiisnull ? ", frahikey" : "",
	    0, 0, /* To be filled in after all tuples put */
	    pgfrag->fraclassid, pgfrag->frastoreid, pgfrag->fralogicalid,
	    getMyHostId(),
	    !loisnull ? ",'" : "",
	    !loisnull ? fmgr(loout, fralokey, gettypelem(pgfragdesc->rd_att[Anum_pg_fragment_fralokey - 1]->atttypid)) : "",
	    !loisnull ? "'::text" : "",
	    !hiisnull ? ",'" : "",
	    !hiisnull ? fmgr(hiout, frahikey, gettypelem(pgfragdesc->rd_att[Anum_pg_fragment_frahikey - 1]->atttypid)) : "",
	    !hiisnull ? "'::text" : "");
    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

    elog(NOTICE, "Leaving FragmentPutCatalog.");

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	FragmentPutData
 *
 *	When a remote site moves a fragment here, this is called to
 *	transfer the actual tuples over into the fragment set up here
 *	by FragmentPutCatalog. Relies on FragmentPutCatalog to set up
 *	xFragDesc, xFragArchDesc and reset moveNumTuplesPut.
 *
 * ----------------------------------------------------------------
 */
RPCErrorCode
FragmentPutData(FragDataBuffer *databuf)
{
    int			i;
    int			status;
    HeapTuple		tup;
    char		*data;
    char		query[1000];
    FragXArg		*fragxarg = &databuf->transferInfo;

    elog(NOTICE, "Entering FragmentPutData (%d tuples.)", databuf->numTuples);

    if (databuf->numTuples == 0) {
	heap_close(xFragArchDesc);
	heap_close(xFragDesc);

	/*
	 * Update fragment statistics. (@@@ Number of pages is clearly wrong,
	 * but we better have an entry because the vacuum cleaner expects the
	 * attribute to be there.)
	 */
	sprintf(query,
		"UPDATE pg_fragment SET frapages = %d, fratuples = %d WHERE\
 frastoreid='%d'::oid",
		0/*@@@*/, moveNumTuplesPut, fragxarg->frastoreid);
        pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	
	return SUCCESS;
    }
    
    data = databuf->data.Data_val;
    
    /*
     * If archive part, we insert into archive. If current part, we insert
     * into new fragment, and keep track of the number of tuples.
     */
    if (databuf->which == ARCHIVE_PART) {
	for(i = 0; i < databuf->numTuples; i++) {
	    tup = (HeapTuple)data;

	    /*
	     * We use the copy update stream access method here, because we
	     * want to preserve the tuples' time info and origin.
	     */
	    updtup_heap_insert(xFragArchDesc, tup);
	    data += tup->t_len;
	}
    }
    else {
	for(i = 0; i < databuf->numTuples; i++) {
	    tup = (HeapTuple)data;

	    /*
	     * We use the copy update stream access method here, because we
	     * want to preserve the tuples' time info and origin.
	     */
	    updtup_heap_insert(xFragDesc, tup);
	    data += tup->t_len;
	}

	moveNumTuplesPut += databuf->numTuples;
    }

    elog(NOTICE, "Leaving FragmentPutData.");

    return SUCCESS;
}

/* ----------------------------------------------------------------
 *    	FragmentGetCatalog
 *
 *	When a remote site takes/copies a fragment from us, this is called
 *	so we can fetch the necessary system catalog information and send
 *	it back to the remote site. Also initializes file-scope variables
 *	for FragmentGetData.
 *
 * ----------------------------------------------------------------
 */
void
FragmentGetCatalog(FragCatalogInfo *catInfo)
{
    Relation		rd;
    Relation		rdesc;
    HeapScanDesc	sdesc;
    HeapTuple		tup;
    ScanKeyData		skey;
    bool		nullAttr;
    FragPartInfo	ofpi;
    int			natts;
    int			i;
    Tuple		*tupdesc;
    FormData_pg_fragment pgFragTuple;
    FragXArg		*fragxarg = &catInfo->transferInfo;

    elog(NOTICE, "Entering FragmentGetCatalog.");

    /*
     * Handle pg_class (2 tuples: class and fragment.)
     */
    
    tup = SearchSysCacheTuple(RELOID,
			      (char *) ObjectIdGetDatum(fragxarg->fraclassid),
			      (char *) NULL,
			      (char *) NULL,
			      (char *) NULL);
    if (!HeapTupleIsValid(tup)) {
	catInfo->error = EXECUTION_ERROR;
	elog(NOTICE,
	     "FragmentGetCatalog: Can't find class tuple in pg_class.");
	return;
    }
    catInfo->classTuple.Tuple_len = tup->t_len;
    catInfo->classTuple.Tuple_val = (char *)malloc(tup->t_len);
    memmove(catInfo->classTuple.Tuple_val, tup, tup->t_len);

    tup = SearchSysCacheTuple(RELOID,
			      (char *) ObjectIdGetDatum(fragxarg->frastoreid),
			      (char *) NULL,
			      (char *) NULL,
			      (char *) NULL);
    if (!HeapTupleIsValid(tup)) {
	catInfo->error = EXECUTION_ERROR;
	elog(NOTICE,
	     "FragmentGetCatalog: Can't find fragment tuple in pg_class.");
	return;
    }
    catInfo->fragTuple.Tuple_len = tup->t_len;
    catInfo->fragTuple.Tuple_val = (char *)malloc(tup->t_len);
    memmove(catInfo->fragTuple.Tuple_val, tup, tup->t_len);

    /*
     * Handle pg_fragment.
     */
    ScanKeyEntryInitialize(&skey, 0x0, Anum_pg_fragment_frastoreid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(fragxarg->frastoreid));
    rd = heap_openr((Name)Name_pg_fragment);
    sdesc = heap_beginscan(rd, false, NowTimeQual, 1, &skey);

    if (!HeapTupleIsValid(tup = heap_getnext(sdesc, 0, NULL))) {
	heap_endscan(sdesc);
	heap_close(rd);

	catInfo->error = EXECUTION_ERROR;
	elog(NOTICE, "FragmentGetCatalog: Can't find pg_fragment tuple.");
	return;
    }
    catInfo->pgfragTuple.Tuple_len = tup->t_len;
    catInfo->pgfragTuple.Tuple_val = (char *)malloc(tup->t_len);
    memmove(catInfo->pgfragTuple.Tuple_val, tup, tup->t_len);

    heap_endscan(sdesc);
    heap_close(rd);

    /*
     * Get fragkey info from fragment reldesc. This relation descriptor also
     * used to handle pg_attributes.
     */

    /*
     * Handle pg_attributes.
     */
    natts = xFragDesc->rd_rel->relnatts;
    tupdesc = (Tuple *)malloc(sizeof(Tuple) * natts);
    catInfo->attrTuples.TupleArray_len = natts;
    catInfo->attrTuples.TupleArray_val = tupdesc;
    for (i = 0; i < natts; i++) {
	tupdesc[i].Tuple_len = ATTRIBUTE_TUPLE_SIZE;
	tupdesc[i].Tuple_val = (char*)xFragDesc->rd_att[i];
    }

    /*
     * Need to do some preparation here. Open archive relation, and set
     * scan descriptors, and set whichscan = ARCHIVE_PART.
     */
    xFragArchDesc = OpenArchiveRelation(fragxarg->frastoreid);
    if (!RelationIsValid(xFragArchDesc)) {
	catInfo->error = EXECUTION_ERROR;
	elog(NOTICE, "FragmentGetCatalog: Can't open fragment %d archive.",
	     fragxarg->frastoreid);
	return;
    }

    xFragscan = heap_beginscan(xFragDesc,
			       0,
			       SelfTimeQual,
			       0, NULL);

    xFetchArchTimeQual = TimeFormRangedTimeQual(EPOCH_ABSTIME,
						CURRENT_ABSTIME);
    xFragArchscan = heap_beginscan(xFragArchDesc,
				   0,
				   xFetchArchTimeQual,
				   0, NULL);
    whichscan = ARCHIVE_PART;

    catInfo->error = SUCCESS;

    elog(NOTICE, "Leaving FragmentGetCatalog.");
}

/* ----------------------------------------------------------------
 *    	FragmentGetData
 *
 *	When a remote site takes/copies a fragment from us, this is called
 *	to send the data in the fragment to the remote site.
 *	Depends on xFragDesc and xFragscan being set up by
 *	FragmentGetCatalog.
 *
 * ----------------------------------------------------------------
 */
void
FragmentGetData(FragDataBuffer *databuf)
{
    Relation		fragdesc;
    HeapScanDesc	fragscan;
    HeapTuple		htup;
    int			nbytes = 0;
    int			numTuples = 0;
    char		*buffer;
    static HeapTuple	lastTuple = NULL;	/* XXX this is gross */
    int			status;
    char		query[1000];
    FragXArg		*fragxarg = &databuf->transferInfo;

    elog(NOTICE, "Entering FragmentGetData.");

    buffer = databuf->data.Data_val;
    databuf->which = whichscan;
    fragscan = (whichscan == ARCHIVE_PART) ? xFragArchscan : xFragscan;

    htup = (lastTuple) ? lastTuple : heap_getnext(fragscan, 0, NULL);

    if (!HeapTupleIsValid(htup) && whichscan == ARCHIVE_PART) {
	/*
	 * Done with archive, switch to current part, scan first tuple.
	 */
	whichscan = CURRENT_PART;
	databuf->which = CURRENT_PART;

	heap_endscan(fragscan);
	pfree(xFetchArchTimeQual);
	heap_close(xFragArchDesc);

	fragscan = xFragscan;
	htup = heap_getnext(fragscan, 0, NULL);
    }

    if (!HeapTupleIsValid(htup)) {
	/*
	 * When we don't have a left-over tuple and we've completed the
	 * scan, signals end of transfer by returning with an empty buffer.
	 */
	databuf->numTuples = 0;
	databuf->data.Data_len = 0;

	heap_endscan(fragscan);
	heap_close(xFragDesc);

	lastTuple = NULL;	/* Reset. */

	if (fragxarg->fragxtype != COPYFRAGMENT) {
	    /*
	     * This a fragment transfer. Remove as many traces of the fragment
	     * as we can since we finished moving it.
	     */

	    /*
	     * Destroy archive relation at our site.
	     */
	    DestroyArchiveRelation(fragxarg->frastoreid);

	    if (fragxarg->fraclassid != fragxarg->frastoreid) {
		heap_destroy(xFragDesc->rd_rel->relname.data);
	    }
	    else {
		/*
		 * We are moving the master fragment. Can't heap destroy it,
		 * but we'll pluck as much out as we can, first the indices,
		 * then the tuples.
		 */
		RelationRemoveIndexes(xFragDesc);

		sprintf(query, "DELETE FROM %s",
			xFragDesc->rd_rel->relname.data);
		pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	    }

	    /*
	     * Now update pg_fragment on our site.
	     */
	    sprintf(query,
		    "UPDATE pg_fragment SET fralocation='%d' WHERE fraclassid='%d'::oid and frastoreid='%d'::oid",
		    fragxarg->sender.hostid,
		    fragxarg->fraclassid,
		    fragxarg->frastoreid);
	    pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
        }

	elog(NOTICE, "Leaving FragmentGetData.");

	databuf->error = SUCCESS;

	return;
    }

    /*
     * Fill in tuples until done.
     */
    while (HeapTupleIsValid(htup)) {
	int len = htup->t_len;

	if ((len + nbytes) > SENDBUFSIZE) {
	    /*
	     * Send the buffer off.
	     */
	    databuf->numTuples = numTuples;
	    databuf->data.Data_len = nbytes;
	    databuf->error = SUCCESS;

	    lastTuple = htup;	/* Save it for next time. */

	    elog(NOTICE, "Leaving FragmentGetData.");

	    return;
	}

	memmove(buffer, htup, len);
	buffer += len;
	numTuples++;
	nbytes += len;

	htup = heap_getnext(fragscan, 0, NULL);
    }

    lastTuple = NULL;	/* nothing left */
    
    databuf->numTuples = numTuples;
    databuf->data.Data_len = nbytes;
    databuf->error = SUCCESS;

    elog(NOTICE, "Leaving FragmentGetData.");
}


/*****************************************************************************
 * Utility Routines
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	CopyFragXArg	
 *
 *	Just a routine to copy a FragXArg struct.
 *
 * ----------------------------------------------------------------
 */
void
CopyFragXArg(FragXArg *source, FragXArg *dest)
{
    dest->fragxtype = source->fragxtype;
    dest->qepFragId = source->qepFragId;
    dest->sender.port = source->sender.port;
    dest->sender.addr = source->sender.addr;
    dest->sender.hostid = source->sender.hostid;
    dest->receiver.port = source->receiver.port;
    dest->receiver.addr = source->receiver.addr;
    dest->receiver.hostid = source->receiver.hostid;
    dest->fraclassid = source->fraclassid;
    dest->frastoreid = source->frastoreid;
    dest->remotexactid = source->remotexactid;
}
