/*-------------------------------------------------------------------------
 *
 * exchange.c--
 *    implements sending and receiving side of inter-exec. tuple streams
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/exchange.c,v 1.32 1996/05/31 17:14:12 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <rpc/rpc.h>
#include <rpcsvc/rusers.h>
#include <sys/time.h>
#include <pthread.h>
#include "c.h"
#include "access/htup.h"
#include "utils/elog.h"
#include "exchange.h"
#include "mariposa_rpc.h"
#include "rpc/rpcutils.h"

#define	TRACETUP 4

typedef void *(*ThreadStartRoutine)(void *);

typedef struct {
    ExchSendChannel channel;
    int threadNo; /* index of activated thread */
} SendThreadArgData;
typedef SendThreadArgData *SendThreadArg;

typedef struct {
    struct sockaddr_in addr;
    bool initialized;		/*
				 * true if addr has been initialized
				 * with the receiving backend's
				 * address after the first call. The
				 * first call always goes to the site
				 * mgr.
				 */
    QepFragmentId qepFragId;
    RpcConn *rpcconn;
    pthread_t sendThread;
    pthread_mutex_t receiverBlockedMutex;
    pthread_cond_t receiverBlockedCond;
    bool receiverIsBlocked;
    
    SendThreadArgData threadArg;/*
				 * Arguments to a thread cannot be
				 * procedure-local variables; they
				 * have to persist until the thread is
				 * finished.
				 */
} SiteInfo;

struct _ExchSendChannelData {
    int nRecvSites;		/*
				 * Number of sites we send to. We'll
				 * have as many elements in the
				 * following array.
				 */
    SiteInfo *receivers;

    int	tableIndex;		/* Index into exchOutChannelTab[] 	*/
    bool threadsActivated;	/* has the thread ever been activated? */

    /*TypeId typeId;*/

    char buffer1[SENDBUFSIZE];
    char buffer2[SENDBUFSIZE];	/*
				 * One of {'buffer1','buffer2'} is
				 * used by the currently active
				 * thread(s) for sending and the other
				 * one is filled up by sendTuple. When
				 * re-starting the sending thread(s),
				 * the buffers exchange roles. The
				 * effect is that the thread can do a
				 * synchronous call with private data,
				 * but no buffers have to be copied.
				 */
    char *clientBuffer;		/*
				 * this points to the buffer that's
				 * currently used by sendTuple.
				 */
    int clientBufferSize;	/*
				 * amount of data in buffer used by
				 * sendTuple.
				 */
    TransmitBuffer transmitBuffer; /* used by the threads for sending */
    int numTuplesSent;		/* total number of tuples sent */
    int numBufferSent;		/* total number of buffers sent */
};
typedef struct _ExchSendChannelData *ExchSendChannel;


typedef struct {
    struct sockaddr_in addr;	/*
				 * address and port of either remote
				 * backend's site mgr or the remote
				 * backend itself.
				 */
    int hostId;
    QepFragmentId qepFragId;	/*
				 * 'hostId' and 'qepFragId' uniquely
				 * identify the data source.
				 */

    bool eos;			/* true if end-of-stream is true. */
    bool initialized;		/*
				 * true if the addr has been updated
				 * from the site mgr's to the remote
				 * backend's direct address.
				 */
} SenderInfo; 

typedef struct {
    TransmitBuffer *buffer;
    SVCXPRT *transportHandle;	/*
				 * used for svc_freeargs with the
				 * respective buffer
				 */
    int sender;			/*
				 * index into 'senders' array of
				 * ExchRecvChannel
				 */
} BufferInfo;

struct _ExchRecvChannelData {
    int nSendSites;
    SenderInfo *senders;

    int tableIndex;			/*
					 * index in the global channel
					 * table.
					 */

    /* buffer scanning related */
    int nextFreeSlot;			/*
					 * top of the allocated slots
					 * in 'buffers'.
					 */
    int currSlot;			/*
					 * bottom of the allocated
					 * slots in 'buffers'.
					 */
    int scanPos;			/*
					 * next position (byte offset)
					 * from which to read in the
					 * current buffer.
					 */

    /* locks and condition var's */
    pthread_mutex_t bufferLock;
    pthread_cond_t bufferNotFull;	/*
					 * a thread waits on this
					 * condition if it finds that
					 * there are no free entries
					 * in 'buffers', i.e., the
					 * next thread that frees a
					 * buffer will signal
					 * 'bufferNotFull'.
					 */
    pthread_cond_t bufferNotEmpty;	/*
					 * a thread waits on this
					 * condition if there are no
					 * buffers with data ready for
					 * it to process, i.e., the
					 * next thread that accepts a
					 * new buffer will signal
					 * 'bufferNotEmpty'.
					 */

    bool needNewBuffer;
    int freeBuffers;			/*
					 * number of vacant slots in
					 * 'buffers'.
					 */
    BufferInfo buffers[MAXBUFFERS];	/*
					 * circular list of buffers.
					 */
};
typedef struct _ExchRecvChannelData *ExchRecvChannel;

/* forward declarations */
static void transmitBuffer(SendThreadArg arg);
static bool joinSenders(ExchSendChannel channel);
static Pointer getNextTuple(ExchRecvChannel channel);

/* file-local variables for receivers */
#define EXCHINCHANNEL_UNINIT ((ExchRecvChannel) -1)
static ExchRecvChannel exchInChannelTab[MAXRECVCHANNELS] = { EXCHINCHANNEL_UNINIT };
/*static TransactionId globalXactId;
static CommandId cmdId;*/
static QepFragmentId qepFragId;

/* and for senders */
#define EXCHOUTCHANNEL_UNINIT ((ExchSendChannel) -1)
static ExchSendChannel exchOutChannelTab[MAXSENDCHANNELS] = { EXCHOUTCHANNEL_UNINIT };

/*
 * openSendChannel --
 *	initialize the channel structure and bind to all receiving sites
 *
 *	data sent through the channel will go to the designated QEP fragments
 *	at all recipient sites 
 */
ExchSendChannel
openSendChannel(SiteAddr recvSites[],
		int nRecvSites,
		/*TransactionId globalXact, CommandId cmd,*/
		QepFragmentId sendQepFrag,
		QepFragmentId recvQepFrags[]/*, TypeId type*/)
{
    ExchSendChannel channel;
    int i;
    bool found = false;
    int channelIndex = 0;
    extern unsigned short tcpPort;
    
    if (exchOutChannelTab[0] == EXCHOUTCHANNEL_UNINIT) {
	(void *) memset(exchOutChannelTab, 0, sizeof(exchOutChannelTab));
    }

    /* find place in channel table */
    while (channelIndex < MAXSENDCHANNELS && !found) {
	if (exchOutChannelTab[channelIndex] == NULL) {
	    found = true;
	} else {
	    channelIndex++;
	}
    }
    if (!found) {
	tcopDebug(0, "openSendChannel: channel limit exceeded.\n");
	/* we've exceeded our supply of channels */
	return (NULL);
    }
    channel = (ExchSendChannel) palloc(sizeof(*channel));
    if (channel == NULL) {
	elog(WARN, "openSendChannel: can't allocate channel");
    }
    channel->receivers = (SiteInfo *) palloc(nRecvSites *
        sizeof(*channel->receivers));
    if (channel->receivers == NULL) {
	elog(WARN, "openSendChannel: can't allocate receivers");
    }

    channel->transmitBuffer.sendingSite.hostid = getMyHostId();
    channel->transmitBuffer.sendingSite.addr = getMyAddr();
    channel->transmitBuffer.sendingSite.port = htons((u_short) tcpPort);
    
    channel->nRecvSites = nRecvSites;
    for (i = 0; i < nRecvSites; i++) {
	channel->receivers[i].addr.sin_family = AF_INET;
	channel->receivers[i].addr.sin_addr.s_addr = recvSites[i].addr;
	if (recvSites[i].port == 0) {
	    channel->receivers[i].initialized = false;
	    channel->receivers[i].addr.sin_port = htons((u_short) SITEMGRPORT);
	} else {
	    /* we already know the backend's port # -> must be the top node */
	    channel->receivers[i].initialized = true;
	    channel->receivers[i].addr.sin_port =
		htons((u_short) recvSites[i].port);
	}
	channel->receivers[i].qepFragId = recvQepFrags[i];
	pthread_mutex_init(&channel->receivers[i].receiverBlockedMutex, pthread_mutexattr_default);
	pthread_cond_init(&channel->receivers[i].receiverBlockedCond, pthread_condattr_default);
	
    }
    channel->threadsActivated = false;
#if 0
    channel->typeId = type;
#endif
    channel->clientBuffer = channel->buffer1;
    channel->clientBufferSize = 0;
#if 0
    channel->transmitBuffer.globalXactId = globalXact;
    channel->transmitBuffer.cmdId = cmd;
#endif
    channel->transmitBuffer.sendQepFragId = sendQepFrag;
    channel->numTuplesSent = channel->numBufferSent = 0;

    /* Generally, bind to site mgrs at all receiving sites.
     * Only if we send to the final destination = top node do we
     * bind directly to the corresponding backend. The reason is that we
     * don't know the port number in any other case; the top node's executing
     * backend is the coordinating backend.
     * The fact that we do *not* send to the top node is indicated by port = 0.
     */
    for (i = 0; i < nRecvSites; i++) {
	tcopDebug(2, "openSendChannel: to (%d, %d, %d)\n",
		  recvSites[i].hostid,
		  channel->receivers[i].addr.sin_addr.s_addr,
		  channel->receivers[i].addr.sin_port);

	if (!channel->receivers[i].initialized) {
	    /* we don't know who to talk to, let's take the site mgr. */

	    channel->receivers[i].rpcconn =
		Rpc_connectSitemgr(recvSites[i].hostid,
				   0,
				   "openSendChannel",
				   RPC_ONERR_WARN);
	} else {
	    channel->receivers[i].rpcconn =
		Rpc_connectExec(channel->receivers[i].addr.sin_addr.s_addr,
				channel->receivers[i].addr.sin_port,
				"openSendChannel",
				RPC_ONERR_WARN);
	}
    }

    tcopDebug(2, "openSendChannel: channelIndex = %i", channelIndex);
    exchOutChannelTab[channelIndex] = channel;
    channel->tableIndex = channelIndex;

    tcopDebug(2, "openSendChannel: DONE");
    
    return (channel);
}


/*
 * sendTuple --
 *	append a tuple to the buffer and possibly flush buffer
 *
 *	tuples are buffered until buffer overflows;
 *	when that happens, send it to all receiving sites;
 *	sending is handled by threads, which do synchronous call
 *	in order to catch errors;
 *	there is one thread per receiving site
 */
void
sendTuple(ExchSendChannel channel, Pointer tuple)
{
    Size tupleLen;
    int status, i;
    extern int nBuffersSent;
    
    ++(channel->numTuplesSent);
    tupleLen = ((HeapTuple) tuple)->t_len;
    while (channel->clientBufferSize + tupleLen > SENDBUFSIZE) { 
	/* buffer is full, send it off */
	if (channel->clientBufferSize == 0) {
	    /* the buffer is empty, but the tuple still won't fit */
	    elog(WARN, "sendTuple: tuple too large");
	}
	if (channel->threadsActivated) {
	    /*
	     * we have to make sure we only have one outstanding 
	     * transmission per receiving site (for flow control;
	     * would also cause race conditions with regard to thread
	     * arguments etc.) 
	     */
	    if (!joinSenders(channel)) {
		/* there was a transmission error 
		elog(WARN, "sendTuple: RPC error");*/
	    }
	}

	/* prepare the buffer and send it */
	channel->transmitBuffer.flags = TransmitMoreData;
	channel->transmitBuffer.data.Data_val = channel->clientBuffer;
	channel->transmitBuffer.data.Data_len = channel->clientBufferSize;
	channel->numBufferSent++;

	tcopDebug(2, "sendTuple: ntup=%d nbuf=%d\n",
		  channel->numTuplesSent, channel->numBufferSent);
	
	channel->threadsActivated = true;
	for (i = 0; i < channel->nRecvSites; i++) {
	    nBuffersSent++;
	    channel->receivers[i].threadArg.channel = channel;
	    channel->receivers[i].threadArg.threadNo = i;
	    channel->transmitBuffer.recvQepFragId =
	        channel->receivers[i].qepFragId;
	    status = pthread_create(&channel->receivers[i].sendThread,
	        pthread_attr_default, (ThreadStartRoutine) transmitBuffer,
		(void *) &channel->receivers[i].threadArg);
	    checkAndElog(status, "sendTuple: can't create thread");
	}

	/* exchange buffers */
	if (channel->clientBuffer == channel->buffer1) {
	    channel->clientBuffer = channel->buffer2;
	} else {
	    channel->clientBuffer = channel->buffer1;
	}
	channel->clientBufferSize = 0; /* reset */
    }
    (void) memcpy((char *) channel->clientBuffer + channel->clientBufferSize,
       (char *) tuple, tupleLen);
    channel->clientBufferSize += tupleLen;
}

/*
 * closeSendChannel --
 *	flush the buffer, if it still has data
 *
 *	last RPCs are synchronous in order to notice errors
 */
void
closeSendChannel(ExchSendChannel channel)
{
    int i;
    TransmitReturn transmitResult;
    SiteAddr addr;
    extern int nBuffersSent;
    
    if (channel->threadsActivated) {
	/*
	 * We wait for the last threads to finish so that if there was
	 * an error, we can report it.
	 */
	if (!joinSenders(channel)) {
	    /* there was a transmission error
	    elog(WARN, "closeSendChannel: RPC error"); */
	}
    }
    /*
     * flush buffer or send an empty buffer. Need to send the empty buffer
     * because the receiving end expects at least one buffer. (see
     * comments on recvTuple)
     */
    if (channel->clientBufferSize > 0 || channel->numBufferSent==0) {
	/* prepare the buffer and send it*/

	nBuffersSent++;
	
	channel->transmitBuffer.flags &= ~TransmitMoreData;
	channel->transmitBuffer.data.Data_val = channel->clientBuffer;
	channel->transmitBuffer.data.Data_len = channel->clientBufferSize;
	channel->numBufferSent++;

	tcopDebug(2, "closeSendChannel: ntup=%d nbuf=%d\n",
		  channel->numTuplesSent, channel->numBufferSent);
	
	for (i = 0; i < channel->nRecvSites; i++) {
	    channel->transmitBuffer.recvQepFragId =
	        channel->receivers[i].qepFragId;

	    Rpc_setCallerName(channel->receivers[i].rpcconn,
			      "closeSendChannel");

	    if (!channel->receivers[i].initialized) {
		tcopDebug(2, "closeSendChannel: calling SiteMgr %d\n",
			  channel->receivers[i].rpcconn->rpc_hostid);
		/* this goes to the site mgr */
		Rpc_call(channel->receivers[i].rpcconn,
			 RPC_SITEMGRTRANSMITDATA, 
			 &channel->transmitBuffer,
			 &transmitResult);
		addr = transmitResult.port;
	    } else {
		ReceiverStatus receiverStatus;
		/* this goes to the receiving backend */
		tcopDebug(2, "closeSendChannel: calling exec %d\n",
			  channel->receivers[i].rpcconn->rpc_hostid);
		Rpc_call(channel->receivers[i].rpcconn,
			 RPC_TRANSMITDATA, 
			 &channel->transmitBuffer,
			 &receiverStatus);
	    }
	}
    }
    for (i = 0; i < channel->nRecvSites; i++) {
	Rpc_disconnect(channel->receivers[i].rpcconn);
    }
    exchOutChannelTab[channel->tableIndex] = NULL;
    pfree(channel->receivers);
    pfree(channel);
}

/* 
 * transmitBuffer --
 *	transmit a single buffer
 *
 *	thread start procedure;
 *	returns true if all went well;
 *	creates a new client handle for the receiving backend
 * 	if this was the first transmit call, in which case
 *	this goes to the site mgr
 */
static void 
transmitBuffer(SendThreadArg arg)
{
    ReceiverStatus receiverStatus;
    ExchSendChannel channel = (ExchSendChannel) arg->channel;
    TransmitReturn transmitResult;
    SiteAddr addr;
    SiteInfo *siteInfo = &channel->receivers[arg->threadNo];

    Rpc_setCallerName(siteInfo->rpcconn, "transmitBuffer");
    Rpc_setOnerr(siteInfo->rpcconn, RPC_ONERR_RETURN);

    tcopDebug(2, "transmitBuffer START: receive qepFragId = %i\n send qepFragId = %i\n threadno = %d\n",
	      channel->transmitBuffer.recvQepFragId,
	      channel->transmitBuffer.sendQepFragId,
	      arg->threadNo
	      );
    
    if (!siteInfo->initialized) {
	/*
	 * We call the site mgr, update our receiving port and 
	 * re-create the client handle, this time for the 'real' backend. 
	 */
	tcopDebug(2, "transmitBuffer: setting up <qep=%d host=%d>\n",
		  channel->transmitBuffer.recvQepFragId,
		  siteInfo->rpcconn->rpc_hostid);
	if (!Rpc_call(siteInfo->rpcconn,
		      RPC_SITEMGRTRANSMITDATA,
		      &channel->transmitBuffer,
		      &transmitResult)) { /* synch. call */
	    pthread_exit((void *) false);
	}
	addr = transmitResult.port;

	siteInfo->initialized = true;
	siteInfo->addr.sin_port = htons((u_short) addr.port);
	Rpc_disconnect(siteInfo->rpcconn);

	siteInfo->rpcconn =
	    Rpc_connectExec(siteInfo->addr.sin_addr.s_addr,
			    htons(siteInfo->addr.sin_port),
			    "transmitBuffer",
			    RPC_ONERR_RETURN);
	if (siteInfo->rpcconn == NULL) {
	    pthread_exit((void *) false);
	}
	    
    } else {
	/* synchronous call to receiveData() at the receiver site */
	int status = Rpc_call(siteInfo->rpcconn,
		      RPC_TRANSMITDATA,
		      &channel->transmitBuffer,
		      &receiverStatus);
	
	if (!status) {
    
	    tcopDebug(2, "transmitBuffer FINISH: receive qepFragId = %i\n send qepFragId = %i\n threadno = %d\n",
		      channel->transmitBuffer.recvQepFragId,
		      channel->transmitBuffer.sendQepFragId,
		      arg->threadNo
		      );
	    
	    pthread_exit((void *) false);

	/*
	 * Rpc call returned ok - check if receiver had a buffer available
	 * or if it had to block.
	 */
	} else {
	    tcopDebug(2, "transmitBuffer: receiver status code = %i\n", receiverStatus);
	    switch (receiverStatus) {
	    case RECEIVER_STATUS_OK:
		break;
	    case RECEIVER_STATUS_BLOCKED: {
		pthread_mutex_lock(&siteInfo->receiverBlockedMutex);
		siteInfo->receiverIsBlocked = TRUE;
		tcopDebug(2, "transmitBuffer: <send=%d recv=%d> waiting for receiver to unblock\n",
			  channel->transmitBuffer.sendQepFragId,
			  channel->transmitBuffer.recvQepFragId);
		while (siteInfo->receiverIsBlocked) {
		    pthread_cond_wait(&siteInfo->receiverBlockedCond, &siteInfo->receiverBlockedMutex);
		}
		pthread_mutex_unlock(&siteInfo->receiverBlockedMutex);
	    }
		break;

	    default: {
		tcopDebug(2, "transmitBuffer: unknown receiver status code %i\n", receiverStatus);
		elog(WARN, "transmitBuffer: unknown receiver status code %i", receiverStatus);
	    }
		break;
	    }
	}
	   
	
    }

    Rpc_setOnerr(siteInfo->rpcconn, RPC_ONERR_DEFAULT);

    
    tcopDebug(2, "transmitBuffer FINISH: receive qepFragId = %i\n send qepFragId = %i\n threadno = %d\n",
	      channel->transmitBuffer.recvQepFragId,
	      channel->transmitBuffer.sendQepFragId,
	      arg->threadNo
	      );

    pthread_exit((void *) true);
}


void
ReceiverUnblocked(void *args) {
    char **argv = (char **)args;
    UnblockBuffer *buffer = (UnblockBuffer *)argv[0];
    SVCXPRT *transportHandle = (SVCXPRT *)argv[1];
    int result = RPC_SUCCESS;
    int sendChannelIndex;
    int receiverIndex;
    ExchSendChannel channel;
    bool found;
    /*
     * Reply to sender
     */
    if (!svc_sendreply(transportHandle, xdr_RPCErrorCode, &result)) {
	elog(WARN, "ReceiverUnblocked: svc_sendreply() failed.");
    }
    
    /*
     * Find which channel in the exchange-out channel table
     * was connected to the receiver.
     */
    found = false;
    for (sendChannelIndex = 0; sendChannelIndex < MAXSENDCHANNELS; sendChannelIndex++) {
	if ((channel = exchOutChannelTab[sendChannelIndex]) != NULL) {
	    if (channel->transmitBuffer.sendQepFragId == buffer->sendQepFragId) {
		found = true;
		break;
	    }
	}
    }

    if (!found) {
	elog(NOTICE, "ReceiverUnblocked: Can't determine send channel... transmission may be complete.");
    } else {

	tcopDebug(3, "ReceiverUnblocked: sendqepFragId determined: %i\n", buffer->sendQepFragId);
    
	/*
	 * Now find which receiver among the (possibly) many
	 * the receiver was.
	 */
	found = false;
	for (receiverIndex = 0; receiverIndex < channel->nRecvSites; receiverIndex++) {
	    if (channel->receivers[receiverIndex].qepFragId == buffer->recvQepFragId) {
		found = true;
		break;
	    }
	}
    
	if (!found) {
	    elog(WARN, "ReceiverUnblocked: Can't determine receiver.");
	}

	tcopDebug(3, "ReceiverUnblocked: recvqepFragId determined: %i\n", buffer->recvQepFragId);

	/*
	 * Wait until the thread in transmitBuffer has acquired the lock,
	 * set receiverIsBlocked to TRUE, and then called cond_wait().
	 */
	while (!channel->receivers[receiverIndex].receiverIsBlocked) {
	}
	
	/*
	 * Now signal the appropriate thread, which must be
	 * waiting inside transmitBuffer()
	 */
	pthread_mutex_lock(&(channel->receivers[receiverIndex].receiverBlockedMutex));
	channel->receivers[receiverIndex].receiverIsBlocked = FALSE;
	pthread_mutex_unlock(&(channel->receivers[receiverIndex].receiverBlockedMutex));
	pthread_cond_signal(&(channel->receivers[receiverIndex].receiverBlockedCond));
    }
}

/*
 * joinSenders --
 *	waits for all sending threads to finish
 *
 *	returns true if all sending threads transmitted successfully
 */
static bool
joinSenders(ExchSendChannel channel)
{
    int i;
    bool success, result;
    int status;

    success = true;
    result = true;
    for (i = 0; i < channel->nRecvSites; i++) {
	status = pthread_join(channel->receivers[i].sendThread,
	    (void *) &success);
	checkAndElog(status, "joinSenders: can't join thread");
	status = pthread_detach(&channel->receivers[i].sendThread);
	checkAndElog(status, "joinSenders: can't detach thread");
	if (success == false) {
	    /*
	     * there was an error in a previous transmission;
	     * join and detach the other threads anyway
	     */
	    result = false;
	}
    }
    tcopDebug(2, "joinSenders: DONE.  result = %i", result);
    
    return (result);
}


/*
 * receiveData --
 *	enqueues the received buffer into the buffer array of 
 *	the receiving channel
 */

void
receiveData(void *args)
{
    char **argv = (char **)args;
    TransmitBuffer *buffer = (TransmitBuffer *)argv[0];
    SVCXPRT *transportHandle = (SVCXPRT *)argv[1];
    int i, j;
    int sender;
    int status;
    ExchRecvChannel channel;
    SenderInfo *sinfo;
    extern int nBuffersReceived;
    ReceiverStatus receiverStatus;
    bool blocked = false; 	/* Set to true if we block and wait for a free buffer */
    int sendQepFragId = buffer->sendQepFragId;
    int recvQepFragId = buffer->recvQepFragId;

    nBuffersReceived++;
    
    tcopDebug(1, "Exec: receiveData called from <qep=%d, host=%d>\n",
	      buffer->sendQepFragId, buffer->sendingSite.hostid);
    
    /*
     * refuse to accept any data until the channel table has been
     * initialized by openRecvChannel.
     */
    if (exchInChannelTab[0] == EXCHINCHANNEL_UNINIT) {
	while (exchInChannelTab[0] == EXCHINCHANNEL_UNINIT) {
	    static const struct timespec tv = { 0, 100000 };
	    pthread_delay_np(&tv);
	}
	tcopDebug(2, "receiveData: channel initialized.\n");
    }
    
    /*
     * find the right sender information for this buffer.  set
     * 'sender' to be the array index of the plan fragment in the
     * channel descriptor 'channel'.
     */
    sender = -1;
    for (i = 0; i < MAXRECVCHANNELS && sender < 0; ++i) {
	if ((channel = exchInChannelTab[i]) != NULL) {
	    for (j = 0; j < channel->nSendSites && sender < 0; ++j) {
		sinfo = &(channel->senders[j]);
		if ((sinfo->qepFragId == buffer->sendQepFragId) &&
		    (sinfo->hostId == buffer->sendingSite.hostid)) {
		    sender = j;
		}
	    }
	}
    }

    /*
     * if we can't figure out the origin of this buffer, drop it on
     * the floor and ignore it (i.e., let the sender time out).
     */
    if (sender < 0) {
	tcopDebug(2, "receiveData: channel table:\n");
	for (i = 0; i < MAXRECVCHANNELS; ++i) {
	    if ((channel = exchInChannelTab[i]) != NULL) {
		for (j = 0; j < channel->nSendSites; ++j) {
		    sinfo = &(channel->senders[j]);
		    tcopDebug(2, "\tvalid sender <qep=%d host=%d>\n",
			      sinfo->qepFragId, sinfo->hostId);
		}
	    }
	}
	tcopDebug(2, "receiveData: discarding buffer from <qep=%d host=%d>",
	     buffer->sendQepFragId, buffer->sendingSite.hostid);
	free(buffer);
	return;
    }
    
    /*
     * place received data into the buffer ring.
     */
    status = pthread_mutex_lock(&channel->bufferLock);
    checkAndExit(status, "receiveData: can't lock bufferLock");
    while (channel->freeBuffers == 0) {
	/* 
	 * We block if we can't take the buffer.
	 */
	tcopDebug(3, "receiveData %i waiting on bufferNotFull...\n", channel->senders[sender].qepFragId);
	receiverStatus = RECEIVER_STATUS_BLOCKED;
	blocked = true;
	tcopDebug(3, "receiveData %i returning blocked.  receiverStatus = %i\n", channel->senders[sender].qepFragId, receiverStatus);
	if (!svc_sendreply(transportHandle, xdr_ReceiverStatus, &receiverStatus)) {
	    free(args);
	    errorExit("receiveData: can't send reply\n");
	}
	status = pthread_cond_wait(&channel->bufferNotFull,
				   &channel->bufferLock);
	checkAndExit(status, "receiveData: couldn't wait for bufferNotFull");
	tcopDebug(3, "receiveData %i DONE waiting on bufferNotFull\n", channel->senders[sender].qepFragId);
    }
    
    /* put received data in channel's buffer list */
    channel->nextFreeSlot = (channel->nextFreeSlot + 1) % MAXBUFFERS;
    channel->buffers[channel->nextFreeSlot].buffer = buffer;
    channel->buffers[channel->nextFreeSlot].transportHandle = transportHandle;
    channel->buffers[channel->nextFreeSlot].sender = sender;
    channel->freeBuffers--;

    tcopDebug(2, "receiveData: sender=<qep=%d host=%d> flags=0x%x curr=%d top=%d free=%d\n",
	      buffer->sendQepFragId, buffer->sendingSite.hostid,
	      buffer->flags,
	      channel->currSlot, channel->nextFreeSlot, channel->freeBuffers);
    
    status = pthread_mutex_unlock(&channel->bufferLock);
    checkAndExit(status, "receiveData: can't unlock bufferLock");
    tcopDebug(2, "receiveData: Broadcasting bufferNotEmpty\n");
    status = pthread_cond_broadcast(&channel->bufferNotEmpty);
    checkAndExit(status, "receiveData: can't signal bufferNotEmpty");

    if (!blocked) {
	receiverStatus = RECEIVER_STATUS_OK;
	tcopDebug(3, "receiveData %i returning unblocked.  receiverStatus = %i\n", channel->senders[sender].qepFragId, receiverStatus);
	if (!svc_sendreply(transportHandle, xdr_ReceiverStatus, &receiverStatus)) {
	    free(args);
	    errorExit("receiveData: can't send reply\n");
	}
    } else {
	UnblockBuffer unblockBuffer;
	RpcConn *rpcc;
	int status;

	tcopDebug(3, "receiveData connecting to exec <qep=%d host=%d>\n", buffer->sendQepFragId, buffer->sendingSite.hostid);
	rpcc = Rpc_connectExec(buffer->sendingSite.addr,
				buffer->sendingSite.port,
				"receiveData",
				RPC_ONERR_WARN);

	if (rpcc == NULL) {
	    elog(WARN, "receiveData: can't connect to exec for receiverUnblock call.");
	}
	
	unblockBuffer.sendQepFragId = sendQepFragId;
	unblockBuffer.recvQepFragId = recvQepFragId;

	tcopDebug(3, "receiveData %i calling ReceiverUnblocked <sendqep=%d recvqep=%d>\n", buffer->sendQepFragId,
		  unblockBuffer.sendQepFragId, unblockBuffer.recvQepFragId);
	
	status = Rpc_call(rpcc,
		      RPC_RECEIVERUNBLOCKED,
		      &unblockBuffer,
		      NULL);

	if (!status) {
	    Rpc_disconnect(rpcc);
	    elog(WARN, "receiveData %i Rpc call to receiverUnblock at host %i failed", buffer->sendQepFragId, buffer->sendingSite.hostid);
	}
	Rpc_disconnect(rpcc);
    }
    
    free(args);
}


/*
 * openRecvChannel --
 *	allocates data structures for receive channel and
 *	registers it in exchInChannelTab
 */
ExchRecvChannel
openRecvChannel(SiteAddr sendSites[],
		int nSendSites,
		/*TransactionId globalXact, CommandId cmd,*/
		QepFragmentId recvQepFrag,
		QepFragmentId sendQepFragIds[]/*, TypeId type*/)
{
    int i = 0;
    int j;
    bool found = false;
    ExchRecvChannel channel;
    int status;

    tcopDebug(2, "openRecvChannel: called.\n");
    
    if (exchInChannelTab[0] == EXCHINCHANNEL_UNINIT) {
	(void *) memset(exchInChannelTab, 0, sizeof(exchInChannelTab));
    }

    /* find place in channel table */
    while (i < MAXRECVCHANNELS && !found) {
	if (exchInChannelTab[i] == NULL) {
	    found = true;
	} else {
	    i++;
	}
    }
    if (!found) {
	tcopDebug(0, "openRecvChannel: channel limit exceeded.\n");
	/* we've exceeded our supply of channels */
	return (NULL);
    }
    channel = (ExchRecvChannel) palloc(sizeof(*channel));
    if (channel == NULL) {
	elog(WARN, "openRecvChannel: can't palloc channel");
    }
    channel->senders = (SenderInfo *) palloc(nSendSites *
        sizeof(*channel->senders));
    if (channel->senders == NULL) {
	elog(WARN, "openRecvChannel: can't palloc senders");
    }

    tcopDebug(2, "openRecvChannel: initializing channel data...\n");
    
    for (j = 0; j < nSendSites; j++) {
	channel->senders[j].addr.sin_family = AF_INET;
	channel->senders[j].addr.sin_port = (u_short) sendSites[j].port;
	channel->senders[j].addr.sin_addr.s_addr = sendSites[j].addr;
	channel->senders[j].hostId = sendSites[j].hostid;
	channel->senders[j].qepFragId = sendQepFragIds[j];
	channel->senders[j].eos = false;
	channel->senders[j].initialized = false;
    }
    channel->nSendSites = nSendSites;
    channel->tableIndex = i;
    channel->nextFreeSlot = channel->currSlot = -1;
    channel->scanPos = 0; 
    pthread_mutex_init(&channel->bufferLock, pthread_mutexattr_default);
    pthread_cond_init(&channel->bufferNotFull, pthread_condattr_default);
    pthread_cond_init(&channel->bufferNotEmpty, pthread_condattr_default);
    channel->needNewBuffer = true;	/* don't have any yet */
    channel->freeBuffers = MAXBUFFERS;
    exchInChannelTab[i] = channel;
#if 0
    globalXactId = globalXact;
    cmdId = cmd;
#endif
    qepFragId = recvQepFrag;

    tcopDebug(2, "openRecvChannel: done.\n");
	
    return (channel);
}

/*
 * recvTuple --
 *	returns the next tuple found in the queue of the specified
 *	channel
 *
 *	blocks on CV bufferNotEmpty if there is nothing to return;
 *	releases a buffer if it is done with it
 *	expects at least one buffer from the sending site, otherwise
 *	it would block indefinitely
 */
void
recvTuple(ExchRecvChannel channel, Pointer *tuple)
{
    int i;
    bool eos;
    int status;

    while (true) {
	status = pthread_mutex_lock(&channel->bufferLock);
	checkAndElog(status, "recvTuple: can't lock bufferLock");

	/*
	 * if the channel has no buffers ready (from any sending
	 * site), wait until a buffer arrives and make 'currSlot'
	 * point to it.
	 */
	if (channel->needNewBuffer == true) {
	    while (channel->freeBuffers == MAXBUFFERS) {
		tcopDebug(2, "recvTuple: waiting on bufferNotEmpty\n");
		status = pthread_cond_wait(&channel->bufferNotEmpty,
					   &channel->bufferLock);
		checkAndElog(status, "recvTuple: can't wait on bufferNotEmpty");
	    }
	    channel->currSlot = (channel->currSlot + 1) % MAXBUFFERS;
	    channel->scanPos = 0;
	    channel->needNewBuffer = false;
	    tcopDebug(TRACETUP, "recvTuple: bufferNotEmpty set, flags=0x%x curr=%d top=%d free=%d\n",
		      channel->buffers[channel->currSlot].buffer->flags,
		      channel->currSlot,
		      channel->nextFreeSlot,
		      channel->freeBuffers);
	} else {
	    if (channel->currSlot < 0) {
		channel->currSlot = 0;
	    }
	    tcopDebug(TRACETUP, "recvTuple: flags=0x%x curr=%d top=%d free=%d\n",
		      channel->buffers[channel->currSlot].buffer->flags,
		      channel->currSlot,
		      channel->nextFreeSlot,
		      channel->freeBuffers);
	}

	status = pthread_mutex_unlock(&channel->bufferLock);
	checkAndElog(status, "recvTuple: can't unlock bufferLock");

	/*
	 * at this point, 'currSlot' points to a valid buffer in this
	 * channel.  extract a tuple from it, if there are any.
	 */
	*tuple = getNextTuple(channel);
	if (*tuple != NULL) {
	    tcopDebug(TRACETUP, "recvTuple: got a tuple\n");
	    return; /* we found a new tuple */
	}

	/*
	 * we had a valid buffer but it ran out of tuples.  if this
	 * sender doesn't have any more data (indicated by clearing
	 * the TransmitMoreData flag), make a note of that fact.
	 *
	 * if all of the senders are done, then we can quit right now.
	 */
	if ((channel->buffers[channel->currSlot].buffer->flags &
	     TransmitMoreData) == 0) {
	    BufferInfo *info = &(channel->buffers[channel->currSlot]);

	    /*
	     * remember that the QEP fragment of the current buffer is
	     * done.
	     */
	    channel->senders[info->sender].eos = true;
	    tcopDebug(2, "recvTuple: EOS for <qep=%d host=%d>\n",
		      info->buffer->sendQepFragId,
		      info->buffer->sendingSite.hostid);

	    /*
	     * if all of our feeding QEP fragments are done, we're
	     * also done.
	     */
	    eos = true;
	    for (i = 0; i < channel->nSendSites; i++) {
		if (channel->senders[i].eos == false) {
		    eos = false;
		}
	    }
	    if (eos == true) {
		tcopDebug(2, "recvTuple: received last EOS\n");
		return; /* returns a NULL tuple to signal EOS */
	    }
	} 

	status = pthread_mutex_lock(&channel->bufferLock);
	checkAndElog(status, "recvTuple: can't lock bufferLock");

	/*
	 * there are still senders who have data pending, so we need a
	 * new buffer from somewhere.  free up 'currSlot', and if some
	 * receiveData thread was waiting for a free buffer, let it
	 * know that 'currSlot' is now free.
	 */
#if 0
	if (!svc_freeargs(channel->buffers[channel->currSlot].transportHandle,
	        xdr_TransmitBuffer,
		channel->buffers[channel->currSlot].buffer)) {
	    elog(WARN, "recvTuple: can't free arguments");
	}
#else
	/* svc_freeargs doesn't seem to be thread-safe */
	if (channel->buffers[channel->currSlot].buffer->data.Data_len > 0) {
	    free(channel->buffers[channel->currSlot].buffer->data.Data_val);
	}
	free(channel->buffers[channel->currSlot].buffer);
#endif
	channel->needNewBuffer = true;
	++channel->freeBuffers;

	status = pthread_mutex_unlock(&channel->bufferLock);
	checkAndElog(status, "recvTuple: can't unlock bufferLock");

	status = pthread_cond_broadcast(&channel->bufferNotFull);
	checkAndElog(status, "recvTuple: can't broadcast bufferNotFull");
    }
}


/* 
 * closeRecvChannel --
 *	deallocates data associated with channel and deregisters 
 *	it from the global table
 */
void
closeRecvChannel(ExchRecvChannel channel)
{
    exchInChannelTab[channel->tableIndex] = NULL; /* deassign it */
    pfree(channel->senders);
    pfree(channel);
}


/* 
 * getNextTuple --
 *	returns the next tuple in the current buffer of the 
 *	specified channel
 */
static Pointer
getNextTuple(ExchRecvChannel channel)
{
    char *data;
    Pointer tuple;
    u_int len;

    data = channel->buffers[channel->currSlot].buffer->data.Data_val;
    len = channel->buffers[channel->currSlot].buffer->data.Data_len;
    if (channel->scanPos < len) {
	tuple = (Pointer) &data[channel->scanPos];
	channel->scanPos += ((HeapTuple) tuple)->t_len;
	return(tuple);
    }
    return(NULL);
}


void
ResetNetworkStats() {
    nBuffersReceived = 0;
    nBuffersSent = 0;
}
