/*-------------------------------------------------------------------------
 *
 * fragutils.c--
 *    Utility routines for transferring fragments.
 *
 * Copyright (c) 1995, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/fragutils.c,v 1.4 1996/01/28 00:35:40 searcher Exp $
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"

#include "access/heapam.h"
#include "catalog/catname.h"
#include "catalog/indexing.h"
#include "catalog/pg_class.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"
#include "utils/syscache.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "utils/elog.h"
#include "mariposa_rpc.h"
#include "rpc/rpcutils.h"
#include "miscadmin.h"		/* For LocalBlockSize */
#include "rpc/fragutils.h"

/* ----------------------------------------------------------------
 *	MakeFragmentRelationDesc
 *
 *	After doing system catalog inserts to establish the existence
 *	of the migrated/copied fragment, creates a relation descriptor
 *	for it and returns it.
 *
 * ----------------------------------------------------------------
 */
Relation
MakeFragmentRelationDesc(HeapTuple classTup, HeapTuple fragTup,
			 int natts, TupleDesc tupdesc)
{
    Relation		pgcldesc;
    Relation		rdesc;
    Form_pg_class	class_pg_class_tup;
    Form_pg_class	frag_pg_class_tup;
    Relation		idescs[Num_pg_class_indices];
    int			i;
    int			smgr;
    MemoryContext	oldcxt;

    class_pg_class_tup = (Form_pg_class)GETSTRUCT(classTup);
    frag_pg_class_tup = (Form_pg_class)GETSTRUCT(fragTup);

    /*
     * Insert into pg_class and pg_attributes.
     */
    pgcldesc = heap_openr(RelationRelationName);

    /*
     * If we are receiving the master fragment of a one fragment relation, or
     * we already have the pg_class tuple of the class, then we need only
     * deal with establishing fragTup in the system catalogs.
     */
    if (classTup->t_oid == fragTup->t_oid ||
	HeapTupleIsValid(SearchSysCacheTuple(RELOID,
					     (char *)
					     ObjectIdGetDatum(classTup->t_oid),
					     NULL, NULL, NULL)))
    {
	/*
	 * We've got the "class" tuple already, or the fragment tuple _is_
	 * the class tuple, only need to insert the fragment tuple into
	 * pg_class.
	 */
	heap_insert(pgcldesc, fragTup);	/* New oid?? */

	/*
	 * First, open the catalog indices and insert index tuples for
	 * the new relation.
	 */
	CatalogOpenIndices(Num_pg_class_indices,
			   Name_pg_class_indices, idescs);
	CatalogIndexInsert(idescs, Num_pg_class_indices, pgcldesc, fragTup);
	CatalogCloseIndices(Num_pg_class_indices, idescs);

	/*
	 * Now restore processing mode.
	 */
	SetProcessingMode(NormalProcessing);

	addNewRelationType(&frag_pg_class_tup->relname, fragTup->t_oid);

	AddNewAttributeTuples(fragTup->t_oid, natts, tupdesc);
    }
    else {
	heap_insert(pgcldesc, classTup);
	heap_insert(pgcldesc, fragTup);	/* New oid?? */

	/*
	 * First, open the catalog indices and insert index tuples for
	 * the new relation.
	 */
	CatalogOpenIndices(Num_pg_class_indices,
			   Name_pg_class_indices, idescs);
	CatalogIndexInsert(idescs, Num_pg_class_indices, pgcldesc, classTup);
	CatalogIndexInsert(idescs, Num_pg_class_indices, pgcldesc, fragTup);
	CatalogCloseIndices(Num_pg_class_indices, idescs);

	/*
	 * Now restore processing mode.
	 */
	SetProcessingMode(NormalProcessing);

	addNewRelationType(&class_pg_class_tup->relname, classTup->t_oid);
	addNewRelationType(&frag_pg_class_tup->relname, fragTup->t_oid);
	    
	AddNewAttributeTuples(classTup->t_oid, natts, tupdesc);
	AddNewAttributeTuples(fragTup->t_oid, natts, tupdesc);
    }

    heap_close(pgcldesc);

    /*
     * The pallocs need to persist at the end of transaction.
     */
    oldcxt = MemoryContextSwitchTo((MemoryContext)CacheCxt);
    
    /*
     * Create the new descriptor.
     */
    rdesc = (Relation)palloc(sizeof(RelationData));
    memset(rdesc, 0, sizeof(RelationData));

    rdesc->rd_att = tupdesc;

    RelationSetReferenceCount(rdesc, 1);
    
    rdesc->rd_rel = (Form_pg_class)palloc(sizeof *rdesc->rd_rel);
    memset(rdesc->rd_rel, 0, sizeof *rdesc->rd_rel);

    smgr = 0;	/* default = magdisk */
    namestrcpy(&(rdesc->rd_rel->relname), frag_pg_class_tup->relname.data); 
    rdesc->rd_rel->relkind = RELKIND_UNCATALOGED;
    rdesc->rd_rel->relnatts = natts;
    rdesc->rd_rel->relarch = 'h'; /* All Mariposa fragments are archived */
    rdesc->rd_rel->relsmgr = smgr;
    rdesc->rd_rel->relpartcmp = frag_pg_class_tup->relpartcmp;
    rdesc->rd_rel->relparthash = frag_pg_class_tup->relparthash;
    rdesc->rd_rel->relpartatt = frag_pg_class_tup->relpartatt;
    rdesc->rd_rel->relpartmode = frag_pg_class_tup->relpartmode;

    rdesc->rd_blocksz = LocalBlockSize;
    
    rdesc->rd_islocal = FALSE;
    rdesc->rd_blocksz = LocalBlockSize;

    rdesc->rd_id = fragTup->t_oid;

    /*
     * this must be the last thing before calling
     * RelationRegisterRelation because smgrcreate needs a
     * properly-formatted rdesc.
     */
    rdesc->rd_fd = (File)smgrcreate(smgr, rdesc);

    RelationRegisterRelation(rdesc);

    MemoryContextSwitchTo(oldcxt);

    return rdesc;
}
