/*-------------------------------------------------------------------------
 *
 * namesvc_be.c--
 *    name service related RPC calls in the exec./backend
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/namesvc_be.c,v 1.7 1995/11/18 01:32:52 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <pthread.h>
#include <rpc/types.h>
#include "utils/elog.h"
#include "mariposa_rpc.h" /* for procedure constants */
#include "rpc/rpcutils.h"

#include "rpc/call_exec.h" /* backend calling facility */

#include "access/heapam.h"
#include "catalog/catname.h"
#include "catalog/pg_class.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_host.h"
#include "catalog/indexing.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "rpc/nameserver.h"
#include "sitemgr/rpcerror.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"

int mariposaNamesvr = 0;
char *NamesvrSiteName = NULL;
int NamesvrSiteAddr;


/*****************************************************************************
 * 
 * Here's how name service is suppose to work. The NAME SERVER is a dedicated
 * process which handles global name service.  A sitemgr or client that needs
 * to access global catalog data can do so in two ways:  NSVC_GLOBALPROBE asks
 * for a specific request specified by the action tag, while NSVC_SEARCH is
 * a general interface similar to SearchSysCacheTuple.  (NSVC_SEARCH has not
 * been implemented yet.)
 *
 * A sitemgr can also find out local name service information by sending a
 * NSVC_LOCALPROBE call to one if its execs, with a similar interface to
 * NSVC_GLOBALPROBE.
 *
 * The NAME SERVER can update its catalogs using NSVC_POLL, which is serviced
 * by the sitemgr.  Right now, only the catalog to be updated can be 
 * specified, and the sitemgr will dump the entire class to the namesvr,
 * but eventually constraints will be available to restrict the poll.
 * 
 * NSVC_PROPOGATE is called to inform a sitemgr of a new catalog tuple; 
 * the sitemgr may forward this information to other sitemgrs.  
 * NSVC_UPDATENAMESVR is used to install this information in a NAME SERVER.
 *
 * Currently available actions for NSVC_GLOBALPROBE are:
 *
 * - DISCOVERCLASS : given a class OID, return the pg_class tuple, all the
 *   pg_attribute tuples and all the pg_fragment tuples for the class.
 *
 * Currently available actions for NSVC_LOCALPROBE are:
 *
 * - FINDLOCALFRAG : given a fragment oid, and copy/master status, is it local?
 *
 * - TAKEINVENTORY : gets all the local catalog info
 *
 *   ---------------  -------NSVC_POLL------->  -------------
 *   -             -                            -           - NSVC_PROPAGATE
 *   - NAME SERVER -  <--NSVC_UPDATENAMESVR---  -  SITEMGR  ------------->
 *   -             -  <---NSVC_GLOBALPROBE----  -           -          other
 *   ---------------  <-----NSVC_SEARCH-------  -------------         sitemgrs
 *         ^                                      ^  |
 *         |NSVC_GLOBALPROBE                      |  |
 *         |NSVC_SEARCH       NSVC_PROPAGATE      |  |NSVC_LOCALPROBE
 *         |             -------------------------   |
 *         |            /                            v
 *   ---------------   /                        -------------
 *   -   CLIENT/   -  /                         -           -
 *   -   BROKER    - /                          -   EXEC    -
 *   -             -                            -           -
 *   ---------------                            -------------
 *
 * When the NAME SERVER starts up, it issues NSVC_POLL to SITEMGRs 
 * for the fragments at the site managed by the SITEMGR. The SITEMGR in turn
 * issues NSVC_LOCALPROBE using the TAKEINVENTORY action to one of its EXEC's 
 * for a catalog of the fragments.
 *
 * For updates, the NAME SERVER can periodically probe the SITEMGRs
 * by sending a NSVC_POLL for an update. Alternatively, the EXEC or
 * SITEMGR can send the NAME SERVER an NSVC_UPDATENAMESVR.
 *
 */

/*****************************************************************************
 *  @@@ For Microsoft Demo 9/27, temporary name service calls @@@
 *
 *  FragIds
 *  handleCall_FRAGIDS
 *  Classes
 *  handleCall_CLASSES
 *****************************************************************************/

#define SENDBUFSIZE 7168		/* copied from exchange.c */

void
FragIds(FragDataBuffer *databuf)
{
    Oid			classoid;
    Relation		fragdesc;
    ScanKeyData		skey;
    HeapScanDesc	fragscan;
    HeapTuple		htup;
    char		*buffer;

    elog(NOTICE, "Entering FragIds.");

    /* Cough, Hack, Hack... Using numTuples to store class oid. */
    classoid = databuf->numTuples;
    buffer = databuf->data.Data_val;
    sprintf(buffer, "{ ");
    buffer += 1;

    ScanKeyEntryInitialize(&skey, 0x0, Anum_pg_fragment_fraclassid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(classoid));

    fragdesc = heap_openr((Name)Name_pg_fragment);
    if (!RelationIsValid(fragdesc)) {
	databuf->error = EXECUTION_ERROR;
	return;
    }

    fragscan = heap_beginscan(fragdesc,
			      0,
			      SelfTimeQual,
			      1, &skey);
    do {
	htup = heap_getnext(fragscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_fragment frag = (Form_pg_fragment)GETSTRUCT(htup);

	    sprintf(buffer, "{%d %d %d} ", frag->fralogicalid, frag->frastoreid,
		    frag->fralocation);
	    buffer += strlen(buffer);
	}
    } while (htup);
    heap_endscan(fragscan);
    heap_close(fragdesc);

    sprintf(buffer, "}");

    databuf->data.Data_len = strlen(databuf->data.Data_val) + 1;
    
    databuf->error = SUCCESS;

    elog(NOTICE, "Leaving FragIds.");
}

void
handleCall_FRAGIDS(void *args)
{
    char		**argv = (char **)args;
    int			*rpcarg = (int *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    FragDataBuffer	result;
    char		buffer[SENDBUFSIZE];

    elog(NOTICE, "Entering handleCall_FRAGIDS.");

    /* Cough, Hack, Hack... Using numTuples to store class oid. */
    result.numTuples = *rpcarg;
    result.data.Data_val = buffer;

    mainThreadExec((thdproc_t)FragIds, &result, NULL);

    if (!svc_sendreply(transp, xdr_FragDataBuffer, &result)) {
  	errorExit("handleCall_FRAGIDS: Can't send reply.\n");
    }

    free(args);
}

void
Classes(FragDataBuffer *databuf)
{
    Oid			classoid;
    Relation		classdesc;
    HeapScanDesc	classscan;
    HeapTuple		htup;
    char		*buffer;

    elog(NOTICE, "Entering Classes.");

    /* Cough, Hack, Hack... Using numTuples to store class oid. */
    classoid = databuf->numTuples;
    buffer = databuf->data.Data_val;
    sprintf(buffer, "{ ");
    buffer += 1;

    classdesc = heap_openr((Name)Name_pg_class);
    if (!RelationIsValid(classdesc)) {
	databuf->error = EXECUTION_ERROR;
	return;
    }

    classscan = heap_beginscan(classdesc,
			       0,
			       SelfTimeQual,
			       0, NULL);
    do {
	htup = heap_getnext(classscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    if (htup->t_oid > 2000) {
		Form_pg_class class = (Form_pg_class)GETSTRUCT(htup);

		sprintf(buffer, "{%d %s} ", htup->t_oid, class->relname.data);
		buffer += strlen(buffer);
	    }
	}
    } while (htup);
    heap_endscan(classscan);
    heap_close(classdesc);

    sprintf(buffer, "}");

    databuf->data.Data_len = strlen(databuf->data.Data_val) + 1;
    
    databuf->error = SUCCESS;

    elog(NOTICE, "Leaving Classes.");
}

void
handleCall_CLASSES(void *args)
{
    char		**argv = (char **)args;
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    FragDataBuffer	result;
    char		buffer[SENDBUFSIZE];

    elog(NOTICE, "Entering handleCall_CLASSES.");

    result.data.Data_val = buffer;

    mainThreadExec((thdproc_t)Classes, &result, NULL);

    if (!svc_sendreply(transp, xdr_FragDataBuffer, (char *)&result)) {
  	errorExit("handleCall_CLASSES: Can't send reply.\n");
    }

    free(args);
}

/*****************************************************************************
 *  NSVC_LOCALPROBE handler (EXEC side)
 *****************************************************************************/

/*
 * handleCall --
 *	thread start routine; does synchronous call to backend to process query
 */
void
handleCall_NSVC_LOCALPROBE(void *args)
{
    char **argv = (char **)args;
    NSVCProbeArg *rpcarg = (NSVCProbeArg *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    NSVCProbeResult *result;

    mainThreadExecInXact((thdproc_t) NsvcLocalProbe, rpcarg, &result);

    if (!svc_sendreply(transp, xdr_NSVCProbeResult, result)) {
  	errorExit("handleCall: can't send reply\n");
    }

    free(rpcarg);
    free(args);
}

/*****************************************************************************
 *  Routines to do the real work for the NSVC_LOCALPROBE
 *****************************************************************************/

static void findLocalFragment(Oid oid, NSVCFragmentSpec fspec,
			      NSVCProbeResult *result);
static void takeInventory(NSVCProbeResult *result);
static void schedGetFragInfo(Oid relid, NSVCProbeResult *result);

NSVCProbeResult *
NsvcLocalProbe(NSVCProbeArg *arg)
{
    NSVCProbeResult *result;

    result = (NSVCProbeResult*)palloc(sizeof(NSVCProbeResult));
    memset(result, 0, sizeof(NSVCProbeResult));

    switch(arg->action) {
    case NSVC_FINDLOCALFRAG:
	findLocalFragment(arg->oid, arg->fspec, result);
	break;
    case NSVC_TAKEINVENTORY:
	takeInventory(result);
	break;
    case NSVC_SCHEDGETFRAGINFO:
	schedGetFragInfo(arg->oid, result);
	break;
    case NSVC_FINDPGHOSTS:
	findPgHosts_be(&result->pgsites);
	break;
    }

    return result;
}

/* ----------------------------------------------------------------
 *   	findLocalFragment
 *
 *	Looks to see if fragment as at this site.
 *
 *	If fspec == ANY, looks for oid in the fralogicalid column of
 *	pg_fragment. If fspec == MASTER, looks for oid in the frastoreid
 *	column of pg_fragment. (ANY is for any copy, MASTER specifically
 *	for the master copy.)
 * ----------------------------------------------------------------
 */
static void
findLocalFragment(Oid oid, NSVCFragmentSpec fspec, NSVCProbeResult *result)
{
    Relation rd;
    HeapScanDesc sdesc;
    HeapTuple tup;
    ScanKeyData skey;
    AttrNumber att;
    int localSiteHostId = getMyHostId();
    
    att = (fspec == ANY) ? Anum_pg_fragment_fralogicalid
	                 : Anum_pg_fragment_frastoreid;

/* @@@ Is there a more efficient way to do this??	-Rex 6.20.1995 */
    ScanKeyEntryInitialize(&skey, 0x0, att,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(oid));
    rd = heap_openr((Name)Name_pg_fragment);
    sdesc = heap_beginscan(rd, false, NowTimeQual, 1, &skey);
    
    if (HeapTupleIsValid(tup = heap_getnext(sdesc, 0, NULL))) {
	Form_pg_fragment frag = (Form_pg_fragment)GETSTRUCT(tup);
	
	if (frag->fralocation == localSiteHostId) {
	    result->found = TRUE;
	}
	else {
	    result->found = FALSE;
	}
    } else {
	result->found = FALSE;
    }
    
    heap_endscan(sdesc);
    heap_close(rd);
}

static void takeInventory(NSVCProbeResult *result)
{
    Relation pgclass, pgfrag;
    HeapTuple classtup, fragtup;
    HeapScanDesc pgclassscan, pgfragscan;
    int ncl=0, maxcl, nfg, maxfg;
    TupleArray *userclasses;
    Tuple *classinfo;
    ScanKeyData skey[1];

    pgclass = heap_openr(RelationRelationName);
    if (!RelationIsValid(pgclass))
	elog(WARN, "cannot open pg_class");

    pgfrag = heap_openr(FragRelationName);
    if (!RelationIsValid(pgfrag))
	elog(WARN, "cannot open pg_fragment");

    maxcl = 8;
    userclasses = (TupleArray *)palloc(sizeof(TupleArray)*maxcl);
    pgclassscan = heap_beginscan(pgclass,
				 0,
				 SelfTimeQual,
				 0,
				 NULL);

    do {	
	classtup = heap_getnext(pgclassscan, 0, NULL);
	if (HeapTupleIsValid(classtup)) {
	    Form_pg_class pgclTup = (Form_pg_class)GETSTRUCT(classtup);
	    
	    /*
	     * leave out built-in relations, indices and system catalogs.
	     */
	    if (classtup->t_oid > 0xffffff &&
		pgclTup->relkind == RELKIND_RELATION &&
		!(pgclTup->relname.data[0] == 'p' &&
		  pgclTup->relname.data[1] == 'g' &&
		  pgclTup->relname.data[2] == '_')) {


		/* 
		 * we now have a user class; find out about the fragments
		 *
		 * classinfo[0] will hold the pg_class tuple,
		 * while classinfo[1, 2, 3, ...] will hold pg_fragment tuples
		 */
		
		nfg = 1;
		maxfg = 8;
		classinfo = (Tuple *)palloc(sizeof(Tuple) * maxfg);
		
		classinfo[0].Tuple_len = classtup->t_len;
		classinfo[0].Tuple_val = (char *)palloc(classtup->t_len);
		memmove(classinfo[0].Tuple_val, classtup, classtup->t_len);

		ScanKeyEntryInitialize(&skey[0],
				       0x0,
		                       Anum_pg_fragment_fraclassid,
				       ObjectIdEqualRegProcedure,
				       ObjectIdGetDatum(classtup->t_oid));

		pgfragscan = heap_beginscan(pgfrag,
					    0,
					    SelfTimeQual,
					    1,
					    (ScanKey)skey);

		do {
		    fragtup = heap_getnext(pgfragscan, 0, NULL);

		    if (HeapTupleIsValid(fragtup)) {
			if (nfg == maxfg) {
			    maxfg *= 2;
			    classinfo = (Tuple *)repalloc(classinfo,
					     sizeof(Tuple) * maxfg);
			}

			classinfo[nfg].Tuple_len = fragtup->t_len;
			classinfo[nfg].Tuple_val = (char *)
			    palloc(fragtup->t_len);
			memmove(classinfo[nfg].Tuple_val, 
				fragtup, fragtup->t_len);
			nfg ++;
		    }
		} while(fragtup);
		    
		    
		heap_endscan(pgfragscan);

		if (ncl==maxcl) {
		    maxcl *= 2;
		    userclasses = (TupleArray *)repalloc(userclasses,
					           sizeof(TupleArray)*maxcl);
		}
		userclasses[ncl].TupleArray_len = nfg;
		userclasses[ncl].TupleArray_val = classinfo;
		ncl++;
	    }
	}
    } while (classtup);
    heap_endscan(pgclassscan);

    heap_close(pgfrag);
    heap_close(pgclass);

    result->classes.TupleArrayArray_len = ncl;
    result->classes.TupleArrayArray_val = userclasses;
}

/* ---------------------------------------------------------------------
 * called by the scheduler to get information about the size of a fragment
 * and the platter on which it resides
 * ----------------------------------------
 */
static void
schedGetFragInfo(Oid relid,NSVCProbeResult *result)
{
    result->fragids.fragids_len = 2;
    result->fragids.fragids_val = (int *) palloc(sizeof(int)*2);
#ifdef HP_JUKEBOX
    result->fragids.fragids_val[0] = RelidGetRelSize(relid);
    result->fragids.fragids_val[1] = RelidGetPlatterId(relid);
#endif
}

/*
 * findPgHosts - wrapper for backend (there is another findPgHosts in 
 *    sitemgr.c for the sitemgr)
 */
SiteAddrList *
findPgHosts()
{
    SiteAddrList *sites;

    sites = calloc(1, sizeof(SiteAddrList));
    findPgHosts_be(sites);
    return sites;
}

findPgHosts_be(SiteAddrList *sites)
{
    int homeSiteHostId = getMyHostId();
    int i = 0;
    Relation hostdesc;
    HeapScanDesc hostscan;
    HeapTuple htup;
	
/* @@@ Must do something about this, sometime */
/* XXX Hardwired - must change. */
#define N_REMOTESITES 3

    sites->SiteAddrList_val =
	(SiteAddr *) calloc(N_REMOTESITES, sizeof(SiteAddr));

    hostdesc = heap_openr(Name_pg_host);
    if (!RelationIsValid(hostdesc))
	elog(WARN, "cannot open pg_host");
    hostscan = heap_beginscan(hostdesc, 0, SelfTimeQual, 0, NULL);
    do {
	htup = heap_getnext(hostscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_host host = (Form_pg_host)GETSTRUCT(htup);

	    /* Philosophy: getBidSites should not include our own site. */
	    if (i < N_REMOTESITES && host->hostid != homeSiteHostId) {
		sites->SiteAddrList_val[i].addr = host->address;
		sites->SiteAddrList_val[i].hostid = host->hostid;
		sites->SiteAddrList_val[i].port = SITEMGRPORT;
		i++;
	    }

	}
    } while (htup);
    heap_endscan(hostscan);
    heap_close(hostdesc);

    sites->SiteAddrList_len = N_REMOTESITES;

    return;
}

/*****************************************************************************
 *  NSVC_GLOBALPROBE handler (NAME SERVER)
 *****************************************************************************/

void
NSVC_GlobalProbe(void *args) 
{
    char **argv = (char **)args;
    NSVCProbeArg *arg = (NSVCProbeArg *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    NSVCProbeResult *result = (NSVCProbeResult *)
	palloc(sizeof(NSVCProbeResult));

    switch(arg->action) {
    case NSVC_DISCOVERCLASS:
	discoverClass((Oid)arg->oid, result);

	if (!svc_sendreply(transp, xdr_NSVCProbeResult, result)) {
	    elog(NOTICE, "NSVC_GlobalProbe: svc_sendreply failed\n");
	}
	break;

    default:
	elog(NOTICE, "NSVC_GlobalProbe: Name server received unrecognized request");

	result->error = UNRECOGNIZED_ACTION;
	if (!svc_sendreply(transp, xdr_NSVCProbeResult, result)) {
	    elog(NOTICE, "NSVC_GlobalProbe: svc_sendreply failed\n");
	}
    }	
}

/*****************************************************************************
 * NSVC_UPDATENAMESVR handler (NAME SERVER)
 *****************************************************************************/

void
NSVC_UpdateNamesvr(void *args)
{
    char **argv = (char **)args;
    NSVCItem *item = (NSVCItem *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    RPCErrorCode result;
    HeapTuple tuple = (HeapTuple)item->tuple.Tuple_val;

    updateNamesvr((Oid)item->classId, tuple, &result);

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	elog(NOTICE, "NSVC_UpdateNamesvr: svc_sendreply failed\n");
    }
}

/*****************************************************************************
 * wrapper for NSVC_UPDATENAMESVR (client side)
 *****************************************************************************/

RPCErrorCode
callNsvcUpdateNamesvr(NSVCItem *item)
{
    RPCErrorCode result;
    RpcConn *rpcc;

    if (NamesvrSiteName == NULL) {
	struct hostent *siteInfo;
	char *siteName = getenv(NSVRSITE_VARNAME);

	if (siteName == NULL) {
	    elog(WARN, "callNsvcUpdateNamesvr: %s not defined", 
		 NSVRSITE_VARNAME);
	}
	
	NamesvrSiteName = (char *)malloc(strlen(siteName) + 1);
	strcpy(NamesvrSiteName, siteName);
	siteInfo = gethostbyname(siteName);
	(void) memmove((char *) &NamesvrSiteAddr,
		      siteInfo->h_addr, siteInfo->h_length);
    }

    rpcc = Rpc_connectNamesvr(NamesvrSiteAddr,
				"callNsvcUpdateNamesvr",
				RPC_ONERR_RETURN);
    if (rpcc == NULL) {
	elog(NOTICE, "callNsvcUpdateNamesvr: couldn't create client to call %s",
	     NamesvrSiteName);
    }

    if (!Rpc_call(rpcc, RPC_NSVC_UPDATENAMESVR, item, &result)) {
	elog(NOTICE, "callNsvcUpdateNamesvr: couldn't call %s", 
	     NamesvrSiteName); 
    }

    return result;
}

/*****************************************************************************
 * wrapper for NSVC_GLOBALPROBE (client side)
 *****************************************************************************/

NSVCProbeResult *
callNsvcGlobalProbe(NSVCProbeArg *probeArg)
{
    NSVCProbeResult *result = (NSVCProbeResult *)
	palloc(sizeof(NSVCProbeResult));
    RpcConn *rpcc;

    if (NamesvrSiteName == NULL) {
	struct hostent *siteInfo;
	char *siteName = getenv(NSVRSITE_VARNAME);
	if (siteName == NULL) {
	    elog(WARN, "callNsvcGlobalProbe: %s not defined", 
		 NSVRSITE_VARNAME);
	}
	
	NamesvrSiteName = (char *)malloc(strlen(siteName) + 1);
	strcpy(NamesvrSiteName, siteName);
	siteInfo = gethostbyname(siteName);
	(void) memmove((char *) &NamesvrSiteAddr,
		      siteInfo->h_addr, siteInfo->h_length);
    }

    rpcc = Rpc_connectNamesvr(NamesvrSiteAddr,
				"callNsvcGlobalProbe",
				RPC_ONERR_RETURN);
    if (rpcc == NULL) {
	elog(NOTICE, "callNsvcGlobalProbe: couldn't create client to call %s",
	     NamesvrSiteName);
    }

    if (!Rpc_call(rpcc,
		  RPC_NSVC_GLOBALPROBE,
		  probeArg,
		  result)) {
	elog(NOTICE, "callNsvcGlobalProbe: couldn't call %s", 
	     NamesvrSiteName); 
    }

    return result;
}
