/*-------------------------------------------------------------------------
 *
 * rexec.c --
 *    interface for remote execs to notify their parents of readiness
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/rexec.c,v 1.5 1996/02/06 18:48:38 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <pthread.h>
#include <rpc/types.h>
#include <stdio.h>

#include "c.h"
#include "postgres.h"

#include "rpctypes.h"
#include "rpc/rpcutils.h"
#include "rpc/rexec.h"
#include "utils/elog.h"
#include "nodes/pg_list.h"
#include "sitemgr/rpcerror.h"

extern unsigned short tcpPort;

static bool condVarInitialized = FALSE;

static pthread_mutex_t RExecStatusLock;
static pthread_cond_t RExecStatus;
static List *RExecSiteAddrList;


/*
 * RExecInit -
 *    Ensures that all synchronization variables are initialized and
 *    that the safeguard variable (which protects from "spurious"
 *    wakeups, see pthread_cond_wait(3) for details) has been set to
 *    the waiting value.
 *
 *    This must be called BEFORE we send the RPC that starts the
 *    remote exec.
 */
void
RExecInit()
{
    if (!condVarInitialized) {
	pthread_mutex_init(&RExecStatusLock, pthread_mutexattr_default);
	pthread_cond_init(&RExecStatus, pthread_condattr_default);
	condVarInitialized = TRUE;
	RExecSiteAddrList = NIL;
    }
}

/*
 * RExecWait -
 *    The main thread of the local exec blocks, waiting for the remote
 *    exec to call RExecSignalReady.  When the thread is unblocked,
 *    the address/port of the remote exec will have been placed in the
 *    static variables RExecAddr and RExecPort.
 */
void
RExecWait(int *addr, int *port, int *hostid)
{
    int status;
    List *lp;
    SiteAddr *sap;

    elog(NOTICE, "RExecWait: Waiting...");

    /* we wait until the remote exec calls back */
    status = pthread_mutex_lock(&RExecStatusLock);
    while (RExecSiteAddrList == NIL) {
	status  = pthread_cond_wait(&RExecStatus, &RExecStatusLock);
    }

    sap = lfirst(RExecSiteAddrList);
    lp = RExecSiteAddrList;
    RExecSiteAddrList = lnext(RExecSiteAddrList);

    status = pthread_mutex_unlock(&RExecStatusLock);
    
    if (addr) {
	*addr = sap->addr;
    }
    if (port) {
	*port = htons(sap->port);
    }
    if (hostid) {
	*hostid = sap->hostid;
    }
    pfree((void *) lp);
    pfree((void *) sap);

    elog(NOTICE, "RExecWait: notified: <%d,%d,%d>",
	 sap->addr, htons(sap->port), sap->hostid);
}

/*
 * RExecSignalReady -
 *    Remote exec makes an RPC call to notify the local exec that it
 *    is ready for the operation.  Assumes local exec is waiting in
 *    RExecWait.
 */
void *
RExecSignalReady(SiteAddr *addr)
{
    RpcConn *rpcc;
    SiteAddr siteaddr;

    rpcc = Rpc_connectExec(addr->addr, htons(addr->port), 
			   "RExecSignalReady", RPC_ONERR_WARN);
    
    siteaddr.addr = getMyAddr();
    siteaddr.port = tcpPort;
    siteaddr.hostid = getMyHostId();

    Rpc_call(rpcc, RPC_FRAGSTARTXFER, &siteaddr, NULL);
    
    Rpc_disconnect(rpcc);

    elog(NOTICE, "RExecSignalReady: notifying: <%d,%d,%d>",
	 siteaddr.addr, htons(siteaddr.port), siteaddr.hostid);

    return (void *)SUCCESS;
}

/*
 * RExecSignalReadyHandler -
 *    Local exec, waiting in RExecWait, receives the RPC call; this
 *    handler frees the main thread from the condition variable.
 */
void 
RExecSignalReadyHandler(struct svc_req *request, SVCXPRT *transportHandle)
{
    SiteAddr *arg;
    int status;

    arg = (SiteAddr *) palloc(sizeof(SiteAddr));
    if (!svc_getargs(transportHandle, xdr_SiteAddr, arg)) {
	errorExit("RExecSignalReadyHandler: can't decode arguments\n");
    }

    /* small job, just do it */
    status = pthread_mutex_lock(&RExecStatusLock);
    RExecSiteAddrList = lcons((void *) arg, RExecSiteAddrList);
    status = pthread_cond_signal(&RExecStatus);
    status = pthread_mutex_unlock(&RExecStatusLock);

    if (!svc_sendreply(transportHandle, xdr_void, NULL)) {
  	errorExit("RExecSignalReadyHandler: can't send reply\n");
    }

    elog(NOTICE, "RExecSignalReadyHandler: notified: <%d,%d,%d>",
	 arg->addr, htons(arg->port), arg->hostid);
}
