/*-------------------------------------------------------------------------
 *
 * rpc_server.c--
 *    RPC server routines for the backend and exec modules.
 *    Not to be confused with sitemgr/server.c, which performs
 *    similar services for the sitemgr.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/rpc_server.c,v 1.43 1996/06/14 20:23:13 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <rpc/rpc.h>
#include <pthread.h>
#include <netdb.h>
#include <sys/socket.h>
#include <string.h>
#include "miscadmin.h" 		/* for IsMariposaExec */
#include "exchange.h" 		/* for handler */
#include "mariposa_rpc.h" 	/* for procedure constants */
#include "rpc/rpcutils.h"

#include "sitemgr/sitemgr.h"	/* for BackendReady structures */
#include "utils/elog.h"
#include "utils/nameserver.h"
#include "tcop/broker.h"
#include "tcop/brokerLong.h"
#include "tcop/pquery.h"
#include "nameserver.h"
#include "databroker/databroker.h"

typedef void *(*ThreadStartRoutine)(void *);


static void rpcBackendHandler(struct svc_req *request,
    SVCXPRT *transp);

unsigned short udpPort;
unsigned short tcpPort;

void namesvr_svc_setup(void);
void namesvr_deregister(void);
extern void broker_svc_setup(void);
extern void databroker_svc_setup(void);
extern void InitExecVars(void);

/*
 * rpcServerLoop --
 *	starts the server loop (svc_run()) in a thread
 */
void
rpcServerLoop(void)
{
    SVCXPRT *transportHandle;
    int status;
    pthread_t thread;
    int pid;
    BackendBuffer backendBuf;
    RPCErrorCode error;
    char hostname[50];
    extern int IsQueryBrokerBackend;

    if (mariposaNamesvr) {
	namesvr_svc_setup();
/*     } else if (IsQueryBrokerBackend) { */
/* 	broker_svc_setup(); */
/*     } else if (IsDataBrokerBackend) { */
/*         databroker_svc_setup(); */
    } else {
      InitExecVars();
	/* create UDP and TCP transport handles */
	transportHandle = svcudp_create(RPC_ANYSOCK);
	if (transportHandle == NULL) {
	    elog(WARN, "rpcServerLoop: cannot create UPD transport handle");
	}
	
	/* this doesn't register the program with the portmapper (proto = 0) */
	if (!svc_register(transportHandle, EXECPROG, EXECVERS, rpcBackendHandler,
			  0)) {
	    elog(WARN, "rpcServerLoop: cannot register UPD transport handle");
	}
	udpPort = transportHandle->xp_port;

	transportHandle = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transportHandle == NULL) {
	    elog(WARN, "rpcServerLoop: cannot create TCP transport handle");
	}

	if (!svc_register(transportHandle, EXECPROG, EXECVERS, rpcBackendHandler,
			  0)) {
	    elog(WARN, "rpcServerLoop: cannot register TCP transport handle");
	}

	tcpPort = transportHandle->xp_port;

#ifdef RPCDEBUG
	fprintf(stderr, "EXECPROG TCP port = %d (%d)\n",
		tcpPort, htons(tcpPort));
#endif	
    }

    if (IsMariposaExec) {
	RpcConn *rpcc;
	int localSiteAddr = getMyAddr();
	
	/* at this point, notify the sitemgr we are ready */
	/* notify local site mgr of our UDP and TCP port numbers */
	backendBuf.pid = getpid();
	backendBuf.udpPort.port = udpPort;
	backendBuf.udpPort.addr = localSiteAddr;
	backendBuf.udpPort.hostid = MyHostId;
	backendBuf.tcpPort.port = tcpPort;
	backendBuf.tcpPort.addr = localSiteAddr;
	backendBuf.tcpPort.hostid = MyHostId;
	gethostname(hostname, 50);
#if 0
	/* We add hostid to program number so we contact the correct sitemgr */
	callrpc(hostname, SITEMGRPROG + getMyHostId(), SITEMGRVERS,
		BACKENDREADY, xdr_BackendBuffer, &backendBuf,
		xdr_RPCErrorCode, &error);
#endif
	rpcc =
	    Rpc_connectSitemgr(getMyHostId(), 0,
			       "rpcServerLoop", RPC_ONERR_WARN);
	Rpc_call(rpcc, RPC_BACKENDREADY, &backendBuf, &error);
	Rpc_disconnect(rpcc);
    }

    /* start server loop and return */
    if (pthread_create(&thread, pthread_attr_default,
		       (ThreadStartRoutine) svc_run, NULL) == -1) {
	elog(WARN, "rpcServerLoop: can't start svc_run()");
    }
    if (pthread_detach(&thread) == -1) {
	elog(WARN, "rpcServerLoop: can't detach");
    }
    return;
}


static void
rpcBackendHandler(struct svc_req *request, SVCXPRT *transp)
{
    switch (request->rq_proc) {
    case NULLPROC:
        if (!svc_sendreply(transp, xdr_void, NULL)) {
	    fprintf(stderr,
	        "rpcBackendHandler: can't send reply to NULLPROC\n");
	}
        return;
    
    case TRANSMITDATA:
	rpcAsyncHandleRequest(transp,
			      xdr_TransmitBuffer,
			      sizeof(TransmitBuffer),
			      receiveData,
			      "receiveData");
	return;
	
    case PROCESSQUERY:
	rpcAsyncHandleRequest(transp,
			      xdr_ProcessQueryArg,
			      sizeof(ProcessQueryArg),
			      handleCall,
			      "handleCall");
	return;
	
    case FRAGXSETUP:
	rpcAsyncHandleRequest(transp,
			      xdr_FragXArg,
			      sizeof(FragXArg),
			      handleCall_FRAGXSETUP,
			      "handleCall_FRAGXSETUP");
	return;
	
    case ESTABLISHCOPY:
	rpcAsyncHandleRequest(transp,
			      xdr_CopySetupInfo,
			      sizeof(CopySetupInfo),
			      handleCall_ESTABLISHCOPY,
			      "handleCall_ESTABLISHCOPY");
	return;
	
    case CP_SENDUPDATESTREAM:
	rpcAsyncHandleRequest(transp,
			      xdr_UpdateContractInfo,
			      sizeof(UpdateContractInfo),
			      handleCall_CP_SENDUPDATESTREAM,
			      "handleCall_CP_SENDUPDATESTREAM");
	return;
	
    case CP_INITIALIZEUPDATESTREAMS:
	rpcAsyncHandleRequest(transp,
			      xdr_void,
			      0,
			      handleCall_CP_INITIALIZEUPDATESTREAMS,
			      "handleCall_CP_INITIALIZEUPDATESTREAMS");
	return;
	
    case CP_PUTUPDATE:
	/* Do this synchronously. see FRAGPUTDATA below for explanation. */
	rpcSyncHandleRequest(transp,
			     xdr_TransferDataBuffer,
			     sizeof(TransferDataBuffer),
			     handleCall_CP_PUTUPDATE,
			     "handleCall_CP_PUTUPDATE");
	return;
	
    case CANCELCONTRACT:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      handleCall_CANCELCONTRACT,
			      "handleCall_CANCELCONTRACT");
	return;
    
    case RECEIVERUNBLOCKED:
	rpcAsyncHandleRequest(transp,
			      xdr_UnblockBuffer,
			      sizeof(UnblockBuffer),
			      ReceiverUnblocked,
			      "ReceiverUnblocked");
	return;

    case EXTRACTPREPARATIONS:
	rpcAsyncHandleRequest(transp,
			      xdr_FragXArg,
			      sizeof(FragXArg),
			      handleCall_EXTRACTPREPARATIONS,
			      "handleCall_EXTRACTPREPARATIONS");
	return;

    case TWOPCSTART:
	rpcAsyncHandleRequest(transp,
			      xdr_SiteAddr,
			      sizeof(SiteAddr),
			      handleCall_TWOPCSTART,
			      "handleCall_TWOPCSTART");
	return;

    case TWOPCPREPARE:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      handleCall_TWOPCPREPARE,
			      "handleCall_TWOPCPREPARE");
	return;
	
    case TWOPCEND:
	rpcAsyncHandleRequest(transp,
			      xdr_EndXactDirective,
			      sizeof(EndXactDirective),
			      handleCall_TWOPCEND,
			      "handleCall_TWOPCEND");
	return;
	
    case FRAGIDS:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      handleCall_FRAGIDS,
			      "handleCall_FRAGIDS");
	break;
	
    case CLASSES:
	rpcAsyncHandleRequest(transp,
			      xdr_void,
			      0,
			      handleCall_CLASSES,
			      "handleCall_CLASSES");
	break;
	
    case SYNCPUTFRAGMENT:
	rpcAsyncHandleRequest(transp,
			      xdr_SyncFragmentArg,
			      sizeof(SyncFragmentArg),
			      handleCall_SYNCPUTFRAGMENT,
			      "handleCall_SYNCPUTFRAGMENT");
	return;
	
    case SYNCGETFRAGMENT:
	rpcAsyncHandleRequest(transp,
			      xdr_SyncFragmentArg,
			      sizeof(SyncFragmentArg),
			      handleCall_SYNCGETFRAGMENT,
			      "handleCall_SYNCGETFRAGMENT");
	return;

	
    case BROKERRECEIVEBID:
	rpcAsyncHandleRequest(transp,
			      xdr_BidBuffer,
			      sizeof(BidBuffer),
			      BrokerReceiveBid,
			      "BrokerReceiveBid");
	break;

    case SUBCONTRACTBROKER:
	rpcAsyncHandleRequest(transp,
			      xdr_BrokerArg,
			      sizeof(BrokerArg),
			      SubContractBroker,
			      "SubContractBroker");
	break;
	
    case FRAGSTARTXFER:	
	/* small job, don't need to start a thread */
	RExecSignalReadyHandler(request, transp);
	break;

    case FRAGPUTCAT:
	rpcAsyncHandleRequest(transp,
			      xdr_FragCatalogInfo,
			      sizeof(FragCatalogInfo),
			      handleCall_FRAGPUTCAT,
			      "handleCall_FRAGPUTCAT");
	break;
	
    case FRAGPUTDATA:
	/*
	 * don't start a thread or we'll run out of threads when moving
	 * large tables. (This will be called many times and we won't
	 * be able to create new threads at some point. This problem occurs
	 * on OSF/1 2.1) Besides, we don't need to do this asynchronously.
	 */
	rpcSyncHandleRequest(transp,
			     xdr_FragDataBuffer,
			     sizeof(FragDataBuffer),
			     handleCall_FRAGPUTDATA,
			     "handleCall_FRAGPUTDATA");
	break;
	
    case FRAGGETCAT:
	rpcAsyncHandleRequest(transp,
			      xdr_FragXArg,
			      sizeof(FragXArg),
			      handleCall_FRAGGETCAT,
			      "handleCall_FRAGGETCAT");
	break;
	
    case FRAGGETDATA:
	/* Do this synchronously. see FRAGPUTDATA above for explanation. */
	rpcSyncHandleRequest(transp,
			     xdr_FragXArg,
			     sizeof(FragXArg),
			     handleCall_FRAGGETDATA,
			     "handleCall_FRAGGETDATA");
	break;
	
    case OPT_GETCOST:
	rpcAsyncHandleRequest(transp,
			      xdr_OptimizerArg,
			      sizeof(OptimizerArg),
			      handleCall_OPT_GETCOST,
			      "handleCall_OPT_GETCOST");
	break;

    case NSVC_LOOKUPCLASS:
	rpcAsyncHandleRequest(transp,
			      xdr_string,
			      sizeof(char *),
			      handleCall_NSVC_LOOKUPCLASS,
			      "handleCall_NSVC_LOOKUPCLASS");
	break;

    case NSVC_LOOKUPFRAGS:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      handleCall_NSVC_LOOKUPFRAGS,
			      "handleCall_NSVC_LOOKUPFRAGS");
	break;

    case NSVC_LOCALPROBE:
	rpcAsyncHandleRequest(transp,
			      xdr_NSVCProbeArg,
			      sizeof(NSVCProbeArg),
			      handleCall_NSVC_LOCALPROBE,
			      "handleCall_NSVC_LOCALPROBE");
	break;
	
    case SCHEDFRAGMENT:
	rpcAsyncHandleRequest(transp,
			      xdr_SchedFragmentArg,
			      sizeof(SchedFragmentArg),
			      handleCall_SCHEDFRAGMENT,
			      "handleCall_SCHEDFRAGMENT");
	break;
	
    default:
        svcerr_noproc(transp);
        return;
    }
}

/****************************************************************************
 *
 * RPC service routines for name server
 *
 ****************************************************************************/


static void namesvrprog_1();

/*
 * namesvr_svc_setup -
 *    Registers name server as rpc server.
 */

void
namesvr_svc_setup(void)
{
    register SVCXPRT *transp;
    int tcpsock;
    struct sockaddr_in *tcpsockaddr = (struct sockaddr_in *)
	                              malloc(sizeof(struct sockaddr));

    (void) pmap_unset(NAMESVRPROG, NAMESVRVERS);

    /* For now mariposa only uses tcp, but we create a udp service
     * anyway in case we change our mind later.
     */
    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
	elog(FATAL, "namesvr_svc_setup: cannot create udp service\n");
    }
    if (!svc_register(transp, NAMESVRPROG, NAMESVRVERS, namesvrprog_1, 
		      IPPROTO_UDP)) {
	elog(FATAL, "namesvr_svc_setup: unable to register (NAMESVRPROG, NAMESVRVERS, udp)\n");
    }

    if ((tcpsock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	elog(FATAL, "namesvr_svc_setup: cannot create tcp socket\n");
    }

    tcpsockaddr->sin_port = 0; /* portmapper will assign port */
    tcpsockaddr->sin_addr.s_addr = gethostid();
    if (bind(tcpsock, tcpsockaddr, sizeof(struct sockaddr_in)) == -1) {
	elog(FATAL, 
		"namesvr_svc_setup: cannot bind tcp socket to nameserver port\n");
    }

    transp = svctcp_create(tcpsock, 0, 0);
    if (transp == NULL) {
	elog(FATAL, "namesvr_svc_setup: cannot create tcp service\n");
    }
    if (!svc_register(transp, NAMESVRPROG, NAMESVRVERS, namesvrprog_1, 
		      IPPROTO_TCP)) {
	elog(FATAL, 
		"namesvr_svc_setup: unable to register (NAMESVRPROG, NAMESVRRVERS, tcp)\n");
	exit(1);
    }

    atexit(namesvr_deregister);
}

/*
 * namesvr_deregister --
 *	Performs cleanup when the nameserver calls exit(3).
 *	- deregister RPC program/version/port mapping
 */

void
namesvr_deregister()
{
    (void) svc_unregister(NAMESVRPROG, NAMESVRVERS);
}

/*
 * namesvrprog_1 -
 *    Handler for RPC requests, called by svc_run.
 */

static void
namesvrprog_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
    switch (rqstp->rq_proc) {
    case NULLPROC:
	(void) svc_sendreply(transp, xdr_void, (char *)NULL);
	return;

    case NSVC_GLOBALPROBE: {
	NSVCProbeArg *arg;
	NSVCProbeResult result;

	arg = (NSVCProbeArg *) calloc(1, sizeof(*arg));
	if (arg == NULL) {
	    elog(WARN, "namesvcprog_1 : cannot allocate arguments to NSVC_GlobalProbe\n");
	    exit(1);
	}
	if (!svc_getargs(transp, xdr_NSVCProbeArg, arg)) {
	    elog(WARN, "namesvcprog_1 : cannot get arguments to NSVC_GlobalProbe\n");
	    exit(1);
	}
	RPCThreadStart(arg, transp, NSVC_GlobalProbe);

	    return;
	}

    case NSVC_UPDATENAMESVR: {
	NSVCItem *arg;
	RPCErrorCode result;

	arg = (NSVCItem *) calloc(1, sizeof(*arg));
	if (arg == NULL) {
	    elog(WARN, "namesvcprog_1 : cannot allocate arguments to NSVC_UpdateNamesvr\n");
	    exit(1);
	}
	if (!svc_getargs(transp, xdr_NSVCItem, arg)) {
	    elog(WARN, "namesvcprog_1 : cannot get arguments to NSVC_UpdateNamesvr\n");
	    exit(1);
	}
	RPCThreadStart(arg, transp, NSVC_UpdateNamesvr);

	    return;
	}

    case NSVC_SEARCH: 
	/* NOT IMPLEMENTED YET */

    default:
	svcerr_noproc(transp);
	return;
    }

    /* NOTREACHED */
}

/*****************************************************************************
 *
 *****************************************************************************/

static void rpcBrokerHandler(struct svc_req *request, SVCXPRT *transp);

/*
 * broker_deregister --
 *	Performs cleanup when the broker calls exit(3).
 *	- deregister RPC program/version/port mapping
 */
#if 0
static void
broker_deregister()
{
    /* We add hostid to program number so we contact the correct sitemgr */
    (void) svc_unregister(BROKERPROG + getMyHostId(), BROKERVERS);
}
#endif

/*
 * broker_svc_setup -
 *    registers rpc services.
 */

#if 0
void
broker_svc_setup(void)
{
    register SVCXPRT *transp;
    int tcpsock;
    struct sockaddr_in tcpsockaddr;


    /* We add hostid to program number so we contact the correct broker */
    (void) pmap_unset(BROKERPROG + getMyHostId(), BROKERVERS);

    /* For now mariposa only uses tcp, but we create a udp service
     * anyway in case we change our mind later.
     */
    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
	elog(WARN, "broker_svc_main: cannot create udp service\n");
    }

    
    if (!svc_register(transp, BROKERPROG + getMyHostId(), BROKERVERS,
		      rpcBrokerHandler, IPPROTO_UDP))
    {
	elog(WARN, "broker_svc_main: unable to register (BROKERPROG, BROKERVERS, udp)\n");
    }
    udpPort = transp->xp_port;

    if ((tcpsock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	elog(WARN, "broker_svc_main: cannot create tcp socket\n");
    }

    tcpsockaddr.sin_port = 0;
    tcpsockaddr.sin_addr.s_addr = gethostid();
    if (bind(tcpsock, &tcpsockaddr, sizeof(struct sockaddr_in)) == -1) {
	elog(WARN, "broker_svc_main: cannot bind tcp socket to broker port\n");
    }

    transp = svctcp_create(tcpsock, 0, 0);
    if (transp == NULL) {
	elog(WARN, "broker_svc_main: cannot create tcp service\n");
    }

    if (!svc_register(transp, BROKERPROG + getMyHostId(), BROKERVERS,
		      rpcBrokerHandler, IPPROTO_TCP))
    {
	elog(WARN, "broker_svc_main: unable to register (BROKERPROG, BROKERVERS, tcp)\n");
    }
    tcpPort = transp->xp_port;

    atexit(broker_deregister);
}

#endif

#if 0
/*
 * JPS - if want to call backend from inside broker, copy stuff
 * from rpcBackendHandler to here.
 * This is called from the rpc server loop, which handles incoming
 * RPC calls from the sitemgr.
 */
static void
rpcBrokerHandler(struct svc_req *request, SVCXPRT *transp)
{
    switch (request->rq_proc) {
    case NULLPROC:
        if (!svc_sendreply(transp, xdr_void, NULL)) {
	    fprintf(stderr,
	        "rpcBrokerHandler: can't send reply to NULLPROC\n");
	}
        break;
    
    case BROKER:
	rpcAsyncHandleRequest(transp,
			      xdr_BrokerArg,
			      sizeof(BrokerArg),
			      Broker,
			      "Broker");
	break;
	
    case BROKERRECEIVEBID:
	rpcAsyncHandleRequest(transp,
			      xdr_BidBuffer,
			      sizeof(BidBuffer),
			      BrokerReceiveBid,
			      "BrokerReceiveBid");
	break;
	
/*     case SUBCONTRACTWON: */
/* 	rpcAsyncHandleRequest(transp, */
/* 			      xdr_int, */
/* 			      sizeof(int), */
/* 			      SubcontractWon, */
/* 			      "SubcontractWon"); */
/* 	break; */
	
/*     case SUBCONTRACTLOST: */
/* 	rpcAsyncHandleRequest(transp, */
/* 			      xdr_int, */
/* 			      sizeof(int), */
/* 			      SubcontractLost, */
/* 			      "SubcontractLost"); */
/* 	break; */
	
    default:
        svcerr_noproc(transp);
        break;
    }

}
#endif

/*****************************************************************************
 *                         Data Broker 
 *****************************************************************************
 */

static void rpcDataBrokerHandler(struct svc_req *request, SVCXPRT *transp);

/*
 * databroker_deregister --
 *	Performs cleanup when the databroker calls exit(3).
 *	- deregister RPC program/version/port mapping
 */
static void
databroker_deregister()
{
    /* We add hostid to program number so we contact the correct sitemgr */
/*     (void) svc_unregister(DATABROKERPROG + getMyHostId(), DATABROKERVERS); */
}

/*
 * databroker_svc_setup -
 *    registers rpc services.
 */
void
databroker_svc_setup(void)
{
    register SVCXPRT *transp;
    int tcpsock;
    struct sockaddr_in tcpsockaddr;


    /* We add hostid to program number so we contact the correct broker */
/*     (void) pmap_unset(DATABROKERPROG + getMyHostId(), DATABROKERVERS); */

    /* For now mariposa only uses tcp, but we create a udp service
     * anyway in case we change our mind later.
     */
    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
	elog(WARN, "databroker_svc_main: cannot create udp service\n");
    }

#ifdef RPCDEBUG
/*     fprintf(stderr, "DATABROKERPROG: %d registered\n", */
/* 	    DATABROKERPROG + getMyHostId()); */
#endif /* RPCDEBUG */
    
/*     if (!svc_register(transp, DATABROKERPROG + getMyHostId(), DATABROKERVERS, */
/* 		      rpcDataBrokerHandler, IPPROTO_UDP)) */
/*     { */
/* 	elog(WARN, "databroker_svc_main: unable to register (BROKERPROG, BROKERVERS, udp)\n"); */
/*     } */
    udpPort = transp->xp_port;

    if ((tcpsock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	elog(WARN, "databroker_svc_main: cannot create tcp socket\n");
    }

    tcpsockaddr.sin_port = 0;
    tcpsockaddr.sin_addr.s_addr = gethostid();
    if (bind(tcpsock, &tcpsockaddr, sizeof(struct sockaddr_in)) == -1) {
	elog(WARN, "databroker_svc_main: cannot bind tcp socket to broker port\n");
    }

    transp = svctcp_create(tcpsock, 0, 0);
    if (transp == NULL) {
	elog(WARN, "databroker_svc_main: cannot create tcp service\n");
    }

/*     if (!svc_register(transp, DATABROKERPROG + getMyHostId(), DATABROKERVERS, */
/* 		      rpcDataBrokerHandler, IPPROTO_TCP)) */
/*     { */
/* 	elog(WARN, "databroker_svc_main: unable to register (DATABROKERPROG, DATABROKERVERS, tcp)\n"); */
/*     } */
    tcpPort = transp->xp_port;

    atexit(databroker_deregister);
}

/*
 * XXXRHPXXX - This is just a dummy function for now.  The details need
 * to be filled in when the RPC interface to the Data Broker becomes
 * more clearly defined.
 *
 * JPS - if we want to call backend from inside broker, copy stuff
 * from rpcBackendHandler to here.
 * This is called from the rpc server loop, which handles incoming
 * RPC calls from the sitemgr.
 */
static void
rpcDataBrokerHandler(struct svc_req *request, SVCXPRT *transp)
{
    switch (request->rq_proc) {
    case NULLPROC:
        if (!svc_sendreply(transp, xdr_void, NULL)) {
	    fprintf(stderr,
	        "rpcDataBrokerHandler: can't send reply to NULLPROC\n");
	}
        break;
    
/*     case DATABROKER: */
/* 	rpcAsyncHandleRequest(transp, */
/* 			      xdr_int, */
/* 			      sizeof(int), */
/* 			      DataBroker, */
/* 			      "DataBroker"); */
/* 	break; */
	
    default:
        svcerr_noproc(transp);
        break;
    }
}


