/*-------------------------------------------------------------------------
 *
 * rpcutils.c--
 *    RPC utilities
 *
 * Copyright (c) 1994-5, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/rpcutils.c,v 1.17 1996/05/31 17:14:17 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/param.h>			/* for MAXPATHLEN */
#include <rpc/types.h>
#include <pthread.h>
#include "miscadmin.h"                  /* for MyHostId */
#include "mariposa_rpc.h"
#include "sitemgr/sitemgr.h"
#include "access/xact.h"
#include "rpc/rpcutils.h"
#include "utils/elog.h"
#include "rpc/call_exec.h"


int localSiteAddr;
int localSiteHostId;

#define RPCDEBUG

static
void
rpcDebug(char *fmt, ...)
{
#ifdef RPCDEBUG
    va_list ap;
    static int pid = -1;
    char s[BUFSIZ];
    pthread_t thisThread = pthread_self();
    int threadID = (thisThread.field1 % 100000);

    va_start(ap, fmt);
    vsprintf(s, fmt, ap);
    va_end(ap);

    fprintf(stderr, "%5d (%5d) %s", (pid < 0) ? (pid = getpid()) : pid, threadID, s);
#endif /* RPCDEBUG */
}

#define	mp_clnt_pcreateerror(f1, f2) \
    rpcDebug("%s: (%s)%s\n", f1, f2, clnt_spcreateerror(""))

#define mp_clnt_perror(c, f1, f2) \
    rpcDebug("%s: (%s)%s\n", f1, f2, clnt_sperror(c, ""))

/*****************************************************************************
 *
 *****************************************************************************/

#define MAXRPCRETRY	1	/* number of times to retry connection/call */

/*
 * Rpc_connectSitemgr -
 *    returns a handle for making subsequent RPC calls to the sitemgr at
 *    site hostid. funcname is the name of the calling function and onerr
 *    indicates what to do on error.
 *
 *    XXX someone please clean me up. addr is here because we can't call
 *        getAddrFromHostId inside the sitemgr. Two things complicate this
 *        implementation:
 *            (1) getAddrFromHostId needs to look up pg_host and doesn't
 *                work in the sitemgr
 *            (2) the broker is called in the sitemgr for subcontracting
 *                and needs to contact other sitemgr.
 *        Seems like these two things are broken and should be fixed.
 *        Until then, please specify addr when calling Rpc_connectSitemgr
 *        in the sitemgr. Otherwise, specify a 0 and we'll look it up for
 *        you.      - 4/95 ay
 */
RpcConn *
Rpc_connectSitemgr(int hostid, int addr, char *funcname, int onerr)
{
    RpcConn *rpcc = (RpcConn*)malloc(sizeof(RpcConn));
    struct sockaddr_in internetAddr;
    int sock = RPC_ANYSOCK;
    int retry = MAXRPCRETRY;

    rpcDebug("Rpc_connectSitemgr: (%s) hostid=%d addr=%d\n",
	     funcname, hostid, addr);
    
    /*
     * set up the internet address for the connection.
     */
    internetAddr.sin_family = AF_INET;
    internetAddr.sin_port = 0;
    if (addr != 0) {
	internetAddr.sin_addr.s_addr = addr;
    } else {
	internetAddr.sin_addr.s_addr = getAddrFromHostId(hostid);
    }

    rpcc->rpc_prog = SITEMGRPROG;

    while (retry > 0) {
	rpcc->rpc_clientHandle =
	    clnttcp_create(&internetAddr, SITEMGRPROG + hostid, SITEMGRVERS,
			   &sock, 0, 0);
	if (rpcc->rpc_clientHandle != NULL)
	    break;

	/* fails to connect, retry. */
	mp_clnt_pcreateerror("Rpc_connectSitemgr", funcname);
	if (--retry) {
	    rpcDebug("Rpc_connectSitemgr: (%s) trying to connect to sitemgr (hostid=%d) once more (retry=%d)\n",
		     funcname, hostid, retry);
	}
    }

    if (rpcc->rpc_clientHandle == NULL) {
	switch (onerr) {
	case RPC_ONERR_WARN:
	    elog(NOTICE, "%s: can't connect to sitemgr (hostid=%d,addr=%d)",
		 funcname, hostid, internetAddr.sin_addr.s_addr);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}
	
	rpcDebug("Rpc_connectSitemgr: (%s) hostid=%d addr=%d FAILED.\n",
		 funcname, hostid, addr);

	free(rpcc);
	return NULL;
    }
    rpcc->rpc_sock = sock;
    rpcc->rpc_caller_name = funcname;
    rpcc->rpc_onerr = onerr;
    rpcc->rpc_hostid = hostid;

    rpcDebug("Rpc_connectSitemgr: (%s) hostid=%d addr=%d done.\n",
	     funcname, hostid, addr);
    
    return rpcc;
}

/*
 * Rpc_connectExec -
 *    returns a handle for making subsequent RPC calls to an exec/backend
 *    at internet address addr and TCP port port. funcname is the name of
 *    the calling function and onerr indicates what to do on error.
 */
RpcConn *
Rpc_connectExec(int addr, int port, char *funcname, int onerr)
{
    RpcConn *rpcc = (RpcConn*)malloc(sizeof(RpcConn));
    struct sockaddr_in internetAddr;
    int sock = RPC_ANYSOCK;
    int retry = MAXRPCRETRY;

    internetAddr.sin_family = AF_INET;
    internetAddr.sin_port = port;
    internetAddr.sin_addr.s_addr = addr;

    rpcDebug("Rpc_connectExec: (%s) addr=%d port=%d\n",
	     funcname, addr, port);

    rpcc->rpc_prog = EXECPROG;

    while (retry > 0) {
	rpcc->rpc_clientHandle =
	    clnttcp_create(&internetAddr, EXECPROG, EXECVERS,
			   &sock, 0, 0);
	if (rpcc->rpc_clientHandle != NULL)
	    break;

	/* fails to connect, retry. */
	mp_clnt_pcreateerror("Rpc_connectExec", funcname);
	if (--retry) {
	    rpcDebug("Rpc_connectExec: (%s) trying to connect to exec once more (retry=%d)\n",
		     funcname, retry);
	}
    }

    if (rpcc->rpc_clientHandle == NULL) {
	switch (onerr) {
	case RPC_ONERR_WARN:
	    elog(WARN, "%s: can't connect to exec (addr=%d, port=%d)",
		 funcname, addr, port);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}

	rpcDebug("Rpc_connectExec: (%s) addr=%d port=%d FAILED\n",
		 funcname, addr, port);
	
	free(rpcc);
	return NULL;
    }

    rpcc->rpc_sock = sock;
    rpcc->rpc_caller_name = funcname;
    rpcc->rpc_onerr = onerr;
    rpcc->rpc_hostid = -1;
    
    rpcDebug("Rpc_connectExec: (%s) addr=%d port=%d DONE\n",
	     funcname, addr, port);
    
    return rpcc;
}

/*
 * Rpc_connectLocalBroker -
 *    returns a handle for making subsequent RPC calls to the broker backend
 *    running at this site.
 *    funcname is the name of the calling function and onerr indicates what
 *    to do on error.
 */

#if 0
RpcConn *
Rpc_connectLocalBroker(char *funcname, int onerr)
{
    RpcConn *rpcc = (RpcConn*)malloc(sizeof(RpcConn));
    struct sockaddr_in internetAddr;
    int sock = RPC_ANYSOCK;
    int retry = MAXRPCRETRY;
    int hostid = getMyHostId();

    internetAddr.sin_family = AF_INET;
    internetAddr.sin_port = 0;
    internetAddr.sin_addr.s_addr = getMyAddr();


    rpcc->rpc_prog = BROKERPROG;

    while (retry > 0) {
	rpcc->rpc_clientHandle =
	    clnttcp_create(&internetAddr, BROKERPROG + hostid, BROKERVERS,
			   &sock, 0, 0);
	if (rpcc->rpc_clientHandle != NULL)
	    break;

	/* fails to connect, retry. */
	mp_clnt_pcreateerror("Rpc_connectLocalBroker", funcname);
	if (--retry) {
	    rpcDebug("Rpc_connectLocalBroker: (%s) trying to connect to broker once more (retry=%d)\n",
		     funcname, retry));
	}
    }

    if (rpcc->rpc_clientHandle == NULL) {
	switch (onerr) {
	case RPC_ONERR_WARN:
	    elog(WARN, "%s: can't connect to broker at host %d",
		 funcname, hostid);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}

	free(rpcc);
	return NULL;
    }

    rpcc->rpc_sock = sock;
    rpcc->rpc_caller_name = funcname;
    rpcc->rpc_onerr = onerr;
    rpcc->rpc_hostid = hostid;
    
    return rpcc;
}
#endif

/*
 * Rpc_connectRemoteBroker -
 *    returns a handle for making subsequent RPC calls to a Broker
 *    at internet address addr and TCP port port. funcname is the name of
 *    the calling function and onerr indicates what to do on error.
 */
RpcConn *
Rpc_connectRemoteBroker(int addr, int port, int hostid, char *funcname, int onerr)
{
    RpcConn *rpcc = (RpcConn*)malloc(sizeof(RpcConn));
    struct sockaddr_in internetAddr;
    int sock = RPC_ANYSOCK;
    int retry = MAXRPCRETRY;

    internetAddr.sin_family = AF_INET;
    internetAddr.sin_port = port;
    internetAddr.sin_addr.s_addr = addr;

    rpcDebug("Rpc_connectRemoteBroker: (%s) addr=%d port=%d\n",
	     funcname, addr, port);

    rpcc->rpc_prog = EXECPROG;

    while (retry > 0) {
#if 0
 	rpcc->rpc_clientHandle =
 	    clnttcp_create(&internetAddr, EXECPROG+hostid, EXECVERS,
 			   &sock, 0, 0);
#endif
	rpcc->rpc_clientHandle =
	    clnttcp_create(&internetAddr, EXECPROG, EXECVERS,
			   &sock, 0, 0);
	if (rpcc->rpc_clientHandle != NULL)
	    break;

	/* fails to connect, retry. */
	mp_clnt_pcreateerror("Rpc_connectRemoteBroker", funcname);
	if (--retry) {
	    rpcDebug("Rpc_connectRemoteBroker: (%s) trying to connect to remote Broker once more (retry=%d)\n",
		     funcname, retry);
	}
    }

    if (rpcc->rpc_clientHandle == NULL) {
	switch (onerr) {
	case RPC_ONERR_WARN:
	    elog(WARN, "%s: can't connect to Broker (addr=%d, port=%d)",
		 funcname, addr, 0);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}

	free(rpcc);
	return NULL;
    }

    rpcc->rpc_sock = sock;
    rpcc->rpc_caller_name = funcname;
    rpcc->rpc_onerr = onerr;
    rpcc->rpc_hostid = -1;
    
    return rpcc;
}

/*
 * Rpc_connectNamesvr -
 *    returns a handle for making subsequent RPC calls to a name server
 *    at internet address addr. funcname is the name of the calling
 *    function and onerr indicates what to do on error.
 */
RpcConn *
Rpc_connectNamesvr(int addr, char *funcname, int onerr)
{
    RpcConn *rpcc = (RpcConn*)malloc(sizeof(RpcConn));
    struct sockaddr_in internetAddr;
    int sock = RPC_ANYSOCK;

    internetAddr.sin_family = AF_INET;
    internetAddr.sin_port = 0;
    internetAddr.sin_addr.s_addr = addr;

    rpcc->rpc_prog = NAMESVRPROG;
    rpcc->rpc_clientHandle =
	clnttcp_create(&internetAddr, NAMESVRPROG, NAMESVRVERS, 
		       &sock, 0, 0);
    if (rpcc->rpc_clientHandle == NULL) {
	mp_clnt_pcreateerror("Rpc_connectNamesvr", funcname);

	switch (onerr) {
	case RPC_ONERR_WARN:
	    elog(WARN, "%s: can't create client handle", funcname);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}

	free(rpcc);
	return NULL;
    }
    rpcc->rpc_sock = sock;
    rpcc->rpc_caller_name = funcname;
    rpcc->rpc_onerr = onerr;
    rpcc->rpc_hostid = -1;
    
    return rpcc;
}

/*
 * Rpc_call -
 *    make an RPC call with the established RPC "connection". The
 *    call is specified by callDesc. arg is the argument and retval, the
 *    return value.
 *    
 *    For each of the calls listed in mariposa_rpc.x, a corresponding
 *    RPC call descriptor variable is generated. The name of the variable
 *    is the name of the call (eg. TRANSMITDATA) prefixed with RPC_ (eg.
 *    RPC_TRANSMITDATA). This is the callDesc to be specified below.
 */
int
Rpc_call(RpcConn *rpcc,
	 RpcCallDesc *callDesc,
	 void *arg,
	 void *retval)
{
    enum clnt_stat status;
    struct timeval timeout = {60, 0};
    int retry = MAXRPCRETRY;

    rpcDebug("Rpc_call: (%s) call=%d to hostid=%d sock=%d\n",
	     rpcc->rpc_caller_name, callDesc->rpc_call_id,
	     rpcc->rpc_hostid, rpcc->rpc_sock);

    while (retry > 0) {
	status = clnt_call(rpcc->rpc_clientHandle,
			   callDesc->rpc_call_id,
			   callDesc->rpc_xdr_in,
			   arg,
			   callDesc->rpc_xdr_out,
			   retval,
			   timeout);
	if (status == RPC_SUCCESS)
	    break;

	/* fails to make the rpccall */
	mp_clnt_perror(rpcc->rpc_clientHandle,
		       "Rpc_call", rpcc->rpc_caller_name);
	if (--retry) {
	    rpcDebug("Rpc_call: (%s) call=%d status=%d to hostid=%d sock=%d failed. retry=%d\n",
		    rpcc->rpc_caller_name, callDesc->rpc_call_id, status,
		    rpcc->rpc_hostid, rpcc->rpc_sock, retry);
	}
    }

    rpcDebug("Rpc_call: (%s) call=%d status=%d to hostid=%d sock=%d done.\n",
	     rpcc->rpc_caller_name, callDesc->rpc_call_id, status,
	     rpcc->rpc_hostid, rpcc->rpc_sock);

    if (status != RPC_SUCCESS) {

	switch(rpcc->rpc_onerr) {
	case RPC_ONERR_WARN:
	    {
		char *p = clnt_sperrno(status);
		char rpcerr[BUFSIZ];
		
		if (p == (char *) NULL) {
		    (void) strcpy(rpcerr, "RPC error");
		} else {
		    (void) strcpy(rpcerr, p);
		    if ((p = strchr(rpcerr, '\n')) != (char *) NULL) {
			*p = '\0';
		    }
		}
		elog(WARN, "%s: %s (%d)", rpcc->rpc_caller_name, rpcerr, status);
	    }
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}
	return 0;
    }

    return 1;
}

/*
 * Rpc_freeres -
 *    free up the memory allocated for the return value of an RPC call
 */
void
Rpc_freeres(RpcConn *rpcc, RpcCallDesc *callDesc, void *retval)
{
    if (clnt_freeres(rpcc->rpc_clientHandle,
		     callDesc->rpc_xdr_out, retval) == 0) {
	switch(rpcc->rpc_onerr) {
	case RPC_ONERR_WARN:
	    elog(WARN, "%s: cannot free return value",
		 rpcc->rpc_caller_name);
	    break;
	case RPC_ONERR_RETURN:
	default:
	    break;
	}
    }
}
    

/*
 * Rpc_disconnect -
 *    destroy the connection
 */
void
Rpc_disconnect(RpcConn *rpcc)
{
    rpcDebug("Rpc_disconnect: (%s) PROG=%s\n",
	     rpcc->rpc_caller_name,
	     (rpcc->rpc_prog==SITEMGRPROG)? "sitemgr":"exec");

    clnt_destroy(rpcc->rpc_clientHandle);

    rpcDebug("Rpc_disconnect: (%s) PROG=%s done\n",
	     rpcc->rpc_caller_name,
	     (rpcc->rpc_prog==SITEMGRPROG)? "sitemgr":"exec");
    
    free(rpcc);
}


/*****************************************************************************
 *
 *****************************************************************************/

/*
 * rpcAsyncHandleRequest -
 *    gets the argument from the transport handle and execute the
 *    function to handle the request in a thread.
 *
 */
void
rpcAsyncHandleRequest(SVCXPRT *transp,	   /* the transport handle */
		      xdrproc_t xdrproc,   /* the xdr function for the arg */
		      int  argsize,	   /* size of the argument */
		      void threadfunc(),  /* the function handling the req */
		      char *reqname)	   /* name of the req for err msg */
{
    void *arg;

    rpcDebug("rpcAsyncHandleRequest: calling %s\n",
	     reqname);
    
    /* allocate memory for svc_getargs */
    if (argsize > 0) {
	arg = calloc(1, argsize);
	if (arg == NULL) {
	    elog(WARN,
		 "rpcAsyncHandleRequest: cannot allocate arguments to %s",
		 reqname);
	}
	
	/* now get the argument */
	if (!svc_getargs(transp, xdrproc, arg)) {
	    elog(WARN,
		 "rpcAsyncHandleRequest: cannot get arguments to %s",
		 reqname);
	}
    } else {
	arg = NULL;
    }

    /* start a thread to handle the call */
    RPCThreadStart(arg, transp, threadfunc);
    
    rpcDebug("rpcAsyncHandleRequest: called %s.  Returning.\n",
	     reqname);
}

/*
 * rpcSyncHandleRequest -
 *    gets the argument from the transport handle and execute the
 *    function to handle the request. (This is synchronous. Call this
 *    for small jobs. You might also want to call this instead of
 *    rpcAsyncHandleRequest for frequent jobs so that we don't run out
 *    of threads.)
 */
void
rpcSyncHandleRequest(SVCXPRT *transp,	 /* the transport handle */
		     xdrproc_t xdrproc,  /* the xdr function for the arg */
		     int  argsize,	 /* size of the argument */
		     void threadfunc(),  /* the function handling the req */
		     char *reqname)	 /* name of the req for err msg */
{
    void *arg;
    char **argv = (char **)malloc(2 * sizeof(char *));

    rpcDebug("rpcSyncHandleRequest: calling %s\n",
	     reqname);
    
    /* allocate memory for svc_getargs */
    if (argsize > 0) {
	arg = calloc(1, argsize);
	if (arg == NULL) {
	    elog(WARN,
		 "rpcSyncHandleRequest: cannot allocate arguments to %s",
		 reqname);
	}
	
	/* now get the argument */
	if (!svc_getargs(transp, xdrproc, arg)) {
	    elog(WARN,
		 "rpcSyncHandleRequest: cannot get arguments to %s",
		 reqname);
	}
    } else {
	arg = NULL;
    }

    /* start a thread to handle the call */
    argv[0] = (char *)arg;
    argv[1] = (char *)transp;

    threadfunc(argv);
    
    rpcDebug("rpcSyncHandleRequest: called %s.  Returning.\n",
	     reqname);
}

/*
 *-------------------------------------------------------------------------
 * RPCThreadStart -
 *    Starts a new thread to handle the RPC service.
 *    The original thread returns to service more clients.
 *-------------------------------------------------------------------------
 */
void
RPCThreadStart(void *args, SVCXPRT *transp, void threadfunc())
{
    char **argv = (char **)malloc(2 * sizeof(char *));
    pthread_t thread;
    pthread_attr_t attr;
    long stacksize;

    pthread_attr_create(&attr);
    stacksize = pthread_attr_getstacksize(attr);
    pthread_attr_setstacksize(&attr, 5 * stacksize);

    argv[0] = (char *)args;
    argv[1] = (char *)transp;

    if (pthread_create(&thread,
		       attr,
		       (void *(*)())threadfunc,
		       (void *)argv) == -1) {
	elog(WARN, "RPCThreadStart: cannot create thread");
    }
    if (pthread_detach(&thread) == -1) {
	elog(WARN, "RPCThreadStart: cannot run thread");
    }
}


/*****************************************************************************
 *
 *****************************************************************************/

/*
 * mainThreadExec -
 *     execute func in main thread with argument arg and return result in
 *     retval. This is synchronous.
 *
 *     All database operations: catalog lookups, heapscans, etc. have to
 *     be done inside the main thread, and should be done by calling
 *     mainThreadExec.
 */
void
mainThreadExec(thdproc_t func, void *arg, void *retval)
{
    callExec(func, FALSE, arg, retval, NULL, NULL);
}

/*
 * mainThreadExecInXact -
 *     same as mainThreadExecInXact except this is done in a transaction.
 */
void
mainThreadExecInXact(thdproc_t func, void *arg, void *retval)
{
    callExec(func, TRUE, arg, retval, NULL, NULL);
}


/*****************************************************************************
 *
 *****************************************************************************/

/*
 * getMyAddr -
 *    returns the internet address of the host
 */
int
getMyAddr(void)
{
    struct hostent *hent;
    int addr;
    char hostname[MAXPATHLEN];

    if (gethostname(hostname, MAXPATHLEN)==-1 ||
	(hent = gethostbyname(hostname)) == NULL) {
	elog (WARN, "getMyAddr: Cannot get host entry for localhost.");
    }
    memcpy((char*)&addr, hent->h_addr, hent->h_length);
    return addr;
}


/*
 * getMyHostId -
 *    returns the host id that postgres is running under
 */
int
getMyHostId(void)
{
    return MyHostId;    /* from miscadmin.h */
}

/*
 * getMyNameServerHostId -
 *    returns the host id of our name server
 */
int
getMyNameServerHostId(void)
{
    return MyNameServerHostId;    /* from miscadmin.h */
}
