/*-------------------------------------------------------------------------
 *
 * twopc.c--
 *    Routines to support a two-phase commit protocol in Mariposa.
 *
 *
 * Copyright (c) 1995, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/rpc/twopc.c,v 1.3 1996/03/05 23:47:37 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdarg.h>

#include <rpc/types.h>
#include <pthread.h>
#include "utils/elog.h"
#include "mariposa_rpc.h"			/* for procedure constants */
#include "rpc/rpcutils.h"
#include "rpc/call_exec.h"			/* backend calling facility */
#include "rpc/rexec.h"
#include "sitemgr/rpcerror.h"

#include "postgres.h"

#include "access/xact.h"


/* ----------------
 *  Support routines for file-scope fragment moving synchronization variables.
 * ----------------
 */

static void SyncVarInitialize();
static void TransactionDone();
static void WaitTransactionToTerminate();

/* ----------------
 *  File-scope synchronization variables.
 * ----------------
 */

static bool condVarInitialized = FALSE;
static pthread_mutex_t xactLock;
static bool xactFinished;
static pthread_cond_t xactDone;


/*****************************************************************************
 * Routines called by coordinating site in two phase commit protocol.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *	StartRemoteSite
 *
 *
 * ----------------------------------------------------------------
 */
void
StartRemoteSite(void *rpcc, int xactid)
{
#if 0
    Rpc_setCallerName(((RpcConn *)rpcc), "StartRemoteSite");

    elog(NOTICE, "StartRemoteSite: Succeeded.");
#endif
}

/* ----------------------------------------------------------------
 *	AbortRemoteSite
 *
 *	Calls remote site on rpc connection passed in to see if it can
 *	take a TWOPCEND message (verifies that xactid is still valid.)
 *	If so, calls TWOPCEND to abort the transaction remotely.
 *
 *	Has side effect of closing rpc connection.
 *
 * ----------------------------------------------------------------
 */
void
AbortRemoteSite(void *rpcc, int xactid)
{
    EndXactDirective	directive = ABORTXACT;
    RPCErrorCode	result;

    Rpc_setCallerName(((RpcConn *)rpcc), "AbortRemoteSite");

    if (!Rpc_call((RpcConn *)rpcc, RPC_TWOPCPREPARE, &xactid, &result)) {
	elog(NOTICE, "AbortRemoteSite: TWOPCPREPARE rpc call failed.");
	Rpc_disconnect((RpcConn *)rpcc);
	return;
    }

    if (result != SUCCESS) {
	elog(NOTICE, "AbortRemoteSite: TWOPCPREPARE error %d.", result);
	Rpc_disconnect((RpcConn *)rpcc);
	return;
    }

    if (!Rpc_call((RpcConn *)rpcc, RPC_TWOPCEND, &directive, &result)) {
	elog(NOTICE, "AbortRemoteSite: TWOPCEND rpc call failed.");
	Rpc_disconnect((RpcConn *)rpcc);
	return;
    }

    if (result != SUCCESS) {
	elog(NOTICE, "AbortRemoteSite: TWOPCEND error %d.", result);
	Rpc_disconnect((RpcConn *)rpcc);
	return;
    }

    elog(NOTICE, "AbortRemoteSite: Succeeded.");

    Rpc_disconnect((RpcConn *)rpcc);
}

/* ----------------------------------------------------------------
 *	CommitRemoteSite
 *
 *	Calls remote site on rpc connection passed in to see if it can
 *	take a TWOPCEND message (verifies that xactid is still valid.)
 *	If so, calls TWOPCEND to commit the transaction	remotely.
 *
 *	Has side effect of closing rpc connection.
 *
 * ----------------------------------------------------------------
 */
void
CommitRemoteSite(void *rpcc, int xactid)
{
    EndXactDirective	directive = COMMITXACT;
    RPCErrorCode	result;

    Rpc_setCallerName(((RpcConn *)rpcc), "CommitRemoteSite");

    if (!Rpc_call((RpcConn *)rpcc, RPC_TWOPCPREPARE, &xactid, &result)) {
	Rpc_disconnect((RpcConn *)rpcc);
	elog(WARN, "CommitRemoteSite: TWOPCPREPARE rpc call failed.");
    }

    if (result != SUCCESS) {
	Rpc_disconnect((RpcConn *)rpcc);
	elog(WARN, "CommitRemoteSite: TWOPCPREPARE error %d.", result);
    }

    if (!Rpc_call((RpcConn *)rpcc, RPC_TWOPCEND, &directive, &result)) {
	Rpc_disconnect((RpcConn *)rpcc);
	elog(WARN, "CommitRemoteSite: TWOPCEND rpc call failed.");
    }

    if (result != SUCCESS) {
	Rpc_disconnect((RpcConn *)rpcc);
	elog(WARN, "CommitRemoteSite: TWOPCEND error %d.", result);
    }

    elog(NOTICE, "CommitRemoteSite: Succeeded.");

    Rpc_disconnect((RpcConn *)rpcc);
}


/*****************************************************************************
 * Routines that handle the RPC calls from coordinator that ensure the
 * two phase commit protocol.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	handleCall_TWOPCSTART
 *
 *	Called for a coordinating site to start a transaction here. They
 *	get back the xactid, and also we notify them of the backend address
 *	so they can contact us. If at any point we detect failure, we abort,
 *	because the last thing we do is send them the backend address, and
 *	there would be no way for them to contact us back.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_TWOPCSTART(void *args)
{
    char		**argv = (char **)args;
    SiteAddr		*rpcarg = (SiteAddr *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    TransactionId	xactid;
    RPCErrorCode	result;
    int			pid;
    
    elog(NOTICE, "Entering handleCall_TWOPCSTART.");

    /*
     * Initialize transaction synchronization variables, if not done already.
     */
    SyncVarInitialize();

    xactFinished = FALSE;

    /*
     * Start the transaction, get the xactid, and return it to the
     * coordinating site so it knows what xactid to call us back with
     * when asking us to prepare and commit.
     */
    StartTransactionCommand();
    xactid = GetCurrentTransactionId();

    elog(NOTICE, "handleCall_TWOPCSTART: svc_sendreply to sitemgr.");
    
    if (!svc_sendreply(transp, xdr_int, &xactid)) {
	errorExit("handleCall_TWOPCSTART: Can't send reply.\n");
    }

    /*
     * Let the coordinating site know the address of our backend. If this
     * fails, we must abort here, because there's no way they can contact
     * us back to tell us anything.
     */
    mainThreadExec((thdproc_t) RExecSignalReady, rpcarg, &result);
    if (result != SUCCESS) {
	mainThreadExec((thdproc_t)AbortCurrentTransaction, NULL, NULL);

	/*
	 * Now tell the sitemgr we're done.
	 */
	pid = getpid();
	sitemgrBackendDone(pid);
    }

    WaitTransactionToTerminate();
    
    /*
     * Now tell the sitemgr we're done.
     */
    pid = getpid();
    sitemgrBackendDone(pid);

/*
 *   if (!svc_freeargs(arg->transportHandle, xdr_ProcessQueryArg, arg->arg)) {
 *	errorExit("handleCall_TWOPCSTART: Can't free arguments.\n");
 *    }
 */

    elog(NOTICE, "Leaving handleCall_TWOPCSTART.");

    free(rpcarg);
    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_TWOPCPREPARE
 *
 *	Tells whether this site can abort or commit, depending on whether
 *	the transaction id passed in is still the current transaction id,
 *	and the transaction state is in progress.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_TWOPCPREPARE(void *args)
{
    char		**argv = (char **)args;
    int			*rpcarg = (int *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result = SUCCESS;

    elog(NOTICE, "Entering handleCall_TWOPCPREPARE.");

    if ((TransactionId)*rpcarg != GetCurrentTransactionId())
	result = EXECUTION_ERROR;
    if (!CurrentXactInProgress())
	result = EXECUTION_ERROR;

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_TWOPCPREPARE: Can't send reply.\n");
    }

    elog(NOTICE, "Leaving handleCall_TWOPCPREPARE.");

    free(args);
}

/* ----------------------------------------------------------------
 *    	handleCall_TWOPCEND
 *
 *	Told to either abort or commit the current transaction.
 *	Does what it's told. Signals done on the transaction condition
 *	variable.
 *
 * ----------------------------------------------------------------
 */
void
handleCall_TWOPCEND(void *args)
{
    char		**argv = (char **)args;
    EndXactDirective	*rpcarg = (EndXactDirective *)argv[0];
    SVCXPRT		*transp = (SVCXPRT *)argv[1];
    RPCErrorCode	result = SUCCESS;
    int			pid;

    elog(NOTICE, "handleCall_TWOPCEND. (%s)",
	 (*rpcarg == ABORTXACT) ? "ABORT" : "COMMIT");

    if (*rpcarg == ABORTXACT) {
	/*
	 * Abort
	 */
	mainThreadExec((thdproc_t)AbortCurrentTransaction, NULL, NULL);
    }
    else {
	/*
	 * Commit
	 */
	mainThreadExec((thdproc_t)CommitTransactionCommand, NULL, NULL);
    }

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
  	errorExit("handleCall_TWOPCEND: Can't send reply.\n");
    }
    
    /*
     * Signal done on the fragment move condition variable.
     */
    TransactionDone();

    elog(NOTICE, "Leaving handleCall_CP_TWOPCEND.");

    free(args);
}


/*****************************************************************************
 * Support routines for file-scope transaction synchronization variables.
 *****************************************************************************/


/* ----------------------------------------------------------------
 *    	SyncVarInitialize
 *
 *	Initializes transaction locks and condition variables if
 *	not done already.
 *
 * ----------------------------------------------------------------
 */
void
SyncVarInitialize()
{
    if (!condVarInitialized) {
	pthread_mutex_init(&xactLock, pthread_mutexattr_default);
	pthread_cond_init(&xactDone, pthread_condattr_default);
	condVarInitialized = TRUE;
    }
}

/* ----------------------------------------------------------------
 *    	TransactionDone
 *
 *	Routine that takes care of the condition variable assignment
 *	and broadcast when finished with a transacation.
 *
 * ----------------------------------------------------------------
 */
void
TransactionDone()
{
    int status;

    status = pthread_mutex_lock(&xactLock);
    checkAndExit(status, "TransactionDone: Can't lock xactLock.");
    xactFinished = TRUE;
    status = pthread_cond_signal(&xactDone);
    checkAndExit(status, "TransactionDone: Couldn't wait for xactDone.");
    status = pthread_mutex_unlock(&xactLock);
    checkAndExit(status, "TransactionDone: Can't unlock xactLock.");
}

/* ----------------------------------------------------------------
 *    	WaitTransactionToTerminate
 *
 *	Wait for xactDone signal.
 *
 * ----------------------------------------------------------------
 */
void
WaitTransactionToTerminate()
{
    int status;
    
    /*
     * Wait to be signalled 'done'.
     */
    status = pthread_mutex_lock(&xactLock);
    checkAndExit(status, "WaitTransactionToTerminate: Can't lock xactLock.");
    while (!xactFinished) {
	status = pthread_cond_wait(&xactDone, &xactLock);
	checkAndExit(status,
		     "WaitTransactionToTerminate: Couldn't wait for xactDone.");
    }
    status = pthread_mutex_unlock(&xactLock);
    checkAndExit(status, "WaitTransactionToTerminate: Can't unlock xactLock.");
}
