#include <stdio.h>
#include "parameter.h"
#include "listSched.h"
#include "querySched.h"
#include "request.h"
#include "schedInterface.h"

typedef struct LruEntry {
    ElemType * next;
    int rel;
} * LruEntryPtr, LruEntryData;

struct {
    int size;         /* size of cache in KB */
    int free;         /* free space in cache */
    ListX reqList;    /* list of pending requests in cache */ 
    ListX lruList;
    LruEntryData * relLruList;
    char *relState;   /* is relation cached, out, io in progress      */
    char *refcnt;     /* number of references to the relation         */
    char *relUsed;
    float * relRefTime; /* time of last access of a relation */
} cache;
 
/* defines for relation state */
#define OUT 0
#define CACHED 1
#define IOINPROG 2

int deadlock = 0;
 
extern int evictionPolicy;
extern RequestPtr * perRelQ;

/*
 * routines exported from this file
 */
void InitCache PROTO((void));
int  RelCached PROTO((int i));
int  CacheRel PROTO((int i));
void  UncacheRel PROTO((int i));
void PinRel PROTO((int i));
void UnpinRel PROTO((int i));
void AccessRelation  PROTO((int i));
int MakeSpaceAndCacheRel PROTO((int i));
int MakeSpaceLRU PROTO((int i));
int FreeCacheSpace PROTO((void));
void PrintLruEntryStruct PROTO((LruEntryPtr lruPtr));


int ReqCached PROTO((char * req));
ListX GetCacheRequestList PROTO((void));
void CacheRequest PROTO(( RequestPtr req));
void UncacheRequest PROTO(( RequestPtr req));
int MakeSpaceInCache PROTO((int size, int forRel));

/*
 * Routines internal to this file
 */
static void LruInsert_SS PROTO((int i));
static void LruDelete_SS PROTO((int i));
static int UncacheLruRel PROTO((void));
static int RelPinned PROTO((int i));

static void DetectDeadlock PROTO((void));
static RequestPtr ChooseReqToEvict PROTO((int forRel, int cond));
static int MeritCriteria PROTO((RequestPtr req, int forRel, int policy));
static RequestPtr LruReq PROTO((int forRel));
static RequestPtr LessRecentlyUsed PROTO((RequestPtr req1, RequestPtr req2));
static RequestPtr MoreSTP PROTO((RequestPtr req1, RequestPtr req2));

/* --------------------------------------------------------------------
 * We do caching on a per relation basis. So, the cache can be
 * modeled simply by stating what relations are currently in
 * cache, the amount of free space available and the total
 * capacity of the  cache.  
 */
void
InitCache()
{
    int i;

    cache.size = cacheSize;
    cache.free = cacheSize;
    InitList(& (cache.reqList));
    cache.relState = (char *) MallocErrorCheck(numRels);
    bzero (cache.relState, numRels);
    cache.refcnt = (char *) MallocErrorCheck(numRels);
    bzero (cache.refcnt, numRels);

    cache.relLruList = (LruEntryData *) MallocErrorCheck
	(numRels * sizeof(LruEntryData));
    
    for (i = 0; i < numRels; i++) {
	cache.relLruList[i].rel = i;
    }    
    InitList (&(cache.lruList));
    
    cache.relUsed = (char *) MallocErrorCheck(numRels);
    bzero (cache.relUsed, numRels);
    cache.relRefTime = (float *) MallocErrorCheck
	(numRels * sizeof(float));
    bzero (cache.relRefTime, numRels*sizeof(float));
}

/* ------------------------------------------------------------
 * return 1 if the relation is cached
 */
int
RelCached(i)
int i;
{
    if (i == -1) return 1;
    return ((cache.relState[i] == CACHED));
}

/*--------------------------------------------------------
 * cache relation with id i in cache. returns 1 if caching is successful
 */
int
CacheRel(i)
int i;
{
    if (RelCached(i)) {
	return (1);
    }
    if (is_debug("l")) {
	printf ("Caching relation %d \n", i);
    }
    cache.relState[i] = CACHED;
    if (!RelPinned(i)) LruInsert_SS(i);
    cache.relUsed[i] = 0;
    cache.relRefTime[i] = MyTime() + 0.5;
    return (1);
}

/*-----------------------------------------------------------
 * evict a relation from cache
 */
void
UncacheRel(i)
int i;
{
    if (!RelCached(i)) {
	return;
    }
    if (is_debug("l")) {
        printf ("Unpinning relation %d \n", i);
    }
    cache.free += RelSize(i);
    cache.relState[i] = OUT;
    LruDelete_SS(i);
    PGUncacheRel(i);
    return;
}

/* ---------------------------------------------------------------
 * access the relation by making it the most recently accessed relation
 * in the lru queue
 */
void
AccessRelation(i)
int i;             /* name of relation accessed */
{
    if (i == -1) return;
    if (!RelCached(i)) {
	return;
    }
    if (is_debug("l")) {
	printf ("MRU relation: %d\n", i);
    }
    LruDelete_SS(i);
    LruInsert_SS(i);
    cache.relUsed[i] = 1;
} 

/* ------------------------------------------------------------
 * check if relation is pinned
 */
static int
RelPinned(i)
int i;
{
    if (cache.refcnt[i] > 0) {
	return 1;
    }
    return 0;
}
/* --------------------------------------------------------------
 * pin relation in cache -- a query is in progress so DONT evict
 * the relation
 */
void
PinRel(i)
int i;
{
    if (i == -1) return;
    if (NeedsFetching(i)) {
	PrintError ("Asking to pin an uncached fragment?");
    }
    LruDelete_SS(i);
    cache.refcnt[i] += 1;
    cache.relUsed[i] = 1;
    cache.relRefTime[i] = MyTime();  
    if (is_debug("l")) {
	printf("Pinning rel %d, refCnt %d\n", i,cache.refcnt[i]);
	PrintList(cache.lruList, PrintLruEntryStruct);
    } 
}

/* ----------------------------------------------------------
 * 
 */
void 
UnpinRel(i)
int i;
{
    if (i == -1) return;
    cache.refcnt[i] -= 1;
    if (cache.refcnt[i] <= 0) LruInsert_SS(i);
    if (is_debug("l")) {
	printf("Unpinning rel %d, refCnt %d\n", i,cache.refcnt[i]);
	PrintList(cache.lruList, PrintLruEntryStruct);
   } 
}


/* ----------------------------------------------------------
 * 
 */
RegisterRel (i)
int i;
{
    cache.refcnt[i] = 0;
    cache.relState[i] = IOINPROG;
    if (RelSize(i) > cache.free) {
	PrintError("No unpinned space left");
    }
    cache.free -= RelSize(i);
}

/* ------------------------------------------------------------
 * check if relation is either present in cache or is begin fetched
 */
int
NeedsFetching(i)
int i;
{
	if (i == -1) return 0;
    if (cache.relState[i] == OUT) {
	return 1;
    }
    return 0;
}
/* -----------------------------------------------------------
 *
 */
ClearCache()
{
    int i;
    
    for (i = 0; i < numRels; i++) {
	UncacheRel(i);
    }
}
/* ---------------------------------------------------------------
 * support routines for keeping the usage ordering of relations
 */

/*
 * insert as the most recently used relation in the lru list
 */
static void
LruInsert_SS(i)
int i;
{
    LruEntryPtr lruPtr;

    lruPtr = &(cache.relLruList[i]);
    AddEltoList (&(cache.lruList), (ElemType *)lruPtr);
}

void
static LruDelete_SS(i)
int i;
{
    LruEntryPtr lruPtr;

    lruPtr = &(cache.relLruList[i]);
    RemoveGivenElfromList (&(cache.lruList), (ElemType *)lruPtr);
}

static int
UncacheLruRel()
{
    LruEntryPtr lruPtr;

    lruPtr = (LruEntryPtr) GetFirstData(cache.lruList);
    
    if (lruPtr == NULL) {
	return 0;
    }
    UncacheRel (lruPtr->rel);
    return 1;
}

void
PrintLruEntryStruct(lruPtr)
LruEntryPtr lruPtr;
{
    printf ("%d ", lruPtr->rel);
}

/* ------------------------------------------------------------------- */
int
MakeSpaceLRU(rel)
int rel;
{
    int size;
    
    if (RelCached(rel)) {
	return;
    }
    /*
     * make space for an uncached relation
     */
    size = RelSize(rel);
    while (FreeCacheSpace() < size) {
	if (UncacheLruRel() == 0) {
	    return 0;
	}
    }
    return 1;
}

/* ---------------------------------------------------------------
 * return the amount of free space available
 */
int
FreeCacheSpace()
{
    return (cache.free);
}

/* -------------------------------------------------------------------- */
/* ------------------------------------------------------------
 * return 1 if the request is in cached list
 */
int
ReqCached(req)
char * req;
{
    Element * reqList;

	return RelCached(((RequestPtr )req)->relname);
    ForEach (reqList, cache.reqList) {
	char * thisReq;
	
	thisReq = (char *) GetData(reqList);
	if (thisReq == req) {
	    return (1);
	}
    }
    return (0);
}

/* -----------------------------------------------------------
 * return the cache request list
 */
ListX
GetCacheRequestList()
{
    return (cache.reqList);
}

/* ------------------------------------------------------------
 * enlist the given request in the cache request list and cache
 * the relation that the request refers to
 */
void
CacheRequest(req)
RequestPtr req;
{
    CacheRel(req->relname);
    AddEltoList(&(cache.reqList), (ElemType *)req);
    if (is_debug("b")) {
	printf ("Caching %d of cost %f and size %d\n", 
		req->relname, req->cost, req->cacheLoad);
	PrintCacheList();
    }
    statxRelsCached++;
    statxQryIn += NumQueries(req); 
}

/* -------------------------------------------------------------
 * remove from cache the relation that the cache refers to and
 * remove the request node from the cache list
 */
void
UncacheRequest(req)
RequestPtr req;
{
    UncacheRel (req->relname);
    RemoveGivenElfromList (&(cache.reqList), (ElemType *)req);
    EnqueueRequest(req);
    if (is_debug("b")) {
	printf ("Uncaching %d\n", req->relname);
    }
    statxQryOut += NumQueries(req);
}


PrintCacheList()
{
    PrintList(cache.reqList, PrintReqStruct);
    printf("\n");
}

/*
 * A deadlock occurs when the cache is filled with relations that are not
 * used even once before fetching and the executor is asking for more
 * relations to cache
 */

int throwingUnused = 0;

static void
DetectDeadlock()
{
    Element * lruEntry;
    LruEntryPtr aEntry;

#if 0
    ForEach (lruEntry, cache.lruList) {
	aEntry = (LruEntryPtr) GetData (lruEntry);
	if (cache.relUsed[aEntry->rel]) {
	    deadlock = 0;
	    return;
	}
    }
    deadlock = 1;
#endif

    if (throwingUnused >= 2) {
	deadlock = 1; 
    } else {
	deadlock = 0;
    }
    return;
}
/* --------------------------------------------------------------------
 * make space in cache to hold relation 'forRel' of size 'size'
 * The relation(s) chosen for eviction are determined using the current
 * eviction policy. If a deadlock is detected, the cache first tries to
 * choose from relations that do not join with the current relation.
 */ 

int 
MakeSpaceInCache(size, forRel)
int size, forRel;
{
    RequestPtr minReq = NULL, ChooseReqToEvict();
    extern int overLap;
    int jobPresent;

    DetectDeadlock();
   
    if (overLap && (QueryAvailable() || !CpuIsFree())) {
	jobPresent = 1;
    } else {
	jobPresent = 0;
    }
    while (FreeCacheSpace() < size) {
	minReq = ChooseReqToEvict(forRel, jobPresent);
	if (minReq == NULL) {
	    return 0;
	}
	if (cache.relUsed[minReq->relname] == 0) {
	    throwingUnused++;
	    statxWastedIOTime += minReq->cost;
	} else {
	    throwingUnused = 0;
	}
  	UncacheRequest(minReq);
    }
    return 1;
}

/*
 * Choose relation to evict -- if in deadlock we first try to choose
 * that are non-adjacent to the new one 
 */
static RequestPtr
ChooseReqToEvict(forRel, jobPresent)
int forRel;
int jobPresent;
{ 

    RequestPtr minReq = NULL;

    switch (evictionPolicy) {
    case LEASTQUERIES:  
    case LEASTOVERLAP:    
    case LEASTWORK: 
    case MAXCL:
    case SIZE_LRU:
    {
	RequestPtr req;
	Element * reqList;
	int leastMerit = INFTY;
	int merit;
	ListX cacheList;
	
	cacheList = GetCacheRequestList();
	ForEach (reqList, cacheList) {
	    req = (RequestPtr) GetData(reqList);
	    
	    if (RelPinned(req->relname)) {
		continue;
	    }
	    if (deadlock && Adjacent(req, forRel)) {
		continue;
	    }
	    if (jobPresent) {
	      if (HasQueries(req)) continue;
	    }
	    merit = MeritCriteria(req, forRel, evictionPolicy);
	    if (merit < leastMerit) {
		minReq = req;
		leastMerit = merit;
		continue;
	    }
	    if (merit == leastMerit) {
		minReq = LessRecentlyUsed (req, minReq);
	    }
	}
	if ((minReq == NULL) && deadlock) {
	    ForEach (reqList, cacheList) {
		req = (RequestPtr) GetData(reqList);
		if (RelPinned(req->relname)) {
		    continue;
		}
		if (jobPresent) {
		  if (HasQueries(req)) continue;
		}

		merit = MeritCriteria(req, forRel, evictionPolicy);
		if (merit < leastMerit) {
		    minReq = req;
		    leastMerit = merit;
		    continue;
		}
		if (merit == leastMerit) {
		    minReq = LessRecentlyUsed (req, minReq);
		}
	    }
	}
	return minReq;
    }
    case MIXED:
	/* select according to numQueries first */
	/* resolve ties by MinShare             */
	/* resolve ties by LRU                  */
    {
	RequestPtr req;
	Element * reqList;
	int leastMerit = INFTY;
	int merit;
	ListX cacheList;
	
	cacheList = GetCacheRequestList();
	ForEach (reqList, cacheList) {
	    req = (RequestPtr) GetData(reqList);
	    
	    if (RelPinned(req->relname)) {
		continue;
	    }
	    if (deadlock && Adjacent(req, forRel)) {
		continue;
	    }
	    if (jobPresent) {
	      if (HasQueries(req)) continue;
	    }

	    merit = MeritCriteria(req, forRel, LEASTWORK);
	    if (merit < leastMerit) {
		minReq = req;
		leastMerit = merit;
		continue;
	    }
	    if (merit == leastMerit) {
		int diffShare = MeritCriteria(req, forRel, LEASTOVERLAP) - 
		    MeritCriteria (minReq, forRel, LEASTOVERLAP);
		if (diffShare < 0) {
		    minReq = req;
		    continue;
		} 
		if (diffShare == 0) {
		    minReq = LessRecentlyUsed (req, minReq);
		}
	    }
	}
	if ((minReq == NULL) && deadlock) {
	    ForEach (reqList, cacheList) {
		req = (RequestPtr) GetData(reqList);
	    
		if (RelPinned(req->relname)) {
		    continue;
		}
		if (jobPresent) {
		  if (HasQueries(req)) continue;
		}

		merit = MeritCriteria(req, forRel, LEASTWORK);
		if (merit < leastMerit) {
		    minReq = req;
		    leastMerit = merit;
		    continue;
		}
		if (merit == leastMerit) {
		    int diffShare = MeritCriteria(req, forRel, LEASTOVERLAP) - 
			MeritCriteria (minReq, forRel, LEASTOVERLAP);
		    if (diffShare < 0) {
			minReq = req;
			continue;
		    } 
		    if (diffShare == 0) {
			minReq = LessRecentlyUsed (req, minReq);
		    }
		}
	    }
	}
	return minReq;
    }
	
    case LRU: 
    default:
	return LruReq(forRel);
    }
}

/* ---------------------------------------------------------------
 * 
 */
static int
MeritCriteria(req, forRel, criteria)
RequestPtr req;
int forRel;
int criteria;
{
    int merit;
    int lastRef;
    int numQueries;

    merit = 0;

    switch (criteria) {
    case LEASTQUERIES:
    { 
	Element * queryList;

	ForEach (queryList, req->singleRelQuery) {
	    merit++;
	}
	ForEach (queryList, req->joinQuery) {
	    merit++;
	}
	return merit;
    }
    case LEASTOVERLAP:
	return OverlapCacheReq (req);

    case LEASTWORK:
    {
	Element * joinList;
	
	ForEach (joinList, req->joinQuery) {
	    JoinEdgePtr joinPtr = (JoinEdgePtr) GetData(joinList);
	    merit += RelSize(joinPtr->reqPtr->relname);
	}
	return merit;
    }
    case MAXCL:
	return (- RelSize(req->relname));

    case SIZE_LRU:
	lastRef = MyTime() - cache.relRefTime[req->relname];
	return (-RelSize(req->relname)*lastRef);

    }
}
	
/*
 * return the least recently used relation
 */
static RequestPtr
LruReq(forRel)
int forRel;
{
    RequestPtr GetRequestPtr();
    Element * lruList;
    LruEntryPtr lruPtr;

    if (deadlock) {
        ForEach (lruList, cache.lruList) {
            lruPtr = (LruEntryPtr) GetData(lruList);
            if (!Adjacent(perRelQ[lruPtr->rel], forRel)) {
                return perRelQ[lruPtr->rel];
            }
        }
    }

    lruPtr = (LruEntryPtr) GetFirstData(cache.lruList);
   
    if (lruPtr == NULL) {
        return (NULL);
    }
    return perRelQ[lruPtr->rel];
}
   
/*
 * return the relation used less recently than the other
 */
static RequestPtr
LessRecentlyUsed(req1, req2)
RequestPtr req1, req2;
{
    LruEntryPtr lruPtr;
    Element * list = NULL;

    ForEach (list, cache.lruList) {
        int rel = ((LruEntryPtr) list)->rel;
        if (rel == req1->relname) {
            return (req1);
        }
        if (rel == req2->relname) {
            return (req2);
        }
    }
    return (req2);
}

int 
HasQueries(req)
RequestPtr req;
{
  if (!IsEmpty(req->singleRelQuery)) return 1;
  if (!IsEmpty(req->joinQuery)) return 1;
  return 0;
}

static RequestPtr
MoreSTP(req1, req2)
RequestPtr req1, req2;
{

    if ((RelSize(req1->relname)*(MyTime() - cache.relRefTime[req1->relname]))
	< (RelSize(req2->relname)*(MyTime() - cache.relRefTime[req2->relname])))
	return req2;
    return req1;
}



