/*---------------------------------------------------
 * scheduling algorithms for queries involving single-relation queries
 * or joins.
 */

#include <stdio.h>
#include "parameter.h"
#include "listSched.h"
#include "querySched.h"
#include "schedInterface.h"
#include "request.h"


#define ANY      1
#define MINBUF   2
#define MAXJOINS 3
#define MAXWORK  4
#define MINCOVER 5

enum source {CACHE, PLAT, TWO_ADJ_CACHE, LOADED_CACHE, LOADED_PLAT, LOADED_2ADJ_CACHE,
		 THIS_PLAT, THIS_PLAT_CACHE, THIS_PLAT_2ADJ_CACHE,
	 THIS_PLAT_SEEK, LOADED_PLAT_SEEK};

enum Domain {ADJ, TWO_ADJ, ALL};

extern ListX GetCacheRequestList();
extern InitialWork();
extern ListX perPlatterQ[MAX_PLAT]; 
extern TimeType epoch;
extern int avgRelSize;
extern int avgFragSize;

int WorkSizeInCache(), WorkSizeDoable(), TwoAdjacentWorkSize();
float magic2Adj = 0.4;
float magicAny = 0.2;
char * relUsed;
char * pendingRelList[MAX_DEVICE];

static enum Domain currDomain;

/* -------------------------------------------------------------
 * relation oriented policies
 */
RequestPtr RemoveBestReq(), SelectFromPlat(), SelectFromCache(), 
SelectFromLoadedPlat(), SelectFromLoadedCache(), SelectFromThisPlat(),
SelectFromThisPlatCache(), SelectFromThisPlatSeek(), SelectFromLoadedPlatSeek(), GetNearestRequest();

int NegCacheLoadReq(), WorkSize(), NegCoverSize(), 
OverlapCacheReq(), NegLatency(), Any();


/* -----------------------------------------------------------------
 * aggregate merit function policies
 */
RequestPtr
AdjPlatAgg(rankFunc)
int (*rankFunc)();
{
    int p;
    RequestPtr req;

    InitialWork();

    currDomain = ADJ;
    req = RemoveBestReq (rankFunc, LOADED_CACHE);
    if (req != NULL) return req;
    p = ChoosePlatter(rankFunc, CACHE);
    if (p != -1) {
	return RemoveBestReq (rankFunc, THIS_PLAT_CACHE, p);
    }
    req = RemoveBestReq(rankFunc, LOADED_PLAT);
    if (req != NULL) return req;
    p = ChoosePlatter(rankFunc, PLAT);
    return RemoveBestReq (rankFunc, THIS_PLAT, p);
}

/* ---------------------------------------------
 *
 */
RequestPtr
AnyPlatAgg(rankFunc)
int (*rankFunc)();
{
    int p;

    RequestPtr req;
    
    if (deadlock) return AdjPlatAgg(rankFunc); 
    InitialWork();

    currDomain = ALL;
    req = RemoveBestReq (rankFunc, LOADED_PLAT);
    if (req != NULL) return req;
    p = ChoosePlatter (rankFunc, PLAT);
    return RemoveBestReq (rankFunc, THIS_PLAT, p);
}

/* ------------------------------------------------------------
 * fetch request with maximum overlap with requests on cache.
 * if no space or no sharing, schedule request with mimimum joins
 * with uncached relations (maximize sharing).
 */
RequestPtr
TwoAdjPlatAgg(rankFunc)
int (*rankFunc)();
{
    RequestPtr req;
    ListX cacheList;
    int sharing = 0;
    int p;

    if (deadlock) return AdjPlatAgg(rankFunc); 
    InitialWork();

    currDomain = TWO_ADJ;

    req = RemoveBestReq (rankFunc, LOADED_2ADJ_CACHE);
    if (req != NULL) return req;
    p = ChoosePlatter (rankFunc, TWO_ADJ_CACHE);
    if (p != -1) {
	req = RemoveBestReq (rankFunc, THIS_PLAT_2ADJ_CACHE, p);
    }
    if (req != NULL) return (req);
    req = RemoveBestReq(rankFunc, LOADED_PLAT);
    if (req != NULL) return req;
    p = ChoosePlatter(rankFunc, PLAT);
    return RemoveBestReq (rankFunc, THIS_PLAT, p);
}

/* ---------------------------------------------
 *
 */
RequestPtr
SeekPlatAgg()
{
    int p;

    RequestPtr req;
    
    if (deadlock) return AdjPlatAgg(WorkSize); 
    InitialWork();

    currDomain = ALL;
    req = RemoveBestReq (NegLatency, LOADED_PLAT);
    if (req != NULL) return req;
    p = ChoosePlatter (WorkSize, PLAT);
    return RemoveBestReq (NegLatency, THIS_PLAT, p);
}

/* ----------------------------------------------------------
 * Merge all three policies
 */
RequestPtr
MixedPlatAgg()
{
    RequestPtr req;
    int p;

    if (deadlock) return AdjPlatAgg(WorkSize); 
    InitialWork();

    req = RemoveBestReq (WorkSizeInCache, LOADED_CACHE);
    if (req != NULL) return req;
    if (ExpandDomain(ALL)) {
	req = RemoveBestReq (TwoAdjacentWorkSize, LOADED_2ADJ_CACHE);
	if (req != NULL) return req;
 	req = RemoveBestReq (WorkSizeDoable, LOADED_PLAT);
	if (req != NULL) return req;
	p = ChoosePlatter (WorkSizeDoable, PLAT);
	return RemoveBestReq (WorkSizeDoable, THIS_PLAT, p);
    }
    p = ChoosePlatter(WorkSizeInCache, CACHE);
    if (p != -1) {
	return RemoveBestReq (WorkSizeInCache, THIS_PLAT_CACHE, p);
    }
    req = RemoveBestReq(WorkSizeDoable, LOADED_PLAT);
    if (req != NULL) return req;
    p = ChoosePlatter(WorkSizeDoable, PLAT);
    return RemoveBestReq (WorkSizeDoable, THIS_PLAT, p);
}    

/* -------------------------------------------------------------
 * ExpandDomain() return non-zero -- when 
 *       pending cache requirement is not too large
 *       cache is not too small compared to rel. size
 *       platter switch/avg transfer time is not too large
 */
int
ExpandDomain(domain)
enum Domain domain;
{
    ListX cacheList;
    Element * reqList, * joinList;
    RequestPtr req, otherReq;
    int i;
    int usedSize = 0;
    int pendingSize = 0;
    float free;
    float temp;

    bzero (relUsed, numRels);
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
	req = (RequestPtr) GetData(reqList);
	ForEach (joinList, req->joinQuery) {
	    JoinEdgePtr joinEdge;
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    
	    otherReq = joinEdge->reqPtr;
	    relUsed[otherReq->relname]++;
	}
	if (! IsEmpty(req->joinQuery)) {
	    usedSize += RelSize (req->relname);
	}
    }
    for (i = 0; i < numRels; i++) {
	if (relUsed[i]) {
	    pendingSize += RelSize(i);
	}
    }
    if (1) {
        float xfrCost =  (float) ((float) (pendingSize + usedSize)* 
				  (float) avgRelSize*tertWeight)/
				      (float)cacheSize;
	if (xfrCost < switchTime) {
	    return 1;
	}
    }
    return 0;
}

/* ----------------------------------------------------------------
 *
 */

NegCacheLoadReq(req)
RequestPtr req;
{
    Element * joinList;
    int cl = 0;
    RequestPtr schedReq;

    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;
	
	joinEdge = (JoinEdgePtr) GetData(joinList);
	schedReq = joinEdge->reqPtr;
	if (!ReqCached(schedReq)) {
	    cl = req->cacheLoad;
	    break;
	}
    }
    return (-cl);
}

WorkSize(req)
RequestPtr req;
{
    switch (currDomain) {
    case ALL:
	return WorkSizeDoable(req);
    case ADJ:
	return WorkSizeInCache(req);
    case TWO_ADJ:
	return TwoAdjacentWorkSize(req);
    }
}

NegCoverSize(req)
RequestPtr req;
{
    Element * joinList;
    int cover = 0;
    RequestPtr schedReq;

    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;

	joinEdge = (JoinEdgePtr) GetData(joinList);
	schedReq = joinEdge->reqPtr;
	if (!ReqCached(schedReq)) {
	    cover += schedReq->cacheLoad;
	}
    }
    return (-cover);
}

OverlapCacheReq(req)
RequestPtr req;
{
    Element * joinList, * joinList2;
    int size = 0;
    RequestPtr schedReq;
    
    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;

	joinEdge = (JoinEdgePtr) GetData(joinList);
	schedReq = joinEdge->reqPtr;
	
	if (schedReq->relname == req->relname) {
	    continue;
	}
	ForEach(joinList2, schedReq->joinQuery) {
	    JoinEdgePtr joinEdge2;
	    
	    joinEdge2 = (JoinEdgePtr) GetData(joinList2);

	    if (ReqCached(joinEdge2->reqPtr)) {
		if (joinEdge2->reqPtr == req) continue;
		size += schedReq->cacheLoad;
		break;
	    }
	}
		
    }
    return (size);
}

NegLatency(req)
RequestPtr req;
{
    int lat;

    lat = - Abs(RelStartBlock(req->relname) - 
		  HeadPos(RelPlatName(req->relname)));
    return lat;
}

Any(req)
RequestPtr req;
{
    return (INFTY);
}

TwoAdjacentWorkSize(req)
RequestPtr req;
{
    return WorkSizeInCache(req) + OverlapCacheReq(req);
}

WorkSizeInCache(req)
RequestPtr req;
{
    Element * joinList;
    int workSize = 0;
    RequestPtr schedReq;

    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;

	joinEdge = (JoinEdgePtr) GetData(joinList);
	schedReq = joinEdge->reqPtr;
	if (ReqCached(schedReq)) {
	    workSize += schedReq->cacheLoad;
	}
    }
    /* A hack to make sure that single relation queries get
     * selected too
     */
    if (!IsEmpty(req->singleRelQuery)) {
	workSize += 1;
    }
    return (workSize);
}


WorkSizeDoable(req)
RequestPtr req;
{
    int workSize = 0;
    Element *joinList;

    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;
	
	joinEdge = (JoinEdgePtr) GetData(joinList);
	workSize += joinEdge->reqPtr->cacheLoad;
    }
    if (!IsEmpty(req->singleRelQuery)) {
	workSize += 1;
    }
    return workSize;
}



PlatHasReq(p)
int p;
{
    int i;

    for (i = 0; i < numDevice; i++) {
	RequestPtr req = (RequestPtr) (pendingRelList[i]);
	if (req != NULL) {
	    if (p == req->platter) {
		return 1;
	    }
	}
    }
    return 0;
}


RequestPtr    
RemoveBestReq(meritFunction, where, i)
int (*meritFunction)();
enum source where;
int i;
{
    int dummy, j;
    RequestPtr selReq;

    for (j = 0; j < numDevice; pendingRelList[j++] = NULL);
    GetPendingReqList(pendingRelList);

    switch (where) {
    case PLAT:
	selReq = SelectFromPlat(meritFunction, &dummy);
	break;
    case CACHE:
	selReq = SelectFromCache(meritFunction, &dummy);
	break;
    case LOADED_PLAT:
	selReq = SelectFromLoadedPlat(meritFunction, &dummy);
	break;
    case LOADED_CACHE:
	selReq = SelectFromLoadedCache(meritFunction, &dummy);
	break;
    case LOADED_2ADJ_CACHE:
	selReq = SelectFromLoadedPlat(meritFunction, &dummy);
	if (dummy == 0) {
	    /* if max merit is 0, we know that the request
	     * is not two adjacent to cache
	     */
	    selReq = NULL;
	}
	break; 
    case LOADED_PLAT_SEEK:
	selReq = SelectFromLoadedPlatSeek(meritFunction, &dummy);
	break;
    case THIS_PLAT:
	selReq = SelectFromThisPlat(meritFunction, &dummy, i);
	break;
    case THIS_PLAT_CACHE:
	selReq = SelectFromThisPlatCache(meritFunction, &dummy, i);
	break;
    case THIS_PLAT_2ADJ_CACHE:
	selReq = SelectFromThisPlat(meritFunction, &dummy, i);
	if (dummy == 0) {
	    /* if max merit is 0, we know that the request
	     * is not two adjacent to cache
	     */
	    selReq = NULL;
	}
	break;
    case THIS_PLAT_SEEK:
	selReq = SelectFromThisPlatSeek(meritFunction, &dummy, i);
	break;
    }
      
    if (selReq == NULL) {
	return (selReq);
    }
    RemoveRequest (selReq);
    return selReq;
}
    
    
RequestPtr    
SelectFromPlat(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    int i;
    Element *reqList;
    RequestPtr bestReq = NULL;
    int merit;

    *maxMerit = -INFTY;
    
    for (i = 0; i < numPlats; i++) {
	if (PlatHasReq(i)) continue;
	ForEach(reqList, perPlatterQ[i]) {
	    RequestPtr req;
	    req = (RequestPtr) GetData(reqList);

	    merit = meritFunction(req);
	    if (merit > *maxMerit) {
		*maxMerit = merit;
		bestReq = req;
	    }
	    if (*maxMerit == INFTY) {
		return bestReq;
	    }
	}
    }
    return bestReq;
}


RequestPtr
SelectFromThisPlat(meritFunction, maxMerit, i)
int (*meritFunction)();
int *maxMerit;
int i;
{
    Element *reqList;
    RequestPtr bestReq = NULL;
    int merit;

    *maxMerit = -INFTY;
    
    if (i == -1) return (bestReq);
    ForEach(reqList, perPlatterQ[i]) {
	RequestPtr req;
	req = (RequestPtr) GetData(reqList);
	
	merit = meritFunction(req);
	if (merit > *maxMerit) {
	    *maxMerit = merit;
	    bestReq = req;
	}
	if (*maxMerit == INFTY) {
	    return bestReq;
	}
    }
    return bestReq;
}


RequestPtr
SelectFromCache(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    Element *reqList, *joinList;
    RequestPtr bestReq = NULL, req2, req;
    ListX cacheList;

    int merit;
    *maxMerit = -INFTY;
 
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
	req2 = (RequestPtr) GetData(reqList);
	ForEach (joinList, req2->joinQuery) {
	    JoinEdgePtr joinEdge;
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    req = joinEdge->reqPtr;

	    if (ReqCached(req)) continue;
	    if (PlatHasReq(req->platter)) continue;

	    merit = meritFunction(req);
	    if (merit > *maxMerit) {
		*maxMerit = merit;
		bestReq = req;
	    }
	    if (*maxMerit == INFTY) {
		return bestReq;
	    }
	}
    }
    return bestReq;
}

RequestPtr
SelectFromThisPlatCache(meritFunction, maxMerit, i)
int (*meritFunction)();
int *maxMerit;
int i;
{
    Element *reqList, *joinList;
    RequestPtr bestReq = NULL, req2;
    int merit;

    *maxMerit = -INFTY;
    
    if (i == -1) return (bestReq);
	    
    ForEach(reqList, perPlatterQ[i]) {
	RequestPtr req;
	req = (RequestPtr) GetData(reqList);
	ForEach (joinList, req->joinQuery) {
	    JoinEdgePtr joinEdge;
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    req2 = joinEdge->reqPtr;
	    if (ReqCached(req2)) {
		
		merit = meritFunction(req);
		if (merit > *maxMerit) {
		    *maxMerit = merit;
		    bestReq = req;
		}
		if (*maxMerit == INFTY) {
		    return bestReq;
		}
		break;
	    }
	}
    }
    return bestReq;
}

RequestPtr    
SelectFromLoadedPlat(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    int i;
    Element *reqList;
    int loadedPlat[MAX_DEVICE];

    RequestPtr bestReq = NULL;
    int merit;
    *maxMerit = -INFTY;
    
    GetLoadedPlatters(loadedPlat);

    for (i = 0; i < numDevice; i++) {
	if (loadedPlat[i] == -1) continue;
	if (PlatHasReq(loadedPlat[i])) continue;
	ForEach(reqList, perPlatterQ[loadedPlat[i]]) {
	    RequestPtr req;
	    req = (RequestPtr) GetData(reqList);

	    merit = meritFunction(req);
	    if (merit > *maxMerit) {
		*maxMerit = merit;
		bestReq = req;
	    }
	    if (*maxMerit == INFTY) {
		return bestReq;
	    }
	}
    }
    return bestReq;
}



RequestPtr
SelectFromLoadedCache(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    Element *reqList, *joinList;
    RequestPtr bestReq = NULL, req2, req;
    ListX cacheList;
    int loadedPlat[MAX_DEVICE];

    int merit;
    *maxMerit = -INFTY;

    GetLoadedPlatters(loadedPlat);
    cacheList = GetCacheRequestList();

    ForEach (reqList, cacheList) {
	req2 = (RequestPtr) GetData(reqList);
	ForEach (joinList, req2->joinQuery) {
	    JoinEdgePtr joinEdge;
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    req = joinEdge->reqPtr;
	    
	    if (ReqCached(req)) continue;
	    if (PlatHasReq(req->platter)) continue;
	    if (!IsAnElementOf(req->platter, loadedPlat, numDevice)) {
		continue;
	    }
	    
	    merit = meritFunction(req);
	    if (merit > *maxMerit) {
		*maxMerit = merit;
		bestReq = req;
	    }
	    if (*maxMerit == INFTY) {
		return bestReq;
	    }
	}
    }
    return bestReq;
}



RequestPtr    
SelectFromLoadedPlatSeek(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    int i;
    int loadedPlat[MAX_DEVICE];
    RequestPtr bestReq = NULL, req;
    int merit;
    *maxMerit = -INFTY;
    
    GetLoadedPlatters(loadedPlat);

    for (i = 0; i < numDevice; i++) {
	if (loadedPlat[i] == -1) continue;
	if (PlatHasReq(loadedPlat[i])) continue;
	req = GetNearestRequest(loadedPlat[i], i);
	if (req == NULL) continue;

	merit = meritFunction(req);
	if (merit > *maxMerit) {
	    *maxMerit = merit;
	    bestReq = req;
	}
	if (*maxMerit == INFTY) {
	    return bestReq;
	}
    }
    return bestReq;
}


RequestPtr
SelectFromTwoAdjCache(meritFunction, maxMerit)
int (*meritFunction)();
int *maxMerit;
{
    Element *reqList, *joinList;
    RequestPtr bestReq = NULL, req2, req;
    ListX cacheList;

    int merit;
    *maxMerit = -INFTY;
 
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
	req2 = (RequestPtr) GetData(reqList);
	ForEach (joinList, req2->joinQuery) {
	    JoinEdgePtr joinEdge;
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    req = joinEdge->reqPtr;

	    if (ReqCached(req)) continue;
	    if (PlatHasReq(req->platter)) continue;

	    merit = meritFunction(req);
	    if (merit > *maxMerit) {
		*maxMerit = merit;
		bestReq = req;
	    }
	    if (*maxMerit == INFTY) {
		return bestReq;
	    }
	}
    }
    return bestReq;
}

RequestPtr
SelectFromThisPlatSeek(meritFunction, maxMerit, i)
int (*meritFunction)();
int *maxMerit;
int i;
{
    Element *reqList;
    RequestPtr bestReq = NULL, req;
    int merit;

    *maxMerit = -INFTY;
    
    if (i == -1) return (bestReq);
    req = (RequestPtr) GetFirstData(perPlatterQ[i]);	

    merit = meritFunction(req);
    if (merit > *maxMerit) {
	*maxMerit = merit;
	bestReq = req;
    }
    if (*maxMerit == INFTY) {
	return bestReq;
    }

    return bestReq;
}


RequestPtr
GetNearestRequest(plat, deviceNum)
int plat;
int deviceNum;
{
    int currPos;
    RequestPtr selReq = NULL;
    int leastDist = INFTY;
    Element * reqList;

    currPos = GetCurrentBlock(deviceNum);

    ForEach (reqList, perPlatterQ[plat]) {
	RequestPtr req;
	int dist;

	req = (RequestPtr) GetData(reqList);

	dist = Abs(RelStartBlock(req->relname) - currPos);
	if (dist < leastDist) {
	    leastDist = dist;
	    selReq = req;
	}
    }
    return (selReq);
}


int
ChoosePlatter(meritFunction, where)
int (*meritFunction)();
enum source where;
{
    int i;
    int merit;
    int bestPlat = -1;
    int maxMerit = -INFTY;
    Element * reqList;

    switch (where) {
    case TWO_ADJ_CACHE:
    case PLAT:
	for (i = 0; i < numPlats; i++) {
	    
	    if (IsEmpty(perPlatterQ[i])) continue;
	    if (PlatHasReq(i)) continue;
	    merit = 0;
	    ForEach (reqList, perPlatterQ[i]) {
		RequestPtr req;
		
		req = (RequestPtr) GetData(reqList);
		merit += meritFunction(req);
	    }
	    
	    if (merit > maxMerit) {
		maxMerit = merit;
		bestPlat = i;
	    }
	    if (maxMerit == INFTY) {
		return bestPlat;
	    }
	}
	return bestPlat;

    case CACHE:
    {
	int perPlatterMerit[MAX_PLAT];
	RequestPtr req2, req;
	ListX cacheList;
	Element * joinList;
	int perPlatterAdj[MAX_PLAT];

	cacheList = GetCacheRequestList();
	for (i = 0; i < numPlats; i++) {
	    perPlatterMerit[i] = 0;
	    perPlatterAdj[i] = 0;
	}
	ForEach (reqList, cacheList) {
	    req2 = (RequestPtr) GetData(reqList);

	    ForEach (joinList, req2->joinQuery) {
		JoinEdgePtr joinEdge;
		joinEdge = (JoinEdgePtr) GetData(joinList);
		req = joinEdge->reqPtr;
		if (ReqCached(req)) continue; 
		
		perPlatterAdj[req->platter] = 1;
		perPlatterMerit[req->platter] += meritFunction (req);
		if (perPlatterMerit[req->platter] == INFTY) {
		    return req->platter;
		}
	    }
	}
	for (i = 0; i < numPlats; i++) {
	    if (! perPlatterAdj[i]) continue;
	    if (PlatHasReq(i)) continue;
	    if (perPlatterMerit[i] > maxMerit) {
		maxMerit = perPlatterMerit[i];
		bestPlat = i;
	    }
	}
	return bestPlat;
    }
    }
}

