/*---------------------------------------------------
 * scheduling algorithms for queries involving single-relation queries
 * or joins.
 */

#include <stdio.h>
#include <string.h>
#include "parameter.h"
#include "listSched.h"
#include "querySched.h"
#include "schedInterface.h"
#include "request.h"

/* 
 * globals 
 */
ListX perPlatterQ[MAX_PLAT]; 
RequestPtr * perRelQ;
TimeType epoch;

static RequestPtr uncacheReq1 = NULL, uncacheReq2 = NULL;

/*
 * routines declared in this file
 */
RequestPtr MinBuffer();
RequestPtr SelectARequest();
int NoCachingRequired();
void IncludeNewQuery();
RequestPtr GetRequestPtr();
RequestPtr CreateNewRequestNode();
RequestPtr FilterDoableJoins();
QueryPtr RemQueryfromRequest();
float ReqCost();

extern char * relUsed;

extern ListX GetCacheRequestList();

/* ----------------------------------------------------------------
 * initialze query scheduler data structures
 */
InitPolicyMaker()
{

    perRelQ = (RequestPtr *) MallocErrorCheck(sizeof(char *)*numRels);
    bzero(perRelQ, sizeof(char *)*numRels);
    relUsed = (char *) MallocErrorCheck(numRels);
}

/*--------------------------------------------------------------------
 * if this request has at least one edge that goes to a request on a
 * fragment that is not in cache, we will need to cache it.  Joins between
 * relations on the same platter both of which are uncached are
 * considered as requiring caching only when `whence` is set to NOW.
 * returns 1 is caching is not required, 0 otherwise. 
 */
int
NoCachingRequired(req, whence)
RequestPtr req;
TimeType whence;
{
    Element *joinList;
    
    if (req->cacheLoad == 0) {
	return (1);
    }
    
    if (ReqCached(req)) {
	return(1);
    }
    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr joinEdge;

	joinEdge = (JoinEdgePtr) GetData(joinList);

	if (ReqCached(joinEdge->reqPtr)) {
	    continue;
	}
	if (whence == NOW) {
	    return (0);
	}
	if (req->platter == joinEdge->reqPtr->platter) {
	    continue;
	}
	return (0);
    }
    return (1);
}

RemoveRequest(req)
RequestPtr req;
{
    RemoveGivenElfromList (&(perPlatterQ[req->platter]), (ElemType *)req);
}

/* -------------------------------------------------------------------
 * ExecuteQuery() -- execute all the single relation queries on the
 * platter first followed by all the join queries that can be executed
 * with the data available in cache now. So, if we do not need to
 * cache the fragment all queries on it can be executed.  assume
 * cacheLoad is set to 0 only when it is not necessary to cache the
 * relation at all -- for example if both components of a join are on
 * the same platter at least one of the request must have cacheload
 * set to a non-0 value, also for a self join the cacheload must be
 * non-zero.  XXX -- make sure cacheLoad field is set properly. 
 */


QueryPtr
NextQueryInCurrReq(req)
RequestPtr req;
{    
    QueryPtr query = NULL;
    Element *reqList;
    ListX cacheList;

    query = RemQueryfromRequest(req);
    if (query != NULL) {
	return (query);
    }
    /*
     * check if any of the requests already cached have
     * some pending queries
     */
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
	RequestPtr nreq;
	
	nreq = (RequestPtr) GetData(reqList);
	query = RemQueryfromRequest(nreq);
	if (query != NULL) {
	    return (query);
	}
    }
    return (NULL);
}


QueryPtr
GetNextQuery()
{    
    ListX cacheList;
    Element * reqList;
    QueryPtr query;

    InitialWork();
    /*
     * check if any of the requests already cached have
     * some pending queries
     */
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
	RequestPtr nreq;
	
	nreq = (RequestPtr) GetData(reqList);
	query = RemQueryfromRequest(nreq);
	if (query != NULL) {
	    return (query);
	}
    }
    return (NULL);
}

int
QueryAvailable()
{
    ListX cacheList;
    Element * reqList;
    Element * joinList;

    /*
     * check if any of the requests already cached have
     * some pending queries
     */
    cacheList = GetCacheRequestList();
    ForEach (reqList, cacheList) {
        RequestPtr req;
	
	req = (RequestPtr) GetData(reqList);
	if (!IsEmpty(req->singleRelQuery)) {
	  return 1;
	}
	ForEach (joinList, req->joinQuery) {
	    JoinEdgePtr joinEdge;
	
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    if (ReqCached(joinEdge->reqPtr)) {
	        return 1;
	    }
	}
    }
    return 0;
}

QueryPtr
RemQueryfromRequest(req)
RequestPtr req;
{
    QueryPtr query = NULL;
    Element * joinList;


    if (req == NULL) return (NULL);
    /*
     * execute all single relation queries
     */
    query = (QueryPtr) RemoveElfromList(&(req->singleRelQuery));
    /*
     * we do not want to execute join queries if both of its
     * components are not available
     */
    if (query == NULL) { 
	ForEach (joinList, req->joinQuery) {
	    JoinEdgePtr joinEdge;
	
	    joinEdge = (JoinEdgePtr) GetData(joinList);
	    if (ReqCached(joinEdge->reqPtr)) {

		RemoveGivenElfromList(&(req->joinQuery), joinList);
		RemovefromOtherReq(joinEdge);

		query = joinEdge->query;
		break;
	    }
	}
    }	
    if (IsEmpty(req->singleRelQuery) && IsEmpty(req->joinQuery)) {
	if (ReqCached(req)) {
	    uncacheReq1 = req;
	}
    }
    return (query);
}
   
/* ---------------------------------------------------------------
 * remove the query from the other cached request, since for joins the
 * query is enqueued in two places. If the cached req does not have
 * any more pending requests uncache it 
 */

RemovefromOtherReq(joinEdge)
JoinEdgePtr joinEdge;
{
    Element *joinList;
    RequestPtr req;



    req = joinEdge->reqPtr;

    ForEach (joinList, req->joinQuery) {
	JoinEdgePtr join;
	
	join = (JoinEdgePtr) GetData(joinList);
	if (join->query == joinEdge->query) {
	    RemoveGivenElfromList(&(req->joinQuery), 
				  joinList);
	    break;
	}
    }

    if (IsEmpty(req->joinQuery) && IsEmpty(req->singleRelQuery)) {
	uncacheReq2 = req;

    }
}

CleanUpCachedReq()
{
    if (uncacheReq1 != NULL) {
	UncacheRequest(uncacheReq1);
    }
    if (uncacheReq2 != NULL) {
	UncacheRequest(uncacheReq2);
    }
    uncacheReq1 = NULL;
    uncacheReq2 = NULL;
}

/* -------------------------------------------------------------------
 * Include all newly arriving queries into the scheduler data structure.
 */    
InitialWork()
{
    while (! IsEmpty(scheduleList)) {
	QueryPtr query;

	query = (QueryPtr) RemoveElfromList(&scheduleList);
	IncludeNewQuery(query);
    }
}

/*-------------------------------------------------------------------
 * A newly arriving query is inserted into the scheduler data
 * structure.  The data structure consists of a list of fragment
 * request on each platter sorted with request to total execution time
 * of all queries on the fragment.
 */

void
IncludeNewQuery(query)
QueryPtr query;
{

    if (!SS_JoinQuery(query)) {
	RequestPtr req;

	req = GetRequestPtr(query->rel[0]);
	if (req == NULL) {
	    /* create a new request node */
	    req = CreateNewRequestNode(query->rel[0]);
	    req->cacheLoad = RelSize(query->rel[0]);
	    InsertSorted(&(perPlatterQ[req->platter]), req, ReqCost); 
	}
	AddEltoList (&(req->singleRelQuery), (ElemType *) query);

    } else {
	RequestPtr req1, req2;
	JoinEdgePtr joinEdge1, joinEdge2;
	
	/* 
	 * A join query, insert first the inner than the outer
	 * relation
	 */
	req1 = GetRequestPtr (query->rel[0]);
	if (req1 == NULL) {
	    /* create a new request node */
	    req1 = CreateNewRequestNode(query->rel[0]);
	    InsertSorted(&(perPlatterQ[req1->platter]), req1, ReqCost); 
	} 
	/* 
	 * At this point we have a request node 
	 */
	req1->cacheLoad = RelSize(query->rel[0]);

	joinEdge1 = (JoinEdgePtr) 
	    MallocErrorCheck(sizeof(JoinEdgeData));
	joinEdge1->direction = FORWARD;
	joinEdge1->query = query;
	AddEltoList (&(req1->joinQuery), (ElemType *) joinEdge1);
	
	/* insert the outer relation request */
	req2 = GetRequestPtr (query->rel[1]);
	if (req2 == NULL) {
	    req2 = CreateNewRequestNode(query->rel[1]);
	    InsertSorted(&(perPlatterQ[req2->platter]), req2, ReqCost); 
	}
	req2->cacheLoad = RelSize(query->rel[1]);
	
	joinEdge2 = (JoinEdgePtr) 
	    MallocErrorCheck(sizeof(JoinEdgeData));
	joinEdge2->query = query;
	joinEdge2->direction = BACKWARD;
	AddEltoList (&(req2->joinQuery), (ElemType *) joinEdge2);
	/*
	 * make the request nodes of the two join components
	 * point to each other
	 */
	joinEdge2->reqPtr = req1;
	joinEdge1->reqPtr = req2;
    }
}


/*----------------------------------------------------------------
 * return the request pointer that contains request on relation `rel`. 
 * If no request on relation `rel` has  been made so far, NULL is 
 * returned.
 */ 
RequestPtr
GetRequestPtr(rel)
int rel;
{
    RequestPtr req = perRelQ[rel];
    
    if (req == NULL) {
	return (NULL);
    }
    
    if (ReqCached(req)) {
	statxQryIn++;
	return (req);
    }
    if (IsEmpty(req->singleRelQuery) && IsEmpty(req->joinQuery)) { 
	return (NULL);
    }
    return (req);
}

/*-----------------------------------------------------------------
 * CreateNewRequestNode() -- create a new node for requests on relation
 * `rel` and initialize the fields
 */
RequestPtr
CreateNewRequestNode(rel)
int rel;
{	   
    RequestPtr req;

    req = perRelQ[rel];
    if (req == NULL) {
	req = (RequestPtr) MallocErrorCheck(sizeof(RequestData));
	req->relname = rel;
	req->platter = RelPlatName(rel);
	req->cost = DataFetchCost(rel);
	perRelQ[rel] = req;
	InitList(&(req->singleRelQuery));
	InitList(&(req->joinQuery));
    }
    return (req);
}


EnqueueRequest(req)
RequestPtr req;
{
    if (!IsEmpty(req->singleRelQuery) || !IsEmpty(req->joinQuery)) {
	InsertSorted(&(perPlatterQ[req->platter]), req, ReqCost); 
    }
}

Adjacent (req, rel)
RequestPtr req;
int rel;
{
    Element * joinList;
    JoinEdgePtr joinEdge;

    ForEach (joinList, req->joinQuery) {
	joinEdge = (JoinEdgePtr) joinList;
	if (joinEdge->reqPtr->relname == rel) {
	    return (1);
	}
    }
    return (0);
}
/* -------------------------------------------------------------------
 *
 */
float ReqCost(req)
RequestPtr req;
{
    return ((float) RelStartBlock(req->relname));
}

/* -------------------------------------------------------------
 *
 */
int 
NumQueries(req)
RequestPtr req;
{
    return ListLength(req->joinQuery) + ListLength(req->singleRelQuery);
}
/* ------------------------------------------------------------------
 * debug and statistics utilities
 */

PrintReqStatus()
{
    int i;

    for (i = 0; i < numPlats; i++) {
	if (IsEmpty(perPlatterQ[i])) {
	    continue;
	}
	printf ("\nPlatter %d: ", i);
	PrintList(perPlatterQ[i], PrintReqStruct);
    }
    printf ("\n Cache:");
    PrintList(GetCacheRequestList(), PrintReqStruct);
    printf ("\n");
}

void
PrintReqStruct(req)
RequestPtr req;
{
    printf ("\n\t req: p%2d r%3d s%4d c%6f \t sq num: %2d \t jq:", 
	    req->platter, req->relname, req->cacheLoad, req->cost,
	    ListLength(req->singleRelQuery));
    
    PrintList (req->joinQuery, PrintJoinStruct);
}

void
PrintJoinStruct(joinEdge)
JoinEdgePtr joinEdge;
{	
    printf ("%d %d %d: ", joinEdge->reqPtr->relname, 
	    joinEdge->query->rel[0], joinEdge->query->rel[1]);
}


StatxEstimateDataDemand()
{
    int i;
    Element * reqList;    
    
    statxSharedAccess = 0.0;
    statxUnsharedAccess = 0.0;
    for (i = 0; i < numPlats; i++) {
	if (IsEmpty(perPlatterQ[i])) {
	    continue;
	}

	ForEach (reqList, perPlatterQ[i]) {
	    RequestPtr req;
	    float relAccessTime;
	    
	    req = (RequestPtr) GetData(reqList);
	    relAccessTime = StatxEstimateAccessTime(RelSize(req->relname));
	    statxSharedAccess += relAccessTime;
	    statxUnsharedAccess += relAccessTime*(ListLength(req->joinQuery) 
				       +ListLength(req->singleRelQuery));
	}
    }
}















