/*
 * main.c ---
 *
 *  Assumptions:
 *   o The only kind of queries handled are
 *     single relation query -- index scan and sequential scan
 *     2-way join query -- nested loop without any qualification on the
 *                         component relation
 *
 *-------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include "parameter.h"
#include "listSched.h"
#include "querySched.h"
#include "request.h"

#define CPU 1
#define IO 2

extern RequestPtr ScheduleNextRequest();
extern QueryPtr ScheduleNextQuery();
extern int firstRelPolicy;
extern QueryPtr GetNextQuery();

extern int numCPUworkers;
extern int numIOworkers;
/*
 * configuration variables
 */
int numUsers = 0;
int maxQueries = MAX_QUERY;
int startMeasurement = 0;
int endMeasurement = MAX_QUERY;
float timeToNextQuery = TIME_TO_NEXT_QUERY;
int joinPercent = 80;
int maxPlats = MAX_PLAT;
int limitPlats = INFTY;
int platCapacity = PLAT_CAPACITY;
int maxPartSize = INFTY;
float switchTime = SWITCH_TIME;
int numPlats = MAX_PLAT;
int numDevice = 3;
int minSize = MIN_SIZE;
int maxSize = MAX_SIZE;
int numLogicalRels = NUM_RELS;
int numRels = NUM_RELS;
int maxFragmentSize = -1;
int cacheSize = CACHE_SIZE;
int cachePercent = 0;
int saveWkld = 0;
int readWkld = 0;
FILE * wkldFile;
int detectMean = 0;
int batchMode  = 0;
int fractionLarge = 0;
int fractionLargeQueries = 0;
int numPerGrp= 1;
int ageFactor = 4;
char outputFormat = 'd';
int seqScanPercent = 20;
int tertNum=1;
int overLap = 1;
char layout = 'a';

static char cpuIsFree = 1;
static char ioIsFree = 1;


/*
 * routines declared in this file
 */
void CPUJobDone PROTO((char * token));
void IOJobDone PROTO((char * token));
void SubmitJobs PROTO((void));
void InitWorld PROTO(());

static void ProcessLocalQuery PROTO((QueryPtr query));
static void ProcessIOReq PROTO((char * req, int rel));
static int NextIOJob PROTO(());
static int NextCPUJob PROTO(());
static int ExecuteNextQOJob PROTO(());
/*-----------------------------------------------------------------
 * InitWorld -- the simulator configuration parameters are read
 * from the command line and initialized appropriately
 */

void
InitWorld(argc, argv)
int argc;
char ** argv;
{
    int schedFlag = 1;
    int evictFlag = 0;
    char queryFlag;
    char wkldFileName[100];
    char * IOFileName = NULL;

    int opt;
    extern char *optarg;
    extern int optind, opterr;

    optind = 1;
    opterr = 0;
    while ((opt = getopt (argc, argv, 
	"a:c:bd:e:f:i:j:k:ln:o:p:q:r:s:tw:A:B:C:D:E:F:I:L:M:N:P:R:S:T:U:V:W:")) 
	   != EOF) {
        switch (opt) {
	case 's':
	    schedFlag = atoi(optarg);
	    break;
	case 'a':
	    seqScanPercent = atoi(optarg);
	    break;
	case 'c':
	    /*
	     * cache size in megabytes
	     */
	    cacheSize = atoi(optarg);
	    break;
	case 'd':
	    debug_set(optarg);
	    break;
	case 'e':
	    evictFlag = atoi(optarg);
	    break;
	case 'p':
	    queryFlag = 's';
	    strcpy (wkldFileName, optarg);
	    break;
	case 'r':
	    queryFlag = 'r';
	    strcpy ( (char *)wkldFileName, optarg);
	    break;	
	case 'j':
	    joinPercent = atoi(optarg);
	    break;
	case 'I':
	    IOFileName = MallocErrorCheck(strlen(optarg) + 1); 
	    strcpy ((char *)IOFileName, optarg);
	    break;
	case 'T':
	    tertNum = atoi(optarg);
	    break;
	case 'b':
	    batchMode = 1;
	    break;
	case 'o':
	    outputFormat = optarg[0];
	    break;
	case 't':
	    detectMean = 1;
	    break;
	case 'k':
	    numPerGrp = atoi(optarg);
	    break;
	case 'l':
	    overLap = 0;
	    break;
	case 'i':
	    srand(atoi(optarg));
	    srand48(atoi(optarg));
	    break;
	case 'f':
	    fractionLarge = atoi(optarg);
	    break;
	case 'q':
	    fractionLargeQueries = atoi(optarg);
	    break;
	case 'P':
	    maxPartSize = atoi(optarg);
	    break;
	case 'F':
	    maxFragmentSize = atoi(optarg);
	    break;
	case 'D':
	    platCapacity = atoi(optarg);
	    break;
	case 'U':
	    numUsers = atoi(optarg);
	    if (numUsers > MAX_USERS) {
		printf ("Number of users cannot be > %d\n", 
			    MAX_USERS);
		exit(0);
	    }
	    break;
	case 'M':
	    maxQueries = atoi(optarg);
	    break;

	case 'B':
	    startMeasurement = atoi(optarg);
	    break;
	case 'E':
	    endMeasurement = atoi(optarg);
	    break;
	case 'R':
	    numLogicalRels = atoi(optarg);
	    break;
	case 'S':
	    minSize = atoi(optarg);
	    break;
	case 'L':
	    maxSize = atoi(optarg);
	    break;
	case 'C':
	    cachePercent = atoi(optarg);
	    cacheSize = (cachePercent*(minSize + maxSize)*numLogicalRels)/200;
	    break;
	case 'A':
	    limitPlats = atoi(optarg);
	    break;
	case 'W':
	    numCPUworkers = atoi(optarg);
	    break;
	case 'V':
	    numIOworkers = atoi(optarg);
	    break;
	default :
	    printf ("<runfile> -s # -p <query store file> -r <query read file> -d <debug>\n");
	    printf ("\t s-opts - 1:Fcfs, 2:MinAvPlat, 3:Sjn, 4:MinAvgGrp\n"); 
	    printf ("\t debug options - q: queries submitted, e: events\n");
	}
    }

    /*-------------------------------------
     * some sanity checks
     */
    if ((endMeasurement == MAX_QUERY) || (endMeasurement > maxQueries)) {
	endMeasurement = maxQueries;
    }

    if (startMeasurement >= endMeasurement) {
	startMeasurement = 0;
    }
    if (queryFlag == 'r') {
	readWkld = 1;
	wkldFile = fopen(wkldFileName, "r");
    }
    if (queryFlag == 's') {
	saveWkld = 1;
	wkldFile = fopen(wkldFileName, "w");
    }
    InitTertiaryMem(IOFileName, tertNum);
    InitScheduler(schedFlag, evictFlag);
    InitCache();
    SS_InitQueries();
    InitLocalPlan(); /* call after InitTM */
    InitPolicyMaker();

}

/* ----------------------------------------------------
 *
 */
    
void
SubmitJobs()
{
    while (IoIsFree()) {
	if (NextIOJob() == 0) break;
    }
    while (CpuIsFree()) {
	if (NextCPUJob() == 0) break;
    }
}

void
IOJobDone(token)
char * token;
{
    ioIsFree = 1;
    if (QueryOriented()) {
	CacheRel(*((int *)token));
	PinRel(*((int *)token));
    } else {
	CacheRequest((RequestPtr) token);
    }
    while (CpuIsFree()) {
	if (NextCPUJob() == 0) break;
    }
    NextIOJob();
}

void
CPUJobDone(token)
char * token;
{
    QueryPtr query = (QueryPtr) token;

    cpuIsFree = 1;
    
    UnpinRel(query->rel[0]);
    UnpinRel(query->rel[1]);
    statxProcessing += query->cost;
    if (is_debug("q")) {
	PrintQueryStruct(query);
    }
    NextCPUJob();
    while (IoIsFree()) {
	if (NextIOJob() == 0) break;
    }
}

static int
NextCPUJob()
{
    QueryPtr query;

    if (!overLap && !ioIsFree) {
	return 0;
    }
    if (QueryOriented()) {
	return ExecuteNextQOJob(CPU);
    }
    query = GetNextQuery();
    if (query == NULL) return 0;
    ProcessLocalQuery(query);
    return 1;
}

static int
NextIOJob()
{
    static RequestPtr req = NULL;
    
    if (!overLap && !cpuIsFree) {
	return 0;
    }
    if (QueryOriented()) {
	return ExecuteNextQOJob(IO);
    }
    if (req == NULL) {
 	req = ScheduleNextRequest();
	if (req == NULL) return 0;
    }
    if (! MakeSpaceInCache(req->cacheLoad, req->relname)) {
	return 0;
    }
    ProcessIOReq((char *) req, req->relname);
    req = NULL;
    return 1;
}

static int
ExecuteNextQOJob(jobType)
int jobType;
{
    static QueryPtr ioQuery = NULL, cpuQuery = NULL;
    static ListX queryList = {NULL, NULL}; 
    QueryPtr submitQuery;
    static int * rel = NULL;
    
    switch (jobType) {
    case CPU:
	if (cpuQuery == NULL) {
	    cpuQuery = (QueryPtr) RemoveElfromList(&queryList);
	    if (cpuQuery == NULL) return 0;
	}
	if (RelCached(cpuQuery->rel[0]) && RelCached(cpuQuery->rel[1])) {
	    submitQuery = cpuQuery;
	    cpuQuery = NULL;
	    ProcessLocalQuery(submitQuery);
	    return (1);
	}
	return 0;
    case IO:
	while (rel == NULL) {
	    if (ioQuery == NULL) {
		ioQuery = ScheduleNextQuery();
		if (ioQuery == NULL) return 0;
		AddEltoList(&queryList, (ElemType *) ioQuery);
		if (NeedsFetching(ioQuery->rel[0])) {
		    rel = &(ioQuery->rel[0]); 
		} else {
		    PinRel(ioQuery->rel[0]);
		}
	    }
	    if (SS_JoinQuery(ioQuery) && NeedsFetching (ioQuery->rel[1])){
		if (rel == NULL) {
		    rel = &(ioQuery->rel[1]);
		    ioQuery = NULL;
		}
	    } else {
		PinRel(ioQuery->rel[1]);
		ioQuery = NULL;
	    }
	    
	}
	if (MakeSpaceLRU(*rel)) {	 
	    ProcessIOReq( (char *)rel, *rel);
	    rel = NULL;
	    return 1;
	}
	return 0;
    }
}


static void
ProcessLocalQuery(query)
QueryPtr query;
{
    cpuIsFree = 0;

    if (!QueryOriented()) {
	PinRel(query->rel[0]);
	PinRel(query->rel[1]);
    }
    PGRunQuery(query);
    return;
}

static void
ProcessIOReq(req, rel)
char * req;
int rel;
{
    ioIsFree = 0;
    RegisterRel(rel);
    PGCacheRel(req, rel);
/* CHANGE
    DataFetchSubmit(req, rel);
*/
}










