#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>

#include "parameter.h"
#include "listSched.h"
#include "querySched.h"

#include "postgres.h"
#include "nodes/pg_list.h"
#include "utils/rel.h"
#include "nodes/plannodes.h"
#include "nodes/primnodes.h"
#include "nodes/parsenodes.h"
#include "tcop/schedule.h"
#include "tcop/dest.h"
#include "mariposa_rpc.h"
#include "rpc/rpcutils.h"

#include "sched.h"

#define MAX_WORKERS     MAX_USERS + MAX_DEVICE
#define MAX_NUM_OPTS    64

/*
 * globals
 */
ListX cpuTokenList[MAX_USERS];
char *ioTokenList[MAX_DEVICE];
int workerBusy[MAX_DEVICE];
int platterIds[MAX_PLAT];
int relationIds[NUM_RELS];
int userIds[MAX_USERS];
int ioIds[MAX_DEVICE];
int avgRelSize; /* CHANGE -- SET IT PROPERLY */

int numCPUworkers = 0;
int numIOworkers = 0;
int numWorkers;

SchedFragmentArg * FormCacheRelArg(Oid relid);
SchedFragmentArg * FormUncacheRelArg(Oid relid);

ListX localList;
pthread_mutex_t schedLock;
pthread_cond_t schedTaskAvail;
pthread_cond_t schedEventAvail;

static int * PGRelInfo(Oid relid);
static char ** ExtractSchedOptions(int argc, char **argv, int *numopts);  
/* ---------------------------------------------------------------
 * called from sitemgr_main routine to initialize the scheduler
 * data stucture -- entry point of the main scheduler thread -- should
 * never return from here.
 */
void
InitSchedMain(int argc, char **argv)
{
    int i;
    int numOptions;

    InitWorld(argc, argv);

    pthread_mutex_init(&schedLock, pthread_mutexattr_default);
    pthread_cond_init(&schedTaskAvail, pthread_condattr_default);
    pthread_cond_init(&schedEventAvail, pthread_condattr_default);

    relationIds[0] = userIds[0] = platterIds[0] = ioIds[0] = -1;

    InitList(&localList);
    for (i = 0; i < MAX_USERS; i++) {
	InitList(&cpuTokenList[i]);
    }

    MainLoop();
}

/* ----------------------------------------------------------------------
 * a list of subqueries has arrived from a client with process id "pid". The
 * query list needs to be parsed and enqueued at proper places
 * -----------------------
 */
void
SchedSubmitQuery(void *arg)
{
    char **argv = (char **) arg;
    SVCXPRT *transp = (SVCXPRT *)argv[1]; /* Requesting site */

    sitemgrDebug(3, "In Submit query ..\n");
    SS_SubmitQuery(argv[0]);
    /*
     * send reply to the caller site
     */
    if (!svc_sendreply(transp, xdr_void, NULL)) {
	sitemgrDebug(0,
		     "sitemgrprog_1 : cannot reply for get_next_sq \n");
	exit(1);
    }
}

/* ---------------------------------------------
 * notify the scheduler that the queries in the "list" are done
 */
void
SchedSubQueryDone(void * arg)
{
    char **argv = (char **) arg;
    SVCXPRT *transp = (SVCXPRT *)argv[1]; /* Requesting site */

    int pid = *((int *)(argv[0]));
    int i;

    sitemgrDebug(3, "In SubQueryDone: pid = %d\n", pid); 
    i = GetInternalUid(pid);
    while (!IsEmpty(cpuTokenList[i])) {
	QueryPtr query = (QueryPtr) RemoveElfromList(&cpuTokenList[i]);
	SS_QueryDone(query);
    }
    /*
     * send reply to the caller site
     */
    if (!svc_sendreply(transp, xdr_void, NULL)) {
	sitemgrDebug(0,
		     "sitemgrprog_1 : cannot reply for get_next_sq \n");
	exit(1);
    }
} 

/* --------------------------------------------------
 * submit in "list" the subqueries which are ready for execution
 * by the backend with whose "pid" is supplied
 */

void
SchedGetNextSQ(void * arg)
{
    char **argv = (char **) arg;
    SVCXPRT *transp = (SVCXPRT *)argv[1]; /* Requesting site */

    int pid = *((int *)(argv[0]));
    List *list = NIL;
    QueryPtr query;
    int i;

    sitemgrDebug(3, "In Get next SQ, pid = %d\n", pid);
    i = GetInternalUid(pid);

    pthread_mutex_lock(&schedLock); 

    while (IsEmpty(cpuTokenList[i])) {
	Element * qlist;
	ForEach (qlist, localList) {
	    query = (QueryPtr) GetData(qlist);
	    if (query->user == i) {
		RemoveGivenElfromList(&localList, qlist);
		AddEltoList(&cpuTokenList[i], qlist);
		list = nconc(list, (List *) query->squery);
	    }
	}
	if (IsEmpty(cpuTokenList[i])) {
	    pthread_cond_wait(&schedTaskAvail, &schedLock);
	} 
    }    
    pthread_mutex_unlock(&schedLock); 

    /*
     * send reply to the caller site
     */
    if (!svc_sendreply(transp, xdr_Node_list, &list)) {
	sitemgrDebug(0,
		     "sitemgrprog_1 : cannot reply for get_next_sq \n");
	exit(1);
    }
    sitemgrDebug(3, "get_next_sq succeeds .... \n");
}

/* ------------------------------------------------------
 * check if all the subqueries submitted by this user are done
 * -------------------
 */
int
CheckIfDone (void * arg)
{
    int pid = *(int *)arg;

    return SS_CheckDone(GetInternalUid(pid));
}

/* ------------------------------------------------------
 * notify backend about completion of an IO job
 */
void
SchedBackendDone(int pid) 
{
    
    int i = GetInternalIOid(pid);
    sitemgrDebug(3, "IO done %d: \n", i);

    workerBusy[i] = 0;
    if (ioTokenList[i] != NULL) {
	SS_CacheRelDone(ioTokenList[i]);
    } else {
	SS_UnCacheRelDone();
    }
    return;
} 

/*---------------------------------------------------------------------
 * run a query using a postgres backend
 */
void
PGRunQuery(query)
QueryPtr query;
{
    int workNum;


    pthread_mutex_lock(&schedLock);
    AddEltoList(&localList, (ElemType *)query);
    pthread_cond_broadcast(&schedTaskAvail);
    pthread_mutex_unlock(&schedLock);
}

/* ----------------------------------------------------------------------
 * fetch a fragment from TM using another postgres backend
 */
void
PGCacheRel(token, rel)
char * token;
int rel;
{
    int workNum;
    int pid;

    sitemgrDebug(3, "In PGCacheRel %d \n", rel);
    if ((pid = SchedRequestSubmit(FormCacheRelArg(GetExternalRelid(rel)))) 
	== INVALID_PID) {
	PrintError("Cache request failed");
    }
    workNum = GetInternalIOid(pid);
    ioTokenList[workNum] = token;
    workerBusy[workNum] = 1;
}

/* --------------------------------------------------------------------
 * Uncache a relation from disk cache
 */
void
PGUncacheRel(int rel)
{
    int workNum;
    int pid;

    sitemgrDebug(3, "In PGUnCacheRel %d \n", rel);
    if ((pid = SchedRequestSubmit(FormUncacheRelArg(GetExternalRelid(rel)))) == INVALID_PID) {
	PrintError("Cache request failed");
    }
    workNum = GetInternalIOid(pid);
    ioTokenList[workNum] = NULL;
    workerBusy[workNum] = 1;
}

/* -----------------------------------------------------------------------*/
CpuIsFree()
{
    return 1;
}

IoIsFree()
{
    int i, numBusy = 0;

    for (i = 0; i < MAX_DEVICE; i++) {
	if (workerBusy[i]) numBusy++;
    }
    if (numBusy >= numDevice) 
	return 0;
    return 1;
}

GetPendingReqList(reqList)
char ** reqList;
{
    int i, j;
    j = 0;
    for (i = numCPUworkers; i < numWorkers; i++) {
	if (workerBusy[i] && ioTokenList[i]) {
	    reqList[j++] = ioTokenList[i];
	}
    }
}


/* ------------------------------------------------------------------
 * parse the arriving query list into the internal data structure of the
 * scheduler
 * ----------------
 */
char *
RegisterArrivingQuery(char *arg)
{
    SchedArg *sarg = (SchedArg *) arg;
    int pid = sarg->pid;
    List * queryList = sarg->list;
    int userid;
    List * sublist;
    QueryPtr parQuery = AllocQueryStruct();

    InitList (&parQuery->subQueryList);
    parQuery->user = userid = GetInternalUid(pid);
    parQuery->arr_time = MyTime();

    foreach (sublist, queryList) {
	SubQuery * squery = (SubQuery *)lfirst(sublist);
	QueryPtr query = AllocQueryStruct();
	query->user = userid;
	query->squery = sublist;
	query->sel[0] = query->sel[1] = 1;
	query->arr_time = parQuery->arr_time;
	
	if (JoinQuery(squery)) {
	    query->rel[1] = GetInternalRelid(RelationId(outerPlan(squery)));
	    query->rel[0] = GetInternalRelid(RelationId(innerPlan(squery)));
	} else {
	    query->rel[1] = -1;
	    query->rel[0] = GetInternalRelid(RelationId(squery));
	}
	SetSS_QueryCost(query);
	AddEltoList(&parQuery->subQueryList, (ElemType *)query);
    }
    {
	QueryPtr query;
	Element * sl;

	ForEach(sl, parQuery->subQueryList) {
	    query = (QueryPtr) GetData(sl);
	    lnext((List *)query->squery) = NIL;
	}
    }
    return (char *) parQuery;
}

/* -------------------------------------------
 * Given the relid of a relation, get the internal relid of the
 * relation and register it in the internal data structures
 * if not present already
 * ---------------------------
 */
static int
GetInternalRelid (Oid relid)
{
    int p;
    int * resArray;

    for (p = 0; p < numRels; p++) {
	if (relationIds[p] == relid) {
	    return p;
	}
	if (relationIds[p] == -1) {
	    break;
	}
    }
    if (p+1 >= NUM_RELS) {
	PrintError("Too many relations!");
    }
    relationIds[p] = relid;
    resArray = PGRelInfo(relid);
    sitemgrDebug(3, "relation size = %d, platter = %d\n", 
	    resArray[0], resArray[1]);
    relList[p].size = IntCeil(resArray[0], SIZE_UNIT);
    /*
     * make appropriate checks for cache size here
     */
    if (relList[p].size >= cacheSize/2) {
	sitemgrDebug(0, "Relation size cannot be more than %d\n", 
		     cacheSize/2);
    }
    relList[p].plat = GetInternalPlid((Oid) (resArray[1]));
    relList[p].relid = p;
    relList[p].startBlock = 0;
    numRels = p + 1;
    return p;
}

static int
GetExternalRelid(int rel)
{
    return relationIds[rel];
}
/* -------------------------------------------
 * Given the relid of a relation, get the internal platter id
 * of the platter on which the relation is stored.
 * ---------------------------
 */
static int
GetInternalPlid (Oid plid)
{
    int p;

    for (p = 0; p < numPlats; p++) {
	if (platterIds[p] == plid) {
	    return p;
	}
	if (platterIds[p] == -1) {
	    break;
	}
    }
    if (p+1 >= MAX_PLAT) {
	PrintError("Too many platters!");
    }
    platterIds[p] = plid;
    numPlats = p + 1;
    return p;
}

/* -------------------------------------------
 * Given the pid of the backend process, map it into an internal uid of the
 * scheduler. 
 * ---------------------------
 */
static int
GetInternalUid (int pid)
{
    int p;

    for (p = 0; p < numUsers; p++) {
	if (userIds[p] == pid) {
	    /*
	     * user already registered -- return
	     */
	    return p;
	}
	if (userIds[p] == -1) {
	    break;
	}
    }
    if (p+1 >= MAX_USERS) {
	PrintError("Too many users!");
    }
    userIds[p] = pid;
    numUsers = p + 1;
    return p;
}


/* -------------------------------------------
 * Given the pid of the backend process, map it into an internal uid of the
 * scheduler. 
 * ---------------------------
 */
static int
GetInternalIOid (int pid)
{
    int p;

    for (p = 0; p < numIOworkers; p++) {
	if (ioIds[p] == pid) {
	    /*
	     * user already registered -- return
	     */
	    return p;
	}
	if (ioIds[p] == -1) {
	    break;
	}
    }
    if (p + 1 >= MAX_DEVICE) {
	PrintError("Too many io backends!");
    }
    ioIds[p] = pid;
    numIOworkers = p + 1;
    return p;
}

/* ----------------------------------------------------------
 * Process query arg stuff....
 */
SchedFragmentArg *
FormCacheRelArg(Oid relid)
{
    static SchedFragmentArg cra;

    cra.type = 'C';
    cra.frastoreid = relid;
    return &cra;
}

SchedFragmentArg *
FormUncacheRelArg(Oid relid)
{
    static SchedFragmentArg cra;

    cra.type = 'U';
    cra.frastoreid = relid;
    return &cra;
}

static int *
PGRelInfo(Oid relid)
{
    NSVCProbeArg arg;
    NSVCProbeResult *result;
    int d[2];

    sitemgrDebug(3, "In PGRelInfo ..%d \n", relid);
    arg.action = NSVC_SCHEDGETFRAGINFO;
    arg.oid = relid;
    result = (NSVCProbeResult *)invokeExec(NSVC_LOCALPROBE, (void *)&arg, 
					   -1, "PGRelInfo");
    if (result==NULL) {
	PrintError("PGRelInfo: NSVC_LOCALPROBE failed");
    }
    return result->fragids.fragids_val;
}


