/*------------------------------------------------------------------
 * schedule.c ---
 *  This file contains all routines that are dependent on the
 *  scheduling policy being tested
 *
 *  Assumptions:
 *   o The only kind of queries handled are
 *     single relation query -- sequential scan
 *     2-way join query -- hash  join without any qualification on the
 *                         component relation
 *   o queries are not grouped together
 *--------------------------------------------------------------------
 */

#include <stdio.h>
#include <limits.h>
#include "parameter.h"
#include "listSched.h"
#include "querySched.h"
#include "schedInterface.h"
#include "request.h"

#define MAXTIME FLT_MAX

extern QueryPtr MinimumAveragePlat(), MinimumAverageGrp(), Fcfs(), Sjn(), 
         LargestNumberNext(), MinAvgKPerPlat(), ArbitraryPlat();

extern QueryPtr FcFs(), QueryOrientedPolicies();
extern RequestPtr AdjPlatAgg(), AnyPlatAgg(), TwoAdjPlatAgg(), MixedPlatAgg(),
                  SeekPlatAgg();
extern int NegCacheLoadReq(), WorkSize(), NegCoverSize(), OverlapCacheReq(), 
           NegLatency(), Any();

extern int PreProcessQuery PROTO((QueryPtr query, ListX * queryList));
/*
 * globals
 */
ListX pendingList; /* queries not yet submitted to scheduler*/
ListX scheduleList; /* queries waiting for execution */
int totalQueriesDone = 0;    
int totalCompleteQueriesDone = 0;

int schedulingPolicy = FCFS;
int evictionPolicy;

int subQueriesPerUser[MAX_USERS];

/*--------------------------------------------------------------
 * InitScheduler()
 *   Initialize the scheduler and the pending list
 */
InitScheduler(schedflag, evictflag)
int schedflag;
int evictflag;
{
    InitList(&pendingList);
    InitList(&scheduleList);

    schedulingPolicy = schedflag;
    evictionPolicy = evictflag;
    if (batchMode) { 
	maxQueries = endMeasurement = numUsers;
    }
}

/*-----------------------------------------------------------------
 * The criteria for stopping the exection engine is  number of 
 * queries executed,
 */
 
int
StoppingCriteria()
{
    if (totalCompleteQueriesDone >= maxQueries)
	return (1);
    else
	return(0);
	
}

/*-------------------------------------------------------------------
 * Do all that is required when a user query has finished execution
 * In the case where the query generator is closed queuing, at this
 * point it needs to be signaled about the completion of this query
 */
void
HandleFinishedQuery(query, curr_time)
QueryPtr query;
float curr_time;
{
    float response, ratio;

    totalQueriesDone++;
    if (detectMean) {
	printf("%d  %f\n", totalQueriesDone, 
	       statxIOTime/(float)totalQueriesDone);
    }
    subQueriesPerUser[query->user]--;

    if (subQueriesPerUser[query->user] == 0) {
	totalCompleteQueriesDone++;
	/*
	 * collect all statistics
         */
	if ((totalCompleteQueriesDone > startMeasurement) && 
	    (totalCompleteQueriesDone <= endMeasurement)) {
	    response = curr_time - query->arr_time;
	    statxCumResponse += response;
	}
    } else {
	;
    }
    ReleaseQueryStruct(query);
}

/*-----------------------------------------------------------------
 * put arriving query in the pending list
 */

HandleArrivingQuery (query, curr_time)
QueryPtr query;
float curr_time;
{
    query->arr_time = curr_time;
    AddEltoList (&pendingList, (ElemType *)query);
}

/*----------------------------------------------------------------
 * It selects a query from the scheduleList() for execution
 */
QueryPtr
ScheduleNextQuery()
{
    HandlePendingQueries();
   
    switch (schedulingPolicy) {

    case FCFS:
	return FcFs();
    case MINCLQ:
    case MINLATENCYQ:
    case MINACCESSCOSTQ:
	return QueryOrientedPolicies();
    default:
	return FcFs();
    }
}


RequestPtr
ScheduleNextRequest()
{
    HandlePendingQueries();
   
    switch (schedulingPolicy) {

    case SIZE_ADJ:
	return AdjPlatAgg(NegCacheLoadReq);
    case SIZE_2ADJ:
	PrintError("Not implemented\n");
	return TwoAdjPlatAgg(NegCacheLoadReq);
    case SIZE_ALL:
	return AnyPlatAgg(NegCacheLoadReq);

    case WORK_ADJ:
	return AdjPlatAgg(WorkSize);
    case WORK_2ADJ:
	return TwoAdjPlatAgg(WorkSize);
    case WORK_ALL:
	return AnyPlatAgg(WorkSize);
    case WORK_MIXED:
	return MixedPlatAgg();

    case COVER_ADJ:
	return AdjPlatAgg(NegCoverSize);
    case COVER_2ADJ:
	PrintError("Not implemented\n");
	return TwoAdjPlatAgg(NegCoverSize);
    case COVER_ALL:
	return AnyPlatAgg(NegCoverSize);

    case OVERLAP_ADJ:
	return AdjPlatAgg(OverlapCacheReq);
    case OVERLAP_2ADJ:
	PrintError("Not implemented\n");
	return TwoAdjPlatAgg(OverlapCacheReq);
    case OVERLAP_ALL:
	return AnyPlatAgg(OverlapCacheReq);

    case SEEK_ADJ:
	return AdjPlatAgg(NegLatency);
    case SEEK_2ADJ:
	PrintError("Not implemented\n");
	return TwoAdjPlatAgg(NegLatency);
    case SEEK_ALL:
	return AnyPlatAgg(NegLatency);
    case SEEK_OPT:
	return SeekPlatAgg();
	
    case ANY_ADJ:
	return AdjPlatAgg(Any);
    case ANY_2ADJ:
	PrintError("Not implemented\n");
	return TwoAdjPlatAgg(Any);

    case ANY_ALL:
    default:
	return AnyPlatAgg(Any);
    }
}

QueryOriented()
{
    switch (schedulingPolicy) {
	case FCFS:
	case MINCLQ:
	case MINLATENCYQ:
	case MINACCESSCOSTQ:
	     return 1;
	default:
	     return 0;
	 }
}

HandlePendingQueries()
{
    QueryPtr query;
    /*
     * preprocess queries in the pending list
     */
    while (1) {
        query = (QueryPtr) RemoveElfromList(&pendingList);
        if (query == NULL) {
	    break;
	}
        subQueriesPerUser[query->user] =
	    ListLength(query->subQueryList);
	
	if (QueryOriented()) {
	    AddEltoList(&scheduleList, (ElemType *) query);
	} else {
	    AppendList (&scheduleList, &(query->subQueryList));
	}
    }
}

int SS_CheckDone(int i)
{
    return (IsEmpty(scheduleList));
} 
