#include <stdio.h>
#include "parameter.h"

/*
 * specification of the tertiary memory
 */
typedef struct _robotSpec {
    char      name[50];
    float     moveArmTime;      /* avg. time to move arm */
    float     loadFromShelf;    /* time to move tape to drive */
    float     unloadToShelf;    /* time to remove tape from drive */
    int       numArms;          /* # of arms per robot */
    int       numCartRows;      /* # of cart. rows in array */
    int       numDataCartCols;  /* # of cart. DATA rows in array */
    int       numCartCols;      /* total # of cart. in array */
    int       numCartridges;    /* total * of cart. per robot */
    int       numReaders;       /* readers per robot */
    float     robotCost;        /* */
} robotSpec;

/*
 * specification of a device driver of the tertiary memory
 */
typedef struct _deviceSpec {
    int                  bytesPerBlock;      
    float                GBytesPerCartridge; /* total bytes on cartridge */
    long int             blocksPerCartridge; 
    float                ejectTime;          /* time to eject cartridge */
    float                loadTime;           /* time to load cartridge */
    float                rewindRate;         /* in Bytes/sec */
    float                rewindStartup;      /* in sec */
    float                searchRate;         /* in bytes/sec */
    float                searchStartup;      /* in sec */
    float                readXferRate;       /* in bytes/sec */
    float                writeXferRate;      /* in bytes/sec */
} deviceSpec;

/*
 * definition of an I/O request
 */
typedef enum{READ, WRITE} reqType;
typedef struct _individualRequest {
  int                deviceNum;
  int                cartridgeNum;
  int                startBlock;
  int                numBlocks;
  reqType            type;
  float              firstByteTime;
  float             serviceTime;
}individualRequest;

/*
 * definition of the device state
 */
typedef enum{ LOADED, EMPTY} deviceState;
typedef struct _perDevice {
  int                cartridgeLoaded;
  int                currentBlock;
  deviceState        state;
}perDevice;

/*
 * static globals
 */
static robotSpec       robot;
static deviceSpec      device;
static perDevice  *deviceArray;
static  int totalArrayBlocks;
static int tapeBased = 0;

static individualRequest req;

/*
 * routines exported from this file
 */
void InitTertiaryMem PROTO((char * hardwareFileName, int num));
float DataFetchSubmit  PROTO((char * reqPtr, int rel));
int ConvertToBlocks PROTO((int size));
float DataFetchCost PROTO((int rel));
float EstimateLatency PROTO((int rel));
float EstimateFetchCost PROTO((int rel));

/*
 * routines internal to this file
 */
static float SubmitRequest PROTO((individualRequest * nreq));
static int ChooseDevice PROTO((individualRequest   *reqPtr));
static float ServiceTime PROTO((perDevice *devicePtr, individualRequest *reqPtr, 
      		 float * setupTime));
static float UnLoadCatridgeTime  PROTO((perDevice *devicePtr));
static float LoadCartridgeTime PROTO(());
static float DeviceAccessTime PROTO((perDevice *devicePtr, 
			      individualRequest *reqPtr, float * setupTime));


/* --------------------------------------------------------------------
 * Initialize the tertiary memory structure. Argument is a file that
 * contains the robotFileName that describes the specs of the tertiary memory
 * a deviceFileName that describes the specs of the device driver and a string
 * that states the type of tertiary memory it is. If the type is tape-based
 * the string should be set to "TAPE"
 */
void
InitTertiaryMem(hardwareFileName, tertNum)
char * hardwareFileName;
int tertNum;
{
    char       deviceFileName[100], robotFileName[100]; 
    char       deviceDirName[50], typeStr[20];
    FILE       *hardwareFilePtr;
    FILE       *deviceFilePtr, *robotFilePtr;
    int numScanned;
    int avgDist;
    int i = -1;


    if (hardwareFileName == NULL) return;

    hardwareFilePtr = fopen(hardwareFileName, "r");
    if (hardwareFilePtr == (FILE *)NULL){
	printf("Could not open hardware file %s\n", hardwareFileName);
	exit(1);
    }
    typeStr[0] = '\0';
    numScanned = fscanf(hardwareFilePtr, "%s", deviceDirName);  

    while (i != tertNum) {
	numScanned = fscanf(hardwareFilePtr, "%d %s %s %s", &i, 
		     deviceFileName, robotFileName, typeStr);
	if (numScanned != 4) {
	    printf ("Could not find tertiary number %d\n", i);
	    exit(0);
	}
    }

    if (strncmp(typeStr, "TAPE", 4) == 0) {
	tapeBased = 1;
    }
    StringAppend((char *) deviceDirName, (char *)deviceFileName);
    StringAppend((char *) deviceDirName, (char *)robotFileName);

    /*
     * Open various configuration files;  set up device, robot, 
     * array configurations
     */
    
    deviceFilePtr = fopen(deviceFileName,"r");
    if (deviceFilePtr == (FILE *)NULL){
	printf("Could not open device configuration file %s\n.",deviceFileName);
	exit(1);
    }
    
    numScanned = fscanf(deviceFilePtr, "%f %f %f %f %f %f %f %f %d %d %f",
			&device.ejectTime,
			&device.loadTime,
			&device.rewindRate,
			&device.rewindStartup,
			&device.searchRate,
			&device.searchStartup,
			&device.readXferRate,
			&device.writeXferRate,
			&device.blocksPerCartridge,
			&device.bytesPerBlock,
			&device.GBytesPerCartridge);
    if (numScanned != 11) {
	printf("Not enough parameters in device file %s\n.",deviceFileName);
	exit(1);
    }
    
    robotFilePtr = fopen(robotFileName,"r");
    if (robotFilePtr == (FILE *)NULL){
	printf("Could not open robot configuration file %s\n.",robotFileName);
	exit(1);
    }
    
    numScanned = fscanf(robotFilePtr, "%s %f %f %f %d %d %f %d",
			robot.name,
			&robot.moveArmTime,
			&robot.loadFromShelf,
			&robot.unloadToShelf,
			&robot.numArms,
			&robot.numReaders,
			&robot.robotCost,
			&robot.numCartridges);
    if (numScanned != 8) {
	printf("Not enough parameters in robot file %s\n.",
	       robotFileName);
	exit(1);
    }
    
    /*
     * initialize the device specific information used in the
     * scheduler
     */
    maxPlats = robot.numCartridges;
    if (maxPlats > MAX_PLAT) {
	PrintError ("maximum number of platters exceeded");
    }
    numDevice = robot.numReaders;
    platCapacity = device.GBytesPerCartridge*((1 << 30)/SIZE_UNIT);
    switchTime = robot.moveArmTime + robot.loadFromShelf + 
	robot.unloadToShelf + device.ejectTime + device.loadTime;
    tertWeight = SIZE_UNIT / device.readXferRate;
    
    if (tapeBased) {
	avgDist =  0.2*device.GBytesPerCartridge*(1 << 30);
	tertFixedCost = 0.5*(avgDist / device.rewindRate + 
			      device.rewindStartup) +
				  0.5*(avgDist / device.searchRate +
				       device.searchStartup);
	switchTime += device.rewindStartup + avgDist / device.rewindRate;
    } else {
	tertFixedCost = device.searchStartup;
	switchTime += device.rewindStartup;
    }
    totalArrayBlocks = device.blocksPerCartridge * robot.numCartridges;

    deviceArray = (perDevice *) MallocErrorCheck
	(sizeof(perDevice) * robot.numReaders);
    
    for (i=0; i< robot.numReaders; i++){
	deviceArray[i].cartridgeLoaded = -1;
	deviceArray[i].currentBlock = 0;
	deviceArray[i].state = EMPTY;
    }


    if (is_debug("i")) {
	printf ("Switch time %f, setup time %f, transfer time/MB %f\n", 
		switchTime, tertFixedCost, tertWeight);
    }

}

char *
GetRobotName()
{
    return (robot.name);
}

/* -------------------------------------------------------------
 *
 */
float
EstimateFetchCost (rel)
int rel;
{
    float cost;

    cost = EstimateLatency(rel);
    cost += tertWeight*RelSize(rel);
    return (cost);
}

float 
EstimateLatency(rel)
int rel;
{
    int loadedPlat[MAX_DEVICE];
    float cost = 0.0;
    
    GetLoadedPlatters(loadedPlat);
    if (! IsAnElementOf (RelPlatName(rel), loadedPlat, numDevice)) {
	cost += switchTime;
    }
    cost += tertFixedCost;
    return (cost);
}

float
DataFetchCost(rel)
int rel;
{
    return tertFixedCost + (float)RelSize(rel)*tertWeight;
}

/* -------------------------------------------------------------------
 * Submit request for fetching a relation to tertiary memory
 */

float
DataFetchSubmit (reqPtr, rel)
char * reqPtr;
int rel;
{
    float iotime;

    req.startBlock = RelStartBlock(rel);
    req.numBlocks = RelNumBlocks(rel);
    req.cartridgeNum = RelPlatName(rel);
    req.type = READ;
    iotime = SubmitRequest (&req);

/*    ScheduleIO(reqPtr, iotime); */
    statxIOTime += req.serviceTime;
    if (is_debug("i")) {
	printf ("I/O: rel: %d, size %d, start %d, plat %d, timeTaken %f\n",
		rel, RelSize(rel), req.startBlock, RelPlatName(rel), 
		req.serviceTime);
    }
    return (req.serviceTime);
}

/* ----------------------------------------------------------------
 * convert size in SIZE_UNIT to number of blocks
 */
int
ConvertToBlocks(size)
int size;            /* size in SIZE_UNIT */
{
    return (size * IntCeil(SIZE_UNIT, device.bytesPerBlock));
}

HeadPos(platNum) 
int platNum;
{
    int i;

    for (i=0;  i < robot.numReaders;  i++){
	if (platNum == deviceArray[i].cartridgeLoaded) {
	    return GetCurrentBlock(i);
	}
    }
    return 0;
}
/* ---------------------------------------------------------------
 *
 */
GetLoadedPlatters(list)
int *list;
{
    int i;

    for (i=0;  i < robot.numReaders;  i++){
	list[i] = deviceArray[i].cartridgeLoaded;
    }
    return;
}

/* -------------------------------------------------------------
 *
 */
int
GetCurrentBlock(deviceNum)
int deviceNum;
{
    return (deviceArray[deviceNum].currentBlock);
}    
/* ---------------------------------------------------------------
 * estimate time taken to satify a request pointed to by 'req' and
 * schedule the I/O event for that time in the simulator
 */
float
SubmitRequest(reqPtr)
individualRequest * reqPtr;
{
    float setupTime;
    float timeTaken;
    int deviceNum;

    deviceNum = ChooseDevice (reqPtr);
    timeTaken = ServiceTime (&(deviceArray[deviceNum]), reqPtr, &setupTime);
    reqPtr->serviceTime = timeTaken;
    reqPtr->firstByteTime = setupTime;
    return (timeTaken);
} 


/* ---------------------------------------------------------------------
 * choose a device for satisfying the request. If the cartridge is already
 * loaded on some drive we choose that drive, otherwise we choose an
 * empty drive, if no empty drive exists choose a random drive
 */

char * LruInit_SS();
char * lruHandle = NULL;

int
ChooseDevice(reqPtr)
individualRequest   *reqPtr;
{
    int deviceNum, emptyDevice;
    int i;
    perDevice   *oneDevice;
    
    if (lruHandle == NULL) {
	lruHandle = LruInit_SS (robot.numReaders);
    }

    deviceNum = -1;
    emptyDevice = -1;
    
    for (i=0;  i < robot.numReaders;  i++){
	oneDevice = &(deviceArray[i]);
	if (oneDevice->state == LOADED){
	    if (oneDevice->cartridgeLoaded == reqPtr->cartridgeNum) {
		deviceNum = i;
		break;
	    }
	} else {
	    emptyDevice = i;
	}	  
    }
    
    if (deviceNum == -1) {
	deviceNum = emptyDevice;
    }
    if (deviceNum == -1) {
	deviceNum = LruEntry_SS(lruHandle);
    }
    
    LruAccess_SS(lruHandle, deviceNum);
    return(deviceNum);
}

/* -------------------------------------------------------------------
 * return the time taken a process an I/O request on the given
 * device driver
 */
float
ServiceTime(devicePtr, reqPtr, setupTime)
perDevice           *devicePtr;
individualRequest   *reqPtr;
float               * setupTime;
{
    float servTime = 0;
    float accessTime;

    if (devicePtr->cartridgeLoaded != reqPtr->cartridgeNum) {
	if (devicePtr->state == LOADED) {
	    servTime +=  UnLoadCatridgeTime(devicePtr);
	}
	servTime += LoadCartridgeTime();
	devicePtr->cartridgeLoaded = reqPtr->cartridgeNum;
	devicePtr->currentBlock = 0;
	devicePtr->state = LOADED;
	statxNumSwitches++;
    }
    
    accessTime = DeviceAccessTime(devicePtr, reqPtr, setupTime);
    devicePtr->currentBlock = reqPtr->startBlock + reqPtr->numBlocks - 1;

    *setupTime += servTime;
    servTime += accessTime;
    return (servTime);
}
/*
 * --------------------------------------------------------------------
 * returns the time to unload a cartridge from a device. For tapebased
 * tertiary memory we need to rewind the tape to the start
 * position. So, includes time to rewind (if tapebased), eject
 * cartridge and move cartridge from device to shelf 
 */
float
UnLoadCatridgeTime (devicePtr)
perDevice           *devicePtr;
{
    float rewind, eject;
    
    if (devicePtr->currentBlock != 0){
	if (tapeBased) {
	    rewind = (device.rewindStartup + 
		      (devicePtr->currentBlock * 
		       (device.bytesPerBlock / device.rewindRate)));
	} else {
	    rewind = device.rewindStartup;
	}
    }
    eject = device.ejectTime;
   
    return (eject +  rewind + robot.unloadToShelf);
}
 
/*
 * ------------------------------------------------------------------------
 * returns the time to load a cartridge into an empty device driver.  
 */
float
LoadCartridgeTime()
{
    return (2 * robot.moveArmTime + robot.loadFromShelf 
	    + device.loadTime);
}

/* ----------------------------------------------------------------------
 * Returns time to access a loaded device -- this includes the time to do any
 * seek/rewind/search and the time to transfer data. Variable 'setupTime' is
 * set to the time to transfer the first byte.
 */

float
DeviceAccessTime(devicePtr, reqPtr, setupTime)
perDevice           *devicePtr;
individualRequest   *reqPtr;
float               * setupTime;
{
  
    float                  backwardSearch = 0.0;
    float                  forwardSearch = 0.0;
    float                  transferTime = 0.0;
    float                  access;
    float                  xferRate;

    /*  
     * Check current position of cartridge.
     * Fast forward or rewind to correct block.
     */
  
    if (tapeBased) {
	if (devicePtr->currentBlock < reqPtr->startBlock) {
	    forwardSearch = (device.searchStartup +
			     ((reqPtr->startBlock - devicePtr->currentBlock) * 
			      (device.bytesPerBlock / device.searchRate)));
	    statxSeekVar += (forwardSearch - device.searchStartup);  
	} else {
	    if (devicePtr->currentBlock > reqPtr->startBlock) {
		backwardSearch = (device.rewindStartup + 
				  ((devicePtr->currentBlock - 
				    reqPtr->startBlock) *
				   (device.bytesPerBlock / 
				    device.rewindRate)));
		statxSeekVar += (backwardSearch - device.rewindStartup);
	    }
	}
    } else {
	forwardSearch = device.searchStartup;
	backwardSearch = 0.0;
    }

    /*
     * Calculate time for data transfer.
     */

    if (reqPtr->type == READ){
	xferRate = device.readXferRate;
    } else {
	xferRate = device.writeXferRate;
    }

    transferTime = (reqPtr->numBlocks / xferRate) * device.bytesPerBlock;
    *setupTime = forwardSearch + backwardSearch;
    access = forwardSearch + backwardSearch + transferTime;
    statxferTime += transferTime;

    return(access);
}





