#-------------------------------------------------------------------------
#
# .sh--
#    
#
# Copyright (c) 1994, Regents of the University of California
#
#
# IDENTIFICATION
#    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/bidder.tcl,v 1.3 1996/05/31 17:14:23 jsidell Exp $
#
#-------------------------------------------------------------------------
###########################################################################
#
# bidder.tcl
#
#        Input:  plan tree, represented as a string
#
#        Output: list containing {response cost delay staleness accuracy}
#
#                response: BID if all data fragments references in the
#                          query are local.  REFUSETOBID otherwise.
#
#                cost:     Based on the per-tuple and per-page charge for
#                          each node in the query plan
#
#                delay:    Based on the per-tuple and per-page delay for
#                          each node in the query plan
#
#                staleness, accuracy: ignored
#
# Recursively descends the plan tree, keeping track of the number of pages
# and number of tuples generated, and adding up the cost and delay until
# the root is reached.  At this point, the total cost and total delay have
# been calculated.
#
###########################################################################


set BID 1
set REFUSETOBID 0

#--------------------------------------------------------------------------
#
# LoadAverage
#
#        Input:  void
#
#        Output: 5-, 30-, and 60-second load averages
#--------------------------------------------------------------------------

proc LoadAverage {} {
    set result [exec uptime]
    puts stderr "LoadAverage: result = $result"
    set len [llength $result]

    set result [lrange $result [expr "$len - 3"] [expr "$len - 1"]]
    regsub -all , $result "" res2

    return $res2
}

#--------------------------------------------------------------------------
#
# CombineBids
#
#        Input:  two bids, bid1 and bid2
#
#        Output: bid that results from combining bid1 and bid2:
#                
#                response:  BID if both bid1 and bid2 responses are BID
#                           REFUSETOBID otherwise
#
#                cost:      bid1.cost + bid2.cost
#
#                delay:     bid1.delay + bid2.delay
#
#                staleness: MAX(bid1.staleness, bid2.staleness)
#
#                accuracy:  MIN(bid1.accuracy, bid2.accuracy)
#
#--------------------------------------------------------------------------
proc CombineBids {bid1 bid2} {

    global BID REFUSETOBID

    puts stderr "CombineBids: $bid1 $bid2"
    flush stderr

    set response1 [lindex $bid1 0]
    set response2 [lindex $bid2 0]
    set cost1 [lindex $bid1 1]
    set cost2 [lindex $bid2 1]
    set delay1 [lindex $bid1 2]
    set delay2 [lindex $bid2 2]
    set stale1 [lindex $bid1 3]
    set stale2 [lindex $bid2 3]
    set acc1 [lindex $bid1 4]
    set acc2 [lindex $bid2 4]

    set response [expr ($response1 && $response2) ? $BID : $REFUSETOBID]
    set cost [expr $cost1 + $cost2]
    set delay [expr $delay1 + $delay2]
    set stale [expr ($stale1 > $stale2) ? $stale1 : $stale2]
    set acc [expr ($acc1 < $acc2) ? $acc1 : $acc2]

    return [list $response $cost $delay $stale $acc]

}

#--------------------------------------------------------------------------
#
# MERGEJOIN
#
#        Input:  left sub-tree, right sub-tree
#
#        Output: bid
#
# Updates nTuples and nPages - guesses one match for each outer
# tuple.
#--------------------------------------------------------------------------
proc MERGEJOIN {nodeNum leftTree rightTree {junk {}} } {

    global BID REFUSETOBID

    global nTuples
    global nPages
    global rtable
    global hostid

    puts stderr "Inside JOIN"
    flush stderr

    set perTupleCharge .001
    set perTupleDelay .000400

    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples
    set leftPages $nPages
    
    set rightSubBid [CostBasedBid $rightTree]
    set rightTuples $nTuples
    set rightPages $nPages

    if {$leftTuples == 0} {
	set leftTuples 10000
    }
    if {$rightTuples == 0} {
	set rightTuples 10000
    }
    if {$leftPages == 0} {
	set leftPages 100
    }
    if {$rightPages == 0} {
	set rightPages 100
    }

    # Each outer and inner tuple is touched once.
    set delay [expr ($leftTuples + $rightTuples) * $perTupleDelay]
    set cost  [expr ($leftTuples + $rightTuples) * $perTupleCharge]

    # Wild guess - one match for each outer tuple
    set nTuples $leftTuples

    set bid [CombineBids $leftSubBid $rightSubBid]

    set bid [CombineBids $bid [list $BID $cost $delay 0.0 0.0]]

    return $bid
}
#--------------------------------------------------------------------------
#
# NESTEDLOOP
#
#        Input:  left sub-tree, right sub-tree
#
#        Output: bid
#
# Updates nTuples and nPages - guesses one match for each outer
# tuple.
#--------------------------------------------------------------------------
proc NESTEDLOOP {nodeNum leftTree rightTree {junk {}} } {

    global BID REFUSETOBID

    global nTuples
    global nPages
    global rtable
    global hostid

    puts stderr "Inside JOIN"
    flush stderr

    set perTupleCharge .001
    set perTupleDelay .000400

    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples
    set leftPages $nPages
    
    set rightSubBid [CostBasedBid $rightTree]
    set rightTuples $nTuples
    set rightPages $nPages
    
    # Each inner tuple is touched once per outer tuple.
    set delay [expr ($leftTuples * $rightTuples) * $perTupleDelay]
    set cost  [expr ($leftTuples * $rightTuples) * $perTupleCharge]

    # Wild guess - one match for each outer tuple
    set nTuples $leftTuples

    set bid [CombineBids $leftSubBid $rightSubBid]

    set bid [CombineBids $bid [list $BID $cost $delay 0.0 0.0]]

    return $bid
}


#--------------------------------------------------------------------------
#  SEQSCAN
#        Input:  scanIndex, fragIndex, left sub-tree
#
#        Output: bid
#
# Updates nTuples and nPages based on information in range table.
#
#--------------------------------------------------------------------------
proc SEQSCAN {nodeNum scanIndex fragIndex {leftTree {}} } {

    global BID REFUSETOBID
    global contract
    global nTuples
    global nPages
    global rtable
    global hostid

    # no extra charge per tuple
    set perTupleCharge 0

    # 5 cents per page
    set perPageCharge .05

    # delay in seconds per tuple retrieved (not including disk I/O)
    set perTupleDelay .000600

    # delay in seconds per disk page accessed
    set perPageDelay .002200

    # Scan on a temporary relation, the result of a sort,
    # join, etc.  Just use the values of nTuples and nPages
    # generated so far.
    if {$scanIndex == -1} {
	set nTuples 10000
	set nPages 100
	set cost [expr $nTuples * $perTupleCharge + $nPages * $perPageCharge]
	set delay [expr $nTuples * $perTupleDelay + $nPages * $perPageDelay]
	
	set bid [CombineBids "$BID $cost $delay 0.0 0.0" [CostBasedBid $leftTree]]

	return $bid
    } else {

    # Scan on a base relation - set nTuples and nPages
    # from rtable information.
    # Only bid if the fragment is stored at this site.

	set rte [lindex $rtable $scanIndex]

	set frags [lindex $rte 3]
	
	set fInfo [lindex $frags $fragIndex]

	# Determine if one of the storage sites is this one.
	set storageSites [lindex $fInfo 4]
	
	puts "storageSites = $storageSites"
	set local false

	foreach site $storageSites {
	    puts "site = $site"
	    set storageHost [lindex $site 2]
	    puts "storageHost = $storageHost, hostid = $hostid"
	    if {$storageHost == $hostid} {
		set local true
		break
	    }
	}

	set nTuples [lindex $fInfo 3]
	set nPages [lindex $fInfo 2]

	if {$local} {
	    
	    set response $BID
	    set cost [expr $nTuples * $perTupleCharge + $nPages * $perPageCharge]
	    set delay [expr $nTuples * $perTupleDelay + $nPages * $perPageDelay]

	    # ignore staleness and accuracy for now.
	    
	} else {
	    set subPlan "{SEQSCAN $nodeNum $scanIndex $fragIndex}"
	    puts stdout "bidder.tcl: subcontracting $subPlan"
	    set subBid [subcontract $subPlan $contract]
	    #set subBid "1 0.0 0.0 0.0 0.0"
	    set response [lindex $subBid 0]
	    set cost [lindex $subBid 1]
	    set delay [lindex $subBid 2]
	    puts stdout "bidder.tcl: done subcontracting $subPlan"
	    puts stdout "subBid = $subBid"
	    puts stdout "first subBid = [lindex $subBid 0]"
	    
	}

    }

    puts stdout "returning [list $response $cost $delay 0.0 0.0]"
    return [list $response $cost $delay 0.0 0.0]

}

#--------------------------------------------------------------------------
#
# SORT
#--------------------------------------------------------------------------
proc SORT {nodeNum {leftTree {}} {junk {}} {junk2 {}} } {

    global BID REFUSETOBID

    global nTuples
    global nPages
    global rtable
    global hostid

    set perTupleCharge .001
    set perTupleDelay .000400
    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples

    if {$leftTuples == 0} {
	set leftTuples 10000
	set nTuples 10000
    }

    set cost [expr $leftTuples * $perTupleCharge]
    set delay [expr $leftTuples * $perTupleDelay]
    set bid "$BID $cost $delay 0.0 0.0"

    set bid [CombineBids $leftSubBid $bid]

    return $bid
}

#--------------------------------------------------------------------------
#
# MERGE
#--------------------------------------------------------------------------
proc MERGE {nodeNum subTreeList {junk {}} {junk2 {}} } {

    global BID REFUSETOBID

    global nTuples
    global nPages
    global rtable
    global hostid

    set perTupleCharge .001
    set perTupleDelay .000400

    set bid [list $BID 0.0 0.0 0.0 0.0]
    set mergeTuples 0
    set mergePages 0

    foreach subPlan $subTreeList {
	puts "MERGE subTreeList = $subTreeList"
	puts "MERGE subPlan = $subPlan"
	set bid [CombineBids $bid [CostBasedBid $subPlan]]
	puts "***MERGE - Done calling CostBasedBid"
	puts "***MERGE subTreeList = $subTreeList"
	puts "***MERGE subPlan = $subPlan"
	puts "***MERGE bid = $bid"
	puts "nTuples = $nTuples"
	puts "nPages = $nPages"
	incr mergeTuples $nTuples
	incr mergePages $nPages
    }
    puts "set cost [expr $mergeTuples * $perTupleCharge]"
    set cost [expr $mergeTuples * $perTupleCharge]

    puts "set delay [expr $mergeTuples * $perTupleDelay]"
    set delay [expr $mergeTuples * $perTupleDelay]

    puts "set bid CombineBids $bid [list $BID $cost $delay 0.0 0.0]]"
    set bid [CombineBids $bid [list $BID $cost $delay 0.0 0.0]]

    return $bid

}

#--------------------------------------------------------------------------
#
# XIN
#--------------------------------------------------------------------------
proc XIN {nodeNum leftTree {junk {}}  {junk2 {}} } {

    global BID REFUSETOBID
    global nTuples

    set perTupleCharge .001
    set perTupleDelay .000400

    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples
    
    if {$leftTuples == 0} {
	set leftTuples 10000
	set nTuples 10000
    }
    
    set cost [expr $leftTuples * $perTupleCharge]
    set delay [expr $leftTuples * $perTupleDelay]

    set bid "$BID $cost $delay 0.0 0.0"

    set bid [CombineBids $leftSubBid $bid]

    return $bid

}

#--------------------------------------------------------------------------
#
# AGG
#--------------------------------------------------------------------------
proc AGG {nodeNum leftTree {junk {}}  {junk2 {}} } {

    global BID REFUSETOBID
    global nTuples

    set perTupleCharge .001
    set perTupleDelay .000400

    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples
    
    if {$leftTuples == 0} {
	set leftTuples 10000
	set nTuples 10000
    }

    set cost [expr $leftTuples * $perTupleCharge]
    set delay [expr $leftTuples * $perTupleDelay]

    set bid "$BID $cost $delay 0.0 0.0"

    set bid [CombineBids $leftSubBid $bid]

    return $bid

}

#--------------------------------------------------------------------------
#
# GROUPBY
#--------------------------------------------------------------------------
proc GROUPBY {nodeNum leftTree {junk {}}  {junk2 {}} } {

    global BID REFUSETOBID
    global nTuples

    set perTupleCharge .001
    set perTupleDelay .000400

    set leftSubBid [CostBasedBid $leftTree]
    set leftTuples $nTuples
    
    if {$leftTuples == 0} {
	set leftTuples 10000
	set nTuples 10000
    }
    
    set cost [expr $leftTuples * $perTupleCharge]
    set delay [expr $leftTuples * $perTupleDelay]

    set bid "$BID $cost $delay 0.0 0.0"

    set bid [CombineBids $leftSubBid $bid]

    return $bid

}

#--------------------------------------------------------------------------
#
# UNKNOWN
#
# Don't bid on plans that contain nodes we can't identify.
#
#--------------------------------------------------------------------------
proc UNKNOWN {nodeNum leftTree rightTree {junk {}} } {

    global BID REFUSETOBID

    return [list $REFUSETOBID 0 0 0 0]

}

#--------------------------------------------------------------------------
#
# CostBasedBid
#
#        Input:  query plan
#
#        Output: bid
#
# Main procedure.  Looks at token representing the node type and calls
# the appropriate bidding routine.
#
#--------------------------------------------------------------------------
proc CostBasedBid {plan} {
    global rtable
    global hostid
    global contract
    global nTuples
    global nPages

    global BID REFUSETOBID

    if {$plan != ""} {
	set nodeType [lindex $plan 0]
	puts stderr "calling $nodeType [lreplace $plan 0 0 ]"
	flush stderr

	set bid [$nodeType [lindex $plan 1 ] [lindex $plan 2 ] [lindex $plan 3 ] [lindex $plan 4]]
    } else {
	set bid [list $BID 0 0 0 0]
    }

    return $bid
}

#--------------------------------------------------------------------------
#
# GetQueryBid
#
#        Input:  query plan
#
#        Output: bid
#
# Main procedure.  Looks at token representing the node type and calls
# the appropriate bidding routine.
#
#--------------------------------------------------------------------------
proc GetQueryBid {plan} {
    global rtable
    global hostid
    global contract
    global nTuples
    global nPages

    global BID REFUSETOBID

    set bid [CostBasedBid $plan]
    set las [exec uptime]
    set len [llength $las]

    set result [lrange $las [expr "$len - 3"] [expr "$len - 1"]]
    regsub -all , $las "" las

    set la [lindex $las 2]
    puts stderr "la = $la"
    flush stderr

    set cost [lindex $bid 1]

    set cost [expr "$cost * (1 + $la)"]

    set bid [lreplace $bid 1 1 $cost]

    puts stderr "GetQueryBid returning $bid"
    flush stderr

    return $bid
}

