/*-------------------------------------------------------------------------
 *
 * cpupdate.c--
 *	Routines that implement the copy updating module. Manages a queue
 *	of update streams (Update Stream Queue) that holds contracts from
 *	pg_contract and the time remaining. When this module wakes up and
 *	detects some contracts are ready, it makes a call to CopyUpdateProcess
 *	which coordinates the update stream sending.
 *     
 * Copyright (c) 1995, Regents of the University of California
 *
 * $Id: cpupdate.c,v 1.5 1996/05/31 17:14:25 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/limits.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <pthread.h>

#include "postgres.h"
#include "utils/nabstime.h"
#include "mariposa_rpc.h"
#include "sitemgr.h"
#include "sitemgr/rpcerror.h"
#include "catalog/pg_contract.h"
#include "lib/queue.h"
#include "lib/usqueue.h"
#include "utils/elog.h"
#include "utils/palloc.h"

#define ITIMER_REAL 0

void CopyUpdateSigHandler();
void timeaddval(struct timeval *t1, struct timeval *t2);

extern pthread_mutex_t updateStreamQueueLock;	/* defined in sitemgr.c */
/* @@@ Debug mechanism */
bool updateSwitch;

struct sigaction sig_act;
int updateInterval = UPDATE_INTERVAL;
USQueue *updateStreams;

/* ----------------------------------------------------------------
 *    	CopyUpdateSetup
 *    
 *	Creates a signal handler that calls CopyUpdateSigHandler
 *	and sets the system timer to wake it up. Creates an Update
 *	Stream Queue that the signal handler processes. Then calls a
 *	backend exec to scan pg_contract and enqueue the contracts
 *	into the Update Stream Queue.
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateSetup(void *args)
{
    struct itimerval ourTimerSet, dummy;

    sig_act.sa_handler = (void *)CopyUpdateSigHandler;
    sig_act.sa_flags = 0;
    sigemptyset(&sig_act.sa_mask);
    sigaddset(&sig_act.sa_mask, SIGALRM);

    /*
     * Set signal handler to handle timer expiration.
     */
    sigaction(SIGALRM, &sig_act, NULL);

    /*
     * Initialize the timer value.
     */
    memset(&ourTimerSet, 0, sizeof(struct itimerval));
    ourTimerSet.it_value.tv_sec = updateInterval;
    ourTimerSet.it_interval.tv_sec = updateInterval;

    sitemgrDebug(3, "UpdateStreamSetup: value %d.%d, interval %d.%d\n",
	 ourTimerSet.it_value.tv_sec, ourTimerSet.it_value.tv_usec,
	 ourTimerSet.it_interval.tv_sec, ourTimerSet.it_interval.tv_usec);

    /*
     * Create our Update Stream Queue.
     */
    updateStreams = USQMakeUSQueue(updateInterval);

    /* @@@ Debug mechanism */
    updateSwitch = true;

    /*
     * Set the timer.
     */
    if (setitimer(ITIMER_REAL, &ourTimerSet, &dummy)) {
	sitemgrDebug(0, "UpdateStreamSetup: set timer failed.");
    }

    /*
     * Loop forever.
     */
    for (;;) pause();
}

/* ----------------------------------------------------------------
 *    	CopyUpdateToggle
 *
 *	Toggles back and forth from turning the update system on and
 *	off. For debugging purposes only!!
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateToggle(Oid key)
{
    updateSwitch = updateSwitch ? false : true;
}

/* ----------------------------------------------------------------
 *    	CopyUpdateScheduleStream
 *
 *	Given a copy update contract (stored in pg_contract),
 * 	schedule the contract into the Update Stream Queue.
 *
 *	Should be called when a new copy is created just after the
 *	contract for its update stream is inserted into pg_contract.
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateScheduleStream(UpdateContractInfo *updatecontract)
{
    USInfo *newStream;

    newStream = (USInfo *)palloc(sizeof(USInfo));
    if (newStream == NULL) {
	sitemgrDebug(0, "CopyUpdateScheduleStream: Out of memory!");
    }

    newStream->updateContract.contractid = updatecontract->contractid;
    newStream->updateContract.sendfrastoreid = updatecontract->sendfrastoreid;
    newStream->updateContract.recvfrastoreid = updatecontract->recvfrastoreid;
    newStream->updateContract.sidefrastoreid = updatecontract->sidefrastoreid;
    newStream->updateContract.lastupdcommit = updatecontract->lastupdcommit;
    newStream->updateContract.otherloc = updatecontract->otherloc;
    newStream->updateContract.period = updatecontract->period;
    newStream->updateContract.typemask = updatecontract->typemask;

    /*
     * If this is a name service contract, we haven't actually
     * received the name service data yet.  Schedule an update
     * stream to be sent right away so the user doesn't have
     * to wait for an entire period before receiving the name
     * service information.
     */
    if (!ContractIsNameService(updatecontract)) {
	newStream->timeRemaining = updatecontract->period;
    } else {
	newStream->timeRemaining = 0;
    }

    pthread_mutex_lock(&updateStreamQueueLock);
    USQInsert(updateStreams, newStream);
    pthread_mutex_unlock(&updateStreamQueueLock);

    sitemgrDebug(3, "CopyUpdateScheduleStream: contract %d period %d sched.",
		 newStream->updateContract.contractid,
		 newStream->updateContract.period);
}

/* ----------------------------------------------------------------
 *    	CopyUpdateUnscheduleStream
 *
 *	Given the key to a copy contract (stored in pg_contract)
 *	unschedule the contract from the Update Stream Queue.
 *
 *	Should be called when a copy is destroyed, just before the
 *	contract is removed from pg_contract.
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateUnscheduleStream(Oid key)
{
    USInfo *oldStream;

    pthread_mutex_lock(&updateStreamQueueLock);
    oldStream = USQRemove(updateStreams, key);
    pthread_mutex_unlock(&updateStreamQueueLock);

    pfree(oldStream);
}

/* ----------------------------------------------------------------
 *    	CopyUpdateProcess
 *
 *	Coordinates the sending of an update stream to a copy-holding site,
 *	based on the information given in the update contract parameter.
 *	Called by USQCycle when it finds an update stream's time is up.
 *
 *	Saves the commit time from sending the updates, for later use by
 *	the same procedure to check for which updates happened since then.
 *	(Unless the stream sending failed, in that case lastupdcommit for
 *	the contract is unchanged so we can try again on those that failed.)
 *
 *	We print out a noisy message if the call to Cp_SendUpdateStream fails.
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateProcess(UpdateContractInfo *updatecontract)
{
    SendUpdateStreamResult *result;

/* @@@ Debug mechanism */
    if (updateSwitch) {

    result = invokeExec(CP_SENDUPDATESTREAM, (void *)updatecontract, -1,
			"CopyUpdateProcess");

    if (result->error != SUCCESS) {
	sitemgrDebug(0, "CopyUpdateProcess: Error %d in CP_SENDUPDATESTREAM\n",
		     result->error);
	sitemgrDebug(0, "*************************************************\n");
	sitemgrDebug(0, "*************************************************\n");
	sitemgrDebug(0, "***** SEND UPDATE STREAM FAILED\n");
	sitemgrDebug(0, "***** contractid = %d\n", updatecontract->contractid);
	sitemgrDebug(0, "***** sendfrastoreid = %d\n",
		     updatecontract->sendfrastoreid);
	sitemgrDebug(0, "***** recvfrastoreid = %d\n",
		     updatecontract->recvfrastoreid);
	sitemgrDebug(0, "***** update-remote-host = %d\n",
		     updatecontract->otherloc);
	sitemgrDebug(0, "***** lastupdcommit = %d\n",
		     updatecontract->lastupdcommit);
	sitemgrDebug(0, "***** current time = %d\n", GetCurrentAbsoluteTime());
	sitemgrDebug(0, "*************************************************\n");
	sitemgrDebug(0, "*************************************************\n");
    }
    else {
	sitemgrDebug(0, "CopyUpdateProcess: Commit Time %d\n",
		     result->committime);

	updatecontract->lastupdcommit = result->committime;
    }

    free(result);

    }
}

/* ----------------------------------------------------------------
 *    	CopyUpdateSignalHandler
 *
 *	Called when a wake-up is scheduled. USQCycle checks the queue for
 *      contracts that are due for sending updates, and for those that are,
 *	calls CopyUpdateProcess on them.
 *
 * ----------------------------------------------------------------
 */
void
CopyUpdateSigHandler(int signo)
{
    switch (signo) {
    case SIGALRM:
	pthread_mutex_lock(&updateStreamQueueLock);
	USQCycle(updateStreams);
	pthread_mutex_unlock(&updateStreamQueueLock);
	break;
    }
}

/* Add two time values. */
void timeaddval(struct timeval *t1, struct timeval *t2)
{
    t1->tv_sec += t2->tv_sec;
    t1->tv_usec += t2->tv_usec;
    if (t1->tv_usec < 0) {
	t1->tv_sec--;
	t1->tv_usec += 1000;
    }
    if (t1->tv_usec >= 1000) {
	t1->tv_sec++;
	t1->tv_usec -= 1000;
    }
}
