/*-------------------------------------------------------------------------
 *
 * execs.c--
 *    Utilities for the sitemgr to manage and keep track of exec processes
 *    under its command.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/execs.c,v 1.33 1996/03/04 09:07:14 aoki Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <rpc/rpc.h>
#include <netdb.h>

#include "mariposa_rpc.h"
#include "sitemgr/sitemgr.h"
#include "queries.h"
#include "rpcerror.h"
#include "exchange.h"
#include "datafrag.h"

#include "rpc/rpcutils.h"

#include <pthread.h>		/* for the fork() jacket */
#include <cma.h>		/* for the fork() jacket */
#include <sys/socket.h>

#include <pwd.h>
#include "libpq/pqcomm.h"
#include "nodes/pg_list.h"

/* These will become Tcl variables */
int numAssignedExecs;
int numReadyExecs;
int numExecsNeeded;

#define EXECDEBUG 1

/*
 * The information on active exec's is kept in the array execs[].
 * The array is indexed from 0...MAX_OPEN_EXECS by execId's (a fancy
 * name for integers).  The execId's are kept in exec_list's.
 *
 * We keep three lists of execId's:
 *
 *	freeExecs  - list of execId's not used for an exec.
 *	readyExecs - list of execId's which have been assigned an exec but
 *		     are not currently executing a query.
 *	busyExecs  - list of execId's for execs which are currently in use.
 *
 */

struct exec_list {
    int execId;
    struct exec_list  *next;
};
typedef struct exec_list *ExecListPtr;

static ExecListPtr freeExecs;
static ExecListPtr readyExecs;
static ExecListPtr busyExecs;

static ExecInfo execs[MAX_OPEN_EXECS];

/* 
 * PidGetExecInfo -
 *    Given the pid of an active exec, return the information about it.
 */

ExecInfo *
PidGetExecInfo(int pid)
{
    int i = 0;

    while (i < MAX_OPEN_EXECS) {
	if (execs[i].pid == pid) {
	    return (&execs[i]);
	}
	i++;
    }

    return (NULL);
}

int
ExecIdGetPid(int execId)
{
    return execs[execId].pid;
}

/*
 * QueryIdGetExecInfo -
 *    Given a query id, get the information on the exec processing
 *    that query.
 *
 *    XXX - eventually we will also use xact id and command id to
 *	    identify the query.
 */

ExecInfo *
QueryIdGetExecInfo(QepFragmentId qepFragId)
{
    ExecListPtr l = busyExecs;
    ExecInfo *execPtr = NULL;
    
    while(l != NULL) {

	if (execs[l->execId].qepFragId == qepFragId) {
	    execPtr = &(execs[l->execId]);
	    sitemgrDebug(2, "QueryIdGetExecInfo: Assigned exec %i\n", l->execId);
	    break;
	}

	l = l->next;
    }

    return execPtr;
}

/*
 * ExecDeletePid--
 *
 * 	Delete the exec with the given pid from the given list
 *	if it's there.  Return TRUE if pid is in list, FALSE else.
 */

bool
ExecDeletePid(int pid, ExecListPtr execList, ExecListPtr *execDeleted) {
    ExecListPtr l, l2;
    
    l = execList;
    l2 = NULL;
    
    while (l != NULL) {
	if (execs[l->execId].pid == pid) {
	    *execDeleted = l;
	    if (l == execList) {
		execList = execList->next;
	    } else {
		l2->next = l->next;
	    }
	    return TRUE;
	}
	l2 = l;
	l = l->next;
    }

    return FALSE;
}

/*
 * ExecKilled--
 *
 *	Called from CleanUpProc() when an exec has been killed.
 *	Takes the exec off the ready list or the busy list and
 *	puts the execId back on the freeList.
 */
void
ExecKilled(int pid) {

    ExecListPtr execKilled;
    
    if (ExecDeletePid(pid, readyExecs, &execKilled)) {
	numReadyExecs--;
	numAssignedExecs--;
	execKilled->next = freeExecs;
	freeExecs = execKilled;
    } else if (ExecDeletePid(pid, busyExecs, &execKilled)) {
	numAssignedExecs--;
	execKilled->next = freeExecs;
	freeExecs = execKilled;
    }
}

/*
 * ExecStart -
 *    Start a new backend to accept queries.
 */
int
ExecStart()
{
    int pid;
    char **argv;
    int execId;
    ExecListPtr l;
    struct passwd *pw;
    StartupInfo si;
    Port beport;
    extern char *backendPath;

    /* make sure there's a free backend,
       if not, either block or error */
    if (numAssignedExecs >= MAX_OPEN_EXECS) {
	if (BLOCK_IF_NO_EXECS) {
	    /* NOT YET IMPLEMENTED */
	    sitemgrDebug(0, "ExecStart: not enough backends available\n");
	    exit(1);
	}
	else {
	    return(SERVICE_REFUSED);
	}
    }

    /* now we have a free backend for sure */
    numAssignedExecs++;
    execId = freeExecs->execId;

    Assert(execId < MAX_OPEN_EXECS);
	   
    /* delete from free list */
    l = freeExecs;
    freeExecs = freeExecs->next;

    /* note: we don't add to the ready list until it actually is ready */
    free(l);

    /* initialize execInfo for this exec */
    execs[execId].alive = 1;
    execs[execId].udpPort = NULL;
    execs[execId].tcpPort = NULL;
    execs[execId].fragid = 0;

    sitemgrLogEvent("mpExec f %d", execId);

    /* set up information for BackendStartup */
    si.database[0] = '\0';
    pw = getpwuid(getuid());
    (void) strcpy(si.user, pw->pw_name);
    (void) strcpy(si.options, " -m ");		/* m for mariposa */
    (void) strcpy(si.execFile, backendPath);
    si.tty[0] = '\0';
    memset(&beport, 0, sizeof(beport));
    
    BackendStartup(&si, &beport, &pid);

    execs[execId].pid = pid;

#ifdef EXECDEBUG    
    fprintf(stderr, "ExecStart: execId=%d pid=%d\n", execId, pid);
#endif

    return(0);
}

void
BrokerStart()
{
    StartupInfo si;
    struct passwd *pw = getpwuid(getuid());
    extern char *backendPath;
    Port beport;
    int bepid;

    (void) memset(&si, 0, sizeof(StartupInfo));
    (void) strcpy(si.user, pw->pw_name);
    (void) strcpy(si.options, " -r ");
    (void) strcpy(si.execFile, backendPath);
    
    (void) memset(&beport, 0, sizeof(Port));

    BackendStartup(&si, &beport, &bepid);
}

/*
 * ExecProcessQuery -
 *    Pass the query to an available exec, modifying exec info
 *    and the ready and busy lists accordingly.
 *
 *    XXX - Eventually we will also use xid and command id to
 *	    identify the query.
 *
 */

void 
ExecProcessQuery(ProcessQueryArg details, QepFragmentId qepFragId)
{
    int execId;
    SiteAddr *execAddr;

    execAddr = popReadyExec(qepFragId, &execId);


    /* call the backend - execProcessQuery is a wrapper for the 
       rpc call */

    sitemgrDebug(3, "ExecProcessQuery: calling execProcessQuery with execId %d, tcpPort %d\n", execId, execAddr->port);
    execProcessQuery(execAddr, &details);
}

/*
 * ExecFragX -
 *    Pass the request to an available exec, modifying exec info
 *    and the ready and busy lists accordingly.
 *
 *    XXX - Eventually we will also use xid and command id to
 *	    identify the query.
 *
 */
RPCErrorCode 
ExecFragX(FragXArg details, QepFragmentId qepFragId)
{
    int execId;
    SiteAddr *execAddr;

    execAddr = popReadyExec(qepFragId, &execId);

    /* call the backend - execFragX is a wrapper for the rpc call */
    return execFragX(execAddr, &details);
}

/*
 * ExecTwopcStart -
 *    Pass the request to an available exec, modifying exec info
 *    and the ready and busy lists accordingly.
 *
 *    XXX - Eventually we will also use xid and command id to
 *          identify the query.
 *
 */
RPCErrorCode
ExecTwopcStart(SiteAddr *sender, QepFragmentId qepFragId)
{
    int execId;
    SiteAddr *execAddr;

    execAddr = popReadyExec(qepFragId, &execId);

    /*
     * call the backend - ExecTwopcStart is a wrapper for the rpc call
     * in sitemgr.c
     */
    return execTwopcStart(execAddr, sender);
}

/*
 * ExecFinishedQuery -
 *    Called when an exec has finshed processing a query, to modify
 *    the necessary records.
 *
 */

void
ExecFinishedQuery(int pid)
{
    ExecListPtr l = busyExecs;
    ExecListPtr l1;

    extern int loadAverage;
    
    while (l != NULL && execs[l->execId].pid != pid) {
	l1 = l;
	l = l->next;
    }

    if (l == NULL) {
	fprintf(stderr, "ExecFinishedQuery: exec not engaged.\n");
	exit(1);
    } else if (l == busyExecs) {
	/* first item */
	/* delete from busy list */
	busyExecs = l->next;
    } else {
	/* not first item */
	/* delete from busy list */
	l1->next = l->next;
    }	

    sitemgrLogEvent("mpExec e %d", l->execId);
    if (execs[l->execId].fragid!=0) {
	sitemgrLogEvent("mpUse e %d", execs[l->execId].fragid);
    }
    
    /* add to ready list */
    l->next = readyExecs;
    readyExecs = l;

#ifdef EXECDEBUG
    fprintf(stderr, "ExecFinishedQuery: execId=%d numready=%d numneeded=%d\n", 
	    l->execId, numReadyExecs, numExecsNeeded);
#endif    

    numReadyExecs++;
    numExecsNeeded--;
    loadAverage--;
}

/* 
 * ExecReady -
 *    Called when a newly started backend indicates that it is ready
 *    to receive queries.
 */

void
ExecReady(int pid)
{
    int i = 0;
    int execId;
    ExecListPtr l;

    while(i < MAX_OPEN_EXECS) {
	if (execs[i].pid == pid) {
	    execId = i;
	    break;
	}
	i++;
    }

    if (i == MAX_OPEN_EXECS) {
	/* error */
	sitemgrDebug(0, "ExecReady: exec not found\n");
	exit(1);
    }

    /* create exec_list entry and add to ready list */
    l = malloc(sizeof(struct exec_list));

    l->execId = execId;
    l->next = readyExecs;
    readyExecs = l;

#ifdef EXECDEBUG    
    fprintf(stderr, "ExecReady: execId=%d\n", execId);
#endif
    
    numReadyExecs++;
}

/* 
 * ExecKillAll -
 *    Called to terminate all exec process.  Not yet implemented.
 *
 */

void
ExecKillAll(void)
{
    int i;

    for(i = 0; i < MAX_OPEN_EXECS; i++) {
	if (execs[i].pid != 0) {
fprintf(stderr, "killing %d\n", execs[i].pid);
	    kill(execs[i].pid, SIGTERM);
	}
    }
}

/* 
 * ExecsInit - 
 *    Initialize the exec information and ready, busy and free lists.
 *
 */

void
ExecsInit(void)
{
    int i;
    ExecListPtr l;

    numAssignedExecs = 0;
    numExecsNeeded = 0;
    numReadyExecs = 0;

    readyExecs = NULL;
    busyExecs = NULL;
    
    for (i=MAX_OPEN_EXECS - 1; i>=0; i--) {
	l = malloc(sizeof(struct exec_list));
	l->execId = i;
	l->next = freeExecs;
	freeExecs = l;

	execs[i].alive = 0;
	execs[i].pid = 0;
	execs[i].fragid = 0;
    }

    /*
     * clean up the mess we leave around.
     */
    on_exitpg(ExecKillAll, 0);
}

/* 
 * GetFreeExec -
 *    This exists so that an exec can be started independently of
 *    the sitemgr, and then report its existence by calling
 *    BackendReady.  This is useful for debugging.
 *
 */

ExecInfo *
GetFreeExec(void)
{
    ExecInfo *result;
    ExecListPtr l;

    if (freeExecs == NULL) {
	sitemgrDebug(0, "GetFreeExec: No execs available\n");
	exit(1);
    }

    l = freeExecs;
    freeExecs = freeExecs->next;

    result = &(execs[l->execId]);
    
    numAssignedExecs++;

    return result;
}

/*****************************************************************************
 *
 *****************************************************************************/

SiteAddr *
popReadyExec(int qepFragId, int *execIdPtr)
{
    ExecListPtr l;
    int execId;

    l = readyExecs;
    execId = l->execId;
    readyExecs = readyExecs->next;
    l->next = busyExecs;
    busyExecs = l;

    execs[execId].qepFragId = qepFragId;
    execs[execId].fragid = 0;

#ifdef EXECDEBUG
    fprintf(stderr, "popExec: execId=%d tcpPort=%d numReady=%d\n",
	    execId, execs[execId].tcpPort->port, numReadyExecs);
#endif    
    
    sitemgrLogEvent("mpExec b %d", execId);
    numReadyExecs--;

    *execIdPtr = execId;

    Assert(execs[execId].tcpPort != NULL);
    Assert(execs[execId].tcpPort->port > 0);

    return execs[execId].tcpPort;
}

void
pushDoneExec(int execId)
{
    ExecListPtr l = busyExecs;
    ExecListPtr l1;

    while (l != NULL && l->execId != execId) {
	l1 = l;
	l = l->next;
    }

    if (l == NULL) {
	fprintf(stderr, "ExecFinishedQuery: exec not engaged.\n");
	exit(1);
    } else if (l == busyExecs) {
	/* first item */
	/* delete from busy list */
	busyExecs = l->next;
    } else {
	/* not first item */
	/* delete from busy list */
	l1->next = l->next;
    }	
    
    sitemgrLogEvent("mpExec e %d", l->execId);

#ifdef EXECDEBUG
    fprintf(stderr, "pushExec: execId=%d numReady=%d\n",
	    execId, numReadyExecs);
#endif    

    /* add to ready list */
    l->next = readyExecs;
    readyExecs = l;

    numReadyExecs++;
    numExecsNeeded--;
}
