
/*-------------------------------------------------------------------------
 *
 * rules.c--
 *    The rule engine.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/rules.c,v 1.28 1996/06/14 20:23:27 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include "libpq/pqsignal.h"	/* substitute for <signal.h> */

#include <pthread.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include "tcl.h"

extern int		errno;
Tcl_Interp* tclInterp = NULL;
static Tcl_DString command;	/* Used to buffer incomplete commands being
				 * read from stdin. */
static int lastRetn;
bool EnableTcl = 1;

void tclRegisterFuncs _ANSI_ARGS_(());

/*
 * single lock around eval'ing Tcl code.
 */
static pthread_mutex_t tclLock;

/* 
 * single place for Tcl interpreter thread to find the next statement to
 * execute.
 */
static char *tclStatement = NULL;

/*
 * main loop thread saying "I have a statement available."
 */
static pthread_cond_t tclStmtAvailCond;

/*
 * Tcl interpreter thread saying "I'm done with the current statement."
 */
static pthread_cond_t tclStmtDoneCond;

/*
 * If you're wondering why we need tclFreeCond, it is to prevent two
 * threads calling tclEvaluate from inserting their statement to tclStatement
 * before the rule engine thread wakes up (note that tclEvaluate gives
 * up tclLock while waiting on tclStmtDoneCond and tclRuleEngineRun
 * gives up tclLock while waiting on tclStmtAvailCond, and a third thread
 * can get tclLock before either wakes up.)
 */
static pthread_cond_t tclFreeCond;

/*
 * tclEvaluate -
 *    evaluate the Tcl statement. This routine is synchronous (ie. returns
 *    only when the tcl statment has been evaluated.  The return is malloc'd().
 */
int
tclEvaluate(char *statement, char** result)
{
    int retn;
    pthread_mutex_lock(&tclLock);

    if (!EnableTcl) {
	pthread_mutex_unlock(&tclLock);
	return TCL_OK;
    }

    /*
     * store a copy of the statement to evaluate in the slot
     * when is free (ie. tcl does not have an outstanding query to evaluate)
     */
    while (tclStatement != NULL) {
	sitemgrDebug(2, "tclEvaluate: waiting for tcl engine.");
	pthread_cond_wait(&tclFreeCond, &tclLock);
    }
    tclStatement = statement;

    /*
     * Tell tclRuleEngineRun that a new statement is ready.
     * Then wait for it to finish.
     */
    pthread_cond_signal(&tclStmtAvailCond);
    pthread_cond_wait(&tclStmtDoneCond, &tclLock);
    if (result != NULL) {
	*result = strdup(tclInterp->result);
    }
    retn = lastRetn;
    pthread_mutex_unlock(&tclLock);
    return retn;
}

/*
 * a stripped-down version of Tcl_Main appropriate for Mariposa.
 */
void
Tcl_Main()
{
    char buffer[1000], *cmd, *args, *fileName;
    int code, gotPartial, tty = 1;
    int exitCode = 0;
    extern char* DataDir;
    extern char* BidderDir;
    extern char* DataBrokerDir;
    extern int MyHostId;
    
    tclInterp = Tcl_CreateInterp();
    tclRegisterFuncs();


    /*
     * Make hostid visible to Tcl interpreter.
     */

    sprintf(buffer, "set hostid %i", MyHostId);
    Tcl_Eval(tclInterp, buffer);

    sprintf(buffer, "%s%s", DataDir, "/files/rules.tcl");
    Tcl_EvalFile(tclInterp, buffer);

    /*
     * Read in GetQueryBid procedure from BIDDERDIR
     */

    sprintf(buffer, "%s%s", BidderDir, "/bidder.tcl");
    Tcl_EvalFile(tclInterp, buffer);

    /*
     * Read in DataBroker procedure from DATABROKERDIR
     */

    sprintf(buffer, "%s%s", DataBrokerDir, "/databroker.tcl");
    Tcl_EvalFile(tclInterp, buffer);
    
    /*
     * Set the "tcl_interactive" variable.
     */

    Tcl_SetVar(tclInterp, "tcl_interactive", "1", TCL_GLOBAL_ONLY);

    /*
     * Process commands from stdin until there's an end-of-file.
     */

    gotPartial = 0;
    Tcl_DStringInit(&command);
    while (1) {
	clearerr(stdin);
	if (tty) {
	    char *promptCmd;

	    promptCmd = Tcl_GetVar(tclInterp,
		gotPartial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
	    if (promptCmd == NULL) {
		defaultPrompt:
		if (!gotPartial) {
		    fputs("tclDMT> ", stdout);
		}
	    } else {
		char *result;
		code = tclEvaluate(promptCmd, &result);
		if (code != TCL_OK) {
		    fprintf(stderr, "%s\n", result);
		    Tcl_AddErrorInfo(tclInterp,
			    "\n    (script that generates prompt)");
		    free(result);
		    goto defaultPrompt;
		}
		free(result);
	    }
	    fflush(stdout);
	}
	if (fgets(buffer, 1000, stdin) == NULL) {
	    if (ferror(stdin)) {
		if (errno == EINTR) {
		    if (tcl_AsyncReady) {
			(void) Tcl_AsyncInvoke((Tcl_Interp *) NULL, 0);
		    }
		    clearerr(stdin);
		} else {
		    goto done;
		}
	    } else {
		if (!gotPartial) {
		    goto done;
		}
	    }
	    buffer[0] = 0;
	}
	cmd = Tcl_DStringAppend(&command, buffer, -1);
	if ((buffer[0] != 0) && !Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    continue;
	}

	gotPartial = 0;
	{
	    char *result;
	    code = tclEvaluate(cmd, &result);
	    Tcl_DStringFree(&command);
	    if (code != TCL_OK) {
		fprintf(stderr, "%s\n", result);
	    } else if (tty && (*result != 0)) {
		printf("%s\n", result);
	    }
	    free(result);
	}
    }

    /*
     * Rather than calling exit(), invoke the "exit" command so that
     * users can replace "exit" with some other command to do additional
     * cleanup on exit.  The Tcl_Eval call should never return.
     */
    done:
/* only kill thread */
#if 0
    sprintf(buffer, "exit %d", exitCode);
    tclEvaluate(buffer, NULL);
#else
    return;
#endif
}


void
tclRuleEngineRun()
{
    /*
     * init tcl, the locks and the condition variables
     */
    pthread_mutex_init(&tclLock, pthread_mutexattr_default);
    pthread_cond_init(&tclFreeCond, pthread_condattr_default);
    pthread_cond_init(&tclStmtAvailCond, pthread_condattr_default);
    pthread_cond_init(&tclStmtDoneCond, pthread_condattr_default);
    tclStatement = NULL;

    /* we're multithreaded anyway - 
     * let's fork a thread to read the tty for tcl statements - asah */
    {
	pthread_t thread;
	sitemgrDebug(2, "tclRuleEngineRun: started readline loop.\n");
	pthread_create(&thread, pthread_attr_default,
		       (void*)Tcl_Main, NULL);
	pthread_detach(&thread);
    }	    

    /*
     * process tcl statements (forever)
     */

    for(;;) {
	pthread_mutex_lock(&tclLock);
	while (tclStatement == NULL) {
	    pthread_cond_wait(&tclStmtAvailCond, &tclLock);
	}

	/*
	 * we got a tcl statement, now process it.
	 */
	lastRetn = Tcl_Eval(tclInterp, tclStatement);

	/*
	 * reset the statement slot and signal that we're done
	 */
	tclStatement = NULL;
	pthread_mutex_unlock(&tclLock);
	pthread_cond_signal(&tclStmtDoneCond);
	pthread_cond_signal(&tclFreeCond);
    }
}
