/*------------------------------------------------------------------------- * server.c -
 *    RPC server routines for the sitemgr.  
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/server.c,v 1.27 1996/06/14 20:23:28 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <stdio.h>
#include <rpc/rpc.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "mariposa_rpc.h"
#include "sitemgr/sitemgr.h"
#include "rpc/rpcutils.h"
#include "databroker.h"

#include "rpctypes.h"
#include "utils/elog.h"
#include "rpc/nameserver.h"
#include "nodes/pg_list.h"
#include "nodes/plannodes.h"
#include "scheduler/sched.h"
#include "sitemgr/cpupdate.h"

static void sitemgrprog_1();

unsigned short udpPort;
unsigned short tcpPort;

/*
 * sitemgr_deregister --
 *	Performs cleanup when the sitemgr calls exit(3).
 *	- deregister RPC program/version/port mapping
 */
static void
sitemgr_deregister()
{
    /* We add hostid to program number so we contact the correct sitemgr */
    (void) svc_unregister(SITEMGRPROG + getMyHostId(), SITEMGRVERS);
}


/*
 * sitemgr_svc_setup -
 *    Registers rpc services.
 */
void
sitemgr_svc_setup(void)
{
    register SVCXPRT *transp;
    int tcpsock;
    struct sockaddr_in tcpsockaddr;

    /* We add hostid to program number so we contact the correct sitemgr */
    (void) pmap_unset(SITEMGRPROG + getMyHostId(), SITEMGRVERS);

    /* For now mariposa only uses tcp, but we create a udp service
     * anyway in case we change our mind later.
     */
    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
	sitemgrDebug(0, "sitemgr_svc_main: cannot create udp service\n");
	exit(1);
    }

#ifdef RPCDEBUG
    fprintf(stderr, "SITEMGRPROG: %d registered\n",
	    SITEMGRPROG + getMyHostId());
#endif    
    
    /* We add hostid to program number so we contact the correct sitemgr */
    if (!svc_register(transp, SITEMGRPROG + getMyHostId(), SITEMGRVERS,
		      sitemgrprog_1, IPPROTO_UDP))
    {
	sitemgrDebug(0,
		     "sitemgr_svc_main: unable to register (SITEMGRPROG, SITEMGRVERS, udp)\n");
	exit(1);
    }
    udpPort = transp->xp_port;

    if ((tcpsock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	sitemgrDebug(0, "sitemgr_svc_main: cannot create tcp socket\n");
	exit(1);
    }

    tcpsockaddr.sin_port = SITEMGRPORT;
    tcpsockaddr.sin_addr.s_addr = gethostid();
    if (bind(tcpsock, &tcpsockaddr, sizeof(struct sockaddr_in)) == -1) {
	sitemgrDebug(0, 
		"sitemgr_svc_main: cannot bind tcp socket to sitemgr port\n");
	exit(1);
    }

    transp = svctcp_create(tcpsock, 0, 0);
    if (transp == NULL) {
	sitemgrDebug(0, "sitemgr_svc_main: cannot create tcp service\n");
	exit(1);
    }
    /* We add hostid to program number so we contact the correct sitemgr */
    if (!svc_register(transp, SITEMGRPROG + getMyHostId(), SITEMGRVERS,
		      sitemgrprog_1, IPPROTO_TCP))
    {
	sitemgrDebug(0, 
		"sitemgr_svc_main: unable to register (SITEMGRPROG, SITEMGRVERS, tcp)\n");
	exit(1);
    }
    tcpPort = transp->xp_port;

    atexit(sitemgr_deregister);
}

/*
 * sitemgrprog_1 -
 *    Handler for RPC requests, called by svc_run.
 */
static void
sitemgrprog_1(struct svc_req *rqstp, SVCXPRT *transp)
{
    sitemgrDebug(3, "sitemgrprog_1 called\n");

    switch (rqstp->rq_proc) {
    case NULLPROC:
	(void) svc_sendreply(transp, xdr_void, (char *)NULL);
	break;

    case CONTRACTWON:
	rpcAsyncHandleRequest(transp,
			      xdr_WinBuffer,
			      sizeof(WinBuffer),
			      ContractWon,
			      "ContractWon");
	break;
	
    case RECEIVESHORTOFFER:
	rpcAsyncHandleRequest(transp,
			      xdr_OfferBuffer,
			      sizeof(OfferBuffer),
			      ReceiveOffer,
			      "ReceiveOffer");
	break;

    case SUBBROKERDONE:
	rpcAsyncHandleRequest(transp,
			      xdr_BrokerAnswer,
			      sizeof(BrokerAnswer),
			      SubBrokerDone,
			      "SubBrokerDone");
	break;

	
    case SITEMGRTRANSMITDATA: 
	rpcAsyncHandleRequest(transp,
			      xdr_TransmitBuffer,
			      sizeof(TransmitBuffer),
			      TransmitData,
			      "TransmitData");
	break;

    case NSVC_SITEMGRLOOKUPCLASS:
	rpcAsyncHandleRequest(transp,
			      xdr_string,
			      sizeof(char *),
			      LookupClass,
			      "LookupClass");
	break;

    case NSVC_SITEMGRLOOKUPFRAGS:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      LookupFrags,
			      "LookupFrags");
	break;

    case BACKENDREADY:
	rpcAsyncHandleRequest(transp,
			      xdr_BackendBuffer,
			      sizeof(BackendBuffer),
			      BackendReady,
			      "BackendReady");
	break;
	
    case BACKENDDONE: 
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      BackendDone,
			      "BackendDone");
	break;
	
    case CONTRACTLOST:
	rpcAsyncHandleRequest(transp,
			      xdr_ContractId,
			      sizeof(ContractId),
			      ContractLost,
			      "ContractLost");
	break;
	
    case REQUESTFORBID:
	rpcAsyncHandleRequest(transp,
			      xdr_OfferBuffer,
			      sizeof(OfferBuffer),
			      ReceiveRequestForBid,
			      "ReceiveRequestForBid");
	break;
	
    case CP_TOGGLE:
	rpcAsyncHandleRequest(transp,
			      xdr_void,
			      0,
			      Cp_Toggle,
			      "Cp_Toggle");
	break;
	
    case CP_SCHEDULESTREAM:
	rpcAsyncHandleRequest(transp,
			      xdr_UpdateContractInfo,
			      sizeof(UpdateContractInfo),
			      Cp_ScheduleStream,
			      "Cp_ScheduleStream");
	break;
	
    case CP_UNSCHEDULESTREAM:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      Cp_UnscheduleStream,
			      "Cp_UnscheduleStream");
	break;
	
    case CP_UPDATETRANSFERSETUP:
	rpcAsyncHandleRequest(transp,
			      xdr_SiteAddr,
			      sizeof(SiteAddr),
			      Cp_UpdateTransferSetup,
			      "Cp_UpdateTransferSetup");
	break;

    case CALLCANCELCONTRACT:
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      CallCancelContract,
			      "CallCancelContract");
	return;
	
    case NSVC_POLL:
	rpcAsyncHandleRequest(transp,
			      xdr_NSVCPollList,
			      sizeof(NSVCPollList),
			      NSVC_Poll,
			      "NSVC_Poll");
	break;
	
    case NSVC_PROPAGATE:
	rpcAsyncHandleRequest(transp,
			      xdr_NSVCItem,
			      sizeof(NSVCItem),
			      NSVC_Propagate,
			      "NSVC_Propagate");
	break;
	
    case NSVC_SITEMGRUPDATE:
	{
	    NSVCUpdate arg;

	    if (!svc_getargs(transp, xdr_NSVCUpdate, &arg)) {
		sitemgrDebug(0,
			     "sitemgrprog_1 : cannot get arguments for NSVC_SITEMGRUPDATE\n");
		exit(1);
	    }


	    /*
	     * right now, only output an event to drive the GUI
	     */
	    if (arg.action==1) { 
		sitemgrLogEvent("mpMove %d", arg.frastoreid); /* departed */
	    } else {
		sitemgrLogEvent("mpFrag %d", arg.frastoreid); /* arrived */
	    }

	    if (!svc_sendreply(transp, xdr_void, NULL)) {
		sitemgrDebug(0,
			     "sitemgrprog_1 : cannot reply for NSVC_SITEMGRUPDATE\n");
		exit(1);
	    }
	}
	break;
	
    case SCHED_SUBMIT_QUERY: 
	rpcAsyncHandleRequest(transp,
			      xdr_SchedArg,
			      sizeof(SchedArg),
			      SchedSubmitQuery,
			      "SchedSubmitQuery");
	break;
	
    case SCHED_GET_NEXT_SQ: 
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      SchedGetNextSQ,
			      "SchedGetNextSQ");
	break;
	
    case SCHED_SQ_DONE: 
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      SchedSubQueryDone,
			      "SchedSubQueryDone");
	break;
	
    case TELLLOST: 
	rpcAsyncHandleRequest(transp,
			      xdr_int,
			      sizeof(int),
			      TellLost,
			      "TellLost");
	break;

    case DATABROKER:
	rpcAsyncHandleRequest(transp,
			      xdr_DataBrokerArg,
			      sizeof(DataBrokerArg),
			      DataBroker,
			      "DataBroker");
	break;

    default:
	svcerr_noproc(transp);
	break;
    }

    return;
}

