/*-------------------------------------------------------------------------
 *
 * sitemgrrpc.c--
 *    handlers for sitemgr RPC calls
 
 * Contains:
 *
 *   ExecuteContract()
 *   receiveShortOffer()
 *   Cp_UpdateTransferSetup()
 *   TransmitData()
 *   BackendReady()
 *   BackendDone()
 *   ContractWon()
 *   ContractLost()
 *   ReceiveRequestForBid()
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/sitemgrrpc.c,v 1.29 1996/06/14 20:23:31 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <rpc/rpc.h>
#include <netinet/in.h>
#include <netdb.h>
#include "mariposa_rpc.h"
#include "sitemgr/sitemgr.h"
#include <pthread.h>
#include "queries.h"
#include "rpc/rpcutils.h"
#include <errno.h>

#include "rpcerror.h"
#include "sitemgr/bidder.h"
#include "access/htup.h"
#include "libpq/pqcomm.h"
#include "syncFrag.h"
#include "nameserver_mk.h"
#include "tcl.h"
#include "nodes/nodes.h"

#define EXECDEBUG 1

/* extern BidCollection *bidsRequested; */
extern int loadAverage;

int nextContractNo = 0;

Contract *contractsMade;

static void *ExecuteContract(OfferBuffer *offerBuffer, SVCXPRT *transp);
void SetAllProcSites(Plan *plan, SiteAddrNode *addr);
void sm_FillInQepFragIds(Plan *plan, int currentQepFragId);
static void sm_FillInQepFragId(Plan *child, Plan *parent, int currentQepFragId);
int sm_newQepFragId();
int GetNextContractNo();

/*
 * -------------------------------------------------------------------
 * ReceiveOffer
 *
 *    Decide whether to accept the offered contract and
 *    inform the broker of our decision.
 *
 * -------------------------------------------------------------------
 */

void
ReceiveOffer(void *args)
{
    char **argv = (char **)args;
    OfferBuffer *offerBuffer = (OfferBuffer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    bool acceptOffer;
    ContractId contract;
    
    sitemgrDebug(3, "ReceiveOffer called\n");

    switch (offerBuffer->protocol) {
	
    case SHORT_PROCESS_QUERY: {
    
	/*
	 * Decide whether or not to accept the offer.
	 */
	acceptOffer = TRUE;
    

	if (acceptOffer) {
	
	    /*
	     * Assign the contract ID.
	     */
	    contract = GetNextContractNo();
	    sitemgrDebug(1, "ReceiveOffer: SHORT_PROCESS_QUERY: Beginning contract %i for site %i.\n",
			 contract, offerBuffer->requester.addr);
	
	}
    }
	break;

	/*
	 * offerBuffer will have the contract ID in it that was generated
	 * here during the bidding process.  Check if there are subcontracts
	 * for this query fragment.  If so, substitute the query plan with
	 * the correct processing sites for the one passed in.  Be careful
	 * to set the root node's qepFragId to be the same as the root node
	 * qepFragId of the plan passed, since the receiving site will be
	 * expecting stuff to come from that qepFragId.
	 */
    case LONG_PROCESS_QUERY: {
	contract = offerBuffer->contract;
	if (contractsMade[contract].subBids) {
	    Plan *oldPlan = (Plan *)offerBuffer->details.OfferDetails_u.longQuery.plan;
	    Plan *newPlan = (Plan *)contractsMade[contract].details_u.queryDetails.plan;
	    sm_FillInQepFragIds(newPlan, oldPlan->qepFragId);
	    
	    offerBuffer->details.OfferDetails_u.longQuery.plan =
		contractsMade[contract].details_u.queryDetails.plan;
	}

	sitemgrDebug(1, "ReceiveOffer: LONG_PROCESS_QUERY: Awarded contract %i from site %i.\n",
		     contract, offerBuffer->requester.addr);
    
	
    }
	break;

    case SHORT_BUY_FRAGMENT: {
	contract = GetNextContractNo();

	sitemgrDebug(1, "ReceiveOffer: SHORT_BUY_FRAGMENT: Beginning contract %i from site %i.\n",
		     contract, offerBuffer->requester.addr);
	
    }
	break;

    case SHORT_SELL_FRAGMENT: {
	contract = GetNextContractNo();

	sitemgrDebug(1, "ReceiveOffer: SHORT_SELL_FRAGMENT: Beginning  contract %i from site %i.\n",
		     contract, offerBuffer->requester.addr);
    }
	break;
	
    case LONG_BUY_FRAGMENT: {
    }

    case LONG_SELL_FRAGMENT: {
    }
	
    default: {
    }
	break;
    }
    
    ExecuteContract(offerBuffer, transp);
}

/*
 *-------------------------------------------------------------------------
 * FillInQepFragIds--
 *
 *	Go through the plan, assigning a qepFragId to each node, depending
 *	on which query fragment it belongs to.  These are the fragments
 *	which will be passed to the executor sites to be processed.
 *-------------------------------------------------------------------------
 */
static
void
sm_FillInQepFragId(Plan *child, Plan *parent, int currentQepFragId)
{
    if (child != (Plan *) NULL) {
	if (child->processingSite.hostid == parent->processingSite.hostid) {
	    sm_FillInQepFragIds(child, currentQepFragId);
	} else {
	    sm_FillInQepFragIds(child, sm_newQepFragId());
	}
    }
}

void
sm_FillInQepFragIds(Plan *plan, int currentQepFragId)
{
    List *lp;
    Plan *p;

    plan->qepFragId = currentQepFragId;
    
    switch (plan->type) {
    case T_Merge:
	foreach(lp, ((Merge *) plan)->mergeplans) {
	    sm_FillInQepFragId((Plan *) lfirst(lp), plan, currentQepFragId);
	}
	break;
    case T_Append:
	foreach(lp, ((Append *) plan)->unionplans) {
	    sm_FillInQepFragId((Plan *) lfirst(lp), plan, currentQepFragId);
	}
	break;
    default:
	sm_FillInQepFragId(plan->lefttree, plan, currentQepFragId);
	sm_FillInQepFragId(plan->righttree, plan, currentQepFragId);
	break;
    }
}


/*
 * -------------------------------------------------------------------
 * ExecuteContract
 *
 *   This is the "just do it" routine.  If this site has contracted
 *   to process a query, process the query.  If it's contracted to
 *   sell a fragment, then sell the fragment.  Long/Short protocol
 *   stuff has all been taken care of.  That is, if the long protocol
 *   was being used, this site won and now has to deliver.  If the
 *   short protocol, this site is going to do the work and bill the
 *   broker.
 *
 *   Increment the number of exec's needed.  Once an exec
 *   becomes available, call ExecProcessQuery(), ExecSellFragment()
 *   or whatever.
 * -------------------------------------------------------------------
 */
static void *
ExecuteContract(OfferBuffer *offerBuffer, SVCXPRT *transp)
{
    Budget budget = offerBuffer->budget;
    OfferDetails details = offerBuffer->details;
    RPCErrorCode result;
    QepFragmentId qepFragId;
    
    switch (details.protocol) {

    case LONG_PROCESS_QUERY: {

	Plan *plan = (Plan *)details.OfferDetails_u.longQuery.plan;
	CopyBudgetIntoPlan(&budget, plan);
	
	qepFragId = plan->qepFragId;
      
	break;
    }

    case SHORT_PROCESS_QUERY: {
	Plan *plan = (Plan *)details.OfferDetails_u.shortQuery.plan;
	qepFragId = plan->qepFragId;
	CopyBudgetIntoPlan(&budget, plan);
	
	break;
    }

    case LONG_SELL_FRAGMENT: {
	qepFragId = details.OfferDetails_u.longSell.qepFragId;
      
	break;
    }

    case SHORT_SELL_FRAGMENT: {
	qepFragId = details.OfferDetails_u.shortSell.qepFragId;
      
	break;
    }
      
    case LONG_BUY_FRAGMENT: {
	qepFragId = details.OfferDetails_u.longBuy.qepFragId;
      
	break;
    }
 
    case SHORT_BUY_FRAGMENT: {
	qepFragId = details.OfferDetails_u.shortBuy.qepFragId;
      
	break;
    }

    case SHORT_COPY_FRAGMENT: {
        qepFragId = details.OfferDetails_u.shortCopy.qepFragId;

        break;
    }


    default: {
	result = BAD_PROTOCOL;
	if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	    sitemgrDebug(0, "ExecuteContract: svc_sendreply failed\n");
	}
	else {
	    sitemgrDebug(2, "ExecuteContract: svc_sendreply succeeded, result = %d\n",
			 result);
	}
	pthread_exit(NULL);
	break;
    }

      
    }
    
    sitemgrDebug(2, "ExecuteContract: QepFragId = %d\n",
		 qepFragId);
    
    if (QueryIdGetExecInfo(qepFragId) != NULL) {
	/* exec already dedicated to this query, ignore */
	sitemgrDebug(1, "ExecuteContract: exec already dedicated to this query (qepFragId=%d)\n",
		     qepFragId);

	result = 0;
	return (&result);
    }
    
    sitemgrDebug(2, "ExecuteContract: acquiring execsLock\n");
    pthread_mutex_lock(&execsLock);
    
    numExecsNeeded++;
    
    sitemgrDebug(2, "ExecuteContract: %d execs assigned, %d needed\n", 
		 numAssignedExecs, numExecsNeeded);
    
    if (numAssignedExecs < numExecsNeeded) {
	/* need to open a new backend */

	sitemgrDebug(1, "ExecuteContract: Starting exec\n");

	if (ExecStart() == SERVICE_REFUSED) {
	    sitemgrDebug(1, "ExecuteContract: Cannot start exec\n");
	    result = SERVICE_REFUSED;
	    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
		sitemgrDebug(0, "ExecuteContract: svc_sendreply failed\n");
	    } else {
		sitemgrDebug(2, "ExecuteContract: svc_sendreply succeeded, result = %d\n",
			     result);
	    }
	    pthread_exit(NULL);
	}

    }
    
    /*
     * before sending a query to an exec we must check
     * that there is one ready to run a query
     */
    while (numReadyExecs == 0) { 
	sitemgrDebug(1, "ExecuteContract: Waiting for ready backend\n");
	pthread_cond_wait(&execAvailableCond, &execsLock);
    } 
    
    sitemgrDebug(1, "ExecuteContract: sending query to backend\n");
    
    result = SUCCESS;

    switch (details.protocol) {
    case LONG_PROCESS_QUERY: {
	ExecProcessQuery(details.OfferDetails_u.longQuery, qepFragId);
      
	break;
    }

    case SHORT_PROCESS_QUERY: {
	ExecProcessQuery(details.OfferDetails_u.shortQuery, qepFragId);
      
	break;
    }

    case LONG_SELL_FRAGMENT: {
	result = ExecFragX(details.OfferDetails_u.longSell, qepFragId);
      
	break;
    }

    case SHORT_SELL_FRAGMENT: {
	result = ExecFragX(details.OfferDetails_u.shortSell, qepFragId);
      
	break;
    }
      
    case LONG_BUY_FRAGMENT: {
	result = ExecFragX(details.OfferDetails_u.longBuy, qepFragId);
      
	break;
    }
 
    case SHORT_BUY_FRAGMENT: {
	result = ExecFragX(details.OfferDetails_u.shortBuy, qepFragId);
      
	break;
    }
      
    case SHORT_COPY_FRAGMENT: {
        result = ExecFragX(details.OfferDetails_u.shortCopy, qepFragId);
      
        break;
    }

    }
    
    
    /*
     * notify waiting threads that an exec has been assigned to a query
     */
    pthread_cond_broadcast(&execAssignedCond);
    
    /*
     * Update load average.  XXX Should make more intelligent
     * in the future.
     */
    loadAverage++;
    
    sitemgrDebug(2, "ExecuteContract: Releasing execsLock\n");
    pthread_mutex_unlock(&execsLock);
    
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "ExecuteContract: svc_sendreply failed\n");
    }
    else {
	sitemgrDebug(2, "ExecuteContract: svc_sendreply succeeded, result = %d\n",
		     result);
    }
}

/* -------------------------------------------------------------------
 *	Cp_UpdateTransferSetup
 *
 *	Called by a site wishing to send its updates to a fragment copy
 *	here. Sets up an exec backend to transfer the data to, and calls
 *	TWOPCSTART to that backend to start the process going.
 *	(The call to TWOPCSTART is actually done in execFragX.)
 *
 * -------------------------------------------------------------------
 */
void
Cp_UpdateTransferSetup(void *args)
{
    char **argv = (char **)args;
    SiteAddr *sender = (SiteAddr *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    RPCErrorCode result;

    sitemgrDebug(3, "Cp_UpdateTransferSetup called\n");
    
    sitemgrDebug(2, "Cp_UpdateTransferSetup: acquiring execsLock...\n");
    pthread_mutex_lock(&execsLock);
    sitemgrDebug(2, "acquired\n");
    
    numExecsNeeded++;
    
    sitemgrDebug(2, "Cp_UpdateTransferSetup: %d execs assigned, %d needed\n", 
		 numAssignedExecs, numExecsNeeded);
    
    if (numAssignedExecs < numExecsNeeded) {
	/* need to open a new backend */

	sitemgrDebug(1, "Cp_UpdateTransferSetup: Starting exec\n");

	if (ExecStart() == SERVICE_REFUSED) {
	    sitemgrDebug(1, "Cp_UpdateTransferSetup: Cannot start exec\n");
	    result = SERVICE_REFUSED;
	    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
		sitemgrDebug(0, "Cp_UpdateTransferSetup: svc_sendreply failed\n");
	    }
	    else {
		sitemgrDebug(2, 
			     "Cp_UpdateTransferSetup: svc_sendreply succeeded, result = %d\n",
			     result);
	    }
	    pthread_exit(NULL);
	}
    }
    
    /*
     * before sending a query to an exec we must check
     * that there is one ready to run a query
     */
    while (numReadyExecs == 0) { 
	sitemgrDebug(1, "Cp_UpdateTransferSetup: Waiting for ready backend\n");

	pthread_cond_wait(&execAvailableCond, &execsLock);
    } 
    
    sitemgrDebug(1, "Cp_UpdateTransferSetup: sending query to backend\n");

    result = ExecTwopcStart(sender, -1);
    
    /*
     * notify waiting threads that an exec has been assigned to a query
     */
    pthread_cond_broadcast(&execAssignedCond);
    
    /*
     * Update load average.  XXX Should make more intelligent
     * in the future.
     */
    loadAverage++;
    
    sitemgrDebug(2, "Cp_UpdateTransferSetup: Releasing execsLock...");
    pthread_mutex_unlock(&execsLock);
    sitemgrDebug(2, "released\n");
    
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "Cp_UpdateTransferSetup: svc_sendreply failed\n");
    }
    else {
	sitemgrDebug(2, 
		     "Cp_UpdateTransferSetup: svc_sendreply succeeded, result = %d\n",
		     result);
    }
}

/*
 * -------------------------------------------------------------------
 * TransmitData -
 *    Main function of thread that handles sitemgr TransmitData requests.
 *    The sitemgr acts as a warehouse for exchanging data from site to
 *    site.  The first time a remote exec has to pass data to a local
 *    one it does it through the sitemgr.  The sitemgr forwards the data
 *    to the correct exec, and returns the exec's port information to
 *    the client so that the data can be sent directly in future.
 *
 * -------------------------------------------------------------------
 */
void
TransmitData(void *args)
{
    char **argv = (char **)args;
    TransmitBuffer *transmitBuffer = (TransmitBuffer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    SiteAddr sender = transmitBuffer->sendingSite;

/*
 *  XXX - later use xids and command ids
 */

    QepFragmentId sendQepFragId = transmitBuffer->sendQepFragId;
    QepFragmentId recvQepFragId = transmitBuffer->recvQepFragId;
    int flags = transmitBuffer->flags;
    Data data = transmitBuffer->data;
    int mutexResult;
    extern int errno;
    
    TransmitReturn *result = (TransmitReturn *)malloc(sizeof(TransmitReturn));
    SiteAddr *execAddr = NULL; /* XXX what should I be? */
    RpcConn *rpcc;
    struct sockaddr_in *execSock = malloc(sizeof(struct sockaddr_in));
    RPCErrorCode error;
    char hostname[50];
    struct hostent *hostEntry;
    ExecInfo *exec = NULL;
    int execId;
    
    sitemgrDebug(1, "TransmitData: Called from <qep=%d host=%d>\n",
		 sendQepFragId, sender.hostid);

    sitemgrDebug(2, "TransmitData: acquiring execsLock... ");
    mutexResult = pthread_mutex_lock(&execsLock);

    if (mutexResult == 0) {
	sitemgrDebug(2, "acquired\n");
    } else {
	sitemgrDebug(0, "pthread_mutex_lock error: ");
	if (errno == EINVAL) {
	    sitemgrDebug(0, "invalid value for execsLock.\n");
	} else {
	    sitemgrDebug(0, "deadlock.\n");
	}
    }

    sitemgrDebug(2, "TransmitData: <qep=%d host=%d> -> <qep=%d host=%d>\n",
		 sendQepFragId, sender.hostid, recvQepFragId, getMyHostId());
    
    while (exec == NULL) {
	/*
	 * here the sendQepFragId identifies the node it came from
	 * and the recvQepFragId identifies the node at this end
	 *
	 * XXX FIX ME this isn't enough information to disambiguate
	 */
	exec = QueryIdGetExecInfo(recvQepFragId);

	if (exec == NULL) {
	    sitemgrDebug(2, "TransmitData: waiting for exec to be assigned\n");
	    pthread_cond_wait(&execAssignedCond, &execsLock);
	}
    }

    sitemgrDebug(2, "TransmitData: got exec for <qep=%d host=%d>\n",
		 recvQepFragId, getMyHostId());
    
    sitemgrDebug(2, "TransmitData: releasing execsLock\n");
    pthread_mutex_unlock(&execsLock);

    execAddr = exec->tcpPort;
    
    execSock->sin_family = AF_INET;
    execSock->sin_port = htons(execAddr->port);
    
    gethostname(hostname, 50);
    if ((hostEntry = gethostbyname(hostname)) == NULL) {
        sitemgrDebug(0, "TransmitData: can't get host entry for %s\n",
		     hostname);
	exit(1);
    }
    (void) memcpy((char *) &execSock->sin_addr.s_addr, hostEntry->h_addr,
		  hostEntry->h_length);

    sitemgrDebug(1, "TransmitData: exec for <qep=%d host=%d>: port=%d\n",
		 recvQepFragId, getMyHostId(), execSock->sin_port);

    rpcc =
	Rpc_connectExec(execAddr->addr, ntohs(execSock->sin_port),
			"TransmitData", RPC_ONERR_RETURN);
    
    if (rpcc == NULL) {
	sitemgrDebug(0, "TransmitData: Couldn't create client\n");
	exit(1);
    }
    
    result->port = *execAddr;
    result->error = SUCCESS;
    
    sitemgrDebug(2, "TransmitData: calling receiveData at exec <qep=%d host=%d>\n",
		 recvQepFragId, getMyHostId());
    Rpc_call(rpcc, RPC_TRANSMITDATA, transmitBuffer, &error);
    
    Rpc_disconnect(rpcc);

    if (!svc_sendreply(transp, xdr_TransmitReturn, result)) {
	sitemgrDebug(0, "TransmitData: svc_sendreply to <qep=%d host=%d> failed\n",
		     recvQepFragId, getMyHostId());
    } else {
	sitemgrDebug(2, "TransmitData: svc_sendreply to <qep=%d host=%d> succeeded, result = <port=%d addr=%d host=%d>\n",
		     recvQepFragId, getMyHostId(),
		     result->port.port, result->port.addr, result->port.hostid);
    }

}

/*
 * -------------------------------------------------------------------
 * BackendReady -
 *    Main function of thread that handles BackendReady requests.
 *    When a newly started backend is ready to receive queries, it
 *    sends a BackendReady message to inform the sitemgr.
 *
 * -------------------------------------------------------------------
 */
void
BackendReady(void *arg)
{
    char **argv = (char **)arg;
    BackendBuffer *backendBuffer = (BackendBuffer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    int pid = backendBuffer->pid;
    SiteAddr *udpPort = &(backendBuffer->udpPort);
    SiteAddr *tcpPort = &(backendBuffer->tcpPort);
    
    ExecInfo *execInfo;
    RPCErrorCode result;

    sitemgrDebug(3, "BackendReady called\n");

    sitemgrDebug(2, "BackendReady: Acquiring execsLock\n");
    pthread_mutex_lock(&execsLock);

    execInfo = PidGetExecInfo(pid);

    if (execInfo == NULL) {
	/* this means that the backend was started independently,
	 * probably for debugging
	 */
	extern ExecInfo *GetFreeExec();

	sitemgrDebug(1, 
		     "BackendReady: backend must have started independently\n");

	execInfo = GetFreeExec();
	execInfo->pid = pid;
    }

    udpPort->port = htons(udpPort->port);
    tcpPort->port = htons(tcpPort->port);
    execInfo->udpPort = udpPort;
    execInfo->tcpPort = tcpPort;

    ExecReady(pid);

#ifdef EXECDEBUG
    fprintf(stderr, "BackendReady: pid=%d tcpPort=%d addr=%d hostid=%d\n",
	    pid, ntohs(tcpPort->port), tcpPort->addr, tcpPort->hostid);
#endif
    
    sitemgrDebug(1, "BackendReady: Backend %d ready!\n", pid);

    /* notify waiting threads that a backend is now ready */
    pthread_cond_broadcast(&execAvailableCond);

    sitemgrDebug(2, "BackendReady: Releasing execsLock...\n");
    pthread_mutex_unlock(&execsLock);

    result = SUCCESS;

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "BackendReady: svc_sendreply failed\n");
    }
    else {
	sitemgrDebug(2, 
		     "BackendReady: svc_sendreply succeeded, result = %d\n",
		     result);
    }

}

/*
 * -------------------------------------------------------------------
 * BackendDone -
 *    Main function of thread that handles BackendDone requests.
 *    When a backend finishes processing a query and is ready to receive
 *    another, it sends a BackendDone message to inform the sitemgr.
 *
 * -------------------------------------------------------------------
 */

void
BackendDone(void *arg)
{
    char **argv = (char **)arg;
    int *pid = (int *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    RPCErrorCode result;
    RpcConn *rpcc;

    
    sitemgrDebug(3, "BackendDone called\n");

    sitemgrDebug(2, "BackendDone: Acquiring execsLock...\n");
    pthread_mutex_lock(&execsLock);

    ExecFinishedQuery(*pid);

    sitemgrDebug(1, "BackendDone: Backend %d done!\n", *pid);

    /* notify waiting threads that a backend is now ready */
    pthread_cond_broadcast(&execAvailableCond);

    sitemgrDebug(2, "BackendDone: Releasing execsLock...\n");
    pthread_mutex_unlock(&execsLock);

    result = SUCCESS;

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "BackendDone: svc_sendreply failed\n");
    }
    else {
	sitemgrDebug(2, 
		     "BackendDone: svc_sendreply succeeded, result = %d\n",
		     result);
    }

    if (scheduleQueries == true) {
       SchedBackendDone(*pid);
   }
}

/*
 * -------------------------------------------------------------------
 * ContractWon -
 *    Main function of thread that handles ContractWon requests.
 *    Only called during long protocol, so bidder has already
 *    agreed to perform the work for the specified price.  Set
 *    up an OfferBuffer, which is what ExecuteContract requires,
 *    and then call ExecuteContract().
 *
 *    Short protocol requests are handled by receiveShortOffer.
 * -------------------------------------------------------------------
 */
void
ContractWon(void *arg)
{
    char **argv = (char **)arg;
    WinBuffer *winBuffer = (WinBuffer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    ContractId contract = winBuffer->contract;
    OfferBuffer offbuf;   
    RPCErrorCode result;

    sitemgrDebug(3, "ContractWon called\n");

    /*
     * JPSXXX Don't return from here because we call ExecuteContract()
     * which will return via svc_sendreply().  This might have to be
     * changed - if ContractWon takes a long time, we don't want the
     * home site to sit around and wait until we're finished.  But this
     * would mean taking the svc_sendreply() call out of ExecuteContract().
     */
    
    /*
     * Fill in OfferBuffer to send to ExecuteContract.
     */
    
    offbuf.requestId = contractsMade[contract].requestId;
    offbuf.requestIndex = contractsMade[contract].requestIndex;
    offbuf.whichBid = contractsMade[contract].whichBid;
    offbuf.requester = contractsMade[contract].requester;
    offbuf.contract = contract;
    offbuf.budget = contractsMade[contract].bid;
    offbuf.details.protocol = contractsMade[contract].protocol;

    switch (offbuf.details.protocol) {

	/*
	 * Everything should be filled in here EXCEPT the recvSite,
	 * which was determined by the Broker after receiving all the
	 * bids.
	 */
    case LONG_PROCESS_QUERY: {
	ProcessQueryArg *qryOffer = &(offbuf.details.OfferDetails_u.longQuery);
	ProcessQueryArg *qryContract = &(contractsMade[contract].details_u.queryDetails);
	List *recvSites = lcons(&winBuffer->recvSite, NIL);
	
	qryOffer->operation = qryContract->operation;
	qryOffer->plan = qryContract->plan;
	qryOffer->rtable = qryContract->rtable;
	qryOffer->recvSites = (PQA_Int_list)recvSites;
	
    }
	break;

	
    case LONG_SELL_FRAGMENT: {
	offbuf.details.OfferDetails_u.longSell = contractsMade[contract].details_u.moveDetails;
    }
	break;
	
    case LONG_BUY_FRAGMENT: {
	offbuf.details.OfferDetails_u.longBuy = contractsMade[contract].details_u.moveDetails;
    }
	break;
	
    }
    
    ExecuteContract(&offbuf, transp);

    
}
 
/*
 * -------------------------------------------------------------------
 * ContractLost -
 *    Main function of thread that handles ContractLost RPC calls.
 *    This will automatically cause a trap to the rule system
 *    once Rush is integrated.
 * -------------------------------------------------------------------
 */

void
ContractLost(void *arg)
{
    char **argv = (char **)arg;
    ContractId contract = (ContractId)*argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    int subBid;
    RPCErrorCode result;
    int i;
    
    sitemgrDebug(3, "ContractLost called\n");
    
    result = SUCCESS;
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "ContractLost: svc_sendreply failed\n");
    } else {
	sitemgrDebug(2, 
		     "ContractLost: svc_sendreply succeeded, result = %d\n",
		     result);
    }

#if 0
    /*
     * If there are sub-bids for this contract, contact our
     * Broker() to indicate that the contract was not awarded to
     * us.  The broker thread that got all the sub-bids in
     * will be waiting inside BiddingDone.
     */
    if (contractsMade[contract].subBids) {
	SyncFragmentArg arg;
	SyncFragmentResult *result;
	int subBidCollIndex =
	    contractsMade[contract].subBidCollIndex;
	RpcConn *rpcc;
	    
	rpcc =
	    Rpc_connectLocalBroker("ContractLost", RPC_ONERR_RETURN);

/* 	if (!Rpc_call(rpcc, RPC_SUBCONTRACTLOST, &subBidCollIndex, NULL)) { */
/* 	    sitemgrDebug(0, "ContractLost: RPC error\n"); */
/* 	} */
	    
	Rpc_disconnect(rpcc);

    }

#endif
}

/*
 * -------------------------------------------------------------------
 * ReceiveRequestForBid -
 *    Main function of thread that handles RFB's.
 *    Calls the appropriate bidder routine, and records the contract
 *    and bid, before calling SubmitBid.
 *
 * -------------------------------------------------------------------
 */
void 
ReceiveRequestForBid(void *arg)
{
    char **argv = (char **)arg;
    OfferBuffer *offbuf = (OfferBuffer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    BidBuffer *bidbuf = (BidBuffer *)malloc(sizeof(BidBuffer));
    SiteAddr requester = offbuf->requester;
    OfferDetails details = offbuf->details;
    int requestIndex = offbuf->requestIndex;
    int whichBid = offbuf->whichBid;
    bool isSubContract = offbuf->isSubContract;
    ContractId contract;
    Budget *bid;
    QepFragmentId qepFragId;
    RPCErrorCode result;
    int status;
    RpcConn *rpcc;
    SiteAddrNode *localAddr;
    
    switch(details.protocol) {
    case LONG_PROCESS_QUERY: {

	ProcessQueryArg *pArg;
	ProcessQueryArg *pArg2;
	List *rtable;
	Plan *plan;
	
	/*
	 * reply immediately so client doesn't time out
	 */
	result = SUCCESS;
	
	if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	    sitemgrDebug(0, "ReceiveRequestForBid: svc_sendreply failed\n");
	}

	/*
	 * Record the information in the offer buffer inside a contract.
	 * If we don't bid, or if the contract is lost, we should make this
	 * slot available again or we'll run out of contracts.  We need to
	 * record the information here because for sub-bidding it has to
	 * be accessible inside SubContract().
	 */

	contract = GetNextContractNo();
	
	sitemgrDebug(1, "ReceiveRequestForBid: Beginning contract %i for bid request %i at site %i.\n",
		     contract, offbuf->requestId, requester.hostid);

	pArg = &offbuf->details.OfferDetails_u.longQuery;
	pArg2 = &contractsMade[contract].details_u.queryDetails;
	rtable = (List *)pArg->rtable;
	plan = (Plan *)pArg->plan;

	/*
	 * Make all processing sites in this plan equal to this
	 * site.  Subcontracting may change this.
	 */
	localAddr = makeNode(SiteAddrNode);
	localAddr->addr = getMyAddr();
	localAddr->hostid = getMyHostId();
	
	SetAllProcSites(plan, localAddr);
	
	contractsMade[contract].requestId = offbuf->requestId;
	contractsMade[contract].requestIndex = offbuf->requestIndex;
	contractsMade[contract].whichChunk = offbuf->whichChunk;
	contractsMade[contract].whichBid = offbuf->whichBid;
	contractsMade[contract].protocol = offbuf->details.protocol;
	contractsMade[contract].requester = requester;
	contractsMade[contract].subBids = FALSE;
	contractsMade[contract].details_u.queryDetails = details.OfferDetails_u.longQuery;
	
	/*
	 * Uses copyObject() inside sitemgr/copyfuncs_sm.c, which I had to modify
	 * slightly from copyObject() inside nodes/copyfuncs.c - see the header
	 * file in copyfuncs_sm.c for details.
	 *						-JPS 2/9/96
	 */
	pArg2->rtable = (PQA_Node_list)copyObject(rtable);
	pArg2->plan = (PQA_Plan_tree)copyObject(plan);

	/*
	 * Get the bid
	 */
 	bid = PlanBid(plan, rtable, contract);
	
	bidbuf->qepFragId = plan->qepFragId;
	bidbuf->bid = *bid;
	bidbuf->subContracted = FALSE;
	bidbuf->subContractedPlan = NULL;
	
	contractsMade[contract].bid.response = bid->response;
	
	if (bid->response == BID) {
	    contractsMade[contract].bid.cost = bid->cost;
	    contractsMade[contract].bid.delay = bid->delay;
	    contractsMade[contract].bid.staleness = bid->staleness;
	    contractsMade[contract].bid.accuracy = bid->accuracy;
	    
	    sitemgrDebug(1, "ReceiveRequestForBid:  Bidding (%f, %f) for:\n", bid->cost, bid->delay);
	    sitemgrDebug(1, "\tcontract %i\n\trequestId %i \n\twhichChunk %i \n\twhichBid %i\n\n",
			 contract, offbuf->requestId, offbuf->whichChunk, offbuf->whichBid);

	    /*
	     * Bug fix - normally, the Bidder doesn't tell the query broker it's
	     * subcontracted anything out.  However, if the bidder and the broker
	     * are at the same site, this causes a problem:  The broker thinks that
	     * the entire plan chunk is being processed locally (because the bidder
	     * agreed to bid).  But, the local backend processes all nodes in the plan
	     * tree that are processed locally and connected to the root node itself.
	     * If the bidder says everything should be done locally, when actually
	     * some things (like scans) have been subcontracted out, the local
	     * backend will try to process things it can't.  So, in this case, we
	     * do the following:
	     *
	     * 	- tell the local backend that we subcontracted by setting
	     *    bidbuf->subContracted to TRUE;
	     *
	     *  - send back the plan tree with all the processing sites.
	     */

	    if (requester.hostid == localAddr->hostid &&
		contractsMade[contract].subBids) {
		bidbuf->subContracted = TRUE;
		bidbuf->subContractedPlan = pArg2->plan;
	    } else {
		bidbuf->subContracted = FALSE;
		bidbuf->subContractedPlan = NULL;
	    }
	    
	} else {
	    sitemgrDebug(1, "ReceiveRequestForBid:  Not bidding on contract %i, whichChunk %i, whichBid %i\n",
			 contract, offbuf->whichChunk, offbuf->whichBid);
	}
	

    }

	break;

    case LONG_SELL_FRAGMENT: {

#if 0
	/*
	 * reply immediately so client doesn't time out
	 */
	result = SUCCESS;
	
	if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	    sitemgrDebug(0, "ReceiveRequestForBid: svc_sendreply failed\n");
	}
    
	contractsMade[contract].details_u.moveDetails = details.OfferDetails_u.longSell;
	
	sitemgrDebug(3, "ReceiveRequestForBid: LONG_SELL_FRAGMENT: Calling GetSellBid for contract %i\n", contract);
	GetSellBid(contractsMade[contract].details_u.moveDetails, contract, bid);

	bidbuf->qepFragId = contractsMade[contract].details_u.moveDetails.qepFragId;
	bidbuf->bid = *bid;
	
	sitemgrDebug(1, "ReceiveRequestForBid (SELLING):  Bidding (%f, %f) for:\n", bid->cost, bid->delay);
	sitemgrDebug(1, "\tcontract %i\n\trequest %i\n\tquery fragment %i \n\tdata fragment %i\n\n",
		     contract, request, bidbuf->qepFragId, contractsMade[contract].details_u.moveDetails.frastoreid);
#endif
	
    }
	break;
	
    case LONG_BUY_FRAGMENT: {
#if 0
	/*
	 * reply immediately so client doesn't time out
	 */
	result = SUCCESS;
	
	if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	    sitemgrDebug(0, "ReceiveRequestForBid: svc_sendreply failed\n");
	}

	contractsMade[contract].details_u.moveDetails = details.OfferDetails_u.longBuy;
	
	sitemgrDebug(3, "ReceiveRequestForBid: LONG_BUY_FRAGMENT: Calling GetBuyBid for contract %i\n", contract);
	GetSellBid(contractsMade[contract].details_u.moveDetails, contract, bid);

	bidbuf->qepFragId = contractsMade[contract].details_u.moveDetails.qepFragId;
	bidbuf->bid = *bid;
	
	sitemgrDebug(1, "ReceiveRequestForBid (BUYING):  Bidding (%f, %f) for:\n", bid->cost, bid->delay);
	sitemgrDebug(1, "\tcontract %i\n\trequest %i\n\tquery fragment %i \n\tdata fragment %i\n\n",
		     contract, request, bidbuf->qepFragId, contractsMade[contract].details_u.moveDetails.frastoreid);
#endif
    }
	break;
	
    default:
	result = BAD_PROTOCOL;
	if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	    sitemgrDebug(0, "ReceiveRequestForBid: svc_sendreply failed\n");
	}
	else {
	    sitemgrDebug(2, 
			 "ReceiveRequestForBid: svc_sendreply succeeded - BAD PROTOCOL, result = %d\n",
			 result);
	}
	pthread_exit(NULL);
    }

    /*
     * Whichever protocol we used (if it was valid),
     * we now have a bid.  Record the outstanding
     * contract, and call BrokerReceiveBid at the site
     * that sent us the request.
     */
    bidbuf->contract = contract;
    bidbuf->requestId = offbuf->requestId;
    bidbuf->requestIndex = offbuf->requestIndex;
    bidbuf->whichChunk = offbuf->whichChunk;
    bidbuf->whichBid = offbuf->whichBid;

    rpcc = Rpc_connectExec(requester.addr, htons(requester.port), "ReceiveRequestForBid", RPC_ONERR_RETURN);
    
    if (rpcc == NULL) {
	sitemgrDebug(0,
		     "ReceiveRequestForBid: can't create client handle (%d,%d,%d)\n",
		     requester.addr,
		     requester.port,
		     requester.hostid
		     );
	pthread_exit(NULL);
    }

    status = Rpc_call(rpcc, RPC_BROKERRECEIVEBID,
		       bidbuf, &result);
    
    if (!status) {
	sitemgrDebug(0, "ReceiveRequestForBid: RPC error when calling BrokerReceiveBid\n");
    }
    
    Rpc_disconnect(rpcc);

}

/*
 *-------------------------------------------------------------------------
 * SetAllProcSites--
 *
 * 	Recursively set all processing sites to the one passed in.
 *-------------------------------------------------------------------------
 */
void
SetAllProcSites(Plan *plan, SiteAddrNode *addr) {
    
    if (plan->type == T_Merge) {
	List *lp, *lp2;
	foreach(lp, ((Merge *)plan)->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp);
	    SetAllProcSites(p, addr);
	}
    } else {
	if (plan->lefttree != NULL) {
	    SetAllProcSites(plan->lefttree, addr);
	}

	if (plan->righttree != NULL) {
	    SetAllProcSites(plan->righttree, addr);
	}
    }
    CopySiteAddrNode(addr, &plan->processingSite);
    
}

void
TellLost(void *arg) {
    char **argv = (char **)arg;
    int logicalid = *(int *)argv[0];
    extern Tcl_Interp *tclInterp;
    char tclSttmt[1024];
    char **tclResult;
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    RPCErrorCode result;
    int tclCond;
    
    sprintf(tclSttmt, "incr losses%i", logicalid);
    tclCond = tclEvaluate(tclSttmt, tclResult);

    free(tclResult);
    
    result = SUCCESS;

    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "TellLost: svc_sendreply failed\n");
    } else {
	sitemgrDebug(2, 
		     "TellLost: svc_sendreply succeeded, result = %d\n",
		     result);
    }

    
}


/*
 * DataBroker--
 *
 */
void
DataBroker(void *arg) {
    char **argv = (char**)arg;
    DataBrokerArg *dbArg = (DataBrokerArg *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    List *rtable = (List *)dbArg->rtable;
    Plan *plan = (Plan *)dbArg->plan;
    extern Tcl_Interp *tclInterp;
    char *tclResult;
    char *rtableStr, *fragStr;
    char planStr[1024];
    char *tclStmt = (char *)malloc(1024 * sizeof(char));
    int nodeNum = 0;
    int tclReturn;
    RPCErrorCode result;
    
    /*
     * reply immediately so client doesn't time out
     */
    result = SUCCESS;
	
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "DataBroker: svc_sendreply failed\n");
    }
    
    /*
     * Make fragments string accessible to data broker Tcl script.
     */
    fragStr = databrokerString(plan, rtable);
    strcpy(tclStmt, "set fragments ");
    strcat(tclStmt, fragStr);
    tclReturn = tclEvaluate(tclStmt, &tclResult);
    
    /*
     * Call data broker and let it do its thing.
     */
    
    strcpy(tclStmt, "DataBroker");

    tclReturn = tclEvaluate(tclStmt, &tclResult);

    /*
     * Data Broker doesn't return anything (other than a condition code).
     */
    
    if (tclReturn != TCL_OK) {
	sitemgrDebug(0, "DataBroker: tclEvaluate returned %i\n", tclReturn);
	sitemgrDebug(0, "Error Msg: %s\n", tclInterp->result);
    }
}


int
sm_newQepFragId()
{
    static int lastQepFragId = 0;
    return ++lastQepFragId;
}

/*
 * LookupClass -
 * forward nameserver request to an available backend
 */
void
LookupClass(void *args)
{
    char **argv = (char **) args;
    char **relname = (char **) argv[0];
    SVCXPRT *transp = (SVCXPRT *) argv[1];
    NsvcLookupClassResult *result;

    result = invokeExec(NSVC_LOOKUPCLASS, relname, -1, "LookupClass");
    if (!svc_sendreply(transp, xdr_NsvcLookupClassResult, result)) {
	sitemgrDebug(0, "LookupClass: svc_sendreply failed\n");
    } else {
	sitemgrDebug(2,
	    "LookupClass: svc_sendreply succeeded, result = %d\n",
	    result);
    }
    free(result);
    free(args);
}

/*
 * LookupFrags -
 * forward nameserver request to an available backend
 */
void
LookupFrags(void *args)
{
    char **argv = (char **) args;
    int *relid = (int *) argv[0];
    SVCXPRT *transp = (SVCXPRT *) argv[1];
    NsvcLookupFragsResult *result;

    result = invokeExec(NSVC_LOOKUPFRAGS, relid, -1, "LookupFrags");
    if (!svc_sendreply(transp, xdr_NsvcLookupFragsResult, result)) {
	sitemgrDebug(0, "LookupFrags: svc_sendreply failed\n");
    } else {
	sitemgrDebug(2,
	    "LookupFrags: svc_sendreply succeeded, result = %d\n",
	    result);
    }
    free(result);
    free(args);
}


int
GetNextContractNo() {

    int contract;
    
    pthread_mutex_lock(&bidderLock);
    contract = nextContractNo++;
    pthread_mutex_unlock(&bidderLock);

    return contract;
    
}
