/*-------------------------------------------------------------------------
 *
 * subcontract.c--
 *
 *    	-- Implementation of Tcl subcontract command - subcontracts out
 *	   a sub-plan.  The sub-plan is passed in, along with the contract
 *	   number, in string form from the Tcl bidder script.  SubContract
 *	   converts it into a Plan and passes it to a Broker to find
 *	   processing sites.  The plan passed back is converted back into
 *	   a string and returned to the Tcl bidder script.
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/sitemgr/subcontract.c,v 1.2 1996/05/31 17:14:29 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <math.h>	/* just for atof */

#include "tcl.h"
#include "postgres.h"
#include "mariposa_rpc.h"
#include "syncFrag.h"
#include "sitemgr/sitemgr.h"
#include "sitemgr/bidder.h"
#include <string.h>

/*
 * Global so it can be passed back to SubBrokerDone
 * from the Broker and accessed in SubContract.
 */
Plan *subContractPlan;		
Budget subContractBid;

/*
 * Pointer to node of original plan that corresponds
 * to root node of subContractPlan
 */
Plan *originalPlan;

extern Contract *contractsMade;

/*
 * Wait in SubContract until woken up
 * by SubBrokerDone.
 */
pthread_mutex_t SubBrokerDoneLock;
pthread_cond_t SubBrokerDoneCond;
bool SubBrokerDoneFlag;

/*
 * Forward declarations of things used only within this file.
 */
void CopyProcessingSites(Plan *from, Plan *to);


/*
 *-------------------------------------------------------------------------
 * SubContract-- Subcontract out a plan tree (which is part of a larger
 *		 plan tree)
 *
 * Input:	String representation of the plan with each node numbered.
 *		(This is the form in which the plan is passed to the Tcl
 *		call GetQueryBid.) and the contract number.
 *
 * Output:	The bid for the plan and the plan, with processing sites
 *		filled in, in string form (passed back to the
 *		Tcl interpreter).
 *
 * NB: This routine is single-threaded.  Since it's called via the Tcl
 * interpreter, and tclEvaluate() (our routine, not TclEval()) is written
 * to serialize Tcl calls, this is okay.  Someday, hopefully, this will
 * not be acceptable and will have to be fixed.  	-JPS 2/6/96
 *-------------------------------------------------------------------------
 */
int
SubContract(ClientData clientData, 
		Tcl_Interp *interp,
		int argc, char *argv[])
{
    char *subPlan = argv[1];
    ContractId contract = atoi(argv[2]);
    List *rtable = (List *)contractsMade[contract].details_u.queryDetails.rtable;
    Plan *plan = (Plan *)contractsMade[contract].details_u.queryDetails.plan;
    int result;
    BrokerArg arg;
    char *rootNodeNumStr;
    int rootNodeNum;
    char planStr[4096];
    char bidStr[1024];
    int n;
    
    if (argc != 3) {
	interp->result = "wrong # args";
	return TCL_ERROR;
    }

    contractsMade[contract].subBids = TRUE;
    
    /*
     * Set up all the stuff we need to make a call to a broker
     * in some backend.
     */

    fprintf(stderr, "Calling StringToPlan\n");
    result = StringToPlan(subPlan, &subContractPlan);
    fprintf(stderr, "Done with StringToPlan\n");
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Get node number of root node of sub-plan passed in.
     */
    deleteLeadingAndTrailingSpaces(&subPlan);
    tclUnlist(&subPlan);
    deleteLeadingAndTrailingSpaces(&subPlan);
    result = lindex(tclList(subPlan), 1, &rootNodeNumStr);
    if (result != TCL_OK) {
	return result;
    }
    rootNodeNum = atoi(rootNodeNumStr);

    /*
     * Find node in original plan that corresponds to
     * root node of subplan passed in.
     */
    n = 0;
    originalPlan = FindPlanNodeNumber(plan, rootNodeNum, &n);
    
    arg.operation = contractsMade[contract].details_u.queryDetails.operation;
    arg.protocol = LONG_PROCESS_QUERY;
    arg.plan = (PQA_Plan_tree)subContractPlan;
    arg.rtable = (PQA_Node_list)rtable;
    arg.parentContract = contract;
    arg.homeSiteAddr.addr = contractsMade[contract].details_u.queryDetails.homeSite.addr;
    arg.homeSiteAddr.port = contractsMade[contract].details_u.queryDetails.homeSite.port;
    arg.homeSiteAddr.hostid = contractsMade[contract].details_u.queryDetails.homeSite.hostid;
    
    /*
     * Now call the backend wrapper procedure that calls the Broker.  When the Broker
     * is finished, the wrapper will pass the plan tree, with the processing sites
     * filled in, back to SubContractReturn, which will fill them into subContractPlan.
     */
    result = (int) invokeExec(SUBCONTRACTBROKER, (void *)&arg, -1, "SubContract");

    /*
     * Go to sleep until the Broker has passed back the processing sites.
     */
    
    /*
     * Initialize broker condition variable
     */
    pthread_mutex_init(&SubBrokerDoneLock, pthread_mutexattr_default);
    pthread_cond_init(&SubBrokerDoneCond, pthread_condattr_default);
    pthread_mutex_lock(&SubBrokerDoneLock);
    SubBrokerDoneFlag = FALSE;
    
    while (!SubBrokerDoneFlag) {
	pthread_cond_wait(&SubBrokerDoneCond, &SubBrokerDoneLock);
    }

    /*
     * Fill in processing sites into original plan.
     */
    CopyProcessingSites(subContractPlan, originalPlan);
    
    bidStr[0] = '\0';
    strcpy(bidStr, bidString(&subContractBid));
    
    /*
     * Fill results into Tcl interpreter.
     */
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, bidStr, (char *)NULL);
    Tcl_AppendResult(interp, " ", planStr, (char *)NULL);
    
    return TCL_OK;
    
}

/*
 *-------------------------------------------------------------------------
 * SubBrokerDone-- Called by the backend doing the subcontract
 * brokering when it's finished.  We place the plan (with the
 * processing sites filled in) in subContractPlan and the bid
 * in subContractBid, then wake up the thread waiting in SubContract().
 *
 *-------------------------------------------------------------------------
 */
void
SubBrokerDone(void *args) {
    char **argv = (char **)args;
    BrokerAnswer *bAnswer = (BrokerAnswer *)argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    RPCErrorCode result;
    
    result = SUCCESS;
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	sitemgrDebug(0, "SubContractDone: svc_sendreply failed\n");
    }
    
    subContractPlan = (Plan *)bAnswer->plan;
    subContractBid = bAnswer->bid;
    SubBrokerDoneFlag = TRUE;
    pthread_cond_signal(&SubBrokerDoneCond);
}

/*
 *-------------------------------------------------------------------------
 * CopySiteAddrNode--
 *
 * Duh.
 *-------------------------------------------------------------------------
 */
void
CopySiteAddrNode(SiteAddrNode *from, SiteAddrNode *to) {
    to->type = T_SiteAddrNode;
    to->port = from->port;
    to->addr = from->addr;
    to->hostid = from->hostid;
}

/*
 *-------------------------------------------------------------------------
 * CopyProcessingSites--
 *
 * 	Recursively copy the processing sites from one plan to another.
 *-------------------------------------------------------------------------
 */
void
CopyProcessingSites(Plan *from, Plan *to) {
    
    if (from->type == T_Merge) {
	List *lp, *lp2;
	lp2 = (((Merge *)to)->mergeplans);
	foreach(lp, ((Merge *)from)->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp);
	    Plan *p2 = (Plan *)lfirst(lp2);
	    CopyProcessingSites(p, p2);
	    lp2 = lp2->next;
	}
    } else {
	if (from->lefttree != NULL) {
	    CopyProcessingSites(from->lefttree, to->lefttree);
	}

	if (from->righttree != NULL) {
	    CopyProcessingSites(from->righttree, to->righttree);
	}
    }
    CopySiteAddrNode(&from->processingSite, &to->processingSite);
    
}
