/*
 *  hp.h -- Private header file for the Hp jukebox storage manager.
 *
 *	This file is only used if you're at Berkeley, and HP_JUKEBOX
 *	is defined.
 *
 *	$Header: /usr/local/devel/mariposa/cvs/src/backend/storage/hp.h,v 1.2 1995/09/05 09:33:22 aoki Exp $
 */

#ifdef HP_JUKEBOX

/*
 *  When the buffer pool requests a particular page, we load a group of
 *  pages from the jukebox into the mag disk cache for efficiency.
 *  HPCACHESIZE is the number of these groups in the disk cache.  Every
 *  group is represented by one entry in the shared memory cache.  HPGRPSIZE
 *  is the number of 8k pages in a group.
 */

#define	HPCACHESIZE	64		/* # groups in mag disk cache */
#define	HPGRPSIZE	16		/* # 8k pages in a group */
#define HPNBLKSIZE	20		/* # sizes cached in nblock cache */

/* misc constants */
#define	HPCACHENAME	"_hp_cache_"	/* relative to $POSTGRESHOME/data */
#define	HPMETANAME	"_hp_meta_"	/* relative to $POSTGRESHOME/data */
#define	HPBLOCKNAME	"_hp_nblocks_"	/* relative to $POSTGRESHOME/data */
#define	HPSIZENAME	"_hp_size_"	/* relative to $POSTGRESHOME/data */

/* for allocations... */
#define HPNEWRELN	0
#define HPOLDRELN	1

/*
 *  HPGroupDesc -- Descriptor block for a cache group.
 *
 *	The first 1024 bytes in a group -- on a platter or in the magnetic
 *	disk cache -- are a descriptor block.  We choose 1024 bytes because
 *	this is the native block size of the jukebox.
 *
 *	This block includes a description of the data that appears in the
 *	group, including relid, dbid, relname, dbname, and a unique OID
 *	that we use to verify cache consistency on startup.  HPGroupDesc
 *	is the structure that contains this information.  It resides at the
 *	start of the 1024-byte block; the rest of the block is unused.
 */

typedef struct HPGroupDesc {
    long	hpgd_magic;
    long	hpgd_version;
    NameData	hpgd_dbname;
    NameData	hpgd_relname;
    Oid	hpgd_dbid;
    Oid	hpgd_relid;
    Oid	hpgd_plid;
    long	hpgd_relblkno;
    long	hpgd_jboffset;
    long	hpgd_extentsz;
    Oid	hpgd_groupoid;
} HPGroupDesc;

#define HPGDMAGIC	0x060362
#define	HPGDVERSION	2
#define JBBLOCKSZ	1024

/* size of HPCacheBuf */
#define HPBUFSIZE	((MAXBLCKSZ * HPGRPSIZE) + JBBLOCKSZ)

/* # of jb blocks in extent */
#define HPEXTENTSZ	(HPBUFSIZE / JBBLOCKSZ)

/*
 *  HPCacheTag -- Unique identifier for individual groups in the magnetic
 *		  disk cache.
 *
 *	We use this identifier to query the shared memory cache metadata
 *	when we want to find a particular group.  
 */

typedef struct HPCacheTag {
    Oid		hpct_dbid;	/* database OID of this group */
    Oid		hpct_relid;	/* relation OID of this group */
    BlockNumber		hpct_base;	/* number of first block in group */
} HPCacheTag;

/*
 *  HPHashEntry -- The hash table code returns a pointer to a structure
 *		   that has this layout.
 */

typedef struct HPHashEntry {
    HPCacheTag		hphe_tag;	/* cache tag -- hash key */
    int			hphe_groupno;	/* which group this is in cache file */
} HPHashEntry;

/*
 *  HPCacheHeader -- Header data for in-memory metadata cache.
 */

typedef struct HPCacheHeader {
    int			hph_nentries;
    int			hph_freehead;
    int			hph_freetail;
    uint32		hph_flags;

#define HPH_INITING	(1 << 0)
#define HPH_INITED	(1 << 1)

#ifdef HAS_TEST_AND_SET

    slock_t		hph_initlock;	/* initialization in progress lock */

#endif /* HAS_TEST_AND_SET */

} HPCacheHeader;

/*
 *  HPCacheItem -- Cache item describing blocks on the magnetic disk cache.
 *
 *	An array of these is maintained in shared memory, with one entry
 *	for every group that appears in the magnetic disk block cache.  We
 *	maintain a consistent copy of this array on magnetic disk whenever
 *	we change the cache contents.  This is because the magnetic disk
 *	cache is persistent, and contains data that logically appears on the
 *	jukebox between backend instances.
 *
 *	The OID that appears in this structure is used to detect corruption
 *	of the cache due to crashes during cache metadata update on disk.
 *	When we detect corruption, we recover by marking the group free.  We
 *	are very careful to do this in a way that guarantees no data is lost,
 *	and that does not require log processing.
 *
 *	We keep a free list of groups to which no references exist.  We
 *	allocate groups off this list on demand.  In general, references
 *	to groups in the cache are very short-lived; we never return pointers
 *	into private structures outside of the code that manages the cache.
 *	The free list is maintained in LRU order, and the least-recently-
 *	used group is allocated first.
 *
 *	Groups on the jukebox include one page (the first) that describes the
 *	group, including its dbid, relid, dbname, relname, and extent size.
 *	This page also includes the OID described above.
 */

typedef struct HPCacheItem {
    HPCacheTag		hpc_tag;		/* dbid, relid, group triple */
    int			hpc_freeprev;		/* free list pointer */
    int			hpc_freenext;		/* free list pointer */
    int			hpc_refcount;		/* number of active refs */
    Oid		hpc_oid;		/* OID of group */
    Oid		hpc_plid;		/* platter OID for group */
    NameData		hpc_plname;		/* platter name for group */
    int			hpc_jboffset;		/* offset of first block */

    uint8		hpc_gflags;		/* flags for entire group */

    uint8		hpc_flags[HPGRPSIZE];	/* flag bytes, 1 per block */

#define HPC_CLEAR	(uint8) 0x0
#define HPC_MISSING	(1 << 0)
#define HPC_ONPLATTER	(1 << 1)
#define	HPC_IOINPROG	(1 << 7)

#ifdef HAS_TEST_AND_SET

    slock_t		hpc_iolock;		/* transfer in progress */

#endif /* HAS_TEST_AND_SET */

} HPCacheItem;

#define MUST_FLUSH(f)	(!(((f) & HPC_ONPLATTER) || ((f) & HPC_MISSING)))

extern BlockNumber	pgjb_offset(char *plname, Oid plid, int extentsz);
extern bool		pgjb_freespc(char *plname, Oid plid, int alloctype);
extern int		pgjb_init();
extern int		pgjb_wrtextent(HPCacheItem *item, int relblocks, char *buf);
extern int		pgjb_rdextent(HPCacheItem *item, char *buf);
extern int		JBShmemSize();

extern BlockNumber	hpmaxseg();

#endif /* HP_JUKEBOX */









