/*
 *  fr.c -- fragmented storage manager
 *          This routine enables a parent relation that consists of a set of
 *          fragments to act like a real relation. It provides a block oriented
 *          interface to a fragmented relation.
 *  author -- Sunita Sarawagi
 *  date   -- 15 April 1995.
 * [missing RCS id]
 */
#include "c.h"
#include "postgres.h"

#ifdef FRAGMENT_SMGR

#include <sys/file.h>
#include <math.h>
#include "machine.h"

#include "miscadmin.h"

#include "storage/ipc.h"
#include "storage/smgr.h"
#include "storage/shmem.h"
#include "storage/spin.h"

#include "utils/hsearch.h"
#include "utils/rel.h"
#include "utils/elog.h"
#include "utils/memutils.h"
#include "utils/relcache.h"

#include "access/htup.h"
#include "access/relscan.h"
#include "access/heapam.h"

#include "catalog/pg_proc.h"
#include "nodes/plannodes.h"
#include "nodes/execnodes.h"
#include "parser/parsetree.h"
#include "executor/execdesc.h"
#include "catalog/pg_fragment.h"

#include "tcop/frag.h"

static Relation GetFragForBlock(Relation reln, BlockNumber blocknum, 
				BlockNumber *newblock);

/*
 *  frcreate() -- Create the requested relation
 *
 */
int
frcreate(Relation reln)
{
    File vfd;
    char *path;

    if (reln->rd_rel->relisshared) {
	path = (char *) palloc(strlen(DataDir) + NAMEDATALEN + 2);
	sprintf(path, "%s/%.*s", DataDir, NAMEDATALEN,
		&(reln->rd_rel->relname.data[0]));
    } else {
	path = (char *) palloc(NAMEDATALEN + 1);
	sprintf(path, "%.*s", NAMEDATALEN, &(reln->rd_rel->relname.data[0]));
    }

    vfd = FileNameOpenFile(path, O_CREAT|O_RDWR|O_EXCL, 0600);

    pfree(path);

    return (vfd);
}

/*
 *  frextend() -- extend a relation by one blocks -- this requires 
 * finding the fragment of the relation and extending it.  
 */

int
frextend(Relation reln, char *buffer)
{
    Relation fragRel;
    
    /* get the last fragment of the relation */
    int fragNum = length(reln->rd_fragrels) - 1;
    
    if (fragNum >= 0) {
	fragRel = nth(fragNum, reln->rd_fragrels);
	return smgrextend(RelationGetSmgr(fragRel), fragRel, buffer);
    }
    return SM_FAIL;
}

int
frunlink(Relation reln)
{
    List * frag;

    foreach (frag, reln->rd_fragrels) {
	Relation fragRel = lfirst(frag);
	
	if (smgrunlink(RelationGetSmgr(fragRel), fragRel) != SM_SUCCESS) {
	    return SM_FAIL;
	}
    }
    return (SM_SUCCESS);
}

int
fropen(Relation reln)
{
    List * frag;
    /* call smgropen on all the fragments */
    List * fragList = getDataFragRelIDs(reln->rd_id);
    reln->rd_fragrels = NIL;
    
    foreach (frag, fragList) {
	FragLoc *floc = lfirst(frag);
	Oid fragId = floc->fragid;
	
	Relation fragRel = RelationIdGetRelation(fragId);
	
	if (! RelationIsValid(fragRel)) {
	    elog(WARN, "fropen: failed on a fragment of %s\n", 
		 RelationGetRelationName(reln));
	}
	reln->rd_fragrels = lcons(fragRel, reln->rd_fragrels);
    }
    return (1);
}

int
frclose(Relation reln)
{
    List * frag;
    foreach (frag, reln->rd_fragrels) {
	Relation fragRel = lfirst(frag);
	
	smgrclose(RelationGetSmgr(fragRel), fragRel);
    }
    return (SM_SUCCESS);
}

int
frread(Relation reln, BlockNumber blocknum, char *buffer)
{
    BlockNumber newblocknum;

    Relation fragRel = GetFragForBlock(reln, blocknum, &newblocknum);
    if (fragRel == NULL) {
	return SM_FAIL;
    }
    return smgrread(RelationGetSmgr(fragRel), fragRel, newblocknum, buffer); 
}

int
frwrite(Relation reln, BlockNumber blocknum, char *buffer)
{
    BlockNumber newblocknum;

    Relation fragRel = GetFragForBlock(reln, blocknum, &newblocknum);
    if (fragRel == NULL) {
	return SM_FAIL;
    }
    return smgrwrite(RelationGetSmgr(fragRel), fragRel, newblocknum, buffer); 
}

int
frflush(Relation reln, BlockNumber blocknum, char *buffer)
{
    BlockNumber newblocknum;

    Relation fragRel = GetFragForBlock(reln, blocknum, &newblocknum);
    if (fragRel == NULL) {
	return SM_FAIL;
    }
    return smgrwrite(RelationGetSmgr(fragRel), fragRel, newblocknum, buffer); 
}

int
frblindwrt(char *dbstr, char *relstr, Oid dbid, Oid relid, BlockNumber 
	   blocknum, Size blocksz, char * buffer)
{
    BlockNumber newblocknum;

    Relation reln = RelationIdGetRelation(relid);

    Relation fragRel = GetFragForBlock(reln, blocknum,  &newblocknum);
    if (fragRel == NULL) {
	return SM_FAIL;
    }
    return smgrblindwrt(RelationGetSmgr(fragRel), dbstr, 
			RelationGetRelationName(fragRel), dbid, relid, 
			newblocknum, blocksz, buffer); 
}

/*
 *  frnblocks() --Return the number of blocks that appear in this relation.
 */

int
frnblocks(Relation reln)
{
    /* sum up the number of blocks of each fragment */
    List * frag;
    int total = 0;

    foreach (frag, reln->rd_fragrels) {
	Relation fragRel = lfirst(frag);
	
	total += smgrnblocks(RelationGetSmgr(fragRel), fragRel);
    }
    return (total);
}

/* -------------------------------------------------
 * find the fragment that contains block number blocknum of
 * the parent relation.
 */
static Relation
GetFragForBlock(Relation reln, BlockNumber blocknum, BlockNumber *newblock)
{
    List * frag;
    foreach (frag, reln->rd_fragrels) {
	Relation fragRel = lfirst(frag);

	BlockNumber blockInFrag = smgrnblocks(RelationGetSmgr(fragRel), 
					       fragRel);
	if (blockInFrag - 1 >= blocknum) {
	    /* assume blocks are numbered from 0 */
	    *newblock = blocknum;
	    return fragRel;
	} else {
	    blocknum -= blockInFrag;
	}
    }
    return NULL;
}
#endif /* FRAGMENT_SMGR */









