/*
 *  sj.c -- sony jukebox storage manager.
 *
 *	This code manages relations that reside on the sony write-once
 *	optical disk jukebox.
 * [missing RCS id]
 */
#include "c.h"
#include "postgres.h"

#ifdef HP_JUKEBOX

#include <sys/file.h>
#include <math.h>
#include "machine.h"

#include "miscadmin.h"

#include "storage/ipc.h"
#include "storage/smgr.h"
#include "storage/shmem.h"
#include "storage/spin.h"
#include "storage/hp.h"

#include "utils/hsearch.h"
#include "utils/rel.h"
#include "utils/elog.h"
#include "utils/memutils.h"
#include "utils/relcache.h"

#include "access/htup.h"
#include "access/relscan.h"
#include "access/heapam.h"

#include "catalog/pg_platter.h"
#include "catalog/pg_plmap.h"
#include "catalog/pg_plrel.h"
#include "catalog/pg_proc.h"


/* globals used in this file */
static File		HPCacheVfd;	/* vfd for cache data file */
static File		HPMetaVfd;	/* vfd for cache metadata file */
static File		HPBlockVfd;	/* vfd for nblocks file */
static HPCacheHeader	*HPHeader;	/* pointer to cache header in shmem */
static HTAB		*HPCacheHT;	/* pointer to hash table in shmem */
static HPCacheItem	*HPCache;	/* pointer to cache metadata in shmem */
static HPCacheTag	*HPNBlockCache;	/* pointer to nblock cache */
static int              HPCacheSize = HPCACHESIZE;
static long             *HPNEntries;

#ifndef	HAS_TEST_AND_SET

/*
 *  If we don't have test-and-set locks, then we need a semaphore for
 *  concurrency control.  This semaphore is in addition to the metadata
 *  lock, HPCacheLock, that we acquire before touching the cache metadata.
 *
 *  This semaphore is used in two ways.  During cache initialization, we
 *  use it to lock out all other backends that want cache access.  During
 *  normal processing, we control access to groups on which IO is in
 *  progress by holding this lock.  When we're done with initialization or
 *  IO, we do enough V's on the semaphore to satisfy all outstanding P's.
 */

static IpcSemaphoreId	HPWaitSemId;	/* wait semaphore */
static long		*HPNWaiting;	/* # procs sleeping on the wait sem */

#endif /* ndef HAS_TEST_AND_SET */

/* static buffer is for data transfer */
static char	HPCacheBuf[HPBUFSIZE];

/*
 *  When we have to do IO on a group, we avoid holding an exclusive lock on
 *  the cache metadata for the duration of the operation.  We do this by
 *  setting a finer-granularity lock on the group itself.  How we do this
 *  depends on whether we have test-and-set locks or not.  If so, it's
 *  easy; we set the TASlock on the item itself.  Otherwise, we use the
 *  'wait' semaphore described above.
 */

#ifdef HAS_TEST_AND_SET
#define SET_IO_LOCK(item) \
    item->hpc_gflags |= HPC_IOINPROG; \
    SpinRelease(HPCacheLock); \
    S_LOCK(&(item->hpc_iolock));
#else /* HAS_TEST_AND_SET */
#define SET_IO_LOCK(item) \
    item->hpc_gflags |= HPC_IOINPROG; \
    (*HPNWaiting)++; \
    SpinRelease(HPCacheLock); \
    IpcSemaphoreLock(HPWaitSemId, 0, 1);
#endif /* HAS_TEST_AND_SET */

#define GROUPNO(item)	(((char *) item) - ((char *) &(HPCache[0])))/sizeof(HPCacheItem)

/* routines declared in this file */
static void		_hpcacheinit();
static void		_hpwait_init();
static void		_hpunwait_init();
static void		_hpwait_io();
static void		_hpunwait_io();
static void		_hptouch();
static void		_hpunpin();
static void		_hpregister();
static void		_hpregnblocks();
static void		_hpnewextent();
static void		_hprdextent();
static void		_hpdirtylast();
static int		_hpfindnblocks();
static int		_hpwritegrp();
static int		_hpreadgrp();
static int		_hpgroupvrfy();
static int		_hpdowrite();
static Form_pg_plmap	_hpchoose(), _hpMychoose();
static HPCacheItem	*_hpallocgrp();
static HPCacheItem	*_hpfetchgrp();
static HPHashEntry	*_hphashop();
static int		_hpgetgrp();
static void		_hpdump();

/*
 *  hpinit() -- initialize the Hp jukebox storage manager.
 *
 *	We need to find (or establish) the mag-disk buffer cache metadata
 *	in shared memory and open the cache on mag disk.  The first backend
 *	to run that touches the cache initializes it.  All other backends
 *	running simultaneously will only wait for this initialization to
 *	complete if they need to get data out of the cache.  Otherwise,
 *	they'll return successfully immediately after attaching the cache
 *	memory, and will let their older sibling do all the work.
 */

int
hpinit()
{
    unsigned int metasize;
    bool metafound, testfound;
    HASHCTL info;
    bool initcache;
    char *cacheblk, *cachesave;
    int status;
    char *path;

    /*
     *  First attach the shared memory block that contains the disk
     *  cache metadata.  At the end of this block in shared memory is
     *  the hash table we use to do fast lookup on groups in the cache.
     */

    SpinAcquire(HPCacheLock);

#ifdef HAS_TEST_AND_SET
    metasize =
	MAXALIGN(HPCacheSize * sizeof(HPCacheItem)) +
	    MAXALIGN(sizeof(HPCacheHeader)) 
+
		MAXALIGN(HPNBLKSIZE * sizeof(HPCacheTag));
#else /* HAS_TEST_AND_SET */
    metasize =
	MAXALIGN(HPCacheSize * sizeof(HPCacheItem)) +
	    MAXALIGN(sizeof(HPCacheHeader)) +
		MAXALIGN(HPNBLKSIZE * sizeof(HPCacheTag)) +
		    MAXALIGN(sizeof(*HPNWaiting));
#endif /* HAS_TEST_AND_SET */
    cachesave = cacheblk = (char *) ShmemInitStruct("Jukebox cache metadata",
						    metasize, &metafound);
    if (cacheblk == (char *) NULL) {
	SpinRelease(HPCacheLock);
	return (SM_FAIL);
    }

    /* ---------------test code ----------------------------*/    
    HPNEntries = ShmemInitStruct("Jukebox test metadata",
				 sizeof(*HPNEntries), &testfound);
    if (HPNEntries == (long *) NULL) {
	SpinRelease(HPCacheLock);
	return (SM_FAIL);
    }

    /* init it if we need to */
    if (!testfound)
	*HPNEntries = 0;
    /*-------------------test code -------------------------*/

    /*
     *  Order of items in shared memory is metadata header, number of
     *  processes sleeping on the wait semaphore (if no test-and-set locks),
     *  nblock cache, and jukebox cache entries.
     */

    HPHeader = (HPCacheHeader *) cacheblk;
    cacheblk += sizeof(HPCacheHeader);

#ifndef HAS_TEST_AND_SET
    HPNWaiting = (long *) cacheblk;
    cacheblk += sizeof(long);
#endif /* ndef HAS_TEST_AND_SET */

    HPNBlockCache = (HPCacheTag *) cacheblk;
    cacheblk += HPNBLKSIZE * sizeof(HPCacheTag);

    HPCache = (HPCacheItem *) cacheblk;

    /*
     *  Now initialize the pointer to the shared memory hash table.
     */

    info.keysize = sizeof(HPCacheTag);
    info.datasize = sizeof(int);
    info.hash = tag_hash;

    HPCacheHT = ShmemInitHash("Jukebox cache hash table",
			      HPCacheSize, HPCacheSize,
			      &info, (HASH_ELEM|HASH_FUNCTION));

    if (HPCacheHT == (HTAB *) NULL) {
	SpinRelease(HPCacheLock);
	return (SM_FAIL);
    }

    /*
     *  Okay, all our shared memory pointers are set up.  If we did not
     *  find the cache metadata entries in shared memory, or if the cache
     *  has not been initialized from disk, initialize it in this backend.
     */

    if (!metafound || !(HPHeader->hph_flags & (HPH_INITING|HPH_INITED))) {
	initcache = true;
	bzero((char *) cachesave, metasize);
	HPHeader->hph_flags = HPH_INITING;
#ifdef HAS_TEST_AND_SET
	S_INIT_LOCK(&(HPHeader->hph_initlock));
	S_LOCK(&(HPHeader->hph_initlock));
#else /* HAS_TEST_AND_SET */
	IpcSemaphoreLock(HPWaitSemId, 0, 1);
	*HPNWaiting = 1;
#endif /* HAS_TEST_AND_SET */
    } else {
	initcache = false;
    }

    /* don't need exclusive access anymore */
    SpinRelease(HPCacheLock);

    path = (char *) palloc(strlen(DataDir) + strlen(HPCACHENAME) + 2);
    sprintf(path, "%s/%s", DataDir, HPCACHENAME);

    HPCacheVfd = PathNameOpenFile(path, O_RDWR|O_CREAT|O_EXCL, 0600);
    if (HPCacheVfd < 0) {
	HPCacheVfd = PathNameOpenFile(path, O_RDWR, 0600);
	if (HPCacheVfd < 0) {

	    /* if we were initializing the metadata, note our surrender */
	    if (!metafound) {
		HPHeader->hph_flags &= ~HPH_INITING;
		_hpunwait_init();
	    }

	    return (SM_FAIL);
	}
    }

    pfree(path);

    path = (char *) palloc(strlen(DataDir) + strlen(HPMETANAME) + 2);
    sprintf(path, "%s/%s", DataDir, HPMETANAME);

    HPMetaVfd = PathNameOpenFile(path, O_RDWR|O_CREAT|O_EXCL, 0600);
    if (HPMetaVfd < 0) {
	HPMetaVfd = PathNameOpenFile(path, O_RDWR, 0600);
	if (HPMetaVfd < 0) {

	    /* if we were initializing the metadata, note our surrender */
	    if (!metafound) {
		HPHeader->hph_flags &= ~HPH_INITING;
		_hpunwait_init();
	    }

	    return (SM_FAIL);
	}
    }

    pfree(path);

    path = (char *) palloc(strlen(DataDir) + strlen(HPBLOCKNAME) + 2);
    sprintf(path, "%s/%s", DataDir, HPBLOCKNAME);

    HPBlockVfd = PathNameOpenFile(path, O_RDWR|O_CREAT|O_EXCL, 0600);
    if (HPBlockVfd < 0) {
	HPBlockVfd = PathNameOpenFile(path, O_RDWR, 0600);
	if (HPBlockVfd < 0) {

	    /* if we were initializing the metadata, note our surrender */
	    if (!metafound) {
		HPHeader->hph_flags &= ~HPH_INITING;
		_hpunwait_init();
	    }

	    return (SM_FAIL);
	}
    }

    pfree(path);

    /*
     *  If it's our responsibility to initialize the shared-memory cache
     *  metadata, then go do that.  Hpcacheinit() will elog(FATAL, ...) if
     *  it can't initialize the cache, so we don't need to worry about a
     *  return value here.
     */

    if (initcache) {
	_hpcacheinit();
    }

    /*
     *  Finally, we need to initialize the data structures we use for
     *  communicating with the jukebox.
     */

    if (pgjb_init() == SM_FAIL)
	return (SM_FAIL);

    return (SM_SUCCESS);
}

static void
_hpcacheinit()
{
    int nbytes, nread;
    int nentries;
    int nblocks;
    int start, stop;
    int i;
    HPCacheItem *cur;
    HPHashEntry *result;
    bool found;

    /* sanity check */
    if ((HPHeader->hph_flags & HPH_INITED)
	|| !(HPHeader->hph_flags & HPH_INITING)) {
	elog(FATAL, "hp cache header metadata corrupted.");
    }

    /* suck in the metadata */
    if (FileSeek(HPMetaVfd, 0L, L_SET) != 0)
	elog(FATAL, "cannot seek on jukebox metadata cache");

    nbytes = HPCacheSize * sizeof(HPCacheItem);
    bzero((char *) HPCache, nbytes);
    nread = FileRead(HPMetaVfd, (char *) HPCache, nbytes);

    /* be sure we got an integral number of entries */
    nentries = nread / sizeof(HPCacheItem);
    if ((nentries * sizeof(HPCacheItem)) != nread) {
	HPHeader->hph_flags &= ~HPH_INITING;
	_hpunwait_init();
	elog(FATAL, "hp cache metadata file corrupted.");
    }

    /*
     *  Clear out the nblock cache
     */
    bzero((char *) HPNBlockCache, HPNBLKSIZE * sizeof(HPCacheTag));

    /* add every group that appears in the cache to the hash table */
    for (i = 0; i < nentries; i++) {
	cur = &(HPCache[i]);
	result = _hphashop(&(cur->hpc_tag), HASH_ENTER, &found);

	/* store the group number for this key in the hash table */
	result->hphe_groupno = i;

	/* no io in progress */
	cur->hpc_gflags &= ~HPC_IOINPROG;
	cur->hpc_refcount = 0;

    }

    /*
     *  Now construct the LRU list (free list).  Extents will be nominated
     *  for reuse in this order.  Since we have no usage information, we
     *  adopt the following policy:  any extents not yet allocated in the
     *  cache are come first in the list, in order.  These are followed by
     *  the allocated extents, in order.  The free list head is the first
     *  unallocated extent, and its tail is the last allocated one.  This
     *  list is doubly-linked and is not circular.
     *
     *  The bizarre loop, and the way that start and stop are set, are
     *  to allow me to do this in a single loop.  Trust me, this is simpler
     *  than the six special cases I used to have here.
     */

    start = nentries % HPCacheSize;
    stop = ((nentries - 1) + HPCacheSize) % HPCacheSize;

    i = start;
    for (;;) {
	cur = &HPCache[i];
	cur->hpc_freeprev = ((i - 1) + HPCacheSize) % HPCacheSize;
	cur->hpc_freenext = ((i + 1) + HPCacheSize) % HPCacheSize;
#ifdef HAS_TEST_AND_SET
	S_INIT_LOCK(&(cur->hpc_iolock));
#endif
	i = (i + 1) % HPCacheSize;
	if (i == start)
	    break;
    }

    /* fix up the list's endpoints */
    HPCache[start].hpc_freeprev = -1;
    HPCache[stop].hpc_freenext = -1;

    /* set up cache metadata header struct */
    HPHeader->hph_freehead = start;
    HPHeader->hph_freetail = stop;
    HPHeader->hph_nentries = 0;
    HPHeader->hph_flags = HPH_INITED;
}

/*
 *  _hpunwait_init() -- Release initialization lock on the jukebox cache.
 *
 *	When we initialize the cache, we don't keep the cache semaphore
 *	locked.  Instead, we set a flag in the metadata to let other
 *	backends know that we're doing the initialization.  This lets
 *	others start running queries immediately, even if the cache is
 *	not yet populated.  If they want to look something up in the
 *	cache, they'll block on the flag we set, and wait for us to finish.
 *	If they don't need the jukebox, they can run unimpeded.  When we
 *	finish, we call _hpunwait_init() to release the initialization lock
 *	that we hold during initialization.
 *
 *	When we do this, either the cache is properly initialized, or
 *	we detected some error we couldn't deal with.  In either case,
 *	we no longer need exclusive access to the cache metadata.
 */

static void
_hpunwait_init()
{
#ifdef HAS_TEST_AND_SET

    S_UNLOCK(&(HPHeader->hph_initlock));

#else /* HAS_TEST_AND_SET */

    /* atomically V the semaphore once for every waiting process */
    SpinAcquire(HPCacheLock);
    IpcSemaphoreUnlock(HPWaitSemId, 0, *HPNWaiting);
    *HPNWaiting = 0;
    SpinRelease(HPCacheLock);

#endif /* HAS_TEST_AND_SET */
}

/*
 *  _hpunwait_io() -- Release IO lock on the jukebox cache.
 *
 *	While we're doing IO on a particular group in the cache, any other
 *	process that wants to touch that group needs to wait until we're
 *	finished.  If we have TASlocks, then a wait lock appears on the
 *	group entry in the cache metadata.  Otherwise, we use the wait
 *	semaphore in the same way as for initialization, above.
 */

static void
_hpunwait_io(HPCacheItem *item)
{
    item->hpc_gflags &= ~HPC_IOINPROG;

#ifdef HAS_TEST_AND_SET
    S_UNLOCK(&(item->hpc_iolock));
#else /* HAS_TEST_AND_SET */

    /* atomically V the wait semaphore once for each sleeping process */
    SpinAcquire(HPCacheLock);

    if (*HPNWaiting > 0) {
	IpcSemaphoreUnlock(HPWaitSemId, 0, *HPNWaiting);
	*HPNWaiting = 0;
    }

    SpinRelease(HPCacheLock);
#endif /* HAS_TEST_AND_SET */
}

/*
 *  _hpwait_init() -- Wait for cache initialization to complete.
 *
 *	This routine is called when we want to access jukebox cache metadata,
 *	but someone else is initializing it.  When we return, the init lock
 *	has been released and we can retry our access.  On entry, we must be
 *	holding the cache metadata lock.
 */

static void
_hpwait_init()
{
#ifdef HAS_TEST_AND_SET
    SpinRelease(HPCacheLock);
    S_LOCK(&(HPHeader->hph_initlock));
    S_UNLOCK(&(HPHeader->hph_initlock));
#else /* HAS_TEST_AND_SET */
    (*HPNWaiting)++;
    SpinRelease(HPCacheLock);
    IpcSemaphoreLock(HPWaitSemId, 0, 1);
#endif /* HAS_TEST_AND_SET */
}

/*
 *  _hpwait_io() -- Wait for group IO to complete.
 *
 *	This routine is called when we discover that some other process is
 *	doing IO on a group in the cache that we want to use.  We need to
 *	wait for that IO to complete before we can use the group.  On entry,
 *	we must hold the cache metadata lock.  On return, we don't hold that
 *	lock, and the IO completed.  We can retry our access.
 */

static void
_hpwait_io(HPCacheItem *item)
{
#ifdef HAS_TEST_AND_SET
    SpinRelease(HPCacheLock);
    S_LOCK(&(item->hpc_iolock));
    S_UNLOCK(&(item->hpc_iolock));
#else /* HAS_TEST_AND_SET */
    (*HPNWaiting)++;
    SpinRelease(HPCacheLock);
    IpcSemaphoreLock(HPWaitSemId, 0, 1);
#endif /* HAS_TEST_AND_SET */
}

/*
 *  hpshutdown() -- shut down the jukebox storage manager.
 *
 *	We want to close the cache and metadata files, release all our open
 *	jukebox connections, and let the  caller know we're done.
 *
 */

int
hpshutdown()
{
    FileClose(HPCacheVfd);
    FileClose(HPMetaVfd);
    jb_exit(0);    
    return (SM_SUCCESS);
}

/*
 *  hpcreate() -- Create the requested relation on the jukebox.
 *
 *	Creating a new relation requires us to make a new cache group,
 *	fill in the descriptor page, make sure everything is on disk,
 *	and create the new relation file to store the last page of data
 *	on magnetic disk.
 */

int
hpcreate(Relation reln)
{
    HPCacheItem *item;
    HPGroupDesc *group;
    HPCacheTag tag;
    Oid dbid;
    Oid relid;
    File vfd;
    int grpno;
    int i;
    char *path;

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    /*
     *  If the cache is in the process of being initialized, then we need
     *  to wait for initialization to complete.  If the cache is not yet
     *  initialized, and no one else is doing it, then we need to initialize
     *  it ourselves.  Hpwait_init() or hp_init() will release the cache
     *  lock for us.
     */

    SpinAcquire(HPCacheLock);
    if (!(HPHeader->hph_flags & HPH_INITED)) {
	if (HPHeader->hph_flags & HPH_INITING) {
	    _hpwait_init();
	} else {
	    hpinit();
	}
	return (hpcreate(reln));
    }
    SpinRelease(HPCacheLock);

    /*
     *  By here, cache is initialized.  We are aggressively lazy, and
     *  will not allocate an initial extent for this relation until it's
     *  actually used.  We just register an initial block count of zero.
     */

    if (reln->rd_rel->relisshared)
	tag.hpct_dbid = (Oid) 0;
    else
	tag.hpct_dbid = MyDatabaseId;

    tag.hpct_relid = reln->rd_id;
    tag.hpct_base = (BlockNumber) 0;

    SpinAcquire(HPCacheLock);
    _hpregnblocks(&tag);
    SpinRelease(HPCacheLock);
    
    /* last thing to do is to create the mag-disk file to hold last page */
    if (reln->rd_rel->relisshared) {
	path = (char *) palloc(strlen(DataDir) + NAMEDATALEN + 2);
	sprintf(path, "%s/%.*s", DataDir, NAMEDATALEN,
		&(reln->rd_rel->relname.data[0]));
    } else {
	path = (char *) palloc(NAMEDATALEN + 1);
	sprintf(path, "%.*s", NAMEDATALEN, &(reln->rd_rel->relname.data[0]));
    }

    vfd = FileNameOpenFile(path, O_CREAT|O_RDWR|O_EXCL, 0600);

    pfree(path);

    return (vfd);
}

/*
 *  _hpregister() -- Make catalog entry for a new extent
 *
 *	When we create a new jukebox relation, or when we add a new extent
 *	to an existing relation, we need to make the appropriate entry in
 *	pg_plmap().  This routine does that.
 *
 *	On entry, we have item pinned; on exit, it's still pinned, and the
 *	system catalogs have been updated to reflect the presence of the
 *	new extent.
 */

static void
_hpregister(HPCacheItem *item, HPGroupDesc *group)
{
    Relation plmap;
    Oid plid;
    Form_pg_plmap plmdata;
    HeapTuple plmtup;

    BlockNumber currBlock;
    BlockNumber platfirst;
    Relation plat;
    long blkno;
    TupleDesc platdesc;
    HeapScanDesc platscan;
    HeapTuple plattup;
    Datum d;
    bool n;
    ScanKeyData skey;
    Datum value[10];
    char nulls[10];
    char repl[10];
    HeapTuple newtup;
    Buffer buf;

    /*
     *  Choose a platter to put the new extent on.  This returns a filled-in
     *  pg_plmap tuple data part to insert into the system catalogs.  The
     *  choose routine also figures out where to place the extent on the
     *  platter.
     *
     *	Hpchoose() palloc's and fills in plmdata; we free it later in this
     *  routine.
     */

    plmdata = _hpMychoose(item, &(group->hpgd_relname));

    /* record plid, offset, extent size for caller */
    group->hpgd_plid = plmdata->plid;
    group->hpgd_jboffset = plmdata->ploffset;
    group->hpgd_extentsz = plmdata->plextentsz;

    plmtup = (HeapTuple) heap_addheader(Natts_pg_plmap,
					sizeof(FormData_pg_plmap),
					(char *) plmdata);

    /* clean up the memory that heap_addheader() palloc()'ed for us */
    plmtup->t_oid = InvalidOid;
    bzero((char *) &(plmtup->t_chain), sizeof(plmtup->t_chain));

    /* open the relation and lock it */
    plmap = heap_openr(Name_pg_plmap);
    RelationSetLockForWrite(plmap);

    /* insert the new catalog tuple */
    heap_insert(plmap, plmtup);

    /* done */
    RelationUnsetLockForWrite(plmap);
    heap_close(plmap);
    pfree((char *) plmtup);

    /* Now we need to set the pg_platter entry right for the next free
     * block -- unlike the HP there is no other way to find out if the
     * block is alread allocated
     */
    
    plid = plmdata->plid;
    currBlock = plmdata->ploffset + plmdata->plextentsz;

    /* first get the tuple we want to replace */
    ScanKeyEntryInitialize(&skey, 0x0, ObjectIdAttributeNumber,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(plid));
    plat = heap_openr(Name_pg_platter);
    RelationSetLockForWrite(plat);
    platdesc = RelationGetTupleDescriptor(plat);
    platscan = heap_beginscan(plat, false, NowTimeQual, 1, &skey);
    plattup = heap_getnext(platscan, false, &buf);
    if (!HeapTupleIsValid(plattup))
	elog(WARN, "missing pg_platter tuple oid %ld", plid);
    
    /* first read the old offset */
    d = (Datum) heap_getattr(plattup, InvalidBuffer, Anum_pg_platter_plstart,
			     platdesc, &n);
    
    /* null means zero to us */
    if (n)
	platfirst = 0;
    else
	platfirst = DatumGetInt32(d);

    if (platfirst < currBlock) {
	int i;
	/* Need to update the pg_platter entry */

	/* form the new tuple */
	for (i = 0; i < 10; repl[i++] = ' ');

	nulls[Anum_pg_platter_plstart-1] = ' ';
	repl[Anum_pg_platter_plstart-1] = 'r';
	value[Anum_pg_platter_plstart-1] = Int32GetDatum(currBlock);

	newtup = heap_modifytuple (plattup, buf, plat, value, nulls, repl);
	heap_replace (plat, &plattup->t_ctid, newtup);
    }
    plattup = heap_getnext(platscan, false, &buf);
    if (HeapTupleIsValid(plattup))
	elog(WARN, "Platter name occurs twice in pg_platter %ld", plid);
    
    ReleaseBuffer(buf);
    RelationUnsetLockForWrite(plat);
    heap_endscan(platscan);
    heap_close(plat);
    pfree((char *) plmdata);
}

/*
 *  _hpallocgrp() -- Allocate a new group in the cache for use by some
 *		    relation.
 *
 *	If there are any unused slots in the cache, we just return one
 *	of those.  Otherwise, we need to kick out the least-recently-used
 *	group and make room for another.
 *
 *	On entry, we hold the cache metadata lock.  On exit, we still hold
 *	it.  In between, we may release it in order to do I/O on the cache
 *	group we're kicking out, if we have to do that.
 */

static HPCacheItem *
_hpallocgrp(int *grpno)
{
    HPCacheItem *item;

    /* free list had better not be empty */
    if (HPHeader->hph_nentries == HPCacheSize)
	elog(FATAL, "_hpallocgrp:  no groups on free list!");

    /*
     *  Get a new group off the free list.  As a side effect, _hpgetgrp()
     *  bumps the ref count on the group for us.
     */

    *grpno = _hpgetgrp();

    item = &HPCache[*grpno];

    return (item);
}

/*
 *  _hpgetgrp() -- Get a group off the free list.
 *
 *	This routine returns the least-recently used group on the free list
 *	to the caller.  If necessary, the (old) contents of the group are
 *	forced to the platter.  On entry, we hold the cache metadata lock.
 *	We release it and mark IOINPROG on the group if we need to do any
 *	io.  We reacquire the lock before returning.
 *
 *	We know that there's something on the free list when we call this
 *	routine.
 *
 *	There's an interesting problem with write-once media that we have
 *	to deal with here.  It is possible in postgres for a half-full
 *	buffer to be flushed to stable storage, then to be reloaded into
 *	the buffer cache, filled completely, and for a new page to be
 *	allocated before the old page is flushed again.  If this happens
 *	to us, it's possible for the half-full page to get flushed all the
 *	way through to an optical disk platter, where it can never be
 *	overwritten.
 *
 *	In order to deal with this, we probe the buffer manager for all
 *	dirty blocks it has that live on an extent before we flush the
 *	extent to permanent storage.
 */

static int
_hpgetgrp()
{
    HPCacheItem *item;
    int grpno;
    int where;
    long loc;
    bool found;
    int grpoffset;
    BlockNumber nblocks;
    Relation reln;
    bool dirty;
    int i;
    int offset;
    Oid dbid;
    Oid relid;
    BlockNumber base;

    /* pull the least-recently-used group off the free list */
    grpno = HPHeader->hph_freehead;
    if (grpno == -1) {
	elog (FATAL, "No free groups in cache");
    }
    item = &(HPCache[grpno]);
    _hptouch(item);

    /* if it was previously a valid group, remove it from the hash table */
    if (item->hpc_oid != InvalidOid)
	_hphashop(&(item->hpc_tag), HASH_REMOVE, &found);

    /*
     *  See if we need to flush the group to the jukebox.  If we're working
     *  with an entirely new item (the corresponding cache slot is empty),
     *  dbid == relid == base == 0, so we can ignore the flags.  Otherwise,
     *  we check every flags entry in the group descriptor to see if anyone
     *  wants to get flushed.
     */

    dirty = false;
    if (item->hpc_tag.hpct_dbid != 0 || item->hpc_tag.hpct_relid != 0) {
	if (MUST_FLUSH(item->hpc_gflags)) {
	    dirty = true;
	} else {
	    for (i = 0; i < HPGRPSIZE; i++) {
		if (MUST_FLUSH(item->hpc_flags[i])) {
		    dirty = true;
		    break;
		}
	    }
	}
    }

    if (!dirty)
	return (grpno);

    /*
     *  By here, we need to force the group to stable storage outside the
     *  cache.  Mark IOINPROG on the group (in fact, this shouldn't matter,
     *  since no one should be able to get at it -- we just got it off the
     *  free list and removed its hash table entry), release our exclusive
     *  lock, and write it out.
     */

    SET_IO_LOCK(item);

    if (_hpreadgrp(item, grpno) == SM_FAIL) {
	_hpunwait_io(item);
	elog(FATAL, "_hpgetgrp:  cannot read group %d", grpno);
    }

    /*
     *  Probe the buffer manager for dirty blocks that belong in this
     *  extent.  The buffer manager will copy them into the space we
     *  pass in, and will mark them clean in the buffer cache.
     */

    dbid = item->hpc_tag.hpct_dbid;
    relid = item->hpc_tag.hpct_relid;
    base = item->hpc_tag.hpct_base;

    for (i = 0; i < HPGRPSIZE; i++) {
	if (MUST_FLUSH(item->hpc_flags[i])) {
	    offset = (i * MAXBLCKSZ) + JBBLOCKSZ;
	    DirtyBufferCopy(dbid, relid, base + i, &(HPCacheBuf[offset]));
	}
    }

    nblocks = _hpfindnblocks(&(item->hpc_tag));

    if (pgjb_wrtextent(item, nblocks, &(HPCacheBuf[0])) == SM_FAIL) {
	_hpunwait_io(item);
	elog(FATAL, "_hpfree:  cannot free group.");
    }

    _hpunwait_io(item);

    /* give us back our exclusive lock */
    SpinAcquire(HPCacheLock);

    return (grpno);
}

static HPCacheItem *
_hpfetchgrp(Oid dbid, Oid relid, int blkno, int *grpno, int blockOnMiss)
{
    HPCacheItem *item;
    HPHashEntry *entry;
    bool found;
    HPCacheTag tag;

    SpinAcquire(HPCacheLock);

    tag.hpct_dbid = dbid;
    tag.hpct_relid = relid;
    tag.hpct_base = blkno;

    entry = _hphashop(&tag, HASH_FIND, &found);

    if (found) {
	*grpno = entry->hphe_groupno;
	item = &(HPCache[*grpno]);

	if (item->hpc_gflags & HPC_IOINPROG) {
	    _hpwait_io(item);
	    return (_hpfetchgrp(dbid, relid, blkno, grpno, blockOnMiss));
	}

	_hptouch(item);

	SpinRelease(HPCacheLock);
    } else {
	item = _hpallocgrp(grpno);

	/*
	 *  Possible race condition:  someone else instantiated the extent
	 *  we want while we were off allocating a group for it.  If that
	 *  happened, we want to put our just-allocated group back on the
	 *  free list for someone else to use.
	 */

	entry = _hphashop(&tag, HASH_FIND, &found);
	if (found) {
	    /*
	     *  Put the just-allocated group back on the free list.  This
	     *  requires us to reenter it into the hash table if it refers
	     *  to actual data.  We only want to do this if we got a different
	     *  free group from the other process.
	     */

	    if (entry->hphe_groupno != *grpno) {
		if (item->hpc_oid != InvalidOid)
		    (void) _hphashop(&(item->hpc_tag), HASH_ENTER, &found);
		_hpunpin(item);
	    }

	    item = &(HPCache[entry->hphe_groupno]);

	    /* if io in progress, wait for it to complete and try again */
	    if (item->hpc_gflags & HPC_IOINPROG) {
		_hpunpin(item);
		_hpwait_io(item);
		return _hpfetchgrp(dbid, relid, blkno, grpno, blockOnMiss);
	    }

	    SpinRelease(HPCacheLock);
	} else {
	    if (blockOnMiss) {
		return (NULL);
	    }
	    /* okay, we need to read the extent from a platter */
	    bcopy((char *) &tag, (char *) &(item->hpc_tag), sizeof(tag));
	    entry = _hphashop(&tag, HASH_ENTER, &found);
	    entry->hphe_groupno = *grpno;

	    SET_IO_LOCK(item);

	    /* read the extent off the optical platter */
	    _hprdextent(item);

	    /* update the magnetic disk cache */
	    _hpwritegrp(item, *grpno);

	    /* done, release IO lock */
	    _hpunwait_io(item);
	}
    }

    return (item);
}

/*
 *  _hprdextent() -- Read an extent from an optical platter.
 *
 *	This routine prepares the HPCacheItem group for the pgjb_rdextent()
 *	routine to work with, and passes it along.  We don't have exclusive
 *	access to the cache metadata on entry, but we do have the IOINPROGRESS
 *	bit set on the item we're working with, so on one else will screw
 *	around with it.
 */

static void
_hprdextent(HPCacheItem *item)
{
    Relation reln;
    HeapScanDesc hscan;
    HeapTuple htup;
    TupleDesc tupdesc;
    Datum d;
    bool n;
    Name plname;
    ScanKeyData skey[3];

    /* first get platter id and offset from pg_plmap */
    reln = heap_openr(Name_pg_plmap);
    tupdesc = RelationGetTupleDescriptor(reln);
    ScanKeyEntryInitialize(&skey[0], 0x0, Anum_pg_plmap_pldbid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(item->hpc_tag.hpct_dbid));
    ScanKeyEntryInitialize(&skey[1], 0x0, Anum_pg_plmap_plrelid,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(item->hpc_tag.hpct_relid));
    ScanKeyEntryInitialize(&skey[2], 0x0, Anum_pg_plmap_plblkno,
			   Integer32EqualRegProcedure,
			   Int32GetDatum(item->hpc_tag.hpct_base));
    hscan = heap_beginscan(reln, false, NowTimeQual, 3, &skey[0]);

    /*
     *  if there is no matching entry in the platter map, then we're
     *  asking for an extent that has not yet been allocated.  in this
     *  case, we return a zero-filled extent.  this happens, for example,
     *  when we try to read the initial block of a relation before one
     *  has been written.
     */

    if (!HeapTupleIsValid(htup = heap_getnext(hscan, false, (Buffer *) NULL))) {
	heap_endscan(hscan);
	heap_close(reln);
	bzero(&(HPCacheBuf[0]), HPBUFSIZE);
	return;
    }

    d = (Datum) heap_getattr(htup, InvalidBuffer, Anum_pg_plmap_plid,
			     tupdesc, &n);
    item->hpc_plid = DatumGetObjectId(d);
    d = (Datum) heap_getattr(htup, InvalidBuffer, Anum_pg_plmap_ploffset,
			     tupdesc, &n);
    item->hpc_jboffset = DatumGetInt32(d);

    heap_endscan(hscan);
    heap_close(reln);

    /* now figure out the platter's name from pg_platter */
    reln = heap_openr(Name_pg_platter);
    tupdesc = RelationGetTupleDescriptor(reln);
    ScanKeyEntryInitialize(&skey[0], 0x0, ObjectIdAttributeNumber,
			   ObjectIdEqualRegProcedure,
			   ObjectIdGetDatum(item->hpc_plid));
    hscan = heap_beginscan(reln, false, NowTimeQual, 1, &skey[0]);

    if (!HeapTupleIsValid(htup = heap_getnext(hscan, false, (Buffer *) NULL))) {
	_hpunwait_io(item);
	elog(WARN, "_hprdextent: cannot find platter oid %d", item->hpc_plid);
    }

    d = (Datum) heap_getattr(htup, InvalidBuffer, Anum_pg_platter_plname,
			     tupdesc, &n);
    plname = DatumGetName(d);
    bcopy(&(plname->data[0]), &(item->hpc_plname.data[0]), sizeof(NameData));

    heap_endscan(hscan);
    heap_close(reln);

    /*
     *  Okay, by here, we have all the fields in item filled in except for
     *  hpc_oid, hpc_gflags, and hpc_flags[].  Those are all filled in by
     *  pgjb_rdextent(), so we call that routine to do the work.
     */

    if (pgjb_rdextent(item, &HPCacheBuf[0]) == SM_FAIL) {
	_hpunwait_io(item);
	elog(WARN, "read of extent <%d,%d,%d> from platter %d failed",
		   item->hpc_tag.hpct_dbid, item->hpc_tag.hpct_relid,
		   item->hpc_tag.hpct_base, item->hpc_plid);
    }
}

/*
 *  _hptouch() -- Increment reference count on the supplied item.
 *
 *	If this is the first reference to the item, we remove it from the
 *	free list.  On entry and exit, we hold HPCacheLock.  If we pulled
 *	the item off the free list, we adjust HPHeader->hph_nentries.
 */

static void
_hptouch(HPCacheItem *item)
{
    /*
     *  Bump the reference count to this group.  If it's the first
     *  reference, pull the group off the free list.
     */

    if (++(item->hpc_refcount) == 1) {

	/* if at the start of the free list, adjust 'head' pointer */
	if (item->hpc_freeprev != -1)
	    HPCache[item->hpc_freeprev].hpc_freenext = item->hpc_freenext;
	else
	    HPHeader->hph_freehead = item->hpc_freenext;

	/* if at the end of the free list, adjust 'tail' pointer */
	if (item->hpc_freenext != -1)
	    HPCache[item->hpc_freenext].hpc_freeprev = item->hpc_freeprev;
	else
	    HPHeader->hph_freetail = item->hpc_freeprev;

	/* disconnect from free list */
	item->hpc_freeprev = item->hpc_freenext = -1;

	/* keep track of number of groups allocated */
	(HPHeader->hph_nentries)++;
    }
}

/*
 *  _hpunpin() -- Decrement reference count on the supplied item.
 *
 *	If we are releasing the last reference to the supplied item, we put
 *	it back on the free list.  On entry and exit, we do not hold the
 *	cache lock.  We must acquire it in order to carry out the requested
 *	release.
 */

static void
_hpunpin(HPCacheItem *item)
{
    int grpno;

    /* exclusive access */
    SpinAcquire(HPCacheLock);

    /* item had better be pinned */
    if (item->hpc_refcount <= 0)
	elog(FATAL, "_hpunpin: illegal reference count");

    /*
     *  Unpin the item.  If this is the last reference, put the item at the
     *  end of the free list.  Implemenation note:  if HPHeader->hph_freehead
     *  is -1, then the list is empty, and HPHeader->hph_freetail is also -1.
     */

    if (--(item->hpc_refcount) == 0) {

	grpno = GROUPNO(item);

	if (HPHeader->hph_freehead == -1) {
	    HPHeader->hph_freehead = grpno;
	} else {
	    item->hpc_freeprev = HPHeader->hph_freetail;
	    HPCache[HPHeader->hph_freetail].hpc_freenext = grpno;
	}

	/* put item at end of free list */
	HPHeader->hph_freetail = grpno;
	(HPHeader->hph_nentries)--;
    }

    SpinRelease(HPCacheLock);
}

static int
_hpwritegrp(HPCacheItem *item, int grpno)
{
    long seekpos;
    long loc;
    int nbytes, i;
    char *buf;

    /* first update the metadata file */
    seekpos = grpno * sizeof(*item);

    if ((loc = FileSeek(HPMetaVfd, seekpos, L_SET)) != seekpos)
	return (SM_FAIL);

    nbytes = sizeof(*item);
    buf = (char *) item;
    while (nbytes > 0) {
	i = FileWrite(HPMetaVfd, buf, nbytes);
	if (i < 0)
	    return (SM_FAIL);
	nbytes -= i;
	buf += i;
    }

    FileSync(HPMetaVfd);

    /* now update the cache file */
    seekpos = grpno * HPBUFSIZE;
    if ((loc = FileSeek(HPCacheVfd, seekpos, L_SET)) != seekpos)
	return (SM_FAIL);

    nbytes = HPBUFSIZE;
    buf = &(HPCacheBuf[0]);
    while (nbytes > 0) {
	i = FileWrite(HPCacheVfd, buf, nbytes);
	if (i < 0)
	    return (SM_FAIL);
	nbytes -= i;
	buf += i;
    }

    FileSync(HPCacheVfd);

    return (SM_SUCCESS);
}

/*
 *  hpextend() -- extend a relation by one block.
 */

int
hpextend(Relation reln, char *buffer)
{
    HPCacheItem *item;
    HPHashEntry *entry;
    HPCacheTag tag;
    int grpno;
    int nblocks;
    int base;
    int offset;
    bool found;
    int grpoffset;
    long seekpos;

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    RelationSetLockForExtend(reln);
    nblocks = hpnblocks(reln);
    base = (nblocks / HPGRPSIZE) * HPGRPSIZE;

    SpinAcquire(HPCacheLock);

    /*
     *  If the highest extent is full, we need to allocate a new group in
     *  the cache.  As a side effect, _hpnewextent will release HPCacheLock.
     *  We need to reacquire it immediately afterwards.
     */

    if ((nblocks % HPGRPSIZE) == 0) {
	_hpnewextent(reln, base);
	SpinAcquire(HPCacheLock);
    }

    if (reln->rd_rel->relisshared)
	tag.hpct_dbid = (Oid) 0;
    else
	tag.hpct_dbid = MyDatabaseId;

    tag.hpct_relid = reln->rd_id;
    tag.hpct_base = base;

    entry = _hphashop(&tag, HASH_FIND, &found);

    if (!found) {
	SpinRelease(HPCacheLock);
	elog(WARN, "hpextend:  hey mao:  your group is missing.");
    }

    /* find the item and block in the item to write */
    grpno = entry->hphe_groupno;
    item = &HPCache[grpno];
    grpoffset = nblocks % HPGRPSIZE;

    /*
     *  Okay, allocate the next block in this extent by marking it 'not
     *  missing'.  Once we've done this, we must hold the extend lock
     *  until end of transaction, since the number of allocated blocks no
     *  longer matches the block count visible to other backends.
     */

    if (!(item->hpc_flags[grpoffset] & HPC_MISSING)) {
	SpinRelease(HPCacheLock);
	elog(WARN, "hpextend: cache botch: next block in group present");
    } else {
	item->hpc_flags[grpoffset] &= ~HPC_MISSING;
    }

    _hptouch(item);

    SET_IO_LOCK(item);

    /* page is allocated */
    item->hpc_flags[grpoffset] = HPC_CLEAR;

    /* verify group descriptor data in the cache file */
    if (_hpgroupvrfy(item, grpno) == SM_FAIL) {
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    /* write the page */
    seekpos = (grpno * HPBUFSIZE) + ((nblocks % HPGRPSIZE) * MAXBLCKSZ)
	      + JBBLOCKSZ;
    if (FileSeek(HPCacheVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpextend: failed to seek to buffer lock (%d)", seekpos);
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    if (FileWrite(HPCacheVfd, buffer, MAXBLCKSZ) != MAXBLCKSZ) {
	elog(NOTICE, "hpextend: can't write page %d", nblocks);
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    /* write the updated cache metadata entry */
    seekpos = grpno * sizeof(*item);

    if (FileSeek(HPMetaVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpextend: seek to %d on metadata file failed", seekpos);
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    if (FileWrite(HPMetaVfd, (char *) item, sizeof(*item)) < 0) {
	elog(NOTICE, "hpextend: write of metadata file failed");
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    /* success */
    _hpunwait_io(item);
    _hpunpin(item);

    SpinAcquire(HPCacheLock);
    tag.hpct_base = ++nblocks;
    _hpregnblocks(&tag);
    SpinRelease(HPCacheLock);

    return (SM_SUCCESS);
}

static int
_hpreadgrp(HPCacheItem *item, int grpno)
{
    long seekpos;
    long loc;
    int nbytes, i;
    char *buf;
    HPGroupDesc *gdesc;

    /* get the group from the cache file */
    seekpos = grpno * HPBUFSIZE;
    if ((loc = FileSeek(HPCacheVfd, seekpos, L_SET)) != seekpos) {
	elog(NOTICE, "_hpreadgrp: cannot seek");
	return (SM_FAIL);
    }

    nbytes = HPBUFSIZE;
    buf = &(HPCacheBuf[0]);
    while (nbytes > 0) {
	i = FileRead(HPCacheVfd, buf, nbytes);
	if (i < 0) {
	    elog(NOTICE, "_hpreadgrp: read failed");
	    return (SM_FAIL);
	}
	nbytes -= i;
	buf += i;
    }

    gdesc = (HPGroupDesc *) &(HPCacheBuf[0]);

    if (gdesc->hpgd_magic != HPGDMAGIC
	|| gdesc->hpgd_version != HPGDVERSION
	|| gdesc->hpgd_groupoid != item->hpc_oid) {

	elog(NOTICE, "_hpreadgrp: trashed cache");
	return (SM_FAIL);
    }

    return (SM_SUCCESS);
}

int
hpunlink(Relation reln)
{
    Assert(reln->rd_blocksz == MAXBLCKSZ);
    
    return (SM_FAIL);
}

/*
 *  _hpnewextent() -- Add a new extent to a relation in the jukebox cache.
 */

static void
_hpnewextent(Relation reln, BlockNumber base)
{
    HPHashEntry *entry;
    HPGroupDesc *group;
    HPCacheItem *item;
    bool found;
    int grpno;
    int i;

    item = _hpallocgrp(&grpno);

    if (reln->rd_rel->relisshared)
	item->hpc_tag.hpct_dbid = (Oid) 0;
    else
	item->hpc_tag.hpct_dbid = MyDatabaseId;

    item->hpc_tag.hpct_relid = (Oid) reln->rd_id;
    item->hpc_tag.hpct_base = base;

    entry = _hphashop(&(item->hpc_tag), HASH_ENTER, &found);
#ifdef HAS_TEST_AND_SET
    S_UNLOCK(&(item->hpc_iolock));
#endif

    entry->hphe_groupno = grpno;

    SET_IO_LOCK(item);

    /* set flags on item, initialize group descriptor block */
    item->hpc_gflags = HPC_CLEAR;
    for (i = 0; i < HPGRPSIZE; i++)
	item->hpc_flags[i] = HPC_MISSING;

    /* should be smarter and only bzero what we need to */
    bzero(HPCacheBuf, HPBUFSIZE);

    group = (HPGroupDesc *) (&HPCacheBuf[0]);
    group->hpgd_magic = HPGDMAGIC;
    group->hpgd_version = HPGDVERSION;

    if (reln->rd_rel->relisshared) {
	group->hpgd_dbid = (Oid) 0;
    } else {
	strncpy(&(group->hpgd_dbname.data[0]),
		GetDatabaseName(),
		sizeof(NameData));
	group->hpgd_dbid = (Oid) MyDatabaseId;
    }

    strncpy(&(group->hpgd_relname.data[0]),
	    &(reln->rd_rel->relname.data[0]),
	    sizeof(NameData));
    group->hpgd_relid = reln->rd_id;
    group->hpgd_relblkno = base;
    item->hpc_oid = group->hpgd_groupoid = newoid();

    /*
     *  Record the presence of the new extent in the system catalogs.  The
     *  plid, jboffset, and extentsz fields are filled in by _hpregister()
     *  or the routines that it calls.  Note that we do not force the new
     *  group descriptor block all the way to the optical platter here.
     *  We do decide where to place it, however, and must go to a fair amount
     *  of trouble elsewhere in the code to avoid allocating the same extent
     *  to a different relation, or block within the same relation.
     */

    _hpregister(item, group);

    /*
     *  Write the new group cache entry to disk.  Hpwritegrp() knows where
     *  the cache buffer begins, and forces out the group descriptor we
     *  just set up.
     */

    if (_hpwritegrp(item, grpno) == SM_FAIL) {
	_hpunwait_io(item);
	elog(FATAL, "_hpnewextent: cannot write new extent to disk");
    }

    SpinAcquire(HPCacheLock);
    _hpregnblocks(&(item->hpc_tag));
    SpinRelease(HPCacheLock);

    /* can now release i/o lock on the item we just added */
    _hpunwait_io(item);

    /* no longer need the reference */
    _hpunpin(item);
}

/*
 *  _hphashop() -- Do lookup, insertion, or deletion on the metadata hash
 *		   table in shared memory.
 *
 *	We don't worry about the number of entries in the hash table here;
 *	that's handled at a higher level (_hpallocgrp and _hpgetgrp).  We
 *	hold HPCacheLock on entry.
 */

static HPHashEntry *
_hphashop(HPCacheTag *tagP, HASHACTION op, bool *foundP)
{
    HPHashEntry *entry;

    entry = (HPHashEntry *) hash_search(HPCacheHT, (char *) tagP, op, foundP);

    if (entry == (HPHashEntry *) NULL) {
	SpinRelease(HPCacheLock);
	elog(FATAL, "_hphashop: hash table corrupt.");
    }

    if (*foundP) {
	if (op == HASH_ENTER) {
	    SpinRelease(HPCacheLock);
	    elog(WARN, "_hphashop: cannot enter <%d,%d,%d>: already exists",
		 tagP->hpct_dbid, tagP->hpct_relid, tagP->hpct_base);
	}
    } else {
	if (op == HASH_REMOVE) {
	    SpinRelease(HPCacheLock);
	    elog(WARN, "_hphashop: cannot delete <%d,%d,%d>: missing",
		 tagP->hpct_dbid, tagP->hpct_relid, tagP->hpct_base);
	}
    }

    /*-------------------test code-------------------------------*/
    if (op == HASH_REMOVE) {
	(*HPNEntries)--;
    }
    if (op == HASH_ENTER) {
	(*HPNEntries)++;
	if (*HPNEntries > HPCacheSize) 
	    elog(WARN, "inserting more entries than declared! %d %d", 
		 *HPNEntries, HPCacheSize);
    }
    /*-------------------test code-------------------------------*/
	
    return (entry);
}

int
hpopen(Relation reln)
{
    char *path;
    int fd;
    extern char *relpath();

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    path = relpath(&(reln->rd_rel->relname.data[0]));

    fd = FileNameOpenFile(path, O_RDWR, 0600);

    return (fd);
}

int
hpclose(Relation reln)
{
    Assert(reln->rd_blocksz == MAXBLCKSZ);

    FileClose(reln->rd_fd);

    return (SM_SUCCESS);
}

int
hpread(Relation reln, BlockNumber blocknum, char *buffer)
{
    HPCacheItem *item;
    Oid reldbid;
    BlockNumber base;
    int offset;
    int grpno;
    long seekpos;

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    /* fake successful read on non-existent data */
    if (hpnblocks(reln) <= blocknum) {
	bzero(buffer, MAXBLCKSZ);
	return (SM_SUCCESS);
    }

    if (reln->rd_rel->relisshared)
	reldbid = (Oid) 0;
    else
	reldbid = MyDatabaseId;

    base = (blocknum / HPGRPSIZE) * HPGRPSIZE;
/*
    item = _hpfetchgrp(reldbid, reln->rd_id, base, &grpno, 1);

    if (item == NULL) {
	InformStateToScheduler();
	waitforio();
	return hpread(reln, blocknum, buffer);
    }
*/
    item = _hpfetchgrp(reldbid, reln->rd_id, base, &grpno, 0);

    /* shd expand _hpfetchgrp() inline to avoid extra semop()s */
    SpinAcquire(HPCacheLock);

    SET_IO_LOCK(item);

    /* First read and verify the group descriptor metadata */
    if (_hpgroupvrfy(item, grpno) == SM_FAIL) {
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    /* By here, group descriptor metadata is okay */
    seekpos = (grpno * HPBUFSIZE) + ((blocknum % HPGRPSIZE) * MAXBLCKSZ)
	      + JBBLOCKSZ;
    if (FileSeek(HPCacheVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpread: failed to seek to buffer lock (%d)", seekpos);
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    /* read the requested page */
    if (FileRead(HPCacheVfd, buffer, MAXBLCKSZ) != MAXBLCKSZ) {
	elog(NOTICE, "hpread: can't read page %d", blocknum);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    _hpunwait_io(item);
    _hpunpin(item);

    return (SM_SUCCESS);
}

static int
_hpgroupvrfy(HPCacheItem *item, int grpno)
{
    long seekpos;
    HPGroupDesc gdesc;

    seekpos = HPBUFSIZE * grpno;
    if (FileSeek(HPCacheVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpgroupvrfy: Cannot seek to %d on hp cache file",
		     seekpos);
	return (SM_FAIL);
    }

    if (FileRead(HPCacheVfd, (char *) &gdesc, sizeof(gdesc)) < 0) {
	elog(NOTICE, "hpgroupvrfy: Cannot read group desc from hp cache file");
	return (SM_FAIL);
    }

    if (gdesc.hpgd_magic != HPGDMAGIC
	|| gdesc.hpgd_version != HPGDVERSION
	|| gdesc.hpgd_groupoid != item->hpc_oid) {

	elog(NOTICE, "hpgroupvrfy: trashed cache");
	return (SM_FAIL);
    }

    return (SM_SUCCESS);
}

int
hpwrite(Relation reln, BlockNumber blocknum, char *buffer)
{
    Oid reldbid;

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    if (reln->rd_rel->relisshared)
	reldbid = (Oid) 0;
    else
	reldbid = MyDatabaseId;

    return (_hpdowrite(reldbid, reln->rd_id, blocknum, buffer));
}

static int
_hpdowrite(Oid reldbid, Oid relid, BlockNumber blocknum, 
	   Size blocksz, char *buffer)
{
    HPCacheItem *item;
    BlockNumber base;
    int offset;
    int grpno;
    int which;
    long seekpos;

    base = (blocknum / HPGRPSIZE) * HPGRPSIZE;

    item = _hpfetchgrp(reldbid, relid, base, &grpno, 0);

    /* shd expand _hpfetchgrp() inline to avoid extra semop()s */
    SpinAcquire(HPCacheLock);

    which = blocknum % HPGRPSIZE;

    if (item->hpc_flags[which] & HPC_ONPLATTER) {
	SpinRelease(HPCacheLock);
	_hpunpin(item);
	elog(WARN, "hpwrite: optical platters are write-once, cannot rewrite");
    }

    SET_IO_LOCK(item);

    item->hpc_flags[which] &= ~HPC_MISSING;

    /* verify group descriptor data in the cache file */
    if (_hpgroupvrfy(item, grpno) == SM_FAIL) {
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    /* write the page */
    seekpos = (grpno * HPBUFSIZE) + ((blocknum % HPGRPSIZE) * MAXBLCKSZ)
	      + JBBLOCKSZ;
    if (FileSeek(HPCacheVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpwrite: failed to seek to buffer lock (%d)", seekpos);
	_hpunpin(item);
	_hpunwait_io(item);
	return (SM_FAIL);
    }

    if (FileWrite(HPCacheVfd, buffer, MAXBLCKSZ) != MAXBLCKSZ) {
	elog(NOTICE, "hpwrite: can't read page %d", blocknum);
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    /* write the updated cache metadata entry */
    seekpos = grpno * sizeof(*item);

    if (FileSeek(HPMetaVfd, seekpos, L_SET) != seekpos) {
	elog(NOTICE, "hpwrite: seek to %d on metadata file failed", seekpos);
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    if (FileWrite(HPMetaVfd, (char *) item, sizeof(*item)) < 0) {
	elog(NOTICE, "hpwrite: write of metadata file failed");
	_hpunwait_io(item);
	_hpunpin(item);
	return (SM_FAIL);
    }

    _hpunwait_io(item);
    _hpunpin(item);

    return (SM_SUCCESS);
}

int
hpflush(Relation reln, BlockNumber blocknum, char *buffer)
{
    Assert(reln->rd_blocksz == MAXBLCKSZ);

    return (hpwrite(reln, blocknum, buffer));
}

int
hpblindwrt(char *dbstr, char *relstr, Oid dbid, Oid relid, BlockNumber blkno,
	   Size blocksz, char * buffer)
{
    Assert(blocksz == MAXBLCKSZ);

    return (_hpdowrite(dbid, relid, blkno, blocksz, buffer));
}

/*
 *  hpnblocks() -- Return the number of blocks that appear in this relation.
 *
 *	Rather than compute this by walking through pg_plmap and fetching
 *	groups off of platters, we store the number of blocks currently
 *	allocated to a relation in a special Unix file.
 */

int
hpnblocks(Relation reln)
{
    HPCacheTag tag;
    int nblocks;

    Assert(reln->rd_blocksz == MAXBLCKSZ);

    if (reln->rd_rel->relisshared)
	tag.hpct_dbid = (Oid) 0;
    else
	tag.hpct_dbid = MyDatabaseId;

    tag.hpct_relid = reln->rd_id;

    tag.hpct_base = (BlockNumber) _hpfindnblocks(&tag);

    return ((int) (tag.hpct_base));
}

/*
 *  _hpfindnblocks() -- Find block count for the (dbid,relid) pair.
 */

static int
_hpfindnblocks(HPCacheTag *tag)
{
    int nbytes;
    int i;
    HPCacheTag *cachetag;
    HPCacheTag mytag;

    cachetag = HPNBlockCache;
    i = 0;
    SpinAcquire(HPCacheLock);
    
    while (i < HPNBLKSIZE && cachetag->hpct_relid != (Oid) 0) {
	if (cachetag->hpct_dbid == tag->hpct_dbid
	    && cachetag->hpct_relid == tag->hpct_relid) {
	    SpinRelease(HPCacheLock);
	    return (cachetag->hpct_base);
	}
	i++;
	cachetag++;
    }
    SpinRelease(HPCacheLock);

    if (FileSeek(HPBlockVfd, 0L, L_SET) != 0) {
	elog(FATAL, "_hpfindnblocks: cannot seek to zero on block count file");
    }

    while ((nbytes = FileRead(HPBlockVfd, (char *)&mytag, sizeof(mytag))) > 0) {
	if (mytag.hpct_dbid == tag->hpct_dbid
	    && mytag.hpct_relid == tag->hpct_relid) { 

	    if (i == HPNBLKSIZE) {
		/* fast pseudo-random function */
		i = mytag.hpct_relid % HPNBLKSIZE;
		cachetag = &(HPNBlockCache[i]);
	    }

	    /* save cache tag */
	    SpinAcquire(HPCacheLock);
	    cachetag->hpct_dbid = mytag.hpct_dbid;
	    cachetag->hpct_relid = mytag.hpct_relid;
	    cachetag->hpct_base = mytag.hpct_base;
	    SpinRelease(HPCacheLock);
	    return (mytag.hpct_base);
	}
    }
    elog(FATAL, "_hpfindnblocks: cannot get block count for <%d,%d>",
		tag->hpct_dbid, tag->hpct_relid);
}

/*
 *  _hpregnblocks() -- Remember the count of blocks for this relid.
 */

static void
_hpregnblocks(HPCacheTag *tag)
{
    int loc;
    int i;
    HPCacheTag *cachetag;
    HPCacheTag mytag;

    cachetag = HPNBlockCache;
    i = 0;
    while (i < HPNBLKSIZE && cachetag->hpct_relid != (Oid) 0) {
	if (cachetag->hpct_dbid == tag->hpct_dbid
	    && cachetag->hpct_relid == tag->hpct_relid)
	    break;

	i++;
	cachetag++;
    }

    if (i == HPNBLKSIZE) {
	i = tag->hpct_relid % HPNBLKSIZE;
	cachetag = &(HPNBlockCache[i]);
    }

    cachetag->hpct_dbid = tag->hpct_dbid;
    cachetag->hpct_relid = tag->hpct_relid;
    cachetag->hpct_base = tag->hpct_base;

    /* update block count file */
    if (FileSeek(HPBlockVfd, 0L, L_SET) < 0) {
	elog(FATAL, "_hpregnblocks: cannot seek to zero on block count file");
    }

    loc = 0;
    mytag.hpct_base = tag->hpct_base;

    /* overwrite existing entry, if any */
    while (FileRead(HPBlockVfd, (char *) &mytag, sizeof(mytag)) > 0) {
	if (mytag.hpct_dbid == tag->hpct_dbid
	    && mytag.hpct_relid == tag->hpct_relid) {
	    if (FileSeek(HPBlockVfd, (loc * sizeof(HPCacheTag)), L_SET) < 0)
		elog(FATAL, "_hpregnblocks: cannot seek to loc");
	    if (FileWrite(HPBlockVfd, (char *) tag, sizeof(*tag)) < 0)
		elog(FATAL, "_hpregnblocks: cannot write nblocks");
	    return;
	}
	loc++;
    }

    /* new relation -- write at end of file */
    if (FileWrite(HPBlockVfd, (char *) tag, sizeof(*tag)) < 0)
	elog(FATAL, "_hpregnblocks: cannot write nblocks for new reln");
}
int
hpcommit()
{
    FileSync(HPMetaVfd);
    FileSync(HPCacheVfd);
    FileSync(HPBlockVfd);

    return (SM_SUCCESS);
}

int
hpabort()
{
    return (SM_SUCCESS);
}

/*
 *  hpmaxseg() -- Find highest segment number occupied by platter id plid
 *		  in the on-disk cache.
 *
 *	This routine is called from _pgjb_findoffset().  On entry here,
 *	we hold JBSpinLock, but not HPCacheLock.  We do something a little
 *	dangerous here; we trust the group descriptor metadata that is in
 *	shared memory to reflect accurately the state of the actual cache
 *	file.  This isn't so bad; if there's an inconsistency, there are
 *	exactly two possibilities:
 *
 *		+  There was a crash between metadata and cache update,
 *		   and we'll figure that out later;
 *
 *		+  Some other backend has IO_IN_PROG set on the group we
 *		   are examining, and we need to look at the group desc
 *		   on disk in order to find out if the group is on plid.
 *
 *	The second case basically means that we wind up holding HPCacheLock
 *	during a disk io, but that's a sufficiently rare event that we don't
 *	care.  I can't think of any cleaner way to do this, anyway.
 *
 *	We return the address of the first block of the highest-numbered
 *	extent that we have cached for plid.  If we have none cached, we
 *	return InvalidBlockNumber.
 */

BlockNumber
hpmaxseg(Oid plid)
{
    int i;
    long seekpos, loc;
    int nbytes;
    BlockNumber last;
    HPGroupDesc *group;

    /* XXX hold the lock for a walk of the entire cache */
    SpinAcquire(HPCacheLock);

    last = InvalidBlockNumber;
    group = (HPGroupDesc *) &(HPCacheBuf[0]);

    /*
     *  Walk backwards along the free list.  If we ever hit an unallocated
     *  block, we can stop searching.  Otherwise, we'll hit the head of the
     *  list when freeprev == -1.
     */

    for (i = HPHeader->hph_freetail;
	 i != -1 && HPCache[i].hpc_oid != InvalidOid;
	 i = HPCache[i].hpc_freeprev) {

	/* if IO_IN_PROG is set, we need to look at the group desc on disk */
	if (HPCache[i].hpc_gflags & HPC_IOINPROG) {
	    seekpos = i * HPBUFSIZE;
	    if ((loc = FileSeek(HPCacheVfd, seekpos, L_SET)) != seekpos) {
		SpinRelease(HPCacheLock);
		elog(NOTICE, "hpmaxseg: cannot seek");
		return (-1);
	    }

	    nbytes = FileRead(HPCacheVfd, (char *) group, sizeof(HPGroupDesc));
	    if (nbytes != sizeof(HPGroupDesc)) {
		SpinRelease(HPCacheLock);
		elog(NOTICE, "hpmaxseg: read of group desc %d failed", i);
		return (-1);
	    }

	    /* sanity checks */
	    if (group->hpgd_magic != HPGDMAGIC
		|| group->hpgd_version != HPGDVERSION) {
		elog(FATAL, "hpmaxseg: cache file corrupt.");
	    }

	    if (group->hpgd_plid == plid) {
		if (group->hpgd_jboffset > last || last == InvalidBlockNumber)
		    last = group->hpgd_jboffset;
	    }
	} else {
	    if (HPCache[i].hpc_plid == plid) {
		if (HPCache[i].hpc_jboffset > last
		    || last == InvalidBlockNumber) {

		    last = HPCache[i].hpc_jboffset;
		}
	    }
	}
    }

    SpinRelease(HPCacheLock);

    return (last);
}

static void
_hpdump()
{
    int i, j;
    int nentries;
    HPCacheItem *item;

    SpinAcquire(HPCacheLock);

    nentries = HPHeader->hph_nentries;

    printf("jukebox cache metdata: size %d, %d entries, free head %d tail %d",
	   HPCacheSize, nentries, HPHeader->hph_freehead,
	   HPHeader->hph_freetail);
    if (HPHeader->hph_flags & HPH_INITING)
	printf(", INITING");
    if (HPHeader->hph_flags & HPH_INITED)
	printf(", INITED");
    printf("\n");

    for (i = 0; i < HPCacheSize; i++) {
	item = &HPCache[i];
	printf("    [%2d] <%ld,%ld,%ld> %d@%d next %d prev %d flags %s Oid %ld\n",
	       i, item->hpc_tag.hpct_dbid, item->hpc_tag.hpct_relid,
	       item->hpc_tag.hpct_base, item->hpc_plid, item->hpc_jboffset,
	       item->hpc_freenext, item->hpc_freeprev,
	       (item->hpc_gflags & HPC_IOINPROG ? "IO_IN_PROG" : "CLEAR"),
	       item->hpc_oid);
	printf("         ");
	for (j = 0; j < HPGRPSIZE; j++) {
	    printf("[%d %c%c]", j,
	    	   (item->hpc_flags[j] & HPC_MISSING ? 'm' : '-'),
	    	   (item->hpc_flags[j] & HPC_ONPLATTER ? 'o' : '-'));
	}
	printf("\n");
    }

    SpinRelease(HPCacheLock);
}

/*
 *  HPInitSemaphore() -- Initialize the 'wait' semaphore for jukebox cache
 *			 pages.
 *
 *	We only do this if we don't have test-and-set locks.
 */

HPInitSemaphore(IPCKey key)
{
#ifndef HAS_TEST_AND_SET
    int status;

    HPWaitSemId = IpcSemaphoreCreate(IPCKeyGetHPWaitSemaphoreKey(key),
				     1, IPCProtection, 0, &status);
    if (HPWaitSemId < 0) {
	elog(FATAL, "cannot create/attach jukebox semaphore");
    }
#else /* ndef HAS_TEST_AND_SET */
    return;
#endif /* ndef HAS_TEST_AND_SET */
}


/* ---------------------------------------------------------------------
 * routines added by me to get fragment oriented interface
 */

Datum GetPlatName();
int HPCacheRel();
int HPUncacheRel();
int GetRelSize();
void ClearTMCache();
void ResetCache();

Oid GetDBid(), GetRelid();
static HPCacheItem * _GetItem();
static bool _IsDirty();


/* hpMyChoose ()
 *     Choose a platter to receive the next extent.
 *   There is basically no allocation policy. The catalog pg_plrel
 *   is read to find the name of the platter to which the extent of
 *   the relation should go. It is assumed for now that all extents
 *   of a relation are placed in the same platter. Space management
 *   is the responsibility of the maker of the catalog pg_plrel.
 *   In future range of extents may be specified for allocation to
 *   the same platter.
 */
static Form_pg_plmap
_hpMychoose(HPCacheItem *item, Name relname)
{
    Relation plat;
    TupleDesc platdesc;
    HeapScanDesc platscan;
    HeapTuple plattup;
    Buffer buf;
    Form_pg_plmap plmdata;
    Oid plid;
    Datum d;
    char *plname;
    bool isnull;
    bool done;
    int alloctype;
    ScanKeyData skey;
    
    
    /* allocate the tuple form */
    plmdata = (Form_pg_plmap) palloc(sizeof(FormData_pg_plmap));
    plname = (char *) palloc(sizeof(NameData) + 1);
    
    d = GetPlatName (relname, plname);
    /*
     * Find the pid of the platter from pg_platter given its name
     */
    plat = heap_openr(Name_pg_platter);
    RelationSetLockForWrite(plat);
    ScanKeyEntryInitialize (&skey, (bits16)0x0, Anum_pg_platter_plname,
			    (RegProcedure)NameEqualRegProcedure, d);
    platdesc = RelationGetTupleDescriptor(plat);
    platscan = heap_beginscan(plat, false, NowTimeQual, 1, &skey);
    plattup = heap_getnext(platscan, false, &buf);
    if (!HeapTupleIsValid(plattup))
	elog(WARN, "_hpMychoose: no %s platter in pg_platter", plname);
    plid = plmdata->plid = plattup->t_oid;
    
    /*
     * make sure there is free space for at least one extent in the
     * named platter
     */
    done = pgjb_freespc(plname, plid, HPOLDRELN);
    
    ReleaseBuffer(buf);
    
    if (!done)
	elog(WARN, "_hpchoose: no space on platter %s", plname);
    
    /* init the rest of the fields */ 
    plmdata->pldbid = item->hpc_tag.hpct_dbid; 
    plmdata->plrelid = item->hpc_tag.hpct_relid; 
    plmdata->plblkno = item->hpc_tag.hpct_base; 
    plmdata->plextentsz = HPEXTENTSZ;
    plmdata->ploffset = pgjb_offset(plname, plmdata->plid,
				     plmdata->plextentsz);
    
    /* no longer need an exclusive lock for the allocation */
    RelationUnsetLockForWrite(plat);

    heap_endscan(platscan);
    heap_close(plat);

    /* save platter name, id, offset in item */
    bcopy(plname, &(item->hpc_plname.data[0]), sizeof(NameData));
    item->hpc_plid = plmdata->plid;
    item->hpc_jboffset = plmdata->ploffset;

    return (plmdata);
}

/* -----------------------------------------------------------------------
 * Get the name of the platter holding the named relation
 */
Datum
GetPlatName(Name relname, char *plname)
{
    Name platname;
    Relation plrel;
    TupleDesc reldesc;
    HeapScanDesc relscan;
    HeapTuple reltup;
    Buffer buf;
    Datum d;
    bool isnull;
    ScanKeyData skey;

    /*
     * Given the name of the relation for which extent is being
     * allocated, find the name of the platter that should hold
     * this extent from pg_plrel
     */
    plrel = heap_openr (Name_pg_plrel);
    ScanKeyEntryInitialize (&skey, (bits16)0x0, Anum_pg_plrel_relname,
			    (RegProcedure)NameEqualRegProcedure, 
			    NameGetDatum(relname));

    reldesc = RelationGetTupleDescriptor(plrel);
    relscan = heap_beginscan (plrel, false, NowTimeQual, 1, &skey);
    reltup = heap_getnext(relscan, false, &buf);
    if (!HeapTupleIsValid(reltup))
	elog(WARN, "_hpMychoose: relation %s not in pg_plrel", relname);
    d = (Datum) heap_getattr(reltup, buf, Anum_pg_plrel_plname,
			     reldesc, &isnull);
    platname = DatumGetName(d);
    strncpy(plname, (char *)platname, sizeof(NameData));
    plname[sizeof(NameData)] = '\0';
    ReleaseBuffer(buf);
    heap_endscan(relscan);
    heap_close(plrel);
    return d;
}

RelidGetPlatterId(Oid relid)
{
    Name platname, relname;
    Relation plrel, plat;
    TupleDesc reldesc, platdesc;
    HeapScanDesc relscan, platscan;
    HeapTuple reltup, plattup;
    Buffer buf;
    Datum d;
    bool isnull;
    ScanKeyData skey;
    Oid plid;

    relname = RelationGetRelationName (RelationIdGetRelation(relid));
    /*
     * Given the name of the relation for which extent is being
     * allocated, find the name of the platter that should hold
     * this extent from pg_plrel
     */
    plrel = heap_openr (Name_pg_plrel);
    ScanKeyEntryInitialize (&skey, (bits16)0x0, Anum_pg_plrel_relname,
			    (RegProcedure)NameEqualRegProcedure, 
			    NameGetDatum(relname));

    reldesc = RelationGetTupleDescriptor(plrel);
    relscan = heap_beginscan (plrel, false, NowTimeQual, 1, &skey);
    reltup = heap_getnext(relscan, false, &buf);
    if (!HeapTupleIsValid(reltup))
	elog(WARN, "_hpMychoose: relation %s not in pg_plrel", relname);
    d = (Datum) heap_getattr(reltup, buf, Anum_pg_plrel_plname,
			     reldesc, &isnull);
    platname = DatumGetName(d);
    
    plat = heap_openr(Name_pg_platter);
    
    ScanKeyEntryInitialize (&skey, (bits16)0x0, Anum_pg_platter_plname,
			    (RegProcedure)NameEqualRegProcedure, d);
    platdesc = RelationGetTupleDescriptor(plat);
    platscan = heap_beginscan(plat, false, NowTimeQual, 1, &skey);
    plattup = heap_getnext(platscan, false, &buf);
    if (!HeapTupleIsValid(plattup))
	elog(WARN, "_hpMychoose: no %s platter in pg_platter", platname);
    plid = plattup->t_oid;
    
    ReleaseBuffer(buf);
    heap_endscan(relscan);
    heap_endscan(platscan);
    heap_close(plrel);
    heap_close(plat);
    return plid;
}
/* --------------------------------------------------------------------
 * get the size of a relation in bytes
 */
int
RelidGetRelSize(Oid relid)
{
    Oid dbid = MyDatabaseId;
    HPCacheTag tag;
    
    tag.hpct_relid = relid;
    tag.hpct_dbid = dbid;

    return MAXBLCKSZ * _hpfindnblocks(&tag);
}
    
/* ---------------------------------------------------------
 * when items are read through hpread() they are not unpinned from
 * cache -- this makes sure that groups are put in free list only
 * when we call HPUncacheRel() or when hpwrite() is called -- we
 * need to fix some policy for writes but so far we just assume
 * LRU since our optimization algos do not involve any writes. later
 * if we want to store temporaries etc we will need to think of all
 * this in detail
 */

/* HPCacheRel() --
 *   It pins into the disk cache all blocks belonging to relation reln.
 *   It is assumed that there is enough space in the buffer to pin
 *   all the blocks of a relation. If not, the latter blocks of the
 *   relation might be displacing some of its own blocks to get space
 *   in the buffer. The upper level routines are supposed to manage
 *   the routines.
 *
 */
int
CacheRelation(Oid relid)
{
    HPCacheItem *item;
    BlockNumber numBlocks, i, base;
    int grpno;
    Oid dbid = MyDatabaseId;
    HPCacheTag tag;
    
    tag.hpct_relid = relid;
    tag.hpct_dbid = dbid;

    numBlocks = (BlockNumber) _hpfindnblocks(&tag);

    for (i = 0; i < numBlocks; i += HPGRPSIZE) {
	item = _hpfetchgrp(dbid, relid, i, &grpno, 0);
    }
}


/* HPUncacheRel () --
 *   Unpin from disk cache all the blocks of given relation -- the way this
 *   is done is to place all the extents belonging to the relation in the
 *   free list
 */
int
UnCacheRelation(Oid relid)
{
    HPCacheItem *item;
    Oid dbid = MyDatabaseId;
    BlockNumber numBlocks, i, base;
    int grpno;
    HPCacheTag tag;
    
    tag.hpct_relid = relid;
    tag.hpct_dbid = dbid;

    numBlocks = (BlockNumber) _hpfindnblocks(&tag);

    for (i = 0; i < numBlocks; i += HPGRPSIZE) {
	item = _GetItem (dbid, relid, i, &grpno);
	if (item == NULL) continue;
	_hpunpin (item);
    }
}

/* ---------------------------------------------------------- 
 * ResetCache() --- reset cache state
 *   For my experiments we want flush all dirty blocks into
 *   tertiary memory and set the cache and metadata files to 0 so that
 *   the next time we use it we are not constrained by what size of cache
 *   we used last time and our measurements do not screw up because of old
 *   data from previous runs
 */
void
ResetCache()
{ 
    char * path;

    ClearTMCache();    
    path = (char *) palloc(strlen(DataDir) + strlen(HPCACHENAME) + 2);
    sprintf(path, "%s/%s", DataDir, HPCACHENAME);
    truncate(path, 0);

    path = (char *) palloc(strlen(DataDir) + strlen(HPMETANAME) + 2);
    sprintf(path, "%s/%s", DataDir, HPMETANAME);
    truncate(path, 0);
}

/* ---------------------------------------------------------------------
 * we need this primitive to make sure that all dirty blocks get
 * flushed in case we decide to change cache structure and do not
 * want to loose data in cache
 * MAKE SURE hpinit() is called before calling this routine
 */

void
ClearTMCache()
{
    int i, j;
    int nentries;
    HPCacheItem *item;
    bool dirty;
    bool found;
    BlockNumber nblocks;
    int offset;
    Oid dbid;
    Oid relid;
    BlockNumber base;


    nentries = HPHeader->hph_nentries;

    for (i = 0; i < HPCacheSize; i++) {
	item = &HPCache[i];
	
	/* if it was previously a valid group, remove it from hash table */
	if (item->hpc_oid != InvalidOid)
	    _hphashop(&(item->hpc_tag), HASH_REMOVE, &found);

	dirty = _IsDirty (item);
	if (!dirty) continue;
	
	if (_hpreadgrp(item, i) == SM_FAIL) {
	    _hpunwait_io(item);
	    elog(FATAL, "_hpgetgrp:  cannot read group %d", i);
	}

	/*
	 *  Probe the buffer manager for dirty blocks that belong in this
	 *  extent.  The buffer manager will copy them into the space we
	 *  pass in, and will mark them clean in the buffer cache.
	 */

	dbid = item->hpc_tag.hpct_dbid;
	relid = item->hpc_tag.hpct_relid;
	base = item->hpc_tag.hpct_base;

	for (j = 0; j < HPGRPSIZE; j++) {
	    if (MUST_FLUSH(item->hpc_flags[j])) {
		offset = (j * MAXBLCKSZ) + JBBLOCKSZ;
		DirtyBufferCopy(dbid, relid, base + j, &(HPCacheBuf[offset]));
	    }
	}

	nblocks = _hpfindnblocks(&(item->hpc_tag));

	if (pgjb_wrtextent(item, nblocks, &(HPCacheBuf[0])) == SM_FAIL) {
	    _hpunwait_io(item);
	    elog(FATAL, "_hpfree:  cannot free group.");
	}
	_hpunwait_io(item);
    }
}

/* --------------------------
 * GetCacheSize -- get size of cache in bytes
 */
int
GetCacheSize()
{
    return HPCacheSize*HPGRPSIZE*MAXBLCKSZ;
}
/* -------------------------------------------------------------------
 */
Oid
GetDBid (Relation reln)
{
    if (reln->rd_rel->relisshared)
	return (Oid) 0;
    else
	return MyDatabaseId;
}

Oid
GetRelid(Relation reln)
{
    return reln->rd_id;
}

static HPCacheItem *
_GetItem (Oid dbid, Oid relid, int blkno, int *grpno)
{
    HPCacheItem *item;
    HPHashEntry *entry;
    bool found;
    HPCacheTag tag;

    item = NULL;
    tag.hpct_dbid = dbid;
    tag.hpct_relid = relid;
    tag.hpct_base = blkno;

    SpinAcquire(HPCacheLock);
    entry = _hphashop(&tag, HASH_FIND, &found);

    if (found) {
	*grpno = entry->hphe_groupno;
	item = &(HPCache[*grpno]);

	if (item->hpc_gflags & HPC_IOINPROG) {
	    item = NULL;
	}
    }
    SpinRelease(HPCacheLock);
    return (item);
}

/*
 *  See if we need to flush the group to the jukebox.  If we're working
 *  with an entirely new item (the corresponding cache slot is empty),
 *  dbid == relid == base == 0, so we can ignore the flags.  Otherwise,
 *  we check every flags entry in the group descriptor to see if anyone
 *  wants to get flushed.
 */
static bool
_IsDirty(HPCacheItem *item)
{
    bool dirty;
    int i;

    dirty = false;
    if (item->hpc_tag.hpct_dbid != 0 || item->hpc_tag.hpct_relid != 0) {
	if (MUST_FLUSH(item->hpc_gflags)) {
	    dirty = true;
	} else {
	    for (i = 0; i < HPGRPSIZE; i++) {
		if (MUST_FLUSH(item->hpc_flags[i])) {
		    dirty = true;
		    break;
		}
	    }
	}
    }
    return dirty;
}

/* --------------------------------------------------
 * the size of the cache = number of extents. Size of an extent in the default
 * case is 129*1024.
 */ 
SetCacheSize(int size) /* size in megabytes */
{
    HPCacheSize = (size*1024*1024)/HPBUFSIZE;
}

#endif /* HP_JUKEBOX */









