/*-------------------------------------------------------------------------
 *
 * us_catalog.c--
 *    MARIPOSA fragment usage statistics routines that interact with the
 *        system catalogs (and therefore, cannot be included in the 
 *        postmaster).
 *
 * Copyright (c) 1994,1995 Regents of the University of California
 *
 * $Id
 * [missing RCS id]
 *-------------------------------------------------------------------------
 */

#include "postgres.h"
#include "access/htup.h"        /* for HeapTuple typedef and GETSTRUCT() macro */
#include "access/relscan.h"     /* for HeapScanDesc typedef */
#include "access/heapam.h"      /* for heap access function prototypes */
#include "catalog/pg_has.h"
#include "catalog/pg_access.h"
#include "storage/ipc.h"
#include "storage/shmem.h"
#include "storage/spin.h"
#include "storage/usage_stat.h"
#include "storage/us_internals.h"
#include "utils/builtins.h"     /* for namecpy function prototype */
#include "utils/hsearch.h"
#include "utils/memutils.h"
#include "utils/elog.h"
#include "utils/rel.h"          /* for Relation typedef */

typedef struct {
    time_t lastAccessTime;
    int index;
} SortingArrayEntry;

/*
 * XXXRHPXXX - Needs to be thought out before implementation.
 */
#define USHL_CreateHotList(ents,len,frags) {}

/*
 * XXXRHPXXX - Needs to be thought out before implementation.
 */
#define USHL_CreateEntry(ptr,idx) {}

#define USEL_CreateEntryFromTuple(ent,tuple) { \
	    ret->fragmentId = tuple->fralogicalid; \
	    ret->prev = INVALID_OFFSET; \
	    ret->next = INVALID_OFFSET; \
	    ret->siteId = tuple->fralocation; \
	    ret->debits = tuple->fradebits; \
	    ret->credits = tuple->fracredits; \
	    ret->expectedPrice = tuple->fraexpprice; \
	    ret->numAccesses = tuple->fraaccesses; \
	    ret->arrivalTime = tuple->fraarrtime; \
	    ret->lastAccessTime = tuple->fralastacctime; \
	    namecpy(&(ret->dbName), &(tuple->datname)); \
	}

#define USEL_CreateSortingArrayEntry(ent,frag,i) {\
	    ent->lastAccessTime = frag->lastAccessTime; \
	    ent->index = i; \
	}

#define USEL_LinkList(len) { \
	    int i; \
	    UsageStatEntry *current, *next; \
	    SHMEM_OFFSET c, n; \
	    \
	    current = UsageStatEntryList; \
	    current->prev = INVALID_OFFSET; \
	    c = (SHMEM_OFFSET)MAKE_OFFSET(UsageStatEntryList); \
	    *UsageStatEntryListHead = c; \
	    \
	    for (i = 1; 1 < len; i++) { \
	        next = &UsageStatEntryList[i]; \
		next->prev = c; \
		n = (SHMEM_OFFSET)MAKE_OFFSET(next); \
		current->next = n; \
		\
		current = next; \
		c = n; \
	    } \
	    \
	    current->next = INVALID_OFFSET; \
	    *UsageStatEntryListTail = c; \
	}

#define US_INIT_STAT_ARRAY(x,size) { \
	    x = (UsageStatEntry *)calloc(100, sizeof(UsageStatEntry)); \
	    Assert(x != NULL); \
	    size = 100; \
	}

#define US_CHECK_STAT_ARRAY_LENGTH(x,idx,size) { \
	    if (idx + 1 >= size) { \
		x = (UsageStatEntry *)realloc((void *)x,size + 100); \
		Assert(x != NULL); \
		memset((void *)&(x[size]), 0, (100 * sizeof(UsageStatEntry *))); \
		size += 100; \
	    } \
	}

static int compar_descending_time(const void *, const void *);
static int compar_descending_revenue(const void *, const void *);
static void USEL_CreateEntry(UsageStatEntry *);
static UsageStatEntry *USEL_ReadDataFromCatalogs(int *);
static void USEL_WriteDataToCatalogs(UsageStatEntry *, int);



/*********************************************************************************
 *                        Exported Interface functions                           *
 *********************************************************************************/

/*
 * This function will be called on startup to read the data from the two shared
 * system catalogs into an array of UsageStatEntry structures, sort the arrays
 * by the time of last access field and write them into the shared memory EntryList
 * and UsageHashTable.  Then, it calculates the initial hot list.
 *
 * XXX - Calculating the hot list really has nothing to do with accessing 
 *       shared memory and should be factored out into a function in the databroker.
 */
void
UsageStatInitShmemData()
{
    UsageStatEntry *frags = NULL;
    int i, j = 0, nstop, num_ents;
    size_t ent_size = (size_t)sizeof(SortingArrayEntry);
    SortingArrayEntry *ents;

    UsageStatAcquireExLock(UsageStatLock);
    if (*UsageStatInitialized) {
        elog(DEBUG, "UsageStatInitShmemData: Usage statistics already initialized");
    }
    else {
        frags = USEL_ReadDataFromCatalogs(&num_ents);
	if (frags != NULL && num_ents > 0) {
	    ents = (SortingArrayEntry *)calloc(num_ents,
					       sizeof(SortingArrayEntry));
	    Assert(ents != NULL);
	    for (i = 0; i < num_ents; i++)
	        USEL_CreateSortingArrayEntry((&ents[i]),(&frags[i]),i);
	    qsort((void *)ents, (size_t)num_ents,
		  ent_size, compar_descending_time);
	    nstop = (num_ents < STABLE_NENTRIES) ? num_ents : STABLE_NENTRIES;

	    for (i = 0; i < nstop; i++)
	        USEL_CreateEntry(&frags[ents[i].index]);

	    USEL_LinkList(nstop);
	    USHL_CreateHotList(ents,n_ents,frags);
	    free(ents);
	}
	*UsageStatHotListNumEntries = j;
	*UsageStatInitialized = 1;
    }
    UsageStatReleaseExLock(UsageStatLock);
    return;
}

/*
 * For now, we aren't going to malloc and bcopy the UsageStatEntryList since
 * we are using a single lock (i.e., no one else can change shared memory
 * while we have the lock).  If the locking strategy changes, we should
 * change this so we can copy shared memory and release the lock while we
 * update the system catalogs. -rhp
 */
void
UsageStatFlushShmemToDisk()
{
    UsageStatEntry *entlist;
    int len;

    UsageStatAcquireExLock(UsageStatLock);
    entlist = UsageStatEntryList;
    len = (int)*UsageStatNumStatStructsUsed;
    USEL_WriteDataToCatalogs(entlist, len);
    UsageStatReleaseExLock(UsageStatLock);
    return;
}

void
UsageStatUpdateHotList()
{
    /*
     * This needs to be thought out before implementation.
     */
    return;
}

/*****************************************************************************
 *                             Internal functions                            *
 *****************************************************************************/

/*
 *Compare functions used in qsort calls.
 */
static int
compar_descending_time(const void *first, const void *second)
{
    SortingArrayEntry *f = (SortingArrayEntry *)first,
                      *s = (SortingArrayEntry *)second;

    if (f->lastAccessTime > s->lastAccessTime)
        return -1;
    else if (f->lastAccessTime < s->lastAccessTime)
        return 1;
    else
        return 0;
}

static void
USEL_CreateEntry(UsageStatEntry *info)
{
    int idx = *UsageStatNumStatStructsUsed;
    UsageHashTableEntry *h_ent;
    UsageStatEntry *ent;
    SHMEM_OFFSET ent_offset;
    long *tmp;
    bool found;

    if (idx >= STABLE_NENTRIES)
        elog(FATAL, "USEL_CreateEntry: All structures already in use");

    ent = &UsageStatEntryList[idx];
    Assert(ShmemIsValid((unsigned long)ent));
    ent_offset = (SHMEM_OFFSET)MAKE_OFFSET(ent);

    tmp = hash_search(UsageStatTable, (char *)&info->fragmentId, HASH_ENTER, &found);

    if (tmp == NULL)
        elog(FATAL, "USEL_CreateEntry: UsageHashTable corrupted");
    if (found)
        elog(WARN, "USEL_CreateEntry: An entry for %d already exists",
	     ent->fragmentId);

    h_ent = (UsageHashTableEntry *)tmp;
    h_ent->location = ent_offset;
    memset((void *)ent, 0, sizeof(UsageStatEntry));
    ent->fragmentId = info->fragmentId;
    ent->prev = INVALID_OFFSET;
    ent->next = INVALID_OFFSET;
    USEL_UpdateEntryData(ent, info);
    *UsageStatNumStatStructsUsed++;
    return;
}

/*
 * This function reads the data from the shared system catalogs pg_has
 * and pg_access and creates an UsageStatEntry array which is used to
 * initialize the shared memory area.
 */
static UsageStatEntry *
USEL_ReadDataFromCatalogs(int *len)
{
    Relation desc;
    HeapScanDesc hscan;
    HeapTuple htup;
    UsageStatEntry *ret = NULL;
    int idx = 0,
        size = 0;

    US_INIT_STAT_ARRAY(ret,size);
    desc = heap_openr((Name)Name_pg_has);
    if (!RelationIsValid(desc))
        elog(WARN, "USEL_ReadDataFromCatalogs: cannot open pg_has");

    hscan = heap_beginscan(desc, 0, SelfTimeQual, 0, NULL);
    do {
        htup = heap_getnext(hscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_has tuple = (Form_pg_has)GETSTRUCT(htup);
	    US_CHECK_STAT_ARRAY_LENGTH(ret,idx,size);
	    USEL_CreateEntryFromTuple((&ret[idx]),tuple);
	    idx++;
	}
    } while (htup);
    heap_endscan(hscan);
    heap_close(desc);

    desc = heap_openr((Name)Name_pg_access);
    if (!RelationIsValid(desc))
        elog(WARN, "USEL_ReadDataFromCatalogs: cannot open pg_access");

    hscan = heap_beginscan(desc, 0, SelfTimeQual, 0, NULL);
    do {
        htup = heap_getnext(hscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_access tuple = (Form_pg_access)GETSTRUCT(htup);
	    US_CHECK_STAT_ARRAY_LENGTH(ret,idx,size);
	    USEL_CreateEntryFromTuple((&ret[idx]),tuple);
	    idx++;
	}
    } while (htup);
    heap_endscan(hscan);
    heap_close(desc);

    if (idx != 0) {
        /*
	 * Don't really need to do this but it's good style.
	 */
        if (idx != size) {
	    ret = (UsageStatEntry *)realloc((void *)ret, (idx * sizeof(UsageStatEntry)));
	    Assert(ret != NULL);
	}
	*len = idx;
	return ret;
    }
    else {
        *len = 0;
        return NULL;
    }
}

static void
USEL_WriteDataToCatalogs(UsageStatEntry *ents, int len)
{
    Relation desc;
    HeapScanDesc hscan;
    HeapTuple htup;
    Buffer buffer;
    int i, h_idx, a_idx; 
    int insHas = 0, insAcc = 0, *has, *access;
    char ent_used[STABLE_NENTRIES];

    Assert(ents != NULL && len != 0);
    memset((void *)&(ent_used[0]), 0, (STABLE_NENTRIES * sizeof(int)));

    desc = heap_openr((Name)Name_pg_has);
    if (!RelationIsValid(desc))
        elog(WARN, "USEL_WriteDataToCatalogs: cannot open pg_has");

    hscan = heap_beginscan(desc, 0, SelfTimeQual, 0, NULL);
    do {
        htup = heap_getnext(hscan, 0, &buffer);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_has tuple = (Form_pg_has)GETSTRUCT(htup);
	    Oid fragId = tuple->fralogicalid;

	    for (i = 0; i < len; i++)
	        if (ents[i].fragmentId == fragId)
		    break;
	    /*
	     * If we found a "matching" entry, we have to jump through 
	     * a few hoops to "update" the tuple.
	     */
	    if (i != len) {
	        int j;
		Datum values[Natts_pg_has];
		char nulls[Natts_pg_has];
		char replaces[Natts_pg_has];
		ItemPointerData tmp_ipd;
		NameData tmp_dbn;

		if (ent_used[i] != (char)0)
		    elog(DEBUG,
			 "USEL_WriteDataToCatalogs: multiple entries for fragment %d in pg_has",
			 (int)fragId);
		/*
		 * Initialize the arrays used to update the tuple data
		 */
		for (j = 0; j < Natts_pg_has; j++) {
		    replaces[j] = ' ';
		    nulls[j] = ' ';
		    values[j] = (Datum)NULL;
		}

		/*
		 * Check to see that the fragment is still at the local site.
		 * If so, just update the tuple and mark the entry as used;
		 * otherwise, remove the tuple and go to the next tuple (the
		 * new information will be inserted into the pg_access table below)
		 */
		if (ents[i].siteId != localSiteHostId) {
		    /* XXX - is this really necessary? */
		    ItemPointerCopy(&htup->t_ctid, &tmp_ipd);
		    heap_delete(desc, &tmp_ipd);
		    continue;
		}
		else {
		    replaces[Anum_pg_has_fralocation - 1] = 'r';
		    values[Anum_pg_has_fralocation - 1] = (Datum)ents[i].siteId;
		}
		if (ents[i].dbName.data[0] != '\0') {
		    if (namecmp(&(ents[i].dbName), &(tuple->datname))) {
			namecpy(&tmp_dbn, &(ents[i].dbName));
			replaces[Anum_pg_has_datname - 1] = 'r';
			values[Anum_pg_has_datname - 1] = (Datum)&tmp_dbn;
		    }
		}
		replaces[Anum_pg_has_fradebits - 1] = 'r';
		values[Anum_pg_has_fradebits - 1] = (Datum)ents[i].debits;
		replaces[Anum_pg_has_fracredits - 1] = 'r';
		values[Anum_pg_has_fracredits - 1] = (Datum)ents[i].credits;
		if (ents[i].expectedPrice != 0) {
		    replaces[Anum_pg_has_fraexpprice - 1] = 'r';
		    values[Anum_pg_has_fraexpprice - 1] = (Datum)ents[i].expectedPrice;
		}
		replaces[Anum_pg_has_fraaccesses - 1] = 'r';
		values[Anum_pg_has_fraaccesses - 1] = (Datum)ents[i].numAccesses;
		if (ents[i].arrivalTime != 0) {
		    replaces[Anum_pg_has_fraarrtime - 1] = 'r';
		    values[Anum_pg_has_fraarrtime - 1] = (Datum)ents[i].arrivalTime;
		}
		replaces[Anum_pg_has_fralastacctime - 1] = 'r';
		values[Anum_pg_has_fralastacctime - 1] = (Datum)ents[i].lastAccessTime;
		/*
		 * Modify the heap tuple.
		 */
		htup = heap_modifytuple(htup, buffer, desc, values, nulls, replaces);
		/* XXX - is this really necessary? */
		ItemPointerCopy(&htup->t_ctid, &tmp_ipd);
		(void)heap_replace(desc, &tmp_ipd, htup);
		ent_used[i] = (char)1;
		insHas++;
	    }
        }
    } while (htup);
    heap_endscan(hscan);
    heap_close(desc);

    desc = heap_openr((Name)Name_pg_access);
    if (!RelationIsValid(desc))
        elog(WARN, "USEL_WriteDataToCatalogs: cannot open pg_access");

    hscan = heap_beginscan(desc, 0, SelfTimeQual, 0, NULL);
    do {
        htup = heap_getnext(hscan, 0, &buffer);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_access tuple = (Form_pg_access)GETSTRUCT(htup);
	    Oid fragId = tuple->fralogicalid;

	    for (i = 0; i < len; i++)
	        if (ents[i].fragmentId == fragId)
		    break;
	    /*
	     * If we found a "matching" entry, we have to jump through 
	     * a few hoops to "update" the tuple.
	     */
	    if (i != len) {
	        int j;
		Datum values[Natts_pg_access];
		char nulls[Natts_pg_access];
		char replaces[Natts_pg_access];
		ItemPointerData tmp_ipd;
		NameData tmp_dbn;

		/*
		 * Initialize the arrays used to update the tuple data
		 */
		for (j = 0; j < Natts_pg_access; j++) {
		    replaces[j] = ' ';
		    nulls[j] = ' ';
		    values[j] = (Datum)NULL;
		}

		/*
		 * Check to see that the fragment is still at a remote site.
		 * If so, just update the tuple and mark the entry as used;
		 * otherwise, remove the tuple and go to the next tuple (the
		 * new information will be inserted into the pg_has table below)
		 */
		if (ents[i].siteId == localSiteHostId) {
		    /* XXX - is this really necessary? */
		    ItemPointerCopy(&htup->t_ctid, &tmp_ipd);
		    heap_delete(desc, &tmp_ipd);
		    continue;
		}
		else {
		    replaces[Anum_pg_access_fralocation - 1] = 'r';
		    values[Anum_pg_access_fralocation - 1] = (Datum)ents[i].siteId;
		}

		if (ent_used[i] != (char)0)
		    elog(DEBUG,
			 "USEL_WriteDataToCatalogs: multiple entries for fragment %d in pg_access?",
			 (int)fragId);
		
		if (ents[i].dbName.data[0] != '\0') {
		    if (namecmp(&(ents[i].dbName), &(tuple->datname))) {
			namecpy(&tmp_dbn, &(ents[i].dbName));
			replaces[Anum_pg_access_datname - 1] = 'r';
			values[Anum_pg_access_datname - 1] = (Datum)&tmp_dbn;
		    }
		}
		replaces[Anum_pg_access_fradebits - 1] = 'r';
		values[Anum_pg_access_fradebits - 1] = (Datum)ents[i].debits;
		replaces[Anum_pg_access_fracredits - 1] = 'r';
		values[Anum_pg_access_fracredits - 1] = (Datum)ents[i].credits;
		replaces[Anum_pg_access_fraexpprice - 1] = 'r';
		if (ents[i].expectedPrice != 0) {
		    replaces[Anum_pg_access_fraexpprice - 1] = 'r';
		    values[Anum_pg_access_fraexpprice - 1] = (Datum)ents[i].expectedPrice;
		}
		values[Anum_pg_access_fraaccesses - 1] = (Datum)ents[i].numAccesses;
		if (ents[i].arrivalTime != 0) {
		    replaces[Anum_pg_access_fraarrtime - 1] = 'r';
		    values[Anum_pg_access_fraarrtime - 1] = (Datum)ents[i].arrivalTime;
		}
		replaces[Anum_pg_access_fralastacctime - 1] = 'r';
		values[Anum_pg_access_fralastacctime - 1] = (Datum)ents[i].lastAccessTime;
		/*
		 * Modify the heap tuple.
		 */
		htup = heap_modifytuple(htup, buffer, desc, values, nulls, replaces);
		/* XXX - is this really necessary? */
		ItemPointerCopy(&htup->t_ctid, &tmp_ipd);
		(void)heap_replace(desc, &tmp_ipd, htup);
		ent_used[i] = (char)1;
		insAcc++;
	    }
        }
    } while (htup);
    heap_endscan(hscan);
    heap_close(desc);

    /*
     * Now, we scan through the ent_used list to see which entries were
     * not found (or their location changed) and need to be inserted
     * into the pg_has or pg_access tables.
     */
    has = (int *)calloc(insHas, sizeof(int));
    Assert(has != NULL);
    access = (int *)calloc(insAcc, sizeof(int));
    Assert(access != NULL);
    h_idx = 0;
    a_idx = 0;

    for (i = 0; i < len; i++) {
        if (!ent_used[i]) {
	    if (ents[i].siteId == localSiteHostId)
	        has[h_idx++] = i;
	    else
	        access[a_idx++] = i;
	}
    }

    if (insHas) {
        desc = heap_openr((Name)Name_pg_has);
	if (!RelationIsValid(desc))
	    elog(WARN, "USEL_WriteDataToCatalogs: cannot open pg_has");

	for (i = 0; i < h_idx; i++) {
	    Datum values[Natts_pg_has];
	    char nulls[Natts_pg_has];
	    NameData tmp_dbn;
	    HeapTuple tup;
	    int j;

	    for (j = 0; j < Natts_pg_has; j++) {
	        values[j] = (Datum)NULL;
		nulls[j] = ' ';
	    }
	    (void)memset((void *)&(tmp_dbn), 0, sizeof(NameData));
	    namecpy(&(tmp_dbn), &(ents[has[i]].dbName));
	    values[Anum_pg_has_datname - 1] = (Datum)&tmp_dbn;
	    values[Anum_pg_has_fralogicalid - 1] = (Datum)ents[has[i]].fragmentId;
	    values[Anum_pg_has_fralocation - 1] = (Datum)ents[has[i]].siteId;
	    values[Anum_pg_has_fradebits - 1] = (Datum)ents[has[i]].debits;
	    values[Anum_pg_has_fracredits - 1] = (Datum)ents[has[i]].credits;
	    values[Anum_pg_has_fraexpprice - 1] = (Datum)ents[has[i]].expectedPrice;
	    values[Anum_pg_has_fraaccesses - 1] = (Datum)ents[has[i]].numAccesses;
	    values[Anum_pg_has_fraarrtime - 1] = (Datum)ents[has[i]].arrivalTime;
	    values[Anum_pg_has_fralastacctime - 1] = (Datum)ents[has[i]].lastAccessTime;
	    tup = heap_formtuple(Natts_pg_has, desc->rd_att, values, nulls);
	    heap_insert(desc, tup);
	    pfree(tup);
	    /* XXXRHPXXX - sanity checking */
	    ent_used[i] = (char)1;
	    insHas--;
	}
	if (insHas)
	    elog(DEBUG, "USEL_WriteDataToCatalogs: insHas = %d instead of 0", insHas);

	heap_close(desc);
    }

    if (insAcc) {
        desc = heap_openr((Name)Name_pg_access);
	if (!RelationIsValid(desc))
	    elog(WARN, "USEL_WriteDataToCatalogs: cannot open pg_access");

	for (i = 0; i < a_idx; i++) {
	    Datum values[Natts_pg_access];
	    char nulls[Natts_pg_access];
	    NameData tmp_dbn;
	    HeapTuple tup;
	    int j;

	    for (j = 0; i < Natts_pg_access; j++) {
	        values[j] = (Datum)NULL;
		nulls[j] = ' ';
	    }
	    (void)memset((void *)&(tmp_dbn), 0, sizeof(NameData));
	    namecpy(&(tmp_dbn), &(ents[access[i]].dbName));
	    values[Anum_pg_access_datname - 1] = (Datum)&tmp_dbn;
	    values[Anum_pg_access_fralogicalid - 1] = (Datum)ents[access[i]].fragmentId;
	    values[Anum_pg_access_fralocation - 1] = (Datum)ents[access[i]].siteId;
	    values[Anum_pg_access_fradebits - 1] = (Datum)ents[access[i]].debits;
	    values[Anum_pg_access_fracredits - 1] = (Datum)ents[access[i]].credits;
	    values[Anum_pg_access_fraexpprice - 1] = (Datum)ents[access[i]].expectedPrice;
	    values[Anum_pg_access_fraaccesses - 1] = (Datum)ents[access[i]].numAccesses;
	    values[Anum_pg_access_fraarrtime - 1] = (Datum)ents[access[i]].arrivalTime;
	    values[Anum_pg_access_fralastacctime - 1] = (Datum)ents[access[i]].lastAccessTime;
	    tup = heap_formtuple(Natts_pg_access, desc->rd_att, values, nulls);
	    heap_insert(desc, tup);
	    pfree(tup);
	    /* XXXRHPXXX - sanity checking */
	    ent_used[i] = (char)1;
	    insAcc--;
	}
	if (insAcc)
	    elog(DEBUG, "USEL_WriteDataToCatalogs: insAcc = %d instead of 0", insAcc);

	heap_close(desc);
    }
    return;
}
