/*-------------------------------------------------------------------------
 *
 * usage_stat.c--
 *    MARIPOSA fragment usage statistics manipulation routines
 *
 * Copyright (c) 1994,1995 Regents of the University of California
 *
 * $Id
 * [missing RCS id]
 *-------------------------------------------------------------------------
 */

#include <string.h>
#include "postgres.h"
#include "storage/ipc.h"
#include "storage/shmem.h"
#include "storage/spin.h"
#include "storage/usage_stat.h"
#include "storage/us_internals.h"
#include "utils/hsearch.h"
#include "utils/memutils.h"
#include "utils/elog.h"

bool NoUsageStats = FALSE;
SPINLOCK UsageStatLock;

/*
 * The first and last entries on the list will change locations,
 * rather than shuffling the data structures, so they must be
 * kept in shared memory.
 *
 * Since the postmaster sets up the shared memory area (under normal
 * circumstances, anyway) and the data broker (aka. storage manager)
 * initializes the region by reading in the initial values from the
 * system catalogs, we need some way of determining when the area is
 * ready for use.  Therefore, we keep a flag in shared memory for this
 * purpose. -rhp
 */
SHMEM_OFFSET *ShmemUsageStatInitialized = NULL;
SHMEM_OFFSET *ShmemUsageStatNumStatStructsUsed = NULL;
SHMEM_OFFSET *ShmemUsageStatFirstEntryOffset = NULL;
SHMEM_OFFSET *ShmemUsageStatLastEntryOffset = NULL;
SHMEM_OFFSET *ShmemUsageStatHotListNumEntries = NULL;

/*
 * Once these have been initialized from shared memory, there values
 * should never change.  However, make sure that you have the appropriate
 * lock before dereferencing them. -rhp
 */
unsigned long *UsageStatInitialized = NULL;
unsigned long *UsageStatNumStatStructsUsed = NULL;
unsigned long *UsageStatHotListNumEntries = NULL;
HTAB *UsageStatTable = NULL;
UsageStatHotListEntry *UsageStatHotList = NULL;
UsageStatEntry *UsageStatEntryList = NULL;
SHMEM_OFFSET *UsageStatEntryListHead = NULL;
SHMEM_OFFSET *UsageStatEntryListTail = NULL;

/*
 * Macros for manipulating the EntryList.  These are macros
 * because we really want inline code but I want to factor
 * this stuff out so I only have to get it right once. -rhp
 */
#define USEL_InsertEntryAtTopOfList(ptr,offset) {\
	    UsageStatEntry *old_first = \
		(UsageStatEntry *)MAKE_PTR(*ShmemUsageStatFirstEntryOffset); \
	    \
	    Assert(ShmemIsValid((unsigned long)old_first)); \
	    old_first->prev = offset; \
	    ptr->next = *ShmemUsageStatFirstEntryOffset; \
	    *ShmemUsageStatFirstEntryOffset = offset; \
	}

#define USEL_RemoveEntryFromBottomOfList(ptr) { \
	    UsageStatEntry *old_prev \
		= (UsageStatEntry *)MAKE_PTR(ptr->prev); \
	    \
	    Assert(ShmemIsValid((unsigned long)old_prev)); \
	    old_prev->next = INVALID_OFFSET; \
	    *ShmemUsageStatLastEntryOffset = ptr->prev; \
	}

#define USEL_RemoveEntryFromMiddleOfList(ptr) {\
	    UsageStatEntry *old_prev = \
		(UsageStatEntry *)MAKE_PTR(ptr->prev); \
	    UsageStatEntry *old_next = \
		(UsageStatEntry *)MAKE_PTR(ptr->next); \
	    \
	    Assert(ShmemIsValid((unsigned long)old_prev)); \
	    Assert(ShmemIsValid((unsigned long)old_next)); \
	    old_prev->next = ptr->next; \
	    old_next->prev = ptr->prev; \
        }

#define USEL_CopyEntry(existing,new) { \
	    Assert(existing != NULL && new != NULL); \
	    memcpy((void *)new, (void *)existing, \
		   sizeof(UsageStatEntry)); \
	}

#define USEL_CopyList(existing,new) { \
	    Assert(new != NULL && existing != NULL); \
	    memcpy((void *)new, (void *)existing, \
		   (STABLE_NENTRIES * sizeof(UsageStatEntry))); \
	}

#define USHL_CopyEntry(existing,new) { \
	    Assert(new != NULL && existing != NULL); \
	    memcpy((void *)new, (void *)existing, \
		   sizeof(UsageStatHotListEntry)); \
	}

static void UST_Init(void);
static void USHL_Init(void);
static void USEL_Init(void);
static void USEL_UpdateList(UsageStatEntry *);
static UsageStatEntry *UST_LookupEntry(Oid, bool *);
static UsageStatEntry *UST_ReplaceEntry(Oid, Oid);
static void USEL_UpdateEntry(UsageStatEntry *, UsageStatEntry *, bool *);



/***************************************************************************
 *                    Exported Interface functions                         *
 ***************************************************************************/

/* -------------------------------------------------------
 * UsageStatShmemSize
 *
 * fashioned after BufferShmemSize(), this function computes
 * the size of shared memory used for all the usage statistics
 * including the offset of the UsageStatTable, the UsageStatTable,
 * the UsageHotList, the head and tail offsets of the global
 * linked list, and the array of fragment usage statistics data 
 * structures. -rhp
 */
int
UsageStatShmemSize()
{
    int size = 0;
    int nbuckets;
    int nsegs;
    int tmp;

    nbuckets = 1 << (int)my_log2((STABLE_NENTRIES - 1) / DEF_FFACTOR + 1);
    nsegs = 1 << (int)my_log2((nbuckets - 1) / DEF_SEGSIZE + 1);

    size = 5 * MAXALIGN(sizeof(SHMEM_OFFSET));

    /* size of the shmem usage statistics hash table */
    size += MAXALIGN(my_log2(STABLE_SIZE) * sizeof(void *));
    size += MAXALIGN(sizeof(HHDR));
    size += MAXALIGN(DEF_SEGSIZE * sizeof(SEGMENT));
    size += BUCKET_ALLOC_INCR *
        (MAXALIGN(sizeof(BUCKET_INDEX)) +
	 MAXALIGN(STABLE_KEYSIZE) +
	 MAXALIGN(STABLE_DATASIZE));

    /* size of the hotlist */
    size += SHOTLIST_NENTRIES *
        MAXALIGN(sizeof(UsageStatHotListEntry));

    /* size of the actual usage statistics data structures */
    size += STABLE_NENTRIES *
        MAXALIGN(sizeof(UsageStatEntry));

    return size;
}

void
UsageStatsInitShmem()
{
    bool foundHotList, foundStatBuffers;
    int i;

    UsageStatAcquireExLock(UsageStatLock);
    UsageStatInitialized =
      (unsigned long *)MAKE_PTR(ShmemUsageStatInitialized);
    Assert(ShmemIsValid((unsigned long)UsageStatInitialized));

    /*
     * We still need to make sure that everything is set up properly
     * in our address space even if the Usage Statistics are 
     * initialized. -rhp
     */
    UST_Init();
    USHL_Init();
    USEL_Init();
    UsageStatReleaseExLock(UsageStatLock);
    return;
}

void
UsageStatUpdateEntry(UsageStatEntry *info)
{
    Oid fragId;
    UsageStatEntry *fragData;
    bool status;

    UsageStatAcquireExLock(UsageStatLock);
    if (!*UsageStatInitialized)
        elog(FATAL, "UsageStatUpdateEntry: Usage statistics not initialized!");

    if (info != NULL) {
        fragId = info->fragmentId;
	fragData = UST_LookupEntry(fragId, &status);
	USEL_UpdateEntry(fragData, info, &status);
	Assert(!status);
    }
    else {
        elog(WARN, "UsageStatUpdateEntry: NULL fragment info");
    }
    UsageStatReleaseExLock(UsageStatLock);
    return;
}

void
UsageStatUpdateEntries(List *ents)
{
    List *lp;
    int len;

    if (length(ents)) {
        UsageStatAcquireExLock(UsageStatLock);
        if (!*UsageStatInitialized)
	    elog(FATAL, "UsageStatUpdateEntries: Usage statistics not initialized!");

	foreach(lp, ents) {
	    UsageStatEntry *shmemData,*fragData = (UsageStatEntry *)lfirst(lp);
	    Oid fragId = fragData->fragmentId;
	    bool status = FALSE;

	    shmemData = UST_LookupEntry(fragId, &status);
	    USEL_UpdateEntry(shmemData, fragData, &status);
	    Assert(!status);
	}
	UsageStatReleaseExLock(UsageStatLock);
    } else {
        elog(DEBUG, "UsageStatUpdateEntries: No fragments to update");
    }
    return;
}

void
UsageStatGetHotListEntry(Oid fragId, UsageStatHotListEntry *out, bool *found)
{
    int i, len;

    Assert(out != NULL && found != NULL);
    UsageStatAcquireExLock(UsageStatLock);

    if (!*UsageStatInitialized)
        elog(FATAL, "UsageStatGetHotListEntry: Usage statistics not initialized!");

    len = *UsageStatHotListNumEntries;
    *found = FALSE;
    for (i = 0; i < len; i++) {
        if (UsageStatHotList[i].fragmentId == fragId) {
	    USHL_CopyEntry((&(UsageStatHotList[i])),out);
	    *found = TRUE;
	    break;
	}
    }
    UsageStatReleaseExLock(UsageStatLock);
    return;
}



/****************************************************************************
 *                         Internal functions                               *
 ****************************************************************************/

static void
UST_Init()
{
    HASHCTL info;
    int hash_flags;
    bool found;
    long *location;

    /*
     * We are using oids as the hash key so we want to use 
     * tag_hash rather than string_hash. -rhp
     */
    info.hash = tag_hash;
    info.keysize = STABLE_KEYSIZE;
    info.datasize = STABLE_DATASIZE;
    info.segbase = (long *)ShmemBase;
    info.alloc = ShmemAlloc;
    info.max_size = STABLE_SIZE;
    hash_flags = (HASH_ELEM | HASH_FUNCTION | HASH_SHARED_MEM);

    /* look it up in the binding table */
    location = ShmemInitStruct("UsageStatTable",
	my_log2(info.max_size) + sizeof(HHDR), &found);

    if (location == NULL) {
	elog(FATAL,"UST_Init: binding table is corrupt");
	return;
    }

    /*
     * If it already exists, attach to it rather than allocating
     * and initializing new space.
     */
    if (found) {
        hash_flags |= HASH_ATTACH;
    }

    /* these structures were allocated or bound in ShmemInitStruct */
    info.hctl = (long *)location;
    info.dir = (long *)(location + sizeof(HHDR));

    /* create/attach to the hash table */
    UsageStatTable = hash_create(STABLE_SIZE, &info, hash_flags);
    
    if (! UsageStatTable) {
	elog(FATAL,"UST_Init: couldn't initialize UsageStatTable");
	return;
    }
    Assert(ShmemIsValid((unsigned long)UsageStatTable->hctl));
    return;
}

static void
USHL_Init()
{
    bool found;
    long *location;
    unsigned long size;
    
    size = SHOTLIST_NENTRIES * MAXALIGN(sizeof(UsageStatHotListEntry));
    location = ShmemInitStruct("UsageStatHotList", size, &found);
    
    if (location == NULL) {
        elog(FATAL,"USHL_Init: binding table is corrupt");
	return;
    }
    if (!found)
      memset((void *)location, 0, (size_t)size);
    UsageStatHotList = (UsageStatHotListEntry *)location;
    UsageStatHotListNumEntries =
        (SHMEM_OFFSET *)MAKE_PTR(ShmemUsageStatHotListNumEntries);
    Assert(ShmemIsValid((unsigned long)UsageStatHotListNumEntries));
    return;
}

static void
USEL_Init()
{
    int i;
    bool found;
    long *location;
    unsigned long size;

    UsageStatNumStatStructsUsed = 
        (unsigned long *)MAKE_PTR(ShmemUsageStatNumStatStructsUsed);
    Assert(ShmemIsValid((unsigned long)UsageStatNumStatStructsUsed));

    size = STABLE_NENTRIES * MAXALIGN(sizeof(UsageStatEntry));
    location = ShmemInitStruct("UsageStatEntryList", size, &found);

    if (location == NULL) {
        elog(FATAL,"USEL_Init: binding table is corrupt");
	return;
    }
    UsageStatEntryList = (UsageStatEntry *)location;
    UsageStatEntryListHead =
        (SHMEM_OFFSET *)MAKE_PTR(ShmemUsageStatFirstEntryOffset);
    Assert(ShmemIsValid((unsigned long)UsageStatEntryListHead));
    UsageStatEntryListTail =
        (SHMEM_OFFSET *)MAKE_PTR(ShmemUsageStatLastEntryOffset);
    Assert(ShmemIsValid((unsigned long)UsageStatEntryListTail));
    return;
}

static void
USEL_UpdateList(UsageStatEntry *ent)
{
    SHMEM_OFFSET current;

    Assert(ShmemIsValid((unsigned long)ent));
    current = (SHMEM_OFFSET)MAKE_OFFSET(ent);

    if (ent->next == INVALID_OFFSET) {
        if (ent->prev == INVALID_OFFSET) { /* Entry is new to the list */
	    if (*ShmemUsageStatFirstEntryOffset == INVALID_OFFSET) {
	        /* First entry to be added to the list */
	        *ShmemUsageStatFirstEntryOffset = current;
                *ShmemUsageStatLastEntryOffset = current;
	    }
	    else {
	        USEL_InsertEntryAtTopOfList(ent,current);
	    }
	}
	else {
	    USEL_RemoveEntryFromBottomOfList(ent);
	    USEL_InsertEntryAtTopOfList(ent,current);
	}
    }
    else {
        if (ent->prev != INVALID_OFFSET) {
	    USEL_RemoveEntryFromMiddleOfList(ent);
	    USEL_InsertEntryAtTopOfList(ent,current);
	    return;
	}
    }
    return;
}

/*
 * This method looks up the offset for a fragment oid in the hash table
 * and returns a pointer to the UsageStatEntry.  If the oid is not
 * found in the hash table AND the the number of UsageStatEntries currently
 * in use is less than the number allocated in shared memory, a new entry
 * is added and initialized to point to the UsageStatEntry to use. Otherwise,
 * if the entry is not found and shared memory if full, we return NULL.
 *
 * It is the caller's job to make sure that the UsageStatEntry gets initialized,
 * which includes the next and prev entries and updating the global linked list.
 */

static UsageStatEntry *
UST_LookupEntry(Oid fragId, bool *found)
{
    UsageStatEntry *retval = NULL;
    UsageHashTableEntry *ent = NULL;
    HASHACTION action;
    int hash_flags, idx;

    if (*UsageStatNumStatStructsUsed < STABLE_NENTRIES)
        action = HASH_ENTER;
    else
        action = HASH_FIND;

    ent = (UsageHashTableEntry *)hash_search(UsageStatTable,
				      (char *)&fragId, action, found);
    if (ent == NULL)
        elog(FATAL, "UST_LookupEntry: UsageHashTable corrupted");
    else if (ent == (UsageHashTableEntry *)TRUE) {
        if (action == HASH_FIND)
	    /* 
	     * The entry was not in the hash table and there is no
	     * empty UsageStatStructs so return NULL and let the
	     * the caller figure out what to do.
	     */
	    return NULL;
	else
	    /*
	     * The entry was not in the table and it was not added,
	     * even though we told hash_search() to add it, so we
	     * need to exit here (if we ever gets here).
	     */
	    elog(FATAL, "UST_LookupEntry: Could not create new UsageHashTable entry");
    }
    else {
        if (!found) {
            if (action == HASH_ENTER) {
                /*
		 * Set the value of the new hash table entry to the offset of the
		 * next free UsageStatStruct.
		 *
		 * NOTE: We are assuming that the structures are allocated contiguously
		 *       so don't do anything to break this assumption or all bets are off.
		 *       -rhp
		 */
	        idx = *UsageStatNumStatStructsUsed;
		ent->location = (SHMEM_OFFSET)MAKE_OFFSET(&UsageStatEntryList[idx]);
		*UsageStatNumStatStructsUsed++;
		retval = (UsageStatEntry *)MAKE_PTR(ent->location);
		Assert(ShmemIsValid((unsigned long)retval));

		/*
		 * Make sure that everything in the new structure is zero.
		 *
		 * NOTE: Only zero out newly allocated structures (i.e., ones
		 *       that were not previously in the list) since otherwise
		 *       we need the prev and next values.
		 */
		memset((void *)retval, 0, sizeof(UsageStatEntry));
		retval->prev = INVALID_OFFSET;
		retval->next = INVALID_OFFSET;
	    }
	    else {
	        elog(FATAL, "UST_LookupEntry: Reached unreachable code");
	    }
	}
	else {
	    /*
	     * This is an old entry, just turn it into a pointer and return.
	     */
	    retval = (UsageStatEntry *)MAKE_PTR(ent->location);
	    Assert(ShmemIsValid((unsigned long)retval));
	}
    }
    return retval;
}

/*
 * This routine basically changes the key for a particular structure in
 * the UsageHashTable from old_frag to new_frag, leaving the data untouched.
 * Then, it finds the structure to which the data refers.  It is the caller's
 * job to reset the values in the UsageStatEntry and update the global list.
 */
static UsageStatEntry *
UST_ReplaceEntry(Oid old_frag, Oid new_frag)
{
    UsageHashTableEntry *ent;
    long *tmp_ent;
    SHMEM_OFFSET old_ent;
    UsageStatEntry *retval = NULL;
    bool found;

    /*
     * Find and save the state of the entry for old_frag.
     */
    tmp_ent = hash_search(UsageStatTable, (char *)&old_frag,
			  HASH_FIND_SAVE, &found);

    if (tmp_ent == NULL)
        elog(FATAL, "UST_ReplaceEntry: UsageHashTable corrupted");

    if (tmp_ent == (long *)TRUE)
        elog(FATAL, "UST_ReplaceEntry: Entry %d not found", old_frag);

    ent = (UsageHashTableEntry *)tmp_ent;
    old_ent = ent->location;

    /*
     * Remove the entry previously found.
     */
    tmp_ent = hash_search(UsageStatTable, (char *)&old_frag,
			  HASH_REMOVE_SAVED, &found);

    if (tmp_ent == NULL)
        elog(FATAL, "UST_ReplaceEntry: UsageHashTable corrupted");

    if (tmp_ent == (long *)TRUE)
        elog(FATAL, "UST_ReplaceEntry: Previously saved entry %d not found", old_frag);

    /*
     * Insert the new entry.
     */
    tmp_ent = hash_search(UsageStatTable, (char *)&new_frag, HASH_ENTER, &found);

    if (tmp_ent == NULL)
        elog(FATAL, "UST_ReplaceEntry: UsageHashTable corrupted");
    if (found)
        elog(WARN, "UST_ReplaceEntry: New entry %d found unexpectedly", new_frag);

    /*
     * Do not reset the values in the structure we're returning because the
     * caller relies on this information to fix up the list.
     */
    ent = (UsageHashTableEntry *)tmp_ent;
    ent->location = old_ent;
    retval = (UsageStatEntry *)MAKE_PTR(ent->location);
    Assert(ShmemIsValid((unsigned long)retval));
    return retval;
}

/*
 * This routine coordinates all the steps necessary to update the stats
 * for a fragment and maintain consistency in the data structures.
 */
static void
USEL_UpdateEntry(UsageStatEntry *ent, UsageStatEntry *info, bool *status)
{
    Oid fragId;
    UsageStatEntry *prev, *next, *first;
    SHMEM_OFFSET current, old_prev, old_next, old_first;

    *status = TRUE;  /* the error condition */
    Assert(info != NULL);
    fragId = info->fragmentId;
    if (ent == NULL) {
        /*
	 * This means that the fragment is not currently in memory AND
	 * the shared memory is full.  Therefore, we need to find the
	 * fragment oid of the last entry and call UsageHashTableReplace().
	 */
        Oid old_fragId;
	UsageStatEntry *old_frag =
	    (UsageStatEntry *)MAKE_PTR(*UsageStatEntryListTail);

	Assert(ShmemIsValid((unsigned long)old_frag));
	old_fragId = old_frag->fragmentId;
	ent = UST_ReplaceEntry(old_fragId, fragId);

	Assert(ent->next == INVALID_OFFSET);
	ent->fragmentId = fragId;

	/* Zero out any old data, leaving the fragmentId, next, and prev untouched */
	USAGE_STAT_ENTRY_BZERO_DATA(ent);

	/*
	 * At this point, the replaced structure should have the correct oid,
	 * have prev set to point to the next to last entry in the list,
	 * have next set to INVALID_OFFSET, and have all other values set to 0.
	 */
    }
    USEL_UpdateEntryData(ent, info);
    USEL_UpdateList(ent);
    *status = FALSE;     /* Success */
    return;
}
