/*-------------------------------------------------------------------------
 *
 * broker.c--
 *    
 *    The Broker module - takes a fragmented query plan and fills in the
 *    processing sites.  Returns an array of sub-plans.  Each array element
 *    is a plan processed by one site.  Two protocols: long and short.
 *
 *    Long protocol:
 *    Sends out requests for bids to bidding sites.  Receives the bids in
 *    and decides when to terminate the bidding process.  Selects the sites
 *    that will process the query parts and puts them into the query
 *    plan.  Notifies loser sites and returns.
 *
 *    Short protocol:
 *    No bidding - broker chooses processing sites, fills them in and returns.
 *
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/tcop/broker.c,v 1.42 1996/05/31 17:14:34 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include "tcop/broker.h"
#include "tcop/brokerLong.h"

/*
 *-------------------------------------------------------------------------
 * Forward declarations.
 *-------------------------------------------------------------------------
 */
void ShortProtocol(List *plans, QueryDesc *qd);
void ShortAssignProcessingSites(Plan *plan, QueryDesc *qd);
void ShortAssignResultSites(QueryDesc *qd);
void ShortSeqScan(SeqScan *scan, List *rtable);
void SetSeqScanSite(SeqScan *scan, RangeTblEntry *rte);
void ShortMergeJoin(MergeJoin *plan, List *rtable);
List *CreatePlanChunkList(Plan *plan);
List *DefinePlanLeaves(Plan *plan);
Plan *CreatePlanChunk(Plan *root, List *leaves);
void CopyPlanChunks(List *bPlans);
void FillInBlankNodes(Plan *plan, Plan *parentPlan, SiteAddr *homeSite);
void FillInQepFragIds(Plan *plan, int currentQepFragId);
FragInfo *DataFragInfo(Oid fralogicalid);
SiteAddrNode *RandomSite(Plan *plan, List *rtable);
void BrokerDebug(int lev, char *fmt, ...);
void printPlanChunks(List *bPlans, List *rtable);

/*
 * chunksize is set using the SET CHUNKSIZE command.
 * It's an integer between 0 and 100, much like parProb in
 * the fragmenter.  It determines the size of plan chunks
 * sent out by the broker.  At each point where the broker
 * could cut off a plan chunk, it generates a random number
 * between 0 and 100, inclusive.  If the number is less than
 * 100 - chunksize, the plan is cut off at that point.  If it's
 * greater than or equal to chunksize, it doesn't.  Setting
 * chunksize to 100 means the whole plan will be sent out as
 * one chunk.  Setting it to 0 means each node will be a
 * separate chunk.
 */
int chunksize = 0;

#define MAKECHUNK ((int)((float)(rand()/(RAND_MAX/100)))) < (100 - chunksize)


/*
 * Jeff PDIS code XXX - remove me.
 */
SiteAddrNode allAddrs[5];



/*
 *-------------------------------------------------------------------------
 * Broker--
 *
 * Shell process that calls the routine to assign a processing
 * site recursively to each node in the plan tree.
 *-------------------------------------------------------------------------
 */
void
Broker(QueryDesc *qd, List *plans, int protocol, SiteAddr *homeSiteAddr) {
    
    extern bool demo;
    List *bPlans = NIL;
    List *lp;
    int saveChunksize;

    /*
     * For the short protocol, keep the plan intact - don't
     * divide it into chunks.  Otherwise, for joins, we don't
     * have any information about the children nodes - makes
     * it difficult to figure out where to process the join.
     */

    if (protocol == SHORT_PROCESS_QUERY) {
	saveChunksize = chunksize;
	chunksize = 100;
    }
    
    /*
     * Take fragmented plan and divide it into chunks.
     */
    foreach (lp, plans) {
	Plan *plan = (Plan *)lfirst(lp);
	bPlans = nconc(CreatePlanChunkList(plan), bPlans);
    }
    
    if (demo) {
	printPlanChunks(bPlans, qd->parsetree->rtable);
    }
    
    /*
     * Assign processing sites to the plans, using the
     * protocol specified.
     */
    switch (protocol) {
    case SHORT_PROCESS_QUERY: {
	ShortProtocol(bPlans, qd);
    }
	break;

    case LONG_PROCESS_QUERY: {
	LongProtocol(bPlans, protocol, qd, homeSiteAddr);
    }
	break;

    default: {
    }
	
    }
    
    /*
     * Put the processing sites into the original plan.
     */
    CopyPlanChunks(bPlans);

    /*
     * Fill in processing sites in nodes which were not bid out,
     * such as merge nodes.
     */
    
    foreach (lp, plans) {
	Plan *plan = (Plan *)lfirst(lp);
	FillInBlankNodes(plan, NULL, homeSiteAddr);
    }
    
    /*
     * Fill in qepFragId in each node.
     */

    if (qd->operation == CMD_SELECT) {
	foreach (lp, plans) {
	    Plan *plan = (Plan *)lfirst(lp);
	    FillInQepFragIds(plan, 0);
	}
    } else {
	foreach (lp, plans) {
	    Plan *plan = (Plan *)lfirst(lp);
	    FillInQepFragIds(plan, newQepFragId());
	}
    }
    
    if (protocol == SHORT_PROCESS_QUERY) {
	chunksize = saveChunksize;
    }

}


/*
 *======================================================================================
 * Short Protocol--
 *
 * Call the appropriate routine to assign a processing site
 * to each node passed in, based on the node type.  The short
 * protocol, in order to save on communication cost, does not
 * contact the processing site it chooses.  Therefore, there is
 * a chance that the processing site can refuse to do the work.
 * This will have to be dealt with by the coordinator.  The other
 * alternative would be for the Broker to contact each site
 * during the short protocol, but that would defeat the purpose
 * of having a short protocol.
 *======================================================================================
 */
void
ShortProtocol(List *plans, QueryDesc *qd)
{
    List *lp;
    BrokerPlan *bp;
    extern int chunksize;
    
    /*
     * Assign the processing sites for each plan chunk.
     */
    foreach (lp, plans) {
	bp = (BrokerPlan *) lfirst(lp);
	ShortAssignProcessingSites(bp->planChunk, qd);
    }

    /*
     * For INSERT queries, assign the sites for the result relations.
     */
    if (qd->operation == CMD_INSERT) {
	ShortAssignResultSites(qd);
    }
    
}

/*
 *-------------------------------------------------------------------------
 * ShortAssignResultSites--
 *
 * Fill in RangeTblEntry->FragInfo->ProcessYN for the RangeTblEntry
 * that is the result relation.  ProcessYN is FALSE for each fragment
 * except the one which will process the INSERT.
 *-------------------------------------------------------------------------
 */
void
ShortAssignResultSites(QueryDesc *qd) {
    int resultIndex;
    RangeTblEntry *rte;
    SiteAddrNode *sAddr;
    List *l;
    
    resultIndex = qd->parsetree->resultRelation;
    Assert(resultIndex > 0);
    
    rte = (RangeTblEntry *)nth(resultIndex-1, qd->parsetree->rtable);

    /*
     * Set all the processYN flags to FALSE, then select one
     * and set it to TRUE.
     */
    foreach(l, rte->fragInfo) {
	FragInfo *fInfo = (FragInfo *)lfirst(l);
	List *l2;
	int nSites = length(fInfo->processYN);
	int whichSite = (random() % nSites);
	int i = 0;
	
	foreach(l2, fInfo->processYN) {
	    if (i == whichSite) {
		l2->elem = (void *)TRUE;
	    } else {
		l2->elem = (void *)FALSE;
	    }
	}
	
    }
    
}

/*
 *-------------------------------------------------------------------------
 * ShortAssignProcessingSites--
 *
 * Visit the nodes in the plan in post-order (children first, then
 * parent), assigning the processing sites.
 *-------------------------------------------------------------------------
 */
void
ShortAssignProcessingSites(Plan *plan, QueryDesc *qd) {

    if (plan->lefttree != NULL) {
	ShortAssignProcessingSites(plan->lefttree, qd);
    }

    if (plan->righttree != NULL) {
	ShortAssignProcessingSites(plan->righttree, qd);
    }
    
    switch (plan->type) {

	/*
	 * Scan - perform at (one of) the storage site(s) if
	 * it's a scan of a base relation.  Otherwise, we'll fix up the
	 * plan later to perform the scan at the same site as the node
	 * above it (at this point the only such nodes are merge-joins).
	 */
    case T_Scan:
    case T_IndexScan:
    case T_SeqScan: {
	if (((SeqScan *)plan)->scanrelid != 0) {
	    ShortSeqScan((SeqScan *)plan, qd->parsetree->rtable);
	}
    }
	break;

	/*
	 * Merge Join - perform at the seq. scan site with the most tuples,
	 * if the sequential scan is chunked with the join.
	 */
    case T_MergeJoin: {
	ShortMergeJoin((MergeJoin *)plan, qd->parsetree->rtable);
    }
	break;

	/*
	 * Merge - leave blank for now - will process a merge node at same site
	 * as its parent.
	 */
    case T_Merge: {
    }
	break;

	/*
	 * Sort, Aggregate, Group-By - process at same site as child node
	 */

    case T_Sort:
    case T_Agg:
    case T_Group:
	{
	    Assert(plan->lefttree != NULL);
	    plan->processingSite.type = plan->lefttree->processingSite.type;
	    plan->processingSite.addr = plan->lefttree->processingSite.addr;
	    plan->processingSite.port = plan->lefttree->processingSite.port;
	    plan->processingSite.hostid = plan->lefttree->processingSite.hostid;
	}
	break;

	/*
	 * T_Result -
	 */
    case T_Result: {
    }
	break;

	
    default: {
    }
	
    }
}

/*
 *-------------------------------------------------------------------------
 * ShortSeqScan--
 *
 * Get list of storage sites for fragment being scanned,
 * call routine to choose scan site.
 *
 * Also fill in size of fragment in plan->cost.  This can
 * be used by merge-join nodes to select a join site.
 *
 *-------------------------------------------------------------------------
 */
void
ShortSeqScan(SeqScan *scan, List *rtable) {

    RangeTblEntry *rte = (RangeTblEntry *)nth(scan->scanrelid-1, rtable);
    
    /*
     * Eventually, this subroutine should be replaced
     * by a RUSH callout.
     */
    SetSeqScanSite(scan, rte);
    
}

/*
 *-------------------------------------------------------------------------
 * SeqScanSite--
 *
 * For now, just return one of the storage sites, chosen at random.
 *-------------------------------------------------------------------------
 */
void
SetSeqScanSite(SeqScan *scan, RangeTblEntry *rte) {
    int whichSite;
    FragInfo *fInfo = (FragInfo *)nth(scan->scanfragid-1, rte->fragInfo);
    int nSites = length(fInfo->storageSites);
    bool *processYN;
    SiteAddrNode *saddr;
    
    whichSite = (random() % nSites);
    saddr = (SiteAddrNode *)nth(whichSite, fInfo->storageSites);
    
    set_nth(fInfo->processYN, whichSite, (void *)TRUE);

    scan->plan.processingSite.type = saddr->type;
    scan->plan.processingSite.addr = saddr->addr;
    scan->plan.processingSite.port = saddr->port;
    scan->plan.processingSite.hostid = saddr->hostid;
    
}

/*
 *-------------------------------------------------------------------------
 * MergeJoinGetChild--
 *
 * Input: Immediate child of a merge join node.
 *
 * Output: If immediate child is index scan, result is immediate child.
 *
 *		else result is immediate child's grandchild.
 *-------------------------------------------------------------------------
 */
Plan *MergeJoinGetChild(Plan *child) {
    Plan *result = NULL;
	
    if (child != NULL) {
	switch(child->type) {
	    /*
	     * Should be SeqScan->Sort->SeqScan
	     */
	case T_SeqScan: {
	    if (child->lefttree != NULL) {
		if (child->lefttree->lefttree != NULL) {
		    result = child->lefttree->lefttree;
		} else {
		    elog(WARN, "MergeJoinGetChild: plan->join.lefttree->lefttree->lefttree = NULL");
		}
	    } else {
		elog(WARN, "MergeJoinGetChild: Unsorted Sequential Scan below merge join.");
	    }
	    
	}
	    break;

	case T_IndexScan: {
	    result = child;
	}
	    break;

	case T_MergeJoin: {
	    result = child;
	}
	    break;
	    
	default: {
	}
	    
	}
	
    } else {
	elog(WARN, "MergeJoinGetChild: Merge join has no child node.");
    }
    
    return result;
}

/*
 *-------------------------------------------------------------------------
 * ShortMergeJoin--
 *
 * Do MergeJoin at site with most tuples, if both children nodes are
 * sequential scans.  If one is a join and the other a sequential scan,
 * perform this join at the site of the sequential scan (thereby assuming
 * that SS's will have more tuples than intermediate join results - not
 * by any means a valid assumption).  If both children are joins themselves,
 * then select from among all processing sites at random.
 *
 *-------------------------------------------------------------------------
 */
void
ShortMergeJoin(MergeJoin *plan, List *rtable) {


    Plan *leftChild;
    Plan *rightChild;

    leftChild = MergeJoinGetChild(plan->join.lefttree);
    rightChild = MergeJoinGetChild(plan->join.righttree);

    /*
     * Make merge-join happen at site with more stuff.
     */
    if ((leftChild->plan_size*leftChild->plan_width) > (rightChild->plan_size * rightChild->plan_width)) {
	CopySiteAddrNode(&leftChild->processingSite, &plan->join.processingSite);
    } else {
	CopySiteAddrNode(&rightChild->processingSite, &plan->join.processingSite);
    }

    /*
     * A merge join's children are always either (seqScan->sort->...)
     * or IndexScan.
     *
     * Make the seqScan and sort happen at the same site as the
     * node below it.
     */
    
    if (plan->join.lefttree != NULL && plan->join.lefttree->type == T_SeqScan) {
	if (plan->join.lefttree->lefttree != NULL && plan->join.lefttree->lefttree->type == T_Sort) {
	    CopySiteAddrNode(&plan->join.lefttree->lefttree->lefttree->processingSite, &plan->join.lefttree->processingSite);
	    CopySiteAddrNode(&plan->join.lefttree->lefttree->lefttree->processingSite, &plan->join.lefttree->lefttree->processingSite);
	}
    }
    
    if (plan->join.righttree != NULL && plan->join.righttree->type == T_SeqScan) {
	if (plan->join.righttree->lefttree != NULL && plan->join.righttree->lefttree->type == T_Sort) {
	    CopySiteAddrNode(&plan->join.righttree->lefttree->lefttree->processingSite, &plan->join.righttree->processingSite);
	    CopySiteAddrNode(&plan->join.righttree->lefttree->lefttree->processingSite, &plan->join.righttree->lefttree->processingSite);
	}
    }
}


/*
 *======================================================================================
 *				      UTILITY ROUTINES
 *======================================================================================
 */

/*
 *-------------------------------------------------------------------------
 * CreatePlanChunkList--
 *
 *	Input: Plan
 *
 *	Output: List of plan chunks, the union of which is the original
 *		plan.  All nodes in a plan chunk will be processed by
 *		the same site, in the absence of subcontracting.  How the
 *		original plan is "chunked" affects the degree of parallelism
 *		of the query execution.
 *
 * 		The Broker will request bids for each chunk.
 * 		A chunk might be one node or many nodes.  For
 * 		example, for a merge-join, a chunk might be any
 * 		of the following:
 *
 *		MERGE_JOIN
 *
 *	       		   |-SS-Sort-Merge-
 *  	     	MERGE_JOIN-|
 *	       		   |-SS-Sort-Merge-
 *
 *
 *			       		   |-SS
 *			       		   |-SS
 *	       		   |-SS-Sort-Merge-|-SS
 *  		MERGE_JOIN-|
 *	       		   |-SS-Sort-Merge-|-SS
 *			       		   |-SS
 *
 *
 * 		CreatePlanChunkList returns a list of BrokerPlans.  A BrokerPlan
 * 		has a pointer to a plan chunk and a pointer to the part of
 * 		the original plan tree to which the plan chunk corresponds.
 * 		This is useful for filling in processing sites in the original
 * 		plan tree later.
 *
 *-------------------------------------------------------------------------
 */

List *
CreatePlanChunkList(Plan *plan)
{
    List *planList = NIL;
    List *leaves = NIL;
    List *p;
    BrokerPlan *bPlan;
    
    /*
     * The Broker doesn't bid out Merge or Append nodes, but will bid
     * out all the plans below the node which feed into it.  The node
     * will be processed at the same site as the node above it or, if
     * it is the root of the plan, at the home site.
     *
     * Eventually, we may want this to change.  Merge nodes especially
     * represent work that has to be done by a processing site.  Hiding
     * them from a bidder site by not passing them during bidding but
     * then including them in the plan actually processed by the winner
     * site doesn't seem quite right.     -JPS 2/20/96
     *
     * For all other node types, find the plan leaves, create a plan
     * chunk, and recurse.
     */
    switch (plan->type) {
    case T_Merge:
	foreach(p, ((Merge *) plan)->mergeplans) {
	    planList = nconc(planList, CreatePlanChunkList((Plan *) lfirst(p)));
	}
	break;
    case T_Append:
	foreach(p, ((Append *) plan)->unionplans) {
	    planList = nconc(planList, CreatePlanChunkList((Plan *) lfirst(p)));
	}
	break;
    default:
	bPlan = (BrokerPlan *) calloc(1, sizeof(BrokerPlan));
	leaves = DefinePlanLeaves(plan);
	bPlan->originalPlan = plan;
	bPlan->planChunk = CreatePlanChunk(plan, leaves);
	planList = lcons(bPlan, NIL);
	break;
    }

    foreach (p, leaves) {
	Plan *parent = (Plan *) lfirst(p);

	if (parent->lefttree != NULL) {
	    planList = nconc(planList, CreatePlanChunkList(parent->lefttree));
	}
	if (parent->righttree != NULL) {
	    planList = nconc(planList, CreatePlanChunkList(parent->righttree));
	}
    }

    return planList;
}

/*
 *-------------------------------------------------------------------------
 * DefinePlanLeaves--
 *
 * 	Input: Plan
 *
 *	Output: List of Plan *'s, which are the leaves of the plan.
 *
 *	DefinePlanLeaves() recursively descends the plan tree.  At each
 * 	execution, it decides whether or not to cut off the recursion
 *	(and therefore the plan chunk).
 *
 **
 *-------------------------------------------------------------------------
 */
List *
DefinePlanLeaves(Plan *plan) {

    List *leaves = NIL;
    
    if (MAKECHUNK) {
	leaves = lappend(NIL, plan);
    } else {
	if (plan->lefttree == NULL && plan->righttree == NULL) {
	    leaves = lappend(NIL, plan);
	} else {
	    if (plan->lefttree != NULL) {
		leaves = nconc(leaves, DefinePlanLeaves(plan->lefttree));
	    }
	    if (plan->righttree != NULL) {
		leaves = nconc(leaves, DefinePlanLeaves(plan->righttree));
	    }
	}
    }

    return leaves;
}

/*
 *-------------------------------------------------------------------------
 * CreatePlanChunk--
 *
 * 	Input: Plan, List of node pointers
 *
 *	Output: The input plan, cropped off at the nodes pointed
 *		to in leaves.
 *
 *		Runs through the list of pointers in leaves and sets
 *		their left and right children to NULL;
 *-------------------------------------------------------------------------
 */
Plan *
CreatePlanChunk(Plan *root, List *leaves)
{
    Plan *planChunk;
    List *rightleaves = NIL, *leftleaves = NIL;
    List *l;
    Plan *p;

    foreach (l, leaves) {
	p = (Plan *) lfirst(l);
	rightleaves = lappend(rightleaves, p->righttree);
	leftleaves = lappend(leftleaves, p->lefttree);
	p->righttree = NULL;
	p->lefttree = NULL;
    }
    planChunk = copyObject(root);
    foreach (l, leaves) {
	p = (Plan *) lfirst(l);
	p->righttree = (Plan *) lfirst(rightleaves);
	rightleaves = lnext(rightleaves);
	p->lefttree = (Plan *) lfirst(leftleaves);
	leftleaves = lnext(leftleaves);
    }
    
    return planChunk;
}

/*
 *-------------------------------------------------------------------------
 * CopyPlanChunks--
 *
 *	Input: List of BrokerPlans
 *
 *	Output: None (void)
 *
 *	Traverses the plan chunks and copies the processing sites and
 *	budgets from their nodes into the original plan nodes.
 *
 *-------------------------------------------------------------------------
 */
void
CopyPlanChunks(List *bPlans)
{
    List *lp;
    BrokerPlan *bp;

    foreach(lp, bPlans) {
	bp = (BrokerPlan *) lfirst(lp);
	CopyProcessingSites(bp->planChunk, bp->originalPlan);
	CopyBudgets(bp->planChunk, bp->originalPlan);
    }
}

/*
 *-------------------------------------------------------------------------
 * FillInBlankNodes--
 *
 * 	Fill in nodes which were not filled in by short/long protocol,
 *	such as merge nodes.
 *-------------------------------------------------------------------------
 */
void
FillInBlankNodes(Plan *plan, Plan *parentPlan, SiteAddr *homeSite)
{
    List *lp;

    if (plan->processingSite.hostid == 0) {
	
	if (parentPlan != NULL) {
	    plan->processingSite.type = T_SiteAddrNode;
	    plan->processingSite.addr = parentPlan->processingSite.addr;
	    plan->processingSite.port = parentPlan->processingSite.port;
	    plan->processingSite.hostid = parentPlan->processingSite.hostid;
	} else {
	    plan->processingSite.type = T_SiteAddrNode;
	    plan->processingSite.addr = homeSite->addr;
	    plan->processingSite.port = homeSite->port;
	    plan->processingSite.hostid = homeSite->hostid;	    
	}
    }

    switch (plan->type) {
    case T_Merge:
	foreach(lp, ((Merge *) plan)->mergeplans) {
	    FillInBlankNodes((Plan *) lfirst(lp), plan, homeSite);
	}
	break;
    case T_Append:
	foreach(lp, ((Append *) plan)->unionplans) {
	    FillInBlankNodes((Plan *) lfirst(lp), plan, homeSite);
	}
	break;
    default:
	if (plan->lefttree != NULL) {
	    FillInBlankNodes(plan->lefttree, plan, homeSite);
	}
	if (plan->righttree != NULL) {
	    FillInBlankNodes(plan->righttree, plan, homeSite);
	}
	break;
    }
}

/*
 *-------------------------------------------------------------------------
 * FillInQepFragIds--
 *
 *	Go through the plan, assigning a qepFragId to each node, depending
 *	on which query fragment it belongs to.  These are the fragments
 *	which will be passed to the executor sites to be processed.
 *-------------------------------------------------------------------------
 */
static
void
FillInQepFragId(Plan *child, Plan *parent, int currentQepFragId)
{
    if (child != (Plan *) NULL) {
	if (child->processingSite.hostid == parent->processingSite.hostid) {
	    FillInQepFragIds(child, currentQepFragId);
	} else {
	    FillInQepFragIds(child, newQepFragId());
	}
    }
}

void
FillInQepFragIds(Plan *plan, int currentQepFragId)
{
    List *lp;
    Plan *p;

    plan->qepFragId = currentQepFragId;
    
    switch (plan->type) {
    case T_Merge:
	foreach(lp, ((Merge *) plan)->mergeplans) {
	    FillInQepFragId((Plan *) lfirst(lp), plan, currentQepFragId);
	}
	break;
    case T_Append:
	foreach(lp, ((Append *) plan)->unionplans) {
	    FillInQepFragId((Plan *) lfirst(lp), plan, currentQepFragId);
	}
	break;
    default:
	FillInQepFragId(plan->lefttree, plan, currentQepFragId);
	FillInQepFragId(plan->righttree, plan, currentQepFragId);
	break;
    }
}



/*
 *-------------------------------------------------------------------------
 * CopyProcessingSites--
 *
 * 	Recursively copy the processing sites from one plan to another.
 *-------------------------------------------------------------------------
 */
void
CopyProcessingSites(Plan *from, Plan *to)
{
    List *lp1, *lp2;

    if (from == (Plan *) NULL || to == (Plan *) NULL) {
	return;
    }

    switch (from->type) {
    case T_Merge:
	lp1 = ((Merge *) from)->mergeplans;
	lp2 = ((Merge *) to)->mergeplans;
	if (length(lp1) != length(lp2)) {
	    elog(WARN, "CopyProcessingSites: plan structures not identical!");
	}
	for (; lp1 != NIL && lp2 != NIL; lp1 = lnext(lp1), lp2 = lnext(lp2)) {
	    CopyProcessingSites((Plan *) lfirst(lp1), (Plan *) lfirst(lp2));
	}

	break;
	
    case T_Append:
	lp1 = ((Append *) from)->unionplans;
	lp2 = ((Append *) to)->unionplans;
	if (length(lp1) != length(lp2)) {
	    elog(WARN, "CopyProcessingSites: plan structures not identical!");
	}
	for (; lp1 != NIL && lp2 != NIL; lp1 = lnext(lp1), lp2 = lnext(lp2)) {
	    CopyProcessingSites((Plan *) lfirst(lp1), (Plan *) lfirst(lp2));
	}
	break;
	
    default:
	CopyProcessingSites(from->lefttree, to->lefttree);
	CopyProcessingSites(from->righttree, to->righttree);
	break;
    }
    CopySiteAddrNode(&from->processingSite, &to->processingSite);
}


/*
 *-------------------------------------------------------------------------
 * CopyBudgets--
 *
 * 	Recursively copy the budgets from one plan to another.
 *-------------------------------------------------------------------------
 */
void
CopyBudgets(Plan *from, Plan *to)
{
    List *lp1, *lp2;

    if (from == (Plan *) NULL || to == (Plan *) NULL) {
	return;
    }

    switch (from->type) {
    case T_Merge:
	lp1 = ((Merge *) from)->mergeplans;
	lp2 = ((Merge *) to)->mergeplans;
	if (length(lp1) != length(lp2)) {
	    elog(WARN, "CopyProcessingSites: plan structures not identical!");
	}
	for (; lp1 != NIL && lp2 != NIL; lp1 = lnext(lp1), lp2 = lnext(lp2)) {
	    CopyBudgets((Plan *) lfirst(lp1), (Plan *) lfirst(lp2));
	}
	break;
    case T_Append:
	lp1 = ((Append *) from)->unionplans;
	lp2 = ((Append *) to)->unionplans;
	if (length(lp1) != length(lp2)) {
	    elog(WARN, "CopyProcessingSites: plan structures not identical!");
	}
	for (; lp1 != NIL && lp2 != NIL; lp1 = lnext(lp1), lp2 = lnext(lp2)) {
	    CopyBudgets((Plan *) lfirst(lp1), (Plan *) lfirst(lp2));
	}
	break;
    default:
	CopyBudgets(from->lefttree, to->lefttree);
	CopyBudgets(from->righttree, to->righttree);
	break;
    }

    to->budget.response = from->budget.response;
    to->budget.cost = from->budget.cost;
    to->budget.delay = from->budget.delay;
    to->budget.staleness = from->budget.staleness;
    to->budget.accuracy = from->budget.accuracy;
    
}

/*
 *-------------------------------------------------------------------------
 * CopyContracts--
 *
 * 	Recursively copy the budgets from one plan to another.
 *-------------------------------------------------------------------------
 */
void
CopyContracts(Plan *from, Plan *to) {
    
    if (from->type == T_Merge) {
	List *lp, *lp2;
	lp2 = (((Merge *)to)->mergeplans);
	foreach(lp, ((Merge *)from)->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp);
	    Plan *p2 = (Plan *)lfirst(lp2);
	    CopyContracts(p, p2);
	    lp2 = lp2->next;
	}
    } else {
	if (from->lefttree != NULL) {
	    CopyContracts(from->lefttree, to->lefttree);
	}

	if (from->righttree != NULL) {
	    CopyContracts(from->righttree, to->righttree);
	}
    }

    to->contract = from->contract;
    
}

/*
 *-------------------------------------------------------------------------
 * CopySiteAddr--
 *
 * Does what it says it does.
 *-------------------------------------------------------------------------
 */
void
CopySiteAddr(SiteAddr *from, SiteAddr *to)
{
    to->port = from->port;
    to->addr = from->addr;
    to->hostid = from->hostid;
}

/*
 *-------------------------------------------------------------------------
 * CopySiteAddrNode--
 *
 * Ditto
 *-------------------------------------------------------------------------
 */
void
CopySiteAddrNode(SiteAddrNode *from, SiteAddrNode *to)
{
    to->type = T_SiteAddrNode;
    to->port = from->port;
    to->addr = from->addr;
    to->hostid = from->hostid;
}

/*
 *-------------------------------------------------------------------------
 * DataFragInfo--
 *
 *	Input: 	Fragment logical OID
 *
 *	Output: List of storage sites for fragment, and number of tuples in
 *		fragment.
 *
 *    	Looks up info in pg_fragment.  This means that there should be one entry
 *	in pg_fragment for each copy of a fragment.  The copies should have the
 *	same logical ID, different storage ID's.
 *
 *-------------------------------------------------------------------------
 */
FragInfo *
DataFragInfo(Oid fralogicalid) {
    
    Relation rd;
    HeapScanDesc sdesc;
    HeapTuple tup;
    Form_pg_fragment pgfform;
    FragInfo *fInfo = (FragInfo *)calloc(1, sizeof(FragInfo));
    SiteAddr *location;
    
    rd = heap_openr((Name)Name_pg_fragment);
    sdesc = heap_beginscan(rd, false, NowTimeQual, 0, NULL);

    while (HeapTupleIsValid(tup = heap_getnext(sdesc, false, NULL))) {
	pgfform = (Form_pg_fragment) GETSTRUCT(tup);
	if (pgfform->fralogicalid == fralogicalid) {
	    location = (SiteAddr *)calloc(1, sizeof(SiteAddr));
	    location->hostid = pgfform->fralocation;
	    location->addr = getAddrFromHostId(location->hostid);
	}
    }
    
    heap_endscan(sdesc);
    heap_close(rd);

    return fInfo;
    
}

/*
 *-------------------------------------------------------------------------
 * RandomSite--
 *
 * Select a site at random.
 *-------------------------------------------------------------------------
 */

SiteAddrNode *
RandomSite(Plan *plan, List *rtable) {
    List *sites = GetBidSites(plan, rtable);
    int whichSite = random() % length(sites);
    return (SiteAddrNode *)(nth(whichSite, sites));
}


/*
 *-------------------------------------------------------------------------
 * GetBidSites--
 *
 * For anything but a base-fragment scan, just run through the range table
 * and pick out all the storage sites therein, returning the values
 * (with no duplicates) in a list.
 *
 * For base-fragment scans, pass back the storage sites for that fragment.
 *
 *-------------------------------------------------------------------------
 */
List *
GetBidSites(Plan *plan, List *rtable) {
    List *lp;
    List *bidSites = NIL;
    extern unsigned short tcpPort;
    
    if (plan->type != T_SeqScan) {
	int i;

	
#if 0
	allAddrs[0].type = T_SiteAddrNode;
	allAddrs[0].hostid = 1;
	allAddrs[0].port = tcpPort;
	allAddrs[0].addr = 1025843328;

	allAddrs[1].type = T_SiteAddrNode;
	allAddrs[1].hostid = 17;
	allAddrs[1].port = 0;
	allAddrs[1].addr = 146446016;

	allAddrs[2].type = T_SiteAddrNode;
	allAddrs[2].hostid = 20;
	allAddrs[2].port = 0;
	allAddrs[2].addr = 96114368;

	allAddrs[3].type = T_SiteAddrNode;
	allAddrs[3].hostid = 19;
	allAddrs[3].port = 0;
	allAddrs[3].addr = 532321984;

	allAddrs[4].type = T_SiteAddrNode;
	allAddrs[4].hostid = 28;
	allAddrs[4].port = 0;
	allAddrs[4].addr = 800757440;

	for (i = 0; i < 5; i++) {
	    bidSites = lappend(bidSites, (void *)&allAddrs[i]);

	}
#endif
	
	foreach(lp, rtable) {
	    RangeTblEntry *rte = (RangeTblEntry *)lfirst(lp);
	    List *lp2;
	    foreach(lp2, rte->fragInfo) {
		FragInfo *fInfo = (FragInfo *)lfirst(lp2);
		List *lp3;
		foreach(lp3, fInfo->storageSites) {
		    SiteAddrNode *saddr = (SiteAddrNode *)lfirst(lp3);
		    bool found = FALSE;
		    List *lp4 = bidSites;
		    while (!found && lp4 != NIL) {
			SiteAddrNode *saddr2 = (SiteAddrNode *)lfirst(lp4);
			if (saddr->hostid == saddr2->hostid) {
			    found = TRUE;
			}
			lp4 = lp4->next;
		    }

		    if (!found) {
			bidSites = lappend(bidSites, (void *)saddr);
		    }
		
		}
	    
	    }
	
	}

	
    } else {
	SeqScan *ss = (SeqScan *)plan;
	RangeTblEntry *rte = (RangeTblEntry *)(nth(ss->scanrelid-1, rtable));
	FragInfo *fInfo = (FragInfo *)(nth(ss->scanfragid-1, rte->fragInfo));
	List *lp3;
	foreach(lp3, fInfo->storageSites) {
	    SiteAddrNode *saddr = (SiteAddrNode *)lfirst(lp3);
	    bool found = FALSE;
	    List *lp4 = bidSites;
	    while (!found && lp4 != NIL) {
		SiteAddrNode *saddr2 = (SiteAddrNode *)lfirst(lp4);
		if (saddr->hostid == saddr2->hostid) {
		    found = TRUE;
		}
		lp4 = lp4->next;
	    }

	    if (!found) {
		bidSites = lappend(bidSites, (void *)saddr);
	    }
		
	}
	
    }

    return bidSites;
}

/*
 * BrokerDebug--
 *	Debugging printout for Broker
 */
void
BrokerDebug(int lev, char *fmt, ...)
{
    va_list ap;
    
    if (lev > BROKERDEBUGLEVEL) {
	return;
    }

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

/*
 *  printPlanChunks--
 *
 *	Prints out each plan in the list of plan chunks
 *	created by the Broker.
 */
void
printPlanChunks(List *bPlans, List *rtable) {
    List *bp;
    
    foreach(bp, bPlans) {
	Plan *p = ((BrokerPlan *)lfirst(bp))->planChunk;
	printf("DEMO PLANCHUNK %i ", p->qepFragId);
	demoPrintPlanTree(p, rtable, 0, FALSE);
	fflush(stdout);
    }
    
}

int
newQepFragId()
{
    static int lastQepFragId = 0;
    return ++lastQepFragId;
}
