/*-------------------------------------------------------------------------
 *
 * brokerLong.c--
 *    
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/tcop/brokerLong.c,v 1.7 1996/05/31 17:14:36 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */

#include "tcop/broker.h"
#include "tcop/brokerLong.h"

/*
 * Forward declarations.
 */
void InitializeBidInfo(int requestIndex, unsigned int requestId, List *bPlans, int protocol, List *rtable, int operation, SiteAddr *homeSiteAddr);
void CallBidders(int requestIndex, List *bPlans);
void *BidOutPlanChunk(void *arg);
void SendOneRFB(int requestIndex, int whichChunk, int whichBid, SiteAddrNode *site);
void RecordBid(Budget *bid, BidInfo *bidInfo, int whichBid, int contract, bool subContracted, Plan *subContractedPlan);
void BiddingDone(BidInfo *bidInfo, int whichChunk);

/*
 * Utilities
 */
void FillInProcessingSites(int requestIndex, List *bPlans);
void CopySiteAddrIntoPlan(SiteAddrNode *site, Plan *plan);
void CopyContractIntoPlan(ContractId contract, Plan *plan);
unsigned int GetNextrequestId();
bool BelowBidCurve(Budget *bid);
Budget *TotalPlanBid(Plan *p);
Budget *AddBids(Budget *b1, Budget *b2);

/*
 *-------------------------------------------------------------------------
 * Long Protocol information.  This stuff has to be global, since the
 * bidders RPC back here.  The subroutine called by the bidders has
 * to access the same info as the one that called them.  The other option
 * would be to send all this stuff over the wire - inefficient, to say the
 * least.
 *
 * requests[requestId]  				Info for all requests.  A request corresponds to one query,
 *							one fragment buy or sell, etc.
 * requests[requestId]->bidInfo[]  			Bidding info for a request.
 * requests[requestId]->bidInfo[partNo]			Bidding info for one part of a request, for example, for one
 *							subplan.
 * requests[requestId]->bidInfo[partNo].bids[bidNo]	One bid for one part of a query.
 *
 * REQUESTSSIZE						Current size of requests[].  Can expand if we run out of space.
 *
 * requestIdLock 					Mutex around incrementing requestId
 *-------------------------------------------------------------------------
 */

BrokerInfo **requests;

REQUESTSSIZE = DEFAULT_REQUEST_STORAGE;

static pthread_mutex_t requestIdLock;

double bidCurveCost1 = 200.0;
double bidCurveDelay1 = 0.0;
double bidCurveCost2 = 0.0;
double bidCurveDelay2 = 10.0;

static pthread_mutex_t brokerLock;					/* Lock around updating cost, delay, etc. */

/*
 *-------------------------------------------------------------------------
 * InitBroker--
 *
 * Called once from postgresMain - initializes global variables.
 *-------------------------------------------------------------------------
 */
void
InitBroker() {
    int i;
    
    pthread_mutex_init(&requestIdLock, pthread_attr_default);
    pthread_mutex_init(&brokerLock, pthread_attr_default);

    requests = calloc(DEFAULT_REQUEST_STORAGE, sizeof(BrokerInfo *));
    
    /*
     * If a slot in requests[] isn't being used, it is set
     * to NULL.
     */
    for (i = 0; i < DEFAULT_REQUEST_STORAGE; i++) {
	requests[i] = NULL;
    }
    
}

/*
 *======================================================================================
 * Long Protocol--
 * 
 * Get list of potential bid sites, based on the type of node passed
 * in then fork a thread to send out requests for bids.  The thread
 * blocks until all the bids have been received.
 *======================================================================================
 */
void
LongProtocol(List *bPlans, int protocol, QueryDesc *qd, SiteAddr *homeSiteAddr) {

    /*
     * Assign next request ID number
     */
    unsigned int requestId = qd->requestId;
    int requestIndex = requestId % REQUESTSSIZE;
    List *rtable = qd->parsetree->rtable;
    int operation = qd->operation;
    
    InitializeBidInfo(requestIndex, requestId, bPlans, protocol, rtable, operation, homeSiteAddr);

    /*
     * CallBidders forks off a thread for each bid.
     * Each thread blocks until the bid comes back.
     * CallBidders() blocks until all bids have been
     * received.  It then determines the winners, notifies
     * the losers, and fills the winners into the plan tree.
     * Therefore, when CallBidders() returns,
     * the processing sites have been filled into the
     * plan tree and we can return.
     */
    CallBidders(requestIndex, bPlans);
    
}


void
InitializeBidInfo(int requestIndex, unsigned int requestId, List *bPlans, int protocol, List *rtable, int operation, SiteAddr *homeSiteAddr) {
    List *bPlan;
    int j = 0;

    requests[requestIndex] = (BrokerInfo *)calloc(1, sizeof(BrokerInfo));
    
    requests[requestIndex]->requestId = requestId;
    requests[requestIndex]->nBidInfo = length(bPlans);
    requests[requestIndex]->bidInfo = (BidInfo *)calloc(requests[requestIndex]->nBidInfo, sizeof(BidInfo));
    requests[requestIndex]->protocol = protocol;
    requests[requestIndex]->nChunksLeft = requests[requestIndex]->nBidInfo;
    pthread_mutex_init(&requests[requestIndex]->biddingDoneLock, pthread_mutexattr_default);
    pthread_cond_init(&requests[requestIndex]->biddingDoneCond, pthread_condattr_default);
    
    foreach(bPlan, bPlans) {
	List *bidders = GetBidSites(((BrokerPlan *)(bPlan->elem))->planChunk, rtable);
	List *bidder;
	int i;
	ProcessQueryArg *qArg;

	requests[requestIndex]->bidInfo[j].nBidders = length(bidders);
	requests[requestIndex]->bidInfo[j].nBiddersContacted = 0;
	requests[requestIndex]->bidInfo[j].nBidsReceived = 0;
	requests[requestIndex]->bidInfo[j].bids = calloc(requests[requestIndex]->bidInfo[j].nBidders,
						      sizeof(Bid));
	i = 0;
	foreach(bidder, bidders) {
	    CopySiteAddrNode((SiteAddrNode *)(bidder->elem), &requests[requestIndex]->bidInfo[j].bids[i].site);
	    requests[requestIndex]->bidInfo[j].bids[i].bidderContacted = FALSE;
	    requests[requestIndex]->bidInfo[j].bids[i].bidReceived = FALSE;
	    i++;
	}
	
	requests[requestIndex]->bidInfo[j].okBidReceived = FALSE;
	requests[requestIndex]->bidInfo[j].biddingDone = FALSE;

	qArg = &requests[requestIndex]->bidInfo[j].BidDetails_u.longQuery;
	qArg->operation = operation;
	qArg->plan = (PQA_Plan_tree) copyObject(((BrokerPlan *)(bPlan->elem))->planChunk);
	qArg->rtable = (PQA_Node_list) copyObject(rtable);
	qArg->homeSite = *homeSiteAddr;
	
	/*
	 * The rest of the ProcessQueryArg can't be filled in until
	 * the Broker is finished, and doesn't need to be filled in
	 * for the Broker to do its job.
	 */
	
	j++;
    }
    
    
}

/*
 *-------------------------------------------------------------------------
 * CallBidders--
 *
 *	Forks off a thread for each bid request, sends out the request,
 *	then blocks.  When all the threads have finished, determines
 *	the winners, notifies the losers, and fills in the processing
 *	sites in the plan trees.
 *
 *-------------------------------------------------------------------------
 */
void
CallBidders(int requestIndex, List *bPlans) {
    int whichChunk;
    pthread_t *thread;
    pthread_attr_t *attr;
    
    /*
     * Iterate through the query fragments, forking off a thread to
     * BidOutPlanChunk for each one.
     */
    for (whichChunk = 0; whichChunk < requests[requestIndex]->nBidInfo; whichChunk++) {
	char **argv = (char **)calloc(3, sizeof(char *));
	
	thread = (pthread_t *)calloc(1, sizeof(pthread_t));
	attr = (pthread_attr_t *)calloc(1, sizeof(pthread_attr_t));
	
	argv[0] = (char *)requestIndex;
	argv[1] = (char *)whichChunk;
	
	/*
	 * Fork off a thread for the bidding process for
	 * each query fragment.
	 */
	pthread_attr_create(attr);
	pthread_create(thread, *attr, BidOutPlanChunk, (void *)argv);
	pthread_detach(thread);
    }

    pthread_mutex_lock(&requests[requestIndex]->biddingDoneLock);
    while (requests[requestIndex]->nChunksLeft > 0) {
	pthread_cond_wait(&requests[requestIndex]->biddingDoneCond,
			  &requests[requestIndex]->biddingDoneLock);
    }

    /*
     * Fill in the processing sites in the plan chunks.
     */

    FillInProcessingSites(requestIndex, bPlans);
    
}

/*
 *-------------------------------------------------------------------------
 * BidOutPlanChunk -
 *    Sends an RFB to all sites in site_list, sending them the
 *    plan for the query fragment and asking them to bid on the query.
 *
 *    For each site, starts a thread that calls the RequestForBid RPC function.
 *
 *    When it returns, control will transfer to the main broker thread.
 *
 * Results: 
 *    none (void)
 *
 * Side effects:
 *    updates bidCollection passed in
 *-------------------------------------------------------------------------
 */
void *
BidOutPlanChunk(void *arg)
{
    char **argv = (char **)arg;
    unsigned int requestIndex = (int)argv[0];               	/* Index into allBids - used to index recvQepFragIds, etc.    */
    int whichChunk = (int)argv[1];		     		/* Which query chunk this bid request is for                 */
    int nBidders = requests[requestIndex]->bidInfo[whichChunk].nBidders;
    int whichBid;
    int bestBidIndex;
    BidInfo *bestBidInfo;
    pthread_t *thread;
    pthread_attr_t *attr;
    extern bool demo;
    
    pthread_mutex_init(&(requests[requestIndex]->bidInfo[whichChunk].bidsLock),
			   pthread_mutexattr_default);
    pthread_cond_init(&(requests[requestIndex]->bidInfo[whichChunk].bidsCond),
			  pthread_condattr_default);
    pthread_mutex_lock(&(requests[requestIndex]->bidInfo[whichChunk].bidsLock));
    
    for (whichBid = 0; whichBid < nBidders; whichBid++) {
	
	if (demo) {
	    demoPrintMsg("DEMO BROKER CHUNKSTATUS %i %i CONTACTED\n", whichChunk, whichBid);
	}

	tcopDebug(3, "BidOutPlanChunk: bidCollection no. %i sending out request %i.\n",
		requestIndex, whichBid);

	SendOneRFB(requestIndex, whichChunk, whichBid, &requests[requestIndex]->bidInfo[whichChunk].bids[whichBid].site);
	
    } 

    /*
     * Block here and wait for all the bids to come in.
     * These condition variables will be signalled in BiddingDone().
     * requests[requestIndex].biddingDone is necessary because
     * condition variables can be spuriously signalled (why? I don't
     * know.  The man pages just say they can.)
     */

    while (!requests[requestIndex]->bidInfo[whichChunk].biddingDone) {
	pthread_cond_wait(&(requests[requestIndex]->bidInfo[whichChunk].bidsCond),
			  &(requests[requestIndex]->bidInfo[whichChunk].bidsLock));
    }

    pthread_mutex_lock(&brokerLock);

    /*
     * Decrement the number of query fragments whose bidding process
     * is not yet finished and then signal the thread waiting in the Broker().
     */
    requests[requestIndex]->nChunksLeft--;
    pthread_cond_signal(&requests[requestIndex]->biddingDoneCond);
    
    pthread_mutex_unlock(&brokerLock);

    
}    


/*
 *-------------------------------------------------------------------------
 * SendOneRFB -
 *    Thread function that calls site manager procedure ReceiveRequestForBid().
 *
 *    Called by BidOutPlanChunk.  Unpacks arguments, sets up and makes
 *    the RPC call to RequestForBid.
 *
 * Results: 
 *    none (void)
 *
 * Side effects:
 *    none - updates no globals
 *-------------------------------------------------------------------------
 */

void
SendOneRFB(int requestIndex, int whichChunk, int whichBid, SiteAddrNode *site)
{
    OfferBuffer *offbuf = calloc(1, sizeof(OfferBuffer));
    extern unsigned short tcpPort;
    RPCErrorCode result;
    RpcConn *rpcc;
    
    int localSiteAddr = getMyAddr();
    int localSiteHostId = getMyHostId();

    /* 
     * fill in some details of the offer buffer 
     * don't fill in budget or contract id ... 
     * these will be supplied by the bidder
     */

    offbuf->details.protocol = requests[requestIndex]->protocol;
    offbuf->requestId = requests[requestIndex]->requestId;
    offbuf->requestIndex = requestIndex;
    offbuf->whichChunk = whichChunk;
    offbuf->whichBid = whichBid;
    offbuf->requester.addr = localSiteAddr;
    offbuf->requester.port = tcpPort;
    offbuf->requester.hostid = localSiteHostId;
    
    switch (requests[requestIndex]->protocol) {
    case LONG_PROCESS_QUERY: {
	ProcessQueryArg *offerDetails = &(offbuf->details.OfferDetails_u.longQuery);
	ProcessQueryArg *reqDetails = &requests[requestIndex]->bidInfo[whichChunk].BidDetails_u.longQuery;
	Plan *p = (Plan *)reqDetails->plan;
	List *lp = (List *)reqDetails->rtable;
	
	offerDetails->operation = reqDetails->operation;
	offerDetails->plan = (PQA_Plan_tree)p;
	offerDetails->rtable = (PQA_Node_list)lp;
	offerDetails->homeSite = reqDetails->homeSite;
	
    }
	break;

     case LONG_SELL_FRAGMENT: {

#if 0
 	FragXArg *offbufMoveDetails = &(offbuf->details.OfferDetails_u.longSell);
	FragXArg *bidcollMoveDetails = &(bidsRequested[bidCollIndex].BidDetails_u.longSell);
	
 	offbufMoveDetails->qepFragId = bidcollMoveDetails->qepFragId;
	/*
	 * Again, this is gross but the rest of the code depends on
	 * the sender being the site requesting the fragment and this
	 * was the easiest way to make this work. - Robert
	 */
	offbufMoveDetails->receiver = bidcollMoveDetails->receiver;
	offbufMoveDetails->sender.addr = bidcollMoveDetails->sender.addr;
	offbufMoveDetails->sender.port = bidcollMoveDetails->sender.port;
	offbufMoveDetails->sender.hostid = bidcollMoveDetails->sender.hostid;
	offbufMoveDetails->frastoreid = bidcollMoveDetails->frastoreid;
	offbufMoveDetails->fraclassid = 0;
#endif
     }
 	break;

     case LONG_BUY_FRAGMENT: {
#if 0
 	FragXArg *offbufMoveDetails = &(offbuf->details.OfferDetails_u.longBuy);
	FragXArg *bidcollMoveDetails = &(bidsRequested[bidCollIndex].BidDetails_u.longBuy);
	
 	offbufMoveDetails->qepFragId = bidcollMoveDetails->qepFragId;
	offbufMoveDetails->receiver = bidcollMoveDetails->receiver;
	offbufMoveDetails->sender.addr = bidcollMoveDetails->sender.addr;
	offbufMoveDetails->sender.port = bidcollMoveDetails->sender.port;
	offbufMoveDetails->sender.hostid = bidcollMoveDetails->sender.hostid;
	offbufMoveDetails->frastoreid = bidcollMoveDetails->frastoreid;
	offbufMoveDetails->fraclassid = 0;
#endif
     }
 	break;
    }


    tcopDebug(3, "SendOneRFB from: (%d, %d, %i) to (%d, %d, %i)\n", 
	localSiteHostId, localSiteAddr, tcpPort,
	site->hostid, site->addr, site->port);

    /*
     * We add hostid to program number so we contact the correct sitemgr
     */
    rpcc =
	Rpc_connectSitemgr(site->hostid, site->addr,
			   "SendOneRFB", RPC_ONERR_RETURN);
    if (rpcc == NULL) {
	/* was elog(WARN) but no good doing longjmps in thread, we just
	   ignore this RFB.  -ay 4/95 */
	elog(NOTICE, "SendOneRFB: can't create client handle\n");
	return;
    }

    if (!Rpc_call(rpcc, RPC_REQUESTFORBID, offbuf, &result)) {
	/* was elog(WARN) but no good doing longjmps in thread, we just
	   ignore this RFB. -ay 4/95 */
	elog(NOTICE, "SendOneRFB: RPC error\n");
    } else {
	requests[requestIndex]->bidInfo[whichChunk].bids[whichBid].bidderContacted = TRUE;
	requests[requestIndex]->bidInfo[whichChunk].nBiddersContacted++;
    }
    
    Rpc_disconnect(rpcc);

    
}

/*
 *-------------------------------------------------------------------------
 * BrokerReceiveBid -
 *    Main function of the thread that handles BrokerReceiveBid RPC requests.
 *    
 *    Called by site manager procedure ReceiveRequestForBid.  Receives bid,
 *    calls RecordBid to update information about the bidding process and 
 *    decides  whether or not to end the bidding process for the specified
 *    query fragment and award the contract.
 *
 * Results: 
 *    none (void)
 *
 * Side effects:
 *    updates allBids[thisQueryFragment].info[thisBid] - that is, the
 *    bidding information for the query fragment and the bid that this
 *    thread is handling.
 *-------------------------------------------------------------------------
 */
void
BrokerReceiveBid(void *arg)	/* thread-style packed arguments */
{
    char **argv = (char **)arg;
    BidBuffer *bidbuf = (BidBuffer *)argv[0];		/* bid information */
    SVCXPRT *transp = (SVCXPRT *)argv[1];		/* Requesting site */
    RPCErrorCode result;				/* RPC call stuff */

    unsigned int requestId = bidbuf->requestId;		/* Unique identifier for this request */
    int requestIndex = bidbuf->requestIndex;    	/* Index into requests[] */
    int whichChunk = bidbuf->whichChunk;		/* Which plan chunk this bid was for. */
    int whichBid = bidbuf->whichBid; 			/* Which bid for the plan chunk this was. */
    ContractId contract = bidbuf->contract; 		/* Contract ID at the bidder site. */
    QepFragmentId qepFragId = bidbuf->qepFragId; 	/* query fragment id */
    Budget *bid = &bidbuf->bid;				/* cost and delay */
    bool subContracted = bidbuf->subContracted;
    Plan *subContractedPlan = (Plan *)bidbuf->subContractedPlan;
    extern bool demo;
    
    /*
     * Send result back to ReceiveRequestForBid() running at remote
     * SiteMgr so RPC call doesn't time out.
     */
    result = SUCCESS;
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	elog(NOTICE, "BrokerReceiveBid: svc_sendreply failed\n");
    }


    /*
     * Update global bidding information.
     */
    RecordBid(bid, &requests[requestIndex]->bidInfo[whichChunk], whichBid, contract, subContracted, subContractedPlan);

    /*
     * Decide whether to end bidding process and award contract.
     * For now, just wait until all bids have been received for
     * this bid collection.
     */

    if (demo) {
	demoPrintMsg("DEMO BROKER CHUNKSTATUS %i %i BIDRECEIVED %f %f\n", whichChunk, whichBid, bid->cost, bid->delay);
    }
    
    tcopDebug(3,
	      "BrokerReceiveBid:  Got bid (%f, %f) for:\n\trequestID %i\n\twhichChunk %i \n\twhichBid %i\nfrom site %i\n\n",
	      bid->cost, bid->delay, requestId, whichChunk, whichBid,
	      requests[requestIndex]->bidInfo[whichChunk].bids[whichBid].site.hostid);
    
    if (requests[requestIndex]->bidInfo[whichChunk].nBidsReceived ==
	requests[requestIndex]->bidInfo[whichChunk].nBidders) {

	tcopDebug(3, "BrokerReceiveBid: All %i bids received for requestId %i, chunk %i.\n", 
		requests[requestIndex]->bidInfo[whichChunk].nBidsReceived, whichChunk, requestIndex);

	/*
	 * Call BiddingDone() to signal thread waiting in BidOutPlanChunk().
	 */
	    
	BiddingDone(&requests[requestIndex]->bidInfo[whichChunk], whichChunk);

    } else {

	tcopDebug(3, "BrokerReceiveBid: %i out of %i bids received for requestId %i, whichChunk %i.\n", 
		requests[requestIndex]->bidInfo[whichChunk].nBidsReceived, 
		requests[requestIndex]->bidInfo[whichChunk].nBidders, 
		requestId, whichChunk);
    }

}


/*
 *-------------------------------------------------------------------------
 * RecordBid -
 *     Reentrant broker function that gets called each time a
 *     bid is received.  Accesses variables declared in rfb.h.
 *
 *     For now, just compares current bid to best bid so far.
 *     If current bid.cost is less, make it the best bid.
 *
 * Results: 
 *     none (void)
 *
 * Side effects:
 *-------------------------------------------------------------------------
 */

void
RecordBid(Budget *bid, BidInfo *bidInfo, int whichBid, int contract, bool subContracted, Plan *subContractedPlan)
{
    

    bidInfo->nBidsReceived++;
    bidInfo->bids[whichBid].contract = contract;
    bidInfo->bids[whichBid].bidReceived = TRUE;
    bidInfo->bids[whichBid].bid = *bid;
    bidInfo->bids[whichBid].subContracted = subContracted;
    if (subContracted) {
	bidInfo->bids[whichBid].subContractedPlan = copyObject(subContractedPlan);
    }
    
    /*
     * If bidder site chose to bid
     */
    if (bid->response == BID) {
	
	/*
	 * If below bid curve and either best bid so far or first
	 * bid received for this query fragment, update bid information.
	 */
	if ((BelowBidCurve(bid)) && (!(bidInfo->okBidReceived) ) || (bid->cost < bidInfo->bids[whichBid].bid.cost)) {
	    bidInfo->bestBidIndex = whichBid;
	    bidInfo->okBidReceived = TRUE;
	    
	    tcopDebug(3, "RecordBid: Best bid so far: no. %i, cost=%f, delay=%f\n", bidInfo->bestBidIndex, bid->cost, bid->delay);
	}
    }
}


/*
 *-------------------------------------------------------------------------
 * BiddingDone
 *
 *	Called when bidding process is finished for one plan chunk.
 *
 *      Part of long protocol - called by BrokerReceiveBid().  If this is a
 *      home site, calls ContractLost at the loser sites.  Doesn't call
 *      ContractWon - that is handled inside ProcessQueryDesc.
 *
 *      Signals thread blocked in Broker().
 *
 *	If this is a subBid, blocks and waits to be notified by SubcontractWon
 *      or SubcontractLost whether the parent contract was won or lost.
 *
 * Results: 
 *     none (void)
 *
 * Side effects:
 *     none - doesn't update any global vars.
 *-------------------------------------------------------------------------
 */

void
BiddingDone(BidInfo *bidInfo, int whichChunk)
{
  int i;
  ProcessQueryArg *queryDetails; 				/* Description of query fragment to be processed, 	*/
                                                                /* if this contract is for processing a query 		*/
  FragXArg *mfArg; 						/* Details about fragment movement, if this contract 	*/
                                                                /* is to buy a fragment 				*/
  SiteAddr *thisSite; 						/* tcp information for this site 			*/
  QepFragmentId *ids; 						/* list of query execution plan fragment ID's 		*/
  struct sockaddr_in internetAddr;  				/* For tcp call 					*/
  RpcConn **rpcc; 					        /* ditto 						*/
  RPCErrorCode result; 						/* ditto 						*/
  int bestBid = bidInfo->bestBidIndex; 				/* Index of winning bidder site. 			*/
  ContractId contract; 						/* Contract number at bidder site 			*/
  int remotehostid;	                                        /* of the sitemgr we're contacting                      */
    
  /*
   * Need one RPC connection for each bidder site.
   */
  rpcc = (RpcConn **)calloc(bidInfo->nBidders, sizeof(RpcConn *));


  /*
   *---------------------------
   * Notify losers, if they submitted a bid.
   * Don't contact sites that refused to bid.
   *---------------------------
   */

  for (i = 0; i < bidInfo->nBidders; i++) {
      if (i != bestBid && bidInfo->bids[i].bid.response == BID) {
	      
	  struct sockaddr_in internetAddress;
	  contract = bidInfo->bids[i].contract;

	  if (demo) {
	      demoPrintMsg("DEMO BROKER CHUNKSTATUS %i %i LOST\n", i, whichChunk);
	  }
	  
	  /* make the RPC call */
	  remotehostid = bidInfo->bids[i].site.hostid;
	  internetAddress.sin_family = AF_INET;
	  internetAddress.sin_port = 0;
	  internetAddress.sin_addr.s_addr =
	      (u_int)bidInfo->bids[i].site.addr;

	  tcopDebug(1, "BiddingDone to:   (%d, %d, %i)\n", remotehostid,
	    internetAddress.sin_addr.s_addr, internetAddress.sin_port);

	  rpcc[i] = Rpc_connectSitemgr(remotehostid,
				       internetAddress.sin_addr.s_addr,
				       "BiddingDone",
				       RPC_ONERR_WARN);

	  tcopDebug(3, "Calling ContractLost: \n\tsite %i\n\tbid no. %i\n\tcontract %i\n\n",
		  remotehostid, i, contract);

	  Rpc_call(rpcc[i], RPC_CONTRACTLOST, &contract, &result);

	  Rpc_disconnect(rpcc[i]);
      }
  }
  
  /*
   * Signal thread waiting in BidOutPlanChunk() to continue.
   */
  bidInfo->biddingDone = TRUE;
  pthread_cond_signal(&(bidInfo->bidsCond));
}


/*
 *-------------------------------------------------------------------------
 * SubContractBroker--
 *
 *	Called by our sitemgr to subcontract out a piece of a plan.
 *	Construct all the stuff we need to pass to the Broker(), let
 *	it do its stuff, then pass the answer back in the form of
 *	a BrokerAnswer.
 *-------------------------------------------------------------------------
 */
void
SubContractBroker(void *arg)	/* thread-style packed arguments */
{
    char **argv = (char **)arg;
    BrokerArg *brokerArg = (BrokerArg *)argv[0];	/* broker information */
    SVCXPRT *transp = (SVCXPRT *)argv[1];		/* Requesting site */
    RPCErrorCode result;				/* RPC call stuff */
    RpcConn *rpcc;
    BrokerAnswer bAnswer;
    QueryDesc qd;
    Query *parsetree = makeNode(Query);
    List *plans;
    Budget *bidp;
    
    int protocol = LONG_PROCESS_QUERY;
    
    /*
     * Send result back to SubContract() running at remote
     * SiteMgr so RPC call doesn't time out.
     */
    result = SUCCESS;
    if (!svc_sendreply(transp, xdr_RPCErrorCode, &result)) {
	elog(NOTICE, "SubContractBroker: svc_sendreply failed\n");
    }

    /*
     * Call the Broker.  It will fill in the processing
     * sites in the plan tree.
     */

    parsetree->commandType = brokerArg->operation;
    parsetree->resultRelation = 0;
    parsetree->resultFragment = 0;
    parsetree->rtable = (List *)brokerArg->rtable;
    qd.operation = brokerArg->operation;
    qd.parsetree = parsetree;
    qd.plantree = (Plan *)brokerArg->plan;
    qd.remoteplans = NIL;
    qd.distPlanList = NIL;

    plans = lcons(qd.plantree, NIL);
    
    Broker(&qd, plans, protocol, &brokerArg->homeSiteAddr);
    
    /*
     * Call SubBrokerDone() inside the sitemgr, passing back the
     * plans tree with the processing sites filled in and the bid.
     */

    bidp = TotalPlanBid(qd.plantree);
    bAnswer.plan = (PQA_Plan_tree)qd.plantree;
    bAnswer.bid = *bidp;
    
    rpcc = Rpc_connectSitemgr(getMyHostId(), getMyAddr(), "SubContractBroker", RPC_ONERR_RETURN);
    
    if (!Rpc_call(rpcc, RPC_SUBBROKERDONE, &bAnswer, &result)) {
	elog(NOTICE, "SubContractBroker: RPC error\n");
    }
    
    Rpc_disconnect(rpcc);
    
}


/*
 *======================================================================================
 *				      UTILITY ROUTINES
 *======================================================================================
 */

/*
 *-------------------------------------------------------------------------
 * FillInProcessingSites--
 *
 *	Fills the processing site from the winning bidders into the
 *	plan chunks.  These will then be copied into the original
 *	plan when the long protocol returns back to the Broker() main
 *	procedure.
 *
 *-------------------------------------------------------------------------
 */
void
FillInProcessingSites(int requestIndex, List *bPlans) {
    int whichChunk, whichBid;
    List *lp = bPlans;
    Plan *planChunk;
    
    for (whichChunk = 0; whichChunk < requests[requestIndex]->nBidInfo; whichChunk++) {
	BidInfo *bidInfo = &requests[requestIndex]->bidInfo[whichChunk];
	planChunk = ((BrokerPlan *)lfirst(lp))->planChunk;

	if (bidInfo->okBidReceived) {
	    Bid *winningBid = &bidInfo->bids[bidInfo->bestBidIndex];
	    if (winningBid->subContracted) {
		CopyProcessingSites(winningBid->subContractedPlan, planChunk);
		CopyBudgets(winningBid->subContractedPlan, planChunk);
		CopyContracts(winningBid->subContractedPlan, planChunk);
	    } else {
		CopySiteAddrIntoPlan(&winningBid->site, planChunk);
		CopyBudgetIntoPlan(&winningBid->bid, planChunk);
		CopyContractIntoPlan(winningBid->contract, planChunk);
	    }
	} else {
	    SiteAddrNode noSite;
	    noSite.type = T_SiteAddrNode;
	    noSite.addr = 0;
	    noSite.port = 0;
	    noSite.hostid = -1;
	    CopySiteAddrIntoPlan(&noSite, planChunk);
	}
	
	lp = lp->next;
    }
    
}

/*
 *-------------------------------------------------------------------------
 * CopySiteAddrIntoPlan--
 *
 *	Recursively fills in the site address into every node
 *	in the plan.  NB: Plans contain SiteAddrNode's, not
 *	SiteAddr's.
 *
 *-------------------------------------------------------------------------
 */
void
CopySiteAddrIntoPlan(SiteAddrNode *site, Plan *plan)
{
    List *lp;

    if (plan != (Plan *) NULL) {
	plan->processingSite.type = T_SiteAddrNode;
	plan->processingSite.addr = site->addr;
	plan->processingSite.port = site->port;
	plan->processingSite.hostid = site->hostid;
	
	switch (plan->type) {
	case T_Merge:
	    foreach(lp, ((Merge *) plan)->mergeplans) {
		CopySiteAddrIntoPlan(site, (Plan *) lfirst(lp));
	    }
	    break;
	case T_Append:
	    foreach(lp, ((Append *) plan)->unionplans) {
		CopySiteAddrIntoPlan(site, (Plan *) lfirst(lp));
	    }
	    break;
	default:
	    CopySiteAddrIntoPlan(site, plan->lefttree);
	    CopySiteAddrIntoPlan(site, plan->righttree);
	    break;
	}
    }
}

/*
 *-------------------------------------------------------------------------
 * CopyContractIntoPlan--
 *
 *	Recursively fills in the contract into every node
 *	in the plan.
 *
 *-------------------------------------------------------------------------
 */
void
CopyContractIntoPlan(ContractId contract, Plan *plan) {
    plan->contract = contract;

    if (plan->type == T_Merge) {
	List *lp;
	foreach(lp, ((Merge *)plan)->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp);
	    CopyContractIntoPlan(contract, p);
	}
	
    } else {
	if (plan->lefttree != NULL) {
	    CopyContractIntoPlan(contract, plan->lefttree);
	}

	if (plan->righttree != NULL) {
	    CopyContractIntoPlan(contract, plan->righttree);
	}
    }
}

/*
 *-------------------------------------------------------------------------
 * GetNextrequestId--
 *
 * Returns next requestId.  The requestId's go from 0...MAXINT.  The requests[]
 * array is indexed by (requestId % REQUESTSSIZE).  As requestId goes from 0...MAXINT,
 * (requestId % REQUESTSSIZE) cycles from 0...REQUESTSSIZE-1, 0...REQUESTSSIZE-1,0...
 * This means that occasionally a slot in requests[] might be occupied when
 * (requestId % REQUESTSSIZE) cycles around to it again.  The code in GetNextrequestId
 * takes care of that.
 *-------------------------------------------------------------------------
 */
unsigned int
GetNextrequestId() {
    static unsigned int nextrequestId = 0;
    unsigned int returnVal;
    int numSlotsTried = 0;
    int startVal;
    bool retry = FALSE;
    
    pthread_mutex_lock(&requestIdLock);

    do {
	do {
	    if (nextrequestId != MAXINT) {
		returnVal = nextrequestId++;
	    } else {
		nextrequestId = 0;
		returnVal = 0;
	    }
	    numSlotsTried++;
	} while ((requests[returnVal % REQUESTSSIZE] != NULL)
		 && numSlotsTried < REQUESTSSIZE);

	/*
	 * All the slots in requests[] are taken.
	 * expand requests[].
	 *
	 * if (requestId % REQUESTSSIZE !=
	 *     requestId % oldREQUESTSSIZE) then
	 * 	requestId % REQUESTSSIZE > oldREQUESTSSIZE-1.
	 *
	 * Therefore, we can just check to see if
	 * requestId % REQUESTSSIZE == requestId % oldREQUESTSSIZE.
	 * If not, reassign requests[requestId % oldREQUESTSSIZE] to
	 * requests[requestId % REQUESTSSIZE].
	 *
	 */
	if (numSlotsTried == REQUESTSSIZE) {
	    int i, oldREQUESTSSIZE;
	    oldREQUESTSSIZE = REQUESTSSIZE;
	    retry = TRUE;
	    REQUESTSSIZE *= 2;
	    requests = realloc(requests, REQUESTSSIZE * sizeof(BrokerInfo *));
	    for (i = 0; i < oldREQUESTSSIZE; i++) {
		if (requests[i]->requestId % oldREQUESTSSIZE !=
		    requests[i]->requestId % REQUESTSSIZE) {
		    int newIndex = requests[i]->requestId % REQUESTSSIZE;
		    requests[newIndex] = requests[i];
		    requests[i] = NULL;
		}
		
	    }
	}
    } while (retry);
    
    pthread_mutex_unlock(&requestIdLock);

    return returnVal;
}

/*
 *-------------------------------------------------------------------------
 * BelowBidCurve
 *
 *     Returns TRUE if bid is below bid curve, FALSE else.
 *
 *-------------------------------------------------------------------------
 */
bool
BelowBidCurve(bid)
     Budget *bid;
{
    extern double bidCurveCost1;
    extern double bidCurveCost2;
    extern double bidCurveDelay1;
    extern double bidCurveDelay2;

    double cost, slope, xdiff;

    /*
     * Calculate what the value for cost is on
     * the bid curve at the delay specified in
     * the bid.
     */

    slope = (bidCurveCost1 - bidCurveCost2)/(bidCurveDelay1 - bidCurveDelay2);

    xdiff = (bid->delay - bidCurveDelay1);
    cost = bidCurveCost1 + (xdiff * slope);

    if (cost >= bid->cost) {
	return TRUE;
    } else {
	return FALSE;
    }
    
}

/*
 *-------------------------------------------------------------------------
 * TotalPlanBid-- Recursively add up the bids for an entire plan tree.
 *
 *-------------------------------------------------------------------------
 */
Budget *
TotalPlanBid(Plan *p) {

    Budget bid, *bidp;

    bid = p->budget;
    
    if (p->type == T_Merge) {
	List *lp;
	foreach (lp, ((Merge *)p)->mergeplans) {
	    Plan *p2 = (Plan *)lfirst(lp);
	    bidp = AddBids(&bid, &(p2->budget));
	}
    } else {
	if (p->lefttree != NULL) {
	    bidp = AddBids(&bid, TotalPlanBid(p->lefttree));
	}

	if (p->righttree != NULL) {
	    bidp = AddBids(&bid, TotalPlanBid(p->righttree));
	}
    }

    return &bid;
    
}

/*
 *-------------------------------------------------------------------------
 * AddBids--
 *
 *	Input: two bid *'s, b1, b2;
 *
 *	Output: b1 = b1 and b2 "added" together:
 *
 *		b1->cost = b1->cost + b2->cost;
 *		b1->delay = b1->delay + b2->delay;
 *		b1->staleness = MAX(b1->staleness, b2->staleness);
 *		b1->accuracy = MIN(b1->accuracy, b2->accuracy);
 *-------------------------------------------------------------------------
 */

Budget *
AddBids(Budget *b1, Budget *b2) {
    b1->cost = b1->cost + b2->cost;
    b1->delay = b1->delay + b2->delay;
    b1->staleness = Max(b1->staleness, b2->staleness);
    b1->accuracy = Min(b1->accuracy, b2->accuracy);

    return b1;
}
