/*-------------------------------------------------------------------------
 *
 * coord.c--
 *    the coordinator module (sending out contracts)
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/tcop/coord.c,v 1.25 1996/05/31 17:14:38 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <sys/socket.h>
#include <rpc/types.h>
#include "nodes/plannodes.h"
#include "nodes/parsenodes.h"
#include "executor/execdesc.h"
#include "tcop/coord.h"
#include "tcop/frag.h"
#include "utils/elog.h"
#include "exchange.h"
#include "queries.h"
#include "bid.h"
#include "mariposa_rpc.h"
#include "rpc/rpcutils.h"
#include "tcop/demo.h"
#include "parser/parsetree.h"

#include "sitemgr/sitemgr.h"
#include "nodes/pqa.h"

extern unsigned short udpPort;
extern unsigned short tcpPort;

extern double bidCurveCost1;
extern double bidCurveDelay1;

/*
 * Forward Declarations
 */
List *GetResultSites(QueryDesc *qd);
List *GetResultQepFragIds(QueryDesc *qd);
List *GetRecvSites(CoordPlan *cPlan, QueryDesc *qd);
List *GetRecvQepFragIds(CoordPlan *cPlan, QueryDesc *qd);
void NotifyReceivingSites(List *recvSites, List *recvQepFragIds, QueryDesc *qd);

/*
 *-------------------------------------------------------------------------
 * SendOffers--
 *
 *	Set up the offer buffer to send to each remote site.
 *	Contact the sites to start processing the subqueries.
 *-------------------------------------------------------------------------
 */
void
SendOffers(QueryDesc *qd)
{
    OfferBuffer offbuf;
    ProcessQueryArg *qry;
    List *lp;
    List *rt;
    RPCErrorCode result;
    RpcConn *rpcc;
    int i = 0;
    extern bool demo;
    List *recvSites = NIL;
    List *recvQepFragIds = NIL;
    
    offbuf.details.protocol = qd->protocol;
    
    switch (qd->protocol) {
    case SHORT_PROCESS_QUERY: {
	qry = &(offbuf.details.OfferDetails_u.shortQuery);
    }
	break;

    case LONG_PROCESS_QUERY: {
	qry = &(offbuf.details.OfferDetails_u.longQuery);
    }
	break;

    default: {
    }
	break;
	
    }

    /*
     * Fill in the fields in the offer buffer.
     */
    
    offbuf.isSubContract = false;

    offbuf.requestId = qd->requestId;
    
    offbuf.requester.addr = getMyAddr();
    offbuf.requester.port = tcpPort;
    offbuf.requester.hostid = getMyHostId();

    offbuf.protocol = qd->protocol;
    
    qry->homeSite.addr = offbuf.requester.addr;
    qry->homeSite.port = offbuf.requester.port;
    qry->homeSite.hostid = offbuf.requester.hostid;

    /*
     * Take care of result relation stuff:
     *
     *	- For DELETE's and UPDATE's, use the result relation
     * 	referred to in the parsetree.  The result fragment
     * 	will be filled in as the remote plans are sent out.
     *
     *	- For INSERT's, the result relation in the remote plan
     * 	sent out is null, but we need to send out separate
     * 	plans to the sites where the INSERT will be processed.
     *
     *	- For SELECT's, there is no result relation.
     */
    switch (qd->operation) {
    case CMD_DELETE:
    case CMD_UPDATE: {
	qry->resultRelation = qd->parsetree->resultRelation;
    }
	break;

    case CMD_INSERT: {
	qry->resultRelation = 0;
	qry->resultFragment = 0;
	recvSites = GetResultSites(qd);
	recvQepFragIds = GetResultQepFragIds(qd);
	NotifyReceivingSites(recvSites, recvQepFragIds, qd);
    }
	break;

    default: {
	qry->resultRelation = 0;
	qry->resultFragment = 0;
    }
	break;
    }
    
    /*
     * Send out each plan, along with the range table
     * and other information necessary for the processing site
     * to execute (or not) the query fragment specified.
     */
    foreach(lp, qd->remoteplans) {
	CoordPlan *cPlan = (CoordPlan *)lfirst(lp);
	Plan *plan = cPlan->plan;
	
	/*
	 * Fill in requestid and budget, along with offer details.
	 */
	offbuf.budget.cost = 0.0;
	offbuf.budget.delay = 0.0;
	offbuf.budget.staleness = 0.0;
	offbuf.budget.accuracy = 0.0;

	switch (qd->protocol) {
	    
	case LONG_PROCESS_QUERY: {
	    offbuf.contract = plan->contract;
	}
	    break;

	default: {
	    offbuf.contract = -1;
	}
	    break;
	}
	
	
	/*
	 * Fill in the operation with whatever the operation for the
	 * original query is.  The remote site will make the operation
	 * into a CMD_EXCHOUT if there are receiving sites specified
	 * in qry->recvSites.
	 */
	qry->operation = qd->operation;

	/*
	 * Fill in the plan tree and range table, casting them
	 * to the right types before sending them over the wire.
	 */
	qry->plan = (PQA_Plan_tree)plan;
	qry->rtable = (PQA_Node_list)qd->parsetree->rtable;

	/*
	 * Fill in the unique request identifier so the remote
	 * backend can identify which query it's processing.
	 */

	qry->requestId = qd->requestId;
	
	/*
	 * If this is an INSERT query and this is the node that will
	 * send to the site(s) processing the INSERT, do the following:
	 *
	 *	1) Set the partition mode so that tuples will be
	 *	   partitioned to the correct sites.
	 *	2) Set the list of receiving sites and receiving
	 *	   query plan fragment ID's so that the connections
	 *	   can be established between the sending site (the
	 *	   one being contacted right now) and the receiving
	 *	   site (which was contacted above in the call to
	 *	   NotifyReceivingSites().
	 *
	 * If this is any other type of query, there will be no partitioning
	 * (we may want to change this later - we could send out intermediate
	 * results to multiple sites and get a higher degree of parallelization).
	 * so partitionMode is NONE.  The receive sites and receive qepFragId's
	 * depend on the operation being performed.
	 *						     -JPS 2/4/96
	 */
	if (qd->operation == CMD_INSERT &&
	    cPlan->parentNode == NULL) {
	    RangeTblEntry *rte = rt_fetch(qry->resultRelation, qd->parsetree->rtable);
	    qry->partitionMode = rte->partitionMode;
	    qry->recvSites = (PQA_Node_list)recvSites;
	    qry->recvQepFragIds = (PQA_Int_list)recvQepFragIds;
	    
	} else {
	    qry->partitionMode = PARTITIONMODE_NONE;
	    qry->recvSites = (PQA_Node_list)GetRecvSites(cPlan, qd);
	    qry->recvQepFragIds = (PQA_Node_list)GetRecvQepFragIds(cPlan, qd);
	}

	/*
	 * Now fill in the fragment ID for the result relation
	 * fragment for a DELETE or UPDATE.
	 */
	if (qd->operation == CMD_DELETE ||
	    qd->operation == CMD_UPDATE) {
	    qry->resultFragment = ((Scan *)plan)->scanfragid;
	}

	if (demo) {
	    /* 	    printf("DEMO BROKER CHUNKSTATUS %i %i WON 0 0\n", ++i, addr.hostid); */
	}

	/*
	 * Now establish the connection to the site manager and
	 * make the call to ReceiveShortOffer() there.
	 */
	rpcc = Rpc_connectSitemgr(plan->processingSite.hostid, 
				  plan->processingSite.addr, 
				  "SendOffers", 
				  RPC_ONERR_WARN); 

	Rpc_call(rpcc, RPC_RECEIVESHORTOFFER, &offbuf, &result); 

	Rpc_disconnect(rpcc);
    }
}

/*
 *-------------------------------------------------------------------------
 * GetResultSites--
 *
 *	For INSERT's, return a list of the sites at which the
 *	insert will take place.
 *-------------------------------------------------------------------------
 */
List *
GetResultSites(QueryDesc *qd) {
	
    List *resultSites = NIL;
    int resultIndex = qd->parsetree->resultRelation;
    RangeTblEntry *rte = (RangeTblEntry *)nth(resultIndex-1, qd->parsetree->rtable);
    List *lp;
	
    foreach(lp, rte->fragInfo) {
	FragInfo *fInfo = (FragInfo *)lfirst(lp);
	List *lp2;
	List *lp3 = fInfo->processYN;
	foreach(lp2, fInfo->storageSites) {
	    SiteAddrNode *addr = (SiteAddrNode *)lfirst(lp2);
	    bool processYN = (bool)lfirst(lp3);
	    if (processYN) {
		SiteAddrNode *newAddr = makeNode(SiteAddrNode);
		CopySiteAddrNode(addr, newAddr);
		resultSites = lcons(newAddr, resultSites);
	    }
	}
    }
}

/*
 *-------------------------------------------------------------------------
 * GetResultQepFragIds--
 *
 *	For INSERT's, return a list of new QepFragIds corresponding
 *	to the INSERT sites' plans.
 *-------------------------------------------------------------------------
 */
List *
GetResultQepFragIds(QueryDesc *qd) {
	
    List *resultQepFragIds = NIL;
    int resultIndex = qd->parsetree->resultRelation;
    RangeTblEntry *rte = (RangeTblEntry *)nth(resultIndex-1, qd->parsetree->rtable);
    int nRecvSites = length(rte->fragInfo);
    int i;

    for (i = 0; i < nRecvSites; i++) {
	resultQepFragIds = lconsi(newQepFragId(), resultQepFragIds);
    }
	
}

/*
 *-------------------------------------------------------------------------
 * GetRecvSites--
 *
 * For non-update queries, the receive site is the processing site for the
 * parent node of the top node in a query plan fragment.
 *-------------------------------------------------------------------------
 */
List *
GetRecvSites(CoordPlan *cPlan, QueryDesc *qd) {
	
    List *recvSites = NIL;

    switch (qd->parsetree->commandType) {
    case CMD_SELECT: {
    
	SiteAddrNode *addr = makeNode(SiteAddrNode);
	addr->type = T_SiteAddrNode;
	
	addr->addr = cPlan->parentNode->processingSite.addr;
	addr->hostid = cPlan->parentNode->processingSite.hostid;

	/*
	 * If this node is sending back to the top-most qep fragment,
	 * then the port is known.  Otherwise, set it to 0.
	 */
	if (cPlan->parentNode->qepFragId == qd->plantree->qepFragId) {
	    addr->port = cPlan->parentNode->processingSite.port; 
	} else {
	    addr->port = 0;
	}
	
	recvSites = lcons(addr, NIL);
    }
	break;

	/*
	 * UPDATE and DELETE queries - no receiving site.
	 */
    case CMD_UPDATE:
    case CMD_DELETE: {
	
    }
	break;

    default:
	break;
    }

    return recvSites;
    
}

/*
 *-------------------------------------------------------------------------
 * GetRecvQepFragIds--
 *
 * For non-update queries, the recvQepFragId is the qepFragId of the
 * parent node of the top node in the current query fragment.
 *
 * For INSERT queries of the form "INSERT INTO foo SELECT * FROM bar;"
 * there is no parent node - the top node is a SELECT node.  So, we
 * assign a new qepFragId to the result node(s).
 *
 *-------------------------------------------------------------------------
 */
List *
GetRecvQepFragIds(CoordPlan *cPlan, QueryDesc *qd) {
	
    List *recvQepFragIds = NIL;

    switch (qd->parsetree->commandType) {
    case CMD_SELECT: {
    
	recvQepFragIds = lconsi(cPlan->parentNode->qepFragId, NIL);
    }
	break;

    case CMD_UPDATE:
    case CMD_DELETE: {
	
    }
	break;

    default:
	break;
    }

    return recvQepFragIds;
}

/*
 *-------------------------------------------------------------------------
 * NotifyReceivingSites--
 *
 * Create a plan consisting of a merge node with X-in nodes below it,
 * set up an rtable and other offer buffer stuff, then send it off.
 *-------------------------------------------------------------------------
 */
void
NotifyReceivingSites(List *recvSites, List *recvQepFragIds, QueryDesc *qd) {
    OfferBuffer offbuf;
    ProcessQueryArg *qry;
    List *lp, *lp2;
    Merge *merge = makeNode(Merge);
    RPCErrorCode result;
    RpcConn *rpcc;
    Plan *templatePlan = (Plan *)lfirst(qd->distPlanList);
    RangeTblEntry *rte;

    merge->mergeplans = NIL;
    merge->plan.targetlist = templatePlan->targetlist;
    merge->plan.lefttree = NULL;
    merge->plan.righttree = NULL;
    
    /*
     * Set up list of sending site addr's and sending
     * qepFragIds.  The sending sites are in the root
     * nodes of the plans in queryDesc->distPlanList.
     */

    foreach(lp, qd->distPlanList) {
	ExchangeIn *exchangeInPlan = makeNode(ExchangeIn);
	Plan *p = (Plan *)lfirst(lp);
	((Plan *)exchangeInPlan)->lefttree = NULL;
	((Plan *)exchangeInPlan)->righttree = NULL;
	((Plan *)exchangeInPlan)->targetlist = templatePlan->targetlist;
	exchangeInPlan->exchsendingSites = lcons(&p->processingSite, NIL);
	exchangeInPlan->exchsendingQepFragIds = lconsi(p->qepFragId, NIL);
	merge->mergeplans = lcons(exchangeInPlan, merge->mergeplans);
    }
    
    
    offbuf.details.protocol = SHORT_PROCESS_QUERY;
    qry = &(offbuf.details.OfferDetails_u.shortQuery);


    /*
     * Fill in the fields in the offer buffer.
     */
    
    offbuf.isSubContract = false;
    offbuf.requestId = qd->requestId;
    
    offbuf.budget.cost = 0.0;
    offbuf.budget.delay = 0.0;
    offbuf.budget.staleness = 0.0;
    offbuf.budget.accuracy = 0.0;

    qry->homeSite.addr = getMyAddr();
    qry->homeSite.port = tcpPort;
    qry->homeSite.hostid = getMyHostId();
    qry->operation = qd->operation;	/* ??? */
	
    offbuf.requester.addr = qry->homeSite.addr;
    offbuf.requester.port = qry->homeSite.port;
    offbuf.requester.hostid = qry->homeSite.hostid;

    qry->recvSites = (PQA_Node_list)NIL;

    qry->recvQepFragIds = (PQA_Int_list)NIL;
	
    qry->rtable = (PQA_Node_list)qd->parsetree->rtable;
    qry->resultRelation = qd->parsetree->resultRelation;
    qry->requestId = qd->requestId;
    
    lp2 = recvQepFragIds;
    
    /*
     * make a copy of the range table entry for the result relation --
     * we're going to modify 'rte' as go to make sure that we only
     * assign each result fragment to one site (actually, one
     * SiteAddrNode, since there can be more than one SiteAddrNode).
     */
    rte = copyObject(rt_fetch(qd->parsetree->resultRelation,
			      qd->parsetree->rtable));

    foreach(lp, recvSites) {
	SiteAddrNode *addr = (SiteAddrNode *)lfirst(lp);
	List *lp3;
	int fragIndex;
	
	CopySiteAddrNode(addr, &(merge->plan.processingSite));

	foreach(lp3, merge->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp3);
	    CopySiteAddrNode(addr, &(p->processingSite));
	}
	
	merge->plan.qepFragId = lfirsti(lp2);

	foreach(lp3, merge->mergeplans) {
	    Plan *p = (Plan *)lfirst(lp3);
	    p->qepFragId = lfirsti(lp2);
	}

	if (length(merge->mergeplans) > 1) {
	    qry->plan = (PQA_Plan_tree)merge;
	} else {
	    qry->plan = (PQA_Plan_tree)((Plan *)lfirst(merge->mergeplans));
	}

	fragIndex = 1;
	qry->resultFragment = 0;
	foreach(lp3, rte->fragInfo) {
	    FragInfo *fInfo = (FragInfo *)lfirst(lp3);
	    List *lp4;
	    List *lp5 = fInfo->processYN;
	    foreach(lp4, fInfo->storageSites) {
		bool processYN = (bool)lfirsti(lp5);
		SiteAddrNode *saddr = (SiteAddrNode *)lfirst(lp4);
		if (processYN && saddr->hostid == addr->hostid) {
		    qry->resultFragment = fragIndex;
		    saddr->hostid = 0;
		    break;
		}
		lp5 = lp5->next;
	    }
	    if (qry->resultFragment > 0) {
		break;
	    }
	    fragIndex++;
	}
	
	rpcc = Rpc_connectSitemgr(addr->hostid, 
				  addr->addr, 
				  "NotifyReceivingSites", 
				  RPC_ONERR_WARN); 

	Rpc_call(rpcc, RPC_RECEIVESHORTOFFER, &offbuf, &result); 

	Rpc_disconnect(rpcc);

	lp2 = lp2->next;
	
    }
    
}

/*
 *======================================================================================
 *				      LONG PROTOCOL
 *======================================================================================
 */
/*
 * I don't think we need to fill in an entire offer buffer -
 * the winning site already has all the information except
 * who the receiving site is.
 */

void NotifyContractWon(QueryDesc *queryDesc) {
    
    int i;

#ifdef 0
    
    for (i = 0; i < bAnswer->processingSites.HostList_len; i++) {
	RpcConn *rpcc;
	WinBuffer winBuffer;
	u_int remotehostid;
	RPCErrorCode result;
	
	/*
	 * Set up buffer to send to ContractWon
	 */
	winBuffer.contract = bAnswer->contracts.ContractList_val[i];
	winBuffer.recvSite.addr = localSiteAddr;
	winBuffer.recvSite.port = tcpPort;
	winBuffer.recvSite.hostid = localSiteHostId;
  
	/*
	 * make the RPC call
	 */
	remotehostid = bAnswer->processingSites.HostList_val[i];

	/*
	 * We add hostid to program number so we contact the correct sitemgr
	 */
	/*
	 * JPSXXX This must be changed - to notify the correct
	 * processing site.  I think we need more than just hostID,
	 * we need to send around the entire address.
	 */
	rpcc =
	    Rpc_connectSitemgr(remotehostid,
			       localSiteAddr,
			       "NotifyContractWon",
			       RPC_ONERR_WARN);

	Rpc_call(rpcc, RPC_CONTRACTWON, &winBuffer, &result);

	Rpc_disconnect(rpcc);

    }
#endif
    
}

