/*
 * demoRules.c--
 *
 *	Written to implement the following two rules, which are specific
 *	to the demo for the VLDB-Journal paper, and for David Hislop's
 *	visit.
 *
 *	RULE #1: if joinsWon - joinsLost < BUYTHRESHHOLD
 *		buy relationToBuy from relationToBuyHost
 *
 *	RULE #2: if nLocalQueryFrags/nQueryFrags > PARALLELTHRESHHOLD
 *		split last relation into nRemoteSites pieces and
 *		distribute to the nRemoteSites sites.
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include "c.h"
#include "postgres.h"
#include "nodes/pg_list.h" /* for List */
#include "tcop/tcopprot.h"
#include "nodes/nodes.h"
#include "nodes/plannodes.h"
#include "nodes/parsenodes.h"
#include <termios.h>


#define BUYTHRESHHOLD 0
#define PARALLELTHRESHHOLD 1

int joinsWon = 0;
int joinsLost = 0;
int relationToBuyStoreId;
int relationToBuyClassId;
int relationToBuyHostId;

bool demo = FALSE;

int newSite1, newSite2;
char *oldFragName;

int nodeNum = 0;		/* For demoPrintPlanTree */

void demoPrintPlanTree2(Plan *p, List *rtable, bool printHostIds);

void
RunDemoRules() {

    if (joinsLost > BUYTHRESHHOLD) {
	char query[1000];

	sprintf(query, "MOVE FRAGMENT %i %i FROM %i",
		relationToBuyClassId,
		relationToBuyStoreId,
		relationToBuyHostId);
	pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	
	joinsWon = 0;
	joinsLost = 0;
    }

    /*
     * JPSV1.0 - no nrecv
     */

#if 0
    if (nrecv == 0 && joinsWon > 0) {
	char newFragName1[100], newFragName2[100];
	char query[1000];

	joinsWon = 0;
	
	oldFragName[0] = 'r';
	
	strcpy(newFragName1, oldFragName);
	strcat(newFragName1, "a");
	strcpy(newFragName2, oldFragName);
	strcat(newFragName2, "b");

	sprintf(query, "SPLIT FRAGMENT %s INTO %s, %s",
		oldFragName,
		newFragName1,
		newFragName2);
	    
	pg_eval(query, (char *)NULL, (Oid *)NULL, 0);

	newSite1 = 2;
	newSite2 = 3;
	    
	sprintf(query, "MOVE FRAGMENT %s to %i",
		newFragName1,
		newSite1);
	pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	    
	sprintf(query, "MOVE FRAGMENT %s to %i",
		newFragName2,
		newSite2);
	pg_eval(query, (char *)NULL, (Oid *)NULL, 0);
	    
	fragmentJoin = TRUE;
    }
#endif
    
}

/*
 * demoPrintPlanTree--
 *
 * 	Prints out each plan node, where it's being processed, and indicates
 *	for sequential scans the relation being scanned.
 */
void
printPlanTree(Plan *p, List *rtable, int level) {

    int i;
    
    if (p == NULL) {
	return;
    }
    
    switch (p->type) {
	
    case T_MergeJoin: {
    
	for (i = 0; i < level; i++) {
	    printf("\t");
	}
	printf("MERGE-JOIN [%i]\n", p->processingSite.hostid);
	printPlanTree(p->lefttree, rtable, level+1);
	printPlanTree(p->righttree, rtable, level+1);
    }
	break;

    case T_SeqScan: {
	int scanrelid = ((Scan *)p)->scanrelid;

    
	for (i = 0; i < level; i++) {
	    printf("\t");
	}
	if (scanrelid > 0) {
	    RangeTblEntry *rte = (RangeTblEntry *)(nth((scanrelid-1), rtable));
	    printf("SCAN(%s) [%i]\n", rte->relname, p->processingSite.hostid);
	} else {
	    printf("SCAN(TEMP)\n");
	}
	
	printPlanTree(p->lefttree, rtable, level+1);
    }
	break;

    case T_Sort: {
    
	for (i = 0; i < level; i++) {
	    printf("\t");
	}
	printf("SORT [%i]\n", p->processingSite.hostid);
	printPlanTree(p->lefttree, rtable, level+1);
    }
	break;
	
    case T_Merge: {
	List *lp;
	
	for (i = 0; i < level; i++) {
	    printf("\t");
	}
	printf("MERGE [%i]\n", p->processingSite.hostid);
	foreach(lp, ((Merge *)p)->mergeplans) {
	    Plan *mp = (Plan *)lp->elem;
	    printPlanTree(mp, rtable, level+1);
	}
    }
	break;

    case T_ExchangeIn: {
    
	for (i = 0; i < level; i++) {
	    printf("\t");
	}
	printf("EXCHANGE-IN [%i]\n", p->processingSite.hostid);
	printPlanTree(p->lefttree, rtable, level+1);
    }
	break;
	
    default: {
	printPlanTree(p->lefttree, rtable, level);
	printPlanTree(p->righttree, rtable, level);
    }
	
    }
    
}

void
demoPrintPlanTree(Plan *p, List *rtable, int startNodeNum, bool printHostIds) {
    nodeNum = startNodeNum;
    demoPrintPlanTree2(p, rtable, printHostIds);
    printf("\n");
}

void
demoPrintPlanTree2(Plan *p, List *rtable, bool printHostIds) {

    int i;
    
    if (p == NULL) {
	return;
    }
    
    switch (p->type) {
	
    case T_MergeJoin: {
    
	printf("{%i MergeJoin ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	demoPrintPlanTree2(p->righttree, rtable, printHostIds);
	printf("} ");
    }
	break;

	
    case T_HashJoin: {
    
	printf("{%i HashJoin ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	demoPrintPlanTree2(p->righttree, rtable, printHostIds);
	printf("} ");
    }
	break;
	
    case T_NestLoop: {
    
	printf("{%i NestLoop ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	demoPrintPlanTree2(p->righttree, rtable, printHostIds);
	printf("} ");
    }
	break;

     case T_SeqScan: {
	int scanrelid = ((Scan *)p)->scanrelid;

    
	if (scanrelid > 0) {
	    RangeTblEntry *rte = (RangeTblEntry *)(nth((scanrelid-1), rtable));
	    printf("{%i Scan(%s) ", ++nodeNum, rte->relname);
	} else {
	    printf("{%i Scan(TEMP) ", ++nodeNum);
	}
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
	
    }
	break;

     case T_IndexScan: {
	int scanrelid = ((Scan *)p)->scanrelid;

    
	if (scanrelid > 0) {
	    RangeTblEntry *rte = (RangeTblEntry *)(nth((scanrelid-1), rtable));
	    printf("{%i I-Scan(%s) ", ++nodeNum, rte->relname);
	} else {
	    printf("{%i IndexScan(TEMP) ", ++nodeNum);
	}
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
	
    }
	break;

    case T_Sort: {
    
	printf("{%i Sort ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;

    case T_Hash: {
    
	printf("{%i Hash ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;
	
    case T_Merge: {
	List *lp;
	
    
	printf("{%i Merge ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}	
	foreach(lp, ((Merge *)p)->mergeplans) {
	    Plan *mp = (Plan *)lp->elem;
	    demoPrintPlanTree2(mp, rtable, printHostIds);
	}
	
	printf("} ");
	
    }
	break;

    case T_ExchangeIn: {
	printf("{%i X-in ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}	
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;

    case T_TupleBuffer: {
	printf("{%i Buffer ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}	
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;

    case T_Group: {
	printf("{%i Group-By ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}	
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;

    case T_Agg: {
	printf("{%i AGG ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}	
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	printf("} ");
    }
	break;
	
    default: {
    
	printf("{%i ??? ", ++nodeNum);
	if (printHostIds) {
	    printf(" %i ", p->processingSite.hostid);
	}
	demoPrintPlanTree2(p->lefttree, rtable, printHostIds);
	demoPrintPlanTree2(p->righttree, rtable, printHostIds);
	printf("} ");
    }
	break;

    }
	    
}

char *
noCR(char *s) {
    char cr;
    int i;
    int l = strlen(s);
    
    sprintf(&cr, "\n");

    for (i = 0; i < l; i++) {
	if (strncmp(&s[i], &cr, 1) == 0) {
	    s[i] = ' ';
	}
    }

    return s;
    
}

void demoPrintMsg(char *fmt, ...) {
    va_list ap;
    int rc;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);

    rc = fflush(stdout);
}
