/*-------------------------------------------------------------------------
 *
 * frag.c--
 *    the fragmenter (generates fragmented plans)
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/tcop/frag.c,v 1.65 1996/06/14 20:23:50 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include "nodes/plannodes.h"
#include "nodes/pg_list.h"
#include "executor/execdesc.h"
#include "access/heapam.h"
#include "tcop/frag.h"
#include "utils/elog.h"
#include "utils/rel.h"
#include "tcop/broker.h"
#include "nodes/parsenodes.h"
#include "catalog/pg_host.h"
#include "utils/nameserver.h"

static List *FragmentPlan(Plan *plan);
List *FragmentSubPlans(Plan *plan);
List *FragmentGenericPlan(Plan *plan);
List *FragmentSeqScan(SeqScan *plan);
List *FragmentMergeJoin(MergeJoin *plan);
void FragmentRTable(List *rt);
void FillInFragInfo(RangeTblEntry *rte, Form_pg_fragment frag, Oid fragOid,
    HeapScanDesc hostscan);
List *GetStorageFrags(int classid, List *rtable);
Plan *CreateMergeNode(Plan *plan, List *plans);

/*
 * Make range table accessible to everything in this file.
 * It's passed into FragmentQuery in the QueryDesc.
 *
 */
List *frag_rtable;

/*
 * Make operation available to entire file.
 */

int operation;

/*
 * fragmentation is an integer between 0 and 100 inclusive which defines
 * the likelihood that parallel plans WON'T be merged together.
 * That is, the higher the value of fragmentation, the more parallel
 * the plan will be.
 */

int fragmentation = 0;

#define MERGEYN ( (((int)((float)(rand()/(RAND_MAX/100)))) > fragmentation) && level > 0 )

/*
 *----------------------------------------------------------------
 *  FragmentQuery--
 *
 *	Top-level fragmenter routine.
 *
 *	Returns a list of plans, the union of which is the original
 *	plan.  If the query is an INSERT, DELETE or UPDATE, then
 *	the subplans are not MERGE'd into one, since the destination
 *	may not be the home site.  For a non-update query, the final
 *	destination is always the home site, so there is only one
 *	plan in the list.
 *----------------------------------------------------------------
 */
List *
FragmentQuery(QueryDesc *qd)
{

    /*
     * find the local address and hostid (XXX should be done only once)
     */
    localSiteAddr = getMyAddr();
    localSiteHostId = getMyHostId();

    /*
     * Initialize frag_rtable.
     */
    frag_rtable = qd->parsetree->rtable;
    operation = qd->operation;
    
    /*
     * Generate fragmented query plan.
     */    
    return FragmentPlan(qd->plantree);

}

/*
 *----------------------------------------------------------------
 * FragmentPlan--
 *
 *    	Walk through the plan tree and create a fragmented plan tree.
 *
 *----------------------------------------------------------------
 */
static List *
FragmentPlan(Plan *plan)
{
    List *subPlans;
    
    if (plan==NULL)
	return NIL;

    subPlans = FragmentSubPlans(plan);
    if (length(subPlans) > 1 && operation == CMD_SELECT) {
	subPlans = lcons(CreateMergeNode(plan, subPlans), NIL);
    }

    
    return subPlans;
}

/*
 *----------------------------------------------------------------
 * FragmentSubPlans--
 *
 * 	Input: 	Plan *
 *
 *	Output:	List of fragmented plans, created from the plan.
 *
 *	For plan types that are fragmented (sequential scans and
 *	joins, so far) after the list is generated, we have the
 *	following choice:
 *
 *	1) add a merge node above it.  This will decrease the number
 *	   of subplans created, at the expense of parallelism.
 *
 *	2) don't add a merge node.  In this case, there will be more
 *	   subplans (greater parallelism).
 *
 * We  consider sequential scans separately from other joins.
 * That is, we allow the option of keeping the parallelism in
 * sequential scans, but merge sub-join results together.
 *----------------------------------------------------------------
 */

List *
FragmentSubPlans(Plan *plan) {

    List *plans = NIL;
    List *lp;
    List *newplans;
    Plan *newplan;
    static int level = 0;
    
    if (plan == NULL) {
	return NULL;
    }
    
    switch (nodeTag(plan)) {

    case T_SeqScan: {

	/*
	 * If this is a scan of a base relation, fragment the
	 * scan.
	 */
	if (((SeqScan *)plan)->scanrelid != 0) {
	    level++;
	    plans = FragmentSeqScan((SeqScan *)plan);
	    level--;
	    if (MERGEYN && length(plans) > 1) {
		Plan *p = CreateMergeNode(plan, plans);
		plans = lcons(p, NIL);
	    }
	}

	/*
	 * Scan on a temporary relation - just process the
	 * left tree (there is no right tree).
	 *
	 * Get the list of plans resulting from fragmenting
	 * the left subtree.  Add a copy of the sequential
	 * scan node above each fragmented subplan.
	 */
	else {
	    List *leftPlans;
	    Plan *leftSave = plan->lefttree;

	    level++;
	    leftPlans = FragmentSubPlans(plan->lefttree);
	    level--;
	    
	    plan->lefttree = NULL;
	    foreach(lp, leftPlans) {
		Plan *newP = copyObject(plan);
		newP->lefttree = (Plan *)lp->elem;
		plans = lcons(newP, plans);
	    }
	    plan->lefttree = leftSave;
	}
	
    }
	break;

    case T_NestLoop:
    case T_MergeJoin: {
	level++;
	plans = FragmentGenericPlan(plan);
	level--;
	
	if (MERGEYN && length(plans) > 1) {
	    newplan = CreateMergeNode(plan, plans);
	    plans = lcons(newplan, NIL);
	}
    }
 	break;
	
    case T_Agg: {
	if (plan->lefttree != (Plan *) NULL && IsA(plan->lefttree,Group)) {
	    Group *gp = (Group *) plan->lefttree;

	    newplans = FragmentPlan(gp->plan.lefttree);
	    if (length(newplans) > 1) {
		gp->plan.lefttree = CreateMergeNode(gp->plan.lefttree, newplans);
	    } else {
		gp->plan.lefttree = lfirst(newplans);
	    }
	} else {
	    newplans = FragmentPlan(plan->lefttree);
	    if (length(newplans) > 1) {
		plan->lefttree = CreateMergeNode(plan->lefttree, newplans);
	    } else {
		plan->lefttree = lfirst(newplans);
	    }
	}
	plans = lcons(plan, NIL);
    }
	break;

    case T_Append: {
	Append *ap = (Append *) plan;

	/*
	 * we need to do the same initialization for these
	 * RangeTblEntry structures as we do for the regular range
	 * table.
	 */
	FragmentRTable(ap->unionrtentries);

	newplans = NIL;
	foreach (lp, ap->unionplans) {
	    newplans = nconc(newplans, FragmentPlan((Plan *) lfirst(lp)));
	}
	ap->unionplans = newplans;

	plans = lcons(plan, NIL);
    }
	break;

    default: {

	plans = FragmentGenericPlan(plan);

    }
	break;
	
    }

    return plans;
}

/*
 *----------------------------------------------------------------
 * FragmentGenericPlan--
 *
 * 	Input: 	Plan * to a generic plan node
 *
 *	Output: List of Plan *'s.  The left and right
 *	subtrees are recursively fragmented.  The number
 *	of plans returned by FragmentGenericPlan is the
 *	number of left plans times the number of right plans.
 *
 *----------------------------------------------------------------
 */

List *
FragmentGenericPlan(Plan *plan) {

    List *plans = NIL;
    
    Plan *leftSave = plan->lefttree;
    Plan *rightSave = plan->righttree;
    List *leftPlans = FragmentSubPlans(plan->lefttree);
    List *rightPlans = FragmentSubPlans(plan->righttree);
    List *lpLeft = leftPlans;
    List *lpRight = rightPlans;
    plan->lefttree = NULL;
    plan->righttree = NULL;
    
    do {
	do {
	    
	    Plan *newP = copyObject(plan);
	    newP->lefttree = (lpLeft != NIL) ? (Plan *)lpLeft->elem: NULL;
	    newP->righttree = (lpRight != NIL) ? (Plan *)lpRight->elem: NULL;
	    plans = lcons(newP, plans);
	    
	    if (lpRight != NIL) {
		lpRight = lpRight->next;
	    }
	} while (lpRight != NIL);

	lpRight = rightPlans;
	
	if (lpLeft != NIL) {
	    lpLeft = lpLeft->next;
	}
	
    } while (lpLeft != NIL);

    return plans;
}

/*
 *----------------------------------------------------------------
 * FragmentSeqScan--
 *
 * 	Input: 	Plan * to a sequential scan node.
 *
 *	Output: List of Plan *'s, one for each storage fragment
 *		of the relation being scanned.  Each plan in
 *		the list has a scanrelid equal to the one in
 *		the original plan.  The Broker and ProcessQueryDesc()
 *		can use the scanrelid to reference the correct
 *		entry in the range table, which contains all the
 * 		fragment information.
 *----------------------------------------------------------------
 */

List *
FragmentSeqScan(SeqScan *plan) {
    
    List *scans = NIL;		/* List of Plan *'s */
    List *f;
    RangeTblEntry *rte = (RangeTblEntry *)(nth(plan->scanrelid-1, frag_rtable));
    Index fragid = 1;
    
    /*
     * Create a plan node for each storage fragment and
     * add it to the list.
     */

    foreach(f, rte->fragInfo) {
	/*
	 * Using copyObject here is a short-cut we can use
	 * because this is a leaf-level plan node.
	 */
	SeqScan *s = copyObject(plan);
	s->scanfragid = fragid++;
	scans = lappend(scans, s);
    }
    
    return scans;
    
}

/*
 *----------------------------------------------------------------
 * FragmentMergeJoin--
 *
 *	Input: 	Plan * to a merge join node.
 *
 *	Output:	List of Plan *'s, one for each sub-join.
 *
 *
 *     	For A JOIN B, if n(A) = number of storage fragments for A,
 *	and likewise for B, there will n(A) * n(B) sub-joins, one
 *	for each storage fragment.  If the outer sub-plan is not a
 *	sequential scan, then assume it is a merge, in which case
 *	there will be n(B) sub-joins, one for each inner plan storage
 *	fragment.
 *----------------------------------------------------------------
 */

List *
FragmentMergeJoin(MergeJoin *plan) {
    List *joins = NIL;
    List *lp, *rp;
    
    /*
     * The merge node is always followed by a sequential scan,
     * then a sort, then a sequential scan, join, etc.  the
     * first sequential scan is on the temporary relation produced
     * by the sort.
     */
    List *leftPlans = FragmentSubPlans(plan->join.lefttree);
    List *rightPlans = FragmentSubPlans(plan->join.righttree);
    
    foreach(lp, leftPlans) {
	Plan *leftPlan = lp->elem;
	foreach(rp, rightPlans) {
	    Plan *rightPlan = rp->elem;
/* 	    MergeJoin *j = CreateMergeJoin(plan, leftPlan, rightPlan); */
/* 	    joins = lcons(j, joins); */
	}
    }
    
    return joins;
}


/*
 *======================================================================================
 *				      UTILITY ROUTINES
 *======================================================================================
 */

/*
 *----------------------------------------------------------------
 * GetStorageFrags--
 *
 *	Input:	OID of a storage class (relation)
 *
 *	Output:	List of RTE's with fralogicalid's of data fragments for class.
 *		I think the Broker will have to switch these logicalID's for
 *		the storeid's.  Or, perhaps the processing site should do this.
 *----------------------------------------------------------------
 */
List *
GetStorageFrags(int classid, List *rtable)
{
    List *frags = NIL;
    Relation fragdesc;
    HeapScanDesc fragscan;
    HeapTuple htup;
    RangeTblEntry *rte = (RangeTblEntry *)(nth((classid-1), rtable));
    Oid classoid = rte->relid;
    
    fragdesc = heap_openr((Name)Name_pg_fragment);
    if (!RelationIsValid(fragdesc))
	elog(WARN, "cannot open pg_fragment");

    fragscan = heap_beginscan(fragdesc,
			      0,
			      SelfTimeQual,
			      0, NULL);
    do {
	htup = heap_getnext(fragscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_fragment frag = (Form_pg_fragment)GETSTRUCT(htup);
	    
	    if (frag->fraclassid == classoid) {
		RangeTblEntry *newRTE = copyObject(rte);
		newRTE->relid = frag->fralogicalid;
		frags = lcons(newRTE, frags);
	    }
	}
    } while (htup);
    heap_endscan(fragscan);
    heap_close(fragdesc);

    return frags;
}

/*
 *----------------------------------------------------------------
 * FragmentRTable--
 *
 *	Input: 	Range Table
 *
 *	Output:	Nothing (void)
 *
 *	Opens pg_fragment and performs a sequential scan on it,
 *	filling in all the fragment information in the range
 *	table, except for the address and port of the fragment
 *	location. After this, fragment information provided by the
 * 	nameserver is added.
 *
 *	If this is a sequential scan directly on a storage fragment,
 * 	we fill in ONE fragInfo entry, even though this is slightly
 *	logically inconsistent.  A storage fragment, however, does
 *	have an entry in both pg_class and pg_fragment, and can
 *	therefore be considered to have one fragment associated with
 * 	it.
 *
 *	Therefore, for each entry in the range table, fragInfo should
 *	not be NIL.
 *
 *----------------------------------------------------------------
 */
void
FragmentRTable(List *rt) {
    Relation fragdesc;
    HeapScanDesc fragscan;
    HeapTuple htup;
    List *r;

    Relation hostdesc;
    HeapScanDesc hostscan;
    
    /*
     * Open pg_fragment and start sequential scan.
     */
    fragdesc = heap_openr((Name)Name_pg_fragment);
    if (!RelationIsValid(fragdesc))
	elog(WARN, "Fragmenter: FragmentRTable() cannot open pg_fragment");
  
    fragscan = heap_beginscan(fragdesc,
			      0,
			      SelfTimeQual,
			      0, NULL);

    /*
     * Open pg_host and start scan.
     */
    hostdesc = heap_openr((Name)Name_pg_host);
    if (!RelationIsValid(hostdesc))
        elog(WARN, "cannot open pg_host");
    hostscan = heap_beginscan(hostdesc, 0, SelfTimeQual, 0, NULL);

    /*
     * First, set all the fragInfo lists to be NIL
     */

    foreach(r, rt) {
	RangeTblEntry *rte = (RangeTblEntry *)r->elem;
	rte->fragInfo = NIL;
    }

    /*
     * Scan pg_fragment.  For each entry, check if it's in the range
     * table.  If so, fill in the fragment information.
     */
    do {
	htup = heap_getnext(fragscan, 0, NULL);
	if (HeapTupleIsValid(htup)) {
	    Form_pg_fragment frag = (Form_pg_fragment)GETSTRUCT(htup);
     
	    /*
	     * See if the current pg_fragment tuple matches any of
	     * the relations referenced in the range table.
	     */
	    foreach(r, rt) {
		RangeTblEntry *rte = (RangeTblEntry *)r->elem;

		if (frag->fraclassid == rte->relid) {
		    FillInFragInfo(rte, frag, htup->t_oid, hostscan);
		}
	    }
	}
    } while (htup);

    /*
     * Go through each entry in the range table and check if
     * each of the fragInfo lists is filled in.
     *
     * If any of the fragInfo lists are NIL, then this might be a
     * sequential scan directly on a storage fragment, rather
     * than on a base relation.  For example, if relation
     * FOO is fragmented into FOO1 and FOO2, and the user
     * types SELECT * from FOO2.  In the loop above, there
     * will be no entry in pg_fragment whose fraclassid equals
     * the rangeTblEntry's relid.  Instead, the rangeTblEntry's
     * relid will be equal to pg_fragment->frastoreid.
     */
    foreach(r, rt) {
	RangeTblEntry *rte = (RangeTblEntry *)r->elem;
	
	if (rte->fragInfo == NIL) {
	    heap_rescan(fragscan, FALSE, NULL);
	    do {
		htup = heap_getnext(fragscan, 0, NULL);
		if (HeapTupleIsValid(htup)) {
		    Form_pg_fragment frag = (Form_pg_fragment)GETSTRUCT(htup);
		    if (frag->frastoreid == rte->relid) {
			FillInFragInfo(rte, frag, htup->t_oid, hostscan);
		    }
		}
	    } while (htup);
	}
    }
    
    /*
     * Go through each entry in the range table and add
     * fragment information from the nameserver.
     */
    foreach(r, rt) {
	RangeTblEntry *rte = (RangeTblEntry *) r->elem;
	NsvcLookupFragsResult *lookupResult;
	int i;

	lookupResult = nsvc_lookupFrags(rte->relname->data, rte->relid);
	for (i = 0; i < lookupResult->fragTuples.TupleArray_len; i++) {
	    HeapTuple tup = (HeapTuple) 
	        lookupResult->fragTuples.TupleArray_val[i].Tuple_val;
	    Form_pg_fragment frag = (Form_pg_fragment) GETSTRUCT(tup);

	    FillInFragInfo(rte, frag, tup->t_oid, hostscan);
	}
	nsvc_freeLookupFragsResult(lookupResult);
    }

    /* 
     * Check if there are any range table entries with
     * fragInfo == NIL left. If so, we have to reject the query.
     */
    foreach(r, rt) {
	RangeTblEntry *rte = (RangeTblEntry *)r->elem;

	if (rte->fragInfo == NIL) {
	    elog(WARN, "FragmentRTable: no fragmentation info on table '%d'",
		 rte->relid);
	}
    }

    heap_endscan(fragscan);
    heap_close(fragdesc);

    heap_endscan(hostscan);
    heap_close(hostdesc);
}

/*
 *----------------------------------------------------------------
 * FillInFragInfo--
 *
 *	Fills in the fragInfo for the range table entry
 *
 * 	If this pg_fragment tuple is for a fragment that
 * 	is part of the relation referenced in the range table,
 * 	check if there is already an entry for this fragment
 * 	in the fragInfo for the range table entry (which means
 * 	that this tuple is for a copy of the same fragment).
 * 	If so, add this location to the list of locations for
 * 	the fragment.  If not, create the fragInfo entry.
 * 	If there's a SiteAddrNode with the same oid, this is a 
 *	duplicate fragment tuple and has to be ignored. Duplicates
 * 	can occur if for instance the nameserver is also subscribed
 * 	to our local metadata.
 *----------------------------------------------------------------
 */

void
FillInFragInfo(RangeTblEntry *rte, Form_pg_fragment frag, Oid fragOid,
    HeapScanDesc hostscan)
{
    List *f = NIL;
    List *s = NIL;
    FragInfo *fInfo;
    SiteAddrNode *sSite;
    bool found = FALSE;
    SiteAddrNode *saddr = (SiteAddrNode *)calloc(1, sizeof(SiteAddrNode));
    bool processYN = FALSE;
    int myHostId = getMyHostId();
    int myAddr = getMyAddr();
    extern unsigned short tcpPort;
    HeapTuple htupHost;

    /* make sure this pg_fragment tuple hasn't been recorded yet */
    foreach(f, rte->fragInfo) {
	fInfo = (FragInfo *) f->elem;
	foreach(s, fInfo->storageSites) {
	    sSite = (SiteAddrNode *) s->elem;
	    if (sSite->pg_frag_oid == fragOid) {
		/* we have seen this before */
		return;
	    }
	}
    }

    foreach(f, rte->fragInfo) {
	fInfo = (FragInfo *)f->elem;
	if (fInfo->fralogicalid == frag->fralogicalid) {
	    found = TRUE;
	    break;
	}
    }
		    
    if (!found) {
	fInfo = calloc(1, sizeof(FragInfo));
	fInfo->type = T_FragInfo;
	fInfo->storageSites = NIL;
	fInfo->processYN = NIL;
	fInfo->fralogicalid = frag->fralogicalid;
	fInfo->frastoreid = frag->frastoreid;
	fInfo->frapages = frag->frapages;
	fInfo->fratuples = frag->fratuples;
	rte->fragInfo = lcons(fInfo, rte->fragInfo);
    }

    saddr->type = T_SiteAddrNode;
    saddr->hostid = frag->fralocation;
    saddr->pg_frag_oid = fragOid;

    /*
     * Assign the address, based on the hostid.
     *
     * If the fragment is stored locally, fill in the address AND
     * the port number for this backend (that is, the backend that
     * includes the Query Broker).  If scans on these fragments are
     * processed by this backend, then this port # is correct.
     * Otherwise, the port needs to be set to 0, because scans will
     * be processed by some other backend at this site.
     */
    if (saddr->hostid == myHostId) {
	saddr->addr = myAddr;
	saddr->port = 0;
    } else {
	saddr->port = 0;
	heap_rescan(hostscan, 0, NULL);
	do {
	    htupHost = heap_getnext(hostscan, 0, NULL);
	    if (HeapTupleIsValid(htupHost)) {
		Form_pg_host host = (Form_pg_host)GETSTRUCT(htupHost);
		if (host->hostid == saddr->hostid) {
		    saddr->addr  = host->address;
		    break;
		}
	    }
	} while (htupHost);
    }
		    
	    
    fInfo->storageSites = lcons((void *)saddr, fInfo->storageSites);
    fInfo->processYN = lconsi(processYN, fInfo->processYN);
}
		    

/*
 *----------------------------------------------------------------
 * CreateMergeNode
 *
 *	Input:	Original plan (for targetlist)
 *		List of plans above which we place a merge node.
 *
 *	Output:	Plan with a merge node as the top.
 *
 *----------------------------------------------------------------
 */
Plan *
CreateMergeNode(Plan *plan, List *plans)
{
    Merge *mergenode;

    mergenode = makeNode(Merge);

    /*
     * JPS - I guess we need this.  The targetlist
     * is the PROJECT fields of the query.  Apparently
     * it's copied all over the place, so we need it
     * here.
     */
    mergenode->plan.targetlist = plan->targetlist;

    
    mergenode->mergeplans = plans;

    /*
     * JPS - Why is this automatically set to 1?
     * If we're just doing a sequential scan, are we
     * necessarily sorting?
     */
    mergenode->keycount = 1;

    return (Plan*)mergenode;

}
