/*-------------------------------------------------------------------------
 *
 * schedule.c--
 *    schedules queries that work on tertiary memory relations
 *
 * [missing RCS id]
 *-------------------------------------------------------------------------
 */
#include "c.h"
#include "postgres.h"
#include "nodes/pg_list.h"
#include "utils/rel.h"
#include "nodes/plannodes.h"
#include "tcop/schedule.h"
#include "storage/smgr.h"

#define BLOCKED    23
#define UN_BLOCKED 0

/* ------------------
 * a join will have both rtree and ltree Non-Null whereas a single rel.
 * scan will have both of them NULL and the relid and node will point to
 * the scanned relation and node respectively.
 * --------------
 */

/* -------------------------------------
 * this mechanism of blocking and unblocking nodes will not work for relations
 * stored on magnetic disk. Suppose if one inner frag. is blocked and outer
 * resides on magnetic disk -- it is necessary to rescan the outer frag when
 * the inner gets unblocked. This might interfere with other inner frags. which
 * are on mag. disk
 * ---------------------------
 */

static List * pendingList = NIL;
static List * unExecQList = NIL;
static List * scheduleList = NIL;
static List * doneList = NIL;
static List * currTasks = NIL;
static int commThreadInitiated = 0;

static List * localList = NIL;
static void ScheduleNextSubQuery();
static void SubmitTaskForExecution(List * task);
static void BlockSubQuery(SubQuery *squery);
static void UnblockSubQuery(SubQuery *squery);
static List * GetQueryIdList(List * list);

void  AllDone(int *isDone);
void  unblockNode();
void  blockNode();

/* ----------------------------------------------------
 * given a list of subqueries -- this routine submits the list
 * to the scheduler
 * ------------------------
 */
informScheduler(List *list)
{
    if (!commThreadInitiated) {
	commThreadInitiated = 1;
    }
    pendingList = nconc(pendingList, list);
    unExecQList = nconc(unExecQList, GetQueryIdList(list));
 
    if (pendingList != NIL) {
	schedulerSubmitQuery(getpid(), pendingList); 
	pendingList = NIL;
    }
}

/* -----------------------------------------------------
 * subQueryDone() -- called by the executor when all the subqueries
 * which are on unblocked nodes have been processed. As a side effect
 * this routines calls the scheduler for more pending queries and
 * if it gets pending subqueries it might unblock some more nodes
 * -----------------
 */
bool
subQueryDone()
{
    SubQuery *squery;
    int isDone;
    List * task;
    /* ---------------------------------
     * set all original nodes in curr task to unavail 
     * again because the scheduler might unpin it 
     * ---------------
     */

    foreach (task, currTasks) { 
	squery = lfirst(task);
    	BlockSubQuery (squery);
    }
    doneList = nconc(doneList, currTasks);
    if (doneList != NIL) {
	schedulerDoneSQ(getpid());
	doneList = NIL;
	currTasks = NIL;
    }
    AllDone(&isDone);
    if (!isDone) {
	ScheduleNextSubQuery();
	return false;
    }
    return true;
}

static void
ScheduleNextSubQuery()
{
    List * task;

    if (scheduleList == NIL) {
	schedulerGetNextSQ(getpid(), &scheduleList);
    }
    currTasks = scheduleList;
    scheduleList = NIL;
    foreach (task, currTasks) {
	SubmitTaskForExecution(task);
    }
}
       
/* -------------------------
 * 
 */
static void
SubmitTaskForExecution(List * task)
{
    SubQuery *squery;
    
    /* set plan nodes appropriately */
    squery = lfirst(task);
    
    unExecQList = intLispRemove(QueryId(squery), unExecQList);
    
    UnblockSubQuery(squery);
    
    /* wake up main thread          */
    return;
}

static void
UnblockSubQuery(SubQuery * squery)
{
    if (squery == NULL) return;

    if (!JoinQuery(squery)) {
	/* single relation query */
	unblockNode((Scan *)NodeStore(squery));
    } else {
	UnblockSubQuery(innerPlan(squery));
	UnblockSubQuery(outerPlan(squery));
    }
}


static void
BlockSubQuery(SubQuery * squery)
{
    if (squery == NULL) return;

    if (!JoinQuery(squery)) {
	/* single relation query */
	blockNode((Scan *)NodeStore(squery));
    } else {
	BlockSubQuery(innerPlan(squery));
	BlockSubQuery(outerPlan(squery));
    }
}

PrintSQ (SubQuery * squery)
{
    if (squery == NULL) return;

    if (!JoinQuery(squery)) {
	printf ("Relid = %d\n", RelationId(squery));
    } else { 
	PrintSQ (innerPlan(squery));
	PrintSQ (outerPlan(squery));
    }
}

void
AllDone(int *isDone)
{
    *isDone = (unExecQList == NIL)?1:0; 
}

/* ---------------------------------------------------------------------
 * interface routines between executor and the comm. thread
 * -----------------------------------------------------
 */
void *
makeSubQuery(Oid relid, void * node, void * oPlan, void * iPlan)
{
    /* create a plan node */
    SubQuery * squery = (SubQuery *)makeNode(Plan);

    QueryType(squery) = (oPlan == NULL)?T_Scan:T_Join;
    outerPlan(squery) = oPlan;
    innerPlan(squery) = iPlan;
    if (!JoinQuery(squery)) {
	RelationId(squery) = relid;
	NodeStore(squery)  = node;
    }
    return (void *) squery;
}

bool    
nodeNotAvail(Scan * node)
{
    return ((node->availState == BLOCKED)?true:false);
}

void
unblockNode(Scan * node)
{
    node->availState = UN_BLOCKED;
    /* rescan node */
    /* this will not work for nested loop where the inner relation has
     * a runtime index scan
     */
    ExecReScan (node, NULL, NULL);
}

void
blockNode(Scan * node)
{
    node->availState = BLOCKED;
}

bool
relationOnTert(Relation rd)
{
    return smgrisTM(rd->rd_rel->relsmgr);
}

/* ----------------------------------------------
 * given a list of subqueries -- assign unique ids to the subqueries
 * and return them in a list form
 */
static List *
GetQueryIdList(List *list)
{
    List * sl, * qList = NIL;
    static int queryId = 0;
    
    foreach (sl, list) {
	SubQuery * squery = lfirst(sl);

	qList = lconsi(queryId, qList);
	QueryId(squery) = queryId++;
    }
    return qList;
}

