/*-------------------------------------------------------------------------
 *
 * syscache.c--
 *    System cache management routines
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/utils/cache/syscache.c,v 0.5 1996/06/07 05:01:27 am Exp $
 *
 * NOTES
 *    These routines allow the parser/planner/executor to perform
 *    rapid lookups on the contents of the system catalogs.
 *
 *    see catalog/syscache.h for a list of the cache id's
 *
 *-------------------------------------------------------------------------
 */
#include "c.h"
 
#include "access/heapam.h"
#include "access/htup.h"
#include "catalog/catname.h"
#include "utils/catcache.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "nodes/pg_list.h"
 
/* ----------------
 *	hardwired attribute information comes from system catalog files.
 * ----------------
 */
#include "catalog/pg_am.h"
#include "catalog/pg_amop.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_group.h"
#include "catalog/pg_index.h"
#include "catalog/pg_inherits.h"
#include "catalog/pg_language.h"
#include "catalog/pg_opclass.h"
#include "catalog/pg_operator.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_class.h"
#include "catalog/pg_type.h"
#include "catalog/pg_rewrite.h"
#include "catalog/pg_aggregate.h"
#include "catalog/pg_user.h"
#include "storage/large_object.h"
#include "catalog/pg_listener.h"
 
extern bool	AMI_OVERRIDE;	/* XXX style */
 
#include "utils/syscache.h"
#include "catalog/indexing.h"
    
typedef HeapTuple (*ScanFunc)();

/* ----------------
 *	Warning:  cacheinfo[] below is changed, then be sure and
 *	update the magic constants in syscache.h!
 * ----------------
 */
static struct cachedesc cacheinfo[] = {
#ifndef P95
    { &AccessMethodOperatorRelationName,	/* AMOPOPID */
	  2,
#else /* P95 */
    { AccessMethodOperatorRelationName,	/* AMOPOPID */
	  3,
#endif /* P95 */
	  { Anum_pg_amop_amopclaid,
		Anum_pg_amop_amopopr,
#ifndef P95
		0,
#else /* P95 */
		Anum_pg_amop_amopid,
#endif /* P95 */
		0 },
	  sizeof(FormData_pg_amop),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &AccessMethodOperatorRelationName,	/* AMOPSTRATEGY */
#else /* P95 */
    { AccessMethodOperatorRelationName,	/* AMOPSTRATEGY */
#endif /* P95 */
	  3,
	  { Anum_pg_amop_amopid,
		Anum_pg_amop_amopclaid,
		Anum_pg_amop_amopstrategy,
		0 },
	  sizeof(FormData_pg_amop),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &AttributeRelationName,			/* ATTNAME */
#else /* P95 */
    { AttributeRelationName,			/* ATTNAME */
#endif /* P95 */
	  2,
	  { Anum_pg_attribute_attrelid,
		Anum_pg_attribute_attname,
		0,
		0 },
	  ATTRIBUTE_TUPLE_SIZE,
#ifndef P95
      &AttributeNameIndex,
#else /* P95 */
      AttributeNameIndex,
#endif /* P95 */
      (ScanFunc) AttributeNameIndexScan  },
#ifndef P95
    { &AttributeRelationName,			/* ATTNUM */
#else /* P95 */
    { AttributeRelationName,			/* ATTNUM */
#endif /* P95 */
	  2,
	  { Anum_pg_attribute_attrelid,
		Anum_pg_attribute_attnum,
		0,
		0 },
	  ATTRIBUTE_TUPLE_SIZE,
#ifndef P95
      &AttributeNumIndex,
#else /* P95 */
      AttributeNumIndex,
#endif /* P95 */
      (ScanFunc) AttributeNumIndexScan  },
#ifndef P95
    { &IndexRelationName,			/* INDEXRELID */
#else /* P95 */
    { IndexRelationName,			/* INDEXRELID */
#endif /* P95 */
	  1,
	  { Anum_pg_index_indexrelid,
		0,
		0,
		0 },
	  offsetof(FormData_pg_index, indpred),
      NULL,
      NULL  },
#ifndef P95
    { &LanguageRelationName,			/* LANNAME */
#else /* P95 */
    { LanguageRelationName,			/* LANNAME */
#endif /* P95 */
	  1,
	  { Anum_pg_language_lanname,
		0,
		0,
		0 },
	  offsetof(FormData_pg_language, lancompiler),
      NULL,
      NULL  },
#ifndef P95
    { &OperatorRelationName,			/* OPRNAME */
#else /* P95 */
    { OperatorRelationName,			/* OPRNAME */
#endif /* P95 */
	  4,
	  { Anum_pg_operator_oprname,
		Anum_pg_operator_oprleft,
		Anum_pg_operator_oprright,
		Anum_pg_operator_oprkind },
	  sizeof(FormData_pg_operator),
      NULL,
      NULL  },
#ifndef P95
    { &OperatorRelationName,			/* OPROID */
#else /* P95 */
    { OperatorRelationName,			/* OPROID */
#endif /* P95 */
	  1,
	  { ObjectIdAttributeNumber,
		0,
		0,
		0 },
	  sizeof(FormData_pg_operator),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &ProcedureRelationName,			/* PRONAME */
#else /* P95 */
    { ProcedureRelationName,			/* PRONAME */
#endif /* P95 */
	  3,
	  { Anum_pg_proc_proname,
	    Anum_pg_proc_pronargs,
	    Anum_pg_proc_proargtypes,
		0 },
	  offsetof(FormData_pg_proc, prosrc),
#ifndef P95
      &ProcedureNameIndex,
#else /* P95 */
      ProcedureNameIndex,
#endif /* P95 */
      (ScanFunc) ProcedureNameIndexScan  },
#ifndef P95
    { &ProcedureRelationName,			/* PROOID */
#else /* P95 */
    { ProcedureRelationName,			/* PROOID */
#endif /* P95 */
	  1,
	  { ObjectIdAttributeNumber,
		0,
		0,
		0 },
	  offsetof(FormData_pg_proc, prosrc),
#ifndef P95
      &ProcedureOidIndex,
#else /* P95 */
      ProcedureOidIndex,
#endif /* P95 */
      (ScanFunc) ProcedureOidIndexScan  },
#ifndef P95
    { &RelationRelationName,			/* RELNAME */
#else /* P95 */
    { RelationRelationName,			/* RELNAME */
#endif /* P95 */
	  1,
	  { Anum_pg_class_relname,
		0,
		0,
		0 },
	  CLASS_TUPLE_SIZE,
#ifndef P95
      &ClassNameIndex,
#else /* P95 */
      ClassNameIndex,
#endif /* P95 */
      (ScanFunc) ClassNameIndexScan  },
#ifndef P95
    { &RelationRelationName,			/* RELOID */
#else /* P95 */
    { RelationRelationName,			/* RELOID */
#endif /* P95 */
	  1,
	  { ObjectIdAttributeNumber,
		0,
		0,
		0 },
	  CLASS_TUPLE_SIZE,
#ifndef P95
      &ClassOidIndex,
#else /* P95 */
      ClassOidIndex,
#endif /* P95 */
      (ScanFunc) ClassOidIndexScan  },
#ifndef P95
    { &TypeRelationName,			/* TYPNAME */
#else /* P95 */
    { TypeRelationName,			/* TYPNAME */
#endif /* P95 */
	  1,
	  { Anum_pg_type_typname,
		0,
		0,
		0 },
	  offsetof(TypeTupleFormData,typalign)+sizeof(char),
#ifndef P95
      &TypeNameIndex,
#else /* P95 */
      TypeNameIndex,
#endif /* P95 */
      TypeNameIndexScan  },
#ifndef P95
    { &TypeRelationName,			/* TYPOID */
#else /* P95 */
    { TypeRelationName,			/* TYPOID */
#endif /* P95 */
	  1,
	  { ObjectIdAttributeNumber,
		0,
		0,
		0},
	  offsetof(TypeTupleFormData,typalign)+sizeof(char),
#ifndef P95
      &TypeOidIndex,
#else /* P95 */
      TypeOidIndex,
#endif /* P95 */
      TypeOidIndexScan  },
#ifndef P95
    { &AccessMethodRelationName,		/* AMNAME */
#else /* P95 */
    { AccessMethodRelationName,		/* AMNAME */
#endif /* P95 */
	  1,
	  { Anum_pg_am_amname,
		0,
		0,
		0},
	  sizeof(FormData_pg_am),
      NULL,
      NULL  },
#ifndef P95
    { &OperatorClassRelationName,		/* CLANAME */
#else /* P95 */
    { OperatorClassRelationName,		/* CLANAME */
#endif /* P95 */
	  1,
	  { Anum_pg_opclass_opcname,
		0,
		0,
		0},
	  sizeof(FormData_pg_opclass),
      NULL,
      NULL  },
#ifndef P95
    { &IndexRelationName,			/* INDRELIDKEY */
#else /* P95 */
    { IndexRelationName,			/* INDRELIDKEY */
#endif /* P95 */
	  2,
	  { Anum_pg_index_indrelid,
		Anum_pg_index_indkey,
		0,
		0},
	  offsetof(FormData_pg_index, indpred),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &InheritsRelationName,			/* INHRELID */
#else /* P95 */
    { InheritsRelationName,			/* INHRELID */
#endif /* P95 */
	  2,
	  { Anum_pg_inherits_inhrel,
		Anum_pg_inherits_inhseqno,
		0,
		0},
	  sizeof(FormData_pg_inherits),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &RewriteRelationName,			/* RULOID */
#else /* P95 */
    { RewriteRelationName,			/* RULOID */
#endif /* P95 */
	  1,
	  { ObjectIdAttributeNumber,
		0,
		0,
		0 },
	  offsetof(FormData_pg_rewrite, ev_qual),
      NULL,
      (ScanFunc) NULL  },
#ifndef P95
    { &AggregateRelationName,			/*AGGNAME*/
	  1,
#else /* P95 */
    { AggregateRelationName,			/*AGGNAME*/
	  2,
#endif /* P95 */
	  { Anum_pg_aggregate_aggname,
#ifndef P95
	        0,
#else /* P95 */
	    Anum_pg_aggregate_aggbasetype,
#endif /* P95 */
		0,
		0 },
	   offsetof(FormData_pg_aggregate, agginitval1),
       NULL,
       (ScanFunc) NULL  },    
#ifndef P95
    { &ListenerRelationName,                  /* LISTENREL */
#else /* P95 */
    { ListenerRelationName,                  /* LISTENREL */
#endif /* P95 */
	2,
	{  Anum_pg_listener_relname,
	     Anum_pg_listener_pid,
	     0,
	     0 },
	sizeof(FormData_pg_listener),
	NULL,
	(ScanFunc) NULL  },
#ifndef P95
    { &UserRelationName,           		/* USENAME */
#else /* P95 */
    { UserRelationName,           		/* USENAME */
#endif /* P95 */
	      1,
	      { Anum_pg_user_usename,
			0,
			0,
			0 },
	      sizeof(FormData_pg_user),
	      NULL,
	      (ScanFunc) NULL  },
#ifndef P95
    { &UserRelationName,           		/* USESYSID */
#else /* P95 */
    { UserRelationName,           		/* USESYSID */
#endif /* P95 */
	      1,
	      { Anum_pg_user_usesysid,
			0,
			0,
			0 },
	      sizeof(FormData_pg_user),
	      NULL,
	      (ScanFunc) NULL  },
#ifndef P95
    { &GroupRelationName,           		/* GRONAME */
#else /* P95 */
    { GroupRelationName,           		/* GRONAME */
#endif /* P95 */
	      1,
	      { Anum_pg_group_groname,
			0,
			0,
			0 },
	      offsetof(FormData_pg_group, grolist[0]),
	      NULL,
	      (ScanFunc) NULL  },
#ifndef P95
    { &GroupRelationName,           		/* GROSYSID */
#else /* P95 */
    { GroupRelationName,           		/* GROSYSID */
#endif /* P95 */
	      1,
	      { Anum_pg_group_grosysid,
			0,
			0,
			0 },
	      offsetof(FormData_pg_group, grolist[0]),
	      NULL,
	      (ScanFunc) NULL  },
#ifndef P95
    { &RewriteRelationName,           		/* REWRITENAME */
#else /* P95 */
    { RewriteRelationName,           		/* REWRITENAME */
#endif /* P95 */
	      1,
	      { Anum_pg_rewrite_rulename,
			0,
			0,
			0 },
	      offsetof(FormData_pg_rewrite, ev_qual),
	      NULL,
	      (ScanFunc) NULL  },
#ifndef P95
    { &ProcedureRelationName,                   /* PROSRC */
#else /* P95 */
    { ProcedureRelationName,                   /* PROSRC */
#endif /* P95 */
	   1,
           { Anum_pg_proc_prosrc,
		  0,
		  0,
		  0 },
	   offsetof(FormData_pg_proc, prosrc),
#ifndef P95
      &ProcedureSrcIndex,
      (ScanFunc) ProcedureSrcIndexScan  },
    { &FragRelationName,			/* FRALOGICALID */
	  1,
	  { Anum_pg_fragment_fralogicalid,
		0,
		0,
		0 },
	  offsetof(FormData_pg_fragment, fralokey),
	  NULL,
	  (ScanFunc) NULL }
#else /* P95 */
      ProcedureSrcIndex,
      (ScanFunc) ProcedureSrcIndexScan  }
#endif /* P95 */
};
 
static struct catcache	*SysCache[lengthof(cacheinfo)];
static int32		SysCacheSize = lengthof(cacheinfo);
    
    
/*
 * zerocaches--
 *
 *    Make sure the SysCache structure is zero'd.
 */
void
zerocaches()
{
    memset((char *) SysCache, 0, SysCacheSize * sizeof(struct catcache *));
}

/*
 * Note:
 *	This function was written because the initialized catalog caches
 *	are used to determine which caches may contain tuples which need
 *	to be invalidated in other backends.
 */
void
InitCatalogCache()
{
    int	cacheId;	/* XXX type */
    
    if (!AMI_OVERRIDE) {
	for (cacheId = 0; cacheId < SysCacheSize; cacheId += 1) {
	    
	    Assert(!PointerIsValid((Pointer)SysCache[cacheId]));
	    
	    SysCache[cacheId] =
#ifndef P95
		InitSysCache((*cacheinfo[cacheId].name)->data, 
#else /* P95 */
		InitSysCache(cacheinfo[cacheId].name,
#endif /* P95 */
			     cacheinfo[cacheId].indname,
			     cacheId,
			     cacheinfo[cacheId].nkeys,
			     cacheinfo[cacheId].key,
			     cacheinfo[cacheId].iScanFunc);
	    if (!PointerIsValid((char *)SysCache[cacheId])) {
		elog(WARN,
		     "InitCatalogCache: Can't init cache %.16s(%d)",
#ifndef P95
		     (*cacheinfo[cacheId].name)->data,
#else /* P95 */
		     cacheinfo[cacheId].name,
#endif /* P95 */
		     cacheId);
	    }
	    
	}
    }
}

/*
 * SearchSysCacheTuple--
 *
 *    A layer on top of SearchSysCache that does the initialization and
 *    key-setting for you.
 *
 * Returns the tuple if one is found, NULL if not.
 *
 * XXX The tuple that is returned is NOT supposed to be pfree'd!
 */
HeapTuple
SearchSysCacheTuple(int cacheId,	/* cache selection code */
#ifndef P95
		    char *key1,
		    char *key2,
		    char *key3,
		    char *key4)
#else /* P95 */
		    Datum key1,
		    Datum key2,
		    Datum key3,
		    Datum key4)
#endif /* P95 */
{
    register HeapTuple	tp;
    
    if (cacheId < 0 || cacheId >= SysCacheSize) {
	elog(WARN, "SearchSysCacheTuple: Bad cache id %d", cacheId);
	return((HeapTuple) NULL);
    }
    
    if (!AMI_OVERRIDE) {
	Assert(PointerIsValid(SysCache[cacheId]));
    } else {
	if (!PointerIsValid(SysCache[cacheId])) {
	    SysCache[cacheId] =
#ifndef P95
		InitSysCache((*cacheinfo[cacheId].name)->data, 
#else /* P95 */
		InitSysCache(cacheinfo[cacheId].name,
#endif /* P95 */
			     cacheinfo[cacheId].indname,
			     cacheId,
			     cacheinfo[cacheId].nkeys,
			     cacheinfo[cacheId].key,
			     cacheinfo[cacheId].iScanFunc);
	    if (!PointerIsValid(SysCache[cacheId])) {
		elog(WARN,
		     "InitCatalogCache: Can't init cache %.16s(%d)",
#ifndef P95
		     (*cacheinfo[cacheId].name)->data,
#else /* P95 */
		     cacheinfo[cacheId].name,
#endif /* P95 */
		     cacheId);
	    }
	    
	}
    }
    
    tp = SearchSysCache(SysCache[cacheId], key1, key2, key3, key4);
    if (!HeapTupleIsValid(tp)) {
#ifdef CACHEDEBUG
	elog(DEBUG,
	     "SearchSysCacheTuple: Search %s(%d) %d %d %d %d failed", 
	     (*cacheinfo[cacheId].name)->data,
	     cacheId, key1, key2, key3, key4);
#endif
	return((HeapTuple) NULL);
    }
    return(tp);
}

/*
 * SearchSysCacheStruct--
 *    Fills 's' with the information retrieved by calling SearchSysCache()
 *    with arguments key1...key4.  Retrieves only the portion of the tuple
 *    which is not variable-length.
 *
 * NOTE: we are assuming that non-variable-length fields in the system
 *	 catalogs will always be defined!
 *
 * Returns 1L if a tuple was found, 0L if not.
 */
int32
SearchSysCacheStruct(int cacheId,		/* cache selection code */
		     char *returnStruct, 	/* (preallocated!) */
#ifndef P95
		     char *key1,
		     char *key2,
		     char *key3,
		     char *key4)
#else /* P95 */
		     Datum key1,
		     Datum key2,
		     Datum key3,
		     Datum key4)
#endif /* P95 */
{
    HeapTuple	tp;
    
    if (!PointerIsValid(returnStruct)) {
	elog(WARN, "SearchSysCacheStruct: No receiving struct");
	return(0);
    }
    tp = SearchSysCacheTuple(cacheId, key1, key2, key3, key4);
    if (!HeapTupleIsValid(tp))
	return(0);
    memmove(returnStruct, (char *) GETSTRUCT(tp), cacheinfo[cacheId].size);
    return(1);
}


/*
 * SearchSysCacheGetAttribute--
 *    Returns the attribute corresponding to 'attributeNumber' for
 *    a given cached tuple.
 *
 * XXX This re-opens a relation, so this is slower.
 *
 * [callers all assume this returns a (struct varlena *). -ay 10/94]
 */
void *
SearchSysCacheGetAttribute(int cacheId,
			   AttrNumber attributeNumber,
#ifndef P95
			   char *key1,
			   char *key2,
			   char *key3,
			   char *key4)
#else /* P95 */
			   Datum key1,
			   Datum key2,
			   Datum key3,
			   Datum key4)
#endif /* P95 */
{
    HeapTuple	tp;
#ifndef P95
    Name cacheName;
#else /* P95 */
    char *cacheName;
#endif /* P95 */
    Relation	relation;
    int32	attributeLength, attributeByValue;
    bool	isNull;
    char	*attributeValue;
    void	*returnValue;
    
    tp = SearchSysCacheTuple(cacheId, key1, key2, key3, key4);
#ifndef P95
     cacheName = *cacheinfo[cacheId].name;
#else /* P95 */
     cacheName = cacheinfo[cacheId].name;
#endif /* P95 */
    
    if (!HeapTupleIsValid(tp)) {
#ifdef	CACHEDEBUG
	elog(DEBUG,
	     "SearchSysCacheGetAttribute: Lookup in %s(%d) failed",
#ifndef P95
	     cacheName->data, cacheId);
#else /* P95 */
	     cacheName, cacheId);
#endif /* P95 */
#endif	/* defined(CACHEDEBUG) */
	return(NULL);
    }
    
    relation = heap_openr(cacheName);
    
    if (attributeNumber < 0 &&
	attributeNumber > FirstLowInvalidHeapAttributeNumber) {
	attributeLength = heap_sysattrlen(attributeNumber);
	attributeByValue = heap_sysattrbyval(attributeNumber);
    } else if (attributeNumber > 0 &&
	       attributeNumber <= relation->rd_rel->relnatts) {
	attributeLength =
#ifndef P95
	    relation->rd_att[attributeNumber-1]->attlen;
#else /* P95 */
	    relation->rd_att->attrs[attributeNumber-1]->attlen;
#endif /* P95 */
	attributeByValue =
#ifndef P95
	    relation->rd_att[attributeNumber-1]->attbyval;
#else /* P95 */
	    relation->rd_att->attrs[attributeNumber-1]->attbyval;
#endif /* P95 */
    } else {
	elog(WARN, 
	     "SearchSysCacheGetAttribute: Bad attr # %d in %s(%d)",
#ifndef P95
	     attributeNumber, cacheName->data, cacheId);
#else /* P95 */
	     attributeNumber, cacheName, cacheId);
#endif /* P95 */
	return(NULL);
    }
    
    attributeValue = heap_getattr(tp,
				  (Buffer) 0,
				  attributeNumber,
#ifndef P95
				  relation->rd_att,
#else /* P95 */
				  RelationGetTupleDescriptor(relation),
#endif /* P95 */
				  &isNull);
    
    if (isNull) {
	/*
	 * Used to be an elog(DEBUG, ...) here and a claim that it should
	 * be a FATAL error, I don't think either is warranted -mer 6/9/92
	 */
	return(NULL);
    }
    
    if (attributeByValue) {
	returnValue = (void *)attributeValue;
    } else {
	char	*tmp;
	int size = (attributeLength < 0)
	    ? VARSIZE((struct varlena *) attributeValue) /* variable length */
		: attributeLength;			 /* fixed length */
	
	tmp = (char *) palloc(size);
	memmove(tmp, attributeValue, size);
	returnValue = (void *)tmp;
    }
    
    heap_close(relation);
    return(returnValue);
}

/*
 * TypeDefaultRetrieve--
 *
 *    Given a type OID, return the typdefault field associated with that
 *    type.  The typdefault is returned as the car of a dotted pair which
 *    is passed to TypeDefaultRetrieve by the calling routine.
 *
 * Returns a fixnum for types which are passed by value and a ppreserve'd
 * vectori for types which are not.
 *
 * [identical to get_typdefault, expecting a (struct varlena *) as ret val.
 *  some day, either of the functions should be removed      -ay 10/94]
 */
void *
TypeDefaultRetrieve(Oid typId)
{
    HeapTuple		typeTuple;
    TypeTupleForm	type;
    int32		typByVal, typLen;
    struct varlena	*typDefault;
    int32		dataSize;
    void		*returnValue;
    
    typeTuple = SearchSysCacheTuple(TYPOID,
#ifndef P95
				    (char *) typId,
				    (char *) NULL,
				    (char *) NULL,
				    (char *) NULL);
#else /* P95 */
				    ObjectIdGetDatum(typId),
				    0,0,0);
#endif /* P95 */
    
    if (!HeapTupleIsValid(typeTuple)) {
#ifdef	CACHEDEBUG
	elog(DEBUG, "TypeDefaultRetrieve: Lookup in %s(%d) failed",
	     (*cacheinfo[TYPOID].name)->data, TYPOID);
#endif	/* defined(CACHEDEBUG) */
	return(NULL);
    }
    
    type = (TypeTupleForm) GETSTRUCT(typeTuple);
    typByVal = type->typbyval;
    typLen = type->typlen;
    
    typDefault = (struct varlena *)
	SearchSysCacheGetAttribute(TYPOID, 
				   Anum_pg_type_typdefault,
#ifndef P95
				   (char *) typId,
				   (char *) NULL,
				   (char *) NULL,
				   (char *) NULL);
#else /* P95 */
				   ObjectIdGetDatum(typId),
				   0,0,0);
#endif /* P95 */
    
    if (typDefault == (struct varlena *)NULL) {
#ifdef	CACHEDEBUG
	elog(DEBUG, "TypeDefaultRetrieve: No extractable typdefault",
	     (*cacheinfo[TYPOID].name)->data, TYPOID);
#endif	/* defined(CACHEDEBUG) */
	return (NULL);
	
    }
    
    dataSize = VARSIZE(typDefault) - VARHDRSZ;
    
    if (typByVal) {
	int8 i8;
	int16 i16;
	int32 i32;
	
	if (dataSize == typLen) {
	    switch (typLen) {
	    case sizeof(int8):
		memmove((char *) &i8, VARDATA(typDefault), sizeof(int8));
		i32 = i8;
		break;
	    case sizeof(int16):
		memmove((char *) &i16, VARDATA(typDefault), sizeof(int16));
		i32 = i16;
		break;
	    case sizeof(int32):
		memmove((char *) &i32, VARDATA(typDefault), sizeof(int32));
		break;
	    }
	    returnValue = (void *)i32;
	} else {
	    returnValue = NULL;
	}
    } else {
	if ((typLen < 0 && dataSize < 0) || dataSize != typLen)
	    returnValue = NULL;
	else {
	    returnValue = (void *)palloc(VARSIZE(typDefault));
	    memmove((char *) returnValue,
		    (char *) typDefault,
		    (int) VARSIZE(typDefault));
	}
    }
    
    return(returnValue);
}


