/*-------------------------------------------------------------------------
 *
 * nameserver_mk.c--
 *    name service related RPC calls in the exec./backend
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/local/devel/mariposa/cvs/src/backend/utils/nameserver/nameserver_mk.c,v 1.5 1996/06/14 20:24:11 jsidell Exp $
 *
 *-------------------------------------------------------------------------
 */
#include <stdlib.h>
#include <pthread.h>
#include <rpc/types.h>
#include "utils/elog.h"
#include "include/miscadmin.h" /* for DebugLvl */
#include "mariposa_rpc.h" /* for procedure constants */
#include "rpc/rpcutils.h"

#include "rpc/call_exec.h" /* backend calling facility */

#include "access/heapam.h"
#include "catalog/catname.h"
#include "catalog/pg_class.h"
#include "catalog/pg_nsvcclass.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_nsvcattr.h"
#include "catalog/pg_nsvcfrag.h"
#include "catalog/pg_host.h"
#include "catalog/indexing.h"
#include "utils/catcache.h"
#include "utils/mcxt.h"
#include "utils/nameserver.h"
#include "sitemgr/rpcerror.h"
#include "catalog/pg_fragment.h"
#include "catalog/pg_proc.h"

/* forward declarations */
static HeapTuple scanPgNsvcClass(char *relname);
static HeapTuple scanPgClass(char *relname);
static void findAttributes(NsvcLookupClassResult *result, Oid relOid,
    int nAtts, ScanKeyData *sKey, Relation rd);
static void lookupNsvcClass(NsvcLookupClassResult *result);
static void lookupClass(NsvcLookupClassResult *result);
static void findFrags(TupleArray *result, ScanKeyData *sKey,
    Relation rd);
static void lookupNsvcFrags(NsvcLookupFragsResult *result);
static void lookupFrags(NsvcLookupFragsResult *result);
static int tupCmp(Tuple *t1, Tuple *t2);

/*
 * handleCall_NSVC_LOOKUPCLASS --
 *	thread start routine; does synchronous call to backend to process query
 */
void
handleCall_NSVC_LOOKUPCLASS(void *args)
{
    char **argv = (char **)args;
    char **relname = (char **) argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    NsvcLookupClassResult result;

    if (DebugLvl > 1) {
	elog(NOTICE, "Entering handleCall_NSVC_LOOKUPCLASS(%s).", *relname);
    }

    (void) memset(&result, 0, sizeof(NsvcLookupClassResult));
    result.relname = *relname;
    mainThreadExecInXact((thdproc_t) lookupClass, &result, NULL);
    if (!svc_sendreply(transp, xdr_NsvcLookupClassResult, &result)) {
  	errorExit("handleCall: can't send reply\n");
    }

    free(args);
}


/*
 * handleCall_NSVC_LOOKUPFRAG --
 *	thread start routine; does synchronous call to backend to process query
 */
void
handleCall_NSVC_LOOKUPFRAGS(void *args)
{
    char **argv = (char **)args;
    int *relid = (int *) argv[0];
    SVCXPRT *transp = (SVCXPRT *)argv[1];
    NsvcLookupFragsResult result;

    if (DebugLvl > 1) {
	elog(NOTICE, "Entering handleCall_NSVC_LOOKUPFRAGS(%d).", *relid);
    }

    (void) memset(&result, 0, sizeof(NsvcLookupFragsResult));
    result.relid = *relid;
    mainThreadExecInXact((thdproc_t) lookupFrags, &result, NULL);
    if (!svc_sendreply(transp, xdr_NsvcLookupFragsResult, &result)) {
  	errorExit("handleCall: can't send reply\n");
    }

    free(args);
}

/*
 * scanPgNsvcClass -
 *
 * Scans pg_nsvcclass sequentially to find tuple for given relation.
 */
static HeapTuple
scanPgNsvcClass(char *relname)
{
    Relation rd;
    HeapScanDesc scanDesc;
    ScanKeyData skey;
    HeapTuple tup;

    ScanKeyEntryInitialize(&skey, 0x0, Anum_pg_nsvcclass_relname,
	(RegProcedure) NameEqualRegProcedure, (Datum) relname);
    rd = heap_openr((Name) Name_pg_nsvcclass);
    scanDesc = heap_beginscan(rd, false, NowTimeQual, 1, &skey);
    tup = heap_getnext(scanDesc, 0, NULL);
    heap_endscan(scanDesc);
    heap_close(rd);
    return tup;
}

/*
 * scanPgClass -
 * 
 * Scans pg_class index to find tuple for given relation.
 */
static HeapTuple
scanPgClass(char *relname)
{
    Relation pgClassDesc;
    HeapTuple result;


    pgClassDesc = heap_openr(Name_pg_class);
    if (!IsInitProcessingMode()) {
	RelationSetLockForRead(pgClassDesc);
    }
    result = ClassNameIndexScan(pgClassDesc, relname);
    if (!IsInitProcessingMode()) {
	RelationUnsetLockForRead(pgClassDesc);
    }
    heap_close(pgClassDesc);
    return result;
}

/*
 * tupCmp -
 *
 * Compares the attribute numbers of two attribute tuples and returns
 * +1 if the first one has a higher number than the second one, otherwise 
 * -1. 
 */
static int
tupCmp(Tuple *t1, Tuple *t2)
{
    HeapTuple htup1, htup2;
    AttributeTupleForm attTup1, attTup2;

    htup1 = (HeapTuple) t1->Tuple_val;
    attTup1 = (AttributeTupleForm) GETSTRUCT(htup1);
    htup2 = (HeapTuple) t2->Tuple_val;
    attTup2 = (AttributeTupleForm) GETSTRUCT(htup2);
    Assert(attTup1->attnum != attTup2->attnum); /* can't be */
    if (attTup1->attnum > attTup2->attnum) {
	return 1;
    } else {
	return -1;
    }
}

/* 
 * findAttributes -
 *
 * Allocates memory for attributes and scans the parameter relation.
 * Attributes are returned in attribute number order.
 */
static void
findAttributes(NsvcLookupClassResult *result, Oid relOid, int nAtts,
    ScanKeyData *sKey, Relation rd)
{
    HeapScanDesc sDesc;
    int needed = nAtts;
    HeapTuple tup;
    AttributeTupleForm attTup;
    Tuple *tupArray;

    /* Allocate memory for attributes */
    result->attrTuples.TupleArray_len = nAtts;
    result->attrTuples.TupleArray_val =
        (Tuple *) malloc(nAtts * sizeof(Tuple));
    tupArray = result->attrTuples.TupleArray_val;

    sDesc = heap_beginscan(rd, 0, NowTimeQual, 1, sKey);
    tup = heap_getnext(sDesc, 0, (Buffer *) NULL);
    while (HeapTupleIsValid(tup) && needed > 0) {
	attTup = (AttributeTupleForm) GETSTRUCT(tup);
	if (attTup->attnum > 0) {
	    /* this is not a system attribute; copy it */
	    tupArray[nAtts-needed].Tuple_len = tup->t_len;
	    tupArray[nAtts-needed].Tuple_val = malloc(tup->t_len);
	    (void) memmove(tupArray[nAtts-needed].Tuple_val, tup, tup->t_len);
	    needed--;
	}
	tup = heap_getnext(sDesc, 0, (Buffer *) NULL);
    }
    if (needed > 0 && DebugLvl > 1) {
	elog(WARN, "catalog is missing %d attributes for relid %d",
	    needed, relOid);
    }

    /* Sort attributes according to attribute number. */
    qsort(tupArray, nAtts, sizeof(Tuple), tupCmp);

    heap_endscan(sDesc);
    heap_close(rd);
}

/*
 * lookupNsvcClass -
 *
 * Looks up the class and its attributes in the nameserver catalogs.
 */
static void
lookupNsvcClass(NsvcLookupClassResult *result)
{
    HeapTuple tup;
    int nAtts;
    Form_pg_class classTuple;
    Oid classOid;
    ScanKeyData sKey;
    Relation rd;

    tup = scanPgNsvcClass(result->relname);
    if (!HeapTupleIsValid(tup)) {
	result->error = EXECUTION_ERROR;
	if (DebugLvl > 1) {
	    elog(NOTICE, "lookupNsvcClass: Can't find class %s.",
		 result->relname);
	}
	return;
    }
    /* this is a nameserver class */
    result->classTuple.Tuple_len = tup->t_len;
    result->classTuple.Tuple_val = (char *) malloc(tup->t_len);
    (void) memmove(result->classTuple.Tuple_val, tup, tup->t_len);

    /* find attributes  */
    classOid = tup->t_oid;
    classTuple = (Form_pg_class) GETSTRUCT(tup);
    nAtts = classTuple->relnatts;
    /* prepare scan */
    ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_nsvcattr_attrelid,
	ObjectIdEqualRegProcedure, ObjectIdGetDatum(classOid));
    rd = heap_openr(Name_pg_nsvcattr);
    findAttributes(result, classOid, nAtts, &sKey, rd);
    result->error = SUCCESS;
}

/*
 * lookupClass -
 *
 * Looks up the class and its attributes in the local and nameserver
 * catalogs (in that order).
 */
static void
lookupClass(NsvcLookupClassResult *result)
{
    HeapTuple tup;
    int nAtts;
    Form_pg_class classTuple;
    Oid classOid;
    ScanKeyData sKey;
    Relation rd;

    result->error = SUCCESS;

    tup = scanPgClass(result->relname);
    if (HeapTupleIsValid(tup)) {
	/* this is one of our local classes */
	result->classTuple.Tuple_len = tup->t_len;
	result->classTuple.Tuple_val = (char *) malloc(tup->t_len);
	(void) memmove(result->classTuple.Tuple_val, tup, tup->t_len);

	/* find attributes  */
	classOid = tup->t_oid;
	classTuple = (Form_pg_class) GETSTRUCT(tup);
	nAtts = classTuple->relnatts;
	/* prepare scan */
	ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_attribute_attrelid,
	    ObjectIdEqualRegProcedure, ObjectIdGetDatum(classOid));
	rd = heap_openr(Name_pg_attribute);
	findAttributes(result, classOid, nAtts, &sKey, rd);

	result->error = SUCCESS;
	nsvc_printLookupClassResult(result);
	return;
    } 

    lookupNsvcClass(result);
    nsvc_printLookupClassResult(result);
}

/*
 * nsvc_lookupFrags -
 *
 * interface procedure to name server for backend;
 * retrieves fragment information from this site's nameserver
 */
NsvcLookupFragsResult *
nsvc_lookupFrags(char *relname, Oid relid)
{
    NsvcLookupFragsResult *result = (NsvcLookupFragsResult *) malloc(sizeof(NsvcLookupFragsResult));
    RpcConn *rpcConn;

    (void) memset(result, 0, sizeof(NsvcLookupFragsResult));
    if (issystem(relname)) {
	/* system tables are purely local */
	return result;
    }

    if (getMyNameServerHostId() == getMyHostId()) {
	/*
	 * we are our own nameserver; query the local
	 * pg_fragment and pg_nsvcfrag tables.
	 */
	result->relid = relid;
	lookupFrags(result);

    } else {
	rpcConn = Rpc_connectSitemgr(getMyNameServerHostId(), 0,
	    "nsvc_lookupFrags", RPC_ONERR_WARN);
	Rpc_call(rpcConn, RPC_NSVC_SITEMGRLOOKUPFRAGS,
	    &relid, result);
	Rpc_disconnect(rpcConn);
    }

    return result;
}

/*
 * nsvc_lookupClass -
 *
 * interface procedure to name server for backend;
 * retrieves class information from this site's nameserver
 */
NsvcLookupClassResult *
nsvc_lookupClass(char *relname)
{
    NsvcLookupClassResult *result = (NsvcLookupClassResult *)malloc(sizeof(NsvcLookupClassResult));

    result->classTuple.Tuple_val = NULL;
    result->attrTuples.TupleArray_val = NULL;
    result->relname = NULL;
    
    if (getMyNameServerHostId() == getMyHostId()) {
	/*
	 * we are our own nameserver; use the local 
	 * nameserver catalogs
	 */
	result->relname = relname;
	lookupNsvcClass(result);

    } else {
	RpcConn *rpcConn = Rpc_connectSitemgr(getMyNameServerHostId(), 0,
	    "nsvc_lookupClass", RPC_ONERR_WARN);
	Rpc_call(rpcConn, RPC_NSVC_SITEMGRLOOKUPCLASS,
	    &relname, result);
	Rpc_disconnect(rpcConn);
    }

    return result;
}

void
nsvc_freeLookupClassResult(NsvcLookupClassResult *result)
{
    xdr_free(xdr_NsvcLookupClassResult, result);
    free(result);
}

void
nsvc_printLookupClassResult(NsvcLookupClassResult *result)
{
    int i;
    HeapTuple tup;
    Form_pg_class class;
    AttributeTupleForm attr;

    fprintf(stderr,"(\"%s\" ", result->relname);

    if (result->error == SUCCESS) {
	tup = (HeapTuple) result->classTuple.Tuple_val;
	class = (Form_pg_class) GETSTRUCT(tup);
	fprintf(stderr, "(%d, \"%.16s\") ", tup->t_oid,
	    class->relname.data);

	fprintf(stderr, "(");
	for (i = 0; i < result->attrTuples.TupleArray_len; i++) {
	    tup = (HeapTuple) result->attrTuples.TupleArray_val[i].Tuple_val;
	    attr = (AttributeTupleForm) GETSTRUCT(tup);
	    fprintf(stderr, "%d \"%.16s\", ", tup->t_oid, attr->attname.data);
	}
	fprintf(stderr, ") ");
    }

    fprintf(stderr, "%d)\n", result->error);
}

void
nsvc_printLookupFragsResult(NsvcLookupFragsResult *result)
{
    int i;
    HeapTuple tup;
    Form_pg_fragment frag;

    fprintf(stderr, "(%d ", result->relid);
    if (result->error == SUCCESS) {
	fprintf(stderr, "(");
	for (i = 0; i < result->fragTuples.TupleArray_len; i++) {
	    tup = (HeapTuple) result->fragTuples.TupleArray_val[i].Tuple_val;
	    frag = (Form_pg_fragment) GETSTRUCT(tup);
	    fprintf(stderr, "(%d %d %d %d)", frag->fraclassid, frag->frastoreid,
	       frag->fralogicalid, frag->fralocation);
	}
	fprintf(stderr, ") ");
    }
    fprintf(stderr, "%d)\n", result->error);
}

void
nsvc_freeLookupFragsResult(NsvcLookupFragsResult *result)
{
    xdr_free(xdr_NsvcLookupFragsResult, result);
    free(result);
}

/* 
 * findFrags -
 *
 * Allocates memory for fragments and scans the parameter relation.
 */
static void
findFrags(TupleArray *result, ScanKeyData *sKey, Relation rd)
{
    HeapScanDesc sDesc;
    List *fragList = NIL;
    Tuple *listElem;
    HeapTuple tup;
    Form_pg_fragment frag;
    int count = 0;
    List *e = NIL;
    int i = 0;

    sDesc = heap_beginscan(rd, 0, NowTimeQual, 1, sKey);
    tup = heap_getnext(sDesc, 0, (Buffer  *) NULL);
    while (HeapTupleIsValid(tup)) {
	frag = (Form_pg_fragment) GETSTRUCT(tup);
	listElem = (Tuple *) malloc(sizeof(Tuple));
	Assert(listElem != NULL);
	listElem->Tuple_len = tup->t_len;
	listElem->Tuple_val = malloc(tup->t_len);
	Assert(listElem->Tuple_val != NULL);
	(void) memmove(listElem->Tuple_val, tup, tup->t_len);
	fragList = lcons(listElem, fragList);
	tup = heap_getnext(sDesc, 0, (Buffer *) NULL);
	count++;
    }
    heap_endscan(sDesc);
    heap_close(rd);

    result->TupleArray_len = count;
    
    if (count > 0) {
	/* move tuples to result structure */
	result->TupleArray_val =
	    (Tuple *) malloc(count * sizeof(Tuple));
	foreach(e, fragList) {
	    listElem = (Tuple *) e->elem;
	    result->TupleArray_val[i].Tuple_len = listElem->Tuple_len;
	    result->TupleArray_val[i].Tuple_val = listElem->Tuple_val;
	    free(listElem);
	    i++;
	}
    }
    freeList(fragList);
}

/*
 * lookupNsvcFrags -
 *
 * Looks up the fragment tuples in the nameserver catalogs.
 */
static void
lookupNsvcFrags(NsvcLookupFragsResult *result)
{
    Relation rd;
    ScanKeyData sKey;

    /* now the nameserver catalogs */
    rd = heap_openr(Name_pg_nsvcfrag);
    ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_nsvcfrag_fraclassid,
	ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
    findFrags(&result->fragTuples, &sKey, rd);
    if (result->fragTuples.TupleArray_len == 0) {
	/*
	 * There aren't any tuples with matching
	 * fraclassid in pg_nsvcfrag. Check for matching fralogicalid.
	 */
	rd = heap_openr(Name_pg_nsvcfrag);
	ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_nsvcfrag_fralogicalid,
	    ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
	findFrags(&result->fragTuples, &sKey, rd);
    }
    if (result->fragTuples.TupleArray_len != 0) {
	result->error = SUCCESS;
    } else {
	result->error = EXECUTION_ERROR;
    }
}

/*
 * lookupFrags -
 *
 * Looks up the fragment tuples in the local and nameserver
 * catalogs, and returns the combined results.
 */
static void
lookupFrags(NsvcLookupFragsResult *result)
{
    Relation rd;
    ScanKeyData sKey;
    TupleArray localFrags, remoteFrags;
    bool directFragScan = FALSE;	/* Set to true if we're doing a scan directly over a storage fragment. */
    
    /*
     * Query the local catalogs
     */
    rd = heap_openr(Name_pg_fragment);
    ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_fragment_fraclassid,
	ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
    
    findFrags(&localFrags, &sKey, rd);
    
    if (localFrags.TupleArray_len == 0) {
	/*
	 * There aren't any tuples with matching
	 * fraclassid in pg_fragment. If there's one with
	 * a matching fralogicalid, we're doing a lookup for
	 * a fragment (not a relation).
	 */
	rd = heap_openr(Name_pg_fragment);
	ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_fragment_fralogicalid,
	    ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
	findFrags(&localFrags, &sKey, rd);
	
	if (result->fragTuples.TupleArray_len != 0) {
	    directFragScan = TRUE;
	}
    }

    /*
     * Now query the nameserver catalogs
     */
    
    rd = heap_openr(Name_pg_nsvcfrag);
    if (!directFragScan) {
	ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_nsvcfrag_fraclassid,
			       ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
	findFrags(&remoteFrags, &sKey, rd);
    }
    
    if (directFragScan ||
	(remoteFrags.TupleArray_len == 0 && localFrags.TupleArray_len == 0)) {
	rd = heap_openr(Name_pg_nsvcfrag);
	ScanKeyEntryInitialize(&sKey, 0x0, Anum_pg_nsvcfrag_fralogicalid,
	    ObjectIdEqualRegProcedure, ObjectIdGetDatum(result->relid));
	findFrags(&remoteFrags, &sKey, rd);
    }

    /*
     * Combine local and remote fragment results
     */

    if (localFrags.TupleArray_len == 0 &&
	remoteFrags.TupleArray_len == 0) {
	result->error = EXECUTION_ERROR;
    } else {

	int i;
	
	result->error = SUCCESS;

	result->fragTuples.TupleArray_len =
	    localFrags.TupleArray_len + remoteFrags.TupleArray_len;
	
	result->fragTuples.TupleArray_val =
	    (Tuple *) malloc(result->fragTuples.TupleArray_len * sizeof(Tuple));

	/*
	 * Local fragment tuples first...
	 */
	for (i = 0; i < localFrags.TupleArray_len; i++) {
	    result->fragTuples.TupleArray_val[i].Tuple_len = localFrags.TupleArray_val[i].Tuple_len;
	    result->fragTuples.TupleArray_val[i].Tuple_val = localFrags.TupleArray_val[i].Tuple_val;
	}

	/*
	 * Followed by remote fragment tuples
	 */
	for (i = 0; i < remoteFrags.TupleArray_len; i++) {
	    result->fragTuples.TupleArray_val[i + localFrags.TupleArray_len].Tuple_len =
		remoteFrags.TupleArray_val[i].Tuple_len;
	    result->fragTuples.TupleArray_val[i + localFrags.TupleArray_len].Tuple_val =
		remoteFrags.TupleArray_val[i].Tuple_val;
	}
    }


    /*
     * Free remoteFrags and localFrags tuple arrays
     */

    if (localFrags.TupleArray_len > 0) {
	free(localFrags.TupleArray_val);
    }
    
    if (remoteFrags.TupleArray_len > 0) {
	free(remoteFrags.TupleArray_val);
    }
}
